!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Bin Data [external members: `_bin`]
!
! DESCRIPTION:
!> Module defining the data structure within the histogram bins. 
!> We use an integer to declare different types of data structures:
!>   itype=0: no channel breakdown: only sclaes
!>   itype=1: channel breakdown + scales
!>   itype=2: legacy compatibility mode
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module BinData_mod
!-------------------------------------------------------------------------------
  use Scales_mod
  implicit none
  private

  public :: ndata_bin
  public :: accumulate_bin
  public :: getLabel_bin, getIndex_bin

  !> charcter constants to construct the labels
  integer, parameter :: max_label_length = 15
  character(len=*), parameter :: str_full = "tot"
  character(len=*), dimension(-3:3), parameter :: &
     & str_chan = (/ "Fb", "fb", "qb", "g ", "q ", "f ", "F " /)
  character(len=*), dimension(1:6), parameter :: &
     & str_f  = (/ "d ", "u ", "s ", "c ", "b ", "t " /)
  character(len=*), dimension(1:6), parameter :: &
     & str_fb = (/ "db", "ub", "sb", "cb", "bb", "tb" /)
  
  !> @interface
  !> @brief
  !> this is the external interface accumulate_bin
  interface accumulate_bin
    module procedure accumulate_raw
    module procedure accumulate_rawErr
  end interface


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> get number of entries in the BinData data structure for a given type
  !
  !> @param[in] itype BinData type
  !> @return number of data entries for `itype`
  !-----------------------------------------------------------------------------
  function ndata_bin(itype)
    integer :: ndata_bin
    integer, intent(in) :: itype

    select case(itype)
    case(0)  ! no channel breakdown: only scales
      ndata_bin = nInp_scl
    
    case(1)  ! channel breakdown + scales
      ndata_bin = nInp_scl*(1+3*3)

    case(2)  ! channel breakdown + scales (nMax_scl)
      ndata_bin = nMax_scl*(1+3*3)

    case(-1,-2,-3,-4,-5,-6)  ! channel breakdown + scales + 1 flavour
      ndata_bin = nInp_scl*(1+5*5)

    case(-66:-11)  ! channel breakdown + scales + 2 flavours
      ndata_bin = nInp_scl*(1+7*7)
      
    case default
      print*, "ndata_bin: unknown BinData type:", itype
      stop
    end select

  end function ndata_bin


  !-----------------------------------------------------------------------------
  !> @brief
  !> accumulate BinData_t
  !
  !> @param[in] itype BinData type
  !> @param[inout] accum the accumulator
  !> @param[in] vals array of length nMax_scl
  !> @param[in] ipart1 first parton identifyer
  !> @param[in] ipart2 secdond parton identifyer
  !-----------------------------------------------------------------------------
  subroutine accumulate_raw(itype, accum, vals,ipart1,ipart2)
    integer, intent(in) :: itype
    double precision, dimension(:), intent(inout) :: accum
    double precision, dimension(nMax_scl), intent(in) :: vals
    integer, intent(in) :: ipart1, ipart2
    integer, dimension(2) :: ipart
    integer :: i, ioffset
    integer :: itype1,itype2

    !> always accumulate the "tot" part
    do i=1,nInp_scl
      accum(i) = accum(i)+vals(i)
    end do

    select case(itype)
    case(0)  ! no channel breakdown: only scales
      !> nothing left to do...
    
    case(1)  ! channel breakdown + scales
      ipart = (/ ipart1, ipart2 /)
      where (ipart /= 0) ipart = sign(1,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+1)*nInp_scl  &
        &     + (ipart(2)+1)*nInp_scl*3 
      do i=1,nInp_scl
        accum(ioffset+i) = accum(ioffset+i)+vals(i)
      end do

    case(2)  ! channel breakdown + scales (nMax_scl)
      ipart = (/ ipart1, ipart2 /)
      where (ipart /= 0) ipart = sign(1,ipart)
      ioffset = nMax_scl  &
        &     + (ipart(1)+1)*nMax_scl  &
        &     + (ipart(2)+1)*nMax_scl*3 
      do i=1,nMax_scl
        accum(ioffset+i) = accum(ioffset+i)+vals(i)
      end do

    case(-1,-2,-3,-4,-5,-6)  ! channel breakdown + scales + 1 flavour
      ipart = (/ ipart1, ipart2 /)
      where ((ipart /= 0) .and. (abs(ipart) /= abs(itype))) ipart = sign(1,ipart)
      where (abs(ipart) == abs(itype))                      ipart = sign(2,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+2)*nInp_scl  &
        &     + (ipart(2)+2)*nInp_scl*5
      do i=1,nInp_scl
        accum(ioffset+i) = accum(ioffset+i)+vals(i)
      end do

    case(-66:-11)  ! channel breakdown + scales + 2 flavours
      ipart = (/ ipart1, ipart2 /)
      !> deduce the two flavours from itype
      itype2 = abs( mod(itype,10) )
      itype1 = abs( (itype+itype2) / 10 )
      where ((ipart /= 0) .and. (abs(ipart) /= abs(itype1)) .and. (abs(ipart) /= abs(itype2))) ipart = sign(1,ipart)
      where (abs(ipart) == abs(itype1))                                                        ipart = sign(2,ipart) 
      where (abs(ipart) == abs(itype2))                                                        ipart = sign(3,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+3)*nInp_scl  &
        &     + (ipart(2)+3)*nInp_scl*7
      do i=1,nInp_scl
        accum(ioffset+i) = accum(ioffset+i)+vals(i)
      end do
      
    case default
      print*, "accumulate_raw: unknown BinData type:", itype
      stop
    end select

  end subroutine accumulate_raw


  !-----------------------------------------------------------------------------
  !> @brief
  !> accumulate BinData_t
  !
  !> @param[in] itype BinData type
  !> @param[inout] accum the accumulator (0: error, 1: sum)
  !> @param[in] vals array of length nMax_scl
  !> @param[in] ipart1 first parton identifyer
  !> @param[in] ipart2 secdond parton identifyer
  !-----------------------------------------------------------------------------
  subroutine accumulate_rawErr(itype, accum, vals,ipart1,ipart2)
    use FPA_mod
    integer, intent(in) :: itype
    double precision, dimension(0:,:), intent(inout) :: accum
    double precision, dimension(nMax_scl), intent(in) :: vals
    integer, intent(in) :: ipart1, ipart2
    double precision :: acc_sum, err
    integer, dimension(2) :: ipart
    integer :: i, ioffset
    integer :: itype1,itype2

    !> always accumulate the "tot" part
    do i=1,nInp_scl
      call twoSum(accum(1,i),vals(i), acc_sum,err)
      accum(1,i) = acc_sum
      accum(0,i) = accum(0,i) + err
    end do

    select case(itype)
    case(0)  ! no channel breakdown: only scales
      !> nothing left to do...
    
    case(1)  ! channel breakdown + scales
      ipart = (/ ipart1, ipart2 /)
      where (ipart /= 0) ipart = sign(1,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+1)*nInp_scl  &
        &     + (ipart(2)+1)*nInp_scl*3 
      do i=1,nInp_scl
        call twoSum(accum(1,ioffset+i),vals(i), acc_sum,err)
        accum(1,ioffset+i) = acc_sum
        accum(0,ioffset+i) = accum(0,ioffset+i) + err
      end do

    case(2)  ! channel breakdown + scales (nMax_scl)
      ipart = (/ ipart1, ipart2 /)
      where (ipart /= 0) ipart = sign(1,ipart)
      ioffset = nMax_scl  &
        &     + (ipart(1)+1)*nMax_scl  &
        &     + (ipart(2)+1)*nMax_scl*3 
      do i=1,nMax_scl
        call twoSum(accum(1,ioffset+i),vals(i), acc_sum,err)
        accum(1,ioffset+i) = acc_sum
        accum(0,ioffset+i) = accum(0,ioffset+i) + err
      end do

    case(-1,-2,-3,-4,-5,-6)  ! channel breakdown + scales + 1 flavour
      ipart = (/ ipart1, ipart2 /)
      where ((ipart /= 0) .and. (abs(ipart) /= abs(itype))) ipart = sign(1,ipart)
      where (abs(ipart) == abs(itype))                      ipart = sign(2,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+2)*nInp_scl  &
        &     + (ipart(2)+2)*nInp_scl*5
      do i=1,nInp_scl
        call twoSum(accum(1,ioffset+i),vals(i), acc_sum,err)
        accum(1,ioffset+i) = acc_sum
        accum(0,ioffset+i) = accum(0,ioffset+i) + err
      end do

    case(-66:-11)  ! channel breakdown + scales + 2 flavours
      ipart = (/ ipart1, ipart2 /)
      !> deduce the two flavours from itype
      itype2 = abs( mod(itype,10) )
      itype1 = abs( (itype+itype2) / 10 )
      where ((ipart /= 0) .and. (abs(ipart) /= abs(itype1)) .and. (abs(ipart) /= abs(itype2))) ipart = sign(1,ipart)
      where (abs(ipart) == abs(itype1))                                                        ipart = sign(2,ipart) 
      where (abs(ipart) == abs(itype2))                                                        ipart = sign(3,ipart)
      ioffset = nInp_scl  &
        &     + (ipart(1)+3)*nInp_scl  &
        &     + (ipart(2)+3)*nInp_scl*7
      do i=1,nInp_scl
        call twoSum(accum(1,ioffset+i),vals(i), acc_sum,err)
        accum(1,ioffset+i) = acc_sum
        accum(0,ioffset+i) = accum(0,ioffset+i) + err
      end do
      
    case default
      print*, "accumulate_raw: unknown BinData type:", itype
      stop
    end select

  end subroutine accumulate_rawErr


  !-----------------------------------------------------------------------------
  !> @brief
  !> Function returning a "meaningful" label for the `index`-th BinData entry.
  !> The order here defines the order of the columns written out in the
  !> histogram data files and must match the ordering in `accumulate_bin`
  !
  !> @param[in] itype BinData type
  !> @param[in] index numbering [1:ndata_bin] of the sub-data
  !> @return string containing the label for the BinData entry
  !-----------------------------------------------------------------------------
  function getLabel_bin(itype, index)
    character(len=max_label_length) :: getLabel_bin
    character(len=max_label_length) :: str_scale
    integer, intent(in) :: itype
    integer, intent(in) :: index
    integer :: iscale, ichan, ipart1, ipart2
    character(len=2) :: str1, str2
    integer :: itype1, itype2
    
    select case(itype)
    case(0)  ! no channel breakdown: only scales
      iscale = index
      ichan  = 0
      ipart1 = 0
      ipart2 = 0

    case(1)  ! channel breakdown + scales
      iscale = mod(index-1,nInp_scl) +1
      ichan  = (index-1)/nInp_scl
      ipart1 = mod(ichan-1,3) -1
      ipart2 = (ichan-1)/3 -1

    case(2)  ! channel breakdown + scales (nMax_scl)
      iscale = mod(index-1,nMax_scl) +1
      ichan  = (index-1)/nMax_scl
      ipart1 = mod(ichan-1,3) -1
      ipart2 = (ichan-1)/3 -1

    case(-1,-2,-3,-4,-5,-6)  ! channel breakdown + scales + 1 flavour
      iscale = mod(index-1,nInp_scl) +1
      ichan  = (index-1)/nInp_scl
      ipart1 = mod(ichan-1,5) -2
      ipart2 = (ichan-1)/5 -2
      
    case(-66:-11)  ! channel breakdown + scales + 2 flavours
      iscale = mod(index-1,nInp_scl) +1
      ichan  = (index-1)/nInp_scl
      ipart1 = mod(ichan-1,7) -3
      ipart2 = (ichan-1)/7 -3
      
    case default
      print*, "getLabel_bin: unknown BinData type:", itype
      stop
    end select

    write(str_scale,'(I2.2)') iscale

    getLabel_bin = ""
    select case (ichan)
      case (0)
        getLabel_bin = trim(str_full)
      case (1:)
        str1 = str_chan(ipart1)
        str2 = str_chan(ipart2)
        if (itype < 0 .and. itype > -7) then
          if (ipart1 == -2) str1 = str_fb(itype)
          if (ipart1 == +2) str1 = str_f(itype)
          if (ipart2 == -2) str2 = str_fb(itype)
          if (ipart2 == +2) str2 = str_f(itype)
        else if (itype < -10 .and. itype > -67) then
          !> deduce the two flavours from itype ...
          itype2 = abs( mod(itype,10) )
          itype1 = abs( (itype+itype2) / 10 )
          !> ...and change labels
          if (ipart1 == -3) str1 = str_fb(itype2)
          if (ipart1 == -2) str1 = str_fb(itype1)
          if (ipart1 == +2) str1 = str_f(itype1)
          if (ipart1 == +3) str1 = str_f(itype2)
          if (ipart2 == -3) str2 = str_fb(itype2)
          if (ipart2 == -2) str2 = str_fb(itype1)
          if (ipart2 == +2) str2 = str_f(itype1)
          if (ipart2 == +3) str2 = str_f(itype2)           
        end if
        getLabel_bin = trim(str1)//trim(str2)
    end select

    getLabel_bin = trim(getLabel_bin)//"_scale"//trim(str_scale)
  end function getLabel_bin


  !-----------------------------------------------------------------------------
  !> @brief
  !> Function returning the BinData entry. If ipart[1|2] are not passed: "tot"
  !
  !> @param[in] itype BinData type
  !> @param[in] iscale i-th scale entry
  !> @param[in] ipart1 optional identifyer for parton 1
  !> @param[in] ipart2 optional identifyer for parton 2
  !> @return value of the BinData entry
  !-----------------------------------------------------------------------------
  function getIndex_bin(itype, iscale, ipart1, ipart2)
    integer :: getIndex_bin
    integer, intent(in) :: itype
    integer, intent(in) :: iscale
    integer, intent(in), optional :: ipart1, ipart2
    integer, dimension(2) :: ipart
    integer :: ioffset
    integer :: itype1,itype2    

    getIndex_bin = -1 ! init to an invalid default value

    if (iscale>nMax_scl .or. iscale<1) return

    if ( present(ipart1) .and. present(ipart2) ) then

      select case(itype)      
      case(1)  ! channel breakdown + scales
        ipart = (/ ipart1, ipart2 /)
        where (ipart /= 0) ipart = sign(1,ipart)
        getIndex_bin = nInp_scl  &
          &          + (ipart(1)+1)*nInp_scl  &
          &          + (ipart(2)+1)*nInp_scl*3  &
          &          + iscale
      case(2)  ! channel breakdown + scales (nMax_scl)
        ipart = (/ ipart1, ipart2 /)
        where (ipart /= 0) ipart = sign(1,ipart)
        getIndex_bin = nMax_scl  &
          &          + (ipart(1)+1)*nMax_scl  &
          &          + (ipart(2)+1)*nMax_scl*3  &
          &          + iscale
      case(-1,-2,-3,-4,-5,-6)  ! channel breakdown + scales + 1 flavour
        ipart = (/ ipart1, ipart2 /)
        where ((ipart /= 0) .and. (abs(ipart) /= abs(itype))) ipart = sign(1,ipart)
        where (abs(ipart) == abs(itype))                      ipart = sign(2,ipart)
        getIndex_bin = nInp_scl  &
          &          + (ipart(1)+2)*nInp_scl  &
          &          + (ipart(2)+2)*nInp_scl*5  &
          &          + iscale
      case(-66:-11)  ! channel breakdown + scales + 2 flavours
        ipart = (/ ipart1, ipart2 /)
        !> deduce the two flavours from itype
        itype2 = abs( mod(itype,10) )
        itype1 = abs( (itype+itype2) / 10 )
        where ((ipart /= 0) .and. (abs(ipart) /= abs(itype1)) .and. (abs(ipart) /= abs(itype2))) ipart = sign(1,ipart)
        where (abs(ipart) == abs(itype1))                                                        ipart = sign(2,ipart) 
        where (abs(ipart) == abs(itype2))                                                        ipart = sign(3,ipart)
        getIndex_bin = nInp_scl  &
          &          + (ipart(1)+3)*nInp_scl  &
          &          + (ipart(2)+3)*nInp_scl*7  &
          &          + iscale
      end select

    else 
      !> do not allow only one of them to be set
      if ( present(ipart1) .or. present(ipart2) ) return
      !> "tot" output
      getIndex_bin = iscale
    end if

  end function getIndex_bin


end module BinData_mod
