!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: CKM
!
! DESCRIPTION:
!> Contains CKM information for charged current processes
!> To add CKM info for a CC process, set CKMflag = true in makeproc and
!> regenerate the process.
!> CKM information is only included for a channel if ckmflag = .true. (i.e.
!> it is an annihiliation crossing) AND the initial parton sets are in the
!> following:
!> [20,-40] [-40,20] [40,-20] [-20,40]
!> aka
!> [u , db] [db ,u ] [d , ub] [ub , d]
!> initiated channels
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module CKM_mod
!-------------------------------------------------------------------------------
  implicit none
  public :: CKM,ckmflag
  public :: fillckm
  double precision, dimension(1:5,1:5) :: CKM
  logical :: ckmflag
!$OMP THREADPRIVATE(ckmflag)

contains

  subroutine fillckm(option)
    integer, intent(in) :: option 
    double precision :: lambda, rho, eta, A
    double precision :: s12, c12

    !> This is the default
    ckmflag = .FALSE.
    !> DIAGONAL CKM
    CKM = 0d0
    CKM(2,1) = 1d0 ! Vud
    CKM(2,3) = 0d0 ! Vus
    CKM(2,5) = 0d0 ! Vub
    CKM(4,1) = 0d0 ! Vcd
    CKM(4,3) = 1d0 ! Vcs
    CKM(4,5) = 0d0 ! Vcb

    !> Modify entries if necessary
    select case (option)
    case (1)
       ckmflag = .TRUE.
       !> WOLFENSTEIN CKM
       !> Wolfenstein parameters [2020 PDG, EQN 12.26]
       lambda = 0.22650d0
       A      = 0.790d0
       rho    = 0.141d0
       eta    = 0.357d0
       CKM(2,1) = 1d0-lambda**2/2d0                ! Vud
       CKM(2,3) = lambda                           ! Vus
       CKM(2,5) = A*lambda**3*DSQRT(rho**2+eta**2) ! Vub
       CKM(4,1) = -lambda                          ! Vcd
       CKM(4,3) = 1d0-lambda**2/2d0                ! Vcs
       CKM(4,5) = A*lambda**2                      ! Vcb
    case (2)
       !> Just Cabibbo mixing angle [sintheta_12, PDG 2020 eq. 12.28]
       ckmflag = .TRUE.
       s12 = 0.2265d0
       c12 = dsqrt(1d0 - s12**2)
       CKM(2,1) = c12  ! Vud
       CKM(2,3) = s12  ! Vus
       CKM(4,1) = -s12 ! Vcd
       CKM(4,3) = c12  ! Vcs
    case (3)
       !> in off-diagonal, only Vcs non-zero
       ckmflag = .TRUE.
       CKM(4,3) = 0.97359d0 ! Vcs
    case (4)
       !> in off-diagonal, only Vcd & Vcs non-zero 
       ckmflag = .TRUE.
       CKM(4,1) = 0.22438d0 ! Vcd
       CKM(4,3) = 0.97359d0 ! Vcs              
    case (5)
       !> benchmark setup from 2104.02400
       ckmflag = .TRUE.
       CKM(2,1) = 0.97401D0  ! Vud
       CKM(2,3) = 0.2265D0   ! Vus
       CKM(2,5) = 0.00361D0  ! Vub
       CKM(4,1) = 0.2265D0   ! Vcd
       CKM(4,3) = 0.97320D0  ! Vcs
       CKM(4,5) = 0.04053D0  ! Vcb
    case (6)
       !> MCFM default
       ckmflag = .TRUE.
       CKM(2,1) = 0.975d0 ! Vud
       CKM(2,3) = 0.222d0 ! Vus
       CKM(2,5) = 0d0     ! Vub
       CKM(4,1) = 0.222d0 ! Vcd
       CKM(4,3) = 0.975d0 ! Vcs
       CKM(4,5) = 0d0     ! Vcb
    end select
    
       
!> #############################################
!> Different parameterisations


!> MCFM CKM
    ! CKM(2,1) = 0.975d0 ! Vud
    ! CKM(2,3) = 0.222d0 ! Vus
    ! CKM(2,5) = 0d0     ! Vub
    ! CKM(4,1) = 0.222d0 ! Vcd
    ! CKM(4,3) = 0.975d0 ! Vcs
    ! CKM(4,5) = 0d0     ! Vcb

!> WOLFENSTEIN CKM
!> Wolfenstein parameters [2016 PDG, EQN 12.26]
    ! lambda = 0.22506d0
    ! A      = 0.811d0
    ! rho    = 0.124d0
    ! eta    = 0.356d0

!> Wolfenstein parameters [2018 PDG, EQN 12.26]
    ! lambda = 0.22453d0
    ! A      = 0.836d0
    ! rho    = 0.122d0
    ! eta    = 0.355d0

!> Wolfenstein parameters [SHERPA defaults]
    ! lambda = 0.2272d0
    ! A      = 0.818d0
    ! rho    = 0.227d0
    ! eta    = 0.349d0

    ! CKM(2,1) = 1d0-lambda**2/2d0                ! Vud
    ! CKM(2,3) = lambda                           ! Vus
    ! CKM(2,5) = A*lambda**3*DSQRT(rho**2+eta**2) ! Vub
    ! CKM(4,1) = -lambda                          ! Vcd
    ! CKM(4,3) = 1d0-lambda**2/2d0                ! Vcs
    ! CKM(4,5) = A*lambda**2                      ! Vcb

!> #############################################


    ! Fill remaining elements
    CKM(1,2) = CKM(2,1)
    CKM(3,2) = CKM(2,3)
    CKM(5,2) = CKM(2,5)
    CKM(1,4) = CKM(4,1)
    CKM(3,4) = CKM(4,3)
    CKM(5,4) = CKM(4,5)

    CKM = CKM**2

  end subroutine fillckm

end module CKM_mod
