!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Channels [external members: `_chan`]
!
! DESCRIPTION:
!> Module managing all channel settings
!> @todo: if we register iprocRR1a & iprocRR2b, both
!>        iprocRR1 & iprocRR2 are integrated over *both* regions!
!>        don't care much about this exceptional case for the moment
!-------------------------------------------------------------------------------


!> preprocessor variable to switch on/off additional accumulators
! #define CHECKACCUM

!-------------------------------------------------------------------------------
module Channels_mod
!-------------------------------------------------------------------------------
#ifdef CHECKACCUM
  use, intrinsic :: iso_fortran_env, dp=>real64, qp=>real128
#endif
  implicit none
  private

  !> global stuff
  public :: init_chan, get_chan, destroy_chan, digest_chan
  public :: digestLumi_chan
  !> multi-channel
  public :: setActive_multi_chan, isActive_multi_chan
  public :: init_multi_chan
  public :: sample_multi_chan, jac_multi_chan
  public :: initBuffer_multi_chan, destroyBuffer_multi_chan
  public :: writeBuffer_multi_chan, flushBuffer_multi_chan
  public :: writeToFile_multi_chan, readFromFile_multi_chan
  public :: adapt_multi_chan, resetAccum_multi_chan
  public :: digest_multi_chan
  public :: ilower_multi_chan, iupper_multi_chan
  ! multi-channel hooks for sockets
  public :: read_multi_chan, write_multi_chan

  !> channel management
  !> allow direct access to member variables...
  public :: ntot_chan, qactive_chan, qselect_chan, contrib_chan
  public :: current_chan, ix_chan, ip1_chan, ip2_chan
  public :: OLiproc_chan, OLntot_chan  ! OL info

  public :: ireject_event_chan
  public :: acc_threshold_OL

  !> management of inclusive calculations
  public :: set_order_inc_chan, get_order_inc_chan

  !> contribution management
  !> integer id's for the different contributions stored in contrib_chan
  integer, public, parameter :: id_contr_inc = -1
  integer, public, parameter :: id_contr_LO  = 1
  integer, public, parameter :: id_contr_R1  = 2
  integer, public, parameter :: id_contr_V1  = 3
  integer, public, parameter :: id_contr_RR  = 4
  integer, public, parameter :: id_contr_RV  = 5
  integer, public, parameter :: id_contr_VV  = 6
  public :: contrib_int_to_str_chan, contrib_str_to_int_chan
  public :: contrib_nV_chan, contrib_nR_chan, contrib_nO_chan
  public :: contrib_to_str_chan
  public :: set_default_region_chan
  public :: parseContrib_chan, enableChannel_chan
  public :: resetCurrentContrib_chan, loadNextContrib_chan
  public :: currentContrib_chan, currentSubContrib_chan
  public :: info_chan

  !> inclusive calculations have their own order management (id_contr_inc)
  !> variable storing bit-flags for the orders we want to compute
  !>  bit-0) LO
  !>  bit-1) NLO coefficient
  !>  bit-2) NNLO coefficient
  !>  ...
  integer :: order_inc = 0  ! no flag set yet

  !> channel information for the process
  logical, dimension(:), allocatable :: qactive_chan  ! selected by runcard
  logical, dimension(:), allocatable :: qselect_chan  ! for multi-channel (& more?)
  integer, dimension(:), allocatable :: contrib_chan
  integer, dimension(:), allocatable :: OLiproc_chan  ! OpenLoops ids
  integer :: ntot_chan = -1
  integer :: OLntot_chan = -1  ! total number of channels calling OL amplitudes
  integer :: default_region_chan = -1  ! -1=all, 1=region a, 2=region b
  integer :: current_chan = -1
  integer :: ix_chan = -1
  integer :: ip1_chan, ip2_chan
!$OMP THREADPRIVATE(qselect_chan)
!$OMP THREADPRIVATE(current_chan, ix_chan, ip1_chan, ip2_chan)

  !> flag events to reject
  !>    0:  event survives
  !>  +ve:  iproc that triggered
  !>   -1:  tcut that triggered
  !>   -2:  encountered NaN
  integer :: ireject_event_chan = 0
  double precision :: acc_threshold_OL = 10
!$OMP THREADPRIVATE(ireject_event_chan)

  !> for multi-channel adaption
  logical :: qmulti_chan = .false.  ! off by default
  !>   0: off [default]
  !>  >0: nselect_multi_chan channels in one group
  !>  <0: divide into |nselect_multi_chan| groups
  integer :: nselect_multi_chan = 0  ! off by default
  integer :: nmulti_chan = 0  ! # channels
  integer :: ilower_multi_chan = -1
  integer :: iupper_multi_chan = -1
  !>  single channel:  [ilower_multi_chan, iupper_multi_chan]
  !>  channel groups:  [1, ngroup_chan]
  integer :: current_multi_chan = -1
!$OMP THREADPRIVATE(current_multi_chan)
  !> Write & read the weights using hexadecimal
  character(len=9), parameter :: file_fmt = "(/(5z16))"
  double precision, dimension(:), allocatable   :: weight_multi_chan
  double precision, dimension(:), allocatable   :: bounds_multi_chan
  double precision, dimension(:,:), allocatable :: buffer_multi_chan  ! (0:1) <-> err & wgt
!$OMP THREADPRIVATE(buffer_multi_chan)
  double precision, dimension(:,:), allocatable :: accum_multi_chan  ! (0:2) <-> err & wgt & wgt^2
  integer, dimension(:), allocatable            :: count_multi_chan
  !> variables for channel groups
  integer :: ngroup_chan = 0
  double precision, dimension(:), allocatable :: weight_group_chan
  double precision, dimension(:), allocatable :: bounds_group_chan
  !> dimension: nmulti_chan -> assign a group to each active channel
  integer, dimension(:), allocatable :: group_multi_chan

#ifdef CHECKACCUM
  real(qp) :: buffer_qp
  real(dp) :: buffer_dp
!$OMP THREADPRIVATE(buffer_qp, buffer_dp)
#endif


  !> type storing all information of a contribution
  type Contrib_t
    !> contr for the full channel: RR = [LO,RR], [R1,R1], [LO,RR]
    integer :: id_full = 0
    integer :: nsub
    !> contr for the sub-processes (prod, decay1, decay2, ...)
    integer, allocatable, dimension(:) :: id_sub
    !> info on the regions: -1=unset, 0=none, 1=region a, 2=region b
    integer, allocatable, dimension(:) :: region_sub
    type(Contrib_t), pointer :: next => null()
  end type Contrib_t

  !> the linked list to store *distinct* contributions
  type(Contrib_t), pointer :: headContrib_p => null()
  !> the pointer for the currently active contribution
  type(Contrib_t), pointer :: currentContrib_p => null()

  !> either parse an integer id or a string
  interface parseContrib_chan
    module procedure parseContrib_int_chan
    module procedure parseContrib_str_chan
  end interface


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the channels module
  !-----------------------------------------------------------------------------
  subroutine init_chan()
    use Process_mod
    use ProcLib_mod

    if ( allocated(qactive_chan) .or.  &
       & allocated(contrib_chan) ) then
      print*, 'init_chan: already allocated?!'
      stop
    end if

    call selectchannel_dl()

    select case(name_proc)

      case ( 'HINC', 'HTO2PINC', 'HTO2TAUINC', 'HTO2L1PINC',  &
           & 'HTO2E2MUINC', 'HTO4EINC', 'HTO2L2NINC' )

        n_alphas_proc = 2
        nsub_proc = 1

        if ( allocated(qactive_chan) .or.  &
           & allocated(contrib_chan) ) then
          print*, 'init_chan: RAPIDIX already allocated'
          stop
        end if

        ntot_chan = 1

        allocate(qactive_chan(1:ntot_chan))
        allocate(contrib_chan(1:ntot_chan))

        qactive_chan(:) = .true.
        contrib_chan(:) = id_contr_inc

      case ( 'HQT', 'HTO2PQT', 'HTO2TAUQT', 'HTO2L1PQT',  &
           & 'HTO2E2MUQT', 'HTO4EQT', 'HTO2L2NQT' )

        n_alphas_proc = 2
        nsub_proc = 1

        if ( allocated(qactive_chan) .or.  &
           & allocated(contrib_chan) ) then
          print*, 'init_chan: Hqt already allocated'
          stop
        end if

        ntot_chan = 1

        allocate(qactive_chan(1:ntot_chan))
        allocate(contrib_chan(1:ntot_chan))

        qactive_chan(:) = .true.
        contrib_chan(:) = id_contr_inc

      case ( 'ZQT', 'WPQT', 'WMQT', 'GGQT' )

        n_alphas_proc = 0
        nsub_proc = 1

        if ( allocated(qactive_chan) .or.  &
           & allocated(contrib_chan) ) then
          print*, 'init_chan: [V]qt already allocated'
          stop
        end if

        ntot_chan = 1

        allocate(qactive_chan(1:ntot_chan))
        allocate(contrib_chan(1:ntot_chan))

        qactive_chan(:) = .true.
        contrib_chan(:) = id_contr_inc

      case ( 'EPEMINC' )

        n_alphas_proc = 0
        nsub_proc = 1

        if ( allocated(qactive_chan) .or.  &
           & allocated(contrib_chan) ) then
          print*, 'init_chan: EEinc already allocated'
          stop
        end if

        ntot_chan = 1

        allocate(qactive_chan(1:ntot_chan))
        allocate(contrib_chan(1:ntot_chan))

        qactive_chan(:) = .true.
        contrib_chan(:) = id_contr_inc


      case default
        if (.not.qloaded_plib) then
          print*, "init_chan: unknown name_proc: ", name_proc
          stop
        end if
    end select

    !> better place to do this?
    if (.not.allocated(accum_multi_chan)) allocate(accum_multi_chan(0:2,1:ntot_chan))
    accum_multi_chan(:,:) = 0d0
    if (.not.allocated(count_multi_chan)) allocate(count_multi_chan(1:ntot_chan))
    count_multi_chan(:) = 0

  end subroutine init_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the channel strings
  !
  !> @param[in] ichan the channel id
  !> @param[out] chan_str string of the channel
  !> @param[out] me_str string name of the Matrix Element
  !> @param[out] sub_str string name for the subtraction term
  !-----------------------------------------------------------------------------
  subroutine get_chan(ichan, chan_str, me_str, sub_str)
    use Process_mod
    use ProcLib_mod
    integer, intent(in) :: ichan
    character(:), allocatable, intent(out) :: chan_str, me_str, sub_str

    call getchannel_dl(ichan, chan_str, me_str, sub_str)
    !return

    select case(name_proc)

!      case('1JET', '2JET')
!        call getchanneljet(ichan, chan_str, me_str, sub_str)
!
!      case('1JETFC', '2JETFC')
!        call getchannel2jetfc(ichan, chan_str, me_str, sub_str)
!
!      case('3JET')
!        call getchannel3jet(ichan, chan_str, me_str, sub_str)
!
!      case('4JET')
!        call getchannel4jet(ichan, chan_str, me_str, sub_str)
!
!      case('5JET')
!        call getchannel5jet(ichan, chan_str, me_str, sub_str)
!
!      case('GG')
!        call getchannelGG(ichan, chan_str, me_str, sub_str)
!
!      case('GG_EW')
!        call getchannelGG_EW(ichan, chan_str, me_str, sub_str)
!
!      case('GGG')
!        call getchannelGGG(ichan, chan_str, me_str, sub_str)
!
!      case('GGJ')
!        call getchannelGGJ(ichan, chan_str, me_str, sub_str)
!
!      case('GGJFC')
!        call getchannelGGJfc(ichan, chan_str, me_str, sub_str)
!
!      case('GGJJ')
!        call getchannelGGJJ(ichan, chan_str, me_str, sub_str)
!
!      case('GGJJFC')
!        call getchannelGGJJfc(ichan, chan_str, me_str, sub_str)
!
!      case('GGJJJFC')
!        call getchannelGGJJJfc(ichan, chan_str, me_str, sub_str)
!
!      case('Z')
!        call getchannelZ(ichan, chan_str, me_str, sub_str)
!
!      case('Z_EW')
!        call getchannelZ_EW(ichan, chan_str, me_str, sub_str)
!
!      case('ZJ')
!        call getchannelZJ(ichan, chan_str, me_str, sub_str)
!
!      case('ZJ_H')
!        call getchannelZJ_h(ichan, chan_str, me_str, sub_str)
!
!      case('ZJJ')
!        call getchannelZJJ(ichan, chan_str, me_str, sub_str)
!
!      case('ZJJJ')
!        call getchannelZJJJ(ichan, chan_str, me_str, sub_str)
!
!      case('WM')
!        call getchannelWm(ichan, chan_str, me_str, sub_str)
!
!      case('WP')
!        call getchannelWp(ichan, chan_str, me_str, sub_str)
!
!      case('WMJ')
!        call getchannelWmJ(ichan, chan_str, me_str, sub_str)
!
!      case('WPJ')
!        call getchannelWpJ(ichan, chan_str, me_str, sub_str)
!
!      case('WMJJ')
!        call getchannelWmJJ(ichan, chan_str, me_str, sub_str)
!
!      case('WPJJ')
!        call getchannelWpJJ(ichan, chan_str, me_str, sub_str)
!
!      case('WMJJJ')
!        call getchannelWmJJJ(ichan, chan_str, me_str, sub_str)
!
!      case('WPJJJ')
!        call getchannelWpJJJ(ichan, chan_str, me_str, sub_str)
!
!      case('H')
!        call getchannelH(ichan, chan_str, me_str, sub_str)
!
!      case('HJ')
!        call getchannelHJ(ichan, chan_str, me_str, sub_str)
!
!      case('HJJ')
!        call getchannelHJJ(ichan, chan_str, me_str, sub_str)
!
!      case('HJJJ')
!        call getchannelHJJJ(ichan, chan_str, me_str, sub_str)
!
!      case('HM')
!        call getchannelHm(ichan, chan_str, me_str, sub_str)
!
!      case('HJM')
!        call getchannelHJm(ichan, chan_str, me_str, sub_str)
!
!      case('HJMREW')
!        call getchannelHJmrew(ichan, chan_str, me_str, sub_str)
!
!      case('HJJM')
!        call getchannelHJJm(ichan, chan_str, me_str, sub_str)
!
!      case('HJJMREW')
!        call getchannelHJJmrew(ichan, chan_str, me_str, sub_str)
!
!      case('HJJJM')
!        call getchannelHJJJm(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2PM', 'HTO2TAUM')
!        call getchannelH2m(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2PJM', 'HTO2TAUJM')
!        call getchannelH2Jm(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2P', 'HTO2TAU')
!        call getchannelH2(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2PJ', 'HTO2TAUJ')
!        call getchannelH2J(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2PJJ', 'HTO2TAUJJ')
!        call getchannelH2JJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2PJJJ', 'HTO2TAUJJJ')
!        call getchannelH2JJJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1P')
!        call getchannelH3(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1PJ')
!        call getchannelH3J(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1PJJ')
!        call getchannelH3JJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1PJJJ')
!        call getchannelH3JJJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1PM')
!        call getchannelH3m(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2L1PJM')
!        call getchannelH3Jm(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MU', 'HTO4E', 'HTO2L2N')
!        call getchannelH4(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MUJ', 'HTO4EJ', 'HTO2L2NJ')
!        call getchannelH4J(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MUJJ', 'HTO4EJJ', 'HTO2L2NJJ')
!        call getchannelH4JJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MUJJJ', 'HTO4EJJJ', 'HTO2L2NJJJ')
!        call getchannelH4JJJ(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MUM', 'HTO4EM', 'HTO2L2NM')
!        call getchannelH4m(ichan, chan_str, me_str, sub_str)
!
!      case('HTO2E2MUJM', 'HTO4EJM', 'HTO2L2NJM')
!        call getchannelH4Jm(ichan, chan_str, me_str, sub_str)
!
!      case('VFH')
!        call getchannelVFH(ichan, chan_str, me_str, sub_str)
!
!      case('VFHFULL')
!        call getchannelVFHfull(ichan, chan_str, me_str, sub_str)
!
!      case('VFHJ')
!        call getchannelVFHJ(ichan, chan_str, me_str, sub_str)
!
!      case('DIS')
!        call getchannelDIS(ichan, chan_str, me_str, sub_str)
!
!      case('DIS1J')
!        call getchannelDIS1j(ichan, chan_str, me_str, sub_str)
!
!      case('EPEM')
!        call getchannelepem(ichan, chan_str, me_str, sub_str)
!
!      case('EPEM_H')
!        call getchannelepem_h(ichan, chan_str, me_str, sub_str)
!
!      case('EPEMJJ')
!        call getchannelepemjj(ichan, chan_str, me_str, sub_str)
!
!      case('EPEMJJ_H')
!        call getchannelepemjj_h(ichan, chan_str, me_str, sub_str)
!
!      case('SFDIS')
!        call getchannelSFDIS(ichan, chan_str, me_str, sub_str)
!
!      case('DISWM')
!        call getchannelDISWm(ichan, chan_str, me_str, sub_str)
!
!      case('DISWP')
!        call getchannelDISWp(ichan, chan_str, me_str, sub_str)
!
!      case('DISWM1J')
!        call getchannelDISWm1j(ichan, chan_str, me_str, sub_str)
!
!      case('DISWP1J')
!        call getchannelDISWp1j(ichan, chan_str, me_str, sub_str)
!
!      case('GJ')
!        call getchannelGJ(ichan, chan_str, me_str, sub_str)
!
!      case('G_EW')
!        call getchannelG_EW(ichan, chan_str, me_str, sub_str)
!
!      case('GJJ')
!        call getchannelGJJ(ichan, chan_str, me_str, sub_str)
!
!      case('GJJJ')
!        call getchannelGJJJ(ichan, chan_str, me_str, sub_str)
!
!      case('DISG')
!        call getchannelDISG(ichan, chan_str, me_str, sub_str)
!
!      case('SFDISWP')
!        call getchannelSFDISWp(ichan, chan_str, me_str, sub_str)
!
!      case('SFDISWM')
!        call getchannelSFDISWm(ichan, chan_str, me_str, sub_str)

      case ( 'HINC', 'HTO2PINC', 'HTO2TAUINC', 'HTO2L1PINC',  &
           & 'HTO2E2MUINC', 'HTO4EINC', 'HTO2L2NINC' )
        if (ichan /= 1) stop "get_chan: invalid ichan for RAPIDIX process"
        chan_str = 'RAPIDIX'
        me_str   = 'RAPIDIX'
        sub_str  = 'RAPIDIX'

      case ( 'HQT', 'HTO2PQT', 'HTO2TAUQT', 'HTO2L1PQT',  &
           & 'HTO2E2MUQT', 'HTO4EQT', 'HTO2L2NQT' )
        if (ichan /= 1) stop "get_chan: invalid ichan for HQT process"
        chan_str = 'HQT'
        me_str   = 'HQT'
        sub_str  = 'HQT'

      case ( 'ZQT' )
        if (ichan /= 1) stop "get_chan: invalid ichan for ZQT process"
        chan_str = 'ZQT'
        me_str   = 'ZQT'
        sub_str  = 'ZQT'

      case ( 'WPQT', 'WMQT' )
        if (ichan /= 1) stop "get_chan: invalid ichan for WQT process"
        chan_str = 'WQT'
        me_str   = 'WQT'
        sub_str  = 'WQT'

      case ( 'GGQT' )
        if (ichan /= 1) stop "get_chan: invalid ichan for GGQT process"
        chan_str = 'GGQT'
        me_str   = 'GGQT'
        sub_str  = 'GGQT'

      case ( 'EPEMINC' )
        if (ichan /= 1) stop "get_chan: invalid ichan for EPEMINC process"
        chan_str = 'EPEMINC'
        me_str   = 'EPEMINC'
        sub_str  = 'EPEMINC'

      ! case default
      !   print*, "get_chan: unknown name_proc", name_proc
      !   stop
    end select

  end subroutine get_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a contribution has regions
  !
  !> @param[in] contr
  !> @return qregions flag if the contribution has phase-space regions
  !-----------------------------------------------------------------------------
  function hasRegions_chan(contr) result(qregions)
    use Process_mod
    use Run_mod, only : ilips_run
    type(Contrib_t), intent(in) :: contr
    logical, dimension(:), allocatable :: qregions
    integer :: npar


    allocate(qregions(contr%nsub))
    qregions(:) = .false.  ! default: no regions

    npar =  n_fs_partons_proc  &
         & +n_fs_spectators_proc  &
         & +n_fs_photons_proc  &
         & +contrib_nR_chan(contr%id_full)  &
         & +2  ! # IS particles

    ! select case (id_class_proc)
    ! !>-----
    ! case ( id_class_proc_jet,  &
    !      & id_class_proc_GJ,  &
    !      & id_class_proc_GG,  &
    !      & id_class_proc_DISG )
    !   qregions(1) = (npar == 6)
    ! !>-----
    ! case ( id_class_proc_Z,  &
    !      & id_class_proc_Wm,  &
    !      & id_class_proc_Wp,  &
    !      & id_class_proc_H,  &
    !      & id_class_proc_Hm,  &
    !      & id_class_proc_Hto2p,  &
    !      & id_class_proc_Hto2pm,  &
    !      & id_class_proc_Hto2tau,  &
    !      & id_class_proc_Hto2taum,  &
    !      & id_class_proc_DIS,  &
    !      & id_class_proc_epem )
    !   qregions(1) = (npar == 7)
    ! !>-----
    ! case ( id_class_proc_Hto2l1p,  &
    !      & id_class_proc_Hto2l1pm )
    !   qregions(1) = (npar == 8)
    ! !>-----
    ! case ( id_class_proc_Hto2e2mu,  &
    !      & id_class_proc_Hto2e2mum,  &
    !      & id_class_proc_Hto4e,  &
    !      & id_class_proc_Hto4em,  &
    !      & id_class_proc_Hto2l2n,  &
    !      & id_class_proc_Hto2l2nm )
    !   qregions(1) = (npar == 9)
    ! end select


    if (     ((npar.eq.6).and.(name_proc(1:4).eq.'1JET'))  &
       & .or.((npar.eq.6).and.(name_proc(1:4).eq.'2JET'))  &
       & .or.((npar.eq.6).and.(name_proc(1:2).eq.'GJ')  &
       &                 .and.(name_proc(1:3).ne.'GJJ')  &
       &                 .and.(name_proc(1:3).ne.'WPH')  &
       &                 .and.(name_proc(1:3).ne.'WMH'))  &
       & .or.((npar.eq.6).and.(name_proc(1:4).eq.'DISG'))  &
       & .or.((npar.eq.7).and.(name_proc(1:5).ne.'HTO4E')  &
       &                 .and.(name_proc(1:2).ne.'HH')  &
       &                 .and.(name_proc(1:7).ne.'HTO2L2N')  &
       &                 .and.(name_proc(1:8).ne.'HTO2E2MU')  &
       &                 .and.(name_proc(1:7).ne.'HTO2L1P')  &
       &                 .and.(name_proc(1:3).ne.'HJJ')  &
       &                 .and.(name_proc(1:7).ne.'HTO2PJJ')  &
       &                 .and.(name_proc(1:9).ne.'HTO2TAUJJ')  &
       &                 .and.(name_proc(1:8).ne.'EPEMJJJJ')  &
       &                 .and.(name_proc(1:4).ne.'WPJJ')  &
       &                 .and.(name_proc(1:4).ne.'WMJJ')  &
       &                 .and.(name_proc(1:4).ne.'4JET')  &
       &                 .and.(name_proc(1:4).ne.'5JET')  &
       &                 .and.(name_proc(1:3).ne.'ZJJ')  &
       &                 .and.(name_proc(1:6).ne.'GGJJFC') &
       &                 .and.(name_proc(1:7).ne.'GGJJJFC') &
       &                 .and.(name_proc     .ne.'VFH')  &
       &                 .and.(name_proc     .ne.'VFHJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2P')  &
       &                 .and.(name_proc     .ne.'VFHTO2PJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2L1P')  &
       &                 .and.(name_proc     .ne.'VFHTO2TAU') &
       &                 .and.(name_proc(1:3).ne.'ZHJ') &
       &                 .and.(name_proc(1:3).ne.'WPH') &
       &                 .and.(name_proc(1:3).ne.'WMH'))  &
       & .or.((npar.eq.8).and.(name_proc(1:5).ne.'HTO4E')  &
       &                 .and.(name_proc(1:2).ne.'HH')  &
       &                 .and.(name_proc(1:7).ne.'HTO2L2N')  &
       &                 .and.(name_proc(1:8).ne.'HTO2E2MU')  &
       &                 .and.(name_proc(1:9).ne.'HTO2L1PJJ')  &
       &                 .and.(name_proc     .ne.'VFHJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2PJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2L1P')  &
       &                 .and.(name_proc     .ne.'VFHTO2E2MU')  &
       &                 .and.(name_proc     .ne.'VFHTO4E')  &
       &                 .and.(name_proc     .ne.'VFHTO2L2N')  &
       &                 .and.(name_proc     .ne.'VFHTO2L1PJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2TAUJ') &
       &                 .and.(name_proc(1:3).ne.'ZHJ') &
       &                 .and.(name_proc(1:3).ne.'WPH') &
       &                 .and.(name_proc(1:3).ne.'WMH'))  &
       & .or.((npar.eq.9).and.(name_proc(1:5).eq.'HTO4E')  &
       &                 .and.(name_proc(6:7).ne.'JJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:2).eq.'HH')  &
       &                 .and.(name_proc(3:4).ne.'JJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:8).eq.'HTO2E2MU')  &
       &                 .and.(name_proc(9:10).ne.'JJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:7).eq.'HTO2L2N')  &
       &                 .and.(name_proc(8:9).ne.'JJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:3).eq.'VFH')  &
       &                 .and.(name_proc     .ne.'VFHTO2E2MU')  &
       &                 .and.(name_proc     .ne.'VFHTO4E')  &
       &                 .and.(name_proc     .ne.'VFHTO2L2N')  &
       &                 .and.(name_proc     .ne.'VFHTO2E2MUJ')  &
       &                 .and.(name_proc     .ne.'VFHTO4EJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2L2NJ')  &
       &                 .and.(name_proc     .ne.'VFHTO2L1PJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:4).eq.'WMHJ'))  &
       & .or.((npar.eq.9).and.(name_proc(1:4).eq.'WPHJ'))  &
       & .or.((npar.eq.10).and.(name_proc(1:3).eq.'VFH')  &
       &                  .and.(name_proc.ne.'VFHTO2E2MUJ')  &
       &                  .and.(name_proc.ne.'VFHTO4EJ')  &
       &                  .and.(name_proc.ne.'VFHTO2L2NJ'))  &
       & ) qregions(1) = .true.


    !> @todo: for prod-decay types we still need to add cases
    if (contr%nsub /= 1) then
      print*, "hasRegions_chan: not set up for prod-decay processes yet"
      stop
    end if

  end function hasRegions_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy the run module by de-allocating everything we allocated
  !> dynamically on the course of the run
  !-----------------------------------------------------------------------------
  subroutine destroy_chan()
    type(Contrib_t), pointer :: it, next

    !> deallocate channel arrays
    if (allocated(qactive_chan)) deallocate(qactive_chan)
    if (allocated(contrib_chan)) deallocate(contrib_chan)

    !> multi-channel global arrays
    if (allocated(weight_multi_chan)) deallocate(weight_multi_chan)
    if (allocated(bounds_multi_chan)) deallocate(bounds_multi_chan)
    if (allocated(accum_multi_chan))  deallocate(accum_multi_chan)
    if (allocated(count_multi_chan))  deallocate(count_multi_chan)

    !> channel group
    if (allocated(weight_group_chan)) deallocate(weight_group_chan)
    if (allocated(bounds_group_chan)) deallocate(bounds_group_chan)
    if (allocated(group_multi_chan))  deallocate(group_multi_chan)

    !> de-allocate the contrib-list
    it => headContrib_p

    do while ( associated(it) )
      !> save next before we destroy it
      next => it%next
      !> deallocate sub-entries and then the entry itself
      if (allocated(it%id_sub)) deallocate(it%id_sub)
      deallocate(it)
      !> loop
      it => next
    end do
    nullify(headContrib_p)
    nullify(currentContrib_p)

  end subroutine destroy_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> output info on the channels settings
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_chan(unit)
    integer, optional, intent(in) :: unit
    integer :: iout, ichan
    character(len=:), allocatable :: chan_str, me_str, sub_str

    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_label =  &
      & '(1x,"*",1x,"  id",T10,"partons",T35,"Matrix Element",T65,"Subtraction",T103,"*")'
    character(len=*), parameter :: fmt_str   =  &
      & '(1x,"*",1x,I4,T10,A,T35,A,T65,A,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_label)
    write(iout,fmt_blank)

    do ichan=1,ntot_chan
      if (.not. qactive_chan(ichan)) cycle
      call get_chan(ichan, chan_str, me_str, sub_str)
      write(iout,fmt_str) ichan, trim(chan_str), trim(me_str), trim(sub_str)
    end do

    ! write(iout,*) "active contributions"
    ! call resetCurrentContrib_chan()
    ! do while (loadNextContrib_chan())
    ! end do

  end subroutine digest_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> output info on the luminosity categories and channel association
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digestLumi_chan(unit)
    use ProcLib_mod, only : getlumiid_dl
    integer, optional, intent(in) :: unit
    integer :: iout, ichan, icont, ilumi, nlumi, jlumi, ip1, ip2
    integer, dimension(:,:), allocatable :: id_lumi

    integer :: id, nsub, nV, nR, nO, nas, npar, ndim
    integer :: nsplit, nfrag
    character(:), allocatable :: str, runid

    character(len=*), parameter :: fmt_chan = '(I0,1x)'
    character(len=*), parameter :: fmt_head = '(A,"_",I0,"  ")'
    character(len=*), parameter :: fmt_tail = '(" ! channel: [",I0,",",I0,"]")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    allocate(id_lumi(-1:6,1:ntot_chan))  ! [-1:6] covers all id_contrib_xyz
    id_lumi(:,:) = 0
    nlumi = 0

    do ichan=1,ntot_chan
      ilumi = getlumiid_dl(ichan, ip1, ip2)
      if (ilumi > nlumi)  nlumi = ilumi
      icont = contrib_chan(ichan)
      id_lumi(icont,ichan) = ilumi
    end do

    ! do icont=-1,6
    !   do ilumi=1,nlumi
    !     if ( all(id_lumi(icont,:) /= ilumi) ) cycle
    !     do ichan=1,ntot_chan
    !       if ( id_lumi(icont,ichan) /= ilumi ) cycle
    !       write(iout,fmt_chan,advance="no") ichan
    !     end do
    !     write(iout,fmt_head) contrib_int_to_str_chan(icont), ilumi
    !   end do
    ! end do

    !> if we have regions this is a smarter way to do it
    !> after registering `ALL` channels
    call resetCurrentContrib_chan()
    do while (loadNextContrib_chan(verbose=0))
      call currentContrib_chan(id=id, nsub=nsub, str=str,  &
        &  nV=nV, nR=nR, nO=nO, nas=nas, npar=npar,  &
        &  ndim=ndim, nsplit=nsplit, nfrag=nfrag, runid=runid )
      do ilumi=1,nlumi
        if ( all(id_lumi(id,:) /= ilumi) ) cycle
        write(iout,fmt_head,advance="no") contrib_to_str_chan(currentContrib_p), ilumi
        do ichan=1,ntot_chan
          if ( id_lumi(id,ichan) /= ilumi ) cycle
          jlumi = getlumiid_dl(ichan, ip1, ip2)  ! overkill but we don't care about performance...
          if (ilumi /= jlumi) then
            print*, "digestLumi_chan: inconsistent lumi idx", ilumi, jlumi
            stop
          end if
          write(iout,fmt_chan,advance="no") ichan
        end do
        write(iout,fmt_tail) ip1, ip2
      end do

    end do

  end subroutine digestLumi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the default region
  !
  !> @param[in] region_str specify fall-back region ("a", "b", "all")
  !-----------------------------------------------------------------------------
  subroutine set_default_region_chan(region_str)
    use StrHelper_mod
    character(len=*), intent(in) :: region_str

    select case(to_upper_str( trim(adjustl( region_str )) ))
      case('A')
        default_region_chan =  1
      case('B')
        default_region_chan =  2
      case('ALL')
        default_region_chan = -1
      case default
        print*, "set_default_region_chan: unknown region: ", region_str
        stop
    end select

  end subroutine set_default_region_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> convert string to a contrib-identifier
  !
  !> @param[in] contr_str string ("LO", "RR", ...)
  !> @return id_contr corresponding to the string
  !-----------------------------------------------------------------------------
  function contrib_str_to_int_chan(contr_str) result(contr_int)
    use StrHelper_mod
    character(len=*), intent(in) :: contr_str
    integer :: contr_int

    contr_int = -1

    select case(to_upper_str( trim(adjustl( contr_str )) ))
      case('LO')
        contr_int = id_contr_LO
      case('R', 'R1')
        contr_int = id_contr_R1
      case('V', 'V1')
        contr_int = id_contr_V1
      case('RR')
        contr_int = id_contr_RR
      case('RV', 'VR')
        contr_int = id_contr_RV
      case('VV')
        contr_int = id_contr_VV
      case default
        print*, "contrib_str_to_int_chan: unknown input:", contr_str
        stop
    end select

  end function contrib_str_to_int_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> convert contrib-identifier to a string
  !
  !> @param[in] contr_int id_contr identifier for a contribution
  !> @return string ("LO", "RR", ...) corresponding to the id_contr
  !-----------------------------------------------------------------------------
  function contrib_int_to_str_chan(contr_int) result(contr_str)
    use StrHelper_mod
    integer, intent(in) :: contr_int
    character(:), allocatable :: contr_str

    contr_str = ''

    select case(contr_int)
      case(id_contr_LO)
        contr_str = 'LO'
      case(id_contr_R1)
        contr_str = 'R'
      case(id_contr_V1)
        contr_str = 'V'
      case(id_contr_RR)
        contr_str = 'RR'
      case(id_contr_RV)
        contr_str = 'RV'
      case(id_contr_VV)
        contr_str = 'VV'
      case(id_contr_inc)
        contr_str = 'inc'
      case default
        print*, "contrib_int_to_str_chan: unknown input:", contr_int
        stop
    end select

  end function contrib_int_to_str_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the number of loops
  !
  !> @param[in] contr_int id_contr identifier for a contribution
  !> @return number of loops
  !-----------------------------------------------------------------------------
  integer function contrib_nV_chan(contr_int)
    integer, intent(in) :: contr_int
    select case(contr_int)
      case(id_contr_LO)
        contrib_nV_chan = 0
      case(id_contr_R1)
        contrib_nV_chan = 0
      case(id_contr_V1)
        contrib_nV_chan = 1
      case(id_contr_RR)
        contrib_nV_chan = 0
      case(id_contr_RV)
        contrib_nV_chan = 1
      case(id_contr_VV)
        contrib_nV_chan = 2
      case(id_contr_inc)
        !> look for highest order
        !> (if mixed setup NLO = LO + NLO_only, we should be careful)
        contrib_nV_chan = 0
        if (get_order_inc_chan(1))  contrib_nV_chan = 1
        if (get_order_inc_chan(2))  contrib_nV_chan = 2
        if (get_order_inc_chan(3))  contrib_nV_chan = 3
      case default
        !> invalid input:
        contrib_nV_chan = -1
        ! stop
    end select
  end function contrib_nV_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the number of real emissions
  !
  !> @param[in] contr_int id_contr identifier for a contribution
  !> @return number of real emissions
  !-----------------------------------------------------------------------------
  integer function contrib_nR_chan(contr_int)
    integer, intent(in) :: contr_int
    select case(contr_int)
      case(id_contr_LO)
        contrib_nR_chan = 0
      case(id_contr_R1)
        contrib_nR_chan = 1
      case(id_contr_V1)
        contrib_nR_chan = 0
      case(id_contr_RR)
        contrib_nR_chan = 2
      case(id_contr_RV)
        contrib_nR_chan = 1
      case(id_contr_VV)
        contrib_nR_chan = 0
      case(id_contr_inc)
        !> inclusive: no real emissions
        contrib_nR_chan = 0
      case default
        !> invalid input:
        contrib_nR_chan = -1
        ! stop
    end select
  end function contrib_nR_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the order = (nV + nR)
  !
  !> @param[in] contr_int id_contr identifier for a contribution
  !> @return order of the contribution
  !-----------------------------------------------------------------------------
  integer function contrib_nO_chan(contr_int)
    integer, intent(in) :: contr_int
    select case(contr_int)
      case(id_contr_LO)
        contrib_nO_chan = 0
      case(id_contr_R1)
        contrib_nO_chan = 1
      case(id_contr_V1)
        contrib_nO_chan = 1
      case(id_contr_RR)
        contrib_nO_chan = 2
      case(id_contr_RV)
        contrib_nO_chan = 2
      case(id_contr_VV)
        contrib_nO_chan = 2
      case(id_contr_inc)
        contrib_nO_chan = contrib_nV_chan(contr_int)  ! inc: no real
      case default
        !> invalid input:
        contrib_nO_chan = -1
        ! stop
    end select
  end function contrib_nO_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> convert a Contrib_t type to a string for print's and filenames
  !
  !> @param[in] contr the contribution
  !> @return string of the format: "RR[R,R]"
  !-----------------------------------------------------------------------------
  function contrib_to_str_chan(contr) result(contr_str)
    use StrHelper_mod
    type(Contrib_t), intent(in) :: contr
    character(:), allocatable :: contr_str, sep
    integer :: i

    contr_str = contrib_int_to_str_chan(contr%id_full)

    if (contr%nsub > 1) then
      contr_str = trim(contr_str) // "["
      sep = ''
      do i=1,contr%nsub
        contr_str = trim(contr_str) //  &
                  & sep //  &
                  & contrib_int_to_str_chan(contr%id_sub(i)) //  &
                  & region_to_str(contr%region_sub(i))
        sep = ','
      end do
      contr_str = trim(contr_str) // "]"
    else
        contr_str = trim(contr_str) //  &
                  & region_to_str(contr%region_sub(1))
    end if

  contains

    !-----------------------------------------------------------------------------
    !> convert the region id to a string suffix
    !-----------------------------------------------------------------------------
    function region_to_str(region) result(region_str)
      integer, intent(in) :: region
      character(:), allocatable :: region_str
      region_str = ''
      select case (region)
      case(-1)
        region_str = ''
      case(0)
        region_str = ''
      case(1)
        region_str = 'a'
      case(2)
        region_str = 'b'
      case default
        print*, "region_to_str: invalid input: ", region
        stop
      end select
    end function region_to_str

  end function contrib_to_str_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> parse an interger encoding a contribution (`icontr` in selectchannelXYZ)
  !
  !> @param[in] icontr integer encoding controbution
  !> @param[inout] contr Contrib_t that corresponds to icontr
  !-----------------------------------------------------------------------------
  subroutine parseContrib_int_chan(icontr, contr)
    integer, intent(in) :: icontr
    type(Contrib_t), intent(inout) :: contr
    integer :: i, tmp
    integer :: nV, nR

    ! print*, "parseContrib_int_chan:", icontr

    if (icontr <= 0) then
      print*, "parseContrib_int_chan: invalid icontr", icontr
      stop
    end if

    !> we keep the `next` pointer in tact and don't nullify it

    !> number of digits
    contr%nsub = int( log10(dble(icontr)) ) + 1

    !> allocate the array containing the sub-id's
    if (allocated(contr%id_sub)) deallocate(contr%id_sub)
    allocate(contr%id_sub(1:contr%nsub))

    !> default-init the array containing the region info
    if (allocated(contr%region_sub)) deallocate(contr%region_sub)
    allocate(contr%region_sub(1:contr%nsub))
    contr%region_sub(:) = -1  ! no info in icontr

    !> read in the digits
    tmp = icontr
    do i=1,contr%nsub
      contr%id_sub(contr%nsub-i+1) = tmp - int(tmp/10) * 10
      tmp = int(tmp/10)
    end do
    ! print*, "parsed icontr"
    ! do i=1,contr%nsub
    !   print*, "  ",i,contr%id_sub(i)
    ! end do

    !> the first index encodes the full id
    contr%id_full = contr%id_sub(1)

    !> reconstruct the id for the production and put it in the array
    nV = contrib_nV_chan(contr%id_full)
    nR = contrib_nR_chan(contr%id_full)

    !> subtract the decays
    do i=2,contr%nsub
      nV = nV - contrib_nV_chan(contr%id_sub(i))
      nR = nR - contrib_nR_chan(contr%id_sub(i))
    end do
    !> build up the production id
    select case(nR)
      case(0)
        select case(nV)
          case(0)
            contr%id_sub(1) = id_contr_LO
          case(1)
            contr%id_sub(1) = id_contr_V1
          case(2)
            contr%id_sub(1) = id_contr_VV
          case default
            print*, "parseContrib_int_chan: invalid nR,nV", nR,nV
            stop
        end select
      case(1)
        select case(nV)
          case(0)
            contr%id_sub(1) = id_contr_R1
          case(1)
            contr%id_sub(1) = id_contr_RV
          case default
            print*, "parseContrib_int_chan: invalid nR,nV", nR,nV
            stop
        end select
      case(2)
        select case(nV)
          case(0)
            contr%id_sub(1) = id_contr_RR
          case default
            print*, "parseContrib_int_chan: invalid nR,nV", nR,nV
            stop
        end select
      case default
        print*, "parseContrib_int_chan: invalid nR,nV", nR,nV
        stop
    end select

  end subroutine parseContrib_int_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> parse a string encoding a contribution (format: "RR[R,R]")
  !
  !> @param[in] scontr string encoding controbution
  !> @param[inout] contr Contrib_t that corresponds to scontr
  !-----------------------------------------------------------------------------
  subroutine parseContrib_str_chan(scontr, contr)
    character(len=*), intent(in) :: scontr
    type(Contrib_t), intent(inout) :: contr
    integer :: i,j,k,l
    integer :: nV, nR

    ! print*, "parseContrib_str_chan:", scontr

    !> we keep the `next` pointer in tact and don't nullify it

    !> allocate the array containing the info on sub-processes
    if (allocated(contr%id_sub))     deallocate(contr%id_sub)
    if (allocated(contr%region_sub)) deallocate(contr%region_sub)

    !> do we have sub-processes?
    i = index(scontr, '[')
    if (i==0) then
      !> no sub-processes
      contr%nsub = 1
      allocate(contr%id_sub(1))
      allocate(contr%region_sub(1))
      call parseSub_chan(scontr, contr%id_sub(1), contr%region_sub(1))
      contr%id_full = contr%id_sub(1)
    else
      ! print*, "full >", scontr(:i-1) , "<"
      !> no "a", "b" suffixes allowed. They belong into the sub-array
      contr%id_full = contrib_str_to_int_chan( scontr(:i-1) )
      !> needed to check for consistency
      nV = contrib_nV_chan(contr%id_full)
      nR = contrib_nR_chan(contr%id_full)
      !> at least one decay
      j = index(scontr, ']')
      if (j==0) then
        print*, "parseContrib_str_chan: couldn't find closing bracket"
        stop
      end if
      !> determine nsub
      contr%nsub = 1
      k = i+1
      do
        ! print*, ">", scontr(k:j-1), "<"
        l = index(scontr(k:j-1), ',')
        ! print*, "l = ", l
        if (l==0) exit
        contr%nsub = contr%nsub+1
        k = k+l
      end do
      !> fill the sub-process array
      allocate(contr%id_sub(1:contr%nsub))
      allocate(contr%region_sub(1:contr%nsub))
      k = i+1
      do i=1,contr%nsub
        l = index(scontr(k:j-1), ',')
        if (l==0) then
          ! print*, "#",scontr(k:j-1),"#"
          call parseSub_chan(scontr(k:j-1), contr%id_sub(i), contr%region_sub(i))
        else
          ! print*, "#",scontr(k:k+l-2),"#"
          call parseSub_chan(scontr(k:k+l-2), contr%id_sub(i), contr%region_sub(i))
          k = k+l
        end if
        !> for consistency
        nV = nV - contrib_nV_chan(contr%id_sub(i))
        nR = nR - contrib_nR_chan(contr%id_sub(i))
      end do
      !> is `full` compatible with the sub-channels?
      if ( nV/=0 .or. nR/=0) then
        print*, "parseContrib_str_chan: invalid input: ", scontr, " > ", nV,nR
        stop
      end if
    end if

    contains

    !-----------------------------------------------------------------------------
    !> parse a sub-contribution "LO", "V", "RRa", ... (no [x,y,...])
    !-----------------------------------------------------------------------------
    subroutine parseSub_chan(str, id, region)
      use StrHelper_mod
      character(len=*), intent(in) :: str
      integer, intent(out) :: id, region
      integer :: iupp
      !> get upper index
      iupp = len(str)
      !> see if the suffix is 'a' or 'b'
      select case ( to_upper_str(str(iupp:iupp)) )
      case('A')
        region = 1
        iupp  = iupp - 1
      case('B')
        region = 2
        iupp  = iupp - 1
      case default
        region = 0
      end select
      id = contrib_str_to_int_chan( str(:iupp) )
    end subroutine parseSub_chan

  end subroutine parseContrib_str_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the order we want (functions will return the *coefficient*)
  !
  !> @param[in] iorder the order we want (0=LO, 1=NLO,...)
  !> @param[in] flag optional value to switch the order on/off
  !-----------------------------------------------------------------------------
  subroutine set_order_inc_chan(iorder, flag)
    integer, intent(in) :: iorder
    logical, optional, intent(in) :: flag
    logical :: qactive
    if (iorder < 0 .or. iorder > 3) then
      print*, "set_order_inc_chan: order unsupported:", iorder
      stop
    end if
    qactive = .true.  ! default: switch on
    if (present(flag)) qactive = flag
    if (qactive) then
      order_inc = ibset(order_inc, iorder)
    else
      order_inc = ibclr(order_inc, iorder)
    end if

  end subroutine set_order_inc_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter function of the order that is currently set
  !
  !> @param[in] iorder the order we want (0=LO, 1=NLO,...)
  !-----------------------------------------------------------------------------
  logical function get_order_inc_chan(iorder)
    integer, intent(in) :: iorder
    get_order_inc_chan = btest(order_inc, iorder)
  end function get_order_inc_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on channels. We can use the channel number or shorthands:
  !>  * LO, V, R, VV, RV, RR
  !>  * with decays: RR[R,R], ...
  !>  * ALL = all channels
  !
  !> @param[in] chan string representing the chanel
  !-----------------------------------------------------------------------------
  subroutine enableChannel_chan(chan)
    use Process_mod
    use StrHelper_mod
    character(len=*), intent(in) :: chan
    integer :: i, ichan
    logical :: fail
    type(Contrib_t) :: contr, contr_in
    logical, dimension(:), allocatable :: qregions
    !> for parsing regions for integer channels
    integer :: iupp, regionint

    ! print*, "enableChannel_chan: >", chan, "<"

    !> inclusive calculations have their own order management
    if ( name_proc.eq.'HINC' .or.  &
       & name_proc.eq.'HTO2PINC'  .or.  &
       & name_proc.eq.'HTO2TAUINC'  .or.  &
       & name_proc.eq.'HTO2L1PINC'  .or.  &
       & name_proc.eq.'HTO2E2MUINC'  .or.  &
       & name_proc.eq.'HTO4EINC'  .or.  &
       & name_proc.eq.'HTO2L2NINC'  .or.  &
       & name_proc.eq.'HQT' .or.  &
       & name_proc.eq.'HTO2PQT'  .or.  &
       & name_proc.eq.'HTO2TAUQT'  .or.  &
       & name_proc.eq.'HTO2L1PQT'  .or.  &
       & name_proc.eq.'HTO2E2MUQT'  .or.  &
       & name_proc.eq.'HTO4EQT'  .or.  &
       & name_proc.eq.'HTO2L2NQT' .or.  &
       & name_proc.eq.'GGQT' .or.  &
       & name_proc.eq.'ZQT' .or.  &
       & name_proc.eq.'WPQT' .or.  &
       & name_proc.eq.'WMQT' .or.  &
       & name_proc.eq.'EPEMINC' ) then
      select case (to_upper_str(chan))
      case("LO")
        call set_order_inc_chan(0)
      case("NLO")
        call set_order_inc_chan(0)
        call set_order_inc_chan(1)
      case("NNLO")
        call set_order_inc_chan(0)
        call set_order_inc_chan(1)
        call set_order_inc_chan(2)
      case("N3LO")
        call set_order_inc_chan(0)
        call set_order_inc_chan(1)
        call set_order_inc_chan(2)
        call set_order_inc_chan(3)
      case("NLO_ONLY")
        call set_order_inc_chan(1)
      case("NNLO_ONLY")
        call set_order_inc_chan(2)
      case("N3LO_ONLY")
        call set_order_inc_chan(3)
      case default
        print*, "enableChannel_chan: invalid channel: ", chan
        print*, "process = ", name_proc
        stop
      end select
      !> register an inclusive contribution
      contr_in%id_full = id_contr_inc
      contr_in%nsub    = 1
      allocate(contr_in%id_sub(1:1))
      contr_in%id_sub = id_contr_inc
      allocate(contr_in%region_sub(1:1))
      contr_in%region_sub(1:1) = 0
      !
      call registerContrib_chan(contr_in)
      return
    endif

    !>----- channel by id
    ! for [prod,decay] there should always only be at most one sub-contribution
    ! that requires rotations... so leave syntax as is?!
    iupp = len(chan)
    !> see if the suffix is 'a' or 'b'
    select case ( to_upper_str(chan(iupp:iupp)) )
    case('A')
      regionint = 1
      iupp  = iupp - 1
    case('B')
      regionint = 2
      iupp  = iupp - 1
    case default
      regionint = -1
    end select
    ichan = to_int_str(chan(:iupp), fail)
    if (.not. fail) then
      !> bound check
      if ( ichan<lbound(qactive_chan,1) .or. ichan>ubound(qactive_chan,1) ) then
        print*, "enableChannel_chan: out of bound:", ichan
        stop
      end if
      !> check if contrib was initialised
      if (contrib_chan(ichan) == 0) then
        print*, "enableChannel_chan: unset contrib for ichan=", ichan
        stop
      end if
      !> switch the channel on
      qactive_chan(ichan) = .true.
      !> parse the contribution
      call parseContrib_chan(contrib_chan(ichan), contr)
      !> turn on regions if there are any
      qregions = hasRegions_chan(contr)
      if ( count(qregions) > 1 ) then
        print*, "enableChannel_chan: at most one sub-contribution with regions!"
        stop
      end if
      if ( count(qregions)==0 .and. regionint>0 ) then
        print*, "enableChannel_chan: channel ", ichan, "has no regions!"
        stop
      end if
      if ( count(qregions)==0 .and. default_region_chan>0 ) then
        print*, "enableChannel_chan: channel ", ichan, "has no regions! [default]"
        stop
      end if
      if ( any(qregions) ) then
        if ( regionint > 0 ) then
          !> default explicitly overridden
          where (qregions) contr%region_sub = regionint
          call registerContrib_chan(contr)
        else if (default_region_chan > 0) then
          !> default with only one region
          where (qregions) contr%region_sub = default_region_chan
          call registerContrib_chan(contr)
        else
          !> fall-back: register both regions
          where (qregions) contr%region_sub = 1
          call registerContrib_chan(contr)
          where (qregions) contr%region_sub = 2
          call registerContrib_chan(contr)
        end if
      else
        !> contribution has no regions
        contr%region_sub = 0
        call registerContrib_chan(contr)
      end if
      return
    end if

    !>----- channels by string
    do ichan=lbound(qactive_chan,1),ubound(qactive_chan,1)
      !> check if everything is set
      if (contrib_chan(ichan) == 0) then
        print*, "enableChannel_chan: unset contrib for ichan=", ichan
        stop
      end if
      !> parse the ichan-contribution
      call parseContrib_chan(contrib_chan(ichan), contr)
      qregions = hasRegions_chan(contr)
      if ( count(qregions) > 1 ) then
        print*, "enableChannel_chan: at most one sub-contribution with regions!"
        stop
      end if
      !> take care of wild-cards
      select case (to_upper_str( chan ))
        !>-----
        case ('ALL')
          qactive_chan(ichan) = .true.
          if ( any(qregions) ) then
            if (default_region_chan > 0) then
              !> default with only one region
              where (qregions) contr%region_sub = default_region_chan
              call registerContrib_chan(contr)
            else
              !> fall-back: register both regions
              where (qregions) contr%region_sub = 1
              call registerContrib_chan(contr)
              where (qregions) contr%region_sub = 2
              call registerContrib_chan(contr)
            end if
          else
            !> contribution has no regions
            contr%region_sub = 0
            call registerContrib_chan(contr)
          end if
        !>-----
        case default
          call parseContrib_chan(chan, contr_in)
          ! print*, ">>>>", 0,": ", contrib_to_str_chan(contr_in)
          !> compare and add if matches
          if (contr%nsub /= contr_in%nsub) then
            print*, "enableChannel_chan: mismatch in the number of sub-processes!", chan
            print*, "processes with corrections to production & decay must be specified like this:"
            print*, "  RV[R,V]  <-->  production=R, decay=V"
            stop
          end if
          !> check that the contribution matches
          if (.not.isSameContrib_chan(contr,contr_in)) cycle
          !> register contribution or die
          if ( all((.not.((contr_in%region_sub>0).eqv.qregions)).eqv.qregions) ) then
            !> regions unset: fall back to default value
            if (default_region_chan > 0) then
              !> default with only one region
              where (qregions) contr_in%region_sub = default_region_chan
              call registerContrib_chan(contr_in)
            else
              !> fall-back: register both regions
              where (qregions) contr_in%region_sub = 1
              call registerContrib_chan(contr_in)
              where (qregions) contr_in%region_sub = 2
              call registerContrib_chan(contr_in)
            end if
          else if ( all((contr_in%region_sub>0).eqv.qregions) ) then
            !> regions already set properly
            call registerContrib_chan(contr_in)
          else
            print*, "enableChannel_chan: set invalid regions: ", chan
            stop
          end if
          !> channel was successfully registered: turn it on
          qactive_chan(ichan) = .true.
      end select
      ! print*, qactive_chan(ichan), ">>>>", ichan,": ", contrib_to_str_chan(contr), " > regions: ", qregions
    end do

    return

  contains

    !-----------------------------------------------------------------------------
    !> check if two Contrib_t are the same (no check on `next` pointer)
    !-----------------------------------------------------------------------------
    logical function isSameContrib_chan(c1,c2)
      type(Contrib_t), intent(in) :: c1, c2
      integer :: i
      isSameContrib_chan = (c1%id_full == c2%id_full) .and.  &
                         & (c1%nsub == c2%nsub)
      if (.not. isSameContrib_chan) return
      do i=1,c1%nsub
        isSameContrib_chan = isSameContrib_chan .and.  &
                           & (c1%id_sub(i) == c2%id_sub(i))
        if (c1%region_sub(i)==-1 .or. c2%region_sub(i)==-1) cycle
        isSameContrib_chan = isSameContrib_chan .and.  &
                           & (c1%region_sub(i) == c2%region_sub(i))
      end do
    end function isSameContrib_chan

    !-----------------------------------------------------------------------------
    !> add a contribution to the global list
    !-----------------------------------------------------------------------------
    subroutine registerContrib_chan(c)
      type(Contrib_t), intent(in) :: c
      type(Contrib_t), pointer :: it, last
      integer :: i

      !> first look if we've already registered this contribution
      it => headContrib_p
      nullify(last)
      do while ( associated(it) )
        if (isSameContrib_chan(c,it)) exit
        !> loop
        last => it  ! save it so I can find the tail
        it   => it%next
      end do

      !> found an old entry: done
      if (associated(it)) return

      !> if we get here, it is null() and last is the tail-element
      allocate(it)
      nullify(it%next)
      if (.not.associated(headContrib_p)) headContrib_p => it
      if (associated(last)) last%next => it
      !> copy over entries
      it%id_full = c%id_full
      it%nsub = c%nsub
      allocate(it%id_sub(it%nsub))
      allocate(it%region_sub(it%nsub))
      do i=1,it%nsub
        it%id_sub(i)    = c%id_sub(i)
        it%region_sub(i) = c%region_sub(i)
      end do

    end subroutine registerContrib_chan

  end subroutine enableChannel_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the pointer to the current contribution
  !-----------------------------------------------------------------------------
  subroutine resetCurrentContrib_chan()
    nullify(currentContrib_p)
  end subroutine resetCurrentContrib_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> load the next contribution
  !
  !> @param[in] verbose  optional flag to control output
  !> @return boolean flag indicating success (.true. = loaded next contrib)
  !-----------------------------------------------------------------------------
  logical function loadNextContrib_chan(verbose)
    use Process_mod
    use ProcLib_mod
    use Run_mod, only : ilips_run
    integer, intent(in), optional :: verbose
    integer :: verbose_
    integer :: npar_, iregion_
    procedure(fillcommonNp_intfce) ::  &
      &  fillcommon2p, fillcommon2pdis, fillcommon2ph,  &
      &  fillcommon3p, fillcommon3pdis, fillcommon3pepem,  &
      &  fillcommon3ph, fillcommon3pvp,  &
      &  fillcommon4allp, fillcommon4allpG, fillcommon4ap,  &
      &  fillcommon4bp, fillcommon4pdis, fillcommon4pepem,  &
      &  fillcommon4ph, fillcommon4pvp, fillcommon4pvv,  &
      &  fillcommon5alln, fillcommon5allph, fillcommon5an,  &
      &  fillcommon5apdis, fillcommon5apepem, fillcommon5aph,  &
      &  fillcommon5bn, fillcommon5bpdis, fillcommon5bpepem,  &
      &  fillcommon5bph, fillcommon5pvp, fillcommon5pvv,  &
      &  fillcommon6allpvp, fillcommon6apvp, fillcommon6bpvp,  &
      &  fillcommon6pvv,  &
      &  fillcommon7allpvv, fillcommon7apvv, fillcommon7bpvv,  &
      &  fillcommonvfh
    procedure(phaseN_intfce) ::  &
      &  phase2dis, phase2h, phase2n,  &
      &  phase3dis, phase3epem, phase3GG, phase3GG_qT, phase3h,  &
      &  phase3h_qT, phase3n, phase3vp,  &
      &  phase4allG, phase4alln, phase4an, phase4bn, phase4dis,  &
      &  phase4epem, phase4GG, phase4GG_qT, phase4h, phase4h_qT,  &
      &  phase4vp, phase4vv,  &
      &  phase5adis, phase5aepem, phase5aGG, phase5aGG_qT, phase5ah,  &
      &  phase5ah_qT, phase5allGG, phase5allh, phase5alln, phase5an,  &
      &  phase5bdis, phase5bepem, phase5bGG, phase5bGG_qT, phase5bh,  &
      &  phase5bh_qT, phase5bn, phase5vp, phase5vv,  &
      &  phase6allvp, phase6avp, phase6bvp, phase6vv,  &
      &  phase7allvv, phase7avv, phase7bvv,  &
      &  phasevfh

    verbose_ = 1
    if (present(verbose))  verbose_ = verbose

    !> get the "next" element
    if (.not.associated(currentContrib_p)) then
      currentContrib_p => headContrib_p
    else
      currentContrib_p => currentContrib_p%next
    end if

    !> success?
    if (.not.associated(currentContrib_p)) then
      loadNextContrib_chan = .false.  ! reached end
      return
    else
      loadNextContrib_chan = .true.  ! success
    end if

    if (verbose_ > 0) then
      print*, ""
      print*, "------------------------------------------------"
      print*, "loadNextContrib_chan: loaded ", contrib_to_str_chan(currentContrib_p), " ..."
    end if

    !> update some global variables for multi-channel adaption
    call init_multi_chan()

    !> load all the contribution-specific stuff here...
    call currentContrib_chan(npar=npar_)
    call currentSubContrib_chan(1, region=iregion_)

    !> load the correct `sig` functions from the dynamic library
    select case (currentContrib_p%id_full)
    case(id_contr_inc)
      call load_sig_plib("???") ! do I even need to call it?
      !> call the variant where we pass the functions manually here
    case(id_contr_LO)
      call load_sig_plib("B")
    case(id_contr_R1)
      call load_sig_plib("R", "SNLO")
    case(id_contr_V1)
      call load_sig_plib("V", "TNLO")
    case(id_contr_RR)
      call load_sig_plib("RR", "S")
    case(id_contr_RV)
      call load_sig_plib("RV", "T")
    case(id_contr_VV)
      call load_sig_plib("VV", "U")
    end select

    !> link phase-space routines
    select case(name_proc)
    !>---
    case( "1JET", "2JET", "1JETFC", "2JETFC", "3JET", "4JET", "5JET",  &
        & "EPEMINC", "G_EW", "GG_EW" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (4)
        call load_phase_plib(phase2n, fillcommon2p)
      case (5)
        call load_phase_plib(phase3n, fillcommon3p)
      case (6)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase4alln, fillcommon4allp)
        case (1)
          call load_phase_plib(phase4an, fillcommon4ap)
        case (2)
          call load_phase_plib(phase4bn, fillcommon4bp)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case (7)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase5alln, fillcommon5alln)
        case (1)
          call load_phase_plib(phase5an, fillcommon5an)
        case (2)
          call load_phase_plib(phase5bn, fillcommon5bn)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case( "GJ", "GJJ", "GJJJ" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (4)
        call load_phase_plib(phase2n, fillcommon2p)
      case (5)
        call load_phase_plib(phase3n, fillcommon3p)
      case (6)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase4allG, fillcommon4allpG)
        case (1)
          call load_phase_plib(phase4an, fillcommon4ap)
        case (2)
          call load_phase_plib(phase4bn, fillcommon4bp)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case ( "GG", "GGJ", "GGJFC", "GGJJ", "GGJJFC", "GGJJJ", "GGJJJFC" )
      select case(npar_)
      case (4)
        call load_phase_plib(phase2n, fillcommon2p)
      case (5)
        select case(ilips_run)
        case (0)
          call load_phase_plib(phase3GG, fillcommon3ph)
        case (1)
          call load_phase_plib(phase3GG_qT, fillcommon3ph)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
          stop
        end select
      case (6)
        select case(ilips_run)
        case (0)
          call load_phase_plib(phase4GG, fillcommon4ph)
        case (1)
          call load_phase_plib(phase4GG_qT, fillcommon4ph)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
          stop
        end select
      case (7)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase5allGG, fillcommon5allph)
        case (1)
          select case(ilips_run)
          case (0)
            call load_phase_plib(phase5aGG, fillcommon5aph)
          case (1)
            call load_phase_plib(phase5aGG_qT, fillcommon5aph)
          case default
            print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
            stop
          end select
        case (2)
          select case(ilips_run)
          case (0)
            call load_phase_plib(phase5bGG, fillcommon5bph)
          case (1)
            call load_phase_plib(phase5bGG_qT, fillcommon5bph)
          case default
            print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
            stop
          end select
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case ( "GGG" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (5)
        call load_phase_plib(phase3GG, fillcommon3ph)
      case (6)
        call load_phase_plib(phase4GG, fillcommon4ph)
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case( "VFH", "VFHJ" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      call load_phase_plib(phasevfh, fillcommonvfh)
    !>---
    case( "DIS1J", "DIS", "DISWM1J", "DISWM", "DISWP1J", "DISWP",  &
        & "SFDIS", "SFDISWM", "SFDISWP" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (4)
        call load_phase_plib(phase2dis, fillcommon2pdis)
      case (5)
        call load_phase_plib(phase3dis, fillcommon3pdis)
      case (6)
        call load_phase_plib(phase4dis, fillcommon4pdis)
      case (7)
        select case(iregion_)
        case (1)
          call load_phase_plib(phase5adis, fillcommon5apdis)
        case (2)
          call load_phase_plib(phase5bdis, fillcommon5bpdis)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case( "EPEMJJ", "EPEM", "EPEM_H" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (4)
        call load_phase_plib(phase2n, fillcommon2p)
      case (5)
        call load_phase_plib(phase3epem, fillcommon3pepem)
      case (6)
        call load_phase_plib(phase4epem, fillcommon4pepem)
      case (7)
        select case(iregion_)
        case (1)
          call load_phase_plib(phase5aepem, fillcommon5apepem)
        case (2)
          call load_phase_plib(phase5bepem, fillcommon5bpepem)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case( "HTO4E",    "HTO4EJ",    "HTO4EJJ",    "HTO4EJJJ",     &
        & "HTO2E2MU", "HTO2E2MUJ", "HTO2E2MUJJ", "HTO2E2MUJJJ",  &
        & "HTO2L2N",  "HTO2L2NJ",  "HTO2L2NJJ",  "HTO2L2NJJJ" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (6)
        call load_phase_plib(phase4vv, fillcommon4pvv)
      case (7)
        call load_phase_plib(phase5vv, fillcommon5pvv)
      case (8)
        call load_phase_plib(phase6vv, fillcommon6pvv)
      case (9)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase7allvv, fillcommon7allpvv)
        case (1)
          call load_phase_plib(phase7avv, fillcommon7apvv)
        case (2)
          call load_phase_plib(phase7bvv, fillcommon7bpvv)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case( "HTO2L1P", "HTO2L1PJ", "HTO2L1PJJ", "HTO2L1PJJJ" )
      if (ilips_run /= 0) then
        print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
        stop
      end if
      select case(npar_)
      case (5)
        call load_phase_plib(phase3vp, fillcommon3pvp)
      case (6)
        call load_phase_plib(phase4vp, fillcommon4pvp)
      case (7)
        call load_phase_plib(phase5vp, fillcommon5pvp)
      case (8)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase6allvp, fillcommon6allpvp)
        case (1)
          call load_phase_plib(phase6avp, fillcommon6apvp)
        case (2)
          call load_phase_plib(phase6bvp, fillcommon6bpvp)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    case default
      select case(npar_)
      case (4)
        call load_phase_plib(phase2h, fillcommon2ph)
      case (5)
        select case(ilips_run)
        case (0)
          call load_phase_plib(phase3h, fillcommon3ph)
        case (1)
          call load_phase_plib(phase3h_qT, fillcommon3ph)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
          stop
        end select
      case (6)
        select case(ilips_run)
        case (0)
          call load_phase_plib(phase4h, fillcommon4ph)
        case (1)
          call load_phase_plib(phase4h_qT, fillcommon4ph)
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
          stop
        end select
      case (7)
        select case(iregion_)
        case (0)
          call load_phase_plib(phase5allh, fillcommon5allph)
        case (1)
          select case(ilips_run)
          case (0)
            call load_phase_plib(phase5ah, fillcommon5aph)
          case (1)
            call load_phase_plib(phase5ah_qT, fillcommon5aph)
          case default
            print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
            stop
          end select
        case (2)
          select case(ilips_run)
          case (0)
            call load_phase_plib(phase5bh, fillcommon5bph)
          case (1)
            call load_phase_plib(phase5bh_qT, fillcommon5bph)
          case default
            print*, "loadNextContrib_chan: ", name_proc, " has no ilips_run = ", ilips_run
            stop
          end select
        case default
          print*, "loadNextContrib_chan: ", name_proc, " has no iregion = ", iregion_
          stop
        end select
      case default
        print*, "loadNextContrib_chan: ", name_proc, " has no npar = ", npar_
        stop
      end select
    !>---
    end select

  end function loadNextContrib_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the current contribution
  !
  !> @param[out] id optional id_contr for the full contribution
  !> @param[out] nsub optional number of sub-processes
  !> @param[out] str optional string (see `contrib_to_str_chan`)
  !> @param[out] nV optional number of loops
  !> @param[out] nR optional number of real emissions
  !> @param[out] nO optional order = (nV+nR)
  !> @param[out] nas optional alphas power = nO + n_alphas_proc
  !> @param[out] npar optional total number of ext. particles
  !> @param[out] ndim optional integration dimension
  !> @param[out] nsplit optional number of IS splittings
  !> @param[out] nfrag optional number of FS fragmentations
  !> @param[out] runid optional string (<-> srunid for backwardscompatibility)
  !-----------------------------------------------------------------------------
  subroutine currentContrib_chan( id, nsub, str, nV, nR, nO, nas,  &
                                & npar, ndim, nsplit, nfrag, runid)
    use Process_mod
    integer, intent(out), optional :: id, nsub, nV, nR, nO, nas, npar, ndim
    integer, intent(out), optional :: nsplit, nfrag
    character(:), allocatable, intent(out), optional :: str, runid

    if (present(id))   id   = currentContrib_p%id_full
    if (present(nsub)) nsub = currentContrib_p%nsub
    if (present(str))  str  = contrib_to_str_chan(currentContrib_p)
    if (present(nV))   nV   = contrib_nV_chan(currentContrib_p%id_full)
    if (present(nR))   nR   = contrib_nR_chan(currentContrib_p%id_full)
    if (present(nO))   nO   = contrib_nO_chan(currentContrib_p%id_full)
    if (present(nas))  nas  = contrib_nO_chan(currentContrib_p%id_full)  &
                            & +n_alphas_proc
    if (present(npar)) npar =  n_fs_partons_proc  &
                            & +n_fs_photons_proc  &
                            & +n_fs_spectators_proc  &
                            & +contrib_nR_chan(currentContrib_p%id_full)  &
                            & +2  ! # IS particles
    if (present(ndim)) ndim = 3*( n_fs_partons_proc  &
                            &    +n_fs_photons_proc  &
                            &    +n_fs_spectators_proc  &
                            &    +contrib_nR_chan(currentContrib_p%id_full) )  &
                            & -4  &             ! E-p conservation
                            & -1  &             ! azimuthal symmetry
                            & +nPDFs_proc  &    ! PDF convolution
                            & +nsplit_chan()  & ! IS splittings
                            & +nfrag_chan()  &  ! FS fragmentation
                            & +ndim_extra_proc  ! extra dimensions
    if (present(nsplit)) nsplit = nsplit_chan()
    if (present(nfrag))  nfrag  = nfrag_chan()

    !> for backwards compatibility
    if (present(runid)) then
      select case (currentContrib_p%id_sub(1))
      case (id_contr_LO, id_contr_V1, id_contr_VV)
        runid = 'vBa'
      case (id_contr_R1, id_contr_RV)
        runid = 'vRa'
      case (id_contr_RR)
        if (currentContrib_p%region_sub(1)==2) then
          runid = 'RRb'
        else
          runid = 'RRa'
        end if
      case default
        runid = '###'
      end select
    end if

    return

    contains

      !> for the virtual PDF convolution we're not interested if the
      !> full process is virtual but if the *production* is.
      !> the production is always the 1st sub-process
      !> convolutions are also on for fully inclusive setups
      integer function nsplit_chan()
        if ( contrib_nV_chan( currentContrib_p%id_sub(1) ) /= 0 .or.  &
           & currentContrib_p%id_full == id_contr_inc ) then
          nsplit_chan = nPDFs_proc
        else
          nsplit_chan = 0
        end if
      end function nsplit_chan

      integer function nfrag_chan()
        use PhotonFragmentation_mod, only : isActive_frag, nz_frag
        use HadronFragmentation_mod, only : isActive_hfrag
        logical :: qvirt
        nfrag_chan = 0
        qvirt = contrib_nV_chan( currentContrib_p%id_full ) /= 0 .or.  &
              & currentContrib_p%id_full == id_contr_inc
        !> hadrons
        if ( isActive_hfrag() ) then
          if ( qvirt ) then
            nfrag_chan = 2
          else
            nfrag_chan = 1
          end if
        end if
        !> photons
        if ( isActive_frag() ) then
          if ( qvirt ) then
            nfrag_chan = 2*nz_frag
          else
            nfrag_chan = nz_frag
          end if
        end if
      end function nfrag_chan

  end subroutine currentContrib_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the `isub`-th sub-preocces of the current contribution
  !
  !> @param[in] isub index for the sub-process to inquire
  !> @param[out] nV optional number of loops
  !> @param[out] nR optional number of real emissions
  !> @param[out] nO optional order = (nV+nR)
  !> @param[out] region optional region flag (-1=unset, 0=none, 1=a, 2=b)
  !-----------------------------------------------------------------------------
  subroutine currentSubContrib_chan(isub, nV, nR, nO, region)
    integer, intent(in) :: isub
    integer, intent(out), optional :: nV, nR, nO, region

    if (isub<1 .or. isub>currentContrib_p%nsub) then
      print*, "currentSubContrib_chan: isub out of range:", isub
      stop
    end if

    if (present(nV)) nV = contrib_nV_chan(currentContrib_p%id_sub(isub))
    if (present(nR)) nR = contrib_nR_chan(currentContrib_p%id_sub(isub))
    if (present(nO)) nO = contrib_nO_chan(currentContrib_p%id_sub(isub))
    if (present(nO)) nO = contrib_nO_chan(currentContrib_p%id_sub(isub))
    if (present(region)) region = currentContrib_p%region_sub(isub)

  end subroutine currentSubContrib_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on a specific channel (not for efficiency!)
  !
  !> @param[in] ichan id for the channel
  !> @param[out] id optional id_contr for the full contribution
  !> @param[out] nsub optional number of sub-processes
  !> @param[out] str optional string (see `contrib_to_str_chan`)
  !> @param[out] nV optional number of loops
  !> @param[out] nR optional number of real emissions
  !> @param[out] nO optional order = (nV+nR)
  !> @param[out] npar optional total number of ext. particles
  !-----------------------------------------------------------------------------
  subroutine info_chan(ichan, id, nsub, str, nV, nR, nO, npar)
    use Process_mod
    integer, intent(in) :: ichan
    integer, intent(out), optional :: id, nsub, nV, nR, nO, npar
    character(:), allocatable, intent(out), optional :: str
    type(Contrib_t) :: contr

    if ( ichan<lbound(qactive_chan,1) .or. ichan>ubound(qactive_chan,1) ) then
      print*, "info_chan: ichan out of bounds:", ichan, lbound(qactive_chan,1), ubound(qactive_chan,1)
      stop
    end if

    !> parse the ichan-contribution
    call parseContrib_chan(contrib_chan(ichan), contr)


    if(present(id))   id   = contr%id_full
    if(present(nsub)) nsub = contr%nsub
    if(present(str))  str  = contrib_to_str_chan(contr)
    if(present(nV))   nV   = contrib_nV_chan(contr%id_full)
    if(present(nR))   nR   = contrib_nR_chan(contr%id_full)
    if(present(nO))   nO   = contrib_nO_chan(contr%id_full)
    if(present(npar)) npar =  n_fs_partons_proc  &
                           & +n_fs_photons_proc  &
                           & +n_fs_spectators_proc  &
                           & +contrib_nR_chan(contr%id_full)  &
                           & +2  ! # IS particles

  end subroutine info_chan





  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate multi-channel adaption
  !
  !> @param[in] active boolean flag to switch on/off
  !> @param[in] nselect setting of the adaption
  !-----------------------------------------------------------------------------
  subroutine setActive_multi_chan(active, nselect)
    logical, intent(in) :: active
    integer, intent(in), optional :: nselect

    qmulti_chan = active

    if (present(nselect)) then
      nselect_multi_chan = nselect

      if ( active .and. (nselect == 0) ) then
        print*, "setActive_multi_chan: nselect = 0 invalid"
        stop
      end if

      if ( active .and. (nselect == -1) ) then
        print*, "setActive_multi_chan: nselect = -1 invalid"
        print*, "no point switching on multi-channel if we have only one group?!"
        stop
      end if

    end if

  end subroutine setActive_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of multi-channel adaption
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_multi_chan()
    isActive_multi_chan = qmulti_chan
    !> separate setting from on/off flag
    ! isActive_multi_chan = (nselect_multi_chan /= 0)
  end function isActive_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire if we need channel groups
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function hasGroups_multi_chan()
    hasGroups_multi_chan = ( abs(nselect_multi_chan) > 1 )
  end function hasGroups_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise global variables for the active contribution
  !-----------------------------------------------------------------------------
  subroutine init_multi_chan()
    integer :: ichan, igroup

    !> optimise the lower & upper bounds of the dynamical array
    !> for the channels that are active in the current contribution
    !> also needed even if multi-channel adaption is off!
    nmulti_chan = 0
    ilower_multi_chan = -1
    iupper_multi_chan = -1
    do ichan=1,ntot_chan
      if (.not. qactive_chan(ichan)) cycle
      if (contrib_chan(ichan) /= currentContrib_p%id_full) cycle
      nmulti_chan = nmulti_chan + 1
      if ((ilower_multi_chan < 0).or.(ichan < ilower_multi_chan)) ilower_multi_chan = ichan
      if ((iupper_multi_chan < 0).or.(ichan > iupper_multi_chan)) iupper_multi_chan = ichan
    end do


    if (.not.isActive_multi_chan()) return


    !> set up the global arrays for multi-channel adaption
    if (allocated(weight_multi_chan)) deallocate(weight_multi_chan)
    if (allocated(bounds_multi_chan)) deallocate(bounds_multi_chan)

    !> initialise to equal weights
    allocate(weight_multi_chan(ilower_multi_chan:iupper_multi_chan))
    weight_multi_chan(:) = 0d0
    where ( qactive_chan(ilower_multi_chan:iupper_multi_chan) .and.  &
          & (contrib_chan(ilower_multi_chan:iupper_multi_chan) == currentContrib_p%id_full) )  &
      weight_multi_chan = 1d0/dble(nmulti_chan)
    ! print*, "weight_multi_chan", weight_multi_chan, nmulti_chan, sum(weight_multi_chan)

    allocate(bounds_multi_chan(ilower_multi_chan:iupper_multi_chan+1))
    call updateBounds_multi_chan()


    !> set up the global arrays for multi-channel groups
    if (hasGroups_multi_chan()) then
      print*, "init_multi_chan -- hasGroups_multi_chan"
      if (allocated(weight_group_chan)) deallocate(weight_group_chan)
      if (allocated(bounds_group_chan)) deallocate(bounds_group_chan)
      if (allocated(group_multi_chan))  deallocate(group_multi_chan)

      !> depermine the number of groups
      if (nselect_multi_chan > 0) then
        ngroup_chan = ceiling( dble(nmulti_chan) / dble(nselect_multi_chan) )
      else
        ngroup_chan = abs(nselect_multi_chan)
      end if

      if (ngroup_chan < 2) then
        print*, "WARNING: only one group?"
        ngroup_chan = 1
      end if
      if (ngroup_chan >= nmulti_chan) then
        print*, "WARNING: more groups than channels?"
        ngroup_chan = nmulti_chan
      end if

      print*, "init_multi_chan:"
      print*, "# of channels: ", nmulti_chan
      print*, "# of groups:   ", ngroup_chan

      allocate(group_multi_chan(ilower_multi_chan:iupper_multi_chan))
      allocate(weight_group_chan(ngroup_chan))
      allocate(bounds_group_chan(ngroup_chan+1))
      call updateGroup_multi_chan()

    end if

  end subroutine init_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the bounds array: [i,i+1] <-> channel i (has "width" = weight)
  !-----------------------------------------------------------------------------
  subroutine updateBounds_multi_chan()
    integer :: ichan
    double precision :: acc
    acc = 0d0
    do ichan = ilower_multi_chan, iupper_multi_chan
      bounds_multi_chan(ichan) = acc
      acc = acc + weight_multi_chan(ichan)
      if (ichan == iupper_multi_chan) then
        if (abs(acc-1d0) > 1d-8) then  !@todo: use min(weights)/10 as discriminator?
          print*, "updateBounds_multi_chan: sum of weights != 1:", acc
          stop
        end if
        bounds_multi_chan(ichan+1) = 1d0
      end if
    end do
    ! print*, "updateBounds_multi_chan: bounds_multi_chan", bounds_multi_chan
  end subroutine updateBounds_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the multi-channel group arrays (weights & boundaries)
  !-----------------------------------------------------------------------------
  subroutine updateGroup_multi_chan()
    integer :: i, max_i, ichan
    integer, dimension(:), allocatable :: sorted_ichan
    double precision :: max_wgt

    integer :: n_chan_per_group, igroup, istart, iedge, ilow, iupp
    double precision :: jump, max_jump

    double precision :: acc

    if (.not. hasGroups_multi_chan()) return

    !> sort from highest to lowest weights
    allocate(sorted_ichan(nmulti_chan))
    sorted_ichan(:) = 0
    do i = 1, nmulti_chan
      max_wgt = -1d0  ! weights are positive (will always be overwritten)
      max_i   = -1  ! the i-th largest weight
      do ichan = ilower_multi_chan, iupper_multi_chan
        if (.not. qactive_chan(ichan)) cycle
        if ( any(sorted_ichan == ichan) ) cycle  ! already assigned
        ! if (weight_multi_chan(ichan) == 0d0) cycle  ! purged
        if (weight_multi_chan(ichan) < max_wgt) cycle  ! not max
        max_wgt = weight_multi_chan(ichan)
        max_i   = ichan
      end do
      sorted_ichan(i) = max_i
    end do

    print*, "sorted_ichan", sorted_ichan
    ! read(*,*)
    !> assign channels to groups
    n_chan_per_group = floor( dble(nmulti_chan) / dble(ngroup_chan) )
    print*, "average channels per group", n_chan_per_group
    group_multi_chan(:) = 0
    weight_group_chan(:) = 0d0
    istart = 1
    do igroup = 1, ngroup_chan-1
      !> edge for equal sized groups
      iedge = igroup * n_chan_per_group
      !> floating edge: find the highest jump in the vicinity
      ilow = iedge - floor( dble(n_chan_per_group) / 2d0 )
      iupp = iedge + floor( dble(n_chan_per_group) / 2d0 )
      max_jump = 0d0
      do i = ilow, iupp-1
        ichan = sorted_ichan(i)
        if (weight_multi_chan(sorted_ichan(i)) == 0d0) cycle  ! purged
        !> safe because n_chan_per_group > 1
        if (weight_multi_chan(sorted_ichan(i+1)) /= 0d0) then
          jump = weight_multi_chan(sorted_ichan(i))  &
            &  / weight_multi_chan(sorted_ichan(i+1))
        else
          jump = 0d0
        end if
        ! print*, "jump:", i, ">", i+1, "=", jump
        if ( (jump > max_jump) .or.  &
          &  (weight_multi_chan(sorted_ichan(i+1)) == 0d0) ) then
          max_jump = jump
          iedge = i
        end if
      end do
      ! print*, "igroup", igroup, "[", ilow, iupp, "]", iedge, "/", nmulti_chan
      !> assign group id's and accumulate weights
      call updateGroup_multi_chan_assign_group(istart, iedge, igroup)
      istart = iedge + 1
    end do
    !> all remaining channels go into the last group
    call updateGroup_multi_chan_assign_group(istart, nmulti_chan, ngroup_chan)

    weight_group_chan = sqrt(weight_group_chan)
    weight_group_chan = weight_group_chan / sum(weight_group_chan)

    deallocate(sorted_ichan)

    print*, "group_multi_chan", group_multi_chan

    !> update group boundaries
    acc = 0d0
    do igroup = 1, ngroup_chan
      bounds_group_chan(igroup) = acc
      acc = acc + weight_group_chan(igroup)
      if (igroup == ngroup_chan) then
        if (abs(acc-1d0) > 1d-8) then  !@todo: use min(weights)/10 as discriminator?
          print*, "updateGroup_multi_chan: sum of weights != 1:", acc
          stop
        end if
        bounds_group_chan(igroup+1) = 1d0
      end if
    end do

    ! print*, "updateGroup_multi_chan: bounds_group_chan", bounds_group_chan

  contains

    !> assign group id's and accumulate weights
    subroutine updateGroup_multi_chan_assign_group(istart, iend, groupid)
      integer :: istart, iend, groupid, i, ichan
      do i = istart, iend
        ichan = sorted_ichan(i)
        group_multi_chan(ichan) = groupid
        weight_group_chan(groupid) = weight_group_chan(groupid)  &
          &                        + weight_multi_chan(ichan)**2
        print*, i, groupid, ">", ichan, weight_multi_chan(ichan)
      end do
    end subroutine updateGroup_multi_chan_assign_group

  end subroutine updateGroup_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise global variables for the active contribution
  !
  !> @param[in] ran random number between [0,1]
  !-----------------------------------------------------------------------------
  subroutine sample_multi_chan(ran)
    use Run_mod, only : currentSweep_run
    use Scales_mod, only : setMuR_scl
    double precision, intent(in) :: ran
    integer :: l,m,u
    !> for muR sampling
    integer :: irun
    double precision :: facR

    if (.not.isActive_multi_chan()) return

    if ( (ran < 0d0) .or. (ran > 1d0) ) then
      print*, "sample_multi_chan: random number out of bounds:", ran
      stop
    end if

    current_multi_chan = -1

    !> kill all channels
    qselect_chan(:) = .false.

    if (hasGroups_multi_chan()) then
      !> sample groups
      l = 1
      u = ngroup_chan+1
      do while ( l < u-1 )
        m = (l+u)/2
        ! print*, "new m", m, bounds_group_chan(m)
        if ( bounds_group_chan(m) >= ran ) then
          u = m
          ! print*, "new u", u
        else
          l = m
          ! print*, "new l", l
        end if
      end do
      current_multi_chan = l
      qselect_chan(ilower_multi_chan:iupper_multi_chan) =  &
        ( group_multi_chan(ilower_multi_chan:iupper_multi_chan) == current_multi_chan )
      ! print*, "pick group id", current_multi_chan, "[", ran, "]"
      ! print*, "group_multi_chan", group_multi_chan
      ! print*, group_multi_chan(ilower_multi_chan:iupper_multi_chan) == current_multi_chan
      ! print*, qselect_chan(ilower_multi_chan:iupper_multi_chan)
      ! read(*,*)
    else
      !> sample single channels
      l = ilower_multi_chan
      u = iupper_multi_chan+1
      do while ( l < u-1 )
        m = (l+u)/2
        ! print*, "new m", m, bounds_multi_chan(m)
        if ( bounds_multi_chan(m) >= ran ) then
          u = m
          ! print*, "new u", u
        else
          l = m
          ! print*, "new l", l
        end if
      end do
      current_multi_chan = l
      qselect_chan(current_multi_chan) = .true.
    end if

    !> if we're doing a warmup, also vary facR
    !> we leave muF fixed to still have the PDF caching advantage
    !> for the costly PDF evolution (we're concerned with log(muR/muF) terms)
    call currentSweep_run(type=irun)
    if (irun==1) then
      !> we use the location of ran in
      !>   [bounds(current_multi_chan), bounds(current_multi_chan+1)]
      !> to map to [0,1]
      if (hasGroups_multi_chan()) then
        facR = ( ran - bounds_group_chan(current_multi_chan) )  &
            &  / ( bounds_group_chan(current_multi_chan+1)  &
            &    - bounds_group_chan(current_multi_chan) )
      else
        facR = ( ran - bounds_multi_chan(current_multi_chan) )  &
          &  / ( bounds_multi_chan(current_multi_chan+1)  &
          &    - bounds_multi_chan(current_multi_chan) )
      end if
      ! print*, "facR[0,1]   = ", facR
      !> now map facR logarithmically between 1/2 and 2
      facR = 0.5d0 * (2d0/0.5d0)**facR
      ! print*, "facR[0.5,2] = ", facR
      ! read(*,*)
      call setMuR_scl(global_fac=facR)
    end if

  end subroutine sample_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> return the jacobian for the active channels (sum weights)
  !-----------------------------------------------------------------------------
  double precision function jac_multi_chan()
    jac_multi_chan = 1d0
    if (.not.isActive_multi_chan()) return
    if (hasGroups_multi_chan()) then
      jac_multi_chan = 1d0 / weight_group_chan(current_multi_chan)
    else
      jac_multi_chan = 1d0 / weight_multi_chan(current_multi_chan)
    end if
    !> this is not correct for groups when weight(group) != sum_i weight(i)
    ! jac_multi_chan = 1d0 / sum( weight_multi_chan,  &
    !                           & mask=qselect_chan(ilower_multi_chan:iupper_multi_chan) )
  end function jac_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the multi-channel arrrays (in parallel block)
  !-----------------------------------------------------------------------------
  subroutine initBuffer_multi_chan()
    integer :: ichan

    ! if ( allocated(qselect_chan) ) then
    !   print*, "initBuffer_multi_chan: already allocated?!"
    !   stop
    ! end if

    if (.not.allocated(qselect_chan)) allocate(qselect_chan(1:ntot_chan))
    qselect_chan(:) = .true.  ! all channels on (default: no multi-channel sampling)

    if (allocated(buffer_multi_chan)) deallocate(buffer_multi_chan)
    allocate(buffer_multi_chan(0:1,ilower_multi_chan:iupper_multi_chan))
    buffer_multi_chan(:,:) = 0d0

#ifdef CHECKACCUM
    buffer_qp = 0_qp
    buffer_dp = 0_dp
#endif

  end subroutine initBuffer_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy the dynamical multi-channel arrays
  !-----------------------------------------------------------------------------
  subroutine destroyBuffer_multi_chan()

    if (allocated(qselect_chan)) deallocate(qselect_chan)
    if (allocated(buffer_multi_chan)) deallocate(buffer_multi_chan)

  end subroutine destroyBuffer_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> accumulate values in the buffer.
  !
  !> @param[in] wgt weight (same as returned by bino) for the current channel
  !-----------------------------------------------------------------------------
  subroutine writeBuffer_multi_chan(wgt)
    use FPA_mod
    double precision, intent(in) :: wgt
    double precision :: acc_sum, err

    !> always fill the channel breakdown?
    ! if (.not.isActive_multi_chan()) return

    if (isnan(wgt)) then
      print*, "writeBuffer_multi_chan: encountered NaN weight!"
      call print_xvegas
      return
    end if

    ! if (wgt /= 0d0) print*, "writeBuffer_multi_chan:", wgt

    !> kahan summation
    ! buffer_multi_chan(current_chan) = buffer_multi_chan(current_chan) + wgt
    call twoSum(buffer_multi_chan(1,current_chan), wgt, acc_sum, err)
    buffer_multi_chan(1,current_chan) = acc_sum
    buffer_multi_chan(0,current_chan) = buffer_multi_chan(0,current_chan) + err

#ifdef CHECKACCUM
    buffer_qp = buffer_qp + wgt
    buffer_dp = buffer_dp + wgt
#endif

  end subroutine writeBuffer_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> flush the buffer into the channel accumulator.
  !> note that any global-global reweighting is irrelevant
  !
  !> @param[in] wgt the Vegas weight
  !> @param[inout] sig value of the integrand: overriden by channel sum
  !>               (more stabl to use buffer sum <-> compensated summation)
  !-----------------------------------------------------------------------------
  subroutine flushBuffer_multi_chan(wgt, sig)
    use FPA_mod
    double precision, intent(in)    :: wgt
    double precision, intent(inout) :: sig
    integer :: ichan
    double precision :: acc_sum, err
    !> check vs. the vegas weight
    double precision :: vegas_wsig, check_wsig, delta_wsig_abs, delta_wsig_rel
    double precision, parameter :: threshold_rel = 1d-5
    double precision, parameter :: threshold_abs = 1d-12 / threshold_rel
    ! double precision, parameter :: threshold_rel = 1d-13
    ! double precision, parameter :: threshold_abs = 1d-10

    !> reject event: clear buffer and die
    if (ireject_event_chan /= 0) then
      if (ireject_event_chan > 0)  &
      &  print*, "flushBuffer_multi_chan: event rejection triggered by channel id = ", ireject_event_chan
      !> clear buffer and zero return to Vegas
      buffer_multi_chan(:,:) = 0d0
      sig = 0d0
      !> reset (also done in sig.f but better be safe)
      ireject_event_chan = 0
      return
    end if

    !> do a check against Vegas to auto-detect issues like missing "subfac"s
    vegas_wsig     = wgt * sig
    check_wsig     = sum(buffer_multi_chan)  ! tot & central scale  (chan sum + comp. errors)
    delta_wsig_abs = abs( check_wsig - vegas_wsig )
    delta_wsig_rel = delta_wsig_abs / abs( check_wsig + vegas_wsig )
    ! print*, "flushBuffer_multi_chan: Vegas vs. bino weights:"
    ! print*, "    Vegas:   ", vegas_wsig
    ! print*, "    bino:    ", check_wsig
    ! read(*,*)
    if ( delta_wsig_abs > threshold_abs .and.  &
       & delta_wsig_rel > threshold_rel ) then
      print*, "flushBuffer_multi_chan: Vegas vs. bino weights don't match!"
      print*, "    Vegas:   ", vegas_wsig
      print*, "    bino:    ", check_wsig
#ifdef CHECKACCUM
      print*, "    quad:    ", buffer_qp, check_wsig/buffer_qp
      print*, "    dble:    ", buffer_dp, check_wsig/buffer_dp
#endif
      print*, "    reldiff: ", abs(vegas_wsig-check_wsig)/abs(vegas_wsig+check_wsig)
      ! if (abs(vegas_wsig)+abs(check_wsig) /= 0d0) read(*,*)
      ! stop
    end if
    !> override the `sig` value as determined from the buffer (more accurate)
    sig = check_wsig / wgt

!$OMP CRITICAL(flushbuffer_multi)
    do ichan = ilower_multi_chan, iupper_multi_chan
      !> note to myself: checking for zero in the buffer is not correct!
      if (.not.qselect_chan(ichan)) cycle
      buffer_multi_chan(1,ichan) = buffer_multi_chan(1,ichan)  &
                               & + buffer_multi_chan(0,ichan)  ! add the error accumulator
      buffer_multi_chan(0,ichan) = 0d0
      count_multi_chan(ichan) = count_multi_chan(ichan) + 1
      ! accum_multi_chan(1,ichan) = accum_multi_chan(1,ichan)  &
      !                           & + buffer_multi_chan(ichan)
      !> use Kahan summation:
      call twoSum(accum_multi_chan(1,ichan), buffer_multi_chan(1,ichan), acc_sum, err)
      accum_multi_chan(1,ichan) = acc_sum
      accum_multi_chan(0,ichan) = accum_multi_chan(0,ichan) + err
      !> no kahan for squares: all positive numbers
      accum_multi_chan(2,ichan) = accum_multi_chan(2,ichan)  &
                              & + buffer_multi_chan(1,ichan)**2
    end do
!$OMP END CRITICAL(flushbuffer_multi)

    buffer_multi_chan(:,:) = 0d0

#ifdef CHECKACCUM
    buffer_qp = 0_qp
    buffer_dp = 0_dp
#endif

  end subroutine flushBuffer_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> hook routine to read out accumulators
  !
  !> @param[out] ilow lower index of arrays
  !> @param[out] iupp upper index of arrays
  !> @param[out] count_arr array(ilow:iupp) counting the hits of channels
  !> @param[out] accum_arr array(ilow:iupp) accumulator for the channels
  !-----------------------------------------------------------------------------
  subroutine read_multi_chan(full_array)
    double precision, intent(out), dimension(1:3,ilower_multi_chan:ilower_multi_chan) :: full_array
    integer :: i

    do i = ilower_multi_chan, iupper_multi_chan
      !> socket read-out induces compensated summation
      full_array(1,i) = accum_multi_chan(1,i) + accum_multi_chan(0,i)
      full_array(2,i) = accum_multi_chan(2,i)
      full_array(3,i) = dble( count_multi_chan(i) )
    end do

  end subroutine read_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> hook routine to override internal accumulators
  !
  !> @param[out] count_arr array(ilow:iupp) counting the hits of channels
  !> @param[out] accum_arr array(ilow:iupp) accumulator for the channels
  !-----------------------------------------------------------------------------
  subroutine write_multi_chan(full_array)
    double precision, intent(in), dimension(1:3,ilower_multi_chan:ilower_multi_chan) :: full_array
    integer :: i

    do i = ilower_multi_chan, iupper_multi_chan
      accum_multi_chan(0,i) = 0d0  ! no errors to compensate
      accum_multi_chan(1,i) = full_array(1,i)
      accum_multi_chan(2,i) = full_array(2,i)
      count_multi_chan(i) = int(full_array(3,i))
    end do

  end subroutine write_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> write the current channel weights into a file
  !
  !> @param[in] file name of the file
  !-----------------------------------------------------------------------------
  subroutine writeToFile_multi_chan(file)
    character(len=*), intent(in) :: file
    integer :: ichan
    logical :: qexists
    print*, "writeToFile_multi_chan: ", trim(file)
    open(unit = 11, file = trim(file), status = 'unknown', action = 'write')
    do ichan = ilower_multi_chan, iupper_multi_chan
      write(11,file_fmt) ichan, weight_multi_chan(ichan)
    enddo
    close(11)
    !> also write out numeical values for plotting
    inquire(file=trim(file)//".txt", exist=qexists)
    if (qexists) then
      open(unit=11, file=trim(file)//".txt", status='old', action='write', position='Append')
    else
      open(unit=11, file=trim(file)//".txt", status='unknown', action='write')
      ! write(11,*,advance='no') "#labels: iteration[1] "
      write(11,*) "#labels: ", (ichan,ichan=ilower_multi_chan,iupper_multi_chan)
    end if
    write(11,*) weight_multi_chan
    close(11)
  end subroutine writeToFile_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> read a weight table from a file
  !
  !> @param[in] file name of the file
  !-----------------------------------------------------------------------------
  subroutine readFromFile_multi_chan(file)
    character(len=*), intent(in) :: file
    integer :: ichan, ichk
    logical :: qexists
    inquire(file=trim(file), exist=qexists)
    if (.not.qexists) return
    print*, "readFromFile_multi_chan: reading in: ", trim(file)
    weight_multi_chan = 0d0
    open(unit=11, file=trim(file), status='unknown')
    do ichan = ilower_multi_chan, iupper_multi_chan
      read(11,file_fmt) ichk, weight_multi_chan(ichan)
      if (ichan /= ichk) then
        print*, "readFromFile_multi_chan: channel mismatch:", ichan, ichk
        stop
      end if
    enddo
    close(11)
    !> got the weights: update the boundaries accordingly
    call updateBounds_multi_chan()
    if (hasGroups_multi_chan()) call updateGroup_multi_chan()
  end subroutine readFromFile_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> adapt the weights according to the accumulated numbers
  !> the optimal new weights are determined by:
  !>   weight[i]_new ~ sqrt(weight[i]_old) * error[i]
  !
  !> @param[in] purge optional flag to purge zero weights
  !-----------------------------------------------------------------------------
  subroutine adapt_multi_chan(purge)
    logical, optional :: purge
    logical :: purge_zero
    integer :: ichan, ndim
    double precision :: min_wgt, rat
    double precision, dimension(:), allocatable :: new_wgt

    if (.not.isActive_multi_chan()) return

    purge_zero = .false.  ! default
    if (present(purge)) purge_zero = purge

    call currentContrib_chan(ndim=ndim)

    allocate(new_wgt(ilower_multi_chan:iupper_multi_chan))
    new_wgt(:) = 0d0

    do ichan = ilower_multi_chan, iupper_multi_chan
      if (.not. qactive_chan(ichan)) cycle
      accum_multi_chan(1,ichan) = accum_multi_chan(1,ichan)  &
                              & + accum_multi_chan(0,ichan)  ! add the error accumulator
      accum_multi_chan(0,ichan) = 0d0
      !> this is the MC error:
      if (count_multi_chan(ichan) > 2) then
        new_wgt(ichan) = sqrt( abs( accum_multi_chan(2,ichan)*dble(count_multi_chan(ichan))  &
               &                  - accum_multi_chan(1,ichan)**2 )   &
               &             / dble(count_multi_chan(ichan)-1) )
      else
        !> zero for no hit, 100% error for a single hit
        new_wgt(ichan) = accum_multi_chan(1,ichan)
      end if
      !> multiply by the sqrt of the # hits (~old weight)
      new_wgt(ichan) = new_wgt(ichan) * sqrt(dble(count_multi_chan(ichan)))
    end do
    !> normalise new distribution: this is the "optimal" new weight
    new_wgt = new_wgt / sum(new_wgt)
    !> safety features
    min_wgt = 1d0/dble(nmulti_chan) * 1d-2  ! channels should not be suppressed below ~1/100 of the uniform case
    do ichan = ilower_multi_chan, iupper_multi_chan
      if (.not. qactive_chan(ichan)) cycle
      if (weight_multi_chan(ichan) == 0d0) cycle  ! no zombies: once dead -> stay dead
      if (purge_zero .and. new_wgt(ichan) == 0d0 .and. count_multi_chan(ichan) > max(100,int(1.5d0**ndim)) ) cycle
      if (new_wgt(ichan) < min_wgt) then
        !> enforce a min (max?)
        new_wgt(ichan) = min_wgt
      else
        !> damping against outliers (both directions [+/-] will induce new_wgt >> weight_multi_chan)
        !rat = new_wgt(ichan) / weight_multi_chan(ichan)
        !> power: 0=no damping, 1=no adaption
        new_wgt(ichan) = new_wgt(ichan) * (weight_multi_chan(ichan)/new_wgt(ichan))**0.75
      end if
    end do
    !> re-normalise
    new_wgt = new_wgt / sum(new_wgt)
    !> copy over
    weight_multi_chan(:) = new_wgt(:)

    deallocate(new_wgt)

    ! if (purge_zero) then
    !   print*, "adapt_multi_chan: purging..."
    !   do ichan = ilower_multi_chan, iupper_multi_chan
    !     if (.not. qactive_chan(ichan)) cycle
    !     if (weight_multi_chan(ichan)==0d0) then
    !       print*, "killing ichan = ", ichan
    !       !> this makes the channel weight file incompatible with future runs...
    !       ! qactive_chan(ichan) = .false.
    !       nmulti_chan = nmulti_chan - 1
    !     end if
    !   end do
    ! end if

    !> done adjusting weights: update the boundaries
    call updateBounds_multi_chan()

    if (hasGroups_multi_chan()) call updateGroup_multi_chan()


  end subroutine adapt_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the channel accumulators
  !-----------------------------------------------------------------------------
  subroutine resetAccum_multi_chan()

    accum_multi_chan(:,:) = 0d0
    count_multi_chan(:) = 0

  end subroutine resetAccum_multi_chan


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest a full channel breakdown
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_multi_chan(unit)
    integer, optional, intent(in) :: unit
    integer :: iout, ichan
    character(len=:), allocatable :: chan_str, me_str, sub_str
    double precision :: fac, sum_wgt, val_ch, err_ch, rat_ch
    integer :: Ntot
    double precision :: sumf, sumf2
    double precision :: val_nve, err_nve

    double precision :: dv2g
    integer :: npg
    common /vegasnumcalls/dv2g,npg

    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_label =  &
      & '(1x,"*",1x,"  id",T10,"partons",T35,"Matrix Element",T65,"Subtraction",T103,"*")'
    ! character(len=*), parameter :: fmt_str   =  &
    !   & '(1x,"*",1x,I4,T10,A,T35,A,T65,A,T103,"*")'
    character(len=*), parameter :: fmt_wgt   =  &
      & '(1x,"*",1x,I0,T10,A,T35,ES20.8," +/- ",ES20.8," (",F9.3,"%)",I0,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_label)
    write(iout,fmt_blank)

    !> not needed and incorrect when not all channels get ncall hits
    ! fac = dble(npg)
    sum_wgt = sum(accum_multi_chan(1,:))

    Ntot = 0
    sumf = 0d0
    sumf2 = 0d0
    val_nve = 0d0
    err_nve = 0d0

    print*, "digest_multi_chan", ilower_multi_chan, iupper_multi_chan, ">", npg

    do ichan = ilower_multi_chan, iupper_multi_chan
      if (.not. qactive_chan(ichan)) cycle
      call get_chan(ichan, chan_str, me_str, sub_str)
      val_ch = accum_multi_chan(1,ichan)  ! / dble(count_multi_chan(ichan)) * fac
      if (count_multi_chan(ichan) > 1) then
        err_ch = sqrt( ( accum_multi_chan(2,ichan)*dble(count_multi_chan(ichan))  &
               &       - accum_multi_chan(1,ichan)**2 )   &
               &     / dble(count_multi_chan(ichan)-1) )
      else
        err_ch = accum_multi_chan(1,ichan)
      end if

      ! err_ch = ( accum_multi_chan(2,ichan)*dble(npg)  &
      !        & - accum_multi_chan(1,ichan)**2 ) * dv2g
      ! err_ch = sqrt(err_ch)

      rat_ch = accum_multi_chan(1,ichan) / sum_wgt * 100d0

      write(iout,fmt_wgt) ichan, trim(chan_str), val_ch, err_ch, rat_ch, count_multi_chan(ichan)

      if (count_multi_chan(ichan) > 0) then
        Ntot = Ntot + count_multi_chan(ichan)
        sumf = sumf + val_ch
        sumf2 = sumf2 + (err_ch**2*dble(count_multi_chan(ichan)-1) + val_ch**2)/dble(count_multi_chan(ichan))
      end if

      val_nve = val_nve + val_ch
      err_nve = err_nve + err_ch**2

    end do

    print*, "full  channel:", sumf,    "+/-", sqrt(sumf2*dble(Ntot)-sumf**2)/sqrt(Ntot-1d0)
    print*, "naive channel:", val_nve, "+/-", sqrt(err_nve)

  end subroutine digest_multi_chan


end module Channels_mod
