!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Check routines [external members: `_chk`]
!
! DESCRIPTION:
!> Collection of check & debug routines
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Checks_mod
!-------------------------------------------------------------------------------
#ifdef USE_OMP
  use omp_lib
#endif
  implicit none
  private


  public :: setActive_point_chk, isActive_point_chk
  public :: point_chk


  !> point check: # of PS points to dump into a *.dat file
  !>   <= 0: inactive
  !>   >= 1: number of PS points
  integer :: n_phasespace_points_chk = 0  ! default: off
  logical :: print_invariants_chk = .false.
  logical :: print_momenta_chk    = .false.
  logical :: print_obs_chk        = .false.


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> activate the point checks
  !
  !> @param[in] npoints number of phase-space points to write out
  !> @param[in] qinv optional flag to toggle printing of invariants
  !> @param[in] qmom optional flag to toggle printing of momenta
  !> @param[in] qobs optional flag to toggle printing of observables
  !-----------------------------------------------------------------------------
  subroutine setActive_point_chk(npoints, qinv, qmom, qobs)
    !> @todo: make it with an bollean flag (on/off) and an optional npoints integer argument
    implicit none
    integer, intent(in) :: npoints
    logical, intent(in), optional :: qinv, qmom, qobs

#ifdef USE_OMP
    !> if openMP is active, make sure there's only one thread
    if (omp_get_max_threads() > 1) then
      print*, "setActive_point_chk: running point checks => re-setting #threads to ONE"
      call omp_set_num_threads(1)
    end if
#endif

    n_phasespace_points_chk = npoints
    if ( present(qinv) ) print_invariants_chk = qinv
    if ( present(qmom) ) print_momenta_chk    = qmom
    if ( present(qobs) ) print_obs_chk        = qobs

  end subroutine setActive_point_chk


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if point checks are active
  !-----------------------------------------------------------------------------
  logical function isActive_point_chk()
    isActive_point_chk = (n_phasespace_points_chk > 0)
  end function isActive_point_chk


  !-----------------------------------------------------------------------------
  !> @brief
  !> activate the point checks
  !
  !> @param[in] npar
  !> @param[in] rnorm
  !> @param[in] x1
  !> @param[in] x2
  !> @param[in] wgt
  !> @param[external] sigR
  !> @param[external] sigS
  !> @param[external] sigV
  !> @param[external] sigT
  !> @param[external] sigVV
  !> @param[external] sigU
  !-----------------------------------------------------------------------------
  subroutine point_chk(npar,rnorm,x1,x2,wgt, sigMAT,sigSUB)
    use Channels_mod
    use KinData_mod !, only: kin
    use Observables_mod, only: printAll_obs
    use Mapping_mod
    implicit none
    integer, intent(in) :: npar
    double precision, intent(in) :: rnorm,x1,x2,wgt
    double precision, external :: sigMAT,sigSUB
    !> counter for the # of PS points
    integer, save :: counter = 0
    !> output
    integer :: iout = 6  ! default: stdout
    integer :: i, ii, jj
    logical :: qnonzero
    double precision :: dMAT, dSUB
    !> vegas random numbers
    double precision xvegas
    common /vdata/xvegas(20)
!$omp threadprivate(/vdata/)
    !> info on the current contribution
    integer :: n, ndim, nv

    !> inactive: return
    !> @todo: use "-1" (or neg. values) to write out until termination?
    if (n_phasespace_points_chk <= 0) return

    call currentContrib_chan(npar=n, ndim=ndim, nsplit=nv)

    !> @todo: set up filenmae and all that

    !> finished the output: die
    if (counter >= n_phasespace_points_chk) then
      print*, "done writing PS points..."
      stop
    end if

    !> use file output
    iout = 501
    if (counter == 0) then
      open (unit=iout,file="point_chk.dat",status="unknown",action="write")
      !> header: channel id, MAT, SUB
      write(iout,'("# NNLOJET point tests")')
      !> add iproc --> process string as comment lines?
    else
      open (unit=iout,file="point_chk.dat",status="old",position="Append",action="write")
    end if

    !> kill all chanels
    !> @todo will be broken when used with multi-channel
    qselect_chan = .false.

    call flushCache_map()  ! clear the cache
    qnonzero = .false.
    do i=1,ntot_chan
      if (.not.qactive_chan(i)) cycle
      dMAT  = 0d0
      dSUB  = 0d0
      qselect_chan(i) = .true.  ! switch on
      dMAT = sigMAT(rnorm,x1,x2,wgt)
      dSUB = sigSUB(rnorm,x1,x2,wgt)
      !> only write out of something is non-zero
      if (dMAT/=0d0 .or. dSUB/=0d0) then
        qnonzero = .true.
        write(iout,'(I0,T10,ES20.10,ES20.10)') i, dMAT, dSUB
      end if
      qselect_chan(i) = .false.  ! switch back off
    end do


    if (qnonzero) then
      counter = counter + 1
      !> evaluate highest-momentum set ME to get the right jets, etc...
      dMAT = sigMAT(rnorm,x1,x2,wgt)
      write(iout, '("# END: ",I0)') counter
      do i=1,ndim
        write(iout, '("# x(",I0,") = ",E25.12)') i, xvegas(i)
      end do
      !> invariants
      if (print_invariants_chk) then
        write(iout, '("# Invariants")')
        do ii=1,npar
        do jj=ii+1,npar
          write(iout, '("# s",I0,"(",I0,",",I0,") = ",E25.12)')  &
            &  npar, ii, jj, kin(npar)%s(ii,jj)
        enddo
        enddo
      end if
      !> momenta
      if (print_momenta_chk) then
        write(iout, '("# Momenta")')
        do ii=1,npar
        do jj=1,4
          write(iout, '("# p",I0,"(",I0,",",I0,") = ",E25.12)')   &
            &  npar, jj, ii, kin(npar)%p(jj,ii)
        enddo
        enddo
        write(iout, '("# Momentum fractions")')
        write(iout, '("# x1_kin    = ",E25.12)') x1_kin
        write(iout, '("# x2_kin    = ",E25.12)') x2_kin
        write(iout, '("# xreg1_kin = ",E25.12)') xreg1_kin
        write(iout, '("# xreg2_kin = ",E25.12)') xreg2_kin
        write(iout, '("# Jets")')
        do ii=1,kin(npar)%njets
        do jj=1,4
          write(iout,'("# pj",I0,"(",I0,") = ",E25.12)')  ii, jj, kin(npar)%pjets(jj,ii)
        enddo
        enddo
      end if
      !> observables
      if (print_obs_chk) then
        write(iout, '("# Observables")')
        call printAll_obs(npar, skipCache=.true., unit=iout)
      end if
    end if

    !> switch back on
    qselect_chan = .false.

    close(iout)

  end subroutine point_chk


end module Checks_mod


