!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: DIS [external members: `_dis`]
!
! DESCRIPTION:
!> Module storing and retrieving all DIS kinematics information
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module DIS_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  use KinData_mod
  implicit none
  private

  public :: boost_to_gamnuc
  public :: boost_to_hera_dis
  public :: boost_to_breit_dis
  public :: eval_trans_Breit_to_gamma_proton_cm
  public :: setetaep, setEs, getEe, getEp
  public :: setDISEVTSHAPE, getDISEVTSHAPE
  public :: setHERA_ORDERETA, getHERA_ORDERETA
  public :: setEhfrac, getEhfrac
  public :: dis
  public :: setxx1, PSold, setPSDIS, setfixT, fixT, setPol
  public :: isSetBreitFrame, useBreitFrame
  public :: boostPGAM
  public :: Pol, isSetPol
  public :: use_EPRC, set_EPRC

  real(kind=wp) :: etaep
  !> electron proton energies in lab frame
  real(kind=wp) :: Ee, Ep
  logical :: PSold = .false.
  logical :: fixT = .false.
  !> are we overwriting the default choice for Breit?
  logical :: isSetBreitFrame = .false.
  !> if yes, what is our choice?
  logical :: useBreitFrame
  real(kind=wp) :: Ehfrac
  logical :: DISEVTSHAPE = .false.
  logical :: HERA_OrderEta = .false.
  logical :: boostPGAM = .false.
  logical :: isSetPol=.false.
  double precision :: Pol = 0d0
  logical :: is_EPRC = .false.

  type DIS_t
    !> This is the boost to the breit frame, the matrix size cannot change
    real(kind=wp), dimension(4, 4) :: breitT
    !> Transformation from BREIT to gamma parton frame
    real(kind=wp), dimension(4, 4) :: BREITgammaprT
    !> jets in HERA frame without any cuts except jet cuts on jetalgorithm applied
    real(kind=wp), dimension(4, 4) :: raw_HERAjets
    !> jets in HERA frame  that survive the cuts
    real(kind=wp), dimension(4, 4) :: HERAjets
    !> jets in gamma proton frame that survive the cuts
    real(kind=wp), dimension(4, 4) :: HERApgammajets
    !> +2 jets in forward jet analysis
    real(kind=wp), dimension(4, 2) :: HERAplus2jets
    integer :: rawjets
    real(kind=wp) :: xx1
  end type DIS_t

  type(DIS_t) dis
!$OMP THREADPRIVATE(dis)

  !> these are "manual" observables of the DIS process
  integer, public :: id_Q2_dis = 0 ! , id_sqrtQ2_dis = 0, id_Ee_dis = 0,  &
  ! id_thetae_dis = 0, id_x_dis = 0, id_y_dis = 0,  &
  ! id_W2_dis = 0, id_W_dis = 0, id_cosgammah_dis = 0,  &
  ! id_nu_dis = 0, id_resrho_dis = 0, id_y_disg = 0

  real(kind=wp), public :: value_q2_dis, value_q_dis, &
                           value_W2_dis, value_W_dis, &
                           value_x_dis, value_Ee_dis, &
                           value_y_dis, value_y_disg, &
                           value_disnu_dis, value_thetae_dis, &
                           value_cosgammah_dis, value_resrho_dis
!$OMP THREADPRIVATE(value_q2_dis, value_q_dis)
!$OMP THREADPRIVATE(value_W2_dis, value_W_dis)
!$OMP THREADPRIVATE(value_x_dis, value_Ee_dis)
!$OMP THREADPRIVATE(value_y_dis, value_y_disg)
!$OMP THREADPRIVATE(value_disnu_dis, value_thetae_dis)
!$OMP THREADPRIVATE(value_cosgammah_dis, value_resrho_dis)

contains

  !-----------------------------------------------------------------------------
  !> @brief
  !> transforms momenta from hera to breit frame + stores transformation matrix
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine boost_to_breit_dis(npar, ifail)
    implicit none
    integer, intent(inout) :: ifail
    integer, intent(in) :: npar
    real(kind=wp) :: pa1(1:4, 1:7), pa2(1:4, 1:7), q(1:4), qr(1:4), pq(1:4), qb(1:4)
    real(kind=wp) :: rmat(1:4, 1:4), bmat1(1:4, 1:4), bmat2(1:4, 1:4), fmat(4, 4)
    real(kind=wp) :: eta, cth, sth
    integer :: i, j, k, l, m

    do i = 1, 4
      do j = 1, npar
        pa1(i, j) = kin(npar)%p(i, j)
      end do
    end do

    do i = 1, 4
      q(i) = pa1(i, 2) - pa1(i, 3)
      pq(i) = pa1(i, 1) + q(i)
    end do
    call boostrestB(pq, bmat1, ifail)
    do i = 1, 4
      qb(i) = 0._wp
      do j = 1, 4
        qb(i) = qb(i) + bmat1(i, j)*q(j)
      end do
    end do

    call rotatetoz(qb, rmat)
    do i = 1, 4
      qr(i) = 0
      do j = 1, 4
        qr(i) = qr(i) + rmat(i, j)*qb(j)
        bmat2(i, j) = 0._wp
      end do
    end do

    eta = atanh(qr(4)/qr(3))
    cth = cosh(eta)
    sth = sinh(eta)

    bmat2(1, 1) = 1._wp
    bmat2(2, 2) = 1._wp
    bmat2(3, 3) = cth
    bmat2(4, 4) = cth
    bmat2(3, 4) = -sth
    bmat2(4, 3) = -sth

    do i = 1, 4
      do j = 1, 4
        fmat(i, j) = 0._wp
        do k = 1, 4
          do l = 1, 4
            fmat(i, j) = fmat(i, j) + bmat2(i, k)*rmat(k, l)*bmat1(l, j)
          end do
        end do
      end do
    end do

    do m = 1, npar
      do i = 1, 4
        pa2(i, m) = 0._wp
        do j = 1, 4
          pa2(i, m) = pa2(i, m) + fmat(i, j)*pa1(j, m)
        end do
      end do
    end do

    do i = 1, 4
      do j = 1, npar
        kin(npar)%p(i, j) = pa2(i, j)
      end do
    end do

    !> set breitT at the end
    dis%breitT = fmat

  contains

    !-----------------------------------------------------------------------------
    !> @brief
    !> boost into the rest frame of 4 vector a
    !>
    !
    !> @param[in] four vector a to be boosted into rest frame
    !> @param[out] bmat transformation matrix from a -> a'
    !-----------------------------------------------------------------------------
    subroutine boostrestB(a, bmat, ifail_)
      implicit none
      integer, intent(inout) :: ifail_
      real(kind=wp), intent(out) :: bmat(4, 4)
      real(kind=wp), intent(in) :: a(4)
      real(kind=wp) :: v(4), rma2, gam, rma, v2
      integer :: i_, j_

      rma2 = dot(a, a)

      if (rma2 .le. 0._wp) then
        write (*, *) "undue boost", rma2
        ifail_ = 1
      end if
      rma = dsqrt(rma2)
      gam = a(4)/rma
      do i_ = 1, 4
        v(i_) = a(i_)/a(4)
      end do
      v2 = v(1)*v(1) + v(2)*v(2) + v(3)*v(3)
      bmat(4, 4) = gam
      do i_ = 1, 3
        bmat(i_, 4) = -gam*v(i_)
        bmat(4, i_) = -gam*v(i_)
        do j_ = 1, 3
          bmat(i_, j_) = (gam - 1._wp)*v(i_)*v(j_)/v2
        end do
        bmat(i_, i_) = bmat(i_, i_) + 1._wp
      end do
    end subroutine boostrestB

  end subroutine boost_to_breit_dis

  !> setting flag if DIS event shapes are calculated
  subroutine setDISEVTSHAPE(bool)
    logical bool
    DISEVTSHAPE = bool
  end subroutine setDISEVTSHAPE

  !> getting flag for DIS event shapes
  logical function getDISEVTSHAPE()
    getDISEVTSHAPE = DISEVTSHAPE
  end function getDISEVTSHAPE

  !> setting flag if DIS event shapes are calculated
  subroutine setHERA_ORDERETA(bool)
    logical bool
    HERA_ORDERETA = bool
  end subroutine setHERA_ORDERETA

  !> getting flag for DIS event shapes
  logical function getHERA_ORDERETA()
    getHERA_ORDERETA = HERA_ORDERETA
  end function getHERA_ORDERETA

  !> set value of Ehfrac
  subroutine setEhfrac(val)
    real(kind=wp) val
    Ehfrac = val
  end subroutine setEhfrac

  !> get value of Ehfrac
  real(kind=wp) function getEhfrac()
    getEhfrac = Ehfrac
  end function getEhfrac

  !-----------------------------------------------------------------------------
  !> @brief
  !> transforms momenta from gamma-parton to gamma-nucleon com
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine boost_to_gamnuc(npar)
    use KinData_mod, only: boostrest
    integer, intent(in) :: npar
    real(kind=wp) :: p_proton_gamma(1:4), bmat(1:4, 1:4)
    integer i, j, k
    !> momenta are in gamma-parton
    p_proton_gamma = kin(npar)%p(:, 2) - kin(npar)%p(:, 3) + kin(npar)%p(:, 1)/dis%xx1
    call boostrest(p_proton_gamma, bmat)
    do i = 1, npar
      do j = 1, 4
        do k = 1, 4
          kin(npar)%p(j, i) = kin(npar)%p(j, i) + bmat(j, k)*kin(npar)%p(k, i)
        end do
      end do
    end do
  end subroutine boost_to_gamnuc

  !-----------------------------------------------------------------------------
  !> @brief
  !> transforms momenta from gamma-parton to hera frame
  !
  !> @param[in] x1 momentum fraction for incoming parton 1
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine boost_to_hera_dis(npar, x1)
    real(kind=wp), intent(in) :: x1
    integer, intent(in) :: npar
    real(kind=wp) :: paux(1:4, 1:7), eta, cth, sth
    integer :: i

    !> Here we do not need to take proton mass into account,
    !> because we are only considering our massless partons!!!
    !> i.e transformation from (p_e+p_parton)_cm to
    eta = -0.5_wp*log(x1*etaep)
    cth = cosh(eta)
    sth = sinh(eta)
    do i = 1, npar
      paux(3, i) = cth*kin(npar)%p(3, i) - sth*kin(npar)%p(4, i)
      paux(4, i) = cth*kin(npar)%p(4, i) - sth*kin(npar)%p(3, i)
      kin(npar)%p(3, i) = paux(3, i)
      kin(npar)%p(4, i) = paux(4, i)
    end do
  end subroutine boost_to_hera_dis

  !-----------------------------------------------------------------------------
  !> @brief
  !> evaluate transformation from breit to proton gamma rest frame
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine eval_trans_Breit_to_gamma_proton_cm(npar)
    use KinData_mod, only: boostrest
    integer, intent(in) :: npar
    real(kind=wp) :: p_proton_gamma(1:4), bmat(1:4, 1:4)
    !> momenta are in Breit frame
    p_proton_gamma = kin(npar)%p(:, 2) - kin(npar)%p(:, 3) + kin(npar)%p(:, 1)/dis%xx1
    call boostrest(p_proton_gamma, bmat)
    dis%BREITgammaprT = bmat

  end subroutine eval_trans_Breit_to_gamma_proton_cm

  !-----------------------------------------------------------------------------
  !> @brief
  !> set eta value
  !
  !> @param[in] eta of 4-vector p_proton+p+electron in electron proton cm
  !-----------------------------------------------------------------------------
  subroutine setetaep(eta)
    real(kind=wp), intent(in) :: eta
    etaep = eta
  end subroutine setetaep

  !-----------------------------------------------------------------------------
  !> @brief
  !> set x1 value
  !
  !> @param[in] x1=x1*xreg1
  !-----------------------------------------------------------------------------
  subroutine setxx1(x1)
    real(kind=wp), intent(in) :: x1
    dis%xx1 = x1
  end subroutine setxx1

  subroutine setPSDIS(flag)
    logical, intent(in) :: flag
    PSold = flag
  end subroutine setPSDIS

  subroutine setfixT(flag)
    logical, intent(in) :: flag
    fixT = flag
  end subroutine setfixT

  !> set value of Polarisations of electron beams
  subroutine setPol(val)
    double precision val
    isSetPol=.true.
    Pol=val
  end subroutine setPol

  !> set value of EPRC
  subroutine set_EPRC(val)
    logical, intent(in) :: val
    is_EPRC=val
  end subroutine set_EPRC

  !> getter routine to see if EPRC should be used
  logical function use_EPRC()
    use_EPRC=is_EPRC
  end function use_EPRC

  !> Getter, Setter routines
  subroutine setEs(sEe, sEp)
    real(kind=wp), intent(in) :: sEe, sEp
    Ee = sEe
    Ep = sEp
  end subroutine SetEs

  real(kind=wp) function getEe()
    getEe = Ee
  end function getEe

  real(kind=wp) function getEp()
    getEp = Ep
  end function getEp

end module DIS_mod

