!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: EvalObs
!
! DESCRIPTION:
!> Module that defines the type & member functions for an observable
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module EvalObs_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  use KinData_mod, only: kin
  implicit none
  public

  !> interface defining the calling structure of all evaluation functions
  abstract interface

    !> signature of the the standard evaluation function
    function eval_intfce(npar) result(res)
      use Kinds_mod, only: wp
      real(kind=wp) :: res
      integer, intent(in) :: npar
    end function eval_intfce

    !> signature of the memoized evaluation function
    !> (can also be used as a parametrized eval function)
    !> if memoization is intended, imem = 0 must reset the cache
    function eval_mem_intfce(npar, imem) result(res)
      use Kinds_mod, only: wp
      real(kind=wp) :: res
      integer, intent(in) :: npar, imem
    end function eval_mem_intfce

    !> decorator functions
    pure function decorator_intfce(val) result(res)
      use Kinds_mod, only: wp
      real(kind=wp) :: res
      real(kind=wp), intent(in) :: val
    end function decorator_intfce

  end interface

  type :: EvalObs_t
    !> function pointers (& decorator) to define the observable
    procedure(eval_intfce), pointer, nopass :: eval => null()
    procedure(eval_mem_intfce), pointer, nopass :: eval_mem => null()
    procedure(decorator_intfce), pointer, nopass :: decorator => null()
    !> flag for the memoized call
    !>  negative value:  w/o memoization -> eval with extra argument
    !>  positive value:  w/  memoization -> cache reset required
    integer :: imem = 0
    !> observable requirements to test validity
    integer :: min_npar = 0
    integer :: min_njets = 0
    integer :: min_nphotons = 0
    integer :: min_flav_njets = 0
#ifdef USEFASTJET
    integer :: ialgo_fj = -1
    integer :: min_njets_fj = 0
    integer :: min_flav_njets_fj = 0
#endif
    !@todo: `dis%rawjets` ?
    !> the type of observable (manual will have no eval function)
    logical :: is_integer = .false.
    logical :: is_manual = .false.
    logical :: is_composite = .false.
    logical :: needs_flav = .false.
    !> flag if the observable should be printed
    logical :: digest = .true.
    !> lazy evaluation (make these private?)
    logical :: cached_ = .false.
    logical :: valid_ = .false.
    real(kind=wp) :: value_ = 0._wp
  contains
    procedure, pass(self) :: get, set, unset
  end type EvalObs_t

contains

  function get(self, npar, valid) result(res)
    real(kind=wp) :: res
    class(EvalObs_t), intent(inout) :: self
    integer, intent(in) :: npar
    logical, intent(out) :: valid

    !> deal with manual observables first
    !> cache state used to check validty
    !> reset option is ignored for manual observables
    if (self%is_manual) then
      res = self%value_
      self%valid_ = self%cached_
      valid = self%valid_
      return
    end if

    !> with stored valid state & how we want to retrieve cache
    !> need to call this first
    if (self%cached_) then
      res = self%value_
      valid = self%valid_
      return
    end if

    !> check if observable valid
    valid = .true.

    if (self%min_npar > 0) &
      valid = valid .and. (npar >= self%min_npar)

    if (self%min_njets > 0) &
      valid = valid .and. (kin(npar)%njets >= self%min_njets)
    if (self%min_nphotons > 0) &
      valid = valid .and. (kin(npar)%nphotons >= self%min_nphotons)
    if (self%min_flav_njets > 0) &
      valid = valid .and. (kin(npar)%flav_njets >= self%min_flav_njets)
#ifdef USEFASTJET
    if ( (self%ialgo_fj >= 0) .and. (self%min_njets_fj > 0) ) &
      valid = valid .and. (kin(npar)%njets_fj(self%ialgo_fj) >= self%min_njets_fj)
    if ( (self%ialgo_fj >= 0) .and. (self%min_flav_njets_fj > 0) ) &
      valid = valid .and. &
         (kin(npar)%flav_njets_fj(self%ialgo_fj) >= self%min_flav_njets_fj)
#endif

    !> actually compute the observable
    res = 0._wp
    if (valid) then
      if (associated(self%eval)) then
        res = self%eval(npar)
      elseif (associated(self%eval_mem)) then
        res = self%eval_mem(npar, abs(self%imem))
      else
        stop "get: no evaluation routine?!"
      end if
      !> add the decorator if present
      if (associated(self%decorator)) res = self%decorator(res)
    end if  ! valid

    !> store in the cache
    self%value_ = res
    self%valid_ = valid
    self%cached_ = .true.

  end function get

  !> GS: Adding valid option as optional argument (with default true)
  !> to allow to set also invalid observables
  subroutine set(self, value, valid)
    class(EvalObs_t), intent(inout) :: self
    real(kind=wp), intent(in) :: value
    logical, intent(in), optional :: valid
    logical :: valid_

    valid_ = .true.
    if (present(valid)) valid_ = valid

    !> in order to allow reading cached values from KinData
    !> need to allow any observable to be `set`

    ! !> only manual observables can be set manually
    ! if ( .not.self%is_manual ) then
    !   stop "set: trying to set a non-manual observable"
    ! end if

    self%value_ = value
    self%cached_ = .true.
    self%valid_ = valid

  end subroutine set

  subroutine unset(self)
    class(EvalObs_t), intent(inout) :: self
    real(kind=wp) :: val

    self%cached_ = .false.

    !> for memoized observables, need to reset memoization
    if (associated(self%eval_mem) .and. (self%imem > 0)) then
      val = self%eval_mem(0, 0)  ! imem=0 to reset local cache
    end if

  end subroutine unset

end module EvalObs_mod
