!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Floating Point Arithmetics
!
! DESCRIPTION:
!> Collection of useful algorithms for more precise FPA
!> We ignore all algorithms that require a sorted list of inputs because that
!> is simply not possible for our use-case and/or is computationally too costly
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module FPA_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  implicit none
  public
  private :: wp

  !> store some `wp` float properties
  real(kind=wp), parameter, private :: epsilon_ = epsilon(1._wp)
  real(kind=wp), parameter, private :: tiny_ = tiny(1._wp)
  real(kind=wp), parameter, private :: huge_ = huge(1._wp)

  real(kind=wp), parameter :: rel_eps = sqrt(epsilon_)

  interface is_nan
    module procedure is_nan_real, is_nan_cmplx
  end interface is_nan

  interface is_eq
    module procedure is_eq_real, is_eq_cmplx
  end interface is_eq

  interface is_zero
    module procedure is_zero_real, is_zero_cmplx
  end interface is_zero

contains

  !> ieee_is_nan very slow?!
  !>  NaN guaranteed to fail all equality tests (including with itself)
  elemental function is_nan_real(x) result(qnan)
    logical :: qnan
    real(kind=wp), intent(in) :: x
    qnan = (x /= x)
  end function is_nan_real

  elemental function is_nan_cmplx(x) result(qnan)
    logical :: qnan
    complex(kind=wp), intent(in) :: x
    qnan = is_nan_real(real(x)) .or. is_nan_real(aimag(x))
  end function is_nan_cmplx

  elemental function is_eq_real(x, y, rel_tol, abs_tol) result(qapprox)
    ! use, intrinsic :: ieee_arithmetic, only: ieee_is_nan
    logical :: qapprox
    real(kind=wp), intent(in) :: x, y
    real(kind=wp), intent(in), optional :: rel_tol, abs_tol
    real(kind=wp) :: rel_tol_, abs_tol_

    rel_tol_ = rel_eps
    if (present(rel_tol)) rel_tol_ = abs(rel_tol)

    abs_tol_ = 0._wp
    if (present(abs_tol)) abs_tol_ = abs(abs_tol)

    !if (ieee_is_nan(x) .or. ieee_is_nan(y)) then
    if (is_nan_real(x) .or. is_nan_real(y)) then
      qapprox = .false.
    else
      qapprox = abs(x - y) <= max(rel_tol_*max(abs(x), abs(y)), abs_tol_)
    end if

  end function is_eq_real

  elemental function is_eq_cmplx(x, y, rel_tol, abs_tol) result(qapprox)
    logical :: qapprox
    complex(kind=wp), intent(in) :: x, y
    real(kind=wp), intent(in), optional :: rel_tol, abs_tol
    qapprox = is_eq_real(real(x), real(y), rel_tol=rel_tol, abs_tol=abs_tol) .and. &
              is_eq_real(aimag(x), aimag(y), rel_tol=rel_tol, abs_tol=abs_tol)
  end function is_eq_cmplx

  elemental function is_zero_real(x, rel_tol, abs_tol) result(qzero)
    logical :: qzero
    real(kind=wp), intent(in) :: x
    real(kind=wp), intent(in), optional :: rel_tol, abs_tol
    qzero = is_eq_real(x, 0._wp, rel_tol=rel_tol, abs_tol=abs_tol)
  end function is_zero_real

  elemental function is_zero_cmplx(x, rel_tol, abs_tol) result(qzero)
    logical :: qzero
    complex(kind=wp), intent(in) :: x
    real(kind=wp), intent(in), optional :: rel_tol, abs_tol
    qzero = is_zero_real(real(x), rel_tol=rel_tol, abs_tol=abs_tol) .and. &
            is_zero_real(aimag(x), rel_tol=rel_tol, abs_tol=abs_tol)
  end function is_zero_cmplx

  !-----------------------------------------------------------------------------
  !> @brief
  !> check if two floats are approximately the "same" (within precision)
  !
  !> @param[in] x 1st number
  !> @param[in] y 2nd number
  !> @return true if the same
  !-----------------------------------------------------------------------------
  pure logical function is_approx(x, y)
    real(kind=wp), intent(in) :: x, y
    real(kind=wp) :: eps_, thr_, diff_, norm_
    if (x == y) then
      is_approx = .true.
      return
    end if
    eps_ = 128._wp*epsilon(x)
    thr_ = tiny(x)
    diff_ = abs(x - y)
    norm_ = min((abs(x) + abs(y)), huge(x))
    is_approx = (diff_ < max((eps_*norm_), thr_))
  end function is_approx

  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a float is zero (within numerical precision)
  !> elemental function so we can feed in arrays too
  !
  !> @param[in] x number to check
  !> @return true if zero
  !-----------------------------------------------------------------------------
  elemental logical function is_approx_zero(x)
    real(kind=wp), intent(in) :: x
    is_approx_zero = is_approx(x, 0._wp)
  end function is_approx_zero

  !-----------------------------------------------------------------------------
  !> @brief
  !> compute the sum of two floats and the error of the addition
  !
  !> @param[in] a 1st number
  !> @param[in] b 2nd number
  !> @param[out] s sum (a+b) rounded to nearest floating point number
  !> @param[out] t error of the summation so that (a+b)=(s+t) exactly
  !-----------------------------------------------------------------------------
  !> TwoSum algorithm for summation of two numbers a,b
  !> where the sum s = (a+b) is rounded to the nearest
  !> floating point number. The algorithm futher computes
  !> the error t of the summation, such that
  !>     a+b = s+t
  !> *exactly*. Compared to the original "FastTwoSum" (below)
  !> algorithm by [Dekker] and [Kahan] (above), this one
  !> requires no comparison to determine the hierarchy
  !> between a,b and is thus quicker, moreover, it works
  !> with any radix
  !> Algorithm by [Knuth] and [Moller].
  !-----------------------------------------------------------------------------
  pure subroutine twoSum(a, b, s, t)
    implicit none
    real(kind=wp), intent(in) :: a, b
    real(kind=wp), intent(out) :: s, t
    real(kind=wp) :: ap, bp, da, db
    s = a + b
    ap = s - b
    bp = s - ap  ! yes, this is supposed to be `ap`!
    da = a - ap
    db = b - bp
    t = da + db
  end subroutine twoSum

  !-----------------------------------------------------------------------------
  !> @brief
  !> perform the cascaded sum by accumulating values and giving the final
  !
  !> @param[in] mode = {-1,0,1} for {init,accumulation,completion}
  !> @param[inout] val elements to be added (0), final result (1)
  !-----------------------------------------------------------------------------
  !> For large cancellations within the terms, which happens
  !> a lot in our subtraction terms, Kahan might not lead to
  !> such a big improvement compared to the naive sum as
  !> pointed out by [Priest].
  !> The improvement by [Pichat] and [Neumaier] is
  !> simply the accumulation of the error and adding it to the
  !> result in the end instead of doing the compensation directly
  !> after each addition.
  !> In order to avoid the floating-point comparison in the
  !> original formulation of the algorithm we can use the twoSum
  !> algorothm above (which is also valid for any radix).
  !> This algorithm is then called the "cascaded summation"
  !> by [Rump, Ogita, Oishi].
  subroutine accumSum(mode, val)
    implicit none
    integer, intent(in) :: mode
    real(kind=wp) :: val
    real(kind=wp), save :: S, E
!$OMP THREADPRIVATE(S,E)
    real(kind=wp) :: T, Ei
    select case (mode)
    case (-1)                 ! reset accumulators
      S = val
      E = 0._wp
    case (0)                  ! add to accumulator
      call twoSum(S, val, T, Ei)
      S = T
      E = E + Ei
    case (1)                  ! retrieve result
      val = S + E
    case default
      stop "accumSum: unknown mode"
    end select
  end subroutine accumSum

end module FPA_mod

!-------------------------------- code cemetery

! pure subroutine fastTwoSum(a,b, s,t)
!   implicit none
!   real(kind=wp), intent(in) :: a,b
!   real(kind=wp), intent(out) :: s,t
!   s  = a + b
!   if ( abs(a) >= abs(b) ) then
!      t = b - (s - a)
!   else
!      t = a - (s - b)
!   end if
! end subroutine fastTwoSum

! !> TwoMultFMA
! !> Multiplication assuming that the
! !> fused multiply-add instruction is available
! pure subroutine twoMult(a,b, m,t)
!   implicit none
!   real(kind=wp), intent(in) :: a,b
!   real(kind=wp), intent(out) :: m,t
!   m  = a * b
!   t  = -m + a*b
! end subroutine twoMult

! !> "compensated summation" by [Kahan]
! !> minor modification by using twoSum from above
! subroutine compSum(mode, val)
!   implicit none
!   integer :: mode
!   real(kind=wp) :: val
!   real(kind=wp), save :: S, C
!   real(kind=wp) :: T, Y
!   select case (mode)
!   case (-1)                 ! reset accumulators
!      S = 0._wp
!      C = 0._wp
!   case (0)                  ! add to accumulator
!      Y = val + C
!      call twoSum(S,Y, T,C)
!      S = T
!   case (1)                  ! retrieve result
!      val = S
!   case default
!      stop "unknown input"
!   end select
! end subroutine compSum
