!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Flavour [external members: `_flav`]
!
! DESCRIPTION:
!> Module providing settings for flavour tagging and a buffer for delayed
!> accumulation in histograms.
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Flavour_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_flav, destroy_flav, isEmpty_flav, digest_flav
  public :: writeObs_flav, readObs_flav
  public :: push_flav, pop_flav
  public :: setActive_flav, isActive_flav
  public :: keepEvenTagged_flav, qKeepEvenTagged_flav
  public :: tagId_flav, dress_flav, getId_flav
  public :: iTag_flav, has_changed_flav
  public :: print_flav
  public :: minnjets_flav, is_leading_flav
  public :: set_flav_perm, unset_flav_perm
  public :: getPDF_flav
  public :: check_flav, set_sub_flav, check_sub_flav
  public :: set_minnpart_flav
  ! public :: map_flav
  ! public :: mapX3_flav

  logical :: qactive_flav = .false.
  !> if true, net summation; if false, modulo_2
  logical :: qkeep_even_tagged_flav = .false.

  !> PDT Id of the flavour to tag (+/-)
  integer :: id_flav = 5 ! default: bottom
  !> minimum number of reconstructed tagged jets
  integer :: minnjets_flav = 0
  integer :: minnpart_flav = 1  ! need a flavoured parton if tag is active
  !> do we request the leading jet to be the tagged one?
  logical :: is_leading_flav = .false.
  !> count the anti-quark as +1 or -1?
  integer :: count_qb_flav = -1

  !> possibility to switch on/off individual `isub` terms
  !> (needed for IR safety test when ME between `isub`s are not the same)
  integer :: pick_sub = 0  ! 0: inactive
  !> keep it a global state (no multirun) !$OMP THREADPRIVATE(pick_sub)

  !> flavour assignment and possible permutations for flavour-checking
  type Flav_t
    integer :: npar
    integer, dimension(:), allocatable :: flavours
    integer, dimension(:), allocatable :: permutation
    integer, dimension(2) :: pdf
    logical :: has_changed = .false.
    logical :: force_has_changed = .false.
  end type Flav_t
  type(Flav_t), target, dimension(:), allocatable :: array_flav
!$OMP THREADPRIVATE(array_flav)

  !> buffer to stage histogram filling for a later bino call
  type FlavBuffer_t
    integer :: ip1, ip2
    double precision, allocatable, dimension(:) :: partons
    double precision, allocatable, dimension(:) :: valueUsed
    logical, allocatable, dimension(:) :: validUsed
  end type FlavBuffer_t

  integer, dimension(:), pointer :: perm_p => null()
!$OMP THREADPRIVATE(perm_p)

  type(FlavBuffer_t), allocatable, dimension(:) :: buffer_flav
  integer :: buffer_size_flav = 0
!$OMP THREADPRIVATE(buffer_flav, buffer_size_flav)

  interface iTag_flav
    module procedure iTag_pdt
    module procedure iTag_npar
  end interface

contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> Init dynamical arrays for Flavour-tagging
  !
  !> @param[in] nmin minimum number of external particles (npar) >= 4
  !> @param[in] nmax maximum number of external particles (npar) >= [4|nmin]
  !> @param[in] nscl number of active scales
  !> @param[in] nobs number of observables in the "used" state
  !-----------------------------------------------------------------------------
  subroutine init_flav(nmin, nmax, nscl, nobs)
    integer, optional, intent(in) :: nmin, nmax
    integer, intent(in) :: nscl, nobs
    integer :: i, max_nsub_flav
    integer :: min_npar, max_npar
    integer, external :: max_nflavourZ, max_nflavourWp, max_nflavourWm
    integer, external :: max_nflavourZJ, max_nflavourWpJ, max_nflavourWmJ
    integer, external :: max_nflavourWmJJ
    integer, external :: max_nflavourZH, max_nflavourWpH, max_nflavourWmH
    integer, external :: max_nflavourepem, max_nflavourepemjj

    if ( .not. qactive_flav ) return  ! nothing to allocate
    ! print*, "tagging flavour", id_flav

    ! print*, "nscl", nscl
    ! print*, "nobs", nobs

    if (present(nmin)) then
      min_npar = max(3,nmin)
    else
      min_npar = 3
    end if
    if (present(nmax)) then
      max_npar = min(10,nmax)
    else
      max_npar = 10
    end if
    ! print*, "init_flav: min/max:", min_npar, max_npar
    allocate(array_flav(min_npar:max_npar))
    do i=min_npar,max_npar
      allocate(array_flav(i)%flavours(1:i))
      allocate(array_flav(i)%permutation(1:i))
      array_flav(i)%npar = i
      array_flav(i)%has_changed = .false.
      array_flav(i)%force_has_changed = .false.
    end do

#ifdef FLAVZ
    max_nsub_flav =  max_nflavourZ()
#elif FLAVZJ
    max_nsub_flav =  max_nflavourZJ()
#elif FLAVWp
    max_nsub_flav =  max_nflavourWp()
#elif FLAVWm
    max_nsub_flav =  max_nflavourWm()
#elif FLAVWpJ
    max_nsub_flav =  max_nflavourWpJ()
#elif FLAVWmJ
    max_nsub_flav =  max_nflavourWmJ()
#elif FLAVWmJJ
    max_nsub_flav =  max_nflavourWmJJ()
#elif FLAVZH
    max_nsub_flav =  max_nflavourZH()
#elif FLAVWpH
    max_nsub_flav =  max_nflavourWpH()
#elif FLAVWmH
    max_nsub_flav =  max_nflavourWmH()
#elif FLAVepem
    max_nsub_flav =  max_nflavourepem()
#elif FLAVepemjj
    max_nsub_flav =  max_nflavourepemjj()
#else
    stop "init_flav: process not fully implemented for flavour-tagging."
#endif
    ! print*, "max_nsub_flav", max_nsub_flav

    buffer_size_flav = 0
    ! if (allocated(buffer_flav)) deallocate(buffer_flav)
    allocate(buffer_flav(max_nsub_flav+1)) !> +1 for the "current" index

    do i=1,max_nsub_flav+1
      allocate(buffer_flav(i)%partons(nscl))
      allocate(buffer_flav(i)%valueUsed(nobs))
      allocate(buffer_flav(i)%validUsed(nobs))
    end do

    ! only for threadloacl init: pick_sub = 0

    ! print*, "init_flav: exited fine"
  end subroutine init_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> Deallocate arrays for Flavour-tagging
  !-----------------------------------------------------------------------------
  subroutine destroy_flav()
    integer :: i

    if ( .not. qactive_flav ) return  ! nothing to deallocate

    do i = lbound(array_flav,1), ubound(array_flav,1)
      deallocate(array_flav(i)%flavours)
      deallocate(array_flav(i)%permutation)
    end do
    deallocate(array_flav)

    !> no buffer needed when histogramming is inactive

    do i = lbound(buffer_flav,1), ubound(buffer_flav,1)
      deallocate(buffer_flav(i)%partons)
      deallocate(buffer_flav(i)%valueUsed)
      deallocate(buffer_flav(i)%validUsed)
    end do
    deallocate(buffer_flav)
    buffer_size_flav = 0

  end subroutine destroy_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if the buffer is empty
  !
  !> @return boolean flag for the status of buffer
  !-----------------------------------------------------------------------------
  logical function isEmpty_flav()
    isEmpty_flav = ( buffer_size_flav == 0 )
  end function isEmpty_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> write observables into the current buffer
  !
  !> @param[in] valueUsed array to copy values for the used observables from
  !> @param[in] validUsed array to copy status for the used observables from
  !> @param[in] offset index shift in buffer array (default = +1)
  !-----------------------------------------------------------------------------
  subroutine writeObs_flav(valueUsed, validUsed, offset)
    double precision, intent(in) :: valueUsed(:)
    logical, intent(in) :: validUsed(:)
    integer, optional, intent(in) :: offset
    integer :: i,j, ioff
    type(FlavBuffer_t), allocatable, dimension(:) :: buffer_tmp
    integer :: new_size
    ioff = +1  ! default: write into the next "un-pushed" entry
    if (present(offset)) ioff = offset
    ! print*, "buffer_size_flav, ioff", buffer_size_flav, ioff
    ! print*, "size value in ", size(valueUsed)
    ! print*, "size valid in ", size(validUsed)
    ! print*, "size value loc", size(buffer_flav(buffer_size_flav+ioff)%valueUsed)
    ! print*, "size valid loc", size(buffer_flav(buffer_size_flav+ioff)%validUsed)

    !> do we need to resize the buffer?
    if (size(buffer_flav)==buffer_size_flav) then
      !> create tmp buffer double the size
      new_size = 2*buffer_size_flav
      allocate(buffer_tmp(new_size))
      do i=1,new_size
        allocate(buffer_tmp(i)%partons(size(buffer_flav(1)%partons)))
        allocate(buffer_tmp(i)%valueUsed(size(buffer_flav(1)%valueUsed)))
        allocate(buffer_tmp(i)%validUsed(size(buffer_flav(1)%validUsed)))
      end do
      ! print*, "doubled the buffer"
      ! print*, "old size: ", buffer_size_flav
      ! print*, "new size: ", new_size
      ! print*,
      !> copy over old stuff
      do j = 1, buffer_size_flav
        do i = 1, size(buffer_flav(j)%partons)
          buffer_tmp(j)%partons(i) = buffer_flav(j)%partons(i)
        end do
        do i = 1, size(buffer_flav(j)%valueUsed)
          buffer_tmp(j)%valueUsed(i) = buffer_flav(j)%valueUsed(i)
        end do
        do i = 1, size(buffer_flav(j)%validUsed)
          buffer_tmp(j)%validUsed(i) = buffer_flav(j)%validUsed(i)
        end do
      end do
      !> overwrite the global flav array
      buffer_flav = buffer_tmp

      !> deallocate tmp flavour buffer
      do i = lbound(buffer_tmp,1), ubound(buffer_tmp,1)
        deallocate(buffer_tmp(i)%partons)
        deallocate(buffer_tmp(i)%valueUsed)
        deallocate(buffer_tmp(i)%validUsed)
      end do
      deallocate(buffer_tmp)
    end if

    !> loop to avoid re-allocation
    do i = 1, size(valueUsed)
      buffer_flav(buffer_size_flav+ioff)%valueUsed(i) = valueUsed(i)
    end do
    do i = 1, size(validUsed)
      buffer_flav(buffer_size_flav+ioff)%validUsed(i) = validUsed(i)
    end do
  end subroutine writeObs_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> read observables from the buffer
  !
  !> @param[out] valueUsed array to copy values for the used observables to
  !> @param[out] validUsed array to copy status for the used observables to
  !> @param[in] offset index shift in buffer array (default = 0)
  !-----------------------------------------------------------------------------
  subroutine readObs_flav(valueUsed, validUsed, offset)
    double precision, intent(out) :: valueUsed(:)
    logical, intent(out) :: validUsed(:)
    integer, optional, intent(in) :: offset
    integer :: i, ioff

    ioff = 0  ! default: red from the last "pushed" entry
    if (present(offset)) ioff = offset
    !> loop to avoid re-allocation
    do i = 1, size(valueUsed)
      valueUsed(i) = buffer_flav(buffer_size_flav+ioff)%valueUsed(i)
    end do
    do i = 1, size(validUsed)
      validUsed(i) = buffer_flav(buffer_size_flav+ioff)%validUsed(i)
    end do
  end subroutine readObs_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> write parton lumilosity and increment the buffer index
  !> the observables need to be set manually using `writeObs_flav` before
  !
  !> @param[in] ip1 PDT id of PDF flavour 1
  !> @param[in] ip2 PDT id of PDF flavour 2
  !> @param[in] partons luminosity of the current flavour assignment
  !-----------------------------------------------------------------------------
  subroutine push_flav(ip1, ip2, partons)
    integer, intent(in) :: ip1, ip2
    double precision, intent(in) :: partons(:)
    integer :: i

    !> should be only on for debugging
    if (buffer_size_flav == size(buffer_flav))  &
    &  stop "push_flav: reached buffer size limit"

    buffer_size_flav = buffer_size_flav + 1  ! next entry
    ! print*, "push_flav: next entry, current size: ", buffer_size_flav
    ! print*,

    !> store PDF flavour information
    buffer_flav(buffer_size_flav)%ip1 = ip1
    buffer_flav(buffer_size_flav)%ip2 = ip2

    !> loop to avoid re-allocation
    do i = 1, size(buffer_flav(buffer_size_flav)%partons)
      buffer_flav(buffer_size_flav)%partons(i) = partons(i)
    end do

  end subroutine push_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> read out parton lumilosity and decrement the buffer index
  !> the observables need to be read out manually using `readObs_flav` before
  !
  !> @param[out] ip1 PDT id of PDF flavour 1
  !> @param[out] ip2 PDT id of PDF flavour 2
  !> @param[out] partons luminosity of the current flavour assignment
  !-----------------------------------------------------------------------------
  subroutine pop_flav(ip1, ip2, partons)
    integer, intent(out) :: ip1, ip2
    double precision, intent(out) :: partons(:)
    integer :: i

    ip1 = buffer_flav(buffer_size_flav)%ip1
    ip2 = buffer_flav(buffer_size_flav)%ip2

    partons = 0d0
    !> loop to avoid re-allocation
    do i = 1, size(buffer_flav(buffer_size_flav)%partons)
      partons(i) = buffer_flav(buffer_size_flav)%partons(i)
    end do

    buffer_size_flav = buffer_size_flav - 1  ! remove entry

  end subroutine pop_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the flavour tagging
  !
  !> @param[in] boolean flag
  !> @param[in] tag optional argument to set the flavour of the tagging
  !-----------------------------------------------------------------------------
  subroutine setActive_flav(qactive, tag)
    logical, intent(in) :: qactive
    integer, intent(in), optional :: tag
    qactive_flav = qactive
    if ( present(tag) ) then
        if (tag < 0) then
          id_flav = -1
          print*, "setActive_flav: negative tag triggers on all quarks"
        else if (tag == 0) then
          print*, "setActive_flav: can't tag gluons"
          stop
        else
          id_flav = tag
          print*, "setActive_flav: turning on tagging for:", id_flav
        end if
    end if
  end subroutine setActive_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of flavour-tagging
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_flav()
    isActive_flav = qactive_flav
  end function isActive_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off whether to keep even-tagged jets
  !
  !> @param[in] boolean flag
  !-----------------------------------------------------------------------------
  subroutine keepEvenTagged_flav(qactive)
    logical, intent(in) :: qactive
    qkeep_even_tagged_flav = qactive
  end subroutine keepEvenTagged_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> inquire status of whether we are keeping even-tagged flavoured jets or not
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function qKeepEvenTagged_flav()
    qKeepEvenTagged_flav = qkeep_even_tagged_flav
  end function qKeepEvenTagged_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to retrieve flavour id of the tagging
  !
  !> @return flavour id of the tagging
  !-----------------------------------------------------------------------------
  integer function tagId_flav()
    tagId_flav = id_flav
  end function tagId_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the PDF interpolation settings
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_flav(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_act_lead =  &
      & '(1x,"*",1x,"tag flavour:",T30,I0," (leading)",T103,"*")'
    character(len=*), parameter :: fmt_act   =  &
      & '(1x,"*",1x,"tag flavour:",T30,I0,T103,"*")'
    character(len=*), parameter :: fmt_isub =  &
      & '(1x,"*",1x,T35," isub = ",I0,T103,"*")'

    if ( .not.isActive_flav() ) return

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    if (is_leading_flav) then
      write(iout,fmt_act_lead) id_flav
    else
      write(iout,fmt_act) id_flav
    end if
    if (pick_sub > 0) write(iout,fmt_isub) pick_sub

  end subroutine digest_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> set thread-local flavour assignment
  !
  !> @param[in] array `npar`-dimensional arrray with PDG-Ids
  !> @param[in] npar number of particles (=lengths of array)
  !> @param[inout] ipass flag to accept/reject sub-channel
  !-----------------------------------------------------------------------------
  subroutine dress_flav(array, npar, ipass, pdf)
    integer, intent(in) :: array(npar)
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer, intent(in) :: pdf(2)
    integer :: i,j
    integer :: new_flavours(npar)

    !> Check whether flavour permutation is set up
    if (.not.associated(perm_p)) stop "dress_flav: flav_perm is not associated!"

    !> npar check
    if (size(perm_p) /= npar) stop "dress_flav: npar mismatch"

    array_flav(npar)%npar = npar  ! this is what Imre wanted
    array_flav(npar)%pdf = pdf ! this is also what Imre wanted

    !> Fill up flavour array according to permutation
    do i=1,npar
      new_flavours(perm_p(i)) = array(i)
    end do

    !> Flip outgoing flavours into their proper direction in the initial state
    new_flavours(:2) = -new_flavours(:2)

    !> potentially easier: just map all flavourless entries to gluons
    !where (abs(new_flavours)/=id_flav) new_flavours = 0

    if (id_flav > 0) then
      !> Do we have the tagged flavour among the final states?
      if ( count(abs(new_flavours(3:)).eq.id_flav) < minnjets_flav ) &
        & ipass = 0
      !> if we want to tag but allow flav_njets = 0
      !> good idea to have a selector on # flavoured partons
      if ( count(abs(new_flavours(3:)).eq.id_flav) < minnpart_flav ) &
        & ipass = 0
    else
      !> same, but any quark counts
      if ( count(new_flavours(3:).ne.0) < minnjets_flav ) &
        & ipass = 0
      if ( count(new_flavours(3:).ne.0) < minnpart_flav ) &
        & ipass = 0
    end if

    if (ipass == 0) return

    array_flav(npar)%has_changed = .false.
    do j=3,npar
      array_flav(npar)%has_changed = array_flav(npar)%has_changed .or.  &
        & .not. dress_flav_is_same(new_flavours(j), array_flav(npar)%flavours(j), id_flav)
    end do
    
    !> GS: In case of minnpart_flav = 0, we need to force the evaluation of observables
    !> at the beginning of each flavour block (otherwise, in absence of flavoured particles, 
    !> the initial array of zero's will be "equivalent" to the first permutation,
    !> and has_changed would result in .false., thus not evaluating observables but trying 
    !> to access a non-existing cache)
    if (array_flav(npar)%force_has_changed) then
       array_flav(npar)%has_changed = .true.
       array_flav(npar)%force_has_changed = .false.
    end if

    !print*, "dress: old  ", array_flav(npar)%flavours
    !print*, "dress: new  ", new_flavours
    !print*, "has_changed?", array_flav(npar)%has_changed
    array_flav(npar)%flavours(:) = new_flavours(:)

    !!! if (ipass /= 0)  &
    !!!   &  print*, npar, "dress_flav: ", array_flav(npar)%flavours(3:), " || ", ipass

  contains

    logical function dress_flav_is_same(flav1_, flav2_, id_)
      integer, intent(in) :: flav1_, flav2_, id_
      if (id_>0) then
        if (abs(flav1_)==id_) then
          dress_flav_is_same = (flav1_ == flav2_)
        else
          dress_flav_is_same = (abs(flav2_) /= id_)
        end if
      else
        dress_flav_is_same = ((flav1_/=0 .and. flav2_/=0) .or. (flav1_==0 .and. flav2_==0))
      end if
    end function dress_flav_is_same


  end subroutine dress_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> setting up momentum-label permutations for correct flavour assignments
  !
  !> @param[in] npar momentum set to consider
  !> @param[in] perm_array permutation array for a given npar
  !-----------------------------------------------------------------------------
  subroutine set_flav_perm(npar, perm_array)
    integer, intent(in) :: npar
    integer, intent(in) :: perm_array(npar)

    if ( .not. isActive_flav() ) return

    !> Fill up permutation array of array_flav
    array_flav(npar)%permutation = perm_array

    !> Associate pointer
    if (associated(perm_p)) &
      & print*, "set_flav_perm: warning, permutation pointer already associated!"

    perm_p => array_flav(npar)%permutation

    !> we reset the flavour for each block of unrolling in qcdnorm
    array_flav(npar)%flavours(:) = 0

    !> forcing observable evaluation at the beginning of each flavour block
    if (minnpart_flav == 0) then
       array_flav(npar)%force_has_changed = .true.
    end if

  end subroutine set_flav_perm

  !-----------------------------------------------------------------------------
  !> @brief
  !> Dissociate permutation pointer. Destruction happens in destroy_flav.
  !-----------------------------------------------------------------------------
  subroutine unset_flav_perm()
    perm_p => null()
  end subroutine unset_flav_perm


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to retrieve flavour id's
  !
  !> @param[in] npar momentum set to consider
  !> @param[in] idx location in momentum set
  !> @return flavour id of the `idx`-th particle
  !-----------------------------------------------------------------------------
  integer function getId_flav(npar, idx)
    integer, intent(in) :: npar, idx
    if ( idx < lbound(array_flav(npar)%flavours,1) .or.  &
       & idx > ubound(array_flav(npar)%flavours,1) )   &
    &  stop "getId_flav: outside of range"
    getId_flav = array_flav(npar)%flavours(idx)
  end function getId_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a parton PDT id matches a tag
  !
  !> @param[in] nflav_q number of quarks
  !> @return +/- 1 for a tag 0 for no match
  !-----------------------------------------------------------------------------
  integer function iTag_pdt(pdt_id)
    integer, intent(in) :: pdt_id
    iTag_pdt = 0
    if ( (id_flav > 0 .and. abs(pdt_id)==id_flav) .or.  &
       & (id_flav < 0 .and. pdt_id/=0) ) iTag_pdt = sign(1,pdt_id)
  end function iTag_pdt


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if flavour id matches the tag
  !
  !> @param[in] npar momentum set to consider
  !> @param[in] idx location in momentum set
  !> @return +/- 1 for a tag 0 for no match
  !-----------------------------------------------------------------------------
  integer function iTag_npar(npar, idx)
    integer, intent(in) :: npar, idx
    iTag_npar = iTag_pdt( getId_flav(npar, idx) )
  end function iTag_npar


  !-----------------------------------------------------------------------------
  !> @brief
  !> inquire if last call of `dress_flav` changed the tagged flavours
  !
  !> @param[in] npar momentum set to consider
  !> @return bool indicating a change
  !-----------------------------------------------------------------------------
  logical function has_changed_flav(npar)
    integer, intent(in) :: npar
    has_changed_flav = array_flav(npar)%has_changed
  end function has_changed_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> use current mapping to determine the flavour for the mapped momentum set
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine print_flav(npar)
    integer, intent(in) :: npar

    write(0,*) "-------- flavour", npar, "--------"
    write(0,*) array_flav(npar)%flavours

  end subroutine print_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to retrieve PDF flavours
  !
  !> @param[in] npar momentum set to consider
  !> @return flavour id of the `idx`-th particle
  !-----------------------------------------------------------------------------
  integer function getPDF_flav(npar, idx)
    integer, intent(in) :: npar, idx
    if (idx > 2 .or. idx < 1) then
      print*, "getPDF_flav: PDF idx must be 1 or 2"
      stop
    end if
    getPDF_flav = array_flav(npar)%pdf(idx)
  end function getPDF_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a flavour array has tagged flavour(s) in it
  !> (only to be used for flav_block = 0, i.e. the ME level!)
  !
  !> @param[in] array `npar`-dimensional arrray with PDG-Ids
  !> @param[in] npar number of particles (=lengths of array)
  !> @param[inout] ipass flag to accept/reject sub-channel
  !-----------------------------------------------------------------------------
  subroutine check_flav(array, npar, ipass)
    integer, intent(in) :: array(npar)
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    if ( count(abs(array(:)).eq.id_flav) < minnpart_flav ) ipass = 0
    ! stop "don't get here..."
  end subroutine check_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> if noselection is true, we do not require any minimum number of 
  !> flavoured particles before clustering (same result, but less performant)
  !
  !> @param[in] boolean
  !-----------------------------------------------------------------------------
  subroutine set_minnpart_flav(noselection)
    logical, intent(in) :: noselection
    if (noselection) then
       print*, "> not requiring any minimum number of flav.part. before clustering!"
       minnpart_flav = 0
    endif
  end subroutine set_minnpart_flav

  !-----------------------------------------------------------------------------
  !> @brief
  !> pick an `isub` flavour sub-channel
  !
  !> @param[in] isub index for the flavour sub-channel
  !-----------------------------------------------------------------------------
  subroutine set_sub_flav(isub)
    integer, intent(in) :: isub
    print*, "set_sub_flav: only picking out sub-channel:", isub
    pick_sub = isub
  end subroutine set_sub_flav


  !-----------------------------------------------------------------------------
  !> @brief
  !> check `isub` selection
  !
  !> @param[in] isub index for the flavour sub-channel
  !> @param[inout] ipass flag to accept/reject sub-channel
  !-----------------------------------------------------------------------------
  subroutine check_sub_flav(isub, ipass)
    integer, intent(in) :: isub
    integer, intent(inout) :: ipass
    if ( pick_sub <= 0 ) return
    if ( isub /= pick_sub ) ipass = 0
  end subroutine check_sub_flav


!!!!!!!!!!!!!!!!!!!!!!!!!!! FLAVTAG CEMETERY !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!




  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> setter routine to set up PDF flavours
  ! !
  ! !> @param[in] npar momentum set to consider
  ! !> @return flavour id of the `idx`-th particle
  ! !-----------------------------------------------------------------------------
  ! subroutine setPDF_flav(npar, idx, flav)
  !   integer, intent(in) :: npar, idx
  !   integer :: flav
  !   if (idx > 2 .or. idx < 1) then
  !     print*, "setPDF_flav: PDF idx must be 1 or 2"
  !     stop
  !   end if
  !   array_flav(npar)%pdf_flavours(idx) = flav

  ! end subroutine setPDF_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> use current mapping to determine the flavour for the mapped momentum set
  ! !
  ! !> @param[in] npar momentum set
  ! !-----------------------------------------------------------------------------
  ! subroutine map_flav(npar)
  !   use Mapping_mod, only : getCurrent_map, hasParent_map, getParent_map
  !   integer, intent(in) :: npar
  !   !> mapping information
  !   logical :: qsuccess
  !   integer :: npar_from, npar_to
  !   integer, allocatable, dimension(:)   :: antenna
  !   integer, allocatable, dimension(:,:) :: mapping
  !   print*, "map_flav: entered"

  !   !> first check if there's a parent mapping
  !   if (hasParent_map()) then
  !     print *, "map_flav: hasParent_map"
  !     qsuccess = getParent_map( npar_from=npar_from, npar_to=npar_to,  &
  !                             & antenna=antenna, mapping=mapping )
  !     if (.not.qsuccess) stop "map_flav: parent map failure"
  !     !> Perform first single unresolved flavour mapping
  !     print*, "map_flav: parent mapping =", mapping
  !     call mapX3_flav(npar_from,npar_to,mapping,antenna)
  !     !> print iterated mapping as sanity check
  !     call print_flav(npar_from)
  !     call print_flav(npar_to)
  !   end if

  !   !> get the current mapping information
  !   qsuccess = getCurrent_map( npar_from=npar_from, npar_to=npar_to,  &
  !                            & antenna=antenna, mapping=mapping)

  !   if (.not.qsuccess) then
  !       print*, "map_flav: no successful mapping"
  !       print*, "map_flav: must correspond to non-subtraction term"
  !       print*, "map_flav: Setting PDFs ..."
  !       call setPDF_flav(npar, 1, array_flav(npar)%flavours(1))
  !       call setPDF_flav(npar, 2, array_flav(npar)%flavours(2))
  !       return
  !   end if

  !   !> check consistency
  !   if (npar /= npar_to) then
  !     print*, "map_flav: npar mismatch!", npar_from, ">", npar_to, npar
  !     stop
  !   end if

  !   !> print as sanity check
  !   call print_flav(npar_from)

  !   !> flavour mapping of integrated subtractions (CalX)
  !   !> single and double unresolved flavour mappings (X3 and X4)
  !   select case (npar_from-npar_to)
  !   case(0)
  !     call mapCalX_flav(npar, mapping, antenna)
  !   case (1)
  !     !call mapX3_flav_old(npar_from,npar_to,mapping,antenna,flavChange)
  !     !> New mapping routine - flavChanges identified within subroutine
  !     call mapX3_flav(npar_from,npar_to,mapping,antenna)
  !   case (2)
  !     !call mapX4_flav_old(npar_from,npar_to,mapping,antenna,flavChange)
  !     !> New mapping routine - flavChanges identified within subroutine
  !     call mapX4_flav(npar_from,npar_to,mapping,antenna)
  !   case default
  !     stop "map_flav: unknown mapping"
  !   end select

  !   !> printing flavours, sanity check
  !   call print_flav(npar_to)

  !   !> Correcting for unphysical matrix element mismatches if necessary
  !   if (associated(perm_p)) then
  !     print*, "map_flav: correcting flavours to match matrix element"
  !     call correct_flav(npar_to)
  !     print*, "map_flav: flavours after correction:"
  !     call print_flav(npar_to)
  !   end if

  ! end subroutine map_flav


  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> Old flavour mapping subroutine for X30 antenna
  ! !
  ! !> @param[in] mapping information
  ! !-----------------------------------------------------------------------------
  ! subroutine mapX3_flav_old(npar_from,npar_to,mapping,antenna,flavChange)
  !   integer, intent(in) :: npar_from, npar_to
  !   integer, dimension(2), intent(in) :: flavChange
  !   integer, dimension(3), intent(in) :: antenna
  !   !RG: Should the mapping be alloctable dimension?
  !   integer, dimension(:,:), intent(in) :: mapping
  !   !integer, optional, intent(in) :: flavChange
  !   !> ih1, ih2
  !   integer :: i
  !   integer :: ih1,ih2, jh1,jh2, iu1,iu2
  !   integer :: fh1,fh2, fu1,fu2
  !   !> Introduce logical to determine if we encounter II FC mapping - used to determine IF a30,g->q type
  !   logical :: IIFC = .false.
  !   print*, "mapX3_flav: Entered"
  !   !> determine mapped flavour indices
  !   ih1 = antenna(lbound(antenna,1)) !hard 1
  !   ih2 = antenna(ubound(antenna,1)) !hard 2
  !   iu1 = antenna(lbound(antenna,1) +1) !unresloved parton 1
  !   iu2 = antenna(ubound(antenna,1) -1) !unresloved parton 2
  !   !> all flavours incoming, rotate the final state flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)
  !   !> determe flavour of hard/unresolved partons
  !   fh1 = array_flav(npar_from)%flavours(ih1) !flavour of hard 1
  !   fh2 = array_flav(npar_from)%flavours(ih2) !flavour of hard 2
  !   fu1 = array_flav(npar_from)%flavours(iu1) !flavour of unresolved 1
  !   fu2 = array_flav(npar_from)%flavours(iu2) !flavour of unresolved 2
  !   !> are we in a flavChange limit
  !   !> in single unresolved limit, flavChange declared as (/iUnresolved,0/)
  !   if( any(flavChange /= 0) ) then
  !     print *, "mapX3_flav: flavChange = ", flavChange
  !     !> First, for x30 there should be no fc2 (flavour changing parton 2)
  !     if( flavChange(2) /= 0) then
  !       print*, "mapX3_flav: flavChange(2) /= 0, error!"
  !       stop
  !     end if
  !     !> Otherwise, find position of unresolved parton for flavour mapping
  !     ih2 = flavChange(1)

  !     !> II mapping, if original hard radiator 2 was initial state parton. I'll map both onto same position
  !     !> This happens for d30 gtoq II mapping
  !     do i=1,npar_from
  !       if( i==mapping(1,ih1) ) jh1 = mapping(2,i)
  !       if( i==mapping(1,ih2) ) jh2 = mapping(2,i)
  !     end do
  !     !> If they are mapping onto same position. Move ih2 > ih1 and re-assign ih2
  !     if( jh1 == jh2 ) then
  !       IIFC = .true. !Store information that we are in an II FC mapping case
  !       ih1 = ih2 ! equiavlent to flavChange(1)
  !       ih2 = antenna(ubound(antenna,1))
  !     end if

  !   end if

  !   !> RG: npar_to upper limit for X30 (if set unresolved parton mapping is npar_to+1 entry)
  !   do i=1,npar_from
  !     if ( ih1==mapping(1,i) ) jh1 = mapping(2,i)
  !     if ( ih2==mapping(1,i) ) jh2 = mapping(2,i)
  !     !> ignore flavours in antenna, mapping given above
  !     if ( any(antenna==mapping(1,i)) ) cycle
  !     !> copy over flavours that are not involved in the mapping
  !     array_flav(npar_to)%flavours(mapping(2,i)) = array_flav(npar_from)%flavours(mapping(1,i))
  !   end do

  !   print*, "mapX3_flav: antenna", antenna
  !   print*, "mapX3_flav: mapping", mapping
  !   print*, "mapX3_flav: hard rad.", fh1, fh2

  !   if (size(antenna)/=3) then
  !     print*, "mapX3_flav: X30, wrong atenna size pal"
  !     stop
  !   else if( fu1 /= fu2 .and. any(flavChange /= 0) ) then
  !     print*, "mapX3_flav: problem in flavchanging"
  !     stop
  !   end if

  !   !> from here on assume: X30
  !   if (fh1==0 .and. fh2==0) then
  !     !> antenna: ggg --> gg
  !     print*, "~~~~~~ antenna: ggg --> gg (F30)"
  !     if (fu2 /= 0)  &
  !     &  stop "mapX3_flav: no such antenna"
  !     !>
  !     array_flav(npar_to)%flavours(jh1) = fh1
  !     array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !   else if ( fh1==0 .or. fh2==0 ) then
  !     if (fu2 == 0) then
  !       !> antenna: qgg --> qg
  !       print*, "~~~~~~ antenna: qgg --> qg (D30)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !       !> nasty case of IF FC configuariton: d30,gtoq
  !       !> II FC case is fine, just to treat speical IF FC case
  !       if( .not. IIFC .and. flavChange(1) /= 0 ) then
  !         print*, "mapX3_flav: encountered a d30,gtoq - corrected flavour assignment"
  !         array_flav(npar_to)%flavours(jh1) = fh2
  !         array_flav(npar_to)%flavours(jh2) = fh1
  !       end if
  !     else
  !       !> antenna: gqq --> gg
  !       print*, "~~~~~~ antenna: gqq --> gg (G30)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = 0!this is fh2+fu1
  !     end if
  !   else
  !     !> Then all hard quark radiator antenna. Distinguish qQQ,
  !     if (fu2 == 0) then
  !       !> antenna: qgq --> qq
  !       print*, "~~~~~~ antenna: qgq --> qq (A30)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2
  !       !> nasty case of IF FC configuration: a30,gtoq
  !       !> II FC case is fine, just to treat special IF FC case
  !       if( .not. IIFC .and. flavChange(1) /= 0 ) then
  !         print*, "mapX3_flav: encountered a a30,gtoq - corrected flavour assignment"
  !         !> See Table 2 arXiv:1301.4693, (i_q,1_g,j_qb) :> (1_q,J_g)
  !         !> jh2 flavour matched with [1_q] and should take quark flav.
  !         !> jh1 is mapped to final state and is gluon
  !         array_flav(npar_to)%flavours(jh1) = 0
  !         array_flav(npar_to)%flavours(jh2) = fh1
  !       end if
  !     else if (fu2 /= 0) then
  !       !> antenna: qQQ --> qg
  !       print*, "~~~~~~ antenna: qQQ --> qg (E30)"
  !       if (fh2 + fu1 /= 0 ) then
  !         stop "mapX3_flav: E30 antennta Q Qbar not equivalent flavours"
  !       end if
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = 0!this is fh2+fu1
  !     end if
  !   end if

  !   !> fix the direction of mapped flavour arrays
  !   array_flav(npar_to)%flavours(3:) = -array_flav(npar_to)%flavours(3:)
  !   !> Also fix direction of original flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)

  ! end subroutine mapX3_flav_old

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> flavour mapping subroutine for X30 antenna
  ! !
  ! !> @param[in] mapping information
  ! !-----------------------------------------------------------------------------
  ! subroutine mapX3_flav(npar_from,npar_to,mapping,antenna)
  !   integer, intent(in) :: npar_from, npar_to
  !   integer, dimension(3), intent(in) :: antenna
  !   integer, dimension(:,:), intent(in) :: mapping
  !   !integer, optional, intent(in) :: flavChange
  !   integer :: i, flavChange
  !   integer :: ih1,ih2, jh1,jh2, iu1
  !   integer :: fh1,fh2, fu1
  !   !> Introduce logical to determine which type of mapping we are in
  !   logical :: II ,IF
  !   II = .false.
  !   IF = .false.
  !   flavChange = 0
  !   print*, "mapX3_flav: Entered"
  !   !> First determine type of mapping
  !   if( any(antenna == 1) .and. any(antenna == 2) ) then
  !     II = .true.
  !   else if( any(antenna == 1) .or. any(antenna == 2) ) then
  !     IF = .true.
  !   end if
  !   !> determine mapped flavour indices
  !   ih1 = antenna(1) !hard 1
  !   iu1 = antenna(2) !unresolved parton
  !   ih2 = antenna(3) !hard 2

  !   !> all flavours incoming, rotate the final state flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)
  !   !> determe flavour of hard/unresolved partons
  !   fh1 = array_flav(npar_from)%flavours(ih1) !flavour of hard 1
  !   fu1 = array_flav(npar_from)%flavours(iu1) !flavour of unresolved 1
  !   fh2 = array_flav(npar_from)%flavours(ih2) !flavour of hard 2
  !   !> derive flavChange from antenna (is it an initial state parton)
  !   if( iu1 .lt. 3) flavChange = iu1

  !   if( flavChange /= 0 ) then
  !     print *, "mapX3_flav: flavChange = ", flavChange
  !     !> Check if hard radiatior 2 is initial (we must be in II mapping for this)
  !     if( ih2 == 1 .or. ih2 == 2) then
  !       ih1 = flavChange
  !       if( .not. II) then
  !         print *, "mapX3_flav: this should be an II mapping, error!"
  !         stop
  !       end if
  !     else
  !       !> Two special cases are IF a30gtoq and d30gtoq which are corrected below
  !       !> these are antenna of form [ i_q, 1_g, X ], where X = j_qb or X = j_g
  !       if( IF .and. fh1 /= 0 .and. fu1 == 0 ) then
  !         ih1 = flavChange
  !       else
  !         ih2 = flavChange
  !       end if
  !       !> Otherwise flavour 2 of mapped antenna should be transferred to initial state
  !     end if
  !   end if

  !   do i=1,npar_from
  !     if ( ih1==mapping(1,i) ) jh1 = mapping(2,i)
  !     if ( ih2==mapping(1,i) ) jh2 = mapping(2,i)
  !     !> ignore flavours in antenna, mapping given above
  !     if ( any(antenna==mapping(1,i)) ) cycle
  !     !> copy over flavours that are not involved in the mapping
  !     array_flav(npar_to)%flavours(mapping(2,i)) = array_flav(npar_from)%flavours(mapping(1,i))
  !   end do

  !   print*, "mapX3_flav: antenna", antenna
  !   print*, "mapX3_flav: mapping", mapping

  !   !Flavour mappings for X30
  !   array_flav(npar_to)%flavours(jh1) = fh1
  !   array_flav(npar_to)%flavours(jh2) = fh2+fu1

  !   !This doesn't make sense - typo in Table 2 of 1301.4693
  !   !Then special case of IF a30,gtoq - a flavour changing IF mapping of qgq > q g
  !   !if( IF .and. fh1 /= 0 .and. fu1 == 0 .and. fh2 /= 0 .and. flavChange /= 0) then
  !   !  array_flav(npar_to)%flavours(jh2) = 0
  !   !end if


  !   !> fix the direction of mapped flavour arrays
  !   array_flav(npar_to)%flavours(3:) = -array_flav(npar_to)%flavours(3:)
  !   !> Also fix direction of original flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)

  !   !> Copy PDF flavours from original
  !   call setPDF_flav(npar_to, 1, getPDF_flav(npar_from, 1))
  !   call setPDF_flav(npar_to, 2, getPDF_flav(npar_from, 2))

  ! end subroutine mapX3_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> flavour mapping subroutine for X40 antenna
  ! !
  ! !> @param[in] mapping information
  ! !-----------------------------------------------------------------------------
  ! subroutine mapX4_flav(npar_from,npar_to,mapping,antenna)
  !   integer, intent(in) :: npar_from, npar_to
  !   integer, dimension(4), intent(in) :: antenna
  !   integer, dimension(:,:), intent(in) :: mapping
  !   integer, dimension(2) :: flavChange
  !   integer :: i
  !   integer :: ih1,ih2, jh1,jh2, iu1,iu2
  !   integer :: fh1,fh2, fu1,fu2

  !   print*, "mapX4_flav: Entered"
  !   flavChange = (/0,0/)

  !   !> determine mapped flavour indices
  !   ih1 = antenna(1) !hard 1
  !   iu1 = antenna(2) !unresloved parton 1
  !   iu2 = antenna(3) !unresloved parton 2
  !   ih2 = antenna(4) !hard 2

  !   !> all flavours incoming, rotate the final state flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)
  !   !> determe flavour of hard/unresolved partons
  !   fh1 = array_flav(npar_from)%flavours(ih1) !flavour of hard 1
  !   fu1 = array_flav(npar_from)%flavours(iu1) !flavour of unresolved 1
  !   fu2 = array_flav(npar_from)%flavours(iu2) !flavour of unresolved 2
  !   fh2 = array_flav(npar_from)%flavours(ih2) !flavour of hard 2

  !   !> derive flavChange from antenna (is it an initial state parton)
  !   if( iu1 .lt. 3) flavChange(1) = iu1
  !   if( iu2 .lt. 3) flavChange(2) = iu2

  !   !> are we in a double flavChange limit (II mapping)
  !   if( all(flavChange /= 0) ) then
  !       ih1 = flavChange(1)
  !       ih2 = flavChange(2)
  !   !> a single flavour changing limit (either II or IF mapping)
  !   else if( any(flavChange /= 0) ) then
  !     !> Check if hard radiatior 2 is initial (must be in II mapping for this)
  !     !> If so, alter ih1 to other initial state
  !     if( ih2 == 1 .or. ih2 == 2) then
  !       if( flavChange(1) /= 0 ) ih1 = flavChange(1)
  !       if( flavChange(2) /= 0 ) ih1 = flavChange(2)
  !     else
  !       if( flavChange(1) /= 0 ) ih2 = flavChange(1)
  !       if( flavChange(2) /= 0 ) ih2 = flavChange(2)
  !     end if
  !   end if

  !   !> RG: sum over mapping up to npar_from. If not, can miss position of hard radiator
  !   do i=1,npar_from
  !     if ( ih1==mapping(1,i) ) jh1 = mapping(2,i)
  !     if ( ih2==mapping(1,i) ) jh2 = mapping(2,i)
  !     !> ignore flavours in antenna, mapping given above
  !     if ( any(antenna==mapping(1,i)) ) cycle
  !     !> copy over flavours that are not involved in the mapping
  !     array_flav(npar_to)%flavours(mapping(2,i)) = array_flav(npar_from)%flavours(mapping(1,i))
  !   end do

  !   print*, "mapX4_flav : antenna", antenna
  !   print*, "mapX4_flav : mapping", mapping
  !   print*, "mapX4_flav : hard rad.", fh1, fh2
  !   print*, "mapX4_flav : unresloved", fu1, fu2

  !   !> RG: for now just do flavour conserving situations
  !   if( any(flavChange /= 0) ) then
  !     print*, "mapX4_flav : flavour changing array", flavChange
  !   end if

  !   array_flav(npar_to)%flavours(jh1) = fh1
  !   array_flav(npar_to)%flavours(jh2) = fh2
  !   !> Now the case of the H40
  !   if( fh1 /= 0 .and. fh2 /= 0 .and. abs(fh1) /= abs(fh2) ) then
  !     array_flav(npar_to)%flavours(jh1) = 0
  !     array_flav(npar_to)%flavours(jh2) = 0
  !     if( fu1 == 0 .or. fu2 == 0) then
  !       print*, "map4_flav: Oh man, your H40 shouldn't be doing this"
  !       stop
  !     end if
  !   end if

  !   !> fix the direction of mapped flavour arrays
  !   array_flav(npar_to)%flavours(3:) = -array_flav(npar_to)%flavours(3:)
  !   !> Also fix direction of original flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)

  !   !> Copy PDF flavours from original
  !   call setPDF_flav(npar_to, 1, getPDF_flav(npar_from, 1))
  !   call setPDF_flav(npar_to, 2, getPDF_flav(npar_from, 2))

  ! end subroutine mapX4_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> flavour mapping subroutine for X40 antenna
  ! !
  ! !> @param[in] mapping information
  ! !-----------------------------------------------------------------------------
  ! subroutine mapX4_flav_old(npar_from,npar_to,mapping,antenna,flavChange)
  !   integer, intent(in) :: npar_from, npar_to
  !   integer, dimension(2), intent(in) :: flavChange
  !   integer, dimension(4), intent(in) :: antenna
  !   !RG: Should the mapping be alloctable dimension?
  !   integer, dimension(:,:), intent(in) :: mapping
  !   !integer, optional, intent(in) :: flavChange
  !   !> ih1, ih2
  !   integer :: i
  !   integer :: ih1,ih2, jh1,jh2, iu1,iu2
  !   integer :: fh1,fh2, fu1,fu2
  !   print*, "mapX4_flav: Entered"
  !   !> determine mapped flavour indices
  !   ih1 = antenna(lbound(antenna,1)) !hard 1
  !   ih2 = antenna(ubound(antenna,1)) !hard 2
  !   iu1 = antenna(lbound(antenna,1) +1) !unresloved parton 1
  !   iu2 = antenna(ubound(antenna,1) -1) !unresloved parton 2
  !   !> all flavours incoming, rotate the final state flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)
  !   !> determe flavour of hard/unresolved partons
  !   fh1 = array_flav(npar_from)%flavours(ih1) !flavour of hard 1
  !   fh2 = array_flav(npar_from)%flavours(ih2) !flavour of hard 2
  !   fu1 = array_flav(npar_from)%flavours(iu1) !flavour of unresolved 1
  !   fu2 = array_flav(npar_from)%flavours(iu2) !flavour of unresolved 2

  !   !> are we in a double flavChange limit (II mapping)
  !   if( all(flavChange /= 0) ) then
  !       ih1 = flavChange(1)
  !       ih2 = flavChange(2)
  !   !> a single flavour changing limit (either II or IF mapping)
  !   else if( any(flavChange /= 0) ) then

  !     !> IF mapping, initial state parton unresolved. both hard radiators map to correct final state
  !     if( flavChange(1) /= 0 ) ih2 = flavChange(1)
  !     if( flavChange(2) /= 0 ) ih2 = flavChange(2)

  !     !> II mapping, if original hard radiator 2 was initial state parton. I'll map both onto same position
  !     !> Therefore, check if they are mapping to same position
  !     do i=1,npar_from
  !       if( i==mapping(1,ih1) ) jh1 = mapping(2,i)
  !       if( i==mapping(1,ih2) ) jh2 = mapping(2,i)
  !     end do
  !     !> If they are mapping onto same position. Move ih2 > ih1 and re-assign ih2
  !     if( jh1 == jh2 ) then
  !       ih1 = ih2
  !       ih2 = antenna(ubound(antenna,1))
  !     end if
  !   end if

  !   !> RG: sum over mapping up to npar_from. If not, can miss position of hard radiator
  !   do i=1,npar_from
  !     if ( ih1==mapping(1,i) ) jh1 = mapping(2,i)
  !     if ( ih2==mapping(1,i) ) jh2 = mapping(2,i)
  !     !> ignore flavours in antenna, mapping given above
  !     if ( any(antenna==mapping(1,i)) ) cycle
  !     !> copy over flavours that are not involved in the mapping
  !     array_flav(npar_to)%flavours(mapping(2,i)) = array_flav(npar_from)%flavours(mapping(1,i))
  !   end do

  !   print*, "mapX4_flav : antenna", antenna
  !   print*, "mapX4_flav : mapping", mapping
  !   print*, "mapX4_flav : hard rad.", fh1, fh2
  !   print*, "mapX4_flav : unresloved", fu1, fu2

  !   !> RG: for now just do flavour conserving situations
  !   if( any(flavChange /= 0) ) then
  !     print*, "mapX4_flav : flavour changing array", flavChange
  !   end if

  !   if (size(antenna)/=4) then
  !     print*, "mapX4_flav : X40, wrong antenna size pal"
  !     stop
  !   end if

  !   print *, "jh1 map to >", jh1
  !   print *, "jh2 map to >", jh2

  !   !> from here on assume: X40
  !   !> [ g, X, Y, g ] block
  !   if (fh1==0 .and. fh2==0) then
  !     !> antenna: gggg --> gg
  !     if(fu1==0 .and. fu2==0) then
  !       print*, "~~~~~~ antenna: gggg --> gg (F40)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !     !> antenna: gQQbg --> gg
  !     else
  !       print*, "~~~~~~ antenna: gQQbg --> gg (G40)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !       if( abs(fu1) /= abs(fu2) ) print*, "Error G40: flavour of Q /= Qb"
  !     end if
  !   !>  [g , X, Y, q], or [q, X, Y, g]
  !   else if(fh1==0 .or. fh2==0) then
  !     !> antenna: q Q Qb g
  !     if( fu1/=0 .and. fu2/= 0) then
  !       print*, "~~~~~~ antenna: qQQbg --> qg (E40)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !       if( abs(fu1) /= abs(fu2) ) print*, "Error E40: flavour of Q /= Qb"
  !     !> antenna: q g g g
  !     else
  !       print*, "~~~~~~ antenna: qggg --> qg (D40)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2!this is fh2+fu1
  !     end if
  !   !> quark radiator antenna (A40, B40, C40)
  !   else
  !     !> antenna qggqb
  !     if( fu1==0 .and. fu2==0) then
  !       print*, "~~~~~~ antenna: qggqb --> qqb (A40)"
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2
  !     !> antenna qqbQQb (MUST be different quark flavours)
  !     else if( abs(fh1) /= abs(fh2) ) then
  !       print*, "~~~~~~ antenna: qqbQQb --> gg (H40)"
  !       array_flav(npar_to)%flavours(jh1) = 0 !this is fh1+fu1
  !       array_flav(npar_to)%flavours(jh2) = 0 !this is fh2+fu2
  !       if( (fh2+fu2) /= 0 ) print *, "mapX4_flav: warning in H40 no Q Qb > g?"
  !     else if( abs(fh1) == abs(fh2) ) then
  !       !> antenna qq'qb'qb
  !       array_flav(npar_to)%flavours(jh1) = fh1
  !       array_flav(npar_to)%flavours(jh2) = fh2
  !       if( abs(fh1) /= abs(fu1) ) print*, "~~~~~~ antenna: qqqbqb --> qqb (B40)"
  !       !> antenna qqqbqb
  !       if( abs(fh1) == abs(fu1) ) then
  !         print*, "~~~~~~ antenna: qqqbqb --> qqb (C40)"
  !         print*, "Flavour changing: check deltaF initial state = 0"
  !         !IF mapping, with flavour change, and four same flavour quarks
  !         !if( flavChange(1) /= 0 .and. flavChange(2) == 0 ) then
  !         !  if( fh2 /= array_flav(npar_from)%flavours( flavChange(1) ) ) then
  !         !    array_flav(npar_to)%flavours(jh1) = fh2
  !         !    array_flav(npar_to)%flavours(jh2) = fh1
  !         !  end if
  !         !Same as above for flav change on other initial parton
  !         !else if( flavChange(1) == 0 .and. flavChange(2) /= 0 ) then
  !         !  if( fh2 /= array_flav(npar_from)%flavours( flavChange(2) ) ) then
  !         !    array_flav(npar_to)%flavours(jh1) = fh2
  !         !    array_flav(npar_to)%flavours(jh2) = fh1
  !         !  end if
  !         !end if
  !         !> RG: Above implementation had been to 'correctly' assign flavours
  !         !> However, this is not correct since the antenna capture limits like:
  !         !> q {qb q^ qb} -> q qb^
  !         !> qb {qb q^ qb} -> qb q^
  !         !So in some limits can change initial state q^ to a qb^
  !       end if
  !     else
  !       print*, "mapX4_flav: did not find an antenna!"
  !     end if
  !   end if

  !   !> fix the direction of mapped flavour arrays
  !   array_flav(npar_to)%flavours(3:) = -array_flav(npar_to)%flavours(3:)
  !   !> Also fix direction of original flavours
  !   array_flav(npar_from)%flavours(3:) = -array_flav(npar_from)%flavours(3:)

  ! end subroutine mapX4_flav_old

!   !-----------------------------------------------------------------------------
!   !> @brief
!   !> flavour mapping subroutine for integrated (CalX) antenna
!   !
!   !> @param[in] mapping information
!   !-----------------------------------------------------------------------------
!   subroutine mapCalX_flav(npar, mapping, antenna)
!     use Mapping_mod, only: hasParent_map
!     integer, intent(in) :: npar
!     integer, dimension(2),   intent(in) :: antenna
!     integer, dimension(:,:), intent(in) :: mapping
!     integer, dimension(npar) :: unperm_flav
!     integer :: ip1, ip2
!     common /idparton/ip1,ip2
! !$omp threadprivate(/idparton/)

!     !> All flavours incoming, rotate the final state flavours
!     !> RG: define all reduced matrix-element as incoming flavours
!     !> Then this permutation not needed
!     !array_flav(npar)%flavours(3:) = -array_flav(npar)%flavours(3:)

!     !> Original flavour array
!     unperm_flav = array_flav(npar)%flavours(mapping(1,:))

!     !> Overwrite the original flavour array with the the permutation (mapping)
!     array_flav(npar)%flavours(mapping(2,:)) = unperm_flav(:)

!     !> Set PDF flavours according to the permuted initial states
!     if (.not.(array_flav(npar)%pdf_flavours_set)) then
!       call setPDF_flav(npar, 1, array_flav(npar)%flavours(1))
!       call setPDF_flav(npar, 2, array_flav(npar)%flavours(2))
!       array_flav(npar)%pdf_flavours_set = .true.
!     end if

!     !> Flavour-change to the predefined PDF flavours where applicable
!     !> Necessary for T, TNLO subtraction terms
!     if (antenna(1)/=0) call setPDF_flav(npar, antenna(1), ip1)
!     if (antenna(2)/=0) call setPDF_flav(npar, antenna(2), ip2)

!     !> Fix the direction of the permuted flavour arrays to be outgoing
!     array_flav(npar)%flavours(3:) = -array_flav(npar)%flavours(3:)

!   end subroutine mapCalX_flav


  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> Dirty fix: corrects flavours in case of reduced matrix-element mismatch
  ! !> that can happen when cancelling spurious singularities of antennae.
  ! !> Corrects free-floating gluons to the left and right to the quark string in
  ! !> case of BNg-type reduced matrix elements. Hopefully does nothing when other
  ! !> matrix element types are encountered. Should also work with decay as long
  ! !> as there are no unphysical reduced decay matrix elements.
  ! !
  ! !> @param[in] npar phase space where the correction should be performed on.
  ! !-----------------------------------------------------------------------------
  ! subroutine correct_flav(npar)
  !   integer, intent(in) :: npar
  !   integer :: me_flav(npar), i, q, ferm_stack, temp

  !   !> If permutation pointer has not been set, do no correction.
  !   if (.not. associated(perm_p)) return

  !   !> All-incoming flavour directions
  !   array_flav(npar)%flavours(3:) = -array_flav(npar)%flavours(3:)

  !   !> Order of flavours that match the matrix element called in the subtraction
  !   !> term.
  !   me_flav = array_flav(npar)%flavours(perm_p(:))
  !   print*, "correct_flav: me_flav = "

  !   print*, "matched flavour: ", me_flav

  !   !> Stack to keep track of whether we are sitting on a fermion line or not
  !   !> while going through the partons from the left to right in the following
  !   !> do loop.
  !   ferm_stack = 0

  !   do i=1,npar
  !     if (me_flav(i)/=0) then
  !       !> Encountered a non-gluon (fermion)
  !       me_flav(i) = me_flav(i)/abs(me_flav(i))
  !       ferm_stack = ferm_stack + me_flav(i)
  !     else
  !       !> Encountered a gluon
  !       if (ferm_stack == 0) then
  !         !> Free-floating gluons found
  !         if (i==1 .and. me_flav(i+1) > 0) then
  !           !> Only i==1, quark to the right (stack is empty)
  !           q = i+1
  !           print*, "correct_flav: floating gluon found, swapping to the right!"
  !         else if (me_flav(i-1) < 0 .and. ferm_stack==0) then
  !           !> Every other case, anti-quark to the left (stack is empty)
  !           print*, "correct_flav: floating gluon found, swapping to the left!"
  !           q = i-1
  !         else
  !           print*, "correct_flav: not expecting this kind of configuration!"
  !           stop
  !         end if
  !         !> Do the swapping
  !         temp = array_flav(npar)%flavours(perm_p(q))
  !         array_flav(npar)%flavours(perm_p(q)) = array_flav(npar)%flavours(perm_p(i))
  !         array_flav(npar)%flavours(perm_p(i)) = temp
  !       end if
  !     end if
  !   end do

  !   !> Fix the directions
  !   array_flav(npar)%flavours(3:) = -array_flav(npar)%flavours(3:)

  ! end subroutine correct_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> lock flavour buffer, preventing reading from and writing to it.
  ! !> this is useful for double virtual subtraction terms where qcdnorm and thus
  ! !> push_flav is being called multiple times before calling bino, thus pop_flav
  ! !
  ! !> @return flavour id of the tagging
  ! !-----------------------------------------------------------------------------
  ! subroutine lockBuffer_flav()
  !   if ( .not.isActive_flav() ) return
  !   buffer_lock_flav = .true.
  ! end subroutine lockBuffer_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> release buffer lock
  ! !
  ! !> @return flavour id of the tagging
  ! !-----------------------------------------------------------------------------
  ! subroutine releaseBuffer_flav()
  !   if ( .not.isActive_flav() ) return
  !   buffer_lock_flav = .false.
  ! end subroutine releaseBuffer_flav

  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> query buffer lock
  ! !
  ! !> @return flavour id of the tagging
  ! !-----------------------------------------------------------------------------
  ! logical function isBufferLocked_flav()
  !   isBufferLocked_flav = buffer_lock_flav
  ! end function isBufferLocked_flav


end module Flavour_mod
