#ifdef USE_PINEAPPL


!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: PineAPPL interface for NNLOJET [external members: `_pine`]
!
! DESCRIPTION:
!> Routines to manage & generate PineAPPL grids
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module PineAPPL_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only : dp
  use pineappl
  use Channels_mod
  implicit none
  private

  public :: init_pine
  public :: register_lumi_pine
  public :: book_grid_pine
  public :: fill_grid_pine
  public :: write_grid_pine

  !>
  integer :: nlumi_
  type(pineappl_lumi) :: lumi_
  integer, dimension(:), allocatable :: lumi_idx_

  integer :: norders_, ord_idx_(0:2)
  integer, dimension(:), allocatable :: orders_

  type(pineappl_keyval) :: key_vals_

  type :: Pine_t
    integer :: id = -1
    type(pineappl_grid) :: grid
    type(Pine_t), pointer :: next => null()
  end type Pine_t

  type(Pine_t), pointer :: head_ => null()


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise PineAPPL module
  !-----------------------------------------------------------------------------
  subroutine init_pine()
    use Run_mod, only : qproductionSweep_run
    use Channels_mod, only : contrib_nO_chan, contrib_nV_chan
    use Process_mod, only : n_alphas_proc, n_alpha_proc
    use ProcLib_mod, only : init_pine_dl
    integer :: ichan, iord, nV, min_nO, max_nO, max_nV_ord(0:2)

    !> NOTE: init_chan is now called *before* Histogram parsing
    !> this should be all set before calling book_grid_pine

    if ( allocated(lumi_idx_) ) then
      stop "init_pine: lumi_idx_ already allocated?!"
    end if
    allocate(lumi_idx_(1:ntot_chan))
    lumi_idx_(:) = -1  ! invalid state
    lumi_ = pineappl_lumi_new()
    nlumi_ = 0

    !> determine the orders & virtual integrations
    min_nO = -1
    max_nO = -1
    max_nV_ord(:) = -1
    do ichan=1,ntot_chan
      if ( .not. qactive_chan(ichan) ) cycle
      iord = contrib_nO_chan( contrib_chan(ichan) )
      if (min_nO<0 .or. iord<min_nO) min_nO = iord
      if (max_nO<0 .or. iord>max_nO) max_nO = iord
      nV = contrib_nV_chan( contrib_chan(ichan) )
      if (max_nV_ord(iord)<0 .or. nV>max_nV_ord(iord)) max_nV_ord(iord) = nV
    end do

    !> determine order array (pure QCD for now)
    if (min_nO<0 .or. max_nO<0) then
      stop "init_pine: no channel selected?! (min/max ord < 0)"
    end if
    ord_idx_(:) = -1
    norders_ = 0
    if (min_nO == 0) then
      ord_idx_(0) = norders_
      norders_ = norders_ + 1
    end if
    if (min_nO <= 1 .and. max_nO >= 1) then
      ord_idx_(1) = norders_
      norders_ = norders_ + 1
      if (max_nV_ord(1) >= 1)  norders_ = norders_ + 2
    end if
    if (min_nO <= 2 .and. max_nO >= 2) then
      ord_idx_(2) = norders_
      norders_ = norders_ + 1
      if (max_nV_ord(2) >= 1)  norders_ = norders_ + 2
      if (max_nV_ord(2) >= 2)  norders_ = norders_ + 3
    end if

    if ( allocated(orders_) ) then
      stop "init_pine: orders_ already allocated?!"
    end if
    !> follows same ordering as in ScaleCoeff::extract_sclc
    allocate(orders_(1:4*norders_))
    if ( ord_idx_(0) >= 0 ) then
      iord = 4*ord_idx_(0)
      orders_(iord+1:iord+4) =  &
        &  [ n_alphas_proc, n_alpha_proc, 0, 0 ]
    end if
    if ( ord_idx_(1) >= 0 ) then
      iord = 4*ord_idx_(1)
      orders_(iord+1:iord+4) =  &
        &  [ n_alphas_proc+1, n_alpha_proc, 0, 0 ]
      if (max_nV_ord(1) >= 1) then
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+1, n_alpha_proc, 1, 0 ]
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+1, n_alpha_proc, 0, 1 ]
      end if
    end if
    if ( ord_idx_(2) >= 0 ) then
      iord = 4*ord_idx_(2)
      orders_(iord+1:iord+4) =  &
        &  [ n_alphas_proc+2, n_alpha_proc, 0, 0 ]
      if (max_nV_ord(2) >= 1) then
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+2, n_alpha_proc, 1, 0 ]
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+2, n_alpha_proc, 0, 1 ]
      end if
      if (max_nV_ord(2) >= 2) then
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+2, n_alpha_proc, 2, 0 ]
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+2, n_alpha_proc, 1, 1 ]
        iord = iord + 4
        orders_(iord+1:iord+4) =  &
          &  [ n_alphas_proc+2, n_alpha_proc, 0, 2 ]
      end if
    end if

    ! if (min_nO == 0) then
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc, n_alpha_proc, 0, 0 ]
    !   iord = iord + 4
    ! end if
    ! if (min_nO <= 1 .and. max_nO >= 1) then
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+1, n_alpha_proc, 0, 0 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+1, n_alpha_proc, 1, 0 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+1, n_alpha_proc, 0, 1 ]
    !   iord = iord + 4
    ! end if
    ! if (min_nO <= 2 .and. max_nO >= 2) then
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 0, 0 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 1, 0 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 0, 1 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 2, 0 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 1, 1 ]
    !   iord = iord + 4
    !   orders_(iord+1:iord+4) =  &
    !     &  [ n_alphas_proc+2, n_alpha_proc, 0, 2 ]
    !   iord = iord + 4
    ! end if
    ! if ( iord /= 4*norders_ ) then
    !   print*, "init_pine: inconsistent iord:", iord, norders_
    !   stop
    ! end if

    !> default interpolation binning parameters
    key_vals_ = pineappl_keyval_new()

    !> this is the place to override the default grid node settings
    !call pineappl_keyval_set_int(key_vals_, "x_bins", 100)

    !> run the process-specific channel registration
    call init_pine_dl()

  end subroutine init_pine


  !-----------------------------------------------------------------------------
  !> @brief
  !> book a grid (== histogram)
  !
  !> @param[in] id  unique id number for the histogram
  !> @param[in] bins  bin edges of the histogram (# = nbins+1)
  !> @param[in] name  file name for the grid
  !-----------------------------------------------------------------------------
  subroutine register_lumi_pine(channels, nlumi, lumi, facs)
    integer, intent(in), dimension(:) :: channels, lumi
    real(dp), intent(in), dimension(:) :: facs
    integer, intent(in) :: nlumi

    if ( .not.allocated(lumi_idx_) ) then
      stop "register_lumi_pine: forgot to run init_pine?!"
    end if

    if ( any(qactive_chan(channels)) ) then
      call pineappl_lumi_add(lumi_, nlumi, lumi, facs)
      lumi_idx_(channels) = nlumi_  ! PineAPPL counting from 0
      nlumi_ = nlumi_ + 1
      print*, "register_lumi_pine: ", nlumi_, ">", lumi
    end if

  end subroutine register_lumi_pine


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise PineAPPL module
  !-----------------------------------------------------------------------------
  subroutine destroy_pine()
    type(Pine_t),  pointer :: it, it_next
    if ( allocated(lumi_idx_) )  deallocate(lumi_idx_)
    if ( allocated(orders_) )    deallocate(orders_)
    it => head_
    do while(associated(it))
      it_next => it%next  ! save before we kill it
      deallocate(it)
      it => it_next  ! loop
    end do
    nullify(head_)
  end subroutine destroy_pine


  !-----------------------------------------------------------------------------
  !> @brief
  !> book a grid (== histogram)
  !
  !> @param[in] id  unique id number for the histogram
  !> @param[in] nbins  number of bins
  !> @param[in] boundaries  bin edges of the histogram (# = nbins+1)
  !-----------------------------------------------------------------------------
  subroutine book_grid_pine(id, nbins, boundaries)
    use ScaleCoeff_mod
#ifdef USE_OMP
    use omp_lib
#endif
    integer, intent(in) :: id, nbins
    real(dp), intent(in), dimension(:) :: boundaries
    type(Pine_t), pointer :: newPine, it

    if (.not.isActive_sclc()) then
      print*, "book_grid_pine: switching on scale coefficients"
      call setActive_sclc(.true.)
    end if

#ifdef USE_OMP
    !> the moment I book a grid, I'm in production mode and
    !> better run on a single thread for safety
    if (omp_get_max_threads() > 1) then
      print*, "book_grid_pine: resetting # threads to ONE!"
      call omp_set_num_threads(1)
    end if
#endif

    if (.not.allocated(lumi_idx_))  call init_pine()

    allocate(newPine)
    newPine%id = id
    newPine%grid = pineappl_grid_new(lumi_, norders_, orders_, nbins, boundaries, key_vals_)
    !> push to stack
    newPine%next => head_
    head_ => newPine

    print*, "book_grid_pine", id, ">"
    print*, "> ", norders_, "|", orders_
    print*, "> ", nbins, "|", boundaries

  end subroutine book_grid_pine


  !-----------------------------------------------------------------------------
  !> @brief
  !> fill weights into grid
  !
  !> @param[in] id  unique id number for the grid/histogram
  !> @param[in] valObs  value of the observable to be binned
  !> @param[in] wgtcoeff  array of weights (as&PFD stripped log coefficients)
  !-----------------------------------------------------------------------------
  subroutine fill_grid_pine(id, valObs, wgtcoeff)
    use Channels_mod, only : current_chan, contrib_chan, ix_chan, currentContrib_chan
    use KinData_mod, only : x1_kin, x2_kin, xreg1_kin, xreg2_kin
    use Scales_mod, only : muR2_scl, muF2_scl
    use Process_mod, only : n_alphas_proc, n_alpha_proc
    use Constants_mod, only : twopi
    use Parameters_mod, only : alpha_param
    integer, intent(in) :: id
    real(dp), intent(in) :: valObs
    real(dp), intent(in), dimension(:) :: wgtcoeff
    type(Pine_t), pointer, save :: it => null()
    logical :: qcycled
    real(dp) :: x1, x2, Q2, fac
    integer :: nas, nV, nO, iord, order_idx, lumi_idx

    ! print*, "fill_grid_pine:", id, valObs,":", wgtcoeff
    ! read(*,*)
    !> we do it in a strange way with save'd pointers because
    !> the order in which we add grids matches the hsitograms
    !> thus we should be iterating the grids in successive orders
    !> since we don't start at the head, we have to allow for one
    !> cycle to happen around the back
    qcycled = .false.
    if (.not.associated(it)) then
      it => head_
      qcycled = .true.
    end if
    !> loop over grids to find matching id
    do while(it%id /= id)
      it => it%next
      if (.not.associated(it)) then
        if (.not.qcycled) then
          it => head_
          qcycled = .true.
        else
          print*, "fill_grid_pine: couldn't find the histogram id", id
          stop
        end if
      end if
    end do

    call currentContrib_chan(nas=nas, nO=nO, nV=nV)
    ! fac = (1d0/twopi)**nas * (1d0/alpha_param())**n_alpha_proc
    fac = (1d0/twopi)**nas
    !> @todo: maybe to this at the end using `pineappl_grid_scale_by_order`

    select case (ix_chan)
    case(1)
       x1 = x1_kin
       x2 = x2_kin
    case(2)
       x1 = x1_kin
       x2 = x2_kin / xreg2_kin
    case(3)
       x1 = x1_kin / xreg1_kin
       x2 = x2_kin
    case(4)
       x1 = x1_kin / xreg1_kin
       x2 = x2_kin / xreg2_kin
    case default
      print*, "x1_hook: invalid ix value:", ix_chan
      stop
    end select
    if ( muR2_scl(1) /= muF2_scl(1) ) then
      print*, "fill_grid_pine: central muR & muF scale need to be the same", muR2_scl(1), muF2_scl(1)
      stop
    end if

    Q2 = muR2_scl(1)

    lumi_idx = lumi_idx_(current_chan)

    !print*, "pineappl_grid_fill", x1, x2, Q2, ord_idx_(nO), valObs, lumi_idx
    if (nV >= 0) then
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO),   valObs, lumi_idx, fac*wgtcoeff(1))
    end if
    if (nV >= 1) then
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO)+1, valObs, lumi_idx, fac*wgtcoeff(2))
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO)+2, valObs, lumi_idx, fac*wgtcoeff(3))
    end if
    if (nV >= 2) then
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO)+3, valObs, lumi_idx, fac*wgtcoeff(4))
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO)+4, valObs, lumi_idx, fac*wgtcoeff(5))
      call pineappl_grid_fill(it%grid, x1, x2, Q2, ord_idx_(nO)+5, valObs, lumi_idx, fac*wgtcoeff(6))
    end if

  end subroutine fill_grid_pine


  !-----------------------------------------------------------------------------
  !> @brief
  !> write the grid to disk
  !
  !> @param[in] id  unique id number for the histogram
  !> @param[in] name  file name
  !> @param[in] factor  optional re-scaling of entire grid (for normalisation)
  !> @param[in] obsName  name of the observable
  !-----------------------------------------------------------------------------
  subroutine write_grid_pine(id, name, factor, obsName)
    use Process_mod, only: beam1_proc, beam2_proc
    integer, intent(in) :: id
    character(*), intent(in) :: name
    real(dp), intent(in), optional :: factor
    character(*), intent(in), optional :: obsName
    type(Pine_t), pointer :: it
    integer :: idx
    character(128) :: beam_str

    print*, "write_grid_pine", id, "[",name,"]", factor

    it => head_
    do while(it%id /= id)
      it => it%next
      if (.not.associated(it)) then
        print*, "write_grid_pine: couldn't find the histogram id", id
        stop
      end if
    end do

    if (present(obsName))  call pineappl_grid_set_key_value(it%grid, 'x1_label', trim(obsName))
    call pineappl_grid_set_key_value(it%grid, 'y_unit', 'fb')

    !> copy over beam information
    write(beam_str, '(I0)') beam1_proc
    call pineappl_grid_set_key_value(it%grid, 'initial_state_1', trim(beam_str))
    write(beam_str, '(I0)') beam2_proc
    call pineappl_grid_set_key_value(it%grid, 'initial_state_2', trim(beam_str))

    !> correct normalisation
    if (present(factor))  call pineappl_grid_scale(it%grid, factor)

    !> optimize the grid representation (makes the file smaller)
    call pineappl_grid_optimize(it%grid)

    !> write the grid to disk - with `.lz4` suffix the grid is automatically LZ4 compressed
    idx = index(name, ".dat")
    if (idx > 1) then
      call pineappl_grid_write(it%grid, trim(name(:idx-1))//'.pineappl.lz4')
    else
      call pineappl_grid_write(it%grid, trim(name)//'.pineappl.lz4')
    end if

  end subroutine write_grid_pine


end module PineAPPL_mod


#endif
