!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: HadronFragmentation [external members: `_hfrag`]
!
! DESCRIPTION:
!> Module providing infrastructure and functions for photon fragmentation
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module HadronFragmentation_mod
!-------------------------------------------------------------------------------
  use, intrinsic :: iso_c_binding
  use Kinds_mod, only: dp
  implicit none
  private

  public :: setActive_hfrag, isActive_hfrag, setIndex_hfrag
  public :: init_hfrag, destroy_hfrag
  public :: unset_hfrag, set_hfrag, isSet_hfrag
  public :: idx_hfrag, iz_hfrag, ip1_hfrag, icorr1_hfrag
  public :: digest_hfrag

  public :: D_hfrag

  logical :: qactive_hfrag = .false.  ! off by default
  integer :: idx_hfrag = 0  ! position of the identified parton in mom set

  !> store local state if we're dealing with a fragmentation term
  logical :: qset_hfrag = .false.  ! off by default
  !$OMP THREADPRIVATE(qset_hfrag)
  integer :: fctn_hfrag  ! pick the parametrisation/extraction we want

  !> analogue of ix:
  !>  1: ~ delta(1-zreg)  <->  Hfrag(z)
  !>  2: ~ regular        <->  Hfrag(z/zreg)
  integer :: iz_hfrag = -1
!$OMP THREADPRIVATE(iz_hfrag)

  !> ip1_hfrag: the identified parton flavour that fragments
  !>    0 -> g
  !> +/-1 -> (anti-)down:  +/- {d,s,b}
  !> +/-2 -> (anti-)up:    +/- {u,c}
  !> +/-3 -> (anti-)quark: +/- {d,u,s,c,b}
  !> icorr1_hfrag: 0 if no corrections; [1|2] to indicate PDF correlation
  integer :: ip1_hfrag, icorr1_hfrag
!$OMP THREADPRIVATE(ip1_hfrag, icorr1_hfrag)

  !> fragmentation set (LHAPDF)
  character(:), allocatable :: lhapdf_set_hfrag
  integer :: lhapdf_mem_hfrag = -1  ! init to an invalid state
  !> pointer to LHAPDF objects
  type(c_ptr) :: lhapdf_hfrag = c_null_ptr
  !$omp threadprivate(lhapdf_hfrag)
  !> set constraints
  real(dp) :: xmin, xmax, qmin, qmax

  !> which fragmentation provider should I use?
  !> -1: from LHAPDF
  !>  0: toy
  !>  1: BKK
  !>  2: KKKS08
  !>  3: toy heavy quark
  integer :: provider_hfrag = 1
  
  !> -1: quark/antiquarks = 1, gluon = 0
  !>  0: quark/antiquarks = 1, gluon = 1
  !>  1: quark/antiquarks = 0, gluon = 1
  integer :: toy_param_hfrag = 0

  !> ih   = 1  : LO (pi^+ + pi^-)/2
  !> ih   = 2  : LO (K^+ + K^-)/2   
  !> ih   = 3  : LO (K^0 + K^0_bar)/2   
  !> ih   = 5  : LO  pi^0  
  !> ih   = 7  : LO  h^+ + h^-       as sum of pions and kaons
  !> ih   = 10 : NLO (pi^+ + pi^-)/2
  !> ih   = 20 : NLO (K^+ + K^-)/2   
  !> ih   = 30 : NLO (K^0 + K^0_bar)/2   
  !> ih   = 50 : NLO  pi^0  
  !> ih   = 70 : NLO  h^+ + h^-       as sum of pions and kaons 
  integer :: bkk_param_hfrag = 70

  !> ih   =  1 : D^*+
  !> ih   =  2 : D^0
  !> ih   =  3 : D^+
  !> ih   = -1 : D^*-
  !> ih   = -3 : D^-
  !> ih   =  0 : sum over charmed hadrons
  integer :: kkks_param_hfrag = 0
  
  !> Simple heavy quark, only one option
  integer :: toyhq_param_hfrag = 0

  !> interface to C LHAPDF routines (copied from LHAPDF.f90)
  interface

    function m_mkPDF(setname, member) result(pdf_p) bind(c, name='mkPDF_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr) :: pdf_p
      character(kind=c_char), dimension(*) :: setname
      integer(c_int) :: member
    end function m_mkPDF

    subroutine m_rmPDF(pdf_p) bind(c, name='rmPDF_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p  ! pass by value; otherwise double-ref in cxx: (LHAPDF::PDF**)
    end subroutine m_rmPDF

    real(c_double) function m_xMin(pdf_p) bind(c, name='xMin_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_xMin

    real(c_double) function m_xMax(pdf_p) bind(c, name='xMax_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_xMax

    real(c_double) function m_qMin(pdf_p) bind(c, name='qMin_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_qMin

    real(c_double) function m_qMax(pdf_p) bind(c, name='qMax_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_qMax

    subroutine m_evolveQ(pdf_p, x, Q, xfxQ) bind(c, name='evolvePDFQ_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
      real(c_double), intent(in) :: x, Q
      real(c_double), dimension(-6:6), intent(out) :: xfxQ
    end subroutine m_evolveQ

  end interface


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate fragmentation
  !
  !> @param[in] set  string for the name of Hfrag
  !> @param[in] mem  id for the set member
  !-----------------------------------------------------------------------------
  subroutine setActive_hfrag(set, mem)
    character(len=*) :: set
    integer, optional :: mem
    !> Activate hfrag
    qactive_hfrag = .true.
    !> Use fake frag fun
    if (set == 'TOY') then
      provider_hfrag = 0
      if ( present(mem) ) then 
        toy_param_hfrag = mem
      end if
      return 
    end if
    !> Use internal BKK set
    if (set == 'BKK') then
      provider_hfrag = 1
      if ( present(mem) ) then 
        bkk_param_hfrag = mem
      end if
      return
    end if
    !> Use internal KKKS08 set
    if (set == 'KKKS') then
      provider_hfrag = 2
      if ( present(mem) ) then 
        kkks_param_hfrag = mem
      end if
      return
    end if   
    !> Use internal toy (simple) heavy quark set
    if (set == 'TOYHQ') then
      provider_hfrag = 3
      if ( present(mem) ) then
        toyhq_param_hfrag = mem
      end if
      return
    end if
    !> Use LHAPDF
    provider_hfrag = -1
    lhapdf_set_hfrag = set
    if ( present(mem) ) then
      lhapdf_mem_hfrag = mem
    else
      lhapdf_mem_hfrag = 0
    end if
  end subroutine setActive_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of the fragmentation
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_hfrag()
    isActive_hfrag = qactive_hfrag
  end function isActive_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the momentum index of the identified parton that fragments
  !
  !> @param[in] idx entry in momentum set
  !-----------------------------------------------------------------------------
  subroutine setIndex_hfrag(idx)
    integer, intent(in) :: idx
    idx_hfrag = idx
  end subroutine setIndex_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise LHAPDF pointers
  !-----------------------------------------------------------------------------
  subroutine init_hfrag()
    if (.not.isActive_hfrag() .or. provider_hfrag>=0) return
    if ( c_associated(lhapdf_hfrag) ) then
      print*, "init_hfrag: lhapdf_hfrag is not null?!"
      print*, "calling destroy_hfrag... (you should be doing this yourself!)"
      call destroy_hfrag()
    end if
    lhapdf_hfrag = m_mkPDF(lhapdf_set_hfrag // c_null_char, lhapdf_mem_hfrag)
    xmin = m_xMin(lhapdf_hfrag)
    xmax = m_xMax(lhapdf_hfrag)
    qmin = m_qMin(lhapdf_hfrag)
    qmax = m_qMax(lhapdf_hfrag)
  end subroutine init_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy dynamically allocated stuff from LHAPDF
  !-----------------------------------------------------------------------------
  subroutine destroy_hfrag()
    if (.not.isActive_hfrag() .or. provider_hfrag>=0) return
    if ( c_associated(lhapdf_hfrag) ) call m_rmPDF(lhapdf_hfrag)
    lhapdf_hfrag = c_null_ptr
  end subroutine destroy_hfrag


  !> @TODO: DIGEST & add to writer!


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the current state of the fragmentation
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine unset_hfrag(npar)
    use KinData_mod, only : resetFrag_kin, clearCacheLocal_kin
    integer, intent(in) :: npar

    if ( .not.isActive_hfrag() ) return

    call resetFrag_kin(npar)
    call clearCacheLocal_kin(npar)

    qset_hfrag = .false.

  end subroutine unset_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the current state of the fragmentation
  !
  !> @param[in] npar momentum set
  !> @param[in] frag parton index that fragments
  !> @param[inout] ipass check ipass state
  !> @param[in] isDfrag optional flag to identify blocks containing frag funcs
  !>            default = .false.
  !-----------------------------------------------------------------------------
  subroutine set_hfrag(npar, frag, ipass, isDfrag)
    use KinData_mod, only : setFrag_kin, resetFrag_kin, clearCacheLocal_kin
    integer, intent(in) :: npar, frag
    integer, intent(inout) :: ipass
    logical, intent(in), optional :: isDfrag
    logical :: isDfrag_

    if ( .not.isActive_hfrag() ) then
      ipass = 0
      return
    end if

    if ( (iz_hfrag/=1) .and. (iz_hfrag/=2) ) then
      print*, "set_hfrag: invalid iz_hfrag", iz_hfrag
      stop
    end if

    !> check if sub-contribution selection triggers
    isDfrag_ = .false.  ! = default
    if ( present(isDfrag) ) isDfrag_ = isDfrag

    if (ipass==0) return

    !> I *need* the z-value stored in KinData (for conv function evaluation)
    call setFrag_kin(npar, frag)  ! z = convolution
    !> but for the delta-piece, need to reset iflag so obs are computed correctly
    !> iz_hfrag == 1 is now the endpoint z=1
    !>if ( iz_hfrag == 1 ) call resetFrag_kin(npar)  ! delta piece!

    call clearCacheLocal_kin(npar)

    qset_hfrag = .true.

  end subroutine set_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire if we're in a fragmentation convolution block
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isSet_hfrag()
    isSet_hfrag = .false.
    if ( .not. isActive_hfrag() ) return
    isSet_hfrag = qset_hfrag
  end function isSet_hfrag


  !-----------------------------------------------------------------------------
  !> @brief
  !> fragmentation from LHAPDF
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA [GeV] fragmentation scale
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D_hfrag(pid, z, muA)
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA
    real(dp), dimension(-6:6) :: zfzQ
    D_hfrag = 0d0
    if (provider_hfrag == 0) then
      D_hfrag = D_hfrag_toy(pid, z) 
      !print*, "D_hfrag", pid, z, muA, ">", D_hfrag
      return
    end if
    if (provider_hfrag == 1) then
      D_hfrag = D_hfrag_bkk(pid, z, muA) 
      !print*, "D_hfrag", pid, z, muA, ">", D_hfrag
      return
    end if
    if (provider_hfrag == 2) then
      D_hfrag = D_hfrag_kkks(pid, z, muA) 
      !print*, "D_hfrag", pid, z, muA, ">", D_hfrag
      return
    end if
    if (provider_hfrag == 3) then
      D_hfrag = D_hfrag_toyhq(pid, z, muA)
      !print*, "D_hfrag", pid, z, muA, ">", D_hfrag
      return
    end if    
    if ((z   < xmin) .or. (z   > xmax)) return
    if ((muA < qmin) .or. (muA > qmax)) return
    call m_evolveQ(lhapdf_hfrag, z, muA, zfzQ)
    D_hfrag = zfzQ(pid) / z
    !print*, "D_hfrag", pid, z, muA, ">", D_hfrag
  end function D_hfrag

  !-----------------------------------------------------------------------------
  !> @brief
  !> toy fragmentation functions for debug purposes
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D_hfrag_toy(pid, z)
    integer, intent(in) :: pid
    double precision, intent(in) :: z
    D_hfrag_toy = 0d0
    select case (toy_param_hfrag)
    !> -1: quark/antiquarks = 1, gluon = 0
    case (-1)
       if (pid==0) then
          D_hfrag_toy = 0d0
          return
       else
          D_hfrag_toy = 1d0
          return
       end if
    !>  0: quark/antiquarks = 1, gluon = 1         
    case (0)
       D_hfrag_toy = 1d0
       return
    !>  1: quark/antiquarks = 0, gluon = 1
    case (1)
       if (pid==0) then
          D_hfrag_toy = 1d0
          return
       else
          D_hfrag_toy = 0d0
          return
       end if
    case default
       print*, "HadronFragmentation: wrong option to toy hfrag: ", toy_param_hfrag
       stop
    end select
  end function D_hfrag_toy


  !-----------------------------------------------------------------------------
  !> @brief
  !> wrapper to TOY heavy-quark fragmentation function
  !> Use for tests of WJ_h process 
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA [GeV] fragmentation scale
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D_hfrag_toyhq(pid, z, muA)
    use Constants_mod, only : pi, gammaE
    !use Scales_mod
    use LHAPDF_mod
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA
    double precision :: t, beta0qcd, epsilon0, epsilon0hat
    integer :: ih, iorder, nf
    D_hfrag_toyhq = 0d0
    nf = 4
    beta0qcd = (33d0 - 2*nf)/(12*pi)
    t = 5d-1/pi/beta0qcd*dlog(alphasQ_lhapdf(127d-2)/alphasQ_lhapdf(muA))
    epsilon0 = 2d0*t
    epsilon0hat = 3d0/2d0*t

    D_hfrag_toyhq = exp(gammaE*epsilon0)*exp(epsilon0hat)/dlgam(1d0+epsilon0)*epsilon0*(1d0-z)**(-1d0+epsilon0)

  end function D_hfrag_toyhq

  !-----------------------------------------------------------------------------
  !> @brief
  !> Log of Gamma function
  !> Used in D_hfrag_toyhq
  !
  !> @return  
  !-----------------------------------------------------------------------------
  function dlgam(xx)
! real logarithm of gamma function
      implicit real*8 (a-h,o-z)
      integer j
      real*8 cof(6),stp,gt,g,cst,tmp,x,xpgt,xmp5,s0,r10
      data cof,stp/76.18009173d0,-86.50532033d0,24.01409822d0, -1.231739516d0,.120858003d-2,-.536382d-5,2.50662827465d0/
      data gt,g/5.5d0,5.0d0/
      data cst/4.081061466d0/
      x = xx - 1
      xpgt = x + gt
      xmp5  = x + .5d0
      s0 = 1
      do 1 j=1,6
        x = x + 1
        tmp = cof(j)/x
        s0  = s0 + tmp
  1   continue
      r10 = log(s0)
      dlgam = xmp5*(log(xpgt)-1) + r10 - cst
      return
  end

  !-----------------------------------------------------------------------------
  !> @brief
  !> wrapper to BKK fragmentation function
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA [GeV] fragmentation scale  
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D_hfrag_bkk(pid, z, muA)
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA
    integer :: ih,iorder
    real(dp), dimension(0:10) :: dh
    D_hfrag_bkk = 0d0
    if (mod(bkk_param_hfrag,10) == 0) then
       iorder = 1
       ih = bkk_param_hfrag/10
    else
       iorder = 0
       ih = bkk_param_hfrag
    end if
    !> the BKK set should alredy return the number density
    call bkk(ih,iorder,z,muA,dh)
    ! Parton label:
    !! 0    1    2    3    4    5    6    7    8     9    10
    !! g    u   ubar  d   dbar  s   sbar  c   cbar   b   bbar
    select case (pid)  
    case (-6) !top case is 0
       D_hfrag_bkk = 0d0
       return
    case (-5)
       D_hfrag_bkk = dh(10)
       return
    case (-4)
       D_hfrag_bkk = dh(8)
       return
    case (-3)
       D_hfrag_bkk = dh(6)
       return
    case (-2)
       D_hfrag_bkk = dh(2)
       return
    case (-1)
       D_hfrag_bkk = dh(4)
       return
    case (0)
       D_hfrag_bkk = dh(0)
       return
    case (1)
       D_hfrag_bkk = dh(3)
       return
    case (2)
       D_hfrag_bkk = dh(1)
       return
    case (3)
       D_hfrag_bkk = dh(5)
       return
    case (4)
       D_hfrag_bkk = dh(7)
       return
    case (5)
       D_hfrag_bkk = dh(9)
       return         
    case (6)
       D_hfrag_bkk = 0d0
       return       
    case default
       print*, "HadronFragmentation: wrong pid in bkk hfrag: ", pid
       stop
    end select
  end function D_hfrag_bkk

  !-----------------------------------------------------------------------------
  !> @brief
  !> wrapper to KKKS08 fragmentation function
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA [GeV] fragmentation scale  
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D_hfrag_kkks(pid, z, muA)
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA
    integer :: ih,iorder
    real(dp), dimension(0:10) :: DH
    real(dp), dimension(0:10) :: DH1,DH2,DH3
    double precision :: muA2
    integer :: i
    integer :: NO
    double precision :: Dg,Dl,Dc,Db
    D_hfrag_kkks = 0d0
    NO = 1 ! include mass effects
    muA2 = muA*muA
    ! sum over charmed hadrons    
    if (kkks_param_hfrag.eq.0) then
       call kkks08_D(NO,1,z,muA2,DH1) 
       call kkks08_D(NO,2,z,muA2,DH2)
       call kkks08_D(NO,3,z,muA2,DH3)
       do i=0,10
          DH(i)=DH1(i)+DH2(i)+DH3(i)
       enddo
    elseif (abs(kkks_param_hfrag).gt.3) then
       print*, "HadronFragmentation: wrong member in kkks hfrag: ", kkks_param_hfrag
       stop
    elseif (abs(kkks_param_hfrag).eq.2) then
       !> the KKKS08 set should alredy return the number density
       call kkks08_D(NO,abs(kkks_param_hfrag),z,muA2,DH) 
    else
       !> the KKKS08 set should alredy return the number density
       call kkks08_D(NO,abs(kkks_param_hfrag),z,muA2,DH)
       !> Here I make an assumption: in 0712.0481, they provide D^+ and D^*+ FFs,
       !> hence the c FF (and b FF) must be different from the cbar FF (and bbar FF).
       !> Instead, in the kkks08 routine, they seem to *not* distinguish charges, D_i = D_ibar
       !> and D_c = D_cbar, D_b = D_bbar, D_u=D_d=D_s.
       !> I assume that they have summed the result, and that:
       !> 
       !> D+
       !! 0    1    2    3    4    5    6    7    8     9    10
       !! g    u   ubar  d   dbar  s   sbar  c   cbar   b   bbar
       !! Dg   Dl   Dl   Dl   Dl  Dl    Dl   Dc  Dl     Dl   Db
       !> 
       !> and
       !>
       !> D-
       !! 0    1    2    3    4    5    6    7    8     9    10
       !! g    u   ubar  d   dbar  s   sbar  c   cbar   b   bbar
       !! Dg   Dl   Dl   Dl   Dl  Dl    Dl   Dl  Dc     Db   Dl
       !>
       !> (Dcbar and Dbbar are not equal to Dl, as they have less evolution,
       !> but they are small anyway, so we can assume they are all the same)
       !>
       !> Hence I can extract Dg, Dl, Dc and Db.
       Dg = DH(0)/2d0
       Dl = DH(1)/2d0
       Dc = DH(7)-Dl
       Db = DH(9)-Dl
       if (kkks_param_hfrag < 0) then
          DH(0)   = Dg
          DH(1:6) = Dl
          DH(7)   = Dl
          DH(8)   = Dc
          DH(9)   = Db
          DH(10)  = Dl
       else
          DH(0)   = Dg
          DH(1:6) = Dl
          DH(7)   = Dc
          DH(8)   = Dl
          DH(9)   = Dl
          DH(10)  = Db
       endif
    endif
    ! Parton label:
    !! 0    1    2    3    4    5    6    7    8     9    10
    !! g    u   ubar  d   dbar  s   sbar  c   cbar   b   bbar
    select case (pid)
    case (-6)
       D_hfrag_kkks = 0d0
    case (-5)
       D_hfrag_kkks = DH(10)
       return
    case (-4)
       D_hfrag_kkks = DH(8)
       return
    case (-3)
       D_hfrag_kkks = DH(6)
       return
    case (-2)
       D_hfrag_kkks = DH(2)
       return
    case (-1)
       D_hfrag_kkks = DH(4)
       return
    case (0)
       D_hfrag_kkks = DH(0)
       return
    case (1)
       D_hfrag_kkks = DH(3)
       return
    case (2)
       D_hfrag_kkks = DH(1)
       return
    case (3)
       D_hfrag_kkks = DH(5)
       return
    case (4)
       D_hfrag_kkks = DH(7)
       return
    case (5)
       D_hfrag_kkks = DH(9)
       return
    case (6)
       D_hfrag_kkks = 0d0
    case default
       print*, "HadronFragmentation: wrong pid in kkks hfrag: ", pid
       stop
    end select
  end function D_hfrag_kkks
  
  !-----------------------------------------------------------------------------
  !> @brief
  !> output info on hadron fragmentation
  !-----------------------------------------------------------------------------
  subroutine digest_hfrag(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",1x,A,T30,A,T103,"*")'
    character(len=*), parameter :: fmt_sweep =  &
      & '(1x,"*",T10,A," = ",I0,"[",I0,"]",T103,"*")'
    character(len=*), parameter :: fmt_pdf =  &
      & '(1x,"*",5x,"PDF",I0," = ",A," [mem = ",I0,"]",T103,"*")'
    character(len=*), parameter :: fmt_opt_str =  &
      & '(1x,"*",5x,A," = ",A,T103,"*")'
    character(len=*), parameter :: fmt_opt_dble =  &
      & '(1x,"*",5x,A," = ",1f6.1,T103,"*")'
    character(len=*), parameter :: fmt_opt_int =  &
      & '(1x,"*",5x,A," = ",I6,T103,"*")'
    integer :: type, ncall, nit

    if (.not.isActive_hfrag()) return

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_str) ""
    write(iout,fmt_str) "> Hadron fragmentation <"
    write(iout,fmt_str) ""

    !> TODO: to be improved    
    if (provider_hfrag == 2) then
       write(iout,fmt_str) "", "FFset: KKKS08"
    end if
    if (provider_hfrag == 1) then
       write(iout,fmt_str) "", "FFset: BKK"
    end if    
    if (provider_hfrag == 0) then
       write(iout,fmt_str) "", "FFset: TOY"
    end if
    if (provider_hfrag == -1) then
       write(iout,fmt_str) "", "FFset: LHAPDF"
    end if
    
  end subroutine digest_hfrag  
  
end module HadronFragmentation_mod






