!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Histograms [external members: `_hist`]
!
! DESCRIPTION:
!> Module storing all the histogram data and providing the interface for
!> everything that we need to do with them.
!-------------------------------------------------------------------------------

!> Preprocessor variable to switch on/off compensated summation in the
!> histogram buffer. Tests show that there is a small performance penalty
!> but no visible improvement, so it's deactivated by default.
!> Uncomment to switch on the error compensation.
#define HISTBUFFKAHAN
#define HISTKAHAN

!-------------------------------------------------------------------------------
module Histograms_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only : int64
  use BinData_mod
  use Selectors_mod
  use Scales_mod
  use Reweight_mod
  implicit none
  private

  public :: n_hist
  public :: setActive_hist, isActive_hist
  public :: init_hist, destroy_hist
  public :: register_hist, registerSub_hist
  public :: digest_hist
  public :: setPrefix_hist, setSuffix_hist
  public :: initBuffer_hist, destroyBuffer_hist
  public :: writeBuffer_hist, flushBuffer_hist
  public :: clear_hist
  public :: writeToFile_hist

  integer, parameter :: max_hist_name_length = 128
  ! flag controlling if histogram accumulation is active
  ! (should be turned off in warmup to save resources)
  logical :: qactive = .true.

  character(len=128) :: filePrefix = ""
  character(len=128) :: fileSuffix = ".dat"

  !> write out all histograms into a single file
  logical :: qsingle = .false.
  character(len=128) :: fileSingle = ""

  type SubHistogram_t
    integer :: obsId = 0
    !> selectors
    ! selector options when an invalid observable is encountered
    !   0: do not bin the event to the histogram at all
    !   1: ignore the given selector and continue with the next & bin
    integer :: selOpt = 0
    type(Selector_t), pointer :: sel_p => null()
    !> set the run_group: = 0 (default) always filled!
    integer :: igroup = 0
    !> reweighting function
    type(ReweightFunction_t) :: rewgtfcn
    !> scale override (initialise to invalid values <-> "unset")
    double precision :: fac_mu0 = 0d0
    integer :: obsId_mu0 = -1
    !> linked list
    type(SubHistogram_t), pointer :: next => null()
  end type SubHistogram_t

  type Histogram_t
    character(len=max_hist_name_length) :: name = ""
    integer :: ihist = 0 ! unique identifier
    !> flag to set the histogram type
    !   -1: composite
    !    0: cross
    !   >0: single observable (=obsId)
    integer :: histType = 0
    integer :: nbins = -1
    !> flag to set the BinData type
    !    0: no channel breakdown: only scales
    !    1:  channel breakdown + scales
    integer :: dataType = 0  ! default: no channel breakdown
    !> flag to switch to a cumulant distribution
    !    0: no channel breakdown: only scales
    !   -1:  cumulant from:    xmin       -->  xi[upper]
    !   +1:  cumulant from:    xi[lower]  -->  xmax
    integer :: icumulant = 0  ! default: no cumulant
    !> set the run_group: = 0 (default) always filled!
    integer :: igroup = 0
    integer :: ndata  ! # of BinData entries
    double precision, allocatable, dimension(:) :: boundaries
    integer, allocatable, dimension(:) :: histCount
    !> (0:2) x (1:ndata) x (0:nbins)  ! with HISTKAHAN
    !> (1:2) x (1:ndata) x (0:nbins)  ! w/o  HISTKAHAN
    double precision, allocatable, dimension(:,:,:) :: histData
    !> selectors
    ! selector options when an invalid observable is encountered
    !   0: do not bin the event to the histogram at all
    !   1: ignore the given selector and continue with the next & bin
    integer :: selOpt = 0
    type(Selector_t), pointer :: sel_p => null()
    !> reweighting function
    type(ReweightFunction_t) :: rewgtfcn
    !> scale override (initialise to invalid values <-> "unset")
    double precision :: fac_mu0 = 0d0
    integer :: obsId_mu0 = -1
    !> sub-histograms (for composite histograms)
    integer :: nsubHist = 0
    type(SubHistogram_t), pointer :: subHist_p => null()
    !> linking to PDF grid-interpolation libraries
    ! flag linking the histogram for grid production
    !   0: off
    !   1: APPLgrid
    !   2: fastNLO
    !   3: both (APPLfast)
    !   4: PineAPPL
    integer :: igrid = 0
    ! name only needed in init... in principle can avoid storing it
    character(len=max_hist_name_length) :: gridname = ""
    !> linked list
    type(Histogram_t), pointer :: next => null()
  end type Histogram_t

  ! keep track of the head of the histogram stack and set counter
  type(Histogram_t), pointer :: head => null()
  integer :: n_hist = 0


  type HistBuffer
    !> HAS TO BE OUTSIDE OF HISTOGRAM_T TO BE DECLARED THREADPRIVATE
    ! for now the buffer is a copy of the full histogram array
    ! obviously this is a bit of a waste since one phase-space point
    ! will typically only fill a handful of different bins
    ! if this turns out to be an issue one could replace it by a
    ! self-balancing binary tree (red-black tree)
    ! https://en.wikipedia.org/wiki/Red%E2%80%93black_tree
    ! https://www.myroms.org/blog/?p=169
    ! either way this is the critical component for OMP parallelization
    ! so we need to be careful. But since this module is a "black box"
    ! supplying a general interface noone needs to know how the data structure
    ! is organized internally and we can swap it out at any time.
    ! actually, repeated allocation and deallocation is not a good idea...
    ! for now, leave it as it is
    !> (0:1) x (1:ndata) x (0:nbins)  ! with HISTBUFFKAHAN
    !> (1:1) x (1:ndata) x (0:nbins)  ! w/o  HISTBUFFKAHAN
    double precision, allocatable, dimension(:,:,:) :: histBuff
    type(Histogram_t), pointer :: histParent
  end type HistBuffer

  type(HistBuffer), allocatable, dimension(:) :: buffer_hist
!$OMP THREADPRIVATE(buffer_hist)

  !> @interface
  !> @brief
  !> this is the external interface to register new histograms
  !> which is overloaded to the [Equal|Variabla]Size procedures
  interface register_hist
    module procedure newVariableWidth
    module procedure newEqualWidth
  end interface


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate histogram binning
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_hist(active)
    logical, intent(in) :: active
    qactive = active
  end subroutine setActive_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of histogram binning
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_hist()
    isActive_hist = qactive
  end function isActive_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new histogram with variable-sized bin widths
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] boudaries array holding all bin boundaries
  !> @param[in] name optional file name to be used in the data files
  !>            (default = observable name)
  !> @param[in] outputType optional flag for channel breakdown settings
  !> @param[in] cumulant optional flag for cumulant distributions
  !> @param[in] run_group optional id for a run group (see MultiRun)
  !> @param[in] selectorsOption optional flag to define selectors treatment
  !> @param[in] selectors_p optional pointer to the head of a selectors
  !>            (linked list)
  !> @param[in] gridname optional filename for
  !>            APPLgrid(`*.appl`)/fastNLO(`*.fast`) link
  !> @param[in] rewgtfcn optional reweighting function
  !> @param[in] fac_mu0 optional factor of the scale override
  !> @param[in] obsId_mu0 optional observable id of the scale override
  !> @param[out] ihist optional return value giving the histogram id
  !>             in the case a COMPOSITE histogram was registered (-1 otherwise)
  !-----------------------------------------------------------------------------
  subroutine newVariableWidth( obsName, boundaries, name,  &
                             & selectorsOption, selectors_p,  &
                             & outputType, cumulant, run_group,  &
                             & gridname, rewgtfcn, fac_mu0, obsId_mu0,  &
                             & ihist)
    use Observables_mod
    use StrHelper_mod, only : to_upper_str
    use MultiRun_mod, only : hasGroup_mult
#ifdef USE_PINEAPPL
    use PineAPPL_mod
#endif
    character(len=*), intent(in) :: obsName
    ! always map the boundaries array to (0:nbins)
    double precision, dimension(0:), intent(in) :: boundaries
    character(len=*), optional, intent(in) :: name
    integer, optional, intent(in) :: outputType
    integer, optional, intent(in) :: cumulant
    integer, optional, intent(in) :: run_group
    integer, optional, intent(in) :: selectorsOption
    type(Selector_t), optional, pointer, intent(in) :: selectors_p
    character(len=*), optional, intent(in) :: gridname
    type(ReweightFunction_t), optional, intent(in) :: rewgtfcn
    double precision, optional, intent(in) :: fac_mu0
    integer, optional, intent(in) :: obsId_mu0
    integer, optional, intent(out) :: ihist
    integer :: histType
    type(Histogram_t), pointer :: newHist, it
    integer :: nbins
    integer :: i
    character(:), allocatable :: gridsuffix


    if (.not.associated(head)) then
      print*, "register_hist: call init_hist first"
      return
    end if

    ! print*, ">>> register_hist obsName >", obsName, "<"
    ! print*, ">>> register_hist name    >", name, "<"
    ! print*, ">>> register_hist boundaries ", boundaries, "<"


    if ( to_upper_str(obsName) == "CROSS" ) then
      histType = 0
    else if ( to_upper_str(obsName) == "COMPOSITE" ) then
      histType = -1
    else
      histType = getIdFromName_obs(obsName)
      if (histType < 1) then ! jet observable
        print*, "register_hist: invalid observable: ", obsName
        stop
      end if
    end if


    !> #(boundaries) = #(bins)+1
    nbins = size(boundaries)-1

    allocate(newHist)

    if ( present(name) .and. name.ne.'' ) then
      newHist%name = name
    else
      newHist%name = obsName
    end if
    !> check if name or gridname already exists
    it => head
    do while ( associated(it) )
      if ( it%name == newHist%name ) then
        print*, "register_hist: histogram filename clash: ", newHist%name
        stop
      end if
      if ( present(gridname) .and. gridname.ne.'' .and. it%gridname == gridname ) then
        print*, "register_hist: histogram gridname clash: ", gridname
        stop
      end if
      it => it%next
    end do
    newHist%histType = histType
    newHist%nbins = nbins

    newHist%dataType = 0  ! default: no channel breakdown
    if ( present(outputType) ) then
      newHist%dataType = outputType
    end if
    newHist%ndata = ndata_bin( newHist%dataType )

    newHist%icumulant = 0  ! default: no cumulant distribution
    if ( present(cumulant) ) then
      newHist%icumulant = cumulant
    end if

    newHist%igroup = 0  ! default: bin all runs
    if ( present(run_group) ) then
      if (.not. hasGroup_mult(run_group) ) then
        print*, "register_hist: error setting run_group:", run_group
        stop
      end if
      newHist%igroup = run_group
    end if

    allocate(newHist%boundaries(0:nbins))
    forall(i=0:nbins) newHist%boundaries(i) = boundaries(i)

    ! bin "0" = overflow
    ! bin "1" = [ boundaries(0), boundaries(1) ]
    ! bin "2" = ( boundaries(1), boundaries(2) ]
    !      :

    allocate(newHist%histCount(0:nbins))
    newHist%histCount = 0

#ifdef HISTKAHAN
    !> 1st dimension: 1 = val, 2 = val**2, 0 = error (comp. summation)
    allocate( newHist%histData(0:2,1:newHist%ndata,0:nbins) )
    newHist%histData = 0d0
#else
    !> 1st dimension: 1 = val, 2 = val**2
    allocate( newHist%histData(1:2,1:newHist%ndata,0:nbins) )
    newHist%histData = 0d0
#endif

    if (present(selectorsOption)) then
      newHist%selOpt = selectorsOption
    else
      newHist%selOpt = 0  ! do not bin event in case of `.not.valid`
    end if

    if (present(selectors_p)) then
      newHist%sel_p => selectors_p
    end if

    if (present(rewgtfcn)) then
      newHist%rewgtfcn = rewgtfcn
    end if

    if (present(fac_mu0)) then
      newHist%fac_mu0 = fac_mu0
    end if

    if (present(obsId_mu0)) then
      newHist%obsId_mu0 = obsId_mu0
    end if

    ! push it to the stack
    newHist%ihist = head%ihist + 1
    newHist%next => head
    head => newHist
    n_hist = n_hist + 1
    if (present(ihist)) then
      ihist = -1
      !> only return id if we have a composite histogram
      if (histType == -1) ihist = newHist%ihist
    end if

    ! parse optional gridname argument
    ! file suffix:
    !    `*.appl`:  APPLgrid
    !    `*.fast`:  fastNLO
    !    `*.both`:  both
    !    `*.pine`:  PineAPPL
    newHist%igrid = 0
    if ( present(gridname) .and. gridname.ne.'' ) then
      newHist%gridname = gridname
      gridsuffix = trim(adjustl( newHist%gridname ))
      do
        i = index(gridsuffix, ".")
        if (i.lt.1) exit
        gridsuffix = trim(adjustl( gridsuffix(i+1:) ))
      end do
      select case (to_upper_str(gridsuffix))
      case ("APPL")
        newHist%igrid = 1
      case ("FAST")
        newHist%igrid = 2
      case ("BOTH")
        newHist%igrid = 3
      case ("PINE")
        newHist%igrid = 4
      case default
        print*, "unknown grid suffix: ", gridname, " [",gridsuffix,"]"
      end select
      !> book the grid here!
      !> note: the equal-size-bins case `newEqualWidth` is mapped to this routine
      !> available information:
      !>    * which mode?:                               `newHist%igrid`
      !>    * unique integer id for the histogram:       `newHist%ihist`
      !>    * name of the observable:                    `getNameFromId_obs(newHist%obsId)`
      !>      [not unique: can be booked multiple times, which is why we have `ihist` above]
      !>    * array of bin boundaries for the histogram: `newHist%boundaries`
      !>    * name of the gridfile:                      `newHist%gridname`
!$OMP CRITICAL(grid_book)
#ifdef USE_PINEAPPL
      call book_grid_pine(newHist%ihist, newHist%nbins, newHist%boundaries)
#else
      !call book_grid( newHist%igrid, newHist%ihist, nbins, newHist%boundaries(0), newHist%boundaries(nbins), gridname )
      call book_grid_varbins( newHist%igrid, newHist%ihist, nbins, newHist%boundaries(0), gridname )
#endif
!$OMP END CRITICAL(grid_book)
    end if

  end subroutine newVariableWidth


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new histogram with equal-sized bin widths
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] nbins number of bins
  !> @param[in] lower lower histogram boundary
  !> @param[in] upper upper histogram boundary
  !> @param[in] type optional specification whether we want lin/log binning
  !> @param[in] name optional file name to be used in the data files
  !>            (default = observable name)
  !> @param[in] outputType optional flag for channel breakdown settings
  !> @param[in] cumulant optional flag for cumulant distributions
  !> @param[in] run_group optional id for a run group (see MultiRun)
  !> @param[in] selectorsOption optional flag to define selectors treatment
  !> @param[in] selectors_p optional pointer to the head of a selectors
  !>            (linked list)
  !> @param[in] gridname optional filename for linking:
  !>            APPLgrid(`*.appl`)
  !>            fastNLO(`*.fast`)
  !>            APPLfast(`*.both`)
  !>            PineAPPL(`*.pine`)
  !> @param[in] rewgtfcn optional reweighting function
  !> @param[in] fac_mu0 optional factor of the scale override
  !> @param[in] obsId_mu0 optional observable id of the scale override
  !> @param[out] ihist optional return value giving the histogram id
  !>             in the case a COMPOSITE histogram was registered (-1 otherwise)
  !-----------------------------------------------------------------------------
  subroutine newEqualWidth( obsName, nbins,lower,upper, binning, name,  &
                          & selectorsOption, selectors_p,  &
                          & outputType, cumulant, run_group,  &
                          & gridname, rewgtfcn, fac_mu0, obsId_mu0,  &
                          & ihist)
    character(len=*), intent(in) :: obsName
    integer, intent(in) :: nbins
    double precision, intent(in) :: lower, upper
    integer, optional, intent(in) :: binning
    character(len=*), optional, intent(in) :: name
    integer, optional, intent(in) :: outputType
    integer, optional, intent(in) :: cumulant
    integer, optional, intent(in) :: run_group
    integer, optional, intent(in) :: selectorsOption
    type(Selector_t), optional, pointer, intent(in) :: selectors_p
    character(len=*), optional, intent(in) :: gridname
    type(ReweightFunction_t), optional, intent(in) :: rewgtfcn
    double precision, optional, intent(in) :: fac_mu0
    integer, optional, intent(in) :: obsId_mu0
    integer, optional, intent(out) :: ihist
    double precision :: delta, fac
    ! #(boundaries) = #(bins)+1
    double precision, dimension(0:nbins) :: boundaries
    integer :: i, ibinning

    if ( lower > upper ) stop "newEqualWidth: min > max?!"

    ibinning = 0  ! default: linear binning
    if (present(binning)) ibinning = binning

    select case(ibinning)

    case(0)  ! linear
      delta = (upper-lower)/dble(nbins)
      boundaries(0) = lower
      do i = 1,nbins-1
        boundaries(i) = lower + i*delta
      end do
      boundaries(nbins) = upper

    case(1)  ! logarithmic
      if (lower <= 0d0) stop "newEqualWidth: non-positive values encountered in log binning"
      fac = (upper/lower)**(1d0/dble(nbins))
      boundaries(0) = lower
      do i = 1,nbins-1
        boundaries(i) = lower * fac**dble(i)
      end do
      boundaries(nbins) = upper

    case default
      stop "newEqualWidth: unkwon type"
    end select

    ! map the EqualSize case to the VariableSize case
    call newVariableWidth( obsName, boundaries, name,  &
                         & selectorsOption, selectors_p,  &
                         & outputType, cumulant, run_group,  &
                         & gridname, rewgtfcn, fac_mu0, obsId_mu0,  &
                         & ihist )

  end subroutine newEqualWidth


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new sub-histogram
  !
  !> @param[in] ihist histogram id to add the subhistogram to
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] selectorsOption optional flag to define selectors treatment
  !> @param[in] selectors_p optional pointer to the head of a selectors
  !>            (linked list)
  !> @param[in] run_group optional id for a run group (see MultiRun)
  !> @param[in] rewgtfcn optional reweighting function
  !> @param[in] fac_mu0 optional factor of the scale override
  !> @param[in] obsId_mu0 optional observable id of the scale override
  !-----------------------------------------------------------------------------
  subroutine registerSub_hist( ihist, obsName,  &
                             & selectorsOption, selectors_p,  &
                             & run_group,  &
                             & rewgtfcn,  &
                             & fac_mu0, obsId_mu0 )
    use Observables_mod, only : isComposite_obs, getIdFromName_obs
    use StrHelper_mod, only : to_upper_str
    use MultiRun_mod, only : hasGroup_mult
    integer, intent(in) :: ihist
    character(len=*), intent(in) :: obsName
    integer, optional, intent(in) :: selectorsOption
    type(Selector_t), optional, pointer, intent(in) :: selectors_p
    integer, optional, intent(in) :: run_group
    type(ReweightFunction_t), optional, intent(in) :: rewgtfcn
    double precision, optional, intent(in) :: fac_mu0
    integer, optional, intent(in) :: obsId_mu0
    integer :: obsId
    type(Histogram_t), pointer :: hist
    type(SubHistogram_t), pointer :: newSubHist

    hist => head
    do while ( associated(hist) )
      if ( hist%ihist == ihist) exit
      hist => hist%next
    end do

    if ( .not. associated(hist) ) then
      print*, "registerSub_hist: no histogram with id = ", ihist
      stop
    end if

    if ( to_upper_str(obsName) == "CROSS" ) then
      obsId = 0
      if (hist%nbins /= 0) then
        print*, "registerSub_hist: cross only possible if parent is a composite-cross"
        stop
      end if
    else
      obsId = getIdFromName_obs(obsName)
      if (obsId < 1) then ! jet observable
        print*, "registerSub_hist: invalid observable: ", obsName
        stop
      end if
      if ( isComposite_obs(obsId) ) then
        print*, "registerSub_hist: composite observable not allowed as sub-histogram: ", obsName
        stop
      end if
      if (hist%nbins == 0) then
        print*, "registerSub_hist: trying to bin observables in a composite-cross histogram?!"
        stop
      end if
    end if

    allocate(newSubHist)
    newSubHist%obsId = obsId

    if (present(selectorsOption)) then
      newSubHist%selOpt = selectorsOption
    else
      newSubHist%selOpt = 0  ! do not bin event in case of `.not.valid`
    end if

    if (present(selectors_p)) newSubHist%sel_p   => selectors_p

    newSubHist%igroup = 0  ! default: bin all runs
    if ( present(run_group) ) then
      if (.not. hasGroup_mult(run_group) ) then
        print*, "register_hist: error setting run_group:", run_group
        stop
      end if
      newSubHist%igroup = run_group
    end if

    if (present(rewgtfcn))  newSubHist%rewgtfcn  = rewgtfcn
    if (present(fac_mu0))   newSubHist%fac_mu0   = fac_mu0
    if (present(obsId_mu0)) newSubHist%obsId_mu0 = obsId_mu0

    !> push it to the list
    newSubHist%next => hist%subHist_p
    hist%subHist_p => newSubHist
    hist%nsubHist = hist%nsubHist + 1

  end subroutine registerSub_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the histograms module by adding the global "storewgt" histogram
  !
  !> @param[in] singleName  optional filename for single-file output
  !-----------------------------------------------------------------------------
  subroutine init_hist(singleName)
    character(len=*), intent(in), optional :: singleName
    type(Histogram_t), pointer :: newHist
    if (associated(head)) then
      print*, "init_hist: already initialised, call destroy_hist first"
      return
    end if
    allocate(newHist)
    newhist%name     = "cross"
    newhist%ihist    = 0
    newhist%histType = 0
    newhist%nbins    = 0
    newhist%dataType = 1  ! always full channel breakdown for "cross"
    newhist%ndata = ndata_bin( newhist%dataType )
    allocate(newHist%boundaries(0:0))
    newHist%boundaries = 0d0
    allocate(newHist%histCount(0:0))
    newHist%histCount = 0
#ifdef HISTKAHAN
    allocate( newHist%histData(0:2,1:newhist%ndata,0:0) )
    newHist%histData = 0d0
#else
    allocate( newHist%histData(1:2,1:newhist%ndata,0:0) )
    newHist%histData = 0d0
#endif
    ! pop it to the stack
    newHist%next => head
    head => newHist
    n_hist = n_hist + 1

    !> do we want to write out all histograms into a single file?
    if ( present(singleName) ) then
      qsingle = .true.
      fileSingle = singleName
      print*, "init_hist: all output into: ", fileSingle, qsingle
    end if

  end subroutine init_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all registered histograms
  !-----------------------------------------------------------------------------
  subroutine destroy_hist()
    type(Histogram_t), pointer :: it, next
    type(SubHistogram_t), pointer :: it_sub, next_sub
    type(Selector_t), pointer :: it_sel, next_sel
    it => head
    do while ( associated(it) )
      ! save next before we destroy it
      next => it%next
      ! deallocate sub-entries and then the entry itself
      deallocate(it%boundaries)
      deallocate(it%histCount)
      deallocate(it%histData)
      !----- kill selectors if registered
      it_sel => it%sel_p
      do while ( associated(it_sel) )
        next_sel => it_sel%next  ! save next
        deallocate(it_sel)
        it_sel => next_sel  ! loop
      end do
      it%sel_p => null()
      !----- kill SubHistograms
      it_sub => it%subHist_p
      do while ( associated(it_sub) )
        next_sub => it_sub%next  ! save next
        !----- kill selectors of SubHistograms if registered
        it_sel => it_sub%sel_p
        do while ( associated(it_sel) )
          next_sel => it_sel%next  ! save next
          deallocate(it_sel)
          it_sel => next_sel  ! loop
        end do
        it_sub%sel_p => null()
        !-----
        deallocate(it_sub)
        it_sub => next_sub  ! loop
      end do
      it%subHist_p => null()
      !-----
      deallocate(it)
      it => next  ! loop
    end do
    head => null()
    n_hist = 0
    ! if there are no histograms we should also destroy the buffers
    if (allocated(buffer_hist)) then
      call destroyBuffer_hist()
    end if
  end subroutine destroy_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to initialize the (threadprivate) buffer for the histograms
  !-----------------------------------------------------------------------------
  subroutine initBuffer_hist()
    type(Histogram_t), pointer :: it
    integer :: i
    allocate(buffer_hist(n_hist))
    it => head
    i = 1
    do while ( associated(it) )
#ifdef HISTBUFFKAHAN
      allocate( buffer_hist(i)%histBuff(0:1,1:it%ndata,0:it%nbins) )
#else
      allocate( buffer_hist(i)%histBuff(1:1,1:it%ndata,0:it%nbins) )
#endif
      buffer_hist(i)%histBuff = 0d0
      buffer_hist(i)%histParent => it
      it => it%next
      i = i + 1
    end do
  end subroutine initBuffer_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to destroy the (threadprivate) buffer for the histograms
  !-----------------------------------------------------------------------------
  subroutine destroyBuffer_hist()
    integer :: i
    do i=1,n_hist
      deallocate(buffer_hist(i)%histBuff)
    end do
    deallocate(buffer_hist)
  end subroutine destroyBuffer_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the registered histograms
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_hist(unit)
    use Observables_mod
    use Selectors_mod
    integer, optional, intent(in) :: unit
    integer :: iout
    type(Histogram_t), pointer :: it
    type(SubHistogram_t), pointer :: it_sub
    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_cross =  &
      & '(1x,"*",1x,I1,1x,A,T30,"(cross)",T103,"*")'
    character(len=*), parameter :: fmt_sub   =  &
      & '(1x,"*",13x,"+ ",A,T35," x ",A,T103,"*")'
    character(len=*), parameter :: fmt       =  &
      & '(1x,"*",1x,I1,1x,A,T30,"(",A15," in ",I4," bins)",5x"["1f12.3,", ",1f12.3,"]",T103,"*")'
    !> for selectors
    type(Selector_t), pointer :: it_sel

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    if ( associated(head) ) then
      it => head
      do while ( associated(it) )
        !> output by type
        if ( it%histType == 0 ) then
          write(iout,fmt_cross) it%dataType, it%name
        else if ( it%histType < 0 ) then
          write(iout,fmt) it%dataType, it%name, "COMPOSITE", &
            &             it%nbins, it%boundaries(0), it%boundaries(size(it%boundaries)-1)
        else
          write(iout,fmt) it%dataType, it%name, getNameFromId_obs(it%histType), &
            &             it%nbins, it%boundaries(0), it%boundaries(size(it%boundaries)-1)
        end if
        !> output optional reweighting function
        if ( it%rewgtfcn%qactive ) then
          write(iout,'(1x,"*",T30," x ",A50,T103,"*")') toString_rewgt(it%rewgtfcn)
        end if
        !> output possible cumulant option
        if ( it%icumulant /= 0 ) then
          write(iout,*) "* cumulant = ", it%icumulant
        end if
        !> output possible run_group option
        if ( it%igroup /= 0 ) then
          write(iout,*) "* run_group = ", it%igroup
        end if
        !> output optional grid association
        if ( it%igrid /= 0 ) then
          write(iout,*) "* igrid = ", it%igrid, ": ", it%gridname
        end if

        !>--- digest Histogram selectors
        if ( associated(it%sel_p) ) then
          it_sel => it%sel_p
          do while ( associated(it_sel) )
            call digest_hist_print_selector(it_sel, it%selOpt)
            it_sel => it_sel%next
          end do
        end if
        !>---

        !>----- loop sub-histograms
        it_sub => it%subHist_p
        do while ( associated(it_sub) )
          if (it_sub%obsId==0) then
            write(iout,fmt_sub) "(cross)", toString_rewgt(it_sub%rewgtfcn)
          else
            write(iout,fmt_sub) getNameFromId_obs(it_sub%obsId), toString_rewgt(it_sub%rewgtfcn)
          end if
          !> output possible run_group option
          if ( it_sub%igroup /= 0 ) then
            write(iout,*) "* run_group = ", it_sub%igroup
          end if
          !>--- digest SubHistogram selectors
          if ( associated(it_sub%sel_p) ) then
            it_sel => it_sub%sel_p
            do while ( associated(it_sel) )
              call digest_hist_print_selector(it_sel, it_sub%selOpt)
              it_sel => it_sel%next
            end do
          end if
          !>---
          it_sub => it_sub%next
        end do
        !>-----

        write(iout, fmt_blank)
        it => it%next
      end do
    else
      write(iout, '(1x,"*",3x,"no histogram registered so far",T103,"*")')
    endif

    contains

      subroutine digest_hist_print_selector(sel, opt)
        type(Selector_t) :: sel
        integer :: opt
        character(len=6) :: seltype
        character(len=*), parameter :: fmt_min_max = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &T30,1f12.3," <= ",T48,A15,T65," <= ",1f12.3,T103,"*")'
        character(len=*), parameter :: fmt_min     = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &T30,1f12.3," <= ",T48,A15,                  T103,"*")'
        character(len=*), parameter :: fmt_max     = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &                  T48,A15,T65," <= ",1f12.3,T103,"*")'

        seltype = ''
        if ( sel%qinvert ) then
          seltype = 'reject'
        else
          seltype = 'accept'
        endif
        if (sel%qlower .and. sel%qupper) then
          write(iout,fmt_min_max) sel%qvalid, seltype, opt, sel%lower, getNameFromId_obs(sel%obsId), sel%upper
        else if (sel%qlower .and. .not.sel%qupper) then
          write(iout,fmt_min) sel%qvalid, seltype, opt, sel%lower, getNameFromId_obs(sel%obsId)
        else if (.not.sel%qlower .and. sel%qupper) then
          write(iout,fmt_max) sel%qvalid, seltype, opt, getNameFromId_obs(sel%obsId), sel%upper
        else
          write(iout,*) "SOMETHING WENT WRONG"
        endif

      end subroutine digest_hist_print_selector

  end subroutine digest_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> set a global prefix for all histogram files.
  !
  !> @param[in] pre prefix for all histtograms
  !-----------------------------------------------------------------------------
  subroutine setPrefix_hist(pre)
    character(len=*) :: pre
    filePrefix = pre
  end subroutine setPrefix_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> set a global suffix for all histogram files.
  !
  !> @param[in] suff suffix for all histtograms
  !-----------------------------------------------------------------------------
  subroutine setSuffix_hist(suff)
    character(len=*) :: suff
    fileSuffix = suff
  end subroutine setSuffix_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine the bin location for a given `x`-value and binning
  !
  !> @param[in] val value to be binned
  !> @param[in] array of bin boundaries
  !> @return bin location
  !-----------------------------------------------------------------------------
  function bin_from_val(val, boundaries)
    integer :: bin_from_val
    double precision, intent(in) :: val
    double precision, dimension(0:), intent(in) :: boundaries
    integer :: l,m,u
    l = lbound(boundaries,1)
    u = ubound(boundaries,1)
    if (l.ne.0) stop "lower array bound not zero!"
    if ( val < boundaries(l) .or. val > boundaries(u) ) then
      bin_from_val = 0 ! overflow bin
      return
    endif
    do while ( l < u-1 )
      m = (l+u)/2
      if ( boundaries(m) >= val ) then
        u = m
      else
        l = m
      end if
    end do
    bin_from_val = u
  end function bin_from_val


  !-----------------------------------------------------------------------------
  !> @brief
  !> accumulate values in the buffer.
  !
  !> @param[in] wts array of length nMax_scl (weights without `partons`)
  !> @param[in] ipart1 first parton identifyer
  !> @param[in] ipart2 second parton identifyer
  !> @param[in] partons optional array of length nMax_scl (parton luminosities)
  !> @param[in] as2pi optional array containing as2pi = alphas/(2pi) values
  !>            of the input scales (for scale-coefficients)
  !-----------------------------------------------------------------------------
  subroutine writeBuffer_hist(wts, ipart1, ipart2, partons, as2pi)
    use Constants_mod
    use Observables_mod
    use MultiRun_mod, only : isActive_mult, currentRun_mult
    use Channels_mod
    use ScaleCoeff_mod
    use ProcLib_mod, only : getlumi_plib
    use nnlo_bridge_mod, only : set_mu2, get_mu2, set_obsId
    double precision, dimension(nMax_scl), intent(in) :: wts
    integer, intent(in) :: ipart1, ipart2
    double precision, dimension(nMax_scl), intent(in), optional :: partons, as2pi
    type(Histogram_t), pointer :: hist
    type(SubHistogram_t), pointer :: it_sub
    type(Selector_t), pointer :: it_sel
    double precision :: valObs
    logical :: valid
    integer :: i, binx
    double precision, dimension(nMax_scl) :: fac_rewgt
    double precision, dimension(nMax_scl) :: fac_local
    integer :: iscl_rewgt, iscl_tmp
    !> composite observables
    integer :: isub, nsub
    double precision :: fac_sub
    !> MultiRun setting for the group
    integer :: run_group
    !> weight for grids: stripped from `partons` & as/(2pi) factors
    double precision, dimension(nMax_scl) :: wtstrip, wtstrip_inp, wtstrip_cpy
    !> scale override option when using scale-coefficients
    double precision, dimension(nMax_scl) :: wthisto, wthisto_inp, wthisto_cpy
    double precision, dimension(nMax_scl) :: as2pi_inp, partons_inp
    double precision, dimension(nMax_scl) :: muR2_inp, muF2_inp
    double precision, dimension(nMax_scl) :: muR2_cpy, muF2_cpy
    double precision, dimension(ndim_sclc) :: coeffs, coeffs_ref1GeV
    double precision :: muR02, muF02, LR02, LF02
    integer :: as2pi_power, nV

    if (.not.qactive) return

    if (isActive_sclc()) then
      !> scale coefficients: wts needs to be converted to coefficients
      !> get default alphas values
      if (present(as2pi)) then
        as2pi_inp = as2pi
      else
        call get_alphas_scl(as2pi_inp, fac=1d0)
        as2pi_inp = as2pi_inp / twopi
      end if
      !> get default luminosity
      if (present(partons)) then
        partons_inp = partons
      else
        call getlumi_plib(ix_chan, current_chan, partons_inp, fac=1d0)
      end if
      call get_muR2_scl(muR2_inp, fac=1d0)
      call get_muF2_scl(muF2_inp, fac=1d0)
      !> get the alphas power
      call currentContrib_chan(nas=as2pi_power, nV=nV)
      !> get the coefficients of the logs
      call extract_sclc(wts, coeffs, nV)
      !> save the values for the central scale
      muR02 = muR2_scl(1)
      muF02 = muF2_scl(1)
      !> time to assemble the full weights
      call writeBuffer_hist_sclc_wgt(wtstrip_inp, wthisto_inp)
      wtstrip = wtstrip_inp
      wthisto = wthisto_inp
      !> fastNLO wants the decomposition w.r.t. muR0=muF0=1GeV
      LR02 = log(muR02)
      LF02 = log(muF02)
      coeffs_ref1GeV(1) = coeffs(1)  &
                      & - coeffs(2) * LR02  &
                      & - coeffs(3) * LF02  &
                      & + coeffs(4) * LR02**2  &
                      & + coeffs(5) * LR02*LF02  &
                      & + coeffs(6) * LF02**2
      coeffs_ref1GeV(2) = coeffs(2)  &
                      & - coeffs(4) * 2*LR02  &
                      & - coeffs(5) * LF02
      coeffs_ref1GeV(3) = coeffs(3)  &
                      & - coeffs(5) * LR02  &
                      & - coeffs(6) * 2*LF02
      coeffs_ref1GeV(4) = coeffs(4)
      coeffs_ref1GeV(5) = coeffs(5)
      coeffs_ref1GeV(6) = coeffs(6)

      ! print*, "wts", wts

      ! print*, "A00", coeffs(1)
      ! print*, "B10", coeffs(2)
      ! print*, "B01", coeffs(3)
      ! print*, "C20", coeffs(4)
      ! print*, "C11", coeffs(5)
      ! print*, "C02", coeffs(6)

      ! print*, "muR2_new", muR2_new(1), sqrt(muR2_new(1))
      ! print*, "muF2_new", muF2_new(1), sqrt(muF2_new(1))
      ! print*, "as2pi_inp", as2pi_inp(1)
      ! print*, "as2pi_power", as2pi_power
      ! print*, "wthisto", wthisto(1)

      ! read(*,*)
    else
      !> no scale coefficients: wts is the full weight from the start
      wthisto = wts
    end if


    !> get the current run_group (zero if no MultiRun)
    if (isActive_mult()) call currentRun_mult(run_group=run_group)

    !> get the "local" reweighting factor
    fac_local = 1d0
    if (isActiveLocal_rewgt()) then
      iscl_rewgt = -1  ! start with an invalid value to make sure it's reset
      fac_local = getLocal_rewgt(iscl_rewgt)
      if ( iscl_rewgt /= 0 .and. iscl_rewgt /= 1 ) then
        print*, "writeBuffer_hist: invalid iscl returned from getLocal_rewgt: ", iscl_rewgt
        stop
      end if
      if (iscl_rewgt /= 0) then
        do iscl_tmp=2,nInp_scl
          iscl_rewgt = iscl_tmp
          fac_local(iscl_rewgt) = getLocal_rewgt(iscl_rewgt)
        end do
      end if
    end if

    !> loop over all registered histograms
    loophist: do i=1,n_hist
      hist => buffer_hist(i)%histParent ! just a temp pointer to `prettify`

      !> check run_group setting
      if (hist%igroup/=0 .and. isActive_mult()) then
        if (hist%igroup/=run_group) cycle loophist
      end if

      !> apply all registered selectors for the histogram
      it_sel => hist%sel_p
      do while ( associated(it_sel) )
        valObs = getValue_obs(it_sel%obsId, valid)
        if ( .not.valid ) then
          if ( it_sel%qvalid .and. hist%selOpt==0 ) then
            cycle loophist
          end if
          !> now either:  it_sel%qvalid=.false.  OR  hist%selOpt==1
          it_sel => it_sel%next
          cycle
        end if
        if ( .not.qpass_sel(it_sel, valObs) ) cycle loophist
        it_sel => it_sel%next
      end do

      !> set the reweighting factor if active
      fac_rewgt  = 1d0
      if (hist%rewgtfcn%qactive) then
        iscl_rewgt = -1  ! start with an invalid value to make sure it's reset
        fac_rewgt = eval_rewgt(hist%rewgtfcn, iscl_rewgt)
        if ( iscl_rewgt /= 0 .and. iscl_rewgt /= 1 ) then
          print*, "writeBuffer_hist: invalid iscl returned from reweighting function"
          stop
        end if
        if (iscl_rewgt /= 0) then
          do iscl_tmp=2,nInp_scl
            iscl_rewgt = iscl_tmp
            fac_rewgt(iscl_rewgt) = eval_rewgt(hist%rewgtfcn, iscl_rewgt)
          end do
        end if
      end if
      !> multiply the local weight
      if (isActiveLocal_rewgt()) then
        fac_rewgt = fac_rewgt * fac_local
      end if

      !> perform the scale-swap if active
      if (isActive_sclc()) then
        if ( (hist%fac_mu0 /= 0d0) .and. (hist%obsId_mu0 >= 0) ) then
          !> runcard parser always sets both at the same time
          !> @todo make more general when only one piece is overridden?
          call writeBuffer_hist_sclc_wgt(wtstrip, wthisto,  &
            & fac=hist%fac_mu0, obsId=hist%obsId_mu0, qgrid=(hist%igrid>0))
        else
          wtstrip = wtstrip_inp
          wthisto = wthisto_inp
          if (hist%igrid>0) call set_mu2(muR2_inp, muF2_inp)
        end if
      end if

      !>----- [1] first deal with non-composite histograms (cross & single-obs)
      if ( hist%histType >= 0 .and. .not.isComposite_obs(hist%histType) ) then
        if ( hist%histType == 0 ) then  ! cross
          binx = 0
        else  ! single-obs
          valObs = getValue_obs(hist%histType, valid)
          if (hist%igrid>0) call set_obsId(hist%histType)
          if (valid) then
            binx = bin_from_val( valObs , hist%boundaries )
          else
            !> not valid for this kinematics:

            !>--- OPTION 1: accumulate in overflow bin
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJ@NNLO
            ! binx = 0
            !>--------------------------------------------

            !>--- OPTION 2: do not accumulate
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJJ@NLO
            cycle loophist
            !>--------------------------------------------

          end if
        end if

        call writeBuffer_hist_accumulate(1d0)
        cycle loophist
      end if

      !>----- [2] deal with composite observables
      if ( isComposite_obs(hist%histType) ) then
        nsub = int(getValueComposite_obs(hist%histType, valid))
        if (.not.valid) cycle loophist
        ! print*, "histo: COMPOSITE", hist%histType, "nsub:", nsub
        loopsub: do isub = 1, nsub
          fac_sub = 1d0
          valObs = getValueComposite_obs(hist%histType, valid, isub, fac_sub)
          ! print*, ">>>", isub, valObs, fac_sub
          if (valid) then
            binx = bin_from_val( valObs , hist%boundaries )
          else
            !> not valid for this kinematics:

            !>--- OPTION 1: accumulate in overflow bin
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJ@NNLO
            ! binx = 0
            !>--------------------------------------------

            !>--- OPTION 2: do not accumulate
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJJ@NLO
            cycle loopsub
            !>--------------------------------------------

          end if
          call writeBuffer_hist_accumulate(fac_sub)
        end do loopsub
        cycle loophist
      end if

      !>----- [3] now deal with composite histograms
      it_sub => hist%subHist_p
      !> the new default for scale-coefficients is the parent weights:
      if ( isActive_sclc() .and. associated(it_sub) ) then
        wtstrip_cpy = wtstrip
        wthisto_cpy = wthisto
        if (hist%igrid>0) call get_mu2(muR2_cpy, muF2_cpy)
      end if
      !> loop over all sub-histograms
      do while ( associated(it_sub) )

        !> check run_group setting
        if (it_sub%igroup/=0 .and. isActive_mult()) then
          if (it_sub%igroup/=run_group) goto 666
        end if

        !> apply all registered selectors for the SubHistogram
        it_sel => it_sub%sel_p
        do while ( associated(it_sel) )
          valObs = getValue_obs(it_sel%obsId, valid)
          if ( .not.valid ) then
            if ( it_sel%qvalid .and. it_sub%selOpt==0 ) then
              goto 666
            end if
            !> now either:  it_sel%qvalid=.false.  OR  it_sub%selOpt==1
            it_sel => it_sel%next
            cycle
          end if
          if ( .not.qpass_sel(it_sel, valObs) ) goto 666
          it_sel => it_sel%next
        end do

        !> set the reweighting factor if active
        !> sub-histograms overwrite reweighting factors of the parent!
        fac_rewgt = 1d0
        if (it_sub%rewgtfcn%qactive) then
          iscl_rewgt = -1  ! start with an invalid value to make sure it's reset
          fac_rewgt = eval_rewgt(it_sub%rewgtfcn, iscl_rewgt)
          if ( iscl_rewgt /= 0 .and. iscl_rewgt /= 1 ) then
            print*, "writeBuffer_hist: invalid iscl returned from reweighting function"
            stop
          end if
          if (iscl_rewgt /= 0) then
            do iscl_tmp=2,nInp_scl
              iscl_rewgt = iscl_tmp
              fac_rewgt(iscl_rewgt) = eval_rewgt(it_sub%rewgtfcn, iscl_rewgt)
            end do
          end if
        end if
        !> multiply the local weight
        if (isActiveLocal_rewgt()) then
          fac_rewgt = fac_rewgt * fac_local
        end if

        !> perform the scale-swap if active
        !> sub-histograms overwrite scales of the parent!
        if (isActive_sclc()) then
          if ( (it_sub%fac_mu0 /= 0d0) .and. (it_sub%obsId_mu0 >= 0) ) then
            !> runcard parser always sets both at the same time
            !> @todo make more general when only one piece is overridden?
            call writeBuffer_hist_sclc_wgt(wtstrip, wthisto,  &
              & fac=it_sub%fac_mu0, obsId=it_sub%obsId_mu0, qgrid=(hist%igrid>0))
          else
            wtstrip = wtstrip_cpy
            wthisto = wthisto_cpy
            if (hist%igrid>0) call set_mu2(muR2_cpy, muF2_cpy)
          end if
        end if

        !> survived all cuts: bin it
        if ( it_sub%obsId == 0 ) then  ! cross
          binx = 0
        else  ! single-obs
          valObs = getValue_obs(it_sub%obsId, valid)
          if (hist%igrid>0) call set_obsId(it_sub%obsId)
          if (valid) then
            binx = bin_from_val( valObs , hist%boundaries )
          else
            !> not valid for this kinematics:
            !> --- option 1: accumulate in overflow bin
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJ@NNLO
            ! binx = 0
            !> --- option 2: do not accumulate
            !> example: ZJ@NNLO --- ptj2
            !>          => sum(bins) + overflow = ZJJ@NLO
            goto 666
          end if
        end if

        call writeBuffer_hist_accumulate(1d0)

        666 it_sub => it_sub%next
      end do

    end do loophist

  contains

    !-----------------------------------------------------------------------------
    !>
    !
    !> @param[in] fac optional
    !> @param[in] obsId optional
    !-----------------------------------------------------------------------------
    subroutine writeBuffer_hist_sclc_wgt(wtstrip_out, wthisto_out, fac, obsId, qgrid)
      double precision, dimension(nMax_scl), intent(out) :: wtstrip_out, wthisto_out
      double precision, optional, intent(in) :: fac
      integer, optional, intent(in) :: obsId
      logical, optional, intent(in) :: qgrid
      !> global variables:
      !>  * for default: partons_inp, as2pi_inp
      !>  * for reconstruction: muR02, muF02, coeff, as2pi_power
      double precision, dimension(nMax_scl) :: partons_new, as2pi_new
      double precision, dimension(nMax_scl) :: muR2_new, muF2_new
      double precision, dimension(nMax_scl, ndim_sclc) :: logs

      !> default input scale
      if ( .not.present(fac) .and.  &
         & .not.present(obsId) ) then
        ! print*, "writeBuffer_hist_sclc_wgt: use default"
        partons_new = partons_inp
        as2pi_new   = as2pi_inp
        call get_muR2_scl(muR2_new, fac=1d0)
        call get_muF2_scl(muF2_new, fac=1d0)
      else
        ! print*, "writeBuffer_hist_sclc_wgt: swap scale", fac, obsId
        call getlumi_plib(ix_chan, current_chan, partons_new, fac=fac, obsId=obsId)
        call get_alphas_scl(as2pi_new, fac=fac, obsId=obsId)
        as2pi_new = as2pi_new / twopi
        call get_muR2_scl(muR2_new, fac=fac, obsId=obsId)
        call get_muF2_scl(muF2_new, fac=fac, obsId=obsId)
      end if

      !> if grid filling is active, save the active scale values
      if (present(qgrid)) then
        ! print*, "writeBuffer_hist_sclc_wgt: save to hook module"
        if (qgrid) call set_mu2(muR2_new, muF2_new)
      end if

      ! print*, "muR02   ", muR02
      ! print*, "muR2_new", muR2_new(1)
      ! print*, "muF02   ", muF02
      ! print*, "muF2_new", muF2_new(1)

      logs(:,:) = 0d0
      logs(1:nInp_scl,1) = 1d0                                    ! A00
      logs(1:nInp_scl,2) = log(muR2_new(1:nInp_scl)/muR02)        ! B10
      logs(1:nInp_scl,3) = log(muF2_new(1:nInp_scl)/muF02)        ! B01
      logs(1:nInp_scl,4) = logs(1:nInp_scl,2)**2                  ! C20
      logs(1:nInp_scl,5) = logs(1:nInp_scl,2)*logs(1:nInp_scl,3)  ! C11
      logs(1:nInp_scl,6) = logs(1:nInp_scl,3)**2                  ! C02
      !> assemble the weights
      wtstrip_out = matmul(logs, coeffs)
      wthisto_out = wtstrip_out * partons_new * as2pi_new**as2pi_power

      ! print*, "wtstrip_out(1)", wtstrip_out(1)
      ! print*, "coeffs(1)     ", coeffs(1)

      ! print*, "partons_new ", partons_new
      ! print*, "as2pi_new   ", as2pi_new
      ! print*, "wtstrip_out ", wtstrip_out
      ! read(*,*)
    end subroutine writeBuffer_hist_sclc_wgt

    !-----------------------------------------------------------------------------
    !> accumulate the weights
    !-----------------------------------------------------------------------------
    subroutine writeBuffer_hist_accumulate(opt_fac)
#ifdef USE_PINEAPPL
      use PineAPPL_mod
#endif
      double precision, intent(in) :: opt_fac
      integer :: binc
      !> get vegas normalisation factor
      double precision :: dv2g
      integer :: npg
      common /vegasnumcalls/ dv2g,npg

#ifdef HISTBUFFKAHAN
      call accumulate_bin( hist%dataType,  &
        &                  buffer_hist(i)%histBuff(:,:,binx), &
        &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#else
      call accumulate_bin( hist%dataType,  &
        &                  buffer_hist(i)%histBuff(1,:,binx),  &
        &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#endif

      !----- cumulant distributions
      !> [xmin, xi]
      if ( binx /= 0 .and. hist%icumulant == -1 ) then
        do binc=(binx+1),hist%nbins,+1
#ifdef HISTBUFFKAHAN
          call accumulate_bin( hist%dataType,  &
            &                  buffer_hist(i)%histBuff(:,:,binc), &
            &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#else
          call accumulate_bin( hist%dataType,  &
            &                  buffer_hist(i)%histBuff(1,:,binc),  &
            &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#endif
        end do
      end if
      !> [xi, xmax]
      if ( binx /= 0 .and. hist%icumulant == +1 ) then
        do binc=(binx-1),1,-1
#ifdef HISTBUFFKAHAN
          call accumulate_bin( hist%dataType,  &
            &                  buffer_hist(i)%histBuff(:,:,binc), &
            &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#else
          call accumulate_bin( hist%dataType,  &
            &                  buffer_hist(i)%histBuff(1,:,binc),  &
            &                  opt_fac*fac_rewgt*wthisto, ipart1, ipart2 )
#endif
        end do
      end if

      !----- begin grid interface
      !> maybe some mode-specific settings?
      ! select case (hist%igrid)
      !   case (1)  ! APPLgrid
      !   case (2)  ! fastNLO
      !   case (3)  ! both
      ! end select
      !> fill the grid here!
      !> availebale information:
      !>    * which mode?:                               `newHist%igrid`
      !>    * unique integer id for the histogram:       `hist%ihist`
      !>    * name of the observable:                    `getNameFromId_obs(hist%obsId)`
      !>      [not unique: can be booked multiple times, which is why we have `ihist` above]
      !>    * value of the observable for current event: `valObs`
      !>    * array of bin boundaries for the histogram: `hist%boundaries`
      !>    * name of the gridfile:                      `hist%gridname`
      !call fill_grid   ( hist%ihist, valObs, coeffs_ref1GeV )
!$OMP CRITICAL(grid_fill)
      if (hist%igrid > 0) then
#ifdef USE_PINEAPPL
        call fill_grid_pine(hist%ihist, valObs, dble(npg)*opt_fac*fac_rewgt(1)*coeffs)
#else
        ! call nnlo_bridge_test( hist%igrid, hist%ihist, valObs,  &
        !   &                    dble(npg)*opt_fac*fac_rewgt*wtstrip,  &
        !   &                    dble(npg)*opt_fac*fac_rewgt*wthisto,  &
        !   &                    dble(npg)*opt_fac*fac_rewgt(1)*coeffs_ref1GeV )
        call fill_grid( hist%igrid, hist%ihist, valObs,  &
          &             dble(npg)*opt_fac*fac_rewgt*wtstrip,  &
          &             dble(npg)*opt_fac*fac_rewgt*wthisto,  &
          &             dble(npg)*opt_fac*fac_rewgt(1)*coeffs_ref1GeV )
#endif
      end if
!$OMP END CRITICAL(grid_fill)
      !----- end grid interface
    end subroutine writeBuffer_hist_accumulate

  end subroutine writeBuffer_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> flush the buffer into the main histogram (histData).
  !
  !> @param[in] vegas_wsig optional argument containing the weight that is
  !>            passed to Vegas. Corresponds to the 'tot' channel for the
  !>            central scale.
  !-----------------------------------------------------------------------------
  subroutine flushBuffer_hist(vegas_wsig)
    use FPA_mod
    use Channels_mod, only : ireject_event_chan
    double precision, optional, intent(in) :: vegas_wsig
    type(Histogram_t), pointer :: hist
    integer :: i, j, k
#ifdef HISTKAHAN
    double precision :: acc_sum, err
#endif
    !> check vs. the vegas weight
    double precision :: check_wsig, delta_wsig_abs, delta_wsig_rel
    double precision, parameter :: threshold_rel = 1d-5
    double precision, parameter :: threshold_abs = 1d-12 / threshold_rel

    if (.not.qactive) return
    if (.not.allocated(buffer_hist)) return

    !> reject event: clear buffer and die
    if (ireject_event_chan /= 0) then
      do i=1,n_hist
        buffer_hist(i)%histBuff = 0d0
      end do
      return
      !> ireject_event_chan reset in flush of Channels_mod & sig.f
    end if

    do i=1,n_hist
      hist => buffer_hist(i)%histParent  ! just a temp pointer to `prettify`

#ifdef HISTBUFFKAHAN
      !> compensated sum: add back the accumulated errors
      do j=0,hist%nbins
      do k=1,hist%ndata
        buffer_hist(i)%histBuff(1,k,j) = buffer_hist(i)%histBuff(1,k,j)  &
          &                            + buffer_hist(i)%histBuff(0,k,j)
        buffer_hist(i)%histBuff(0,k,j) = 0d0
      end do
      end do
#endif

      !> do a check against Vegas to auto-detect issues like missing "subfac"s
      if ( present(vegas_wsig) .and. .not.isActive_rewgt() ) then
        !> look for the "master" cross histogram at the tail
        if ( vegas_wsig /= 0d0 .and. .not.associated(hist%next) ) then
          !> cross only has 1 bin (0:0)  [not even oveflow]
          check_wsig     = buffer_hist(i)%histBuff(1,1,0)  ! tot & central scale
          delta_wsig_abs = abs( (check_wsig-vegas_wsig) )
          delta_wsig_rel = delta_wsig_abs /  abs( (check_wsig+vegas_wsig) )
          if ( delta_wsig_abs > threshold_abs .and.  &
             & delta_wsig_rel > threshold_rel ) then
            print*, "flushBuffer_hist: Vegas vs. bino weights don't match!"
            print*, "    Vegas: ", vegas_wsig
            print*, "    bino:  ", check_wsig
            ! stop
          end if
        end if
      end if

      !> do the flush
!$OMP CRITICAL(flushbuffer)
      do j=0,hist%nbins
        hist%histCount(j) = hist%histCount(j) + 1
        do k=1,hist%ndata
#ifdef HISTKAHAN
          call twoSum(hist%histData(1,k,j), buffer_hist(i)%histBuff(1,k,j), acc_sum, err)
          hist%histData(1,k,j) = acc_sum
          hist%histData(0,k,j) = hist%histData(0,k,j) + err
#else
          hist%histData(1,k,j) = hist%histData(1,k,j) + buffer_hist(i)%histBuff(1,k,j)
#endif
          hist%histData(2,k,j) = hist%histData(2,k,j) + buffer_hist(i)%histBuff(1,k,j)**2
        end do
      end do
!$OMP END CRITICAL(flushbuffer)

      buffer_hist(i)%histBuff = 0d0
    end do  ! end `i`
  end subroutine flushBuffer_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> clear all histograms.
  !-----------------------------------------------------------------------------
  subroutine clear_hist()
    type(Histogram_t), pointer :: it
    call flushBuffer_hist() ! first reset buffer
    it => head
    do while ( associated(it) )
      it%histCount = 0
      it%histData  = 0d0
      it => it%next
    end do
  end subroutine clear_hist


  !-----------------------------------------------------------------------------
  !> @brief
  !> write histograms to data files. The first line is a comment labelling the
  !> columns. The second line in a comment line containing the overflow. With
  !> this information the total cross section can be otained via:
  !>     sum( bin(i) * binwidth(i) ) + overflow = total xsection
  !> The x-values are given by the first *three* columns: The lower bound, the
  !> center, and the upper bound of the bin. In gnuplot this means:
  !>     variable size:  plot "..." u 1:Y with steps
  !>     variable size:  plot "..." u 3:Y with fsteps
  !>     equal size:     plot "..." u 2:Y with histeps
  !
  !> @param[in] scaleFac optional global factor on histogram
  !> @param[in] commentLines optional array of strings written out as comment
  !>            lines at the end of each histogram file.
  !> @param[in] verbose optional flag to control verbosity
  !-----------------------------------------------------------------------------
  subroutine writeToFile_hist(scaleFac, commentLines, verbose)
    use Observables_mod
    use Scales_mod, only: nInp_scl,nMax_scl
#ifdef USE_PINEAPPL
    use PineAPPL_mod
#endif
    double precision, optional, intent(in) :: scaleFac
    character(len=*), dimension(:), optional, intent(in) :: commentLines
    logical, optional, intent(in) :: verbose
    integer, parameter :: fileUnit = 20
    character(len=*), parameter :: fmt_colNum = '("[",I0,"]")'
    character(len=100) :: fileName
    character(len=5) :: str_colNum ! #(columns) < 1000
    type(Histogram_t), pointer :: it
    double precision :: delta, fac, tmp_tot,tmp_err,tmp_chn
    integer :: i,j,k, neval, nxbins, idx_tot,idx_chn
    logical :: verb

    !> if we want to index each iteration
    ! integer, save :: nhist = 1
    ! character(len=5) :: s_nhist
    ! write(s_nhist,"(I3.3)") nhist
    ! nhist = nhist+1
    ! print*, "s_nhist: >", trim(adjustl(s_nhist)), "<"

    if (present(scaleFac)) then
      fac = scaleFac
    else
      fac = 1d0
    end if

    if (present(verbose)) then
      verb = verbose
    else
      verb = .false.
    end if

    if (qsingle) then
      fileName = trim(filePrefix)//trim(fileSingle)//trim(fileSuffix)
      if (verb) print*, "writing histogram: ", trim(fileName)
      open(fileUnit, file=trim(fileName))
    end if

    it => head
    do while ( associated(it) )

#ifdef HISTKAHAN
     !> compensated sum: add back the accumulated errors
      do j=0,it%nbins
      do k=1,it%ndata
        it%histData(1,k,j) = it%histData(1,k,j) + it%histData(0,k,j)
        it%histData(0,k,j) = 0d0
      end do
      end do
#endif
      if (.not.qsingle) then
        fileName = trim(filePrefix)//trim(it%name)//trim(fileSuffix)
        !fileName = trim(filePrefix)//trim(it%name)//"_it"//trim(adjustl(s_nhist))//trim(fileSuffix)
        if (verb) print*, "writing histogram: ", trim(fileName)
        open(fileUnit, file=trim(fileName))
      else
        write(fileUnit,*)
        write(fileUnit,'("#name: ",a)') trim(it%name)
      end if
      !> comment line labelling the columns
      write(fileUnit,'("#labels:")',advance='no')
      nxbins = 0
      if ( it%histType /= 0 .and. it%nbins > 0 ) then
        !> cumulant?
        select case (it%icumulant)
        case (0)
          nxbins = 3
          write(str_colNum,fmt_colNum) 1
          write(fileUnit,'(" ",a)',advance='no') &
             & trim(it%name)//"_lower"//str_colNum
          write(str_colNum,fmt_colNum) 2
          write(fileUnit,'(" ",a)',advance='no') &
             & trim(it%name)//"_center"//str_colNum
          write(str_colNum,fmt_colNum) 3
          write(fileUnit,'(" ",a)',advance='no') &
             & trim(it%name)//"_upper"//str_colNum
        case (-1,+1)
          nxbins = 1
          write(str_colNum,fmt_colNum) 1
          write(fileUnit,'(" ",a)',advance='no') &
             & trim(it%name)//str_colNum
        case default
          print*, "writeToFile_hist: unknown icumulant", it%icumulant
        end select
      end if
      do k=1,it%ndata
        write(str_colNum,fmt_colNum) nxbins + 2*k-1
        write(fileUnit,'(" ",a)',advance='no') &
           & trim(getLabel_bin(it%dataType,k))//str_colNum
        write(str_colNum,fmt_colNum) nxbins + 2*k
        write(fileUnit,'(" ",a)',advance='no') &
           & trim(getLabel_bin(it%dataType,k))//"_Err"//str_colNum
      end do
      write(fileUnit,*)
      !> comment line giving the total statistics of the run
      neval = it%histCount(0)
      do j=1,it%nbins
        if (neval /= it%histCount(j)) then
          print*, "writeToFile_hist: !!! WARNIMG !!! neval mismatch"
        end if
      end do
      write(fileUnit,'("#neval: ",I0)') neval
      !> histogram bins & errors
      do j=0,it%nbins
        delta = 1d0
        !if (j == 0 .and. it%obsId /= 0 ) then
        if (j == 0) then
          if ( it%histType /= 0 .and. it%nbins > 0 ) then
            !> cumulant?
            select case (it%icumulant)
            case (0)
              write(fileUnit,'("#overflow:lower center upper")',advance='no')
            case (-1,+1)
              write(fileUnit,'("#overflow:")',advance='no')
            end select
          else
            write(fileUnit,'(" ")',advance='no') ! apparently this is needed...
          end if
        else
            !> cumulant (delta=1 for cumulant distributions)
            select case (it%icumulant)
            case (0)
              write(fileUnit,'(3(" ",es20.10e3))',advance='no') &
                 & it%boundaries(j-1), &
                 & (it%boundaries(j-1)+it%boundaries(j))/2d0, &
                 & it%boundaries(j)
              delta = abs(it%boundaries(j)-it%boundaries(j-1))
            case (-1)
              write(fileUnit,'(" ",es20.10e3)',advance='no') &
                 & it%boundaries(j)
            case (+1)
              write(fileUnit,'(" ",es20.10e3)',advance='no') &
                 & it%boundaries(j-1)
            end select
        end if
        do k=1,it%ndata
          !> mean
          write(fileUnit,'(" ",es20.10e3)',advance='no') &
             & it%histData(1,k,j) / dble(it%histCount(j)) &
             & * fac / delta
          !> standard deviation of the mean
          write(fileUnit,'(" ",es20.10e3)',advance='no')   &
             & sqrt( it%histData(2,k,j) &
             &     - it%histData(1,k,j)**2/dble(it%histCount(j)) ) &
             & / dble(it%histCount(j)) * fac / delta
        end do
        write(fileUnit,*)
      end do
      !> print out number of bins
      write(fileUnit,'("#nx: ",I0)') nxbins
      !> if grid ouput is active, write out gridname
      if ( it%igrid /= 0 ) then
        write(fileUnit,'("#grid: ",A)') it%gridname
      end if
      !> optional trailing comment lines
      if(present(commentLines)) then
        do i=1,size(commentLines)
          write(fileUnit,'("# ")',advance='no')
          write(fileUnit,*) commentLines(i)
        end do
      end if
      if (.not.qsingle) close(fileUnit)
      !> closing call to APPLfast-NNLO
      if ( it%igrid /= 0 ) then
#ifdef USE_PINEAPPL
        !> only write in the last iteration
        if (verbose) then
          call write_grid_pine( it%ihist,  &
            &  trim(filePrefix)//trim(it%name)//trim(fileSuffix),  &
            &  factor=1d0/dble(neval),  &
            &  obsName=trim(it%name) )
        end if
#else
        call write_grid( it%igrid, it%ihist, 1d0/dble(neval) )
        ! select case (it%igrid)
        ! case(1) ! applgrid
        !    call write_grid( it%ihist )
        ! case(2) ! fastNLO
        !    call write_grid( it%ihist )
        ! end select
#endif
      end if
      it => it%next
    end do
    if (qsingle) close(fileUnit)

    !> storewgt type output
    if(.not.verb) return
    !> grab the tail ("cross" histogram)
    it => head
    do
      if ( .not.associated(it%next) ) exit
      it => it%next
    end do
    if ( it%histType /= 0 ) then
      print*, "writeToFile_hist: no cross histogram registered?! run init_hist!"
      return
    end if
    !> output
    write(*,*)
    neval = it%histCount(0)
    do i=1,nInp_scl
      idx_tot = getIndex_bin(it%dataType,i)
      tmp_tot = it%histData(1,idx_tot,0)*fac/dble(neval)
      tmp_err = sqrt( it%histData(2,idx_tot,0) &
              &     - it%histData(1,idx_tot,0)**2/dble(neval) ) &
              & *fac/dble(neval)
      do j=-1,1
      do k=-1,1
        idx_chn = getIndex_bin(it%dataType,i,j,k)
        tmp_chn = it%histData(1,idx_chn,0)*fac/dble(neval)
        write(*,'(" ",a15," ",g14.6," ",f8.2,"%")') &
           & trim(getLabel_bin(it%dataType,idx_chn)), tmp_chn, 100d0*tmp_chn/tmp_tot
      end do
      end do
      write(*,'(" ",a15," ",g14.6," +- ",g14.6," fb")') &
         & trim(getLabel_bin(it%dataType,idx_tot)), tmp_tot, tmp_err
      write(*,*)
    end do

  end subroutine writeToFile_hist


end module Histograms_mod
