!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: I/O Helper [external members: `_io`]
!
! DESCRIPTION:
!> Module collecting routines for parsing the runcard simplifying IO
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module IOHelper_mod
!-------------------------------------------------------------------------------
  use StrHelper_mod
  use Kinds_mod
  implicit none
  public

  !> IO units
  integer, parameter :: out_unit_io = 6
  integer, parameter :: err_unit_io = 0
  integer, parameter :: log_unit_io = 42

  ! character(:), allocatable :: logfile_io

contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> get filenames, prefixes, suffixes, ...
  !
  !> @param[out] logfile
  !> @param[out] gridfile
  !> @param[out] histogram_prefix
  !> @param[out] histogram_suffix
  !-----------------------------------------------------------------------------
  subroutine file_io(logfile, gridfile, histogram_prefix, histogram_suffix)
    use Process_mod
    use Run_mod
    use Channels_mod
    character(len=128), intent(out), optional :: logfile
    character(len=128), intent(out), optional :: gridfile
    character(len=128), intent(out), optional :: histogram_prefix
    character(len=128), intent(out), optional :: histogram_suffix
    character(:), allocatable :: str_prefix, str_seed, str_tcut
    character(:), allocatable :: name_contr
    character(len=128) :: input_str

    str_prefix = trim(in_name_proc) // '.' // trim(name_run)

    write(input_str, *) iseed_run
    str_seed = '.s' // trim(adjustl(input_str))

    write(input_str, '(ES10.2E2)') tcut_run
    str_tcut = '.y' // trim(adjustl(input_str))

    if (present(gridfile) .or. present(histogram_prefix)) then
      call currentContrib_chan(str=name_contr)
    end if


    if (present(logfile)) then
      logfile = str_prefix // str_seed // '.log'
      print*, "logfile: >", logfile, "<"
    end if

    if (present(gridfile)) then
      gridfile = str_prefix // str_tcut // '.' // name_contr
    end if

    if (present(histogram_prefix)) then
      histogram_prefix = str_prefix // '.' // name_contr // '.'
    end if

    if (present(histogram_suffix)) then
      histogram_suffix = str_seed // '.dat'
    end if

  end subroutine file_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Reads a complete line from `unit` and further process the line by
  !> stripping unwanted blanks and comment lines at the end (start by `!`)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !> @param[out] line the line
  !> @param[out] stat error code: positive on error
  !-----------------------------------------------------------------------------
  recursive subroutine getline_io(unit, line, stat)
    use, intrinsic :: iso_fortran_env, only: iostat_eor
    integer, intent(in) :: unit
    character(:), intent(out), allocatable :: line
    integer, intent(out) :: stat
    character(256) :: buffer
    integer :: size
    integer :: i,j
    !> list of characers where any blanks after/before it will be eliminated
    character(*), parameter :: killtrail = "=,>[*+ "
    character(*), parameter :: killlead  = "=,>]*+"
    character(2) :: pattern
    integer :: stat2
    character(:), allocatable :: line2

    !> read the full line
    line = ''
    do
      read(unit, "(A)", advance='no', iostat=stat, size=size) buffer
        if (stat > 0) return ! error
        line = line // buffer(:size)
        if (stat < 0) then
          if (stat == iostat_eor) stat = 0
          exit
        end if
    end do

    !> replace all `tab` characters by a blank
    do
      i = index(line, char(9))
      if ( i.eq.0 ) exit
      line(i:i) = " "
    end do

    line = trim(adjustl( line ))  ! kill leading blanks
    !> kill possible comments
    i = index(line, "!")
    if ( i.gt.0 ) line = trim(adjustl( line(:i-1) ))  ! kill trailing blanks

    !> do we have a line continuation?
    i = len(line)
    if ( i>0 ) then
      if ( line(i:i) == "&" ) then
        call getline_io(unit, line2, stat2)
        if ( stat2 /= 0 ) then
          stat = stat2
          return
        end if
        if ( line2(1:1)/="&" ) then
          stat = 42
          return
        end if
        line = line // line2
        !> remove continuation symbols
        do
          i = index(line, "&&")
          if ( i.eq.0 ) exit
          line = trim( line(:i-1) // adjustl(line(i+2:)) )
        end do
      end if
    end if

    !> kill all blanks after characters in `killtrail`
    do j = 1,len(killtrail)
      pattern = killtrail(j:j)//" "
      do
        i = index(line, pattern)
        if ( i.eq.0 ) exit
        line = trim( line(:i) // adjustl(line(i+1:)) )
      end do
    end do

    !> kill all blanks before characters in `killlead`
    do j = 1,len(killlead)
      pattern = " "//killlead(j:j)
      do
        i = index(line, pattern)
        if ( i.eq.0 ) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do
    end do

  end subroutine getline_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Rewind a `line`
  !-----------------------------------------------------------------------------
  recursive subroutine rewindline_io(unit)
    use, intrinsic :: iso_fortran_env, only: iostat_eor
    integer, intent(in) :: unit
    character(:), allocatable :: line
    integer :: stat
    character(256) :: buffer
    integer :: size
    integer :: i,j

    backspace(unit) ! un-do the last read
    !> read the line
    line = ''
    do
      read(unit, "(A)", advance='no', iostat=stat, size=size) buffer
        if (stat > 0) return ! error
        line = line // buffer(:size)
        if (stat < 0) then
          if (stat == iostat_eor) stat = 0
          exit
        end if
    end do
    backspace(unit) ! un-do the last read

    !> replace all `tab` characters by a blank
    do
      i = index(line, char(9))
      if ( i.eq.0 ) exit
      line(i:i) = " "
    end do

    line = trim(adjustl( line ))  ! kill leading blanks
    !> kill possible comments
    i = index(line, "!")
    if ( i.gt.0 ) line = trim(adjustl( line(:i-1) ))  ! kill trailing blanks

    !> do we have a line continuation?
    if ( line(1:1)=="&" ) then
      call rewindline_io(unit)
    end if

  end subroutine rewindline_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> count the number of columns in a line delimited by blanks.
  !> possible comments starting with `!` are ignored
  !
  !> @param[in] line input string containing the line
  !> @return ncols # of columns
  !-----------------------------------------------------------------------------
  pure function ncols_io (line) result (ncols)
    integer :: ncols
    character(*), intent(in) :: line
    character(len(line)) :: str

    Integer :: i
    logical :: qfound

    str = adjustl(line) ! kill all leading blanks

    ncols = 0
    qfound = .true.

    do i = 1, len_trim(str)
       if ( str(i:i).eq."!" ) return
       if ( str(i:i).eq." " ) then
          qfound = .true.
       endif
       if ( qfound .and. str(i:i).ne." " ) then
          ncols = ncols + 1
          qfound = .false.
       endif
    end do

  end function ncols_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> extract a flag: check if `optname` exists & remove
  !> optionally the flag can also have suboptions:  `flag[opt=val,...]`
  !
  !> @param[inout] line a line of the runcard to extract a flag from
  !> @param[in] optname name of the option to extract
  !> @param[out] qfail signal success of parsing the line
  !> @param[out] suboptions optional string to parse sub-options
  !-----------------------------------------------------------------------------
  subroutine extract_flag_io(line, optname, qfail, suboptions)
    character(:), allocatable, intent(inout) :: line
    character(*), intent(in) :: optname
    logical, intent(out) :: qfail
    character(:), allocatable, intent(out), optional :: suboptions
    integer :: idx, idx1, idx2, optlen
    logical :: shift
    character(:), allocatable :: value

    qfail = .false.
    optlen = len( optname )

    !> find matching option
    idx1 = 1
    idx2 = 0
    do
      idx2 = index(to_upper_str(line(idx1:)), to_upper_str(optname))
      ! couldn't find a match
      if (idx2 == 0) then
        idx1 = 0
        exit
      end if
      idx1 = idx1 + idx2 - 1
      !> match was in a substring of another option?
      shift = .false.
      idx2 = idx1 + optlen
      if ( idx1 > 1 ) then
        if ( line(idx1-1:idx1-1) /= " " ) shift = .true.
      end if
      if ( idx2 < len(line) ) then
        if ( line(idx2:idx2) /= " " ) then
          !> is there a suboption?
          if  ( line(idx2:idx2) == "[" ) then
            idx = index(line(idx2:)," ")
            if (idx > 0) then
              idx2 = idx2 + idx - 1
            else
              idx2 = 0  ! go till the end
            end if
          else
            !> not space, nor "["  => not the flag we're looking for
            shift = .true.
          end if
        else if ( idx2+1 < len(line) ) then
          !> is there a space-separted suboption block?
          if  ( line(idx2+1:idx2+1) == "[" ) then
            idx = index(line(idx2+1:)," ")
            if (idx > 0) then
              idx2 = idx2 + idx
            else
              idx2 = 0  ! go till the end
            end if
          end if
        end if
      end if
      if ( shift ) then
        idx1 = idx1 + 1 ! shift one
        cycle  ! this is not the match you are looking for  *wavehand*
      end if
      ! found the match we were after
      exit
    end do

    !> remove option
    if (idx1 > 0) then
      if ( present(suboptions) ) then
        if (idx2 > 0) then
          value = line(idx1:idx2-1)
        else
          value = line(idx1:)
        end if
        call extract_suboptions_io(value, suboptions)
      end if
      if (idx2 > 0) then
        line = trim(adjustl(  &
             & trim( line(:idx1-1) ) // " " //  &
             & adjustl( line(idx2:) ) ))
      else
        line = trim(adjustl( line(:idx1-1) ))
      end if
    else
      !> couldn't find a matching optname
      qfail = .true.
    end if

  end subroutine extract_flag_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> extract a long option-value from the structure: "optname=value[suboptions]"
  !
  !> @param[inout] line a line of the runcard to extract an option from
  !> @param[in] optname name of the option to extract
  !> @param[out] qfail signal success of parsing the line
  !> @param[out] value value of the option as a string
  !> @param[out] suboptions optional string to parse sub-options
  !-----------------------------------------------------------------------------
  subroutine extract_option_io(line, optname, qfail, value, suboptions)
    character(:), allocatable, intent(inout) :: line
    character(*), intent(in) :: optname
    logical, intent(out) :: qfail
    character(:), allocatable, intent(out) :: value
    character(:), allocatable, intent(out), optional :: suboptions
    integer :: idx1, idx2, idx, optlen

    qfail = .false.
    optlen = len( optname )

    !> find matching option
    idx1 = 1
    do
      idx2 = index(to_upper_str(line(idx1:)), to_upper_str(optname)//"=")
      !> couldn't find a match
      if (idx2 == 0) then
        idx1 = 0
        exit
      end if
      idx1 = idx1 + idx2 - 1
      !> match was in a substring of another option ("=" already matched above)
      if ( idx1 > 1 ) then
        if ( line(idx1-1:idx1-1) /= " " ) then
          idx1 = idx1 + 1 ! shift one
          cycle  ! this is not the match you are looking for  *wavehand*
        end if
      end if
      !> found the match we were after
      exit
    end do

    !> get value & remove option
    if (idx1.gt.0) then
      value = line(idx1+optlen+1:)
      idx2 = index(value," ")
      !> are there sub-options separated by a space?
      if (idx2.gt.0) then
        if (value(idx2+1:idx2+1)=="[") then
          idx = index(value(idx2+1:)," ")
          if (idx>0) then
            idx2 = idx2 + idx
          else
            idx2 = 0  ! go till the end
          end if
        end if
      end if
      if (idx2.gt.0) then
        value = value(:idx2-1)
        line = trim(adjustl(  &
             & trim( line(:idx1-1) ) // " " //  &
             & adjustl( line(idx1+idx2+optlen+1:) ) ))
      else
        line = trim(adjustl( line(:idx1-1) ))
      end if

      !> time to look for sub-options
      if ( present(suboptions) ) call extract_suboptions_io(value, suboptions)

    else
      !> couldn't find a matching optname
      qfail = .true.
    end if

  end subroutine extract_option_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> extract a sub-options from the structure: "value[A,B,...]"
  !
  !> @param[inout] value string of the option (return: stripped by "[A,B,...]")
  !> @param[out] suboptions string to parse sub-options ("A B ...")
  !-----------------------------------------------------------------------------
  subroutine extract_suboptions_io(value, suboptions)
    character(:), allocatable, intent(inout) :: value
    character(:), allocatable, intent(out) :: suboptions
    integer :: idx1, idx2, idx

    suboptions = ''
    idx1 = index(value,"[")
    if (idx1>0) then
      idx2 = index(value,"]")
      if (idx2==0) then
        print*, "extract_suboptions_io: no closing bracket for suboptions: ", value
        stop
      end if
      suboptions = value(idx1+1:idx2-1)
      value = trim(adjustl( value(:idx1-1) ))
      !> convert commas into spaces so we can use `extract_option` on this string
      do
        idx = index(suboptions,",")
        if (idx==0) exit  ! done
        suboptions(idx:idx) = " "
      end do
    end if

  end subroutine extract_suboptions_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> extract an option-value
  !
  !> @param[inout] line a line of the runcard to extract an option from
  !> @param[in] optname name of the option to extract
  !> @param[out] qfail signal success of parsing the line
  !> @param[out] valueINT optional integer value of the option
  !> @param[out] valueDBLE optional double value of the option
  !> @param[out] valueBOOL optional bool value of the option
  !-----------------------------------------------------------------------------
  subroutine extract_option_type_io( line, optname, qfail,  &
                                   & valueINT,  &
                                   & valueDBLE,  &
                                   & valueBOOL )
    character(:), allocatable, intent(inout) :: line
    character(*), intent(in) :: optname
    logical, intent(out) :: qfail
    integer, intent(out), optional :: valueINT
    double precision, intent(out), optional :: valueDBLE
    logical, intent(out), optional :: valueBOOL
    integer :: ierr
    character(:), allocatable :: valueSTR, subopt

    !> get the value as a string
    call extract_option_io(line, optname, qfail, valueSTR, suboptions=subopt)
    if (qfail) return

    if ( subopt /= '' ) then
      print*, "extract_option_type_io: ignoring suboptions: ", subopt
    end if

    !> try to read in an integer
    if ( present(valueINT) ) then
      read(valueSTR,*,iostat=ierr) valueINT
      if (ierr/=0) qfail = .true.
    end if
    !> try to read in a double
    if ( present(valueDBLE) ) then
      read(valueSTR,*,iostat=ierr) valueDBLE
      if (ierr/=0) qfail = .true.
    end if
    !> try to read in a bool
    if ( present(valueBOOL) ) then
      read(valueSTR,*,iostat=ierr) valueBOOL
      if (ierr/=0) qfail = .true.
    end if

  end subroutine extract_option_type_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse a line for set up parameter
  !
  !> @param[inout] line one line of the runcard containing a selector
  !> @param[out] varname name of the variable
  !> @param[out] opt value of the variable
  !> @param[out] qfail signal success of parsing the line
  !-----------------------------------------------------------------------------
  subroutine readline_setup_io(line, varname, opt, opt1, qfail)
    use Observables_mod
    character(:), allocatable, intent(inout) :: line
    character(:), allocatable, intent(out) :: varname,opt,opt1
    logical, intent(out) :: qfail
    integer :: ncols, i,j,k

    !> init return values
    varname = ''
    qfail = .false.

    ncols = ncols_io(line)
    if ( ncols.gt.3 .or. ncols.lt.2) then
      qfail = .true.
      return
    endif

    !> get obsname
    i = index(line," ")
    varname = to_upper_str(line(:i-1))
    line = trim(adjustl( line(i+1:) ))
    if(ncols.eq.2) then
       opt = to_upper_str(line)
       opt1=""
    else
       i = index(line," ")
       opt = to_upper_str(line(:i-1))
       line = trim(adjustl( line(i+1:) ))
       opt1 = to_upper_str(line)
    endif
  end subroutine readline_setup_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse a line for a selector
  !
  !> @param[inout] line one line of the runcard containing a selector
  !> @param[out] obsname name of the observable
  !> @param[out] qmin flag if a min value was found
  !> @param[out] qmax flag if a max value was found
  !> @param[out] qinv flag if the invert flag was set
  !> @param[out] dmin value of the min cut
  !> @param[out] dmax value of the max cut
  !> @param[out] qfail signal success of parsing the line
  !-----------------------------------------------------------------------------
  subroutine readline_sel_io(line, obsname, qmin,qmax,qinv,qval, dmin,dmax, qfail)
    use Selectors_mod
    character(:), allocatable, intent(inout) :: line
    character(:), allocatable, intent(out) :: obsname
    logical, intent(out) :: qmin,qmax,qinv,qval, qfail
    double precision, intent(out) :: dmin,dmax
    character(:), allocatable :: cuttype, opt
    integer :: ncols, i,j,k

    !> init return values
    obsname = ''
    qmin = .false.
    qmax = .false.
    qinv = .false.
    qval = .true.  ! obs must be valid by default
    dmin = 0d0
    dmax = 0d0
    qfail = .false.

    ncols = ncols_io(line)
    if ( ncols.lt.3 .or. ncols.gt.4 ) then
      qfail = .true.
      return
    endif

    !> get cuttype
    i = index(line," ")
    cuttype = to_upper_str(line(:i-1))
    line = trim(adjustl( line(i+1:) ))
    select case (cuttype)
    case ("SELECT","ACCEPT")
      qinv = .false.
    case ("REJECT")
      qinv = .true.
    !> these guys don't need to be valid observables
    case ("SELECT_IF_VALID","ACCEPT_IF_VALID")
      qinv = .false.
      qval = .false.
    case ("REJECT_IF_VALID")
      qinv = .true.
      qval = .false.
    end select

    !> get obsname
    i = index(line," ")
    obsname = line(:i-1)
    line = trim(adjustl( line(i+1:) ))

    !> get options (one or two)
    do k = 3,ncols
      i = index(line," ")
      if (i.gt.0) then
        opt = to_upper_str(line(:i-1))
        line = trim(adjustl( line(i+1:) ))
      else
        opt = to_upper_str(line)
        line = ''
      endif
      select case (opt(1:3))
      case ("MIN")
        qmin = .true.
        read(opt(5:),*) dmin
      case ("MAX")
        qmax = .true.
        read(opt(5:),*) dmax
      end select
    enddo

    !> check selector for consistency
    if ( (qmin .and. qmax .and. dmax.lt.dmin) .or. &
       & (ncols.eq.4 .and. .not.(qmin.and.qmax)) ) then
      qfail = .true.
      return
    endif

  end subroutine readline_sel_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse a selectors block
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !> @param[in] label string label that indicates the start of the block
  !> @param[out] sel_p pointer to the linked list of selectors in the block
  !> @param[inout] selOpt option (will be unchanged if no option found)
  !-----------------------------------------------------------------------------
  subroutine parse_sel_block_io(unit, label, sel_p, selOpt)
    use Selectors_mod
    use Observables_mod
    integer, intent(in) :: unit
    character(*), intent(in) :: label
    type(Selector_t), intent(out), pointer :: sel_p
    integer, intent(inout) :: selOpt
    !> selectors
    type(Selector_t), pointer :: newSel
    character(:), allocatable :: line_sel, spec_sel, obsname_sel
    double precision :: dmin_sel, dmax_sel
    logical :: qmin,qmax,qinv,qval, qfail
    !> local vars
    integer :: i, j
    integer :: stat

    !> check if the next line starts a list of selectors
    nullify(sel_p)  ! reset
    loop_sel: do
      call getline_io(unit, line_sel, stat)
      if (stat /= 0) then
        print*, "parse_sel_block_io: error"
        stop
      end if
      if (line_sel.eq.'') cycle ! skip empty lines

      !> selectors block can have options:
      !> grab the specifier
      j = ncols_io(line_sel)
      if (j > 1) then
        !> look for the specifier
        i = index(line_sel," ")
        spec_sel = line_sel(:i-1)
      else
        spec_sel = line_sel
      end if

      !> found selectors list: parse them
      if (to_upper_str(spec_sel).eq.to_upper_str(trim(label))) then

        !> parse options for histogram selectors
        call extract_option_type_io(line_sel, "option", qfail, valueINT=selOpt)
        !> vaue passed to routine is the default!!!
        ! if (qfail) selOpt = 0  ! default value

        do
          call getline_io(unit, line_sel, stat)
          if (stat /= 0) then
            print*, "parse_sel_block_io: error"
            stop
          end if
          if (line_sel.eq.'') cycle ! skip empty lines
          if (to_upper_str(line_sel).eq.to_upper_str('END_'//trim(label))) exit loop_sel

          call readline_sel_io( line_sel, obsname_sel, &
                              & qmin,qmax,qinv,qval, &
                              & dmin_sel,dmax_sel, qfail)
          if (qfail) then
            print*, "parse_sel_block_io: invalid histogram selector: ", line_sel
            stop
          endif

          allocate(newSel)
          newSel%obsId = getIdFromName_obs(obsname_sel)
          if (newSel%obsId < 1) then
            deallocate(newSel)
            call cleanselectors
            print*, "parse_sel_block_io: observable not allowed in histogram selectors:", obsname_sel
            stop
          end if
          if (qmin) newSel%lower = dmin_sel
          if (qmax) newSel%upper = dmax_sel
          newSel%qlower  = qmin
          newSel%qupper  = qmax
          newSel%qinvert = qinv
          newSel%qvalid  = qval
          newSel%qisint  = isInteger_obs(newSel%obsId)
          ! pop it to the stack
          newSel%next => sel_p
          sel_p => newSel

        end do

      !> no selector list undo read and continue with histogram init
      else
        call rewindline_io(unit) ! un-do the last read
        exit loop_sel
      end if
    end do loop_sel

  contains

    !> kill selectors
    subroutine cleanselectors()
      type(Selector_t), pointer :: it_sel, next_sel
      it_sel => sel_p
      do while ( associated(it_sel) )
        next_sel => it_sel%next  ! save next
        deallocate(it_sel)
        it_sel => next_sel  ! loop
      end do
      sel_p => null()
    end subroutine cleanselectors

  end subroutine parse_sel_block_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse an isolation option `algo1[...]+algo2[...] ...`
  !
  !> @param[inout] line string containing the r.h.s. of `isolation = `
  !> @param[inout] isol_p optional pointer to the linked list of isolation algorithms
  !>             (default: register to the main PhotonIsolation head)
  !> @param[in] leptons flag to switch to lepton isolation
  !-----------------------------------------------------------------------------
  subroutine parse_isol_io(line, isol_p, leptons)
    use PhotonIsolation_mod
    use LeptonIsolation_mod
    character(:), allocatable, intent(inout) :: line
    type(Isol_t), intent(inout), optional, pointer :: isol_p
    logical, intent(in), optional :: leptons
    character(1), parameter :: pattern = "+"
    character(:), allocatable :: value, subopt
    integer :: i
    logical :: qfail
    !> PHOTON:
    integer :: ialgo, iopt
    double precision :: R
    double precision :: ETmax_const, ETmax_epsilon
    double precision :: R_min, ET_threshold  ! fixed
    double precision :: n  ! smooth
    double precision :: R_inner  ! hybrid
    double precision :: zcut, beta  ! softdrop
    integer :: jetalgo2  ! softdrop
    !> LEPTON:
    logical :: qleptons

    qleptons = .false.
    if ( present(leptons) ) qleptons = leptons

    if ( present(isol_p) ) then
      if (qleptons) then
        print*, "parse_isol_io: leptons don't have a pointer!"
        stop
      end if
      nullify(isol_p)  ! reset
    end if

    do
      ! print*, " line: >", line, "<"
      i = index(line, pattern)
      if ( i > 0 ) then
        if (qleptons) then
          print*, "parse_isol_io: leptons only have one isolation!"
          stop
        end if
        value = trim( line(:i-1) )
        line = trim(adjustl( line(i+len(pattern):) ))
      else
        value = line
        line = ""
      end if

      !print*, " value + line: >",value,"< >",line, "<"
      call extract_suboptions_io(value, subopt)
      !print*, " extracted: >", value,"[",subopt,"]", "<"

      ialgo = getiAlgo_isol(value, iopt=iopt)
      if (ialgo < 0) then
        print*, "parse_isol_io: unknown isolation algorithm: ", value
        stop
      end if

      if (ialgo>0 .and. ialgo/=5) then
        !> all (real) algorithms have a cone size (exception: democratic[=5])
        !! > WF's modification on runcard parser syntax for the release:
        !! > A backward-compatible implementation that makes sure all the previous runcards work
        !! > Using R_cone as the default cone size option;
        !! > Search for R instead with a warning message, if Rcone is nowhere to be found
        !! > Reason: cleaner syntax when describing photon isolation prescription in the paper
        !! > even though in the backend all of these are still called `R`
        ! call extract_option_type_io(subopt, "R", qfail, valueDBLE=R)
        call extract_option_type_io(subopt, "R_cone", qfail, valueDBLE=R)
        if (qfail) then
          write(*,*) "Warning: parse_isol_io: missing cone size `R_cone=...`, looking for `R=` instead..."
          call extract_option_type_io(subopt, "R", qfail, valueDBLE=R)
          if (qfail) then
            print*, "parse_isol_io: invalid/missing cone size `R_cone=...` OR `R=...` specification"
            stop
          endif
        end if
        !> better be positive
        if ( R <= 0d0 ) then
          print*, "parse_isol_io: invalid R =", R
          stop
        end if
      else
        R = 0d0
      end if

      !> get ETmax properties (needed by most)
      if ( (ialgo==1) .or. (ialgo==2) .or. (ialgo==3) ) then
        call extract_option_type_io(subopt, "ETmax_const", qfail, valueDBLE=ETmax_const)
        if (qfail) ETmax_const = 0d0
        call extract_option_type_io(subopt, "ETmax_epsilon", qfail, valueDBLE=ETmax_epsilon)
        if (qfail) ETmax_epsilon = 0d0
        !> better be positive
        if ( ETmax_const < 0d0 ) then
          print*, "parse_isol_io: invalid ETmax_const =", ETmax_const
          stop
        end if
        if ( ETmax_epsilon < 0d0 ) then
          print*, "parse_isol_io: invalid ETmax_epsilon =", ETmax_epsilon
          stop
        end if
        !> if both are zero: that's no good
        if ( (ETmax_const==0d0) .and. (ETmax_epsilon==0d0) ) then
          print*, "parse_isol_io: invalid/missing ETmax (const and/or epsilon) specification"
          stop
        end if
      end if

      !> standard cone can have more options
      if (ialgo==1) then
        call extract_option_type_io(subopt, "R_min", qfail, valueDBLE=R_min)
        if (qfail) R_min = 0d0
        call extract_option_type_io(subopt, "ET_threshold", qfail, valueDBLE=ET_threshold)
        if (qfail) ET_threshold = 0d0
        !> better be positive
        if ( R_min < 0d0 ) then
          print*, "parse_isol_io: invalid R_min =", R_min
          stop
        end if
        if ( ET_threshold < 0d0 ) then
          print*, "parse_isol_io: invalid ET_threshold =", ET_threshold
          stop
        end if
      end if

      !> smooth variants have a parameter to control the profile
      if ( (ialgo==2) .or. (ialgo==3) ) then
        call extract_option_type_io(subopt, "n", qfail, valueDBLE=n)
        if (qfail) then
          print*, "parse_isol_io: invalid/missing smooth-cone `n=...` specification"
          stop
        end if
        !> better be positive
        if ( n <= 0d0 ) then
          print*, "parse_isol_io: invalid n =", n
          stop
        end if
      end if

      !> hybrid algorithms need to specify the inner cone
      if (ialgo==3) then
        call extract_option_type_io(subopt, "R_inner", qfail, valueDBLE=R_inner)
        if (qfail) then
          print*, "parse_isol_io: invalid/missing hybrid `R_inner=...` specification"
          stop
        end if
        !> better be smaller than the outer cone size
        if ( R_inner > R ) then
          print*, "parse_isol_io: R_inner > R_cone in hybrid algorithm?!"
          stop
        end if
      end if

      !> zcut for soft drop and democratic
      if ( (ialgo==4) .or. (ialgo==5) ) then
        call extract_option_type_io(subopt, "zcut", qfail, valueDBLE=zcut)
        if (qfail) then
          print*, "parse_proc_io: invalid/missing softdrop zcut=... specification"
          stop
        end if
        !> better be sensible
        if ( (zcut < 0d0) .or. (zcut > 1d0) ) then
          print*, "parse_isol_io: zcut not in [0,1] ?!"
          stop
        end if
      end if

      !> soft-drop has its own properties
      if (ialgo==4) then
        call extract_option_type_io(subopt, "beta", qfail, valueDBLE=beta)
        if (qfail) then
          print*, "parse_proc_io: invalid/missing softdrop beta=... specification"
          stop
        end if
        call extract_option_type_io(subopt, "jetalgo2", qfail, valueINT=jetalgo2)
        if (qfail) then
          print*, "parse_proc_io: invalid/missing softdrop jetalgo2=... specification"
          stop
        end if
      end if

      !> actually register the non-trivial guys:
      if (qleptons) then
        !> LEPTONS
        select case (ialgo)
        case (1)  ! fixed
          call setLeptonIsol_lepisol( R,  &
            & ETmax_const=ETmax_const, ETmax_epsilon=ETmax_epsilon,  &
            & R_min=R_min, ET_threshold=ET_threshold )
        case default
          print*, "parse_proc_io: invalid algorithm for leptons?!"
          stop
        end select
      else
        !> PHOTONS
        select case (ialgo)
        case (0)  ! none
          call registerPhotonIsol_isol( ialgo, iopt, R, isol_p=isol_p )
        case (1)  ! fixed
          call registerPhotonIsol_isol( ialgo, iopt, R,  &
            & ETmax_const=ETmax_const, ETmax_epsilon=ETmax_epsilon,  &
            & R_min=R_min, ET_threshold=ET_threshold, isol_p=isol_p )
        case (2)  ! smooth
          call registerPhotonIsol_isol( ialgo, iopt, R,  &
            & ETmax_const=ETmax_const, ETmax_epsilon=ETmax_epsilon,  &
            & n=n, isol_p=isol_p )
        case (3)  ! hybrid
          call registerPhotonIsol_isol( ialgo, iopt, R,  &
            & ETmax_const=ETmax_const, ETmax_epsilon=ETmax_epsilon,  &
            & n=n, R_inner=R_inner, isol_p=isol_p )
        case (4)  ! softdrop
          call registerPhotonIsol_isol( ialgo, iopt, R,  &
            & zcut=zcut, beta=beta, ja2=jetalgo2, isol_p=isol_p )
        case (5)  ! democratic
          call registerPhotonIsol_isol( ialgo, iopt, zcut=zcut )
        case default
          print*, "parse_proc_io: invalid algorithm for photons?!"
          stop
        end select
      end if

      if ( line == "" ) exit
    end do  ! .and.

  end subroutine parse_isol_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for the process specifications in readruncard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_proc_io(unit)
    use Process_mod
    use ProcLib_mod
    use Observables_mod
    use Jets_mod
    use Selectors_mod
    use PhotonIsolation_mod
    use PhotonFragmentation_mod
    use HadronFragmentation_mod
    use Flavour_mod
    use DIS_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier, opt, subopt
    integer :: stat, ncols, i,j,k, ispec
    logical :: qfail, qfound
    integer :: valINT
    double precision :: valDBLE
    logical :: valBOOL
    character(:), allocatable :: valSTR
    character(2), parameter :: pattern = " ["
    !> COLLIDER:
    character(:), allocatable :: beam1, beam2
    double precision :: sqrts, Ebeam1, Ebeam2
    !> JET:
    character(:), allocatable :: jet_algo, jet_recomb
    double precision :: jet_param
    !> PHOTON FRAGMENTATION:
    integer :: nfrag

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a valid `PROCESS` block!"
        stop
      end if

      !> look for specifier
      i = index(line," ")
      if (i/=0) then
        specifier = line(:i-1)
        line = trim(adjustl( line(i+1:) ))
      else
        specifier = line
        line = ''
      end if
      if (to_upper_str(specifier).eq.'PROCESS') exit
    end do

    if (line == '') then
      print*, "parse_proc_io: cannot find the process name after `PROCESS`"
      stop
    end if


    !> found a process block
    call init_proc(line)
    call init_plib()
    !> init beam names
    beam1 = ''
    beam2 = ''
    !> init energies
    sqrts  = 0d0
    Ebeam1 = 0d0
    Ebeam2 = 0d0
    !> init jet
    jet_algo   = 'none'
    jet_recomb = 'v4'
    jet_param  = 0d0

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_PROCESS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_PROCESS') exit

      !> kill all blanks before "[" that can be used to specify a sub-option
      do
        i = index(line, pattern)
        if (i==0) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      !>----- parse all options

      !> collider
      call extract_option_io(line, "collider", qfail, valSTR)
      if (.not.qfail) then
        select case (to_upper_str(valSTR))
          case ('LHC','PP')
            beam1 = 'p'
            beam2 = 'p'
          case ('TEVATRON','PAP','PPBAR')
            beam1 = 'p'
            beam2 = 'ap'
          case ('DIS','EP')
            beam1 = 'em'
            beam2 = 'p'
          case ('PE')
            beam1 = 'p'
            beam2 = 'em'
          case ('LEP','EPEM')
            beam1 = 'ep'
            beam2 = 'em'
          case default
            print*, "parse_proc_io: don't know your collider: ", valSTR
            stop
        end select
      end if

      !> beam1
      call extract_option_io(line, "beam1", qfail, valSTR)
      if (.not.qfail) then
        beam1 = valSTR
      end if

      !> beam2
      call extract_option_io(line, "beam2", qfail, valSTR)
      if (.not.qfail) then
        beam2 = valSTR
      end if

      !> sqrts
      call extract_option_type_io(line, "sqrts", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        sqrts = valDBLE
      end if

      !> Ebeam1
      call extract_option_type_io(line, "Ebeam1", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        Ebeam1 = valDBLE
      end if

      !> Ebeam2
      call extract_option_type_io(line, "Ebeam2", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        Ebeam2 = valDBLE
      end if

      !> jet algorithm
      call extract_option_io(line, "jet", qfail, valSTR, suboptions=subopt)
      if (.not.qfail) then
        jet_algo = valSTR
        !> sub-option is the parameter
        select case (to_lower_str(jet_algo))
          case ('none')
            !> nothing to do
          case ('antikt','akt','ca','cam','kt')
            !> dynamical soft cone (increase R dynamically for soft emissions)
            call extract_option_type_io(subopt, "dyn_soft", qfail, valueDBLE=valDBLE)
            if (.not.qfail)  then
               call setParam_jet("DYN_SOFT", valueDBLE=valDBLE)
            end if
            !> sub-option is the cone size
            jet_param = to_dble_str(subopt, qfail)
            if (qfail) then
              print*, "parse_proc_io: missing Rcone option for jet algorithm ", jet_algo
              stop
            end if
          case ('flavkt')
            !> flavour-kt has additional options
            call extract_option_type_io(subopt, "alpha", qfail, valueDBLE=valDBLE)
            if (.not.qfail)  then
              call setParam_jet("FLAVKT_ALPHA", valueDBLE=valDBLE)
            end if
            call extract_option_type_io(subopt, "include_colourless", qfail, valueBOOL=valBOOL)
            if (.not.qfail)  then
              call setParam_jet("FLAVKT_INCLUDE_COLOURLESS", valueBOOL=valBOOL)
            end if
            call extract_option_type_io(subopt, "only_proto", qfail, valueBOOL=valBOOL)
            if (.not.qfail)  then
              call setParam_jet("FLAVKT_ONLY_PROTO", valueBOOL=valBOOL)
            end if
            call extract_option_type_io(subopt, "sum_nonqcd", qfail, valueBOOL=valBOOL)
            if (.not.qfail)  then
              call setParam_jet("FLAVKT_SUM_NONQCD", valueBOOL=valBOOL)
            end if
            !> remaining sub-option is the cone size
            jet_param = to_dble_str(subopt, qfail)
            if (qfail) then
              print*, "parse_proc_io: missing Rcone option for jet algorithm ", jet_algo
              stop
            end if
          case ('flavantikt')
            !> flavour anti-kt has additional options
            call extract_option_type_io(subopt, "a", qfail, valueDBLE=valDBLE)
            if (.not.qfail)  then
              call setParam_jet("FLAVANTIKT_A", valueDBLE=valDBLE)
            end if
            !> sub-option is the cone size
            jet_param = to_dble_str(subopt, qfail)
            if (qfail) then
              print*, "parse_proc_io: missing Rcone option for jet algorithm ", jet_algo
              stop
            end if
          case ('jade','durham')
            !> sub-option is the ycut value
            jet_param = to_dble_str(subopt, qfail)
            if (qfail) then
              print*, "parse_proc_io: missing ycut options for jet algorithm"
              stop
            end if
          case default
            print*, "parse_proc_io: unknown jet algorithm"
            stop
        end select
      end if

      !> jet exclusive flag
      call extract_option_type_io(line, "jet_exclusive", qfail, valueBOOL=valBOOL)
      if (.not.qfail) then
        qexclusive_proc = valBOOL
      end if

      !> jet recombination scheme
      call extract_option_io(line, "jet_recomb", qfail, valSTR)
      if (.not.qfail) then
        jet_recomb = valSTR
      end if

      !> jet flavour dressing
      call extract_option_io(line, "jet_flavour", qfail, valSTR, suboptions=subopt)
      if (.not.qfail) then
        print*, "JET_FLAVOUR: ", line, " > ", valSTR
        !> initialise the flavour dressing
        call init_dress_jet(valSTR)
        !> sub-option contains settings
        if ( algo_dress_jet() < 0 ) then
          !>----- recursive softdrop algorithms
          call extract_option_type_io(subopt, "zcut", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_SD_ZCUT", valueDBLE=valDBLE)
          end if
          call extract_option_type_io(subopt, "beta", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_SD_BETA", valueDBLE=valDBLE)
          end if
          !> remaining sub-option is the declustering algorithm
          if ( subopt /= '' ) then
            select case (to_lower_str(subopt))
              case ('antikt','akt')
                call setParam_jet("FLAV_DRESS_SD_ALGO", valueINT=-1)
              case ('ca','cam')
                call setParam_jet("FLAV_DRESS_SD_ALGO", valueINT=-2)
              case ('kt')
                call setParam_jet("FLAV_DRESS_SD_ALGO", valueINT=-3)
              case ('jade')
                call setParam_jet("FLAV_DRESS_SD_ALGO", valueINT=-4)
              case default
                print*, "parse_proc_io: unknown jet_flavour option: ", subopt
                stop
            end select
          end if
        end if
        if ( algo_dress_jet() > 0 ) then
          !>----- flavour clustering algorithms
          call extract_option_type_io(subopt, "proto", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_CLUSTER_PROTO", valueBOOL=valBOOL)
          end if
          call extract_option_type_io(subopt, "beam", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_CLUSTER_BEAM", valueBOOL=valBOOL)
          end if
          call extract_option_io(subopt, "beam_nonqcd", qfail, valSTR)
          if (.not.qfail)  then
            select case (to_lower_str(valSTR))
              case ('no')
                call setParam_jet("FLAV_DRESS_CLUSTER_BEAM_NONQCD", valueINT=0)
              case ('sum_et')
                call setParam_jet("FLAV_DRESS_CLUSTER_BEAM_NONQCD", valueINT=1)
              case ('individual_pt')
                call setParam_jet("FLAV_DRESS_CLUSTER_BEAM_NONQCD", valueINT=2)
              case default
                print*, "parse_proc_io: unknown dress option: ", valSTR
                stop
            end select
          end if
          call extract_option_type_io(subopt, "alpha", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_CLUSTER_ALPHA", valueDBLE=valDBLE)
          end if
          call extract_option_io(subopt, "dress", qfail, valSTR)
          if (.not.qfail)  then
            select case (to_lower_str(valSTR))
              case ('demo','democratic')
                call setParam_jet("FLAV_DRESS_CLUSTER_IDRESS", valueINT=0)
              case ('topdown','td')
                call setParam_jet("FLAV_DRESS_CLUSTER_IDRESS", valueINT=1)
              case default
                print*, "parse_proc_io: unknown dress option: ", valSTR
                stop
            end select
          end if
          call extract_option_io(subopt, "fj_penalty", qfail, valSTR)
          if (.not.qfail)  then
            select case (to_lower_str(valSTR))
              case ('none','off')
                call setParam_jet("FLAV_DRESS_CLUSTER_PENALTY", valueINT=0)
              case ('wilson')
                call setParam_jet("FLAV_DRESS_CLUSTER_PENALTY", valueINT=1)
              case ('mult')
                call setParam_jet("FLAV_DRESS_CLUSTER_PENALTY", valueINT=2)
              case default
                print*, "parse_proc_io: unknown fj_penalty option: ", valSTR
                stop
            end select
          end if
          call extract_option_type_io(subopt, "beta", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_DRESS_CLUSTER_BETA", valueDBLE=valDBLE)
          end if
          !> the underdress
          call extract_option_type_io(subopt, "udress", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            call setParam_jet("FLAV_QUNDERDRESS", valueBOOL=valBOOL)
          end if
          call extract_option_type_io(subopt, "udress_R", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_UNDERDRESS_R", valueDBLE=valDBLE)
          end if
          call extract_option_type_io(subopt, "udress_zcut", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_UNDERDRESS_ZCUT", valueDBLE=valDBLE)
          end if
          call extract_option_type_io(subopt, "udress_beta", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            call setParam_jet("FLAV_UNDERDRESS_BETA", valueDBLE=valDBLE)
          end if
          call extract_option_type_io(subopt, "keep_even_tag", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            call setParam_jet("FLAV_KEEP_EVEN_TAG", valueBOOL=valBOOL)
          end if
          !> remaining sub-option is the declustering algorithm
          !> (this should come last b/c it can override the `alpha`)
          if ( subopt /= '' ) then
            select case (to_lower_str(subopt))
              case ('jade')
                call setParam_jet("FLAV_DRESS_CLUSTER_ALGO", valueINT=4)
              case ('flavkt')
                call setParam_jet("FLAV_DRESS_CLUSTER_ALGO", valueINT=5)
              case default
                print*, "parse_proc_io: unknown jet_flavour option: ", subopt
                stop
            end select
          end if
        end if
      end if

      !> lepton isolation
      call extract_option_io(line, "lepton_isolation", qfail, valSTR)
      if (.not.qfail) then
        call parse_isol_io(valSTR, leptons=.true.)
      end if

      !> photon isolation
      call extract_option_io(line, "photon_isolation", qfail, valSTR)
      if (.not.qfail) then
        call parse_isol_io(valSTR)
      end if

      !> photon recombination
      call extract_option_type_io(line, "photon_recomb", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        call setRecombPhoton_isol(valDBLE)
      end if

      !> photon fragmentation
      call extract_option_io(line, "photon_fragmentation", qfail, valSTR, suboptions=subopt)
      if (.not.qfail) then
        !> find suboption
        nfrag = -1  ! switch on all parts
        call extract_option_type_io(subopt, "frag", qfail, valueINT=valINT)
        if (.not.qfail)  nfrag = valINT
        !> better have no left overs
        if ( subopt /= '' ) then
          print*, "parse_proc_io: left-over photon_fragmentation subopt: ", subopt
          stop
        end if
        !> first see if it's an bool
        valBOOL = to_bool_str(valSTR, qfail)
        if (.not.qfail) then
          call setActive_frag(valBOOL, frag=nfrag)
        else
          call setActive_frag(.true., fctn=valSTR, frag=nfrag)
        end if
      end if

      !> hadron fragmentation
      call extract_option_io(line, "hadron_fragmentation", qfail, valSTR, suboptions=subopt)
      if (.not.qfail) then
        !> sub-option is the member
        valINT = to_int_str(subopt, qfail)
        if (qfail) valINT = 0  ! default value
        call setActive_hfrag(valSTR, valINT)
      end if

      !> V_NC
      call extract_option_io(line, "V_NC", qfail, valSTR)
      if (.not.qfail) then
        valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then  ! user has passed the int flag directly
          igamma_proc = valINT
        else  ! try to convert string to int flag
          select case (to_upper_str(valSTR))
            case ('ZGAMMA','Z_GAMMA','Z+GAMMA')
              igamma_proc = 0
            case ('GAMMA')
              igamma_proc = 1
            case ('Z')
              igamma_proc = 2
            case default
              print*, "parse_proc_io: unknown V_NC option: ", valSTR
              stop
          end select
        end if
      end if
      !> L_NC
      call extract_option_io(line, "L_NC", qfail, valSTR)
      if (.not.qfail) then
        valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then  ! user has passed the int flag directly
          ilepton_proc = valINT
        else  ! try to convert string to int flag
          select case (to_upper_str(valSTR))
            case ('N','NU','NEUTRINO')
              ilepton_proc = 0
            case ('L','LEPTON')
              ilepton_proc = 1
            case default
              print*, "parse_proc_io: unknown L_NC option: ", valSTR
              stop
          end select
        end if
      end if

      !> decay option
      call extract_option_io(line, "decay_type", qfail, valSTR, suboptions=subopt)
      if (.not.qfail) valINT = to_int_str(valSTR, qfail)
      if (.not.qfail) then
        decay_type_proc = valINT
        if (decay_type_proc == 4) then
          !> need the Q value
          call extract_option_type_io(subopt, "Q", qfail, valueDBLE=valDBLE)
          if (.not.qfail)  then
            Q_OS_proc = valDBLE
          else
            print*, "parse_proc_io: (decay_type_proc == 4) requires a `Q=...` suboption!"
            stop
          end if
        end if
      end if

      !>----- DIS specific options
      !> PSDIS
      call extract_option_io(line, "psdis", qfail, valSTR)
      if (.not.qfail) then
        if(name_proc(1:3)/="DIS") then
          print*, "parse_run_io: psdis option not available for process: "  &
                  &   , name_proc
          stop
        end if
        if(to_upper_str(valSTR)=="OLD") then
          call setPSDIS(.True.)
        else
          write(*,*) "Only accept OLD as a valid option for PSDIS otherwise current PS will be used"
          stop
        endif
        qfound = .true.
      end if
      !> POL
      call extract_option_type_io(line, "pol", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        if(name_proc/="DIS" .and. name_proc/="DIS1J") then
          print*, "parse_run_io: pol option not available for process: "  &
                  &   , name_proc
          stop
        end if
        call setPol(valDBLE)
        qfound = .true.
      end if
      !> HERA_ORDERETA
      !> GS: renaming hera_ordereta -> lab_ordereta in runcard for release
      call extract_option_type_io(line, "lab_ordereta", qfail, valueBOOL=valBOOL)
      if (.not.qfail) then
        if(name_proc(1:3)/="DIS") then
          print*, "parse_run_io: lab_ordereta option not available for process: "  &
                  &   , name_proc
          stop
        end if
        call setHERA_ORDERETA(valBOOL)
        qfound = .true.
      end if
      !> DIS_EVENTSHAPES
      call extract_option_type_io(line, "dis_eventshapes", qfail, valueDBLE=valDBLE)
      if (.not.qfail) then
        if(name_proc(1:3)/="DIS") then
          print*, "parse_run_io: dis_eventshapes option not available for process: "  &
                  &   , name_proc
          stop
        end if
        call setEhfrac(valDBLE)
        call setDISEVTSHAPE(.true.)
        qfound = .true.
      end if
      !> EPRC
      call extract_option_type_io(line, "eprc", qfail, valueBOOL=valBOOL)
      if (.not.qfail) then
        if(name_proc(1:3)/="DIS") then
          print*, "parse_run_io: eprc option not available for process: "  &
                  &   , name_proc
          stop
        end if
        call set_EPRC(valBOOL)
        qfound = .true.
      end if
      !> DIS_FRAME
      call extract_option_io(line, "dis_frame", qfail, valSTR)
      if (.not.qfail) then
        if(name_proc(1:3)/="DIS") then
          print*, "parse_run_io: dis_frame option not available for process: "  &
                  &   , name_proc
          stop
        end if
        select case(to_upper_str(valSTR))
        case( "FIXT" )
          call setfixT(.true.)
          isSetBreitFrame = .true.
          useBreitFrame = .false.
        case( "BREIT" )
          call setfixT(.false.)
          isSetBreitFrame = .true.
          useBreitFrame = .true.
        case( "LAB" )
          call setfixT(.false.)
          isSetBreitFrame = .true.
          useBreitFrame = .false.
        case default
          print*, "parse_run_io: unknown option for DIS_FRAME, choose either FIXT, BREIT or LAB"
          stop
        end select
      end if

      !> gotta catch em all!
      if ( trim(line) /= "" ) then
        print*, "parse_proc_io: left over: ", line
        stop
      end if

    enddo

    !> done parsing: propagate accumulated settings
    !> beams
    if ( (beam1 == '') .or. (beam2 == '') ) then
      print*, "parse_proc_io: forgot to set `collider` or `beam[1|2]`?"
      stop
    else
      call setBeams_proc(beam1, beam2)
    end if
    !> energy
    if ( (sqrts == 0d0) .and. (Ebeam1*Ebeam2 == 0d0) ) then
      print*, "parse_proc_io: forgot to set `sqrts` or `Ebeam[1|2]`?"
      stop
    else if ( (sqrts /= 0d0) .and. (Ebeam1*Ebeam2 /= 0d0) ) then
      print*, "parse_proc_io: set both `sqrts` and `Ebeam[1|2]` at the same time?"
      stop
    else if ( sqrts /= 0d0 ) then
      call setSqrts_proc(sqrts)
    else if (Ebeam1*Ebeam2 /= 0d0 ) then
      call setEnergies_proc(Ebeam1, Ebeam2)
    else
      stop "parse_proc_io: should never get here..."
    end if
    !> jet algorithm
    call init_jet(jet_algo, jet_param, jet_recomb)

  end subroutine parse_proc_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for the run specifications in readruncard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_run_io(unit, smember)
    use Run_mod
    use ScaleCoeff_mod
    use Channels_mod
    use Checks_mod
    use Flavour_mod
    use LIPS_mod
    use Mapping_mod
    !> for process-specific settings
    use Process_mod, only : name_proc
    use sigHqt_mod  ! qt_cut & boost flag
    use sigZqt_mod  ! qt_cut & boost flag
    use sigGGqt_mod  ! qt_cut & boost flag
    use sigWqt_mod  ! qt_cut & boost flag
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier, opt, subopt
    character(5), optional, intent(in) :: smember
    integer :: stat, ncols, i,j,k, ispec
    logical :: qfail, qfound
    integer :: valINT
    integer :: member
    double precision :: valDBLE
    logical :: valBOOL
    character(:), allocatable :: valSTR
    character(2), parameter :: pattern = " ["
    !> PDF
    character(:), allocatable :: pdf_set
    integer :: pdf_mem
    !> SWEEP
    integer :: sweep_type, sweep_ncall, sweep_nit
    !> point check
    logical qinv, qmom, qobs


    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a valid `RUN` block!"
        stop
      end if

      !> look for specifier
      i = index(line," ")
      if (i/=0) then
        specifier = line(:i-1)
        line = trim(adjustl( line(i+1:) ))
      else
        specifier = line
        line = ''
      end if
      if (to_upper_str(specifier).eq.'RUN') exit
    end do

    if (line == '') then
      print*, "parse_run_io: cannot find a jobname name after `RUN`"
      stop
    end if

    !> found a run block
    call init_run(line)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_RUN` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_RUN') exit

      !> kill all blanks before "[" that can be used to specify a sub-option
      do
        i = index(line, pattern)
        if (i==0) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      !> loop over specs (multiple entries per line allowed)
      ncols = ncols_io(line)
      do ispec = 1,ncols
        !>----- get one specification
        i = index(line," ")
        if (i.gt.0) then
          opt = line(:i-1)
          line = trim(adjustl( line(i+1:) ))
        else
          opt = line
          line = ''
        endif
        !>----- parse the opt
        qfound =.false.
        !> PDF
        call extract_option_io(opt, "pdf", qfail, pdf_set, suboptions=subopt)
        if (.not.qfail) then
          !> sub-option is the member
          pdf_mem = to_int_str(subopt, qfail)
          if (qfail) pdf_mem = 0  ! default value
          !> was there an override from the CMD option `-imember`?
          if ( trim(smember) /= 'No') then
            member = to_int_str(smember, qfail)
            if (.not.qfail) pdf_mem = member
            if (.not.qfail) name_run = trim(name_run) // "." // trim(smember)
          end if
          call set_pdf_run(set=pdf_set, mem=pdf_mem)
          qfound = .true.
        end if
        !> possibility to set different PDFs
        call extract_option_io(opt, "pdf1", qfail, pdf_set, suboptions=subopt)
        if (.not.qfail) then
          !> sub-option is the member
          pdf_mem = to_int_str(subopt, qfail)
          if (qfail) pdf_mem = 0  ! default value
          call set_pdf_run(set=pdf_set, mem=pdf_mem)
          qfound = .true.
        end if
        call extract_option_io(opt, "pdf2", qfail, pdf_set, suboptions=subopt)
        if (.not.qfail) then
          !> sub-option is the member
          pdf_mem = to_int_str(subopt, qfail)
          if (qfail) pdf_mem = 0  ! default value
          call set_pdf_run(set2=pdf_set, mem2=pdf_mem)
          qfound = .true.
        end if
        !> the technical cutoff (a.k.a. y0)
        call extract_option_io(opt, "tcut", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) valDBLE = to_dble_str(valSTR, qfail)
        if (.not.qfail) then
          tcut_run = valDBLE
          select case(to_upper_str(subopt))
          case ("MAP","LOCAL")
            !> reject individually on the mapped momentum sets
            qtcut_veto_run = .false.
          case ("VETO","GLOBAL")
            !> switch to a global veto
            qtcut_veto_run = .true.
          case ("")
            !> default: backwards compatibility for now
            qtcut_veto_run = .false.
          case default
            print*, "parse_run_io: unrecognised tcut option:", subopt
            stop
          end select
          qfound = .true.
        end if
        !> seed id
        call extract_option_type_io(opt, "iseed", qfail, valueINT=valINT)
        if (.not.qfail) then
          iseed_run = valINT
          qfound = .true.
        end if
        !> imethod
        call extract_option_type_io(opt, "imethod", qfail, valueINT=valINT)
        if (.not.qfail) then
          imethod_run = valINT
          qfound = .true.
        end if
        !> iplot
        call extract_option_type_io(opt, "iplot", qfail, valueINT=valINT)
        if (.not.qfail) then
          iplot_run = valINT
          qfound = .true.
        end if
        !> phase-space parametrisation
        call extract_option_io(opt, "phase_space", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) then
          valDBLE = to_dble_str(subopt,qfail=qfail)
          if (.not.qfail) then
            call set_lips_run(valSTR, valDBLE)
          else
            call set_lips_run(valSTR)
          end if
          qfound = .true.
        end if
        !> perform an average over rotations to kill ang-corr
        call extract_option_type_io(opt, "angular_average", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          angular_average_run = valBOOL
          qfound = .true.
        end if
        !> activate cache
        call extract_option_type_io(opt, "cache_kinematics", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          call setActiveCache_map(valBOOL)
          qfound = .true.
        end if
        !> reset vegas grid
        call extract_option_type_io(opt, "reset_vegas_grid", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          reset_vegas_grid_run = valBOOL
          qfound = .true.
        end if
        !> keep track of the max/min weights and print
        call extract_option_type_io(opt, "print_max_weight", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          print_max_weight_run = valBOOL
          qfound = .true.
        end if
        !> switch to performing numerical pole checks
        call extract_option_type_io(opt, "pole_check", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          pole_check_run = valBOOL
          qfound = .true.
        end if
        !> switch to performing point checks
        call extract_option_io(opt, "point_check", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then
          call extract_flag_io(subopt, "print_invariants", valBOOL)
          qinv = .not. valBOOL
          call extract_flag_io(subopt, "print_momenta", valBOOL)
          qmom = .not. valBOOL
          call extract_flag_io(subopt, "print_obs", valBOOL)
          qobs = .not. valBOOL
          call setActive_point_chk(valINT, qinv=qinv, qmom=qmom, qobs=qobs)
          qfound = .true.
        end if
        !> switch to performing a decomposition into scale coefficients
        call extract_option_type_io(opt, "scale_coefficients", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          call setActive_sclc(valBOOL)
          qfound = .true.
        end if
        !> switch to performing a multi-channel adaption
        call extract_option_io(opt, "multi_channel", qfail, valSTR)
        if (.not.qfail) then
          !> first see if it's an int
          valINT = to_int_str(valSTR, qfail)
          if (.not.qfail) then
            call setActive_multi_chan(.true., nselect=valINT)
            qfound = .true.
          end if
          !> otherwise bool on/off flag
          valBOOL = to_bool_str(valSTR, qfail)
          if (.not.qfail) then
            call setActive_multi_chan(valBOOL)
            qfound = .true.
          end if
        end if
        !> observables-matching
        call extract_option_io(opt, "match_obs", qfail, valSTR)
        if (.not.qfail) then
          !> set the threshold & turn on
          valDBLE = to_dble_str(valSTR, qfail)
          if (.not.qfail) then
            call setActiveMatchObs_run(.true., threshold=valDBLE)
            qfound = .true.
          end if
          !> use default threshold: switch on/off
          valBOOL = to_bool_str(valSTR, qfail)
          if (.not.qfail) then
            call setActiveMatchObs_run(valBOOL)
            qfound = .true.
          end if
        end if
        !> flavour tagging
        call extract_option_io(opt, "tag_flavour", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then
          call setActive_flav(.true., tag=valINT)
          !> optional settings for the flavour tag
          call extract_option_type_io(subopt, "is_leading", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            is_leading_flav = valBOOL
          end if
          call extract_option_type_io(subopt, "isub", qfail, valueINT=valINT)
          if (.not.qfail)  then
            call set_sub_flav(valINT)
          end if
          call extract_option_type_io(subopt, "no_flav_sel", qfail, valueBOOL=valBOOL)
          if (.not.qfail)  then
            call set_minnpart_flav(valBOOL)
          end if
          if (subopt /= '') then
            print*, "parse_run_io: tag_flavour invalid options: ", subopt
            stop
          end if
          qfound = .true.
        end if
        !> switch to reduce LIPS
        call extract_option_type_io(opt, "lips_reduce", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          call setActive_reduce_lips(valBOOL)
          qfound = .true.
        end if
        !> SWEEP: warmup
        call extract_option_io(opt, "warmup", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then
          sweep_type  = 1
          sweep_ncall = valINT
          !> sub-option is the number of iterations
          valINT = to_int_str(subopt, qfail)
          if (.not.qfail) then
            sweep_nit = valINT
            call registerSweep_run(sweep_type, sweep_ncall, sweep_nit)
            qfound = .true.
          end if
          !> sub-option auto
          select case(to_upper_str(subopt))
          case ("AUTO")
            call registerSweep_run(sweep_type, sweep_ncall/168, 7)
            call registerSweep_run(sweep_type, sweep_ncall/24,  5)
            call registerSweep_run(sweep_type, sweep_ncall/4,   3)
            qfound = .true.
          end select
        end if
        !> SWEEP: production
        call extract_option_io(opt, "production", qfail, valSTR, suboptions=subopt)
        if (.not.qfail) valINT = to_int_str(valSTR, qfail)
        if (.not.qfail) then
          sweep_type  = 2
          sweep_ncall = valINT
          !> sub-option is the number of iterations
          valINT = to_int_str(subopt, qfail)
          if (.not.qfail) then
            sweep_nit = valINT
            call registerSweep_run(sweep_type, sweep_ncall, sweep_nit)
            qfound = .true.
          end if
        end if
        !>----- Process-specific options:
        !> qt_cut
        call extract_option_type_io(opt, "qt_cut", qfail, valueDBLE=valDBLE)
        if (.not.qfail) then
          select case (name_proc)
          case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
               & "HTO2L1PQT",  &
               & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
            call set_qtcut_Hqt(valDBLE)
          case ( "ZQT" )
            call set_qtcut_Zqt(valDBLE)
          case ( "GGQT" )
            call set_qtcut_GGqt(valDBLE)
          case ( "WPQT", "WMQT" )
            call set_qtcut_Wqt(valDBLE)
          case default
            print*, "parse_run_io: qt_cut option not available for process: "  &
              &   , name_proc
            stop
          end select
          qfound = .true.
        end if
        !> boost
        call extract_option_type_io(opt, "boost", qfail, valueBOOL=valBOOL)
        !> simple case of setting boost flag by true or false
        if (.not.qfail) then
          select case (name_proc)
          case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
               & "HTO2L1PQT",  &
               & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
            call set_boost_Hqt(valBOOL)
          case ( "ZQT" )
            call set_boost_Zqt(valBOOL)
          case ( "GGQT" )
            call set_boost_GGqt(valBOOL)
          case ( "WPQT", "WMQT" )
            call set_boost_Wqt(valBOOL)
          case default
            print*, "parse_run_io: boost option not available for process: "  &
              &   , name_proc
            stop
          end select
          qfound = .true.
        end if
        !> boostmultiply (do not use this flag if the 'boost' flag above was already used)
        call extract_option_type_io(opt, "boostmultiply", qfail, valueINT=valINT)
        !> can also set boostmultiply flag as an integer multiplier of qt_cut
        !> default value is set in sigXqt.f90
        if (.not.qfail) then
          select case (valINT)
          case(0)!> Same as boost = false: only consider accumulated contribution below qt_cut
            select case (name_proc)
            case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
                 & "HTO2L1PQT",  &
                 & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
              call set_boost_Hqt(.FALSE.)
            case ( "ZQT" )
              call set_boost_Zqt(.FALSE.)
            case ( "GGQT" )
              call set_boost_GGqt(.FALSE.)
            case ( "WPQT", "WMQT" )
              call set_boost_Wqt(.FALSE.)
            case default
              print*, "parse_run_io: boostmultiply option not available for process: "  &
                &   , name_proc
              stop
            end select
          case(1) !> Same as boost = true: consider both accumulated and differential (account for fiducial power correction) below qt_cut
            select case (name_proc)
            case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
                 & "HTO2L1PQT",  &
                 & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
              call set_boost_Hqt(.TRUE.)
            case ( "ZQT" )
              call set_boost_Zqt(.TRUE.)
            case ( "GGQT" )
              call set_boost_GGqt(.TRUE.)
            case ( "WPQT", "WMQT" )
              call set_boost_Wqt(.TRUE.)
            case default
              print*, "parse_run_io: boostmultiply option not available for process: "  &
                &   , name_proc
              stop
            end select
          case default
          !> no accumulated contribution below qt_cut and being differential in [qt_cut,qt_cut*boostmultiply] region.
          !> no contribution above qt_cut*boostmultiply value.
            select case (name_proc)
            case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
                 & "HTO2L1PQT",  &
                 & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
              call set_boost_Hqt(.TRUE.,valINT)
            case ( "ZQT" )
              call set_boost_Zqt(.TRUE.,valINT)
            case ( "GGQT" )
              call set_boost_GGqt(.TRUE.,valINT)
            case ( "WPQT", "WMQT" )
              call set_boost_Wqt(.TRUE.,valINT)
            case default
              print*, "parse_run_io: boostmultiply option not available for process: "  &
                &   , name_proc
              stop
            end select
          end select
          qfound = .true.
        end if
        !>----- success?
        if (.not. qfound) then
          print*, "parse_run_io: unknown option ", opt
          stop
        end if

      enddo  ! ispec

    enddo

    !> compatibility check
    if (isActive_flav() .and. isActive_reduce_lips()) then
      stop "parse_run_io: flavtag & lips-reduction cannot be active at the same time!"
    end if

  end subroutine parse_run_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse the selectors in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_sel_io(unit)
    use Run_mod
    use Selectors_mod
    use PhotonIsolation_mod
    use Jets_mod
    use Scales_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, obsname, obsname_padded
    integer :: stat, ncols, i,j,k
    double precision :: dmin,dmax
    logical :: qmin,qmax,qinv,qval, qfail
    integer :: ORgroup, nORgroups

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a `SELECTORS` block!"
        stop
      end if
      if (to_upper_str(line).eq.'SELECTORS') exit
    end do

    nORgroups = 0
    ORgroup   = 0

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_SELECTORS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_SELECTORS') exit

      !> look for OR group statements
      !> open
      if (to_upper_str(line).eq.'OR') then
        if (ORgroup /= 0) then
          print *, "opening a new OR block before closing the previous one"
          stop
        end if
        nORgroups = nORgroups + 1
        ORgroup   = nORgroups
        cycle
      end if
      !> close
      if (to_upper_str(line).eq.'END_OR') then
        if (ORgroup == 0) then
          print *, "closed an OR block before opening one"
          stop
        end if
        ORgroup = 0
        cycle
      end if

      call readline_sel_io(line, obsname, qmin,qmax,qinv,qval, dmin,dmax, qfail)
      if (qfail) then
        !print*, "skipping invalid selector"
        !cycle
        print*, "encountered invalid selector"
        stop
      endif

      !> register selectors
      obsname_padded = obsname // repeat(' ',10)  ! padding for substring operation
      !>-----
      if ( to_upper_str(obsname_padded(1:8)) == "PHOTONS_" ) then
      !>-----
        print*, "registered a photon selector: ", obsname
        if ( ORgroup /= 0 ) then
          print*, "photons_... specifiers in OR groups not possible!"
          stop
        end if
        if (qmin) then
          if (qmax) then
            call registerSelector_isol(obsname,lower=dmin,upper=dmax,invert=qinv)
          else
            call registerSelector_isol(obsname,lower=dmin,invert=qinv)
          endif
        else
          if (qmax) then
            call registerSelector_isol(obsname,upper=dmax,invert=qinv)
          endif
        endif
      !>-----
      else if ( to_upper_str(obsname_padded(1:5)) == "JETS_" ) then
      !>-----
        print*, "registered a jet selector: ", obsname
        if ( ORgroup /= 0 ) then
          print*, "jets_... specifiers in OR groups not possible!"
          stop
        end if
        if (qmin) then
          if (qmax) then
            call setSelector_jet(obsname,lower=dmin,upper=dmax,invert=qinv)
          else
            call setSelector_jet(obsname,lower=dmin,invert=qinv)
          endif
        else
          if (qmax) then
            call setSelector_jet(obsname,upper=dmax,invert=qinv)
          endif
        endif
      !>-----
      else  ! "normal" selector
      !>-----
        if (qmin) then
          if (qmax) then
            call register_sel(obsname,lower=dmin,upper=dmax,invert=qinv,valid=qval,ORgroup=ORgroup)
          else
            call register_sel(obsname,lower=dmin,invert=qinv,valid=qval,ORgroup=ORgroup)
          endif
        else
          if (qmax) then
            call register_sel(obsname,upper=dmax,invert=qinv,valid=qval,ORgroup=ORgroup)
          endif
        endif
      !>-----
      end if
      !>-----

    enddo

    !> check if all OR blocks have benn closed
    if (ORgroup /= 0) then
      print*, "unclosed OR block detected!"
      stop
    end if

    !> update smin now that all selectors are parsed
    call set_smin_run(min_shat_sel())
    !> attention: not ready yet: do it in `init_isol`
    !DO NOT DO: call set_smin_run(min_shat_isol())

  end subroutine parse_sel_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse the histograms in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_hist_io(unit)
    use Histograms_mod
    use Selectors_mod
    use Observables_mod
    use Reweight_mod
    use ScaleCoeff_mod
    use Run_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier, opt
    character(:), allocatable :: obsstring,obsname,fname, binstring
    double precision :: dmin, dmax
    integer :: stat, ncols, i,j,k, nbins
    double precision, dimension(:), allocatable :: binarray
    logical :: qfail
    !> selectors
    integer :: selOpt
    type(Selector_t), pointer :: sel_p!, newSel
    ! character(:), allocatable :: line_sel, spec_sel, obsname_sel
    ! double precision :: dmin_sel, dmax_sel
    ! logical :: qmin,qmax,qinv, qfail
    !> optional grid specification
    character(:), allocatable :: gridname
    integer :: current_ihist
    !> run group
    integer :: igroup
    !> reweighting function
    character(:), allocatable :: reweight, obsname_rewgt
    double precision :: fac_rewgt, pow_rewgt, off_rewgt
    logical :: qfail_rewgt
    type(ReweightFunction_t) :: rewgtfcn
    !> scale override
    double precision :: fac_mu0
    integer :: obsId_mu0
    logical :: qfail_mu0
    !> ouput level & cumulant option & type id
    integer :: outLevel, icumulant, ibinning
    integer :: itype1,itype2

    if (.not.qproductionSweep_run) then
      print*, "parse_hist_io: only warmup run => skipping HISTOGRAMS parsing..."
      return
    end if

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a `HISTOGRAMS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines

      opt = ''
      i = index(line,">")
      if (i.gt.0) then
        specifier = line(:i-1)
        opt = trim(adjustl(line(i+1:)))
      else
        specifier = line
      endif

      if (to_upper_str(specifier).eq.'HISTOGRAMS') exit
    end do

    !> init to no sub-histogram
    current_ihist = -1

    if (opt /= '') then
      call init_hist(opt)
    else
      call init_hist()
    end if

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_HISTOGRAMS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_HISTOGRAMS') exit
      !> end of a composite block
      if (to_upper_str(line).eq.'END_COMPOSITE') then
        current_ihist = -1
        cycle
      end if

      !> extract the optional reweighting specification
      rewgtfcn%qactive = .false.  ! init to an invalid state
      call extract_option_io(line, "fac", qfail, reweight)
      if (.not.qfail) then
        call readstring_rewgt_io( reweight, obsname_rewgt, &
                                & fac_rewgt, pow_rewgt, off_rewgt, qfail_rewgt )
        if (.not. qfail_rewgt) then
          call set_rewgt( obsname_rewgt, fac=fac_rewgt, pow=pow_rewgt, off=off_rewgt, &
                        & rewgtfcn=rewgtfcn )
        end if
      end if

      !> extract the optional scale override
      call extract_option_io(line, "mu0", qfail, opt)
      if (.not.qfail) then
        call readstring_scl_io(opt, fac_mu0, obsId_mu0, qfail_mu0)
      end if
      if ( qfail .or. qfail_mu0 ) then
        !> default values indicating "unset" state
        fac_mu0   = 0d0
        obsId_mu0 = -1
      else
        print*, "found scale setting:"
        print*, "fac_mu0  ", fac_mu0
        print*, "obsId_mu0", obsId_mu0
        !read(*,*)
        if (.not.isActive_sclc()) then
          print*, "parse_hist_io: scale override only availabe when scale-coefficients are on!"
          stop
        end if
      end if

      !> extract the optional `grid = ...` specification
      call extract_option_io(line, "grid", qfail, gridname)
      if (qfail) then
        gridname = ''
      else
        if (.not.isActive_sclc()) then
          print*, "parse_hist_io: grid filling needs scale-coefficients to be on!"
          !> can't do this because init_scl must be called *before* parsing histograms
          !> not ok:  call setActive_sclc(.true.)
          stop
          !call setActive_sclc(.true.)
        end if
      end if

      !> extract the optional `output_type = ...` specification
      call extract_option_type_io(line, "output_type", qfail, valueINT=outLevel)
      if (qfail) outLevel = 0  ! default: no channel breakdown
      ! if (qfail) outLevel = 2  ! default: legacy mode
      if (outLevel < -66 .or. outLevel > +2) then
        print*, "parse_hist_io: invalid output_type:", outLevel
        stop
      end if
      if (outLevel < -10 .and. outLevel > -67) then
        !> additional check in the 2 flavours case
        itype2 = abs( mod(outLevel, 10) )
        itype1 = abs( (outLevel+itype2) / 10 )
        if ( itype1 < 1 .or. itype1 > 6 .or. itype2 < 1 .or. itype2 > 6 ) then
          print*, "parse_hist_io: invalid output_type:", outLevel
          print*, "deduced itypes are invalid:", itype1, itype2
          stop
        end if
      end if

      !> extract the optional `cumulant = ...` specification
      call extract_option_type_io(line, "cumulant", qfail, valueINT=icumulant)
      if (qfail) icumulant = 0  ! default: no cumulant distribution
      if (icumulant < -1 .or. icumulant > +1) then
        print*, "parse_hist_io: invalid cumulant flag:", icumulant
        stop
      end if

      !> extract optional `run_group = ...` specification
      call extract_option_type_io(line, "run_group", qfail, valueINT=igroup)
      if (qfail) igroup = 0  ! default

      !> histogram parsing
      ncols = ncols_io(line)

      obsstring = ''
      obsname = ''
      fname = ''
      opt = ''
      nbins = 0
      dmin = 0d0
      dmax = 0d0
      if (allocated(binarray)) deallocate(binarray)

      !> parse optional selector block
      selOpt = 0  ! default value
      call parse_sel_block_io(unit, 'HISTOGRAM_SELECTORS', sel_p, selOpt)

      !> get obsname
      i = index(line," ")
      if (i>0) then
        obsstring = line(:i-1)
        line = trim(adjustl( line(i+1:) ))
      else
        obsstring = line
        line = ''
      end if
      i = index(obsstring,">")
      if (i.gt.0) then
        obsname = obsstring(:i-1)
        fname   = obsstring(i+1:)
      else
        obsname = obsstring
      endif

      !----- equal-size bins
      if ( current_ihist<0 .and. ncols>=4 ) then
        !> this guy is an optional flag to switch between linear/logarithmic binning
        call extract_option_io(line, "binning", qfail, opt)
        ibinning = 0  ! default: linear
        if (.not.qfail) then
          select case (to_upper_str(opt))
          case ("LIN","LINEAR")
            ibinning = 0
          case ("LOG","LOGARITHMIC")
            ibinning = 1
          case default
            stop "parse_hist_io: unknown `binning` specification"
          end select
        end if
        !> get mandatory options
        call extract_option_type_io(line, "nbins", qfail, valueINT=nbins)
        if (qfail) stop "parse_hist_io: missing/invalid `bins = ...` option"
        call extract_option_type_io(line, "min", qfail, valueDBLE=dmin)
        if (qfail) stop "parse_hist_io: missing/invalid `min = ...` option"
        call extract_option_type_io(line, "max", qfail, valueDBLE=dmax)
        if (qfail) stop "parse_hist_io: missing/invalid `max = ...` option"
        call register_hist( obsname, nbins,dmin,dmax, &
                          & binning=ibinning, &
                          & name=fname, &
                          & selectorsOption=selOpt, &
                          & selectors_p=sel_p, &
                          & gridname=gridname, &
                          & outputType=outLevel, &
                          & cumulant=icumulant, &
                          & run_group=igroup, &
                          & rewgtfcn=rewgtfcn, &
                          & fac_mu0=fac_mu0, &
                          & obsId_mu0=obsId_mu0, &
                          & ihist=current_ihist )

      !----- variable-size bins
      else if ( current_ihist<0 .and. ncols==2 ) then
        binstring = line(2:len(line)-1) ! trim away brackets
        do
          i = index(binstring, ",")
          if (i.lt.1) exit
          binstring(i:i) = " "
        end do
        nbins = ncols_io(binstring)
        allocate(binarray(nbins))
        read(binstring,*) binarray
        call register_hist( obsname, binarray, &
                          & name=fname, &
                          & selectorsOption=selOpt, &
                          & selectors_p=sel_p, &
                          & gridname=gridname, &
                          & outputType=outLevel, &
                          & cumulant=icumulant, &
                          & run_group=igroup, &
                          & rewgtfcn=rewgtfcn, &
                          & fac_mu0=fac_mu0, &
                          & obsId_mu0=obsId_mu0, &
                          & ihist=current_ihist )
        deallocate(binarray)

      !----- cross or composite-cross
      else if ( current_ihist<0 .and. ncols==1 .and.  &
              & ( to_upper_str(obsname).eq."CROSS" .or.  &
              &   to_upper_str(obsname).eq."COMPOSITE" ) ) then
        allocate(binarray(1))
        call register_hist( obsname, binarray, &
                          & name=fname, &
                          & selectorsOption=selOpt, &
                          & selectors_p=sel_p, &
                          & gridname=gridname, &
                          & outputType=outLevel, &
                          & run_group=igroup, &
                          & rewgtfcn=rewgtfcn, &
                          & fac_mu0=fac_mu0, &
                          & obsId_mu0=obsId_mu0, &
                          & ihist=current_ihist )
        deallocate(binarray)

      !----- sub-histogram
      else if ( current_ihist>0 .and. ncols==1 ) then
        call registerSub_hist( current_ihist, obsname, &
                             & selectorsOption=selOpt, &
                             & selectors_p=sel_p, &
                             & run_group=igroup, &
                             & rewgtfcn=rewgtfcn, &
                             & fac_mu0=fac_mu0, &
                             & obsId_mu0=obsId_mu0)

      !----- invalid input
      else
        print*, "invalid histogram specification: ", line
        print*, "current_ihist = ", current_ihist
        print*, "ncols = ", ncols
        stop
      endif

    end do

  end subroutine parse_hist_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse the scales in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_scl_io(unit)
    use Scales_mod
    use ScaleCoeff_mod
    use Observables_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, opt, subopt, valSTR, pref, scale_str
    integer :: stat, ncols, nsub, i,j,k
    double precision, target  :: facR,facF
    double precision, pointer :: fac_p
    integer, target  :: iobsR,iobsF
    integer, pointer :: iobs_p
    logical :: qfail, qfound
    !> for scale arrays
    character(2), parameter :: pattern = " ["
    integer :: nsubR, nsubF
    type(Scale_t), allocatable, dimension(:), target :: arrayR, arrayF
    type(Scale_t), pointer, dimension(:) :: arrayScl_p

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a `SCALES` block!"
        stop
      end if
      if (to_upper_str(line).eq.'SCALES') exit
    end do

    !> init scale-array variables
    nsubR = -1
    nsubF = -1

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_SCALES` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_SCALES') exit

      !> kill all blanks before "[" used for specifying scale arrays
      ! pattern = " ["
      do
        i = index(line, pattern)
        if ( i.eq.0 ) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      ncols = ncols_io(line)
      if ( ncols.ne.2 ) then
        print*, "invalid scale specification: ", line
        stop
      end if

      facR = 1d0
      facF = 1d0
      iobsR = -1  ! flag < 0 for un-initialised option: mu[F|R]
      iobsF = -1

      ! !>--- @todo: prettier way to do the thing below...
      ! call extract_option_io(line, "mur", qfail, opt)
      ! if (qfail) then
      !   print*, "parse_scl_io: couldn't fine mur = ... specification: ", line
      !   stop
      ! end if
      ! print*, "mur >", opt, "<"
      ! call extract_suboptions_io(opt, subopt)
      ! print*, "opt >", opt, "<"
      ! print*, "subopt >", subopt, "<", ncols_io(subopt)
      ! nsub = 1
      ! stop


      !> get options
      do k = 1,ncols
        !>----- get the specification
        i = index(line," ")
        if (i.gt.0) then
          opt = line(:i-1)
          line = trim(adjustl( line(i+1:) ))
        else
          opt = line
          line = ''
        endif
        !>----- extract the sub-optification if we have one
        subopt = ''
        nsub   = 1
        i = index(opt, "[")
        if (i/=0) then
          j = index(opt, "]")
          if (j==0) then
            print*, "parse_scl_io: couldn't find closing bracket"
            stop
          end if
          subopt = opt(i+1:j-1)
          opt    = trim(adjustl( adjustr(opt(:i-1)) // opt(j+1:) ))
          !> convert commas to spaces and count # of sub-scales
          do
            i = index(subopt, ",")
            if (i.lt.1) exit
            subopt(i:i) = " "
          end do
          nsub = ncols_io(subopt)
        end if
        ! print*, "scl: pre  >", opt, "<"
        ! print*, "scl: sub  >", subopt, "<", nsub
        !>----- parse the scale
        qfound =.false.
        nullify(arrayScl_p)
        pref = ""
        !> muR
        call extract_option_io(opt, "mur", qfail, valSTR)
        if (.not.qfail) then
          if (nsubR < 0) then  ! first time
            nsubR = nsub
            !> @todo check agains nsub in Process
            !> nsub >= nsub_proc to accomodate *at least* as many scales as sub-procs!!
            allocate( arrayR(nsubR) )
          end if
          if (nsub /= nsubR) then
            print*, "parse_scl_io: # mismatch in the muR scale array", nsub, nsubR
            stop
          end if
          pref = valSTR
          arrayScl_p(1:nsubR) => arrayR(1:nsubR)
        end if
        !> muF
        call extract_option_io(opt, "muf", qfail, valSTR)
        if (.not.qfail) then
          if (nsubF < 0) then  ! first time
            nsubF = nsub
            !> @todo check agains nsub in Process
            !> nsub >= nsub_proc to accomodate *at least* as many scales as sub-procs!!
            allocate( arrayF(nsubF) )
          end if
          if (nsub /= nsubF) then
            print*, "parse_scl_io: # mismatch in the muF scale array", nsub, nsubF
            stop
          end if
          pref = valSTR
          arrayScl_p(1:nsubF) => arrayF(1:nsubF)
        end if
        !> array set up?
        if ( .not. associated(arrayScl_p) ) then
            print*, "parse_scl_io: error parsing scale, unknown option"
            stop
        end if
        !> set up the corresponding array
        do i=1,nsub
          j = index(subopt," ")
          if (j.gt.0) then
            scale_str = pref // subopt(:j-1)
            subopt = trim(adjustl( subopt(j+1:) ))
          else
            scale_str = pref // subopt
            subopt = ''
          endif
          call readstring_scl_io(scale_str, arrayScl_p(i)%fac, arrayScl_p(i)%obsId, qfail)
          if (qfail) then
            print*, "parse_scl_io: error parsing scale"
            stop
          else
            qfound = .true.
          end if
        end do  ! do i [nsub]
        !>----- success?
        if (.not. qfound) then
          print*, "parse_scl_io: unknown option ", opt
          stop
        end if
      enddo  ! end loop k [columns: muR & muF]

      !> done parsing line in SCALES block: register
      !> (init will be called inside registerPair_scl)
      call registerPair_scl(arrayR, arrayF)

    enddo  ! end loop line

    if (nsubR<0.or.nsubF<0) then
      stop "FATAL ERROR: factorization or renormalisation scale unspecified"
    end if

    deallocate(arrayR)
    deallocate(arrayF)

    call init_scl()

  end subroutine parse_scl_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse a string for a scale
  !
  !> @param[in] string containing a scale
  !> @param[out] fac prefactor or the value of the constant scale
  !> @param[out] obsId integer Id of the observable
  !> @param[out] qfail signal success of parsing the string
  !-----------------------------------------------------------------------------
  subroutine readstring_scl_io(string, fac, obsId, qfail)
    use Observables_mod, only : getIdFromName_obs, needsFlav_obs
    character(:), allocatable, intent(in) :: string
    double precision, intent(out) :: fac
    integer, intent(out) :: obsId
    logical, intent(out) :: qfail
    character(:), allocatable :: obsname
    integer :: ncols, i
    double precision :: readDBLE
    logical :: qfail_str

    !> init values
    fac     = 1d0
    obsId   = -1
    qfail   = .false.
    obsname = ''

    ncols = ncols_io(string)
    if ( ncols.ne.1 ) then
      qfail = .true.
      return
    endif

    !> product?
    i = index(string,'*')
    if ( i > 0 ) then  ! A*B
      !> get "A"
      readDBLE = to_dble_str(string(:i-1), qfail_str)
      if (qfail_str) then
        obsname = trim(adjustl( string(:i-1) ))
        obsId   = getIdFromName_obs(obsname)
        if (obsId==0) then
          print*, "readstring_scl_io: invalid observable: ", obsname
          stop
        end if
        ! if (needsflav_obs(obsid)) then
        !   print*, "no flavour-sensitive observables allowed in the scales for now", obsid
        !   stop
        ! end if
      else
        fac = fac * readDBLE
        if (obsId < 0) obsId = 0
      end if
      !> get "B"
      readDBLE = to_dble_str(string(i+1:), qfail_str)
      if (qfail_str) then
        if (obsId > 0) then
          print*, "readstring_scl_io: product of two observables as scale not supported"
          stop
        end if
        obsname = trim(adjustl( string(i+1:) ))
        obsId   = getIdFromName_obs(obsname)
        if (obsId==0) then
          print*, "readstring_scl_io: invalid observable: ", obsname
          stop
        end if
        ! if (needsflav_obs(obsid)) then
        !   print*, "no flavour-sensitive observables allowed in the scales for now", obsid
        !   stop
        ! end if
      else
        fac = fac * readDBLE
        if (obsId < 0) obsId = 0
      end if
    else  ! A
      readDBLE = to_dble_str(string, qfail_str)
      if (qfail_str) then
        obsname = trim(adjustl( string ))
        obsId   = getIdFromName_obs(obsname)
        if (obsId==0) then
          print*, "readstring_scl_io: invalid observable: ", obsname
          stop
        end if
        ! if (needsflav_obs(obsid)) then
        !   print*, "no flavour-sensitive observables allowed in the scales for now", obsid
        !   stop
        ! end if
      else
        fac = fac * readDBLE
        if (obsId < 0) obsId = 0
      end if
    end if

  end subroutine readstring_scl_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse a string for a reweighting function
  !
  !> @param[inout] string containing a reweighting function
  !> @param[out] obsname name of the observable
  !> @param[out] fac prefactor
  !> @param[out] pow power
  !> @param[out] off offset
  !> @param[out] qfail signal success of parsing the string
  !-----------------------------------------------------------------------------
  subroutine readstring_rewgt_io(string, obsname, fac, pow, off, qfail)
    character(:), allocatable, intent(inout) :: string
    character(:), allocatable, intent(out) :: obsname
    double precision, intent(out) :: fac, pow, off
    logical, intent(out) :: qfail
    integer :: ncols, i,j,k, ierr, ierr2
    double precision :: readDBLE, readDBLE2

    !> init return values
    obsname = ''
    fac = 1d0
    pow = 1d0
    off = 0d0
    qfail = .false.

    ncols = ncols_io(string)
    if ( ncols.ne.1 ) then
      print*, "readstring_rewgt_io: invalid spaces in string: ", string
      qfail = .true.
      return
    endif

    ! !> parse reweighting function
    ! print*, "parsing a reweighting string: ", string

    !>----- find the offset
    i = index(string,"+")
    if (i > 0) then
      ! left:   string(:i-1)
      ! right:  string(i+1:)
      read(string(:i-1),*,iostat=ierr)  readDBLE
      read(string(i+1:),*,iostat=ierr2) readDBLE2
      !> both not real
      if ( ierr/=0 .and. ierr2/=0 ) then
        print*, "readstring_rewgt_io: only constant offsets allowed: ", string
        qfail = .true.
        return
      end if
      !> both real
      if ( ierr==0 .and. ierr2==0 ) then
        off = readDBLE + readDBLE2
        return
      end if
      !> now only one of the two is a real
      if ( ierr==0 ) then
        ! left: offset
        off = readDBLE
        string = trim(adjustl( string(i+1:) ))
      else
        ! right: offset
        off = readDBLE2
        string = trim(adjustl( string(:i-1) ))
      end if
    else
      ! let's try parsing it as a constant
      read(string,*,iostat=ierr) readDBLE
      !> it's just a real
      if ( ierr==0 ) then
        off = readDBLE
        return
      end if
    endif

    !>----- find the power
    i = index(string,"**")
    if (i > 0) then
      !> we have a power: look for prefactors
      j = index(string(:i-1),"*")
      k = index(string(i+2:),"*")
      !> factors on both sides
      if ( j>0 .and. k>0 ) then
        print*, "readstring_rewgt_io: only one prefactor allowed: ", string
        qfail = .true.
        return
      end if
      !> factor to the left
      if ( j > 0 ) then
        ! print*, "looking in >", string(:i-1), "<"
        ! print*, "left factor >", string(:j-1), "<"
        ! print*, "remainder >", string(j+1:), "<"
        read(string(:j-1),*,iostat=ierr)  fac
        if ( ierr/=0 ) then
          print*, "readstring_rewgt_io: prefactor must be a constant: ", string
          qfail = .true.
          return
        end if
        string = trim(adjustl( string(j+1:) ))
      end if
      !> factor to the right
      if ( k > 0 ) then
        ! print*, "looking in >", string(i+2:), "<"
        ! print*, "right factor >", string(i+2+k:), "<"
        ! print*, "remainder >", string(:i+k), "<"
        read(string(i+2+k:),*,iostat=ierr)  fac
        if ( ierr/=0 ) then
          print*, "readstring_rewgt_io: prefactor must be a constant: ", string
          qfail = .true.
          return
        end if
        string = trim(adjustl( string(:i+k) ))
      end if
      !> now the rest is "{obs}**pow"
      j = index(string,"**")
      if (j==0) stop "readstring_rewgt_io: shouldn't get here..."
      !> get the power
      read(string(j+2:),*,iostat=ierr) pow
      if ( ierr/=0 ) then
        print*, "readstring_rewgt_io: the exponent must be a constant: ", string
        qfail = .true.
        return
      end if
      string = trim(adjustl( string(:j-1) ))  ! only "{obs now}"
      !> is {obs} a real value?
      read(string,*,iostat=ierr) readDBLE
      if (ierr==0) then
        !> everything was reals, accumulate to the offset:
        off = off + fac * readDBLE ** pow  ! works because init: pow=1
        return
      else
        !> it's an observable
        obsname = string
      end if

    else
      !> there is no power: the string is "fac*{obs}" or "{obs}*fac" or only "{obs}"
      j = index(string,"*")
      if (j>0) then
        read(string(:j-1),*,iostat=ierr)  readDBLE
        read(string(j+1:),*,iostat=ierr2) readDBLE2
        !> both not real
        if ( ierr/=0 .and. ierr2/=0 ) then
          print*, "readstring_rewgt_io: prefactor must be a constant: ", string
          qfail = .true.
          return
        end if
        !> everything was reals, accumulate to the offset:
        if ( ierr==0 .and. ierr2==0 ) then
          off = off + readDBLE * readDBLE2
          return
        end if
        !> now only one of the two is a real
        if ( ierr==0 ) then
          ! left: prefactor
          fac = readDBLE
          string = trim(adjustl( string(j+1:) ))
        else
          ! right: prefactor
          fac = readDBLE2
          string = trim(adjustl( string(:j-1) ))
        end if
      end if
      !> is {obs} a real value?
      read(string,*,iostat=ierr) readDBLE
      if (ierr==0) then
        !> everything was reals, accumulate to the offset:
        off = off + fac * readDBLE ** pow  ! works because init: pow=1
        return
      else
        !> it's an observable
        obsname = string
      end if
    endif

  end subroutine readstring_rewgt_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for an optional re-weigthing function in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_rewgt_io(unit)
    use Reweight_mod, only: set_rewgt
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier,obsname
    integer :: stat, ncols
    integer :: i
    double precision :: fac, pow, off
    logical :: qfail
    !> optional settings
    integer :: intValue
    logical :: boolValue


    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) return
      if (line.eq.'') cycle ! skip empty lines
      ncols = ncols_io(line)
      !> we now allow for optional arguments
      !if (ncols.ne.2) cycle ! not correct match

      !> look for the specifier
      i = index(line," ")
      specifier = line(:i-1)
      line = trim(adjustl( line(i+1:) ))
      if (to_upper_str(specifier).eq."REWEIGHT") then

        call readstring_rewgt_io(line, obsname, fac, pow, off, qfail)
        if (.not. qfail) then
          call set_rewgt(obsname, fac=fac, pow=pow, off=off)
          !> check for optional arguments depending on the "observable"
          select case (to_upper_str(obsname))
          !>
          case default
            ! print*, "parse_rewgt_io: skip optional argument for: ", obsname
          end select
        end if
      else
        cycle
      end if

    end do

  end subroutine parse_rewgt_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for an optional PDF interpolation setting in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_pdfi_io(unit)
    use PDFInterp_mod, only : setActive_pdfi, init_pdfi
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier
    integer :: stat, ncols
    integer :: i
    logical :: qfail
    !> optional settings
    double precision :: dbleValue

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) return
      if (line.eq.'') cycle ! skip empty lines
      ncols = ncols_io(line)

      if(ncols.gt.1)then
         i = index(line," ")
         specifier = to_upper_str(line(:i-1))
         line = trim(adjustl(line(i+1:)))
      else
         specifier = to_upper_str(trim(adjustl(line)))
      endif

      if (specifier.eq."PDF_INTERPOLATION") then

        call setActive_pdfi(.true.)

        !> check for optional argument
        call extract_option_type_io(line, "stepfac", qfail, valueDBLE=dbleValue)
        if (.not.qfail) then
          call init_pdfi(stepfac=dbleValue)
        else
          call init_pdfi()
        endif

      else
        cycle
      end if

    end do

  end subroutine parse_pdfi_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for an optional unit phase space volume flag in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_phase_unit_io(unit)
    use UnitPhase_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier,opt
    integer :: stat,ncols
    integer :: i,k

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) return
      if (line.eq.'') cycle ! skip empty lines

      ncols = ncols_io(line)

      if(ncols.gt.1)then
         i = index(line," ")
         specifier = to_upper_str(line(:i-1))
         line = trim(adjustl(line(i+1:)))
      else
         specifier = to_upper_str(trim(adjustl(line)))
      endif

      if (specifier.eq."UNIT_PHASE") then
         phase_unit = .true.
         if(ncols.gt.1)then
            do k=2,ncols
               i = index(line," ")
               if (i.gt.0) then
                  opt = to_upper_str(line(:i-1))
                  line = trim(adjustl(line(i+1:)))
               else
                  opt = to_upper_str(line)
                  line = ''
               endif
               select case (opt)
               case ("NOCUTS")
                  nocuts = .true.
               case ("FIXS")
                  fixS = .true.
               end select
            end do
         end if
         return
      else
        cycle
      end if
    end do

  end subroutine parse_phase_unit_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse for an optional collider energy in readruncard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_coll_io(unit)
    use Process_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, specifier, opt
    integer :: stat, ncols, i,j,k
    logical :: qfail
    character(:), allocatable :: beam1, beam2
    double precision :: sqrts, E1, E2

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) return
      if (line.eq.'') cycle ! skip empty lines

      !> look for specifier
      i = index(line," ")
      if (i.eq.0) cycle ! something has to come after the specifier
      specifier = line(:i-1)
      line = trim(adjustl( line(i+1:) ))
      if (to_upper_str(specifier).eq."COLLIDER") then

        !> init beam names
        beam1 = ''
        beam2 = ''
        !> get collider name
        i = index(line," ")
        opt = line(:i-1)
        select case (to_upper_str(opt))
          case ('LHC')
            print*, "LHC..."
            beam1 = 'p'
            beam2 = 'p'
          case ('TEVATRON')
            beam1 = 'p'
            beam2 = 'ap'
          case ('EP')
            beam1 = 'em'
            beam2 = 'p'
          case ('EPEM')
            beam1 = 'ep'
            beam2 = 'em'
          case ('LHCASY')!LHC experiment with asymmetric bean like p-Pb collision
            beam1 = 'p'
            beam2 = 'ions'
          case default
            beam1 = ''
            beam2 = ''
        end select
        !> successfully parsed a collider?
        !> if not, look for beam[i] = ... options
        if ( beam1 == '' .and. beam2 == '' ) then
          call extract_option_io(line, "beam1", qfail, beam1)
          if (qfail) stop "parse_coll_io: error reading beam1 option"
          call extract_option_io(line, "beam2", qfail, beam2)
          if (qfail) stop "parse_coll_io: error reading beam2 option"
        else if ( (beam1 == 'ions') .or. (beam2 == 'ions') ) then
          call extract_option_io(line, "beam1", qfail, beam1)
          if (qfail) stop "parse_coll_io: error reading beam1 ion type"
          call extract_option_io(line, "beam2", qfail, beam2)
          if (qfail) stop "parse_coll_io: error reading beam2 ion type"
        else
          !> eliminate the collider name
          line = trim(adjustl( line(i+1:) ))
        end if
        !> done getting beam information: set in the Process module
        call setBeams_proc(beam1, beam2)

        !> # cols after stripping specifier & collider / beams
        ncols = ncols_io(line)

        if (ncols == 1) then
          !> sqrts = ...
          call extract_option_type_io(line, "sqrts", qfail, valueDBLE=sqrts)
          if (qfail) stop "parse_coll_io: error reading sqrts"
          call setSqrts_proc(sqrts)

        else if (ncols == 2) then
          !> E1 = ... E2 = ...
          call extract_option_type_io(line, "E1", qfail, valueDBLE=E1)
          if (qfail) stop "parse_coll_io: error reading E1"
          call extract_option_type_io(line, "E2", qfail, valueDBLE=E2)
          if (qfail) stop "parse_coll_io: error reading E2"
          call setEnergies_proc(E1, E2)

        else
          stop "parse_coll_io: invalid COLLIDER specification"
        end if

      else
        cycle
      end if

    end do

  end subroutine parse_coll_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse the selectors in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_chan_io(unit)
    use Channels_mod, only : enableChannel_chan, set_default_region_chan
    integer, intent(in) :: unit
    character(:), allocatable :: line, region
    integer :: stat, ncols, i,j
    character(2) :: pattern
    logical :: qfail

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find a `CHANNELS` block!"
        stop
      end if

      !> look for an optional region=[a|b|all] fall-back override
      call extract_option_io(line, "region", qfail, region)
      if (.not.qfail) then
        call set_default_region_chan(region)
      end if
      if (to_upper_str(line).eq.'CHANNELS') exit
    end do

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_CHANNELS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_CHANNELS') exit

      !> kill all blanks before "[" that can be used to specify a contribution
      pattern = " ["
      do
        i = index(line, pattern)
        if ( i.eq.0 ) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      !> register all channels (multiple entries per line allowed)
      ncols = ncols_io(line)
      do j = 1,ncols
        i = index(line," ")
        if (i.gt.0) then
          call enableChannel_chan(line(:i-1))
          line = trim(adjustl( line(i+1:) ))
        else
          call enableChannel_chan(line)
          line = ''
        endif
      enddo

    enddo

  end subroutine parse_chan_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse an optional multi-run block in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_mult_io(unit)
    use MultiRun_mod
    use Reweight_mod
    use Selectors_mod
    use PhotonIsolation_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line, opt, subopt
    integer :: stat, ncols, i,j,k
    logical :: qfail
    !> run setting
    integer :: current_irun, igroup
    !> reweighting factor
    character(:), allocatable :: reweight, obsname_rewgt, obsname_scl
    double precision :: fac_rewgt, pow_rewgt, off_rewgt
    logical :: qfail_rewgt
    type(ReweightFunction_t) :: rewgtfcn
    !> selectors
    integer :: selOpt
    type(Selector_t), pointer :: sel_p
    !> isolation
    character(:), allocatable :: isol_options
    type(Isol_t), pointer :: isol_p
    !> value types
    integer :: intValue
    double precision :: dbleValue
    logical :: boolValue

  rewind(unit)

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) return
      if (to_upper_str(line).eq.'MULTI_RUN') exit
    end do

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_MULTI_RUN` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_MULTI_RUN') exit

      !> extract the reweighting factor
      rewgtfcn%qactive = .false.  ! init to an invalid state
      call extract_option_io(line, "fac", qfail, reweight)
      if (qfail) then
        print*, "parse_mult_io: missing 'fac=...' specification"
        stop
      end if
      call readstring_rewgt_io( reweight, obsname_rewgt, &
                              & fac_rewgt, pow_rewgt, off_rewgt, qfail_rewgt )
      if (qfail_rewgt) then
        print*, "parse_mult_io: invalid 'fac=...' specification"
        stop
      end if
      call set_rewgt( obsname_rewgt, fac=fac_rewgt, pow=pow_rewgt, off=off_rewgt,  &
                    & rewgtfcn=rewgtfcn )

      !> get the reweighting group
      call extract_option_type_io(line, "run_group", qfail, valueINT=igroup)
      if (qfail) then
        igroup = 0  ! defualt group id
      end if

      !> parse optional selector block
      selOpt = 0  ! default value
      call parse_sel_block_io(unit, 'RUN_SELECTORS', sel_p, selOpt)

      !> parse optional isolation setting
      nullify(isol_p) ! reset (needed if there's no isolation specification)
      call extract_option_io(line, "photon_isolation", qfail, isol_options)
      if (.not.qfail) then
        call parse_isol_io(isol_options, isol_p=isol_p)
      end if

      !> create a new run object
      current_irun = register_mult( rewgtfcn, run_group=igroup,  &
                                  & selectorsOption=selOpt, selectors_p=sel_p,  &
                                  & isolation_p=isol_p )

      !> extract the optional settings for the run
      !----- RCONE
      call extract_option_type_io(line, "rcone", qfail, valueDBLE=dbleValue)
      if (.not.qfail) call addopt_mult(current_irun, "rcone", valueDBLE=dbleValue)
      !----- MUR0
      call extract_option_io(line, "mur0", qfail, opt)
      if (.not.qfail) then
        call readstring_scl_io(opt, dbleValue, intValue, qfail)
        if (.not.qfail) call addopt_mult(current_irun, "mur0", valueDBLE=dbleValue, valueINT=intValue)
      end if
      !----- MUF0
      call extract_option_io(line, "muf0", qfail, opt)
      if (.not.qfail) then
        call readstring_scl_io(opt, dbleValue, intValue, qfail)
        if (.not.qfail) call addopt_mult(current_irun, "muf0", valueDBLE=dbleValue, valueINT=intValue)
      end if
      !----- MU0 = MUR0 & MUF0
      call extract_option_io(line, "mu0", qfail, opt)
      if (.not.qfail) then
        call readstring_scl_io(opt, dbleValue, intValue, qfail)
        if (.not.qfail) then
          call addopt_mult(current_irun, "mur0", valueDBLE=dbleValue, valueINT=intValue)
          call addopt_mult(current_irun, "muf0", valueDBLE=dbleValue, valueINT=intValue)
        end if
      end if
      !----- P2B
      call extract_flag_io(line, "p2b", qfail, suboptions=subopt)
      if (.not.qfail) then
        !> P2B can have suboptions
        call extract_option_type_io(subopt, "qt_cut", qfail, valueDBLE=dbleValue)
        if (.not.qfail) then
          call addopt_mult(current_irun, "p2b", valueDBLE=dbleValue)
        else
          call addopt_mult(current_irun, "p2b")
        end if
      end if
      !----- any left-overs are invalid options:
      if (line /= '') then
        print*, "parse_mult_io: invalid options: ", line
        stop
      end if

    enddo

  end subroutine parse_mult_io


  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse the SM parameters and settings in the runcard (unit)
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_param_io(unit)
    use Parameters_mod
    integer, intent(in) :: unit
    character(:), allocatable :: line
    integer :: stat, ncols, ispec, i,j, iparam
    character(:), allocatable :: spec, subspec, subopt
    logical :: qfound, qfail
    integer :: valINT
    double precision :: valDBLE
    logical :: valBOOL
    character(:), allocatable :: valSTR, param
    character(2), parameter :: pattern = " ["

    rewind(unit)

    do
      call getline_io(unit, line, stat)
      !> don't make the parameters block mandatory just yet
      if (stat /= 0) return
      ! if (stat /= 0) then
      !   print*, "invalid runcard: can't find a `PARAMETERS` block!"
      !   stop
      ! end if
      if (to_upper_str(line).eq.'PARAMETERS') exit
    end do

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_PARAMETERS` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_PARAMETERS') exit

      !> kill all blanks before "[" that can be used to specify a sub-specification
      do
        i = index(line, pattern)
        if (i==0) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      !> loop over specs (multiple entries per line allowed)
      ncols = ncols_io(line)
      do ispec = 1,ncols
        !>----- get the specification
        i = index(line," ")
        if (i.gt.0) then
          spec = line(:i-1)
          line = trim(adjustl( line(i+1:) ))
        else
          spec = line
          line = ''
        endif
        !>----- extract the sub-specification if we have one
        subspec = ''
        i = index(spec, "[")
        if (i/=0) then
          j = index(spec, "]")
          if (j==0) then
            print*, "parse_param_io: couldn't find closing bracket"
            stop
          end if
          subspec = spec(i+1:j-1)
          spec    = trim(adjustl( adjustr(spec(:i-1)) // spec(j+1:) ))
        end if
        !>----- extract the sub-option if we have one
        subopt = ''
        i = index(subspec, ",")
        if (i/=0) then
          subopt  = trim(subspec(i+1:))
          subspec =trim(adjustl( adjustr(subspec(:i-1)) ))
        end if
        ! print*, "spec     >", spec, "<"
        ! print*, "subspec  >", subspec, "<"
        ! print*, "subopt   >", subopt, "<"
        !>----- parse the spec
        qfound =.false.
        !> masses
        call extract_option_type_io(spec, "mass", qfail, valueDBLE=valDBLE)
        if (.not.qfail) then
          if (subopt /= '') then
            call set_mass_param(subspec, valDBLE, scheme=subopt)
          else
            call set_mass_param(subspec, valDBLE)
          end if
          qfound = .true.
        end if
        !> widths
        call extract_option_type_io(spec, "width", qfail, valueDBLE=valDBLE)
        if (.not.qfail) then
          call set_width_param(subspec, valDBLE)
          qfound = .true.
        end if
        !> more parameters
        do iparam = 1,4
          select case(iparam)
            case(1)
              param = 'alpha'
            case(2)
              param = 'alpha0'
            case(3)
              param = 'alphaMZ'
            case(4)
              param = 'GF'
            case default
              stop "parse_param_io: wrong iparam bounds"
          end select
          call extract_option_type_io(spec, param, qfail, valueDBLE=valDBLE)
          if (.not.qfail) then
            call set_value_param(param, valDBLE)
            qfound = .true.
          end if
        end do
        !> boolean flags
        call extract_option_type_io(spec, "hard_photon_alpha0", qfail, valueBOOL=valBOOL)
        if (.not.qfail) then
          qalpha0_hard_photons = valBOOL
          qfound = .true.
          print*, "hard_photon_alpha0 = ", qalpha0_hard_photons
        end if
        !> schemes
        call extract_option_io(spec, "scheme", qfail, valSTR, suboptions=param)
        if (.not.qfail) then
          call set_scheme_param(subspec, valSTR, param)
          qfound = .true.
        end if
        call extract_option_io(spec, "ckm", qfail, valSTR)
        if (.not.qfail) then
           call set_ckm_param(valSTR)
           qfound = .true.
        end if
        !>----- success?
        if (.not. qfound) then
          print*, "parse_param_io: skipping unknown option ", spec
        end if

      enddo

    enddo

    ! stop "done parsing for PARAMETERS block"

  end subroutine parse_param_io


#ifdef USEFASTJET
  !-----------------------------------------------------------------------------
  !> @brief
  !> Parse FastJet options
  !
  !> @param[in] unit logical unit connected for formatted input from a file
  !-----------------------------------------------------------------------------
  subroutine parse_fastjet_io(unit)
    use FastJet_mod!, only : book_fj
    integer, intent(in) :: unit
    character(:), allocatable :: line
    integer :: stat, ncols, ispec, i,j, iparam
    character(:), allocatable :: spec, subopt
    character(:), allocatable :: abbr, algo
    !> parameters for ppgenkt
    real(kind=wp) :: palg, R, ptcut
    logical :: qfound, qfail
    integer :: valINT
    double precision :: valDBLE
    logical :: valBOOL
    character(:), allocatable :: valSTR, param
    character(2), parameter :: pattern = " ["


    rewind(unit)

    do
      call getline_io(unit, line, stat)
      !> don't make the fastjet block mandatory
      if (stat /= 0) return
      if (to_upper_str(line).eq.'FASTJET') exit
    end do

    do
      call getline_io(unit, line, stat)
      if (stat /= 0) then
        print*, "invalid runcard: can't find closing `END_FASTJET` block!"
        stop
      end if
      if (line.eq.'') cycle ! skip empty lines
      if (to_upper_str(line).eq.'END_FASTJET') exit

      !> kill all blanks before "[" that can be used to specify a contribution
      do
        i = index(line, pattern)
        if ( i.eq.0 ) exit
        line = trim(adjustl( adjustr(line(:i)) // line(i+1:) ))
      end do

      !> loop over specs (multiple entries per line allowed)
      ncols = ncols_io(line)
      do ispec = 1,ncols
        !>----- get the specification
        i = index(line," ")
        if (i.gt.0) then
          spec = line(:i-1)
          line = trim(adjustl( line(i+1:) ))
        else
          spec = line
          line = ''
        endif
        !>----- get the abbreviation
        i = index(spec, "=")
        if (i/=0) then
          abbr = trim(adjustl( adjustr(spec(:i-1)) ))
        else
          print*, "parse_fastjet_io: invalid specification", line
          stop
        end if
        !>----- extract specification
        call extract_option_io(spec, abbr, qfail, algo, suboptions=subopt)
        qfound =.false.
        if (.not.qfail) then
          !>@todo do a switch on the name of the algorithm
          !>      to extract different options and call booking
          call extract_option_type_io(subopt, "R", qfail, valueDBLE=valDBLE)
          if (.not.qfail) then
            R = valDBLE
          else
            print*, "parse_fastjet_io: missing R: ", line
            stop
          end if
          palg = 0.0
          ptcut = 0.0
          call extract_option_type_io(subopt, "palg", qfail, valueDBLE=valDBLE)
          if (.not.qfail) then
            palg = valDBLE
          end if
          call extract_option_type_io(subopt, "ptcut", qfail, valueDBLE=valDBLE)
          if (.not.qfail) then
            ptcut = valDBLE
          end if
          !> better have no left overs
          if ( subopt /= '' ) then
            print*, "parse_fastjet_io: left-over subopt: ", subopt
            stop
          end if
          call book_fj(abbr, algo, in_palg=palg, in_R=R, in_ptcut=ptcut)
          qfound =.true.
        end if
        !>----- success?
        if (.not. qfound) then
          print*, "parse_fastjet_io: skipping unknown option ", spec
        end if

      enddo

    enddo

  end subroutine parse_fastjet_io
#endif

end module IOHelper_mod
