!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Jets [external members: `_jet`]
!
! DESCRIPTION:
!> Module implementing jet reconstruction
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module Jets_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  use ObsHelper_mod
  use KinData_mod
  implicit none
  private

  !-------------------!
  !  Jet Observables  !
  !-------------------!

  public :: init_jet, init_dress_jet
  public :: algo_jet, setRcone_jet, getRcone_jet
  public :: setParam_jet, getIdFromName_jet
  public :: setSelector_jet, qpass_jet, qpass_mom_jet
  public :: digest_jet
  public :: min_shat_jet
  public :: initBuffer_jet, destroyBuffer_jet
  public :: initRecomb_jet
  public :: cluster_jet, cluster_jade_jet
  public :: cluster_flavjet, cluster_flav_dress_jet
  public :: digestActive_jet
  public :: recombine_jet, recombine_democratic_jet
  public :: jetalg, Rcone, RconeSq
  public :: algo_dress_jet

  !------------------------------!
  !  Jet-Observable identifiers  !
  !------------------------------!

  integer, public, parameter :: idj_pt = 1
  integer, public, parameter :: idj_y = 2
  integer, public, parameter :: idj_abs_y = 3
  integer, public, parameter :: idj_eta = 4
  integer, public, parameter :: idj_abs_eta = 5
  integer, public, parameter :: idj_et = 6
  integer, public, parameter :: idj_dR_l1 = 7
  integer, public, parameter :: idj_dR_l2 = 8
  integer, public, parameter :: idj_dR_l3 = 9
  integer, public, parameter :: idj_dR_l4 = 10
  integer, public, parameter :: idj_min_dR_l12 = 11
  integer, public, parameter :: idj_min_dR_l34 = 12
  integer, public, parameter :: idj_min_dR_l1234 = 13
  integer, public, parameter :: idj_dR_exp_l1 = 14
  integer, public, parameter :: idj_dR_exp_l2 = 15
  integer, public, parameter :: idj_dR_exp_l3 = 16
  integer, public, parameter :: idj_dR_exp_l4 = 17
  integer, public, parameter :: idj_min_dR_exp_l12 = 18
  integer, public, parameter :: idj_min_dR_exp_l34 = 19
  integer, public, parameter :: idj_min_dR_exp_l1234 = 20
  integer, public, parameter :: idj_min_dR_gam = 21
  integer, public, parameter :: idj_min_dR_exp_gam = 22
  !> DIS: Jet-Observable identifiers in HERA frame
  integer, public, parameter :: idj_hera_ptj = 23
  integer, public, parameter :: idj_hera_yj = 24
  integer, public, parameter :: idj_hera_abs_yj = 25
  integer, public, parameter :: idj_hera_etaj = 26
  integer, public, parameter :: idj_hera_abs_etaj = 27
  integer, public, parameter :: idj_hera_etj = 28
  integer, public, parameter :: idj_jets_xhera = 29
  integer, public, parameter :: idj_jets_H1xhera = 30
  !> TEEC: rapidity selector for partons
  integer, public, parameter :: idj_TEEC_y = 31
  integer, public, parameter :: idj_TEEC_abs_y = 32
  !> for democratic clustering: zgam = ET fraction of the photon
  integer, public, parameter :: idj_zgam = 33
  !> flavoured jet observables
  integer, public, parameter :: idfj_pt = 34
  integer, public, parameter :: idfj_y = 35
  integer, public, parameter :: idfj_abs_y = 36
  integer, public, parameter :: idfj_eta = 37
  integer, public, parameter :: idfj_abs_eta = 38
  integer, public, parameter :: idfj_et = 39

  !> number of jet-observables
  integer, public, parameter :: n_obsj = 39

  logical :: softdrop = .false.

  !> flag choosing the algorithm
  !>   0: none
  !>   1: anti-kt
  !>   2: C/A
  !>   3: kt
  !>   4: jade
  !>   5: flavour-kt
  !>   6: durham
  !>   7: flavour anti-kt ! from [2011.01011]
  integer :: jetalg = -1
  integer :: ip_jet ! power of the generalized kt algorithm
  !> option for a soft-inclusive jet to test IR sensitive observables
  !>   power of the ratio (max/min)^p that widens the cone (<= 0 <-> off)
  real(kind=wp) :: dyn_soft_jet = 0._wp
  !>-----
  !> flavour-kT settings
  real(8) :: flavkt_alpha = 2._wp
  !> include colourless particles in the beam distance
  logical :: flavkt_include_colourless = .true.
  !>  include them individually or sum them and use ET
  !>   0: sum them and then use ET
  !>   1: sum them and then use MT
  logical :: flavkt_sum_nonqcd = .true.
  !> include only projets in the beam distance
  logical :: flavkt_only_proto = .false.
  !>-----
  !> Flavour anti-kt 'a' parameter
  real(8) :: flavantikt_a = 0.1_wp
  !>-----
  !> R parameter of the algorithm set in the runcard
  real(kind=wp) :: inputRcone = -1._wp
  !> ycut in the jade/durham algorithm
  real(kind=wp) :: jet_ycut = -1._wp
  !> recombination scheme: ET?
  logical :: qEtrecomb = .false.
  !> threadlocal copies for R [& squared]
  !> typically will be set to inputRcone but multi-runs can
  !> have their own values
  real(kind=wp) :: Rcone = -1._wp
  real(kind=wp) :: RconeSq = -1._wp
!$OMP THREADPRIVATE(Rcone, RconeSq)

  !>-----
  !> flavour dressing options (absolute value as above)
  !>    0: naive dressing (potentially IR unsafe!)
  !>  (<0) => recursive soft drop
  !>   -2: [C/A de-clustering]
  !>  (>0) => flavour clustering
  !>    4: [jade] (pi.pj) (no beam distance) != flavour-kt(alpha=1)
  !>    5: [flavour-kt]
  !>    6: [durham] == flavour-kt(alpha=0)
  !>    7: revised flavour dressing (08/2023)
  integer :: iflav_dress = 0
  real(kind=wp) :: flav_dress_SD_zcut = 0.1_wp
  real(kind=wp) :: flav_dress_SD_beta = 2._wp
  ! apply dressing at the level of proto-jets or reco jets (after cuts)?
  logical :: flav_dress_cluster_qproto = .false.
  logical :: flav_dress_cluster_qbeam = .true.
  integer :: flav_dress_cluster_idress = 0  ! 0: democratic, 1: top->down
  !> include nonQCD particles in the beam?
  !>  0: no
  !>  1: sum them and use ET
  !>  2: include them individually
  integer :: flav_dress_cluster_beam_nonqcd = 0
  real(kind=wp) :: flav_dress_cluster_alpha = 2._wp
  ! flavour-jet distance "penalty"
  !  0: no penalty
  !  1: Wilson-like picture - beta in [0,1]
  !  2: multiplicative penalty - beta in [0,infty)
  integer :: flav_dress_cluster_ipenalty = 0
  real(kind=wp) :: flav_dress_cluster_beta = 1._wp
  ! clustering needs an under-dress
  logical :: flav_qunderdress = .true.
  real(kind=wp) :: flav_underdress_R = 0.1_wp
  real(kind=wp) :: flav_underdress_zcut = 0.1_wp
  real(kind=wp) :: flav_underdress_beta = 2._wp
  ! omega parameter introduced in 2306.07314
  logical :: flav_dress_cluster_use_omega = .false.
  real(kind=wp) :: flav_dress_cluster_omega = 2._wp
  !> type to collect information needed in the recombination
  type Jet_t
    !> flag to identify status
    !>   < 0: ProtoJet (subject to further recombinations)
    !>   = 0: Nothing (merged into another Protojet)
    !>   > 0: Jet (identified as a hard jet)
    integer :: isJet
    real(kind=wp), dimension(4) :: p
    real(kind=wp) :: pt2, y, phi
    !> properties only used when flavour-tagging is active
    logical :: isFlavoured
    integer :: nflav_q
    integer :: nflav_qb
  end type Jet_t

  type NonQCD_t
    real(8) :: p(4)
    real(8) :: pt2, y, m2
  end type NonQCD_t

  !> threadprivate buffer to minimize re-allocation.
  type(Jet_t), allocatable, dimension(:) :: jets, flavjets
  type(NonQCD_t) :: nonQCDsum
  type(NonQCD_t), allocatable, dimension(:) :: nonQCDvec
  integer :: nonQCDvecsize
  real(kind=wp), allocatable, dimension(:) :: diB  ! "B" = beams
  real(kind=wp), allocatable, dimension(:, :) :: dij
  !> information on currently stored data
  integer :: npar_jet, nproto_jet, njets_jet, njets_flavjet
  integer :: flav_njets_jet = -1, flav_njets_flavjet = -1
  logical :: qskip_init_jet = .false.
  !> store the clustering history
  !> if parton `i` is merged into `j`:
  !>  clusterHist_jet(i) = j
  !> final jet will have
  !>  clusterHist_jet(i) = i
  integer, allocatable, dimension(:) :: clusterHist_jet, clusterHist_flavjet
  !> ordering according to pt (descending)
  !>  leading jet: pt_sort_jet(1)
  !>  => ptj1 = jets(pt_sort_jet(1))%pt
  !>  etc...
  integer, allocatable, dimension(:) :: pt_sort_jet
!$OMP THREADPRIVATE(jets, flavjets, nonQCDsum, nonQCDvec, nonQCDvecsize)
!$OMP THREADPRIVATE(dij, diB)
!$OMP THREADPRIVATE(npar_jet, nproto_jet, njets_jet, njets_flavjet)
!$OMP THREADPRIVATE(flav_njets_jet, flav_njets_flavjet)
!$OMP THREADPRIVATE(qskip_init_jet)
!$OMP THREADPRIVATE(clusterHist_jet, clusterHist_flavjet, pt_sort_jet)

  type JetSelector_t
    logical :: qactive = .false.
    character(len=32) :: name
    real(kind=wp) :: lower = 0._wp
    real(kind=wp) :: upper = 0._wp
    logical :: qlower = .false.
    logical :: qupper = .false.
    logical :: qinvert = .false.
    logical :: qflavobs = .false.
  end type JetSelector_t

  type(JetSelector_t), dimension(n_obsj) :: list_obsj

  !> ycut transition observables
  real(kind=wp), public :: value_y01, value_y12, value_y23, value_y34, value_y45
  logical, public :: valid_y01, valid_y12, valid_y23, valid_y34, valid_y45
!$OMP THREADPRIVATE(value_y01, value_y12, value_y23, value_y34, value_y45)
!$OMP THREADPRIVATE(valid_y01, valid_y12, valid_y23, valid_y34, valid_y45)

contains

  !-----------------------------------------------------------------------------
  !> @brief
  !> determine if the associated jet observable is of `extra` type
  !> extra variables are only used to facilitate the infrastructure of
  !> registering new observables / selectors... evaluation of these observables
  !> is left to the user
  !
  !> @param[in] obsId unique identifier for the jet observabes
  !-----------------------------------------------------------------------------
  elemental function isExtraj_obs(obsId) result(is_extraj)
    integer, intent(in) :: obsId
    logical :: is_extraj
    select case (obsId)
      !> DIS
    case (idj_hera_abs_etaj, idj_hera_etj, idj_hera_ptj, idj_hera_yj, &
          idj_hera_abs_yj, idj_hera_etaj, idj_jets_xhera, idj_jets_H1xhera)
      is_extraj = .true.
      !> TEEC observable in jets
    case (idj_TEEC_y, idj_TEEC_abs_y)
      is_extraj = .true.
      !> photon ET fraction
    case (idj_zgam)
      is_extraj = .true.
      !> *
    case default
      is_extraj = .false.
    end select
  end function isExtraj_obs

  !-----------------------------------------------------------------------------
  !> @brief
  !> Wrapper function for jet observables.
  !> These are evaluated during the recombination algorithm in order to
  !> evaluate jet-veto conditions and cannot be used for histograms!
  !
  !> @param[in] obsIdj unique identifier for the jet observabes
  !> @param[in] npar kinematics specified by the number of particles
  !> @param[in] v four-vector of the jet
  !> @param[in] qflav is the jet flavoured or not?
  !-----------------------------------------------------------------------------
  pure real(kind=wp) function eval_obsj(obsIdj, npar, v)
    integer, intent(in) :: obsIdj, npar
    real(kind=wp), intent(in), dimension(4) :: v
    integer :: i
    real(kind=wp) :: dummy
    select case (obsIdj)
    !> jet properties
    case (idj_pt)
      eval_obsj = v1_pt(v)
    case (idj_y)
      eval_obsj = v1_y(v)
    case (idj_abs_y)
      eval_obsj = abs(v1_y(v))
    case (idj_eta)
      eval_obsj = v1_eta(v)
    case (idj_abs_eta)
      eval_obsj = abs(v1_eta(v))
    case (idj_et)
      eval_obsj = v1_et(v)
    !> properties dependent on npar kinematics
    case (idj_dR_l1)
      eval_obsj = v2_delta_R(kin(npar)%p(:, npar - 1), v)
    case (idj_dR_l2)
      eval_obsj = v2_delta_R(kin(npar)%p(:, npar), v)
    case (idj_dR_l3)
      eval_obsj = v2_delta_R(kin(npar)%p(:, npar - 3), v)
    case (idj_dR_l4)
      eval_obsj = v2_delta_R(kin(npar)%p(:, npar - 2), v)
    case (idj_min_dR_l12)
      eval_obsj = min(v2_delta_R(kin(npar)%p(:, npar - 1), v), &
                      v2_delta_R(kin(npar)%p(:, npar), v))
    case (idj_min_dR_l34)
      eval_obsj = min(v2_delta_R(kin(npar)%p(:, npar - 3), v), &
                      v2_delta_R(kin(npar)%p(:, npar - 2), v))
    case (idj_min_dR_l1234)
      eval_obsj = min(v2_delta_R(kin(npar)%p(:, npar - 1), v), &
                      v2_delta_R(kin(npar)%p(:, npar), v), &
                      v2_delta_R(kin(npar)%p(:, npar - 3), v), &
                      v2_delta_R(kin(npar)%p(:, npar - 2), v))
    case (idj_dR_exp_l1)
      eval_obsj = v2_delta_R_exp(kin(npar)%p(:, npar - 1), v)
    case (idj_dR_exp_l2)
      eval_obsj = v2_delta_R_exp(kin(npar)%p(:, npar), v)
    case (idj_dR_exp_l3)
      eval_obsj = v2_delta_R_exp(kin(npar)%p(:, npar - 3), v)
    case (idj_dR_exp_l4)
      eval_obsj = v2_delta_R_exp(kin(npar)%p(:, npar - 2), v)
    case (idj_min_dR_exp_l12)
      eval_obsj = min(v2_delta_R_exp(kin(npar)%p(:, npar - 1), v), &
                      v2_delta_R_exp(kin(npar)%p(:, npar), v))
    case (idj_min_dR_exp_l34)
      eval_obsj = min(v2_delta_R_exp(kin(npar)%p(:, npar - 3), v), &
                      v2_delta_R_exp(kin(npar)%p(:, npar - 2), v))
    case (idj_min_dR_exp_l1234)
      eval_obsj = min(v2_delta_R_exp(kin(npar)%p(:, npar - 1), v), &
                      v2_delta_R_exp(kin(npar)%p(:, npar), v), &
                      v2_delta_R_exp(kin(npar)%p(:, npar - 3), v), &
                      v2_delta_R_exp(kin(npar)%p(:, npar - 2), v))
    !> photon separation
    case (idj_min_dR_gam)
      eval_obsj = -1._wp  ! some invalid state: assume there's at least one photon
      do i = 1, kin(npar)%nphotons
        dummy = v2_delta_R(kin(npar)%pphotons(:, i), v)
        if ((eval_obsj < 0._wp) .or. (dummy < eval_obsj)) eval_obsj = dummy
      end do
    case (idj_min_dR_exp_gam)
      eval_obsj = -1._wp  ! some invalid state: assume there's at least one photon
      do i = 1, kin(npar)%nphotons
        dummy = v2_delta_R_exp(kin(npar)%pphotons(:, i), v)
        if ((eval_obsj < 0._wp) .or. (dummy < eval_obsj)) eval_obsj = dummy
      end do
    !> flavoured jet properties
    case (idfj_pt)
      eval_obsj = v1_pt(v)
    case (idfj_y)
      eval_obsj = v1_y(v)
    case (idfj_abs_y)
      eval_obsj = abs(v1_y(v))
    case (idfj_eta)
      eval_obsj = v1_eta(v)
    case (idfj_abs_eta)
      eval_obsj = abs(v1_eta(v))
    case (idfj_et)
      eval_obsj = v1_et(v)
      !> escape
    case default
      eval_obsj = 0._wp
      ! stop "eval_obsj: unknown obsIdj"
    end select
  end function eval_obsj

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the algorithm and cone size
  !
  !> @param[in] algorithm string identifying the algorithm
  !> @param[in] param parameter of the jet algorithm
  !>              * ycut for JADE/Durham
  !>              * R else
  !> @param[in] recomb string identifying the recombination scheme
  !-----------------------------------------------------------------------------
  subroutine init_jet(algorithm, param, recomb)
    use StrHelper_mod
    use Process_mod
    character(len=*), intent(in) :: algorithm, recomb
    real(kind=wp), intent(in) :: param

    !> global setting for the algorithm
    select case (to_lower_str(algorithm))
    case ("none")
      jetalg = 0
      ip_jet = 0
    case ("antikt", "akt")
      jetalg = 1
      ip_jet = -1
    case ("ca", "cam")
      jetalg = 2
      ip_jet = 0
    case ("kt")
      jetalg = 3
      ip_jet = +1
    case ("jade")
      jetalg = 4
      ip_jet = 0
      jet_ycut = param
    case ("flavkt")
      jetalg = 5
    case ("durham")
      jetalg = 6
      ip_jet = 0
      jet_ycut = param
    case ("flavantikt")
      jetalg = 7
      ip_jet = -1
    case default
      print *, "unknown jet algorithm: ", algorithm
      stop
    end select

    inputRcone = param
    call setRcone_jet()

    !> recomb scheme
    select case (to_upper_str(recomb))
    case ("ET")
      qEtrecomb = .true.
    case ("V4")
      qEtrecomb = .false.
    case default
      print *, "unknown jet recomb: ", recomb
      stop
    end select

    !> compatibility checks
    if ((beam1_proc == pid_ep .and. beam2_proc == pid_em) .and. &
        (jetalg /= 4 .and. jetalg /= 6)) then
      print *, "WARNING: init_jet: epem process without JADE/Durham algorithm?!"
      !stop
    end if

    !> register observables
    list_obsj(idj_pt) = JetSelector_t(.true., "jets_pt")
    list_obsj(idj_y) = JetSelector_t(.true., "jets_y")
    list_obsj(idj_abs_y) = JetSelector_t(.true., "jets_abs_y")
    list_obsj(idj_eta) = JetSelector_t(.true., "jets_eta")
    list_obsj(idj_abs_eta) = JetSelector_t(.true., "jets_abs_eta")
    list_obsj(idj_et) = JetSelector_t(.true., "jets_et")
    !> register flavoured jet observables
    list_obsj(idfj_pt) = JetSelector_t(.true., "jets_flav_pt", qflavobs=.true.)
    list_obsj(idfj_y) = JetSelector_t(.true., "jets_flav_y", qflavobs=.true.)
    list_obsj(idfj_abs_y) = JetSelector_t(.true., "jets_flav_abs_y", qflavobs=.true.)
    list_obsj(idfj_eta) = JetSelector_t(.true., "jets_flav_eta", qflavobs=.true.)
    list_obsj(idfj_abs_eta) = JetSelector_t(.true., "jets_flav_abs_eta", qflavobs=.true.)
    list_obsj(idfj_et) = JetSelector_t(.true., "jets_flav_et", qflavobs=.true.)

    !> process-specific jet-observables (default init is qactive=.false.)
    select case (name_proc)

    case ("Z", "ZJ", "ZJJ", "ZJJJ", "Z_EW")
      list_obsj(idj_min_dR_l12) = JetSelector_t(.true., "jets_min_dr_lj")

    case ("WM", "WMJ", "WMJUNSYM", "WMJJ", "WMJJJ")
      list_obsj(idj_dR_l1) = JetSelector_t(.true., "jets_min_dr_lj")

    case ("WP", "WPJ", "WPJUNSYM", "WPJJ", "WPJJJ")
      list_obsj(idj_dR_l2) = JetSelector_t(.true., "jets_min_dr_lj")

    case ("HTO2PM", "HTO2PJM", "HTO2P", "HTO2PJ", "HTO2PJJ", "HTO2PJJJ", "HTO2PINC", "HTO2PQT")
      list_obsj(idj_min_dR_l12) = JetSelector_t(.true., "jets_min_dr_gj")

    case ("HTO2TAUM", "HTO2TAUJM", "HTO2TAU", "HTO2TAUJ", "HTO2TAUJJ", "HTO2TAUJJJ")
      list_obsj(idj_min_dR_l12) = JetSelector_t(.true., "jets_min_dr_tauj")

    case ("HTO4EM", "HTO4EJM", "HTO4E", "HTO4EJ", "HTO4EJJ", "HTO4EJJJ", "HTO4EINC", "HTO4EQT", &
      "HH", "HHJ", "HHJJ", "HHJJJ")
      list_obsj(idj_min_dR_l1234) = JetSelector_t(.true., "jets_min_dr_lj")

    case ("HTO2E2MUM", "HTO2E2MUJM", "HTO2E2MU", "HTO2E2MUJ", "HTO2E2MUJJ", &
          "HTO2E2MUJJJ", "HTO2E2MUINC", "HTO2E2MUQT")
      list_obsj(idj_min_dR_l12) = JetSelector_t(.true., "jets_min_dr_muj")
      list_obsj(idj_min_dR_l34) = JetSelector_t(.true., "jets_min_dr_ej")
      list_obsj(idj_min_dR_l1234) = JetSelector_t(.true., "jets_min_dr_lj")

    case ("GJ", "GJ_FRAG", "GJJ", "GJJJ", "G_EW", "GG", "GG_FRAG", "GGJ", "GGJFC", &
          "GGJJ", "GGJJFC", "GGJJJFC", "GG_EW", "GGG", "GGQT")
      list_obsj(idj_min_dR_gam) = JetSelector_t(.true., "jets_min_dr_gamj")
      list_obsj(idj_min_dR_exp_gam) = JetSelector_t(.true., "jets_min_dr_exp_gamj")
      list_obsj(idj_zgam) = JetSelector_t(.true., "jets_zgam")

    case ("DIS", "DISWM", "DISWP")
      list_obsj(idj_hera_ptj) = JetSelector_t(.true., "jets_hera_pt")
      list_obsj(idj_hera_yj) = JetSelector_t(.true., "jets_hera_y")
      list_obsj(idj_hera_abs_yj) = JetSelector_t(.true., "jets_hera_abs_y")
      list_obsj(idj_hera_etaj) = JetSelector_t(.true., "jets_hera_eta")
      list_obsj(idj_hera_abs_etaj) = JetSelector_t(.true., "jets_hera_abs_eta")
      list_obsj(idj_hera_etj) = JetSelector_t(.true., "jets_hera_et")
      list_obsj(idj_jets_xhera) = JetSelector_t(.true., "jets_xhera")
      list_obsj(idj_jets_H1xhera) = JetSelector_t(.true., "jets_H1xhera")

    ! case ("1JET", "2JET", "1JETFC", "2JETFC", "3JET", "4JET")
    !   list_obsj(idj_TEEC_y) = JetSelector_t(.true., "teec_y")
    !   list_obsj(idj_TEEC_abs_y) = JetSelector_t(.true., "teec_abs_y")

    end select

  end subroutine init_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> set up the flavour dressing algorthim
  !
  !> @param[in] algorithm string identifying the algorithm
  !-----------------------------------------------------------------------------
  subroutine init_dress_jet(algorithm)
    use StrHelper_mod
    use Process_mod
    character(len=*), intent(in) :: algorithm

    select case (to_lower_str(algorithm))
    case ('naive', 'none')
      iflav_dress = 0
    case ('sd', 'softdrop', 'soft_drop')
      iflav_dress = -2  ! C/A decluster by default
    case ('cluster_jade')
      iflav_dress = +4
      flav_dress_cluster_qbeam = .false.
    case ('cluster', 'cluster_flavkt')
      iflav_dress = +5
    case ('cluster_kt', 'cluster_durham')
      iflav_dress = +6
      flav_dress_cluster_alpha = 0._wp
    case ('revised_flavour_dressing')
      iflav_dress = +7
      flav_dress_cluster_alpha = 1._wp
      flav_dress_cluster_use_omega = .true.
      flav_dress_cluster_omega = 2._wp
      flav_qunderdress = .false.
    case default
      print *, "init_dress_jet: unknown algorithm: ", algorithm
      stop
    end select

    !> no: tag_flavour not yet parsed when this is called...
    ! if ( .not.isActive_flav() .and. (iflav_dress/=0) ) then
    !   print*, "init_dress_jet: tagging must be active for: ", algorithm
    !   stop
    ! end if

    print *, "init_dress_jet: ", iflav_dress

  end subroutine init_dress_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the width for a particle
  !
  !> @param[in] name string identifying parameter
  !> @param[in] valueINT optional integer value of the parameter
  !> @param[in] valueDBLE optional double value of the parameter
  !> @param[in] valueBOOL optional bool value of the parameter
  !-----------------------------------------------------------------------------
  subroutine setParam_jet(name, &
                          valueINT, &
                          valueDBLE, &
                          valueBOOL)
    use StrHelper_mod, only: to_upper_str
    use Flavour_mod, only: keepEvenTagged_flav
    character(len=*), intent(in) :: name
    integer, intent(in), optional :: valueINT
    real(kind=wp), intent(in), optional :: valueDBLE
    logical, intent(in), optional :: valueBOOL

    select case (to_upper_str(trim(adjustl(name))))
    case ('DYN_SOFT')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting dyn_soft requires valueDBLE!"
      end if
      if (valueDBLE < 0._wp) stop "setParam_jet: dyn_soft must be positive!"
      dyn_soft_jet = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAVKT_ALPHA')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flavkt_alpha requires valueDBLE!"
      end if
      flavkt_alpha = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAVKT_INCLUDE_COLOURLESS')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flavkt_include_colourless requires valueBOOL!"
      end if
      flavkt_include_colourless = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
    case ('FLAVKT_ONLY_PROTO')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flavkt_only_proto requires valueBOOL!"
      end if
      flavkt_only_proto = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
    case ('FLAVKT_SUM_NONQCD')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flavkt_sum_nonqcd requires valueBOOL!"
      end if
      flavkt_sum_nonqcd = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
      !> Flavour anti-kt algo
    case ('FLAVANTIKT_A')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flavantikt_a requires valueDBLE!"
      end if
      flavantikt_a = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
      !> settings for flavour dressing
    case ('FLAV_DRESS_SD_ALGO')
      if (.not. present(valueINT)) then
        stop "setParam_jet: setting iflav_dress requires valueINT!"
      end if
      if (valueINT >= 0) then
        stop "setParam_jet: setting iflav_dress for SD must be negative!"
      end if
      iflav_dress = valueINT
      print *, "setParam_jet: ", name, " = ", valueINT
    case ('FLAV_DRESS_CLUSTER_ALGO')
      if (.not. present(valueINT)) then
        stop "setParam_jet: setting iflav_dress requires valueINT!"
      end if
      if (valueINT <= 0) then
        stop "setParam_jet: setting iflav_dress for CLUSTER must be positive!"
      end if
      iflav_dress = valueINT
      select case (iflav_dress)
      case (4)  ! Jade
        flav_dress_cluster_qbeam = .false.
      case (6)  ! Durham/kt
        flav_dress_cluster_alpha = 0._wp
      end select
      print *, "setParam_jet: ", name, " = ", valueINT
    case ('FLAV_DRESS_SD_ZCUT')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_dress_SD_zcut requires valueDBLE!"
      end if
      flav_dress_SD_zcut = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_DRESS_SD_BETA')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_dress_SD_beta requires valueDBLE!"
      end if
      flav_dress_SD_beta = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_DRESS_CLUSTER_PROTO')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flav_dress_cluster_qproto requires valueBOOL!"
      end if
      flav_dress_cluster_qproto = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
    case ('FLAV_DRESS_CLUSTER_BEAM')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flav_dress_cluster_qbeam requires valueBOOL!"
      end if
      flav_dress_cluster_qbeam = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
    case ('FLAV_DRESS_CLUSTER_BEAM_NONQCD')
      if (.not. present(valueINT)) then
        stop "setParam_jet: setting flav_dress_cluster_beam_nonqcd requires valueINT!"
      end if
      flav_dress_cluster_beam_nonqcd = valueINT
      print *, "setParam_jet: ", name, " = ", valueINT
    case ('FLAV_DRESS_CLUSTER_ALPHA')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_dress_cluster_alpha requires valueDBLE!"
      end if
      flav_dress_cluster_alpha = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_DRESS_CLUSTER_IDRESS')
      if (.not. present(valueINT)) then
        stop "setParam_jet: setting flav_dress_cluster_idress requires valueINT!"
      end if
      flav_dress_cluster_idress = valueINT
      print *, "setParam_jet: ", name, " = ", valueINT
    case ('FLAV_DRESS_CLUSTER_PENALTY')
      if (.not. present(valueINT)) then
        stop "setParam_jet: setting flav_dress_cluster_ipenalty requires valueINT!"
      end if
      flav_dress_cluster_ipenalty = valueINT
      print *, "setParam_jet: ", name, " = ", valueINT
    case ('FLAV_DRESS_CLUSTER_BETA')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_dress_cluster_beta requires valueDBLE!"
      end if
      flav_dress_cluster_beta = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
      !>
    case ('FLAV_QUNDERDRESS')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flav_qunderdress requires valueBOOL!"
      end if
      flav_qunderdress = valueBOOL
      print *, "setParam_jet: ", name, " = ", valueBOOL
    case ('FLAV_UNDERDRESS_R')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_underdress_R requires valueDBLE!"
      end if
      flav_underdress_R = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_UNDERDRESS_ZCUT')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_underdress_zcut requires valueDBLE!"
      end if
      flav_underdress_zcut = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_UNDERDRESS_BETA')
      if (.not. present(valueDBLE)) then
        stop "setParam_jet: setting flav_underdress_beta requires valueDBLE!"
      end if
      flav_underdress_beta = valueDBLE
      print *, "setParam_jet: ", name, " = ", valueDBLE
    case ('FLAV_KEEP_EVEN_TAG')
      if (.not. present(valueBOOL)) then
        stop "setParam_jet: setting flav_keep_even_tag requires valueBOOL!"
      end if
      call keepEvenTagged_flav(valueBOOL)
      print *, "setParam_jet: ", name, " = ", valueBOOL
      !>
    case default
      print *, "setParam_jet: no parameter by the name: ", name
      stop
    end select

  end subroutine setParam_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> get the algorithm id
  !-----------------------------------------------------------------------------
  integer function algo_jet()
    algo_jet = jetalg
  end function algo_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> get the flavour dressing algorithm id
  !-----------------------------------------------------------------------------
  integer function algo_dress_jet()
    algo_dress_jet = iflav_dress
  end function algo_dress_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the cone size
  !
  !> @param[in] R optional argument for the R-parameter in the jet algorithm
  !-----------------------------------------------------------------------------
  subroutine setRcone_jet(R)
    real(kind=wp), optional, intent(in) :: R
    real(kind=wp) :: Rcone_val
    if (present(R)) then
      !> used in multi-runs to set different thread-local values
      Rcone_val = R
    else
      !> default to using the input Rcone value
      Rcone_val = inputRcone
    end if
    Rcone = Rcone_val
    RconeSq = Rcone_val**2
  end subroutine setRcone_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> get the algorithm cone size
  !-----------------------------------------------------------------------------
  real(kind=wp) function getRcone_jet()
    getRcone_jet = Rcone
  end function getRcone_jet

  subroutine setSoftdrop_jet(R0, zcut, beta, ip)
    real(kind=wp), intent(in) :: R0, zcut, beta
    integer, intent(in) :: ip
    softdrop = .true.
  end subroutine setSoftdrop_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Convert a name string to the jet observable identifier
  !
  !> @param[in] name string containing the name of a jet observable
  !> @return observable id (0 if unsuccessful)
  !-----------------------------------------------------------------------------
  function getIdFromName_jet(name)
    use StrHelper_mod, only: to_upper_str
    integer :: getIdFromName_jet
    character(len=*), intent(in) :: name
    integer :: i
    getIdFromName_jet = 0
    do i = 1, n_obsj
      if (.not. list_obsj(i)%qactive) cycle
      if (trim(to_upper_str(list_obsj(i)%name)) == trim(to_upper_str(name))) then
        getIdFromName_jet = i
        return
      end if
    end do
    if (getIdFromName_jet == 0) then
      print *, "getIdFromName_jet: unmatched observable: ", name
    end if
  end function getIdFromName_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the selector for obsName
  !> @note omit both `lower` and `upper` to disable the selector
  !
  !> @param[in] obsName unique name of the jet observable
  !> @param[in] lower optional lower cut
  !> @param[in] upper optional upper cut
  !> @param[in] invert optional flag controlling whether we "accept" or "reject"
  !-----------------------------------------------------------------------------
  subroutine setSelector_jet(obsName, lower, upper, invert)
    use Flavour_mod, only: setActive_flav
    character(len=*), intent(in) :: obsName
    real(kind=wp), optional, intent(in) :: lower, upper
    logical, optional, intent(in) :: invert
    integer :: obsId

    obsId = getIdFromName_jet(obsName)

    if (.not. list_obsj(obsId)%qactive) then
      print *, "setSelector_jet: inactive jet observable: ", obsName
      stop
    end if

    if (list_obsj(obsId)%qlower .or. list_obsj(obsId)%qupper) then
      print *, "setSelector_jet: jet observable ", obsName, " already set?!"
      stop
    end if

    if (present(lower)) list_obsj(obsId)%lower = lower
    if (present(upper)) list_obsj(obsId)%upper = upper
    list_obsj(obsId)%qlower = present(lower)
    list_obsj(obsId)%qupper = present(upper)
    if (present(invert)) then
      list_obsj(obsId)%qinvert = invert
    end if ! default init is .false. (see above)

    !> turn on flavour-tagging if jet selector is flavour-sensitive
    if (list_obsj(obsId)%qflavobs) call setActive_flav(.true.)

  end subroutine setSelector_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> test for an jet cut if the value passes the selector
  !
  !> @param[in] obsId integer identifyer for the extra jet observable
  !> @param[in] val value of the observable associated with `obsId`
  !-----------------------------------------------------------------------------
  function qpass_jet(obsId, val)
    logical :: qpass_jet
    integer, intent(in) :: obsId
    real(kind=wp), intent(in) :: val

    qpass_jet = .true.
    if (.not. list_obsj(obsId)%qactive) return
    if ((.not. list_obsj(obsId)%qlower) .and. &
        (.not. list_obsj(obsId)%qupper)) return  ! no cut registered

    if (.not. list_obsj(obsId)%qinvert .and. &
        ( &
        (list_obsj(obsId)%qlower .and. val < list_obsj(obsId)%lower) .or. &
        (list_obsj(obsId)%qupper .and. val > list_obsj(obsId)%upper) &
        ) &
        .or. &
        list_obsj(obsId)%qinvert .and. &
        ( &
        (list_obsj(obsId)%qlower .and. val >= list_obsj(obsId)%lower .and. &
         list_obsj(obsId)%qupper .and. val <= list_obsj(obsId)%upper) .or. &
        (list_obsj(obsId)%qlower .and. val >= list_obsj(obsId)%lower .and. &
         .not. list_obsj(obsId)%qupper) .or. &
        (list_obsj(obsId)%qupper .and. val <= list_obsj(obsId)%upper .and. &
         .not. list_obsj(obsId)%qlower) &
        )) then
      qpass_jet = .false.
    end if

    ! if (isExtraj_obs(obsId)) then
    !   print*, "qpass_jet: extra obs:", obsId, val, ">", qpass_jet
    ! end if

  end function qpass_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the registered selectors
  !
  !> @param[in] unit optional logical unit connected for formatted output (file, screen)
  !-----------------------------------------------------------------------------
  subroutine digest_jet(unit)
    use Flavour_mod, only: isActive_flav, qKeepEvenTagged_flav
    use Process_mod, only: nPDFs_proc
    integer, optional, intent(in) :: unit
    integer :: i, iout
    character(len=*), parameter :: fmt_txt = '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_algo = '(1x,"*",3x,"jet algorithm:",A10,4x,"R = ",1f4.2,T103,"*")'
    character(len=*), parameter :: fmt_jade = '(1x,"*",3x,"jet algorithm:",A10,4x,"ycut = ",1f4.2,T103,"*")'
    character(len=*), parameter :: fmt_valDBLE = '(1x,"*",7x,A20," = ",1f12.2,T103,"*")'
    character(len=*), parameter :: fmt_valINT = '(1x,"*",7x,A20," = ",I0,T103,"*")'
    character(len=*), parameter :: fmt_valBOOL = '(1x,"*",7x,A20," = ",L,T103,"*")'
    character(len=*), parameter :: fmt_valTXT = '(1x,"*",17x,A,T103,"*")'
    character(len=*), parameter :: fmt_min_max = '(1x,"*",3x,A6,":",1f12.2," <= ",A15," <= ",1f12.2,T103,"*")'
    character(len=*), parameter :: fmt_min = '(1x,"*",3x,A6,":",1f12.2," <= ",A15,T103,"*")'
    character(len=*), parameter :: fmt_max = '(1x,"*",3x,A6,":",16x,A15," <= ",1f12.2,T103,"*")'
    character(len=6) :: seltype

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    end if

    select case (jetalg)
    case (0)
      write (iout, fmt_algo) "none", Rcone
    case (1)
      write (iout, fmt_algo) "anti-kt", Rcone
    case (2)
      write (iout, fmt_algo) "C/A", Rcone
    case (3)
      write (iout, fmt_algo) "kT", Rcone
    case (4)
      write (iout, fmt_jade) "JADE", jet_ycut
    case (5)
      write (iout, fmt_algo) "flavour-kt", Rcone
      write (iout, fmt_valDBLE) "alpha", flavkt_alpha
      write (iout, fmt_valBOOL) "include_colourless", flavkt_include_colourless
      write (iout, fmt_valBOOL) "only_proto", flavkt_only_proto
      write (iout, fmt_valBOOL) "sum_nonqcd", flavkt_sum_nonqcd
    case (6)
      write (iout, fmt_jade) "Durham", jet_ycut
    case (7)
      write (iout, fmt_algo) "Flavour anti-kt", Rcone
      write (iout, fmt_valDBLE) "a", flavantikt_a
    end select

    if (isActive_flav()) then
      select case (iflav_dress)
      case (0)
        write (iout, fmt_txt) "naive flavour dressing"
      case (-1)
        write (iout, fmt_txt) "recursive soft drop: anti-kt"
      case (-2)
        write (iout, fmt_txt) "recursive soft drop: C/A"
      case (-3)
        write (iout, fmt_txt) "recursive soft drop: kt"
      case (-4)
        write (iout, fmt_txt) "recursive soft drop: Jade"
      case (4)
        write (iout, fmt_txt) "flavour clustering: Jade"
      case (5)
        write (iout, fmt_txt) "flavour clustering: flavour-kt"
        write (iout, fmt_valDBLE) "alpha", flav_dress_cluster_alpha
      case (6)
        write (iout, fmt_txt) "flavour clustering: Durham"
        write (iout, fmt_valDBLE) "alpha", flav_dress_cluster_alpha
      case (7)
        write (iout, fmt_txt) "flavour clustering: revised 08/2023"
        write (iout, fmt_valDBLE) "alpha", flav_dress_cluster_alpha
        write (iout, fmt_valDBLE) "omega", flav_dress_cluster_omega
      case default
        write (iout, fmt_txt) "?????"
        stop
      end select
      if (iflav_dress < 0) then
        write (iout, fmt_valDBLE) "zcut", flav_dress_SD_zcut
        write (iout, fmt_valDBLE) "beta", flav_dress_SD_beta
      end if
      if (iflav_dress > 0) then
        if (flav_qunderdress) then
          write (iout, fmt_valDBLE) "underdress[R]   ", flav_underdress_R
          write (iout, fmt_valDBLE) "underdress[zcut]", flav_underdress_zcut
          write (iout, fmt_valDBLE) "underdress[beta]", flav_underdress_beta
        else
          write (iout, fmt_valTXT) "no underdress"
        end if
        if (flav_dress_cluster_qproto) then
          write (iout, fmt_valTXT) "dress proto-jets"
          if (flav_dress_cluster_idress /= 1) &
            stop "digest_jet: we require top-down!"
        else
          write (iout, fmt_valTXT) "dress reconstructed jets"
        end if
        if (flav_dress_cluster_qbeam .and. (nPDFs_proc > 0)) then
          write (iout, fmt_valTXT) "include beam distances"
        else
          write (iout, fmt_valTXT) "no beam distances"
        end if
        select case (flav_dress_cluster_beam_nonqcd)
        case (0)
          write (iout, fmt_valTXT) "no non-QCD particles in the beam"
        case (1)
          write (iout, fmt_valTXT) "Et of sum of non-QCD particles in the beam"
        case (2)
          write (iout, fmt_valTXT) "Individual non-QCD particles in the beam"
        case default
          write (iout, fmt_valTXT) "invalid idress", flav_dress_cluster_idress
          stop
        end select
        select case (flav_dress_cluster_idress)
        case (0)
          write (iout, fmt_valTXT) "democratic dressing"
        case (1)
          write (iout, fmt_valTXT) "top-down dressing"
        case default
          write (iout, fmt_valTXT) "invalid idress", flav_dress_cluster_idress
          stop
        end select
        select case (flav_dress_cluster_ipenalty)
        case (0)
          write (iout, fmt_valTXT) "f-j penalty: off"
        case (1)
          write (iout, fmt_valTXT) "f-j penalty: wilson"
          write (iout, fmt_valDBLE) "beta", flav_dress_cluster_beta
          if ((flav_dress_cluster_beta < 0._wp) .or. &
              (flav_dress_cluster_beta > 1._wp)) &
            stop "digest_jet: invalid beta value!"
        case (2)
          write (iout, fmt_valTXT) "f-j penalty: mult"
          write (iout, fmt_valDBLE) "beta", flav_dress_cluster_beta
          if (flav_dress_cluster_beta < 0._wp) &
            stop "digest_jet: invalid beta value!"
        case default
          write (iout, fmt_valTXT) "f-j penalty: ????"
          stop
        end select
      end if
      if (qKeepEvenTagged_flav()) then
        write (iout, fmt_valTXT) "flavour net summation"
      else
        write (iout, fmt_valTXT) "flavour modulo 2 summation"
      end if
    end if

    do i = 1, n_obsj
      if (.not. list_obsj(i)%qactive) cycle
      seltype = ''
      if (list_obsj(i)%qinvert) then
        seltype = 'reject'
      else
        seltype = 'accept'
      end if
      if (list_obsj(i)%qlower .and. list_obsj(i)%qupper) then
        write (iout, fmt_min_max) seltype, list_obsj(i)%lower, list_obsj(i)%name, list_obsj(i)%upper
      else if (list_obsj(i)%qlower .and. .not. list_obsj(i)%qupper) then
        write (iout, fmt_min) seltype, list_obsj(i)%lower, list_obsj(i)%name
      else if (.not. list_obsj(i)%qlower .and. list_obsj(i)%qupper) then
        write (iout, fmt_max) seltype, list_obsj(i)%name, list_obsj(i)%upper
      end if
    end do

  end subroutine digest_jet


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine from the jet selectors the minimal partonic
  !> centre-of-mass energy (squared)
  !
  !> @return minimal value for shat [GeV^2]
  !-----------------------------------------------------------------------------
  real(kind=wp) function min_shat_jet()
    real(kind=wp) :: min_val
    min_shat_jet = 0._wp
    !> only pt and et of the jets are dimensionful variables I have
    !> note: if minnjets = 0 this does not impose a shat_min cut!!!
    if (list_obsj(idj_pt)%qactive .and. &
        .not. list_obsj(idj_pt)%qinvert .and. &
        list_obsj(idj_pt)%qlower) then
      min_val = list_obsj(idj_pt)%lower
      min_shat_jet = max(min_shat_jet, (2._wp*min_val)**2)
    end if
    if (list_obsj(idj_et)%qactive .and. &
        .not. list_obsj(idj_et)%qinvert .and. &
        list_obsj(idj_et)%qlower) then
      min_val = list_obsj(idj_et)%lower
      min_shat_jet = max(min_shat_jet, (2._wp*min_val)**2)
    end if
    if (list_obsj(idj_hera_ptj)%qactive .and. &
        .not. list_obsj(idj_hera_ptj)%qinvert .and. &
        list_obsj(idj_hera_ptj)%qlower) then
      min_val = list_obsj(idj_hera_ptj)%lower
      min_shat_jet = max(min_shat_jet, (2._wp*min_val)**2)
    end if
    !>---------------------------------------------------
    !> !!! this is WRONG !!!
    !> never uncomment! Left in so noone get's the idea of implementing such a line
    !> see comments in min_shat_sel
    ! if (      list_obsj(idj_pt)%qinvert .and. list_obsj(idj_pt)%qupper ) then
    !   min_valpt = list_obsj(idj_pt)%upper
    ! end if
    !>---------------------------------------------------
    return
  end function min_shat_jet


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to initialize the (threadprivate) buffer for the jet-algorithm
  !
  !> @param[in] maxnproto largest array size needed for the jet-algorithm
  !-----------------------------------------------------------------------------
  subroutine initBuffer_jet(maxnproto)
    use Process_mod, only: nPDFs_proc, n_fs_spectators_proc, n_fs_photons_proc
    use Flavour_mod, only: isActive_flav
    integer, intent(in) :: maxnproto
    integer :: maxnonqcd
    !> DIS using JADE/Durham needs another pseudo-particle
    integer :: size

    !> no jet algo
    if (jetalg == 0) return

    size = maxnproto
    if (nPDFs_proc == 1) then
!      size = maxnproto + 1
    end if

    allocate (jets(size))
    allocate (diB(size))
    allocate (dij(size, size))
    allocate (clusterHist_jet(size))
    allocate (pt_sort_jet(size))

    !> allocate flavour-tagging arrays
    if (isActive_flav()) then
      allocate (flavjets(size))
      allocate (clusterHist_flavjet(size))
      !> TODO: dealing with photons
      maxnonqcd = n_fs_spectators_proc
      allocate (nonQCDvec(maxnonqcd))
    end if

  end subroutine initBuffer_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to deallocate the (threadprivate) buffer for the jet-algorithm
  !-----------------------------------------------------------------------------
  subroutine destroyBuffer_jet()
    use Flavour_mod, only: isActive_flav

    !> no jet algo
    if (jetalg == 0) return

    deallocate (jets)
    deallocate (diB)
    deallocate (dij)
    deallocate (clusterHist_jet)
    deallocate (pt_sort_jet)

    !> de-allocate flavour-tagging arrays
    if (isActive_flav()) then
      deallocate (flavjets)
      deallocate (clusterHist_flavjet)
      deallocate (nonQCDvec)
    end if

  end subroutine destroyBuffer_jet


  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialise & prepare the local arrays for the jet recombination algorithm
  !
  !> @param[in] npar kinematics specified by the number of particles
  !-----------------------------------------------------------------------------
  subroutine initRecomb_jet(npar)
    use Flavour_mod, only: isActive_flav, iTag_flav
    use PhotonFragmentation_mod, only: isActive_frag
    integer, intent(in) :: npar
    type(Kin_t), pointer :: kd
    integer :: i, k, nonQCDdim
    real(kind=wp), dimension(4) :: p_dummy

    !> prettify code
    kd => kin(npar)

    !> copy over the momenta and initialize
    npar_jet = npar
    nproto_jet = 0
#ifdef USEFASTJET
    !> reset cache
    kd%njets_fj(0) = -1
#endif
    njets_jet = -1
    flav_njets_jet = -1
    do i = kd%ijets_lower, kd%ijets_upper
      nproto_jet = nproto_jet + 1
      jets(nproto_jet)%isJet = -1 ! all proto-jets
      if (isActive_frag()) then
        !> can only happen when `ifrag` set; only for photon frag, not hfrag.
        if (i == kd%ifrag) then
          p_dummy(:) = (1._wp - kd%zfrag)*kd%p(:, i)  ! parton carries (1-z_gamma)
          jets(nproto_jet)%p(:) = p_dummy
          jets(nproto_jet)%pt2 = v1_pt2(p_dummy)
          jets(nproto_jet)%y = v1_y(p_dummy)
          jets(nproto_jet)%phi = v1_phi(p_dummy)
          ! print*, "initRecomb_jet: fragmentation", kd%ifrag, kd%zfrag
          ! print*, "> orig", kd%p(:,i)
          ! print*, "> frag", p_dummy(:)
          ! read(*,*)
        end if
        if (i == kd%i2frag) then
          p_dummy(:) = (1._wp - kd%z2frag)*kd%p(:, i)  ! parton carries (1-z_gamma)
          jets(nproto_jet)%p(:) = p_dummy
          jets(nproto_jet)%pt2 = v1_pt2(p_dummy)
          jets(nproto_jet)%y = v1_y(p_dummy)
          jets(nproto_jet)%phi = v1_phi(p_dummy)
        end if
      else
        jets(nproto_jet)%p(:) = kd%p(:, i)
        jets(nproto_jet)%pt2 = v1_pt2(kd%p(:, i))
        jets(nproto_jet)%y = v1_y(kd%p(:, i))
        jets(nproto_jet)%phi = v1_phi(kd%p(:, i))
      end if
#ifdef USEFASTJET
      !> populate parton momenta
      kd%pjets_fj(:,nproto_jet,0) = jets(nproto_jet)%p(:)
      if (isActive_flav()) then
         !> populate parton flavours
         if (iTag_flav(npar,i) > 0) then
            kd%fljets_fj(nproto_jet,0) = +1
         else if (iTag_flav(npar,i) < 0) then
            kd%fljets_fj(nproto_jet,0) = -1
         else
            kd%fljets_fj(nproto_jet,0) = 0
         end if
      end if
#endif
    end do
#ifdef USEFASTJET
    !> set number of parton momenta
    kd%njets_fj(0) = nproto_jet
#endif

    jets(:)%nflav_q = 0
    jets(:)%nflav_qb = 0
    jets(:)%isFlavoured = .false.

    !> flavour-tagging (working on flavjets)
    flav_njets_flavjet = -1
    if (isActive_flav()) then
      !> create singlet from all the non-QCD spectator particles
      nonQCDsum%p(:) = 0._wp
      nonQCDdim = 0
      do k = 3, npar
        !> loop over all non-QCD particles in the final state
        if ((k >= kd%ijets_lower) .and. (k <= kd%ijets_upper)) cycle
        nonQCDsum%p(:) = nonQCDsum%p(:) + kd%p(:, k)
        nonQCDdim = nonQCDdim + 1
        nonQCDvec(nonQCDdim)%p(:) = kd%p(:, k)
        nonQCDvec(nonQCDdim)%pt2 = v1_pt2(nonQCDvec(nonQCDdim)%p)
        nonQCDvec(nonQCDdim)%y = v1_y(nonQCDvec(nonQCDdim)%p)
        nonQCDvec(nonQCDdim)%m2 = v1_minv2(nonQCDvec(nonQCDdim)%p)
      end do
      if (kd%ifrag > 0) then
        nonQCDsum%p(:) = nonQCDsum%p(:) + kd%zfrag*kd%p(:, kd%ifrag)  ! photon carries (z_gamma)
        nonQCDdim = nonQCDdim + 1
        nonQCDvec(nonQCDdim)%p(:) = kd%zfrag*kd%p(:, kd%ifrag)
      end if
      if (kd%i2frag > 0) then
        nonQCDsum%p(:) = nonQCDsum%p(:) + kd%z2frag*kd%p(:, kd%i2frag)  ! photon2 carries (z2_gamma)
        nonQCDdim = nonQCDdim + 1
        nonQCDvec(nonQCDdim)%p(:) = kd%z2frag*kd%p(:, kd%i2frag)
      end if
      nonQCDvecsize = nonQCDdim
      nonQCDsum%pt2 = v1_pt2(nonQCDsum%p)
      nonQCDsum%y = v1_y(nonQCDsum%p)
      nonQCDsum%m2 = v1_minv2(nonQCDsum%p)
      flavjets = jets
      do i = 1, nproto_jet
        !> find initial partonic flavours
        if (iTag_flav(npar, kd%ijets_lower + i - 1) > 0) then
          flavjets(i)%nflav_q = 1
          flavjets(i)%nflav_qb = 0
          flavjets(i)%isFlavoured = .true.
        else if (iTag_flav(npar, kd%ijets_lower + i - 1) < 0) then
          flavjets(i)%nflav_q = 0
          flavjets(i)%nflav_qb = 1
          flavjets(i)%isFlavoured = .true.
        else
          flavjets(i)%nflav_q = 0
          flavjets(i)%nflav_qb = 0
          flavjets(i)%isFlavoured = .false.
        end if
      end do
    end if

    if (nproto_jet .ne. kd%nproto) then
      print *, "initRecomb_jet: nproto don't match"
      stop
    end if

  end subroutine initRecomb_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Append another "protojet" to the list
  !> has to be called *after* `initRecomb_jet` and *before* `cluster_jet`
  !
  !> @param[in] o the four-momentum of the new parton
  !-----------------------------------------------------------------------------
  integer function pushProto_jet(p)
    real(kind=wp), dimension(1:4), intent(in) :: p

    nproto_jet = nproto_jet + 1

    jets(nproto_jet)%isJet = -1
    jets(nproto_jet)%p(:) = p(:)
    jets(nproto_jet)%pt2 = v1_pt2(p(:))
    jets(nproto_jet)%y = v1_y(p(:))
    jets(nproto_jet)%phi = v1_phi(p(:))

    pushProto_jet = nproto_jet

  end function pushProto_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> merge two proto-jets
  !> this routine does not touch the flavour information of the jets!
  !
  !> @param[inout] j1 first proto-jet (will absorb j2)
  !> @param[inout] j2 second proto-jet (will be set inactive)
  !-----------------------------------------------------------------------------
  pure subroutine merge_jet(j1, j2)
    use Constants_mod
    type(Jet_t), intent(inout) :: j1, j2
    !> for Et-weighted recombination
    real(kind=wp) :: Et1, Et2, Etsum, eta, phi1, phi2, phi

    !> only for debugging: in order to keep this routine `pure`,
    !> we can't have print or stop statements
    ! !> are we really dealing with proto-jets?
    ! if ( j1%isJet>=0 .or. j2%isJet>=0 ) then
    !   print*, "merge_jet: not proto-jets"
    !   stop
    ! end if

    if (.not. qEtrecomb) then
      !>----- add four momenta
      j1%p = j1%p + j2%p
      j1%pt2 = v1_pt2(j1%p)
      j1%y = v1_y(j1%p)
      j1%phi = v1_phi(j1%p)

    else
      !>----- ET weighted recombination
      !> all proto-jets are *massless*: ET = pt, eta = y
      Et1 = sqrt(j1%pt2)
      Et2 = sqrt(j2%pt2)
      Etsum = Et1 + Et2
      eta = (Et1*j1%y + Et2*j2%y)/Etsum
      phi1 = j1%phi
      phi2 = j2%phi
      if (abs(phi1 - phi2) > pi) then
        if (phi1 < phi2) then
          phi1 = twopi + phi1
        else
          phi2 = twopi + phi2
        end if
      end if
      phi = (Et1*phi1 + Et2*phi2)/Etsum
      if (phi > twopi) phi = phi - twopi
      !> new massless proto-jet
      j1%p(1) = Etsum*cos(phi)
      j1%p(2) = Etsum*sin(phi)
      j1%p(3) = Etsum*sinh(eta)
      j1%p(4) = Etsum*cosh(eta)
      j1%pt2 = Etsum**2
      j1%y = eta
      j1%phi = phi
    end if

    !> second jet eliminated
    j2%isJet = 0

  end subroutine merge_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply the jet recombination
  !
  !> @param[in] RSq optional argument to override the Rcone value (squared!)
  !-----------------------------------------------------------------------------
  subroutine cluster_jet(RSq)
    real(kind=wp), optional, intent(in) :: RSq
    real(kind=wp) :: RSq_val
    !>-----
    real(kind=wp) :: diB_min, dij_min
    integer :: iB_min, ij_min(2)
    integer :: i, j, k

    if (present(RSq)) then
      RSq_val = RSq
    else
      RSq_val = RconeSq
    end if

    if (softdrop) then
      ! if cluster_jet is being called for the first time (from apply_softdrop),
      ! then clusterHist_jet = -1; if not in this state, skip clustering.
      if (.not. (all(clusterHist_jet .eq. -1))) then
        return
      end if
    end if

    !> a valid value for `njets_jet` indicates that clustering has already been done
    !> this is reset in `initRecomb_jet`
    if (njets_jet >= 0) return

    !> init distance measures & clustering history
    do i = 1, nproto_jet
      clusterHist_jet(i) = i
      diB(i) = del_iB(jets(i))
      do j = i + 1, nproto_jet
        dij(i, j) = del_ij(jets(i), jets(j))
      end do
    end do

    !> perform the recombination algorithm
    njets_jet = 0
    do k = 1, nproto_jet
      ! find the minimum
      diB_min = -1._wp
      iB_min = -1
      dij_min = -1._wp
      ij_min = -1
      do i = 1, nproto_jet
        if (.not. jets(i)%isJet < 0) cycle ! not a proto-jet
        if ((diB(i) < diB_min) .or. (diB_min < 0._wp)) then
          diB_min = diB(i)
          iB_min = i
        end if
        do j = i + 1, nproto_jet
          if (.not. jets(j)%isJet < 0) cycle ! not a proto-jet
          if ((dij(i, j) < dij_min) .or. (dij_min < 0._wp)) then
            dij_min = dij(i, j)
            ij_min = (/i, j/)
          end if
        end do
      end do
      ! merge with beam or two proto-jets
      if ((dij_min >= 0._wp) .and. (dij_min < diB_min)) then
        ! dij smallest -> merge the two jets (`j` into `i`)
        i = ij_min(1)
        j = ij_min(2)
        call merge_jet(jets(i), jets(j))
        ! store clustering
        clusterHist_jet(j) = i  ! 2nd argument merged into 1st
        ! update distance measures
        diB(i) = del_iB(jets(i))
        do j = 1, nproto_jet
          if (.not. jets(j)%isJet < 0) cycle ! not a proto-jet
          if (j == i) cycle
          if (i < j) dij(i, j) = del_ij(jets(i), jets(j))
          if (i > j) dij(j, i) = del_ij(jets(i), jets(j))
        end do
      else
        ! diB smallest -> make it a jet
        jets(iB_min)%isJet = 1
        njets_jet = njets_jet + 1
      end if
    end do

    return

  contains  ! distance measures for antikt, cam, kt

    pure real(kind=wp) function del_iB(j_)
      type(Jet_t), intent(in) :: j_
      del_iB = j_%pt2**(ip_jet)
    end function del_iB

    pure real(kind=wp) function del_ij(j1, j2)
      use Constants_mod
      type(Jet_t), intent(in) :: j1, j2
      real(kind=wp) :: del_y, del_phi, deltaRSq
      real(kind=wp) :: RSq_dyn
      del_y = j1%y - j2%y
      del_phi = abs(j1%phi - j2%phi)
      if (del_phi .gt. pi) del_phi = twopi - del_phi
      deltaRSq = del_y**2 + del_phi**2
      RSq_dyn = RSq_val
      if (dyn_soft_jet > 0._wp) then
        RSq_dyn = RSq_dyn &
                  *(max(j1%pt2, j2%pt2)/min(j1%pt2, j2%pt2))**dyn_soft_jet
      end if
      del_ij = min(j1%pt2**(ip_jet), j2%pt2**(ip_jet))*deltaRSq/RSq_dyn
    end function del_ij

  end subroutine cluster_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> reclustering geared towards soft drop implementation (grooming mode)
  !-----------------------------------------------------------------------------
  subroutine recluster_jet()
    use Softdrop_mod
    implicit none
    integer, dimension(:), allocatable :: pinjet
    integer, dimension(:), allocatable :: jet_list
    integer, dimension(:, :), allocatable :: j_pass
    integer, dimension(:, :), allocatable :: merge_list
    real(kind=wp) :: RSq_val
    integer :: np, i, j, k, l, n
    real(kind=wp) :: dij_min, Rij
    integer :: ij_min(2)

    if (.not. allocated(clusterHist_orig)) then
      error stop "recluster_jet uses clusterHist_orig, which must have been allocated."
    end if

    RSq_val = R0_sdrop**2

    allocate (merge_list(2, nproto_jet))
    allocate (jet_list(njets_orig))
    merge_list = 0
    ! populate jet_list
    j = 0
    do i = 1, nproto_jet
      if (jets(i)%isJet .eq. 1) then
        j = j + 1
        jet_list(j) = i
      else
        cycle
      end if
    end do

    ! reset jets; calculate distances using second distance measure
    call initRecomb_jet(npar_jet)
    dij = 0._wp
    njets_jet = 0
    do i = 1, nproto_jet
      clusterHist_jet(i) = i
      do j = i + 1, nproto_jet
        dij(i, j) = del_ij(jets(i), jets(j), ip_jet2_sdrop)
      end do
    end do
    ! recluster particles within each jet
    n = 0
    do l = 1, njets_orig
      pinjet = partons_in_jet(clusterHist_orig, jet_list(l), nproto_jet)
      np = size(pinjet)
      allocate (j_pass(3, np))
      j_pass = 0
      ! just one parton in jet - reclustering is trivial
      if (np .eq. 1) then
        ! promote immediately to jet
        jets(jet_list(l))%isJet = 1
        njets_jet = njets_jet + 1
        merge_list(1, n + 1) = jet_list(l)
        merge_list(2, n + 1) = jet_list(l)
        n = n + 1
      else
        ! tweaked version of cluster_jet code
        ! np particles need (np-1) pairwise merges to result in 1 jet
        do k = 1, np - 1
          ! find the minimum distance
          dij_min = -1._wp
          ij_min = -1
          ! loop over constituent non-jet partons to find closest pair i,j
          do i = 1, np
            if (jets(pinjet(i))%isJet >= 0) then
              cycle ! not a proto-jet
            end if
            do j = i + 1, np
              if (jets(pinjet(j))%isJet >= 0) cycle ! not a proto-jet
              if ((dij(pinjet(i), pinjet(j)) < dij_min) .or. (dij_min < 0._wp)) then
                dij_min = dij(pinjet(i), pinjet(j))
                ij_min = (/pinjet(i), pinjet(j)/)
              end if
            end do
          end do
          ! dij smallest -> merge the two jets (`j` into `i`)
          ! here we depart from cluster_jet and always merge softer (j) into harder (i)
          if (jets(ij_min(1))%pt2 .ge. jets(ij_min(2))%pt2) then
            i = ij_min(1)
            j = ij_min(2)
          else
            i = ij_min(2)
            j = ij_min(1)
          end if
          j_pass(1, k) = i
          j_pass(2, k) = j
          ! test soft-drop criterion (can substitute any other test function)
          if (sdtest) then
            ! can replace this with any other function returning 1 or 0
            ! this is just the usual distance; independent of jet algorithm (consider using the usual function instead of del_ij)
            Rij = sqrt(del_ij(jets(i), jets(j), 0)*RSq_val)
            j_pass(3, k) = softdrop_test(sqrt(jets(i)%pt2), sqrt(jets(j)%pt2), zcut_sdrop, Rij, R0_sdrop, beta_sdrop)
!                write(*,*) "jets + pass:", i, j, j_pass(3,k)
          else
            j_pass(3, k) = 1
          end if
          call merge_jet(jets(i), jets(j))
          ! store clustering
          clusterHist_jet(j) = i  ! 2nd argument merged into 1st - i.e. j into i
          ! update distance measures
          do j = 1, nproto_jet
            if (.not. jets(j)%isJet < 0) cycle ! not a proto-jet
            if (j == i) cycle
            if (i < j) dij(i, j) = del_ij(jets(i), jets(j), ip_jet2_sdrop)
            if (i > j) dij(j, i) = del_ij(jets(i), jets(j), ip_jet2_sdrop)
          end do

        end do ! k loop (from 1 to np-1)

        ! remaining particle is a jet: promote
        j_pass(1, np) = j_pass(1, np - 1)
        j_pass(2, np) = j_pass(1, np - 1)
        j_pass(3, np) = 2
        jets(j_pass(1, np))%isJet = 1
        njets_jet = njets_jet + 1

        ! have completed soft-drop test table, now reconstruct desired merges
        merge_list(:, n + 1:n + np) = return_merges(j_pass, np)
        n = n + np
      end if
      deallocate (pinjet)
      deallocate (j_pass)
    end do ! l loop (over jets from 1 to nj)

    ! final reset
    call initRecomb_jet(npar_jet)
    njets_jet = 0
    do i = 1, nproto_jet
      clusterHist_jet(i) = i
    end do

    do i = 1, nproto_jet
      if (merge_list(2, i) .ne. merge_list(1, i)) then
        call merge_jet(jets(merge_list(1, i)), jets(merge_list(2, i)))
        clusterHist_jet(merge_list(2, i)) = merge_list(1, i)
      else if (merge_list(2, i) .ne. 0) then
        jets(merge_list(2, i))%isJet = 1
        njets_jet = njets_jet + 1
      end if
    end do

    ! clean up
    do i = 1, nproto_jet
      if ((clusterHist_jet(i) .eq. i) .and. (jets(i)%isJet .ne. 1)) then
        jets(i)%isJet = 0
        !not needed? clusterHist_jet(i) = 0
      end if
    end do

    return

  contains

    ! distance measures etc.
    pure real(kind=wp) function del_ij(j1, j2, ip)
      use Constants_mod
      type(Jet_t), intent(in) :: j1, j2
      real(kind=wp) :: del_y, del_phi, deltaRSq
      integer, intent(in) :: ip
      del_y = j1%y - j2%y
      del_phi = abs(j1%phi - j2%phi)
      if (del_phi .gt. pi) del_phi = twopi - del_phi
      deltaRSq = del_y**2 + del_phi**2
      del_ij = min(j1%pt2**(ip), j2%pt2**(ip))*deltaRSq/RSq_val
    end function del_ij

    ! returns 1 if the softdrop condition (arXiv: 1402.2657, eq. 1.1) is satisfied, else 0.
    pure integer function softdrop_test(pt1, pt2, zcut, R12, R0, beta)
      real(kind=wp), intent(in) :: pt1, pt2, zcut, R12, R0, beta
      real(kind=wp) :: LH, RH

      LH = min(pt1, pt2)/(pt1 + pt2)
      RH = zcut*((R12/R0)**beta)

      if (LH .gt. RH) then
        softdrop_test = 1
      else
        softdrop_test = 0
      end if

    end function softdrop_test

    pure function return_merges(jet_pass, np_)
      implicit none
      integer, dimension(:, :), intent(in) :: jet_pass
      integer, intent(in) :: np_
      integer, dimension(:, :), allocatable :: return_merges
      integer, dimension(:), allocatable :: jet_p_ind_func
      integer :: jet
      integer :: i, j, k, l, m

      allocate (return_merges(2, np_))
      return_merges = 0

      jet = jet_pass(1, np_)
      allocate (jet_p_ind_func(nproto_jet))
      jet_p_ind_func = 0
      jet_p_ind_func(jet) = 1
      m = np_ - 1
      ! find last valid (ie passing) merge in merge tree:
      ! search backward through list of merges until we find a merge that
      ! (i) produces the desired jet, and
      ! (ii) passes the soft-drop condition.
      ! The choice to always merge softer into harder means that we choose the correct
      ! subjet in the clustering history by testing only column 1.
      do while ((jet_pass(1, m) .ne. jet) .or. (jet_pass(3, m) .ne. 1))
        m = m - 1
        if (m .eq. 0) exit
      end do
      ! reconstruct prior merges up to last valid merge
      if (m .eq. 0) then
        ! jet returned is a singleton (grooming mode)
        ! note: for tagging mode, this would be vetoed.
        return_merges(:, 1) = (/jet, jet/)
        return
      else
        do i = m, 1, -1
          ! go through previous merges and keep only those which contribute to clustered jet
          if (jet_p_ind_func(jet_pass(1, i)) .eq. 1) then
            jet_p_ind_func(jet_pass(2, i)) = 1
          else if (jet_p_ind_func(jet_pass(2, i)) .eq. 1) then
            jet_p_ind_func(jet_pass(1, i)) = 1
          end if
        end do
      end if
      j = 0
      do i = 1, np_
        if ((jet_p_ind_func(jet_pass(1, i)) .eq. 1) .and. (jet_p_ind_func(jet_pass(2, i)) .eq. 1)) then
          j = j + 1
          return_merges(1, j) = jet_pass(1, i)
          return_merges(2, j) = jet_pass(2, i)
        end if
      end do

    end function return_merges

    ! returns index of jet into which parton i is clustered, using cluster_Hist array
    pure integer function parent_jet(cluster_Hist, i)
      integer, dimension(:), intent(in) :: cluster_Hist
      integer, intent(in) :: i
      integer :: j, k, n

      n = size(cluster_Hist)

!      if ((i .lt. 1) .or. (i .gt. n)) then
!       ! 'error stop' in pure subroutine apparently only valid in Fortran 2015 or later; removed for compatibility
!         error stop "parent_jet: invalid particle index"
!      end if

      j = i
      k = cluster_Hist(i)

      do while (k .ne. j)
        j = k
        k = cluster_Hist(j)
      end do

      parent_jet = k

    end function parent_jet

    ! returns array containing indices of partons clustered into jet i
    pure function partons_in_jet(cluster_Hist, i, nproto)
      integer, dimension(:), intent(in) :: cluster_Hist
      integer, intent(in) :: i, nproto
      integer :: j, k, np_in_jet
      integer, dimension(:), allocatable :: pij_indicator_func
      integer, dimension(:), allocatable :: partons_in_jet

      allocate (pij_indicator_func(nproto))
      pij_indicator_func = 0
      np_in_jet = 0
      do j = 1, nproto
        if (parent_jet(cluster_Hist, j) .eq. i) then
          pij_indicator_func(j) = 1
          np_in_jet = np_in_jet + 1
        end if
      end do
      allocate (partons_in_jet(np_in_jet))
      partons_in_jet = 0
      k = 0

      do j = 1, nproto
        if (pij_indicator_func(j) == 1) then
          k = k + 1
          partons_in_jet(k) = j
        end if
      end do
    end function partons_in_jet

  end subroutine recluster_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply the JADE/Durham recombination
  !
  !> @param[in] ycut optional argument to override the ycut value
  !> @param[in] compute_ytrans optional argument to switch to the computation
  !>            of the ycut transition values y_{n n+1}
  !>            the reconstructed jet momenta correspond to ycut --> infty
  !-----------------------------------------------------------------------------
  subroutine cluster_jade_jet(ycut, compute_ytrans)
    use Process_mod, only: nPDFs_proc
    use DIS_mod
    real(kind=wp), optional, intent(in) :: ycut
    logical, optional, intent(in) :: compute_ytrans
    real(kind=wp) :: ycut_
    logical :: qytrans, qdone
    !>-----
    real(kind=wp) :: dij_min, dij_last
    integer :: ij_min(2)
    integer :: i, j, k
    !> sum up all 4-momenta of particles involved in the clustering to
    !> determine the normalisation in the distance measure ( = minv)
    real(kind=wp), dimension(4) :: psum
    real(kind=wp) :: M2_norm
    !> DIS using JADE/Durham needs another pseudo-particle
    type(Kin_t), pointer :: kd
    real(kind=wp) :: pL_miss
    real(kind=wp), dimension(4) :: prem
    !> in DIS JADE/Durham we want to ignore "+1" (remnant) in the counting
    integer :: count_offset

    if (present(ycut)) then
      ycut_ = ycut
    else
      ycut_ = jet_ycut
    end if

    if (present(compute_ytrans)) then
      qytrans = compute_ytrans
    else
      qytrans = .false.  ! default: normal clustering till ycut_
    end if

    !!! print*, "cluster_jade_jet: ", ycut_, qytrans

    count_offset = 0
    !> DIS with the JADE/Durham algorithm registers a pseudo-particle
    !> which goes along the beam direction (proton remnant)
    if (nPDFs_proc == 1) then
      !> don't count the remnant in the observables (y45,...)
      count_offset = 1
      !> prettify code
      kd => kin(npar_jet)
      !> determine missing long. momentum of the event
      pL_miss = getEp() - getEe()  ! incoming proton & electrom
      do i = 3, npar_jet
        pL_miss = pL_miss - kd%p(3, i)  ! all outgoing particles
      end do
      !> set up the momentum
      prem(:) = 0._wp
      prem(3) = pL_miss       ! z direction
      prem(4) = abs(pL_miss)  ! energy
      !> register as a proto-jet
      nproto_jet = nproto_jet + 1
      jets(nproto_jet)%isJet = -1 ! init as proto-jet
      jets(nproto_jet)%p(:) = prem(:)
      jets(nproto_jet)%pt2 = v1_pt2(prem(:))
      jets(nproto_jet)%y = v1_y(prem(:))
      jets(nproto_jet)%phi = v1_phi(prem(:))

      ! print*, "check momentum"
      ! call print_momenta(npar_jet)
      ! print*, "x1_kin   ", x1_kin
      ! print*, "parton   ", kd%p(:,1)
      ! print*, "proton   ", 0._wp, 0._wp,  getEp(), getEp()
      ! print*, "         ", kd%p(:,1) / x1_kin
      ! print*, "electron ", 0._wp, 0._wp, -getEe(), getEe()
      ! print*, "         ", kd%p(:,2)
      ! print*, "remnant  ", prem
      ! print*, '# partons', nproto_jet
      ! do k=1,nproto_jet
      !   print*, ">", k, jets(k)%isJet, jets(k)%p(:)
      ! end do
    end if

    !> get the invariant mass of the particles involved in the clustering
    psum = 0._wp
    do i = 1, nproto_jet
      psum(:) = psum(:) + jets(i)%p(:)
    end do
    M2_norm = v1_minv2(psum)
    ! !> for DIS, this is the same as W^2
    ! !> for epem, this is the same as roots^2
    ! print*, "M2_norm", M2_norm
    ! print*, "DIS_W2 ", getManual_obs(id_DIS_W2)
    ! read(*,*)
    !> init the ycut transitions
    !> everything must be valid all the time... default transition = 0
    !> better solution requires some major rewrite in how Observables are treated
    if (qytrans) then
      ! print*,
      ! print*, "------------------------ cluster_jade_jet", M2_norm
      value_y45 = 0._wp
      valid_y45 = .false.
      value_y34 = 0._wp
      valid_y34 = .false.
      value_y23 = 0._wp
      valid_y23 = .false.
      value_y12 = 0._wp
      valid_y12 = .false.
      value_y01 = 0._wp
      valid_y01 = .false.
    end if

    !> init distance measures & clustering history
    do i = 1, nproto_jet
      clusterHist_jet(i) = i
      !print*, "|", i, jets(i)%isJet
      !print*, "|", jets(i)%p(:)
      do j = i + 1, nproto_jet
        dij(i, j) = del_ij(jets(i), jets(j))
        !print*, "# dij", i, j, dij(i,j)
      end do
    end do

    !> perform the recombination algorithm
    dij_last = -1._wp
    qdone = .false.
    do k = 1, nproto_jet

      !> find the minimum distance
      dij_min = -1._wp
      ij_min = -1
      do i = 1, nproto_jet
        if (.not. jets(i)%isJet < 0) cycle ! not a proto-jet
        do j = i + 1, nproto_jet
          if (.not. jets(j)%isJet < 0) cycle ! not a proto-jet
          if ((dij(i, j) < dij_min) .or. (dij_min < 0._wp)) then
            dij_min = dij(i, j)
            ij_min = (/i, j/)
          end if
        end do
      end do

      !> we're done: nothing left to recombine
      !> for the transitions, we cannot exit yet
      if (dij_min < 0._wp) qdone = .true.

      !> set the ycut transitions
      if (qytrans) then
        !> dij_min cannot become smaller than the value of the last clustering
        if (dij_min > dij_last) then
          dij_last = dij_min
        else
          dij_min = dij_last
        end if
        select case (nproto_jet - k - count_offset)
        case (4)
          value_y45 = dij_min
          valid_y45 = .true.
          !!! print*, "y45", dij_min, "[",k,"/",nproto_jet,"(",count_offset,")]"
        case (3)
          value_y34 = dij_min
          valid_y34 = .true.
          !!! print*, "y34", dij_min, "[",k,"/",nproto_jet,"(",count_offset,")]"
        case (2)
          value_y23 = dij_min
          valid_y23 = .true.
          !!! print*, "y23", dij_min, "[",k,"/",nproto_jet,"(",count_offset,")]"
        case (1)
          value_y12 = dij_min
          valid_y12 = .true.
          !!! print*, "y12", dij_min, "[",k,"/",nproto_jet,"(",count_offset,")]"
        case (0)
          value_y01 = dij_min
          valid_y01 = .true.
          !!! print*, "y01", dij_min, "[",k,"/",nproto_jet,"(",count_offset,")]"
        end select
        if (qdone) cycle
      end if

      !> if we're doing a "normal" clustering, check for termination
      if (qdone) exit
      if (.not. qytrans .and. (dij_min > ycut_)) then
        exit  ! the recombination do-loop
      end if

      !> merge smallest dij
      i = ij_min(1)
      j = ij_min(2)
      call merge_jet(jets(i), jets(j))
      !!! print*, "merge", j, "into", i, "| dij", dij_min
      ! print*, ">", i, jets(i)%isJet, jets(i)%p(:)
      !> store clustering
      clusterHist_jet(j) = i  ! `j` merged into `i`
      do j = 1, nproto_jet
        if (.not. jets(j)%isJet < 0) cycle ! not a proto-jet
        if (j == i) cycle
        if (i < j) dij(i, j) = del_ij(jets(i), jets(j))
        if (i > j) dij(j, i) = del_ij(jets(i), jets(j))
        ! print*, "$ dij", min(i,j), max(i,j), dij(min(i,j),max(i,j))
      end do
    end do  ! k=1,nproto_jet

    !> the clustering is complete => all proto-jets are jets
    njets_jet = 0
    do i = 1, nproto_jet
      if (.not. jets(i)%isJet < 0) cycle  ! not a proto-jet
      jets(i)%isJet = 1
      njets_jet = njets_jet + 1
    end do

    ! !> DEBUG consistency check
    ! if ( qytrans ) then
    !   if ( (value_y01 < value_y12) .or.  &
    !      & (value_y12 < value_y23) .or.  &
    !      & (value_y23 < value_y34) .or.  &
    !      & (value_y34 < value_y45) ) then
    !     print*, "cluster_jade_jet: hierarchy wrong:"
    !     print*, "> y01", value_y01
    !     print*, "> y12", value_y12
    !     print*, "> y23", value_y23
    !     print*, "> y34", value_y34
    !     print*, "> y45", value_y45
    !     stop
    !   end if
    ! end if

    !> DIS with JADE/Durham algorithm:
    !> we should remove the jet associated with the proton remnant
    !> this is the "+1" in the DIS "(n+1) jet" event which we don't count
    if (nPDFs_proc == 1) then
      ! find out where the entry `nproto_jet` (slot for the remnant) was clustered
      j = nproto_jet
      do while (jets(j)%isJet == 0)
        j = clusterHist_jet(j)
      end do
      ! eliminate the beam jet
      jets(j)%isJet = 0
      njets_jet = njets_jet - 1
    end if

    return

  contains  ! distance measures for jade/durham

    pure real(kind=wp) function del_ij(j1, j2)
      use Constants_mod
      type(Jet_t), intent(in) :: j1, j2
      real(kind=wp) :: Ei, Ej, cos_ij
      Ei = j1%p(4)
      Ej = j2%p(4)
      cos_ij = dot_eucl(j1%p(1:3), j2%p(1:3)) &
               /sqrt(dot_eucl(j1%p(1:3), j1%p(1:3))) &
               /sqrt(dot_eucl(j2%p(1:3), j2%p(1:3)))
      if (jetalg == 4) then
        del_ij = 2._wp*Ei*Ej*(1._wp - cos_ij)/M2_norm
      elseif (jetalg == 6) then
        del_ij = 2._wp*min(Ei, Ej)**2*(1._wp - cos_ij)/M2_norm
      else
        del_ij = 0._wp
      end if
      if (del_ij < 0._wp) del_ij = 0._wp  ! numerical stability guard
    end function del_ij

  end subroutine cluster_jade_jet

  subroutine cluster_flavjet(Rsq, a)
    use Flavour_mod, only: qKeepEvenTagged_flav
    real(8), optional, intent(in) :: RSq, a
    real(8) :: RSq_val, diB_min, dij_min, alpha, a_param
    real(8) :: pt2_max
    integer :: iB_min, ij_min(2)
    integer :: i, j, k, l

    !> Optional input for jet cone size
    if (present(Rsq)) then
      RSq_val = RSq
    else
      RSq_val = RconeSq
    end if

    !> Flavour-kt parameter
    if (present(a)) then
      alpha = a
    else
      alpha = flavkt_alpha  ! default value
    end if

    !> Flavour anti-kt 'a' parameter
    if (present(a)) then
      a_param = a
    else
      a_param = flavantikt_a  ! default value
    end if

    !> Initiate distance measures & clustering history
    pt2_max = 0._wp  ! keep track of the highest pt^2 before clustering
    do i = 1, nproto_jet
      if (i .eq. 1) then
        do l = 1, nproto_jet
          if (pt2_max < flavjets(l)%pt2) pt2_max = flavjets(l)%pt2
        end do
      end if
      clusterHist_flavjet(i) = i
      diB(i) = del_iB(flavjets(i))
      do j = i + 1, nproto_jet
        dij(i, j) = del_ij(flavjets(i), flavjets(j))
      end do
    end do
    pt2_max = 0._wp !(flavantikt) restart to take a new value in the next clustering step

    !> perform the recombination algorithm
    njets_flavjet = 0
    flav_njets_flavjet = 0

    do k = 1, nproto_jet ! loop: maximum possible recombinations
      !> find minimum
      dij_min = -1._wp
      diB_min = -1._wp
      ij_min = -1
      iB_min = -1

      do i = 1, nproto_jet
        if (.not. flavjets(i)%isJet < 0) cycle ! not a proto-jet, next one!

        !> Find minimal beam distance
        if (diB(i) < diB_min .or. diB_min < 0._wp) then
          iB_min = i
          diB_min = diB(i)
        end if

        !> Find minimal distance between proto-jets
        do j = i + 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle ! not a proto-jet, next one!
          if (dij(i, j) < dij_min .or. dij_min < 0._wp) then
            dij_min = dij(i, j)
            ij_min = (/i, j/)
          end if
        end do
      end do

      !> Merge proto-jets with each other, or declare proto-jet as jet!
      if (dij_min >= 0._wp .and. dij_min < diB_min) then
        !> dij is the smallest -> merge the two jets ('ij_min(2)' into 'ij_min(1)')
        call merge_flav(ij_min(1), ij_min(2))
        ! store clustering
        clusterHist_flavjet(ij_min(2)) = ij_min(1)
        ! update distance measures
        ! all beam distance measures changed -> put them in the loop too
        do i = 1, nproto_jet
          if (i .eq. 1) then
            do l = 1, nproto_jet
              if (.not. flavjets(l)%isJet < 0) cycle ! not a proto-jet, next one!
              if (pt2_max < flavjets(l)%pt2) pt2_max = flavjets(l)%pt2
            end do
          end if
          if (.not. flavjets(i)%isJet < 0) cycle
          diB(i) = del_iB(flavjets(i))
          do j = i + 1, nproto_jet
            dij(i, j) = del_ij(flavjets(i), flavjets(j))
          end do
        end do
        pt2_max = 0._wp
      else
        !> beam distance is the smallest -> declare as jet
        flavjets(iB_min)%isJet = 1
        njets_flavjet = njets_flavjet + 1
        if (flavjets(iB_min)%isFlavoured) flav_njets_flavjet = flav_njets_flavjet + 1
        !> if final-state jets not included in beam distance, we need to update distances
        !> and this has to be done after iB_min has been declared as jet
        if (flavkt_only_proto) then
          do i = 1, nproto_jet
            if (.not. flavjets(i)%isJet < 0) cycle
            diB(i) = del_iB(flavjets(i))
          end do
        end if
      end if

    end do

  contains

    real(kind=wp) function del_iB(fj)
      type(Jet_t), intent(in) :: fj
      del_iB = -1._wp
      select case (jetalg)
      case (5)  ! flavour kt
        del_iB = del_iB_fkt(fj)
      case (7)  ! flavour anti-kt
        del_iB = del_iB_fakt(fj)
      end select
    end function del_iB

    real(kind=wp) function del_ij(fj1, fj2)
      use Constants_mod, only: pi
      type(Jet_t), intent(in) :: fj1, fj2
      del_ij = -1._wp
      select case (jetalg)
      case (5)  ! flavour kt
        del_ij = del_ij_fkt(fj1, fj2)
      case (7)  ! flavour anti-kt
        del_ij = del_ij_fakt(fj1, fj2)
      end select
    end function del_ij

    !>---------------------------------
    !>  distance measures for flavour kt
    !>---------------------------------

    !> Flavour-kt beam measure
    real(8) function del_iB_fkt(fj)
      type(Jet_t), intent(in) :: fj
      if (fj%isFlavoured) then
        !> fj is flavoured
        if (alpha /= 2._wp) then
          del_iB_fkt = min(ptB(fj), ptBb(fj))**alpha*fj%pt2**(1._wp - alpha/2._wp)
        else
          del_iB_fkt = min(ptB(fj), ptBb(fj))**2
        end if
      else
        !> fj is not flavoured
        del_iB_fkt = fj%pt2
      end if

    end function del_iB_fkt

    !> beam scale for flavour-kt
    real(kind=wp) function ptB(fj)
      type(Jet_t), intent(in) :: fj
      integer :: i
      real(8) :: boost_fac

      ptB = 0._wp

      !> first the QCD partons
      do i = 1, nproto_jet
        if (flavkt_only_proto) then
          if (flavjets(i)%isJet >= 0) cycle  ! proto-jets (<0) only
        else
          if (flavjets(i)%isJet == 0) cycle  ! both jets & proto-jets
        end if
        if (fj%y > flavjets(i)%y) then
          boost_fac = exp(flavjets(i)%y - fj%y)
        else
          boost_fac = 1._wp
        end if
        ptB = ptB + sqrt(flavjets(i)%pt2)*boost_fac
      end do

      !> do not include non-QCD particles
      if (.not. flavkt_include_colourless) return

      if (flavkt_sum_nonqcd) then
        !> sum nonQCD particles
        if (fj%y > nonQCDsum%y) then
          boost_fac = exp(nonQCDsum%y - fj%y)
        else
          boost_fac = 1._wp
        end if
        !> use ET
        ptB = ptB + sqrt(nonQCDsum%pt2 + nonQCDsum%m2)*boost_fac
      else
        !> include them individually
        do i = 1, nonQCDvecsize
          if (fj%y > nonQCDvec(i)%y) then
            boost_fac = exp(nonQCDvec(i)%y - fj%y)
          else
            boost_fac = 1._wp
          end if
          ptB = ptB + sqrt(nonQCDvec(i)%pt2)*boost_fac
        end do
      end if

    end function ptB

    real(kind=wp) function ptBb(fj)
      type(Jet_t), intent(in) :: fj
      integer :: i
      real(8) :: boost_fac

      ptBb = 0._wp

      !> first the QCD partons
      do i = 1, nproto_jet
        if (flavkt_only_proto) then
          if (flavjets(i)%isJet >= 0) cycle  ! proto-jets (<0) only
        else
          if (flavjets(i)%isJet == 0) cycle  ! both jets & proto-jets
        end if
        if (flavjets(i)%y > fj%y) then
          boost_fac = exp(fj%y - flavjets(i)%y)
        else
          boost_fac = 1._wp
        end if
        ptBb = ptBb + sqrt(flavjets(i)%pt2)*boost_fac
      end do

      !> do not include non-QCD particles if alt distance
      if (.not. flavkt_include_colourless) return

      if (flavkt_sum_nonqcd) then
        !> sum nonQCD particles
        if (nonQCDsum%y > fj%y) then
          boost_fac = exp(fj%y - nonQCDsum%y)
        else
          boost_fac = 1._wp
        end if
        !> use ET
        ptBb = ptBb + sqrt(nonQCDsum%pt2 + nonQCDsum%m2)*boost_fac
      else
        !> include them individually
        do i = 1, nonQCDvecsize
          if (nonQCDvec(i)%y > fj%y) then
            boost_fac = exp(fj%y - nonQCDvec(i)%y)
          else
            boost_fac = 1._wp
          end if
          ptBb = ptBb + sqrt(nonQCDvec(i)%pt2)*boost_fac
        end do
      end if

    end function ptBb

    !> Flavour-kt distance measure
    real(8) function del_ij_fkt(fj1, fj2)
      use Constants_mod, only: pi
      type(Jet_t), intent(in) :: fj1, fj2
      real(8) :: del_phi, del_y, deltaRSq

      del_y = fj1%y - fj2%y
      del_phi = abs(fj1%phi - fj2%phi)
      if (del_phi > pi) del_phi = 2._wp*pi - del_phi
      deltaRSq = del_y**2 + del_phi**2

      if ((fj1%pt2 < fj2%pt2 .and. fj1%isFlavoured) .or. &
          (fj2%pt2 < fj1%pt2 .and. fj2%isFlavoured)) then
        !> Softer of fj1 and fj2 is flavoured
        if (alpha /= 2._wp) then
          del_ij_fkt = max(fj1%pt2, fj2%pt2)**(alpha/2._wp)* &
                       min(fj1%pt2, fj2%pt2)**(1._wp - alpha/2._wp)
        else ! alpha = 2
          del_ij_fkt = max(fj1%pt2, fj2%pt2)
        end if
      else
        !> Softer of fj1 and fj2 is flavourless
        del_ij_fkt = min(fj1%pt2, fj2%pt2)
      end if
      del_ij_fkt = del_ij_fkt*(deltaRSq/Rsq_val)

    end function del_ij_fkt

    !>---------------------------------
    !>  distance measures for flavour anti-kt
    !>---------------------------------

    !> normal kt-type beam distance
    real(8) function del_iB_fakt(fj)
      type(Jet_t), intent(in) :: fj
      del_iB_fakt = fj%pt2**(ip_jet)
    end function del_iB_fakt

    real(8) function del_ij_fakt(fj1, fj2)
      use Constants_mod, only: pi, twopi
      type(Jet_t), intent(in) :: fj1, fj2
      real(kind=wp) :: del_y, del_phi, deltaRSq
      real(kind=wp) :: Sfac, x
      del_y = fj1%y - fj2%y
      del_phi = abs(fj1%phi - fj2%phi)
      if (del_phi .gt. pi) del_phi = twopi - del_phi
      deltaRSq = del_y**2 + del_phi**2
      del_ij_fakt = min(fj1%pt2**(ip_jet), fj2%pt2**(ip_jet))*(deltaRSq/RSq_val)
      !> if the pair is flavoured, dress with a soft factor
      if (fj1%isFlavoured .and. fj2%isFlavoured) then
        x = (fj1%pt2 + fj2%pt2)/(2._wp*a_param*pt2_max)
        Sfac = 1._wp
        if (x < 1._wp) Sfac = 1._wp - cos(0.5_wp*pi*x)
        if (qKeepEvenTagged_flav()) then
          if ((fj1%nflav_q + fj2%nflav_q) == (fj1%nflav_qb + fj2%nflav_qb)) &
            del_ij_fakt = del_ij_fakt*Sfac
        else
          del_ij_fakt = del_ij_fakt*Sfac
        end if
      end if
    end function del_ij_fakt

    !> Proto-jet merger routine
    subroutine merge_flav(i, j)
      use Flavour_mod, only: qKeepEvenTagged_flav
      integer, intent(in) :: i, j
      integer :: k

      !> update kinematics
      flavjets(i)%p = flavjets(i)%p + flavjets(j)%p
      flavjets(i)%pt2 = v1_pt2(flavjets(i)%p)
      flavjets(i)%y = v1_y(flavjets(i)%p)
      flavjets(i)%phi = v1_phi(flavjets(i)%p)

      !> Add flavour-tagged quark numbers
      flavjets(i)%nflav_q = flavjets(i)%nflav_q + flavjets(j)%nflav_q
      flavjets(i)%nflav_qb = flavjets(i)%nflav_qb + flavjets(j)%nflav_qb

      !> Is the resulting jet tagged?
      !> isFlavoured = .true.  <--> odd number of flavoured quarks
      !> isFlavoured = .false. <--> even number of flavoured quarks
      if (modulo(flavjets(i)%nflav_q + flavjets(i)%nflav_qb, 2) == 0) then
        flavjets(i)%isFlavoured = .false.
      else
        flavjets(i)%isFlavoured = .true.
      end if

      !> Or use regular flavour-kt instead of even-tag exclusion?
      if (qKeepEvenTagged_flav()) then
        if (flavjets(i)%nflav_q /= flavjets(i)%nflav_qb) then
          flavjets(i)%isFlavoured = .true.
        else
          flavjets(i)%isFlavoured = .false.
        end if
      end if

      !> Eliminate jet 'j' that has been merged into 'i'
      flavjets(j)%isJet = 0

    end subroutine merge_flav

  end subroutine cluster_flavjet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply jet cuts to a single jet four momentum
  !
  !> @param[in] is_flavoured optional argument to tag flavoured jet
  !-----------------------------------------------------------------------------
  function qpass_mom_jet(v, is_flavoured)
    use Flavour_mod, only: isActive_flav
    real(kind=wp), intent(in) :: v(1:4)
    logical, intent(in), optional :: is_flavoured
    logical :: qpass_mom_jet
    integer :: i, k
    real(kind=wp) :: obs_val

    qpass_mom_jet = .false.

    do k = 1, n_obsj
      if (.not. list_obsj(k)%qactive) cycle
      if ( list_obsj(k)%qflavobs ) then
        if ( present(is_flavoured) ) then
          if (.not.is_flavoured) cycle
        else
          cycle
        end if
      end if
      if (isExtraj_obs(k)) stop "qpass_mom_jet: extra not supported yet!"

      obs_val = eval_obsj(k, npar_jet, v)
      ! print*, "qpass_mom_jet: obs_val [",k,"]", obs_val
      if (.not. qpass_jet(k, obs_val)) then
        ! print*, "qpass_mom_jet: kill", i, "[",k,"]", njets_jet, "->", (njets_jet - 1)
        return
      end if
    end do  ! k = 1, n_obsj

    qpass_mom_jet = .true.

  end function qpass_mom_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply jet cuts (rejecting jets not the event!)
  !
  !> @param[in] skipFlav optional argument to skip flavour-sensitive cuts
  !-----------------------------------------------------------------------------
  subroutine applyCuts_jet(skipFlav)
    use Flavour_mod, only: isActive_flav
    logical, optional :: skipFlav
    logical :: skipFlav_
    integer :: i, k
    real(kind=wp) :: obs_val

    skipFlav_ = .false.  ! default
    if (present(skipFlav)) skipFlav_ = skipFlav

    do i = 1, nproto_jet
      if (.not. jets(i)%isJet > 0) cycle ! not a jet
      do k = 1, n_obsj
        if (.not. list_obsj(k)%qactive) cycle
        if (list_obsj(k)%qflavobs .and. &
            (skipFlav_ .or. .not. jets(i)%isFlavoured)) cycle
        if (isExtraj_obs(k)) cycle  ! extra cuts are applied *manually*
        ! if ( .not. list_obsj(k)%qactive ) then
        !   print*, "applyCuts_jet: inactive [",k,"]", njets_jet
        !   cycle
        ! end if
        ! if ( isExtraj_obs(k) ) then
        !   print*, "applyCuts_jet: extra [",k,"]", njets_jet
        !   cycle
        ! end if
        obs_val = eval_obsj(k, npar_jet, jets(i)%p)
        ! print*, "applyCuts_jet: obs_val [",k,"]", obs_val
        if (.not. qpass_jet(k, obs_val)) then
          ! print*, "applyCuts_jet: kill", i, "[",k,"]", njets_jet, "->", (njets_jet - 1)
          jets(i)%isJet = 0
          njets_jet = njets_jet - 1
          if (isActive_flav()) then
            if (jets(i)%isFlavoured) flav_njets_jet = flav_njets_jet - 1
          end if

          exit
        end if
      end do
    end do

  end subroutine applyCuts_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Sort the jets in descending order of pt
  !> result of the sorting is stored in the "pt_sort_jet" array:
  !> jets( pt_sort_jet(1) ) = leading jet, ...
  !-----------------------------------------------------------------------------
  subroutine sort_jet()
    integer :: i, k
    real(kind=wp) :: pt2_max
    integer :: i_max

    pt_sort_jet = 0
    do k = 1, njets_jet
      ! find highest pt
      pt2_max = -1._wp
      i_max = -1
      do i = 1, nproto_jet
        ! not a jet or already written out
        if (.not. jets(i)%isJet > 0 .or. jets(i)%isJet == 2) cycle
        if (jets(i)%pt2 > pt2_max) then
          pt2_max = jets(i)%pt2
          i_max = i
        end if
      end do
      ! write out
      !print*, "sort:", k, i_max, pt2_max
      jets(i_max)%isJet = 2  ! flag with "2" for sorted jets
      pt_sort_jet(k) = i_max
    end do

  end subroutine sort_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> print out a list of active observables
  !-----------------------------------------------------------------------------
  subroutine digestActive_jet(unit)
    use Flavour_mod, only: isActive_flav
    integer, optional, intent(in) :: unit
    integer :: i
    integer :: iout
    character(len=*), parameter :: fmt_str = '(1x,"*",3x,A,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    end if

    write (iout, fmt_str) ""
    write (iout, fmt_str) "> jet acceptance cuts <  "
    write (iout, fmt_str) ""
    do i = 1, n_obsj
      if (.not. list_obsj(i)%qactive) cycle
      if ((.not. isActive_flav()) .and. list_obsj(i)%qflavobs) cycle
      write (iout, fmt_str) list_obsj(i)%name
    end do

  end subroutine digestActive_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply the jet recombination and store the jets in descending order of
  !> transverse momentum in kd%pjets.
  !
  !> @param[in] npar kinematics specified by the number of particles
  !> @param[in] skipInit optional flag to skil initialisation `initRecomb_jet`
  !>            This ensures that `njets_jet` is not reset and subsequent
  !>            `cluster_jet` calls have no side effect (needed for democratic)
  !-----------------------------------------------------------------------------
  recursive subroutine recombine_jet(npar, skipInit)
    use Flavour_mod, only: isActive_flav
    integer, intent(in) :: npar
    logical, optional, intent(in) :: skipInit
    logical :: skipInit_
    type(Kin_t), pointer :: kd
    integer :: i, j, f, f_id, ipart
    !>----- for special 2->2 config in dijets
    real(kind=wp) :: pt2j1, pt2j2, del_pt2_abs, del_pt2_rel
    real(kind=wp), parameter :: threshold_del_pt2_abs = 1e-3_wp**2  ! GeV**2
    real(kind=wp), parameter :: threshold_del_pt2_rel = 1e-7_wp
    integer :: npar_copy
    integer :: test_fj, test_mj

    ! skipInit_ =.false.
    skipInit_ = qskip_init_jet
    qskip_init_jet = .false.
    if (present(skipInit)) skipInit_ = skipInit

    !> prettify code
    kd => kin(npar)

    !> recombination has been performed already in the past
    ! if ( kd%qcachedObs .and. kd%njets >= 0 ) then
    if (kd%njets >= 0) then
      return
    end if

    !> no jet algo
    if (jetalg == 0) then
      kd%njets = 0
      return
    end if

    if ((.not. skipInit_) .and. (.not. softdrop)) then
      call initRecomb_jet(npar)
    end if

    !> cluster normal jets
    select case (jetalg)
    case (4, 6)
      !> JADE/Durham
      call cluster_jade_jet()
    case (5)
      !> flavour kt
      if (.not. isActive_flav()) then
        stop "recombine_jet: tag a flavour for using the flavour-kt algorithm"
      end if
      call cluster_flavjet()
    case (7)
      !> flavour anti-kt
      if (.not. isActive_flav()) then
        stop "recombine_jet: flavour anti-kt algorithm, tag a flavour or use normal anti-kt"
      end if
      call cluster_flavjet()
    case default
      call cluster_jet()
    end select

    if (isActive_flav()) then
      if (.not. flav_dress_cluster_qproto) then
        !> only retain resolved (flavour-insensitive) jets for the dressing
        call applyCuts_jet(skipFlav=.true.)
      end if
      call flav_dress_jet(npar)
    end if
    call applyCuts_jet()
    call sort_jet()

    !------ WRITE OUT INTO KINDATA (need to do this before any matching)
    kd%RconeSq = RconeSq
    kd%njets = njets_jet
    kd%qflavjet = .false.
    if (isActive_flav()) kd%flav_njets = flav_njets_jet
    do i = 1, njets_jet
      kd%pjets(:, i) = jets(pt_sort_jet(i))%p(:)
      if (isActive_flav()) then
        if (jets(pt_sort_jet(i))%isFlavoured) then
          kd%qflavjet(i) = .true.
        end if
        kd%sumqflavjet(i) = jets(pt_sort_jet(i))%nflav_q - jets(pt_sort_jet(i))%nflav_qb
      end if
    end do

    !------ treat ambiguous 2->2(+soft) configuration where pt2j1~pt2j2
    ! if ( njets_jet >= 2 ) then
    ! IR-safe jet algorithm  should always cluster down to two jets
    ! in the unresolved limits where cancellations are relevant
    if (njets_jet == 2) then

      ! note: these are pt^2 [GeV^2]
      pt2j1 = jets(pt_sort_jet(1))%pt2
      pt2j2 = jets(pt_sort_jet(2))%pt2
      del_pt2_abs = abs(pt2j1 - pt2j2)
      del_pt2_rel = del_pt2_abs/abs(pt2j1 + pt2j2)
      if (del_pt2_abs > threshold_del_pt2_abs .or. &
          del_pt2_rel > threshold_del_pt2_rel) then
        return
      end if

      ! !_______ DEBUG
      ! call randJJ(kd)
      ! return
      ! !_______ DEBUG

      !-- try resolving the ambiguity to match to the higher momentum set
      npar_copy = kd%npar_from

      if (npar_copy <= kd%npar) then
        ! no mapping: out of options -> randomize
        ! if (del_pt2_abs == 0._wp) call randJJ
        call randJJ
        return
      end if

      !> we're in a subtraction term
      !>  => determine assignment from real-emission kinemtics
      !> make sure clustering was performed in the higher momentum set
      call recombine_jet(npar_copy)
      if (kin(npar_copy)%njets == 0) then
        ! if ( kin(npar_copy)%njets /= 2 ) then
        !> in an unresolved limit (where cancellations are relevant)
        !> njets must match, i.e. must be two
        !> if the higher momentum set has a different # of jets
        !> we can just ramdomise the assignment
        ! if (del_pt2_abs == 0._wp) call randJJ
        call randJJ
      else if (deltaJJ(kd%pjets(:, 1), kin(npar_copy)%pjets(:, 1)) &
               > deltaJJ(kd%pjets(:, 2), kin(npar_copy)%pjets(:, 1))) then
        !> align the leading jets
        call flipJJ
      end if

    end if

    return

  contains ! functions local to `recombine_jet`

    !> funtion to measure "closeness" of jet momenta (for matching)
    pure real(kind=wp) function deltaJJ(p1, p2)
      real(kind=wp), dimension(4), intent(in) :: p1, p2
      integer :: i

      ! deltaJJ = 0._wp
      ! do i=1,4
      !   deltaJJ = deltaJJ + (p1(i)-p2(i))**2
      ! end do

      deltaJJ = v2_delta_R(p1, p2)

      ! deltaJJ = 0._wp
      ! do i=1,3
      !   deltaJJ = deltaJJ + (p1(i)-p2(i))**2
      ! end do

      !deltaJJ = abs( dot(p1,p2) )

    end function deltaJJ

    !> routine to flip j1,j2
    subroutine flipJJ
      real(kind=wp) :: pi
      integer :: i
      !print*, "FLIP"
      do i = 1, 4
        pi = kd%pjets(i, 1)
        kd%pjets(i, 1) = kd%pjets(i, 2)
        kd%pjets(i, 2) = pi
      end do
    end subroutine flipJJ

    !> routine to randomise j1,j2
    subroutine randJJ
      real(kind=wp) :: rnd
      call random_number(rnd)
      !print*, "RAND"
      !> default assignment is deterministic: j1=3, j2=4
      !> flip the two with a 50% probability
      if (rnd < 0.5_wp) call flipJJ
    end subroutine randJJ

  end subroutine recombine_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Apply the democratic jet clustering
  !
  !> @param[in] npar kinematics specified by the number of particles
  !> @param[in] nphotons number of photons
  !> @param[inout] pphotons momentum of photon `i` <-> pphotons(1:4,i)
  !> @param[in] zcut minimum (transverse) energy fraction for a photon:
  !>            E_T[photon] > zcut * EE_T[jet]  =>  "jet" identified as a photon
  !-----------------------------------------------------------------------------
  recursive subroutine recombine_democratic_jet(npar, nphotons, pphotons, zcut)
    use Flavour_mod, only: isActive_flav
    integer, intent(in) :: npar, nphotons
    real(kind=wp), dimension(1:4, nphotons), intent(inout) :: pphotons
    real(kind=wp), intent(in) :: zcut
    type(Kin_t), pointer :: kd
    integer, dimension(nphotons) :: idxphotons
    integer :: i, j, f, f_id, ipart

    !> prettify code
    kd => kin(npar)

    !> recombination has been performed already in the past
    ! if ( kd%qcachedObs .and. kd%njets >= 0 ) then
    if (kd%njets >= 0) then
      return
    end if

    !> no jet algo
    if ((jetalg == 0) .or. softdrop) then
      print *, "recombine_democratic_jet: invalid settings"
      print *, "jet algorithm:", jetalg
      print *, "softdrop:", softdrop
      stop
    end if

    call initRecomb_jet(npar)

    !> register all photons as additional proto-jets
    do i = 1, nphotons
      idxphotons(i) = pushProto_jet(pphotons(1:4, i))
    end do

    !> cluster normal jets
    call cluster_jet()

    call applyZcut_jet()

    !> can't call recombination because photons are not yet
    !> stored in KinData at this point! need to flag the next
    !> call to skip the initialisation instead
    qskip_init_jet = .true.
    ! call recombine_jet(npar, skipInit=.true.)

  contains

    subroutine applyZcut_jet()
      integer :: i, j, k, igam, ijet, iproto, jetroot, gamout
      integer :: jet_nphotons, jet_nparticles
      real(kind=wp), dimension(1:4) :: pgam
      real(kind=wp) :: ET_gam, ET_jet, zval

      do ijet = 1, nproto_jet
        if (jets(ijet)%isJet /= 1) cycle
        jet_nphotons = 0
        jet_nparticles = 0
        pgam(:) = 0._wp
        do iproto = 1, nproto_jet
          jetroot = iproto
          do while (.true.)
            if (clusterHist_jet(jetroot) == jetroot) exit
            jetroot = clusterHist_jet(jetroot)
          end do
          if (jetroot /= ijet) cycle
          !> `iproto` is constituten of jet `ijet`
          jet_nparticles = jet_nparticles + 1
          !> check if the particle was a photon
          do igam = 1, nphotons
            if (iproto == idxphotons(igam)) then
              jet_nphotons = jet_nphotons + 1
              pgam(1:4) = pgam(1:4) + pphotons(1:4, igam)
              exit
            end if
          end do  ! igam
        end do  ! iproto
        !> know all constituents, apply zcut
        ET_jet = v1_et(jets(ijet)%p(1:4))
        ET_gam = v1_et(pgam)
        zval = ET_gam/ET_jet
        ! print*, "# jet_nparticles:", jet_nparticles
        ! print*, "# jet_nphotons:  ", jet_nphotons
        ! print*, "# zval: ", zval, "(",ET_gam,ET_jet,")"
        if ((zval > zcut) .or. (jet_nphotons == jet_nparticles)) then
          ! print*, "IS PHOTON"
          !> reset all photons but one (= carries jet momentum)
          gamout = -1
          do igam = 1, nphotons
            jetroot = idxphotons(igam)
            do while (.true.)
              if (clusterHist_jet(jetroot) == jetroot) exit
              jetroot = clusterHist_jet(jetroot)
            end do
            if (jetroot /= ijet) cycle
            pphotons(1:4, igam) = 0._wp
            gamout = igam
          end do  ! igam
          pphotons(1:4, gamout) = jets(ijet)%p(1:4)
          !> kill the jet
          jets(ijet)%isJet = 0
          njets_jet = njets_jet - 1
        else
          ! print*, "IS JET..."
          !> reset all photons (they are part of the jet now)
          do igam = 1, nphotons
            jetroot = idxphotons(igam)
            do while (.true.)
              if (clusterHist_jet(jetroot) == jetroot) exit
              jetroot = clusterHist_jet(jetroot)
            end do
            if (jetroot /= ijet) cycle
            pphotons(1:4, igam) = 0._wp
          end do  ! igam
          !> independent of zcut (<=> used to *identify* if jet/photon)
          !> want a cut to reject jets based on ET fraction
          !> *without* identifying it as a photon
          if (.not. qpass_jet(idj_zgam, zval)) then
            ! print*, "NOT!"
            jets(ijet)%isJet = 0
            njets_jet = njets_jet - 1
          end if
        end if
      end do  ! ijet

      ! print*, "# jets surviving zcut:", njets_jet

      ! do igam=1,nphotons
      !   ET_gam = v1_et( pphotons(1:4,igam) )
      !   !> find in which jet it ended up
      !   ijet = idxphotons(igam)
      !   do while (.true.)
      !     if ( clusterHist_jet(ijet) == ijet ) exit
      !     ijet = clusterHist_jet(ijet)
      !   end do
      !   ET_jet = v1_et( jets(ijet)%p(1:4) )
      !   !> decide if it's a photon or a jet
      !   zval = ET_gam / ET_jet
      !   !> `or`: photon was never clustered with anything
      !   !>  => zval=1 => always identify as photon (no rounding errors)
      !   print*, ">", igam, ":", zval, ijet, idxphotons(igam)
      !   if ( (zval > zcut) .or. (ijet == idxphotons(igam)) ) then
      !     print*, "IS PHOTON"
      !     pphotons(1:4,igam) = jets(ijet)%p(1:4)
      !     jets(ijet)%isJet = 0
      !     njets_jet = njets_jet - 1
      !   else
      !     print*, "IS JET..."
      !     pphotons(1:4,igam) = 0._wp
      !   end if
      !   !> independent of zcut (<=> used to *identify* if jet/photon)
      !   !> want a cut to reject jets based on ET fraction
      !   !> *without* identifying it as a photon
      !   if ( .not. qpass_jet(idj_zgam,zval) ) then
      !     print*, "NOT!"
      !     jets(ijet)%isJet = 0
      !     njets_jet = njets_jet - 1
      !   end if
      ! end do  ! igam (nphotons)

    end subroutine applyZcut_jet

  end subroutine recombine_democratic_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> Dress recombined jets (any algorithm) with flavour information.
  !
  !> @param[in] npar kinematics specified by the number of particles
  !-----------------------------------------------------------------------------
  subroutine flav_dress_jet(npar)
    use Flavour_mod, only: qKeepEvenTagged_flav
    integer, intent(in) :: npar
    integer :: j

    !> no dressing for flavour-kt; simple copy
    if (jetalg == 5 .or. jetalg == 7) then
      jets = flavjets
      njets_jet = njets_flavjet
      flav_njets_jet = flav_njets_flavjet
      return
    end if

    !> reset any flavour information before dressing
    jets(:)%nflav_q = 0
    jets(:)%nflav_qb = 0
    jets(:)%isFlavoured = .false.

    if (iflav_dress < 0) then
      call softdrop_flav_dress_jet(npar)
    else if (iflav_dress > 0) then
      ! print*, "calling cluster_flav_dress_jet ", npar
      call cluster_flav_dress_jet(npar)
    else  ! iflav_dress = 0
      call naive_flav_dress_jet(npar)
    end if

    !> determine the flavour tag of the jet and count
    flav_njets_jet = 0
    do j = 1, nproto_jet
      if (jets(j)%isJet <= 0) cycle
      if (qKeepEvenTagged_flav()) then
        jets(j)%isFlavoured = (jets(j)%nflav_q /= jets(j)%nflav_qb)
      else
        jets(j)%isFlavoured = &
          (modulo(jets(j)%nflav_q + jets(j)%nflav_qb, 2) /= 0)
      end if
      !!! print*, "# jet", j, "flav?", jets(j)%isFlavoured, jets(j)%p
      if (jets(j)%isFlavoured) flav_njets_jet = flav_njets_jet + 1
    end do

  end subroutine flav_dress_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> naive flavour dressing (in general, IR *unsafe*)
  !
  !> @param[in] npar kinematics specified by the number of particles
  !-----------------------------------------------------------------------------
  subroutine naive_flav_dress_jet(npar)
    use Flavour_mod, only: iTag_flav, getId_flav, qKeepEvenTagged_flav
    integer, intent(in) :: npar
    type(Kin_t), pointer :: kd
    integer :: i, j, imom, itag

    if (iflav_dress /= 0) then
      stop "naive_flav_dress_jet: incompatible algorithm!"
    end if

    kd => kin(npar)

    do i = 1, nproto_jet
      imom = kd%ijets_lower + i - 1  ! idx in momentum set
      itag = iTag_flav(npar, imom)
      !> "This isn't the flavour you're looking for."
      if (itag == 0) cycle
      !> find the jet `j` that parton `i` was clustered into
      j = i
      do while (clusterHist_jet(j) /= j)
        j = clusterHist_jet(j)
      end do
      if (itag > 0) jets(j)%nflav_q = jets(j)%nflav_q + 1
      if (itag < 0) jets(j)%nflav_qb = jets(j)%nflav_qb + 1
    end do

  end subroutine naive_flav_dress_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> flavour dressing using recursive softdrop
  !> repurpose (*_flavjet) datastructure for declustering
  !
  !> @param[in] npar kinematics specified by the number of particles
  !-----------------------------------------------------------------------------
  subroutine softdrop_flav_dress_jet(npar)
    use Flavour_mod, only: iTag_flav, qKeepEvenTagged_flav
    integer, intent(in) :: npar
    type(Kin_t), pointer :: kd
    integer :: i, j, k, ijet, ijet_sub, imom, itag
    integer, dimension(npar) :: j_target  ! size large enough
    real(8) :: dij_min
    integer ::  ij_min(2)
    integer, dimension(npar) :: ij_seq

    if (iflav_dress >= 0) then
      stop "softdrop_flav_dress_jet: incompatible algorithm!"
    end if

    kd => kin(npar)

    !!! print*, "softdrop_flav_dress_jet:", npar

    !> figure out in which jet each partons ends up in & what flavour it has
    j_target = 0
    do i = 1, nproto_jet
      ijet = i  ! the target jet
      do while (clusterHist_jet(ijet) /= ijet)
        ijet = clusterHist_jet(ijet)
      end do
      j_target(i) = ijet  ! `i` is in jet `ijet`
    end do

    !!! print*, "> j_target", j_target

    !> the main loop over jets
    do ijet = 1, nproto_jet
      if (jets(ijet)%isJet <= 0) cycle

      !!! print*, ">> process jet", ijet

      !> init (flavjet) datastructure for jet `ijet`
      flavjets(:)%isJet = 0  ! off by default
      flavjets(:)%nflav_q = 0
      flavjets(:)%nflav_qb = 0
      flavjets(:)%isFlavoured = .false.
      do i = 1, nproto_jet
        if (j_target(i) /= ijet) cycle
        imom = kd%ijets_lower + i - 1  ! idx in momentum set
        flavjets(i)%isJet = -1  ! protojet
        flavjets(i)%p(:) = kd%p(:, imom)
        flavjets(i)%pt2 = v1_pt2(kd%p(:, imom))
        flavjets(i)%y = v1_y(kd%p(:, imom))
        flavjets(i)%phi = v1_phi(kd%p(:, imom))
        itag = iTag_flav(npar, imom)
        if (itag == 0) cycle
        !> `i` is flavoured
        if (itag > 0) flavjets(i)%nflav_q = 1
        if (itag < 0) flavjets(i)%nflav_qb = 1
        flavjets(i)%isFlavoured = .true.
      end do

      !> init distance measure & clustering history
      clusterHist_flavjet = 0  ! inactive
      dij = -1._wp  ! invalid state (don't use zero!)
      do i = 1, nproto_jet
        if (.not. flavjets(i)%isJet < 0) cycle
        clusterHist_flavjet(i) = i
        do j = i + 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle
          dij(i, j) = del_ij(flavjets(i), flavjets(j))
        end do
      end do

      !> re-cluster the jet
      ij_seq = 0  ! store the order in which the sequence occured
      do k = 1, nproto_jet
        dij_min = -1._wp
        ij_min = -1
        do i = 1, nproto_jet
          if (.not. flavjets(i)%isJet < 0) cycle
          do j = i + 1, nproto_jet
            if (.not. flavjets(j)%isJet < 0) cycle
            if ((dij(i, j) < dij_min) .or. (dij_min < 0._wp)) then
              dij_min = dij(i, j)
              ij_min = (/i, j/)
            end if
          end do
        end do
        if (dij_min < 0._wp) exit  ! done
        !!! print*, ">> cluster: ", ij_min, "| dij", dij_min
        ! merge the two jets (`j` into `i`)
        i = ij_min(1)
        j = ij_min(2)
        call merge_flav(flavjets(i), flavjets(j))
        ! store clustering
        clusterHist_flavjet(j) = i  ! 2nd argument merged into 1st
        ij_seq(k) = j  ! the one that got merged away (k-th clustering)
        ! update distance measures
        do j = 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle
          if (j == i) cycle
          if (i < j) dij(i, j) = del_ij(flavjets(i), flavjets(j))
          if (i > j) dij(j, i) = del_ij(flavjets(i), flavjets(j))
        end do
      end do  ! k proto jets
      if (count(ij_seq .ne. 0) + 1 /= count(j_target .eq. ijet)) then
        stop "softdrop_flav_dress_jet: iproto_subjet /= nproto_subjet"
      end if

      ijet_sub = 0
      do i = 1, nproto_jet
        if (clusterHist_flavjet(i) == i) then
          ijet_sub = i
          exit  ! only one jet
        end if
      end do

      !!! print*, ">> clusterHist_flavjet", clusterHist_flavjet, "|", ijet_sub

      !! all done inside the merge_flav routine!
      !! only need to count flavours at this point...
      !> recursively traverse `clusterHist_flavjet` and apply softdrop
      !> clusterHist_flavjet not enough to get relative ordering:
      !> 3->1 & 2->1, but which one was first?
      !> leaf condition: was there ever someting clustered into `i`?
      !> .not.any(clusterHist_flavjet.eq.i)

      !> copy over the flavour tag to the `jets` array for ijet
      !> the reclustered jet is where clusterHist_flavjet(I) = I
      !> where I = ijet_sub
      jets(ijet)%nflav_q = flavjets(ijet_sub)%nflav_q
      jets(ijet)%nflav_qb = flavjets(ijet_sub)%nflav_qb
      !> isFlavoured is updated outside of this routine

      !!! print*, ">> ijet [nq,nqb]", jets(ijet)%nflav_q, jets(ijet)%nflav_qb

    end do  ! ijet [jets]

    return

  contains

    !> jet distance measure for re-clustering; stripped from R_cone
    pure real(kind=wp) function del_ij(j1, j2)
      use Constants_mod
      type(Jet_t), intent(in) :: j1, j2
      real(kind=wp) :: Ei, Ej, cos_ij
      real(kind=wp) :: del_y, del_phi, deltaRSq

      !> epem distance measure
      Ei = j1%p(4)
      Ej = j2%p(4)
      cos_ij = dot_eucl(j1%p(1:3), j2%p(1:3)) &
               /sqrt(dot_eucl(j1%p(1:3), j1%p(1:3))) &
               /sqrt(dot_eucl(j2%p(1:3), j2%p(1:3)))
      del_ij = 1._wp - cos_ij  ! C/A for the moment

      ! del_y = j1%y - j2%y
      ! del_phi = abs(j1%phi - j2%phi)
      ! if(del_phi.gt.pi) del_phi = twopi-del_phi
      ! deltaRSq = del_y**2 + del_phi**2
      ! ! del_ij = min(j1%pt2**(ip_jet),j2%pt2**(ip_jet)) * deltaRSq
      ! del_ij = deltaRSq  ! C/A for the moment

    end function del_ij

    !> Proto-jet merger routine: i1 <- i2
    subroutine merge_flav(j1, j2)
      use Constants_mod
      type(Jet_t), intent(inout) :: j1, j2
      real(kind=wp) :: E1, E2, cos_ij
      real(kind=wp) :: del_y, del_phi, dR, pt_j1, pt_j2
      logical :: qSD
      integer :: k

      !> check softdrop condition
      !> epem measure
      E1 = j1%p(4)
      E2 = j2%p(4)
      cos_ij = dot_eucl(j1%p(1:3), j2%p(1:3)) &
               /sqrt(dot_eucl(j1%p(1:3), j1%p(1:3))) &
               /sqrt(dot_eucl(j2%p(1:3), j2%p(1:3)))
      dR = 1._wp - cos_ij  ! C/A for the moment
      if (flav_dress_SD_beta /= 2._wp) &
        dR = dR**(0.5_wp*flav_dress_SD_beta)
      qSD = (min(E1, E2)/(E1 + E2) > flav_dress_SD_zcut*dR)

      ! pt_j1 = sqrt(j1%pt2)
      ! pt_j2 = sqrt(j2%pt2)
      ! del_y = j1%y - j2%y
      ! del_phi = abs(j1%phi - j2%phi)
      ! if (del_phi > pi )  del_phi = twopi - del_phi
      ! dR = del_y**2 + del_phi**2
      ! if ( flav_dress_SD_beta /= 2._wp )  &
      !   &  dR = dR**(0.5_wp*flav_dress_SD_beta)
      ! qSD = ( min(pt_j1,pt_j2)/(pt_j1+pt_j2) > flav_dress_SD_zcut*dR )

      !!! print*, ">>> qSD", qSD, "|", j1%isFlavoured, j2%isFlavoured, "|", min(E1,E2)/(E1+E2)

      !> if it fails, strip off all flavours from the softer branch
      if (.not. qSD) then
        if (E1 > E2) then
          !!! print*, ">>> strip jet 2"
          j2%nflav_q = 0
          j2%nflav_qb = 0
        else
          !!! print*, ">>> strip jet 1"
          j1%nflav_q = 0
          j1%nflav_qb = 0
        end if
        !> traverse the sub-tree of `isoft` and kill flavour info
        !!!! NO NEED TO TRAVERSE. just kill flavour for this pseudo-jet !!!!
      end if

      !> just do the normal clustering as usual from here on out

      !> update kinematics
      j1%p = j1%p + j2%p
      j1%pt2 = v1_pt2(j1%p)
      j1%y = v1_y(j1%p)
      j1%phi = v1_phi(j1%p)
      !> Add flavour-tagged quark numbers
      j1%nflav_q = j1%nflav_q + j2%nflav_q
      j1%nflav_qb = j1%nflav_qb + j2%nflav_qb

      !> Eliminate jet 'j' that has been merged into 'i'
      j2%isJet = 0

    end subroutine merge_flav

  end subroutine softdrop_flav_dress_jet

  !-----------------------------------------------------------------------------
  !> @brief
  !> flavour dressing using clustering
  !> we'll repurpose the existing datastructures:
  !>   dij <-> d(fi,fj)
  !>   diB <-> d(fi,jet[fi])  <- only one jet where fi was clustered into
  !
  !> @param[in] npar kinematics specified by the number of particles
  !-----------------------------------------------------------------------------
  subroutine cluster_flav_dress_jet(npar)
    use Flavour_mod, only: iTag_flav, qKeepEvenTagged_flav
    use Process_mod, only: nPDFs_proc
    integer, intent(in) :: npar
    type(Kin_t), pointer :: kd
    integer :: i, j, k, l, ijet, jjet, ijet_sub, imom, itag
    integer, dimension(npar) :: j_target  ! size large enough
    real(8) :: diB_min, dij_min
    integer :: iB_min, ij_min(2)
    !> for wilson-line inspired penanty
    real(kind=wp) :: HTfull, pt_balance(1:2)
    !> for top-down dressing
    integer, dimension(npar) :: j_order
    real(kind=wp) :: pt2_max
    integer:: i_max
    !> for the underdress
    real(kind=wp) :: dRfj, ptf, ptj

    if (iflav_dress <= 0) then
      stop "cluster_flav_dress_jet: incompatible algorithm!"
    end if

    kd => kin(npar)

    !!! print*, "cluster_flav_dress_jet:", npar

    !> figure out in which jet each parton ends up in
    !> & copy over all flavoured particles into the (flavjet) datastructure
    j_target = 0
    flavjets(:)%isJet = 0
    flavjets(:)%nflav_q = 0
    flavjets(:)%nflav_qb = 0
    flavjets(:)%isFlavoured = .false.
    do i = 1, nproto_jet
      ijet = i  ! the target jet
      do while (clusterHist_jet(ijet) /= ijet)
        ijet = clusterHist_jet(ijet)
      end do
      j_target(i) = ijet  ! `i` is in jet `ijet`
      !> In the revised version, we use 0 to indicate "association" with the beam
      if (iflav_dress == 7) then
        if (.not. jets(ijet)%isJet > 0) then
          j_target(i) = 0
        end if
      end if
      imom = kd%ijets_lower + i - 1  ! idx in momentum set
      !> check the flavour
      itag = iTag_flav(npar, imom)
      !> In the revised version, we include all particles as "flavjets"
      if (iflav_dress /= 7) then
        if (itag == 0 .and. .not. flav_qunderdress) cycle
      end if
      !> `i` is flavoured
      flavjets(i)%isJet = -1  ! protojet
      flavjets(i)%p(:) = kd%p(:, imom)
      flavjets(i)%pt2 = v1_pt2(kd%p(:, imom))
      flavjets(i)%y = v1_y(kd%p(:, imom))
      flavjets(i)%phi = v1_phi(kd%p(:, imom))
      if (itag > 0) flavjets(i)%nflav_q = 1
      if (itag < 0) flavjets(i)%nflav_qb = 1
      if (itag /= 0) flavjets(i)%isFlavoured = .true.
    end do
    !!! print*, "> j_target", j_target

    !> the flavour underdress
    if (flav_qunderdress) then
      !> init distances
      dij = -1._wp  ! invalid state (don't use zero!)
      do i = 1, nproto_jet
        if (.not. flavjets(i)%isJet < 0) cycle
        do j = i + 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle
          dij(i, j) = del_udress(flavjets(i), flavjets(j))
        end do
      end do
      !> sequential clustering
      loop_udress: do k = 1, nproto_jet  ! loop: maximum possible recombinations
        !> find minimum distance among "protojets"
        dij_min = -1._wp
        ij_min = -1
        do i = 1, nproto_jet
          if (.not. flavjets(i)%isJet < 0) cycle
          do j = i + 1, nproto_jet
            if (.not. flavjets(j)%isJet < 0) cycle
            if (dij(i, j) < 0._wp) stop "udress: unexpected neg. distance?!"
            if (dij(i, j) < dij_min .or. dij_min < 0._wp) then
              dij_min = dij(i, j)
              ij_min = (/i, j/)
            end if
          end do
        end do
        !> nothing left or Rcut termination reached
        if ((dij_min < 0._wp) .or. (dij_min > flav_underdress_R)) &
          exit loop_udress
        i = ij_min(1)
        j = ij_min(2)
        !> check cases
        if (.not. flavjets(i)%isFlavoured .and. &
            .not. flavjets(j)%isFlavoured) then
          !> flavless-flavless  [j killed]
          call merge_jet(flavjets(i), flavjets(j))
          do j = 1, nproto_jet
            if (.not. flavjets(j)%isJet < 0) cycle
            if (j == i) cycle
            if (i < j) dij(i, j) = del_udress(flavjets(i), flavjets(j))
            if (i > j) dij(j, i) = del_udress(flavjets(i), flavjets(j))
          end do
        else if (flavjets(i)%isFlavoured .and. &
                 flavjets(j)%isFlavoured) then
          !> flav-flav  (if collinear: flav dress will merge later)
          flavjets(i)%isJet = 1
          flavjets(j)%isJet = 1
        else
          !> flav[i]-flavless[j]  (ensure assignment)
          if (flavjets(j)%isFlavoured) then
            i = ij_min(2)
            j = ij_min(1)
          end if
          ptf = sqrt(flavjets(i)%pt2)
          ptj = sqrt(flavjets(j)%pt2)
          dRfj = v2_delta_R(flavjets(i)%p, flavjets(j)%p)
          if (min(ptf, ptj)/(ptf + ptj) > &
              flav_underdress_zcut &
              *(dRfj/flav_underdress_R)**flav_underdress_beta) then
            !> cluster condition satisfied => absorb into flav[i]
            call merge_jet(flavjets(i), flavjets(j))
            do j = 1, nproto_jet
              if (.not. flavjets(j)%isJet < 0) cycle
              if (j == i) cycle
              if (i < j) dij(i, j) = del_udress(flavjets(i), flavjets(j))
              if (i > j) dij(j, i) = del_udress(flavjets(i), flavjets(j))
            end do
          end if
        end if
      end do loop_udress  ! k
      !> compatibility (expect flavoured: `isJet = -1`)
      do i = 1, nproto_jet
        !> double-check the flavour (debugging)
        imom = kd%ijets_lower + i - 1  ! idx in momentum set
        itag = iTag_flav(npar, imom)
        if (flavjets(i)%isFlavoured .and. itag == 0) &
          stop "udress: flavoured not tagged?"
        if (.not. flavjets(i)%isFlavoured .and. itag /= 0) &
          stop "udress: flavourless tagged?"
        !> flavoured can still be a "protojet"
        if (flavjets(i)%isFlavoured .and. flavjets(i)%isJet == 0) &
          stop "udress: flavoured killed?!"
        if (flavjets(i)%isFlavoured) then
          flavjets(i)%isJet = -1
        else
          flavjets(i)%isJet = 0
        end if
      end do
    end if  ! udress

    !> HTfull for wilson-line inspired penalty
    HTfull = 0._wp
    if (flav_dress_cluster_ipenalty == 1) then
      pt_balance = 0._wp
      do i = 1, nproto_jet
        if (.not. jets(i)%isJet > 0) cycle
        HTfull = HTfull + sqrt(jets(i)%pt2)
        pt_balance(1:2) = pt_balance(1:2) + jets(i)%p(1:2)
      end do
      !> add the balance to HTfull (possibly unresolved jets, colour-neutral FS)
      HTfull = HTfull + sqrt(pt_balance(1)**2 + pt_balance(2)**2)
      !!! print*, "> HTfull", HTfull
    end if

    !> jet ordering for top-down approach
    j_order = 0
    if (flav_dress_cluster_idress == 1) then
      do k = 1, nproto_jet
        pt2_max = -1._wp
        i_max = -1
        do i = 1, nproto_jet
          if (.not. jets(i)%isJet > 0 .or. any(j_order == i)) cycle
          if (jets(i)%pt2 > pt2_max) then
            pt2_max = jets(i)%pt2
            i_max = i
          end if
        end do
        if (i_max < 0) exit  ! ran out of jets
        j_order(k) = i_max
      end do
    end if

    !> In the revised version, we use a separate routine to update distances
    if (iflav_dress == 7) then
      call update_distances(j_target)
    else
      !> init distance measures & clustering history
      diB = -1._wp  ! invalid state (don't use zero!)
      dij = -1._wp  ! invalid state (don't use zero!)
      clusterHist_flavjet = 0  ! inactive (don't really need it)
      do i = 1, nproto_jet
        if (.not. flavjets(i)%isJet < 0) cycle
        ijet = j_target(i)
        if (jets(ijet)%isJet > 0) then
          diB(i) = del_fj(flavjets(i), jets(ijet))
        !!! print*, ">> diJ ", diB(i), "|", i
        elseif (flav_dress_cluster_qbeam) then
          diB(i) = del_fB(flavjets(i))
        !!! print*, ">> diB ", diB(i), "|", i
        end if
        do j = i + 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle
          !> check if flavours are compatible
          if (qKeepEvenTagged_flav() .and. &
              ((flavjets(i)%nflav_q + flavjets(j)%nflav_q) /= &
               (flavjets(i)%nflav_qb + flavjets(j)%nflav_qb)) &
              ) cycle  ! only allow q+qb flavour clustering
          dij(i, j) = del_ff(flavjets(i), flavjets(j))
        !!! print*, ">> dij ", dij(i,j), "|", i, j
        end do
      end do
    end if

    !> perform the clustering
    loop_jets: do l = 1, nproto_jet
      ijet = j_order(l)  ! the l-th leading jet
      if (flav_dress_cluster_idress == 1) then
        if (ijet <= 0) exit loop_jets
      else
        if (l > 1) exit loop_jets
      end if

      loop_reco: do k = 1, nproto_jet  ! loop: maximum possible recombinations

        !> find minimum distance
        dij_min = -1._wp
        diB_min = -1._wp
        ij_min = -1
        iB_min = -1
        do i = 1, nproto_jet
          if (.not. flavjets(i)%isJet < 0) cycle
          !> flavour-jet distance
        !!! print*, ">>> testing diB", diB(i), "|", i
          if (flav_dress_cluster_idress == 1) then
            if ((j_target(i) == ijet) .or. &
                (.not. jets(j_target(i))%isJet > 0)) then
              if ((diB(i) >= 0._wp) .and. &
                  (diB(i) < diB_min .or. diB_min < 0._wp)) then
                iB_min = i
                diB_min = diB(i)
              end if
            end if
          else
            if ((diB(i) >= 0._wp) .and. &
                (diB(i) < diB_min .or. diB_min < 0._wp)) then
              iB_min = i
              diB_min = diB(i)
            end if
          end if
          !> flavour-flavour
          do j = i + 1, nproto_jet
            if (.not. flavjets(j)%isJet < 0) cycle
            if (dij(i, j) < 0._wp) cycle  ! even tag -> distances can be invalid
          !!! print*, ">>> testing dij", dij(i,j), "|", i, j
            if (dij(i, j) < dij_min .or. dij_min < 0._wp) then
              dij_min = dij(i, j)
              ij_min = (/i, j/)
            end if
          end do
        end do

        !> no clustering left over
        if (flav_dress_cluster_idress == 1) then
          !> top-down:  there's no flavour <--> l-th jet distance left over
          !>   =>  move on to (l+1)-th leading jet
          if (diB_min < 0._wp) exit loop_reco
        else
          !> this must be inside the else block:
          !>  top down can have a single flavour left over that
          !>  should be matched to a sub-leading jet later down the line.
          if ((diB_min < 0._wp) .and. (dij_min < 0._wp)) exit loop_reco
        end if

        !> "clustering" step
      !!! print*, ">> diB", diB_min, "v.s. dij", dij_min
        !> setting isJet = 0 already skips distances in subsequent loop!
        if ((dij_min >= 0._wp .and. dij_min < diB_min) .or. &
            (diB_min < 0._wp)) then
          !> dij is the smallest -> flavour annihilation
        !!! print*, ">>> smallest: dij", dij_min, "|", ij_min
          if (any(ij_min <= 0)) &
            stop "cluster_flav_dress_jet: ij_min invalid"
          if (iflav_dress /= 7) then
            flavjets(ij_min(1))%isJet = 0
            flavjets(ij_min(2))%isJet = 0
          else
            i = ij_min(1)
            j = ij_min(2)
            !> determine new association information before merging jets
            ijet = j_target(i)
            jjet = j_target(j)
            if (ijet == jjet) then
              j_target(i) = ijet
            else
              j_target(i) = 0
            end if
            !> this routine does not touch the flavour information of the jets!
            !> second jet is eliminated i.e. isJet = 0
            call merge_jet(flavjets(i), flavjets(j))
            !> merge flavour of particles involved
            flavjets(i)%nflav_q = flavjets(i)%nflav_q + flavjets(j)%nflav_q
            flavjets(i)%nflav_qb = flavjets(i)%nflav_qb + flavjets(j)%nflav_qb
            !> Recompute all distances w.r.t. jets and beam distances
            call update_distances(j_target)
          end if
        else
          !> diB smallest -> absorb flavour i into the jet
          !>              -> just discard flavour if target is a beam
        !!! print*, ">>> smallest: diB", diB_min, "|", iB_min
          if (iB_min <= 0) &
            stop "cluster_flav_dress_jet: iB_min invalid"
          if (j_target(iB_min) > 0) then
            if (jets(j_target(iB_min))%isJet > 0) then
              jets(j_target(iB_min))%nflav_q = &
                jets(j_target(iB_min))%nflav_q + flavjets(iB_min)%nflav_q
              jets(j_target(iB_min))%nflav_qb = &
                jets(j_target(iB_min))%nflav_qb + flavjets(iB_min)%nflav_qb
            end if
          end if
          !> isFlavoured is updated outside this routine later
          flavjets(iB_min)%isJet = 0
        end if

      end do loop_reco  ! k
    end do loop_jets  ! l

    return

  contains

    !> jet distance measure for the underdress (= C/A for now)
    pure real(kind=wp) function del_udress(j1, j2)
      use Constants_mod, only: pi, twopi
      type(Jet_t), intent(in) :: j1, j2
      real(kind=wp) :: del_y, del_phi, deltaRSq
      real(kind=wp) :: RSq_dyn
      del_y = j1%y - j2%y
      del_phi = abs(j1%phi - j2%phi)
      if (del_phi .gt. pi) del_phi = twopi - del_phi
      deltaRSq = del_y**2 + del_phi**2
      del_udress = deltaRSq
    end function del_udress

    !> jet distance measure between flavoured partons (stripped from R_cone)
    pure real(kind=wp) function del_ff(f1, f2)
      use Constants_mod, only: pi, twopi
      type(Jet_t), intent(in) :: f1, f2
      real(kind=wp) :: dang_1, dang_2, delta, deltaphi, deltay

      del_ff = 0._wp
      delta = 0._wp

      select case (nPDFs_proc)
      case (0, 1)
        !> epem & DIS distance measure
        delta = dot_eucl(f1%p(1:3), f2%p(1:3)) &
                /sqrt(dot_eucl(f1%p(1:3), f1%p(1:3))) &
                /sqrt(dot_eucl(f2%p(1:3), f2%p(1:3)))
        delta = (1._wp - delta)  ! 1 - cos(th_ij)
        del_ff = &
          max(f1%p(4), f2%p(4))**(flav_dress_cluster_alpha)* &
          min(f1%p(4), f2%p(4))**(2._wp - flav_dress_cluster_alpha)* &
          delta
      case default
        !> hadron collider distance measure
        if (iflav_dress == 4) then
          !> jade (pi.pj)
          del_ff = dot(f1%p, f2%p)
        else
          !> flavour-kt
          deltaphi = abs(f1%phi - f2%phi)
          if (deltaphi > pi) deltaphi = twopi - deltaphi
          deltay = f1%y - f2%y
          if (flav_dress_cluster_use_omega) then
            delta = 2._wp*(1._wp/flav_dress_cluster_omega/flav_dress_cluster_omega* &
                         (cosh(flav_dress_cluster_omega*deltay) - 1._wp) - &
                         (cos(deltaphi) - 1._wp))
          else
            delta = deltay**2 + deltaphi**2  ! (delta R)^2
          end if
          del_ff = &
            max(f1%pt2, f2%pt2)**(flav_dress_cluster_alpha/2._wp)* &
            min(f1%pt2, f2%pt2)**(1._wp - flav_dress_cluster_alpha/2._wp)* &
            delta
        end if
      end select

      ! !> only relative ordering matters: drop factors of 2, 1/Q2, ...
      ! if (flav_dress_cluster_alpha == 2._wp) then
      !   del_ff = max(engy_1,engy_2)**2 * delta
      ! elseif (flav_dress_cluster_alpha == 0._wp) then
      !   del_ff = min(engy_1,engy_2)**2 * delta  ! Durham
      ! elseif (flav_dress_cluster_alpha == 1._wp) then
      !   del_ff = engy_1 * engy_2 * delta  ! Jade
      ! else
      !   del_ff =  &
      !     &  max(engy_1,engy_2)**(flav_dress_cluster_alpha) *  &
      !     &  min(engy_1,engy_2)**(2._wp-flav_dress_cluster_alpha) *  &
      !     &  delta
      ! end if

      !> numerical stability (flavour-jet distance can be exactly zero)
      if (del_ff < 0._wp) del_ff = 0._wp

    end function del_ff

    !> flavoured parton to jet distance
    pure real(kind=wp) function del_fj(f, j)
      type(Jet_t), intent(in) :: f, j
      type(Jet_t) :: j_wilson
      real(kind=wp) :: fac

      select case (flav_dress_cluster_ipenalty)
      case (1)  ! wilson
        fac = 1._wp - flav_dress_cluster_beta + &
              flav_dress_cluster_beta*HTfull/sqrt(j%pt2)
        j_wilson = j
        j_wilson%p = fac*j%p
        j_wilson%pt2 = (fac**2)*j%pt2
        del_fj = del_ff(f, j_wilson)
      case (2)  ! mult
        fac = (max(j%pt2, f%pt2)/f%pt2)**(flav_dress_cluster_beta/2._wp)
        del_fj = fac*del_ff(f, j)
      case default
        del_fj = del_ff(f, j)
      end select

    end function del_fj

    !> flavoured parton to beam distance
    real(kind=wp) function del_fB(f)
      type(Jet_t), intent(in) :: f
      real(kind=wp) :: ktBp_, ktBm_, kti_
      integer :: i_
      real(kind=wp) :: Et_, pt_

      del_fB = -1._wp  ! invalid by default
      select case (nPDFs_proc)
      case (0, 1)
        !> epem has no beam
        !> DIS uses `E` and not `pt`: no need for the beam distance
        !>  (also remnant usually added as a separate "jet")
        return
      case default
        !> hadron colliders have two beams:
        ktBp_ = 0._wp
        ktBm_ = 0._wp
        do i_ = 1, nproto_jet
          if (.not. jets(i_)%isJet > 0) cycle
          kti_ = sqrt(jets(i_)%pt2)
          if (jets(i_)%y > f%y) then
            ktBp_ = ktBp_ + kti_
            ktBm_ = ktBm_ + kti_*exp(f%y - jets(i_)%y)
          else
            ktBp_ = ktBp_ + kti_*exp(jets(i_)%y - f%y)
            ktBm_ = ktBm_ + kti_
          end if
        end do
        !> deal with nonQCD particles
        if (flav_dress_cluster_beam_nonqcd == 1) then
          !> sum of them
          Et_ = sqrt(nonQCDsum%pt2 + nonQCDsum%m2)
          if (nonQCDsum%y > f%y) then
            ktBp_ = ktBp_ + Et_
            ktBm_ = ktBm_ + Et_*exp(f%y - nonQCDsum%y)
          else
            ktBp_ = ktBp_ + Et_*exp(nonQCDsum%y - f%y)
            ktBm_ = ktBm_ + Et_
          end if
        elseif (flav_dress_cluster_beam_nonqcd == 2) then
          !> individually
          do i_ = 1, nonQCDvecsize
            pt_ = sqrt(nonQCDvec(i_)%pt2)
            if (nonQCDvec(i_)%y > f%y) then
              ktBp_ = ktBp_ + pt_
              ktBm_ = ktBm_ + pt_*exp(f%y - nonQCDvec(i_)%y)
            else
              ktBp_ = ktBp_ + pt_*exp(nonQCDvec(i_)%y - f%y)
              ktBm_ = ktBm_ + pt_
            end if
          end do
        end if
        !> **************************************************
        if (ktBp_ < ktBm_) then
          del_fB = &
            max(f%pt2, ktBp_**2)**(flav_dress_cluster_alpha/2._wp)* &
            min(f%pt2, ktBp_**2)**(1._wp - flav_dress_cluster_alpha/2._wp)
        else
          del_fB = &
            max(f%pt2, ktBm_**2)**(flav_dress_cluster_alpha/2._wp)* &
            min(f%pt2, ktBm_**2)**(1._wp - flav_dress_cluster_alpha/2._wp)
        end if
      end select

    end function del_fB

    subroutine update_distances(jass)
      use Flavour_mod, only: qKeepEvenTagged_flav
      integer, dimension(npar_jet), intent(in) :: jass
      logical :: i_is_flav, j_is_flav, sum_is_flav
      integer sumq, sumqb

      diB = -1._wp  ! invalid state (don't use zero!)
      dij = -1._wp  ! invalid state (don't use zero!)
      do i = 1, nproto_jet
        if (.not. flavjets(i)%isJet < 0) cycle
        ijet = jass(i)
        if (ijet > 0) then
          if (jets(ijet)%isJet > 0) then
            diB(i) = del_fj(flavjets(i), jets(ijet))
          else
            diB(i) = del_fB(flavjets(i))
          end if
        else
          diB(i) = del_fB(flavjets(i))
        end if
        ! diB(i) = del_flav_cluster( flavjets(i) , jets(j_target(i)) )
        ! !!! print*, ">> diB ", diB(i), "|", i
        do j = i + 1, nproto_jet
          if (.not. flavjets(j)%isJet < 0) cycle
          if (qKeepEvenTagged_flav()) then
            j_is_flav = (flavjets(j)%nflav_q /= flavjets(j)%nflav_qb)
            i_is_flav = (flavjets(i)%nflav_q /= flavjets(i)%nflav_qb)
          else
            j_is_flav = modulo(flavjets(j)%nflav_q + flavjets(j)%nflav_qb, 2) /= 0
            i_is_flav = modulo(flavjets(i)%nflav_q + flavjets(i)%nflav_qb, 2) /= 0
          end if
          jjet = jass(j)
          if ((.not. i_is_flav .or. .not. j_is_flav) .and. (ijet /= jjet)) cycle
          !> check if flavours are compatible
          sumq = flavjets(i)%nflav_q + flavjets(j)%nflav_q
          sumqb = flavjets(i)%nflav_qb + flavjets(j)%nflav_qb
          if (qKeepEvenTagged_flav()) then
            sum_is_flav = (sumq /= sumqb)
          else
            sum_is_flav = modulo(sumq + sumqb, 2) /= 0
          end if
          ! only allow q+qb flavour clustering
          if (i_is_flav .and. j_is_flav .and. sum_is_flav) cycle
          dij(i, j) = del_ff(flavjets(i), flavjets(j))
          !print*, ">> dij ", dij(i,j), "|", i, j
          ! print*, ">> dij ", dij(i,j), "|", i, j
        end do
      end do
    end subroutine update_distances

    !> commenting this 'pure' implementation without nonQCD particles in the beam
    !> could become useful again in the future in case we understand that
    !> nonQCD particles in the beam distance are not really needed with flav. dressing
!!!    !> flavoured parton to beam distance
!!!    pure real(kind=wp) function del_fB(f)
!!!      type(Jet_t), intent(in) :: f
!!!      real(kind=wp) :: ktBp_, ktBm_, kti_
!!!      integer :: i_
!!!
!!!      del_fB = -1._wp  ! invalid by default
!!!      select case(nPDFs_proc)
!!!      case (0,1)
!!!        !> epem has no beam
!!!        !> DIS uses `E` and not `pt`: no need for the beam distance
!!!        !>  (also remnant usually added as a separate "jet")
!!!        return
!!!      case default
!!!        !> hadron colliders have two beams:
!!!        ktBp_ = 0._wp
!!!        ktBm_ = 0._wp
!!!        do i_ = 1, nproto_jet
!!!          if ( .not. jets(i_)%isJet > 0 ) cycle
!!!          kti_ = sqrt(jets(i_)%pt2)
!!!          if ( jets(i_)%y > f%y ) then
!!!            ktBp_ = ktBp_ + kti_
!!!            ktBm_ = ktBm_ + kti_ * exp( f%y - jets(i_)%y )
!!!          else
!!!            ktBp_ = ktBp_ + kti_ * exp( jets(i_)%y - f%y )
!!!            ktBm_ = ktBm_ + kti_
!!!          end if
!!!        end do
!!!        if ( ktBp_ < ktBm_ ) then
!!!          del_fB =  &
!!!            &  max(f%pt2,ktBp_**2)**(flav_dress_cluster_alpha/2._wp) *  &
!!!            &  min(f%pt2,ktBp_**2)**(1._wp-flav_dress_cluster_alpha/2._wp)
!!!        else
!!!          del_fB =  &
!!!            &  max(f%pt2,ktBm_**2)**(flav_dress_cluster_alpha/2._wp) *  &
!!!            &  min(f%pt2,ktBm_**2)**(1._wp-flav_dress_cluster_alpha/2._wp)
!!!        end if
!!!      end select
!!!
!!!    end function del_fB
  end subroutine cluster_flav_dress_jet

end module Jets_mod
