!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Kinematics Data [external members: `_kin`]
!
! DESCRIPTION:
!> Module storing and retrieving all kinematics information
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module KinData_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  implicit none
  private

  public :: init_kin, destroy_kin
  public :: setParFrac_kin
  ! external access to the storage.
  !> @note comparison to legacy code: p6(i,j) <-> kin(6)%p(i,j), ...
  !>       (no need for cascade if..then..elif.. constructs)
  public :: kin
  public :: spinAB_kin, fillS_kin
  public :: swapzAB_kin, resetzAB_kin
  public :: transposezA_kin, transposezB_kin
  public :: Kin_t
  !> momentum fractions
  public :: x1_kin, x2_kin, xreg1_kin, xreg2_kin
  !> fragmentation convolution
  public :: zfrag_kin, zregfrag_kin
  public :: z2frag_kin, zreg2frag_kin
  public :: resetFrag_kin
  public :: setFragFrac_kin, setFrag_kin
  public :: setFrag2Frac_kin, setFrag2_kin
  !> routines to manage the Observables cache
  public :: setCacheObsSize_kin
  public :: writeCacheObs_kin, storeCacheObs_kin, readCacheObs_kin
  public :: printCacheObs_kin, compCacheObs_kin, copyCacheObs_kin
  public :: clearCacheObs_kin
  !> three routines intended for use with the dynamic cache (mostly for pstore)
  public :: copy_kin, overwrite_kin, clearCache_kin
  !> wrapper
  public :: clearCacheLocal_kin
  !> helper to fill OL momenta
  public :: fill_pOL
  public :: fill_pOLHinc
  public :: fill_pOLH3decay
  !> helper routines
  public :: dot, dot_eucl
  public :: boost_rest
  public :: boostrest, unboostrest, unrotatetoz, rotatetoz
  !> pointers for legacy support
  public :: p3, p4, p5, p6, p7, p8, p9, p10
  public :: s3, s4, s5, s6, s7, s8, s9, s10
  public :: zA3, zA4, zA5, zA6, zA7, zA8, zA9, zA10
  public :: zB3, zB4, zB5, zB6, zB7, zB8, zB9, zB10

  !> size of the observables chache.
  !> should be identical to n_obs in `Observables_mod`
  !> can't `use Observables_mod` because of a cyclic dependence :(
  integer :: nObs = 0

  !> parton momentum fractions
  real(kind=wp) :: x1_kin, x2_kin, xreg1_kin, xreg2_kin
!$OMP THREADPRIVATE(x1_kin, x2_kin, xreg1_kin, xreg2_kin)

  !> convolution variable with fragmentation function
  real(kind=wp) :: zfrag_kin, zregfrag_kin  ! momentum fraction of the *photon*
!$OMP THREADPRIVATE(zfrag_kin, zregfrag_kin)
  real(kind=wp) :: z2frag_kin, zreg2frag_kin  ! momentum fraction of the 2nd(?) *photon*
!$OMP THREADPRIVATE(z2frag_kin, zreg2frag_kin)

  !> @todo: eventually mode to Events_mod ("modules3")
  !> bremsstrahlung option: variable number of photons
  integer, public :: max_nphotons_brems = 0  ! max # of bremsstrahlung photons in this process
  integer, public :: nphotons_brems = 0      ! # of additional bremsstrahlung photons (EW_real)
!$OMP THREADPRIVATE(nphotons_brems)
  !> photon fragmentation option: variable number of photons
  integer, public :: max_nphotons_frag = 0  ! max # of fragmentation photons in this process
  integer, public :: nphotons_frag = 0      ! # of additional fragmentation photons (EW_real)
!$OMP THREADPRIVATE(nphotons_frag)

  type Kin_t
    !>--------------------
    ! kinematics
    !>--------------------
    integer :: npar = -1
    integer :: npar_from = -1  ! mapping information
    !> first entry in momentum `p` matrix is the Lorentz index:
    !>    p(1,:)  -->  px
    !>    p(2,:)  -->  py
    !>    p(3,:)  -->  pz
    !>    p(4,:)  -->  E
    real(kind=wp), allocatable, dimension(:, :) :: p, s
#ifdef USE_OL
    real(kind=wp), allocatable, dimension(:, :) :: pOL, pOLHinc, pOLH3decay
#endif
    ! but conforming to the standard ;p
    complex(kind=wp), private, allocatable, dimension(:, :) :: zA_data, zB_data
    ! we don't allow direct access to the z[AB]_data but use pointers
    ! this makes swapping cheap (request by Tom)
    complex(kind=wp), dimension(:, :), pointer :: zA, zB
    !>--------------------
    ! cache information
    !>--------------------
    !>    ID:   unique number of the cache
    !>    PTR:  pointer to the Kin_t object in the cache
    integer :: cacheID = 0
    type(Kin_t), pointer :: cachePTR => null()
    !>--------------------
    ! reconstructed jets
    !>--------------------
    real(kind=wp) :: RconeSq = -1._wp
    integer :: njets = -1
    integer :: flav_njets = -1
    !> lower/upper indices of proto-jets (also see comments in `initialiseproc.f`)
    !> jet recombination should act on the momenta [ijets_lower, ijets_upper]
    integer :: ijets_lower, ijets_upper, nproto
    real(kind=wp), allocatable, dimension(:, :) :: pjets
    logical, allocatable, dimension(:) :: qflavjet
    integer, allocatable, dimension(:) :: sumqflavjet

#ifdef USEFASTJET
    !>--------------------
    ! reconstructed jets from fastjet
    ! (ialgo=0 is the input <-> parton momenta)
    !>--------------------
    !>     dim 1: (0:n_fj): jet algorithm
    integer, allocatable, dimension(:) :: njets_fj
    !>     dim 1: (1:4): px,py,pz,E
    !>     dim 2: (1:njets+): jet index (sorted in pt)
    !>     dim 3: (0:n_fj): jet algorithm
    real(kind=wp), allocatable, dimension(:, :, :) :: pjets_fj
    !> flavour info: +1: q, -1: qb, 0: flavourless
    !>     dim 1: (1:njets+): jet index (sorted in pt)
    !>     dim 2: (0:n_fj): jet algorithm
    integer, allocatable, dimension(:, :) :: fljets_fj
    !> number of flavoured jets
    !>     dim 1: (0:n_fj): jet algorithm
    integer, allocatable, dimension(:) :: flav_njets_fj
#endif

    !>--------------------
    ! isolated photons
    !>--------------------
    integer :: nphotons = -1
    !> ET-ordered list of photon four-momenta
    real(kind=wp), allocatable, dimension(:, :) :: pphotons
    !>--------------------
    ! fragmentation information
    !>--------------------
    integer :: ifrag = -1  ! index of parent: inactive <= 0
    integer :: i2frag = -1  ! index of parent: inactive <= 0
    real(kind=wp) :: zfrag  ! momentum fraction of 1st *photon* (in general /= zfrag_kin above)
    real(kind=wp) :: z2frag ! momentum fraction of 2nd *photon* (in general /= z2frag_kin above)
    !>--------------------
    ! Observables cache
    !>--------------------
    logical :: qcachedObs = .false.
    real(kind=wp), allocatable, dimension(:) :: cacheObs
  end type Kin_t

  type(Kin_t), allocatable, dimension(:), target :: kin
!$OMP THREADPRIVATE(kin)

  ! for backwards compatibility define some array pointers
  real(kind=wp), dimension(:, :), pointer :: p3, p4, p5, p6, p7, p8, p9, p10
  real(kind=wp), dimension(:, :), pointer :: s3, s4, s5, s6, s7, s8, s9, s10
  complex(kind=wp), dimension(:, :), pointer :: zA3, zA4, zA5, zA6, zA7, zA8, zA9, zA10
  complex(kind=wp), dimension(:, :), pointer :: zB3, zB4, zB5, zB6, zB7, zB8, zB9, zB10
!$OMP THREADPRIVATE(p3,p4,p5,p6,p7,p8,p9,p10)
!$OMP THREADPRIVATE(s3,s4,s5,s6,s7,s8,s9,s10)
!$OMP THREADPRIVATE(zA3,zA4,zA5,zA6,zA7,zA8,zA9,zA10)
!$OMP THREADPRIVATE(zB3,zB4,zB5,zB6,zB7,zB8,zB9,zB10)

  !> this is a double-linked list to eventually allow for the deletion of one
  !> single cache entry without clearing everything... if pstore is the only
  !> use-case of the cache, however, we will probably never need it.
  type KinCache_t
    type(Kin_t) :: cache
    !type(KinCache_t), pointer :: prev
    type(KinCache_t), pointer :: next
  end type KinCache_t

  type(KinCache_t), pointer :: headCache => null()
!$OMP THREADPRIVATE(headCache)

contains

  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialize storage for kinematics
  !
  !> @note with nmin & nmax we can optimize memory allocation for the different
  !> contributions/processes, i.e. only allocate:
  !>    [5:7] for ZJ @ RR
  !>    [4:5] for DrellYan @ RV
  !>      :    :
  !
  !> @param[in] nmin minimum number of external particles (npar) >= 4
  !> @param[in] nmax maximum number of external particles (npar) >= [4|nmin]
  !>            (for double-real I should register at least npar(Born)+2)
  !> @param[in] jetOffsetLower lower offset for protojets
  !> @param[in] jetOffsetUpper upper offset for protojets
  !>            (see comments in `initialiseproc.f`)
  !> @param[in] nphotons number of Born-level photons (=n_fs_photons_proc)
  !> @param[in] nfj number of fastjet algorithms registered
  !-----------------------------------------------------------------------------
  subroutine init_kin(nmin, nmax, jetOffsetLower, jetOffsetUpper, nphotons, nfj)
    integer, optional, intent(in) :: nmin, nmax
    integer, optional, intent(in) :: jetOffsetLower, jetOffsetUpper
    integer, optional, intent(in) :: nphotons, nfj
    integer :: min_npar, max_npar, delta_lower, delta_upper
    integer :: i
    integer :: size_pjets

    if (present(nmin)) then
      min_npar = max(3, nmin)
    else
      min_npar = 3
    end if

    if (present(nmax)) then
      max_npar = min(10, nmax)
    else
      max_npar = 10
    end if

    if (present(jetOffsetLower)) then
      delta_lower = jetOffsetLower
    else
      delta_lower = 2  ! 2 incoming particles
    end if

    if (present(jetOffsetUpper)) then
      delta_upper = jetOffsetUpper
    else
      delta_upper = 0
    end if

    if (allocated(kin)) then
      print *, "clearing KinData before new allocation (you should be doing this yourself!) ..."
      call destroy_kin()
    end if

    allocate (kin(min_npar:max_npar))

    do i = min_npar, max_npar
      kin(i)%npar = i
      kin(i)%npar_from = i
#ifdef USE_OL
      allocate (kin(i)%pOL(0:3, i))
      allocate (kin(i)%pOLHinc(0:3, i - 1))
      allocate (kin(i)%pOLH3decay(0:3, 4))
#endif
      allocate (kin(i)%p(4, i))
      allocate (kin(i)%s(i, i))
      allocate (kin(i)%zA_data(i, i))
      allocate (kin(i)%zB_data(i, i))
      kin(i)%zA => kin(i)%zA_data(:, :)
      kin(i)%zB => kin(i)%zB_data(:, :)
      !> cache id set to zero (value for the highest momentum set)
      kin(i)%cacheID = 0
      nullify (kin(i)%cachePTR)
      !> frequent dynamic allocation is costly
      !> so we will reserve the memory beforehand
      !> njets = -1 indicates that no recombination has been performed so far
      kin(i)%RconeSq = -1._wp
      kin(i)%njets = -1
      kin(i)%flav_njets = -1
      !> determine lower/upper proto-jet indices from the offsets
      kin(i)%ijets_lower = delta_lower + 1
      kin(i)%ijets_upper = i - delta_upper
      kin(i)%nproto = i - delta_lower - delta_upper
      !> only allocate the needed storage
      !> observables-matching needs a larger storage
      !> for mapped momentum sets
      size_pjets = max_npar - delta_lower - delta_upper
      size_pjets = max(size_pjets, kin(i)%nproto + 3)  ! P2B can go up to three emissions
      ! print*, "init_kin", i, ":", kin(i)%nproto, " > ", size_pjets
      allocate (kin(i)%pjets(4, size_pjets))
      allocate (kin(i)%qflavjet(size_pjets))
      allocate (kin(i)%sumqflavjet(size_pjets))
      kin(i)%qflavjet(:) = .false.
      kin(i)%sumqflavjet(:) = 0
      !> storage for photons
      !> nphotons = -1 indicates that no isolation has been run so far
      kin(i)%nphotons = -1
      if (present(nphotons)) then
        !> possibility of all emissions being photons
        allocate (kin(i)%pphotons(4, nphotons + (i - min_npar)))
      end if
      !> fragmentation
      kin(i)%ifrag = -1
      kin(i)%zfrag = 0._wp
      !> cached Observables
      kin(i)%qcachedObs = .false.
      allocate (kin(i)%cacheObs(nObs))
#ifdef USEFASTJET
      if ( present(nfj) ) then
        allocate (kin(i)%njets_fj(0:nfj))
        allocate (kin(i)%pjets_fj(4, size_pjets, 0:nfj))
        allocate (kin(i)%fljets_fj(size_pjets, 0:nfj))
        allocate (kin(i)%flav_njets_fj(0:nfj))
      else
        stop "init_kin: missing nfj argument"
      end if
#else
      if ( present(nfj) ) then
        stop "init_kin: didn't expect nfj argument!"
      end if
#endif
    end do

    nullify (headCache)

    !> legacy support is a pain...
    nullify (p3, p4, p5, p6, p7, p8, p9, p10)
    nullify (s3, s4, s5, s6, s7, s8, s9, s10)
    nullify (zA3, zA4, zA5, zA6, zA7, zA8, zA9, zA10)
    nullify (zB3, zB4, zB5, zB6, zB7, zB8, zB9, zB10)
    if (min_npar <= 3 .and. max_npar >= 3) then
      p3 => kin(3)%p(:, :)
      s3 => kin(3)%s(:, :)
      zA3 => kin(3)%zA_data(:, :)
      zB3 => kin(3)%zB_data(:, :)
    end if
    if (min_npar <= 4 .and. max_npar >= 4) then
      p4 => kin(4)%p(:, :)
      s4 => kin(4)%s(:, :)
      zA4 => kin(4)%zA_data(:, :)
      zB4 => kin(4)%zB_data(:, :)
    end if
    if (min_npar <= 5 .and. max_npar >= 5) then
      p5 => kin(5)%p(:, :)
      s5 => kin(5)%s(:, :)
      zA5 => kin(5)%zA_data(:, :)
      zB5 => kin(5)%zB_data(:, :)
    end if
    if (min_npar <= 6 .and. max_npar >= 6) then
      p6 => kin(6)%p(:, :)
      s6 => kin(6)%s(:, :)
      zA6 => kin(6)%zA_data(:, :)
      zB6 => kin(6)%zB_data(:, :)
    end if
    if (min_npar <= 7 .and. max_npar >= 7) then
      p7 => kin(7)%p(:, :)
      s7 => kin(7)%s(:, :)
      zA7 => kin(7)%zA_data(:, :)
      zB7 => kin(7)%zB_data(:, :)
    end if
    if (min_npar <= 8 .and. max_npar >= 8) then
      p8 => kin(8)%p(:, :)
      s8 => kin(8)%s(:, :)
      zA8 => kin(8)%zA_data(:, :)
      zB8 => kin(8)%zB_data(:, :)
    end if
    if (min_npar <= 9 .and. max_npar >= 9) then
      p9 => kin(9)%p(:, :)
      s9 => kin(9)%s(:, :)
      zA9 => kin(9)%zA_data(:, :)
      zB9 => kin(9)%zB_data(:, :)
    end if
    if (min_npar <= 10 .and. max_npar >= 10) then
      p10 => kin(10)%p(:, :)
      s10 => kin(10)%s(:, :)
      zA10 => kin(10)%zA_data(:, :)
      zB10 => kin(10)%zB_data(:, :)
    end if

  end subroutine init_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> Deallocate all memory storing the kinematics
  !-----------------------------------------------------------------------------
  subroutine destroy_kin()
    integer :: i

    do i = lbound(kin, 1), ubound(kin, 1)
#ifdef USE_OL
      deallocate (kin(i)%pOL)
      deallocate (kin(i)%pOLHinc)
      deallocate (kin(i)%pOLH3decay)
#endif
      deallocate (kin(i)%p)
      deallocate (kin(i)%s)
      deallocate (kin(i)%zA_data)
      deallocate (kin(i)%zB_data)
      deallocate (kin(i)%pjets)
      deallocate (kin(i)%qflavjet)
      deallocate (kin(i)%sumqflavjet)
      deallocate (kin(i)%cacheObs)
      if (allocated(kin(i)%pphotons)) deallocate (kin(i)%pphotons)
#ifdef USEFASTJET
      deallocate (kin(i)%njets_fj)
      deallocate (kin(i)%pjets_fj)
      deallocate (kin(i)%fljets_fj)
      deallocate (kin(i)%flav_njets_fj)
#endif
    end do

    deallocate (kin)

    nullify (p3, p4, p5, p6, p7, p8, p9, p10)
    nullify (s3, s4, s5, s6, s7, s8, s9, s10)
    nullify (zA3, zA4, zA5, zA6, zA7, zA8, zA9, zA10)
    nullify (zB3, zB4, zB5, zB6, zB7, zB8, zB9, zB10)
    !> no need for a cache if storage for kinematics is destroyed
    call clearCache_kin()
  end subroutine destroy_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the parton momentum fraction variables
  !
  !> @param[in] x1 momentum fraction for incoming parton 1
  !> @param[in] x2 momentum fraction for incoming parton 2
  !> @param[in] xreg1 optional convolution fraction for incoming parton 1
  !> @param[in] xreg2 optional convolution fraction for incoming parton 2
  !-----------------------------------------------------------------------------
  subroutine setParFrac_kin(x1, x2, xreg1, xreg2)
    real(kind=wp), intent(in) :: x1, x2
    real(kind=wp), intent(in), optional :: xreg1, xreg2

    x1_kin = x1
    x2_kin = x2

    if (present(xreg1)) then
      xreg1_kin = xreg1
    else
      xreg1_kin = 0._wp
    end if

    if (present(xreg2)) then
      xreg2_kin = xreg2
    else
      xreg2_kin = 0._wp
    end if

  end subroutine setParFrac_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the convolution variable for the fragmentation
  !
  !> @param[in] z momentum fraction of the photon
  !> @param[in] zreg optional convolution fraction of the photon
  !-----------------------------------------------------------------------------
  subroutine setFragFrac_kin(z, zreg)
    real(kind=wp), intent(in) :: z
    real(kind=wp), intent(in), optional :: zreg
    !print*, "setFragFrac_kin: z = ", z

    zfrag_kin = z

    if (present(zreg)) then
      zregfrag_kin = zreg
    else
      zregfrag_kin = 0._wp
    end if

  end subroutine setFragFrac_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the convolution variable for the fragmentation
  !
  !> @param[in] z2 momentum fraction of the photon
  !> @param[in] zreg2 optional convolution fraction of the photon
  !-----------------------------------------------------------------------------
  subroutine setFrag2Frac_kin(z2, zreg2)
    real(kind=wp), intent(in) :: z2
    real(kind=wp), intent(in), optional :: zreg2
    !print*, "setFrag2Frac_kin: z2 = ", z2

    z2frag_kin = z2

    if (present(zreg2)) then
      zreg2frag_kin = zreg2
    else
      zreg2frag_kin = 0._wp
    end if

  end subroutine setFrag2Frac_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the momentum fraction of the photon in a fragmentation piece
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine resetFrag_kin(npar)
    integer, intent(in) ::npar
    kin(npar)%ifrag = -1
    kin(npar)%i2frag = -1
    !> do *not* reset kin(npar)%zfrag here!!!
  end subroutine resetFrag_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the momentum fraction of the photon in a fragmentation piece
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !> @param[in] iparent index of the parent parton --> (parton[1-z] + photon[z])
  !> @param[in] z momentum fraction of the photon
  !-----------------------------------------------------------------------------
  subroutine setFrag_kin(npar, iparent, z)
    integer, intent(in) ::npar, iparent
    real(kind=wp), intent(in), optional :: z
    !> ATTENTION: can't do this because photon-isolation
    !>            will change the parton indices *after* the mapping
    ! !> check if iparent is actually a parton
    ! if ( (iparent<kin(npar)%ijets_lower) .or.   &
    !    & (iparent>kin(npar)%ijets_upper) ) then
    !   print*, "setFrag_kin: iparent is not a parton?!",   &
    !     & iparent, kin(npar)%ijets_lower, kin(npar)%ijets_upper
    !   stop
    ! end if
    !> store info
    kin(npar)%ifrag = iparent
    !print*, "setFrag_kin: iparent = ", iparent
    if (present(z)) then
      !print*, "setFrag_kin: over = ", z
      kin(npar)%zfrag = z
    else
      !print*, "setFrag_kin: conv = ", zfrag_kin
      kin(npar)%zfrag = zfrag_kin
    end if
    !> changing the fragmentation info can impact all observables
    call clearCacheLocal_kin(npar)
  end subroutine setFrag_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the momentum fraction of the 2nd(?) photon in a fragmentation piece
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !> @param[in] iparent index of the parent parton --> (parton[1-z] + photon[z])
  !> @param[in] z momentum fraction of the photon
  !-----------------------------------------------------------------------------
  subroutine setFrag2_kin(npar, iparent, z2)
    integer, intent(in) ::npar, iparent
    real(kind=wp), intent(in), optional :: z2
    kin(npar)%i2frag = iparent
    if (present(z2)) then
      kin(npar)%z2frag = z2
    else
      kin(npar)%z2frag = z2frag_kin
    end if
    call clearCacheLocal_kin(npar)
  end subroutine setFrag2_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> fill the spinor matrices zA(i,j), zB(i,j)
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine spinAB_kin(npar)
    integer, intent(in) :: npar
    integer :: i, j
    integer, dimension(npar) :: ss
    integer, parameter :: iaxis = 2
    ! !> local memory on the stack: compute then copy (better against cache misses)
    ! complex(kind=wp), dimension(npar, npar) :: zA_loc

    ss(:) = +1
    ss(1) = -1
    ss(2) = -1

    do i = 1, npar
      kin(npar)%zA_data(i, i) = cmplx(0._wp, kind=wp)
      kin(npar)%zB_data(i, i) = cmplx(0._wp, kind=wp)
      do j = i + 1, npar
        kin(npar)%zA_data(i, j) = eval_zA(i, j, npar)
        kin(npar)%zA_data(j, i) = -kin(npar)%zA_data(i, j)
        kin(npar)%zB_data(i, j) = -ss(i)*ss(j)*conjg(kin(npar)%zA_data(i, j))
        kin(npar)%zB_data(j, i) = -kin(npar)%zB_data(i, j)
      end do
    end do

    ! !> compute zA *locally*
    ! do i = 1, npar
    !   zA_loc(i, i) = cmplx(0._wp, kind=wp)
    !   do j = i + 1, npar
    !     zA_loc(i, j) = eval_zA(i, j, npar)
    !     zA_loc(j, i) = -zA_loc(i, j)
    !   end do
    ! end do
    ! kin(npar)%zA_data(:, :) = zA_loc(:, :)  ! and copy over
    ! !> compute zB from *local* zA
    ! do i = 1, npar
    !   kin(npar)%zB_data(i, i) = cmplx(0._wp, kind=wp)
    !   do j = i + 1, npar
    !     kin(npar)%zB_data(i, j) = -ss(i)*ss(j)*conjg(zA_loc(i, j))
    !     kin(npar)%zB_data(j, i) = -kin(npar)%zB_data(i, j)
    !   end do
    ! end do
    ! !> can also directly compute s_ij from *local* zA
    ! do i = 1, npar
    !   kin(npar)%s(i, i) = 0._wp
    !   do j = i + 1, npar
    !     kin(npar)%s(i, j) = ss(i)*ss(j)*(real(zA_loc(i, j))**2 + aimag(zA_loc(i, j))**2)
    !     kin(npar)%s(j, i) = kin(npar)%s(i, j)
    !   end do
    ! end do

  contains

    complex(kind=wp) function eval_zA(ia, ib, npar_)
      integer, intent(in) :: ia, ib, npar_
      real(kind=wp), dimension(:), pointer :: a, b
      real(kind=wp) :: at2, at, ap, am, bt2, bt, bp, bm
      complex(kind=wp) :: zea, zeb
      integer :: i1, i2, i3

      select case (iaxis)
      case (1)
        !> lightlike rotation along x direction
        i1 = 2
        i2 = 3
        i3 = 1
      case (2)
        !> lightlike rotation along y direction
        i1 = 3
        i2 = 1
        i3 = 2
      case (3)
        !> lightlike rotation along z direction
        i1 = 1
        i2 = 2
        i3 = 3
      case default
        stop "eval_zA: invalid axis"
      end select

      a => kin(npar_)%p(:, ia)
      b => kin(npar_)%p(:, ib)

      at2 = a(i1)**2 + a(i2)**2
      ap = a(4) + a(i3)
      if (ap < 0.5_wp*a(4)) ap = at2/(a(4) - a(i3))

      bt2 = b(i1)**2 + b(i2)**2
      bp = b(4) + b(i3)
      if (bp < 0.5_wp*b(4)) bp = bt2/(b(4) - b(i3))

      if (ap == 0._wp .or. bp == 0._wp) then
        !> treat the special case separately
        if (ap == 0._wp) then
          am = a(4) - a(i3)
          ! if (am < 0.5._wp*a(4)) am = at2 / (a(4) + a(i3))
          zea = sqrt(am*bp)*cmplx(0._wp, 1._wp, kind=wp)
        else
          zea = cmplx(a(i2), -a(i1), kind=wp)*sqrt(bp/ap)
        end if
        if (bp == 0._wp) then
          bm = b(4) - b(i3)
          ! if (bm < 0.5._wp*b(4)) bm = bt2 / (b(4) + b(i3))
          zeb = sqrt(bm*ap)*cmplx(0._wp, 1._wp, kind=wp)
        else
          zeb = cmplx(b(i2), -b(i1), kind=wp)*sqrt(ap/bp)
        end if
        eval_zA = zea - zeb
      else
        eval_zA = (cmplx(a(i2), -a(i1), kind=wp)*bp - cmplx(b(i2), -b(i1), kind=wp)*ap)/sqrt(ap*bp)
      end if
      !> global phase to match old eval_zA
      eval_zA = eval_zA*cmplx(0._wp, +1._wp, kind=wp)
      !> cross initial states because we use the convention of all ingoing => more phases
      if (ia == 1 .or. ib == 1) eval_zA = eval_zA*cmplx(0._wp, +1._wp, kind=wp)
      if (ia == 2 .or. ib == 2) eval_zA = eval_zA*cmplx(0._wp, +1._wp, kind=wp)

    end function eval_zA

  end subroutine spinAB_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> fill the momentum-product matrix s(i,j)
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine fillS_kin(npar)
    integer, intent(in) :: npar
    integer :: i, j

    call spinAB_kin(npar)

    do i = 1, npar
      kin(npar)%s(i, i) = 0._wp
      do j = i + 1, npar
        kin(npar)%s(i, j) = real(kin(npar)%zA(i, j)*kin(npar)%zB(j, i))
        kin(npar)%s(j, i) = kin(npar)%s(i, j)
      end do
    end do

    !> do NOT reset fragmentation information here!!!

  end subroutine fillS_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> swap zA and zB for `npar` kinematics
  !> (practically no cost since it's just a pointer re-assignment)
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine swapzAB_kin(npar)
    integer, intent(in) :: npar
    complex(kind=wp), dimension(:, :), pointer :: zX
    zX => kin(npar)%zA
    kin(npar)%zA => kin(npar)%zB
    kin(npar)%zB => zX
    select case (npar)
    case (3)
      zA3 => kin(npar)%zA
      zB3 => kin(npar)%zB
    case (4)
      zA4 => kin(npar)%zA
      zB4 => kin(npar)%zB
    case (5)
      zA5 => kin(npar)%zA
      zB5 => kin(npar)%zB
    case (6)
      zA6 => kin(npar)%zA
      zB6 => kin(npar)%zB
    case (7)
      zA7 => kin(npar)%zA
      zB7 => kin(npar)%zB
    case (8)
      zA8 => kin(npar)%zA
      zB8 => kin(npar)%zB
    case (9)
      zA9 => kin(npar)%zA
      zB9 => kin(npar)%zB
    case (10)
      zA10 => kin(npar)%zA
      zB10 => kin(npar)%zB
    case default
      stop "npar in swapzAB_kin not supported"
    end select
  end subroutine swapzAB_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the zA,zB pointers to "correct" references
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine resetzAB_kin(npar)
    integer, intent(in) :: npar
    kin(npar)%zA => kin(npar)%zA_data(:, :)
    kin(npar)%zB => kin(npar)%zB_data(:, :)
    select case (npar)
    case (3)
      zA3 => kin(npar)%zA
      zB3 => kin(npar)%zB
    case (4)
      zA4 => kin(npar)%zA
      zB4 => kin(npar)%zB
    case (5)
      zA5 => kin(npar)%zA
      zB5 => kin(npar)%zB
    case (6)
      zA6 => kin(npar)%zA
      zB6 => kin(npar)%zB
    case (7)
      zA7 => kin(npar)%zA
      zB7 => kin(npar)%zB
    case (8)
      zA8 => kin(npar)%zA
      zB8 => kin(npar)%zB
    case (9)
      zA9 => kin(npar)%zA
      zB9 => kin(npar)%zB
    case (10)
      zA10 => kin(npar)%zA
      zB10 => kin(npar)%zB
    case default
      stop "npar in swapzAB_kin not supported"
    end select
  end subroutine resetzAB_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> perform a transposition of the zA matrix
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine transposezA_kin(npar)
    integer, intent(in) :: npar
    !> utterly slow:
    !kin(npar)%zA_data = transpose(kin(npar)%zA_data)
    !> for zA/zB matrix transposition is just a sign
    kin(npar)%zA_data = -kin(npar)%zA_data
  end subroutine transposezA_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> perform a transposition of the zB matrix
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine transposezB_kin(npar)
    integer, intent(in) :: npar
    !> utterly slow:
    !kin(npar)%zB_data = transpose(kin(npar)%zB_data)
    !> for zA/zB matrix transposition is just a sign
    kin(npar)%zB_data = -kin(npar)%zB_data
  end subroutine transposezB_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> set the size of the cache in KinData_t
  !
  !> @param[in] nsize number of entries in the Observables-cache array
  !-----------------------------------------------------------------------------
  subroutine setCacheObsSize_kin(nsize)
    integer, intent(in) :: nsize
    nObs = nsize
  end subroutine setCacheObsSize_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> write to the Observables cache
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !> @param[in] obsVals array containing the Observable values
  !-----------------------------------------------------------------------------
  subroutine writeCacheObs_kin(npar, obsVals)
    integer, intent(in) :: npar
    real(kind=wp), dimension(nObs), intent(in) :: obsVals
    integer :: i
#ifndef NDEBUG
    if (kin(npar)%qcachedObs) then
      print *, "writeCacheObs_kin: already cached?!"
      stop
    end if
#endif
    do i = 1, nObs
      kin(npar)%cacheObs(i) = obsVals(i)
    end do
    kin(npar)%qcachedObs = .true.
    !> copy over to the "real" cache if exists
    call storeCacheObs_kin(npar)
  end subroutine writeCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> after evaluating all observables, make sure everything is copied in
  !> the KinCache
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine storeCacheObs_kin(npar)
    integer, intent(in) :: npar
    type(Kin_t), pointer :: kin_p
    integer :: i, j
    if (.not. kin(npar)%qcachedObs) then
      print *, "storeCacheObs_kin: try to store uncached observables..."
      return
    end if
    !> is there a KinCache?
    if (associated(kin(npar)%cachePTR)) then
      kin_p => kin(npar)%cachePTR  ! shorthand
      ! print*, ">>> pushing observables into cache with id:", kin_p%cacheID
      kin_p%npar_from = kin(npar)%npar_from
      kin_p%RconeSq = kin(npar)%RconeSq
      kin_p%njets = kin(npar)%njets
      kin_p%flav_njets = kin(npar)%flav_njets
      do j = 1, kin(npar)%njets
        do i = 1, 4
          kin_p%pjets(i, j) = kin(npar)%pjets(i, j)
        end do
        kin_p%qflavjet(j) = kin(npar)%qflavjet(j)
        kin_p%sumqflavjet(j) = kin(npar)%sumqflavjet(j)
      end do
      kin_p%nphotons = kin(npar)%nphotons
      if (allocated(kin_p%pphotons) .and. allocated(kin(npar)%pphotons)) then
        do j = 1, kin(npar)%nphotons
          do i = 1, 4
            kin_p%pphotons(i, j) = kin(npar)%pphotons(i, j)
          end do
        end do
      end if
      kin_p%ifrag = kin(npar)%ifrag
      kin_p%i2frag = kin(npar)%i2frag
      kin_p%zfrag = kin(npar)%zfrag
      kin_p%z2frag = kin(npar)%z2frag
      do i = 1, nObs
        kin_p%cacheObs(i) = kin(npar)%cacheObs(i)
        ! print*, ">>>", kin_p%cacheObs(i)
      end do
      kin_p%qcachedObs = .true.
    end if
  end subroutine storeCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> read from the Observables cache
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !> @param[out] obsVals array where the Observable values are written to
  !-----------------------------------------------------------------------------
  subroutine readCacheObs_kin(npar, obsVals)
    integer, intent(in) :: npar
    real(kind=wp), dimension(nObs), intent(out) :: obsVals
    integer :: i
#ifndef NDEBUG
    if (.not. kin(npar)%qcachedObs) then
      print *, "readCacheObs_kin: tried to read from an uncached array"
      stop
    end if
#endif
    do i = 1, nObs
      obsVals(i) = kin(npar)%cacheObs(i)
    end do
  end subroutine readCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> print content of the observables cache
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine printCacheObs_kin(npar)
    integer, intent(in) :: npar
    integer :: i
    print *, "printCacheObs_kin", npar
    print *, "cacheID", kin(npar)%cacheID
    print *, "npar_from", kin(npar)%npar_from
    if (.not. kin(npar)%qcachedObs) then
      print *, "<<< NOT CACHED >>>"
    end if
    do i = 1, nObs
      print *, i, kin(npar)%cacheObs(i)
    end do
  end subroutine printCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> compare the observables between the two set of kinematics
  !
  !> @param[in] npar1 defines kinematics by the number of particles
  !> @param[in] npar2 defines kinematics by the number of particles
  !> @param[in] threshold optional value for the threshold
  !> @param[in] verbose optional integer flag to control verbosity
  !> @param[out] fail optional flag to check if obs agree within threshold
  !-----------------------------------------------------------------------------
  subroutine compCacheObs_kin(npar1, npar2, threshold, verbose, fail)
    integer, intent(in) :: npar1, npar2
    real(kind=wp), intent(in), optional :: threshold
    integer, intent(in), optional :: verbose
    logical, intent(out), optional :: fail
    real(kind=wp) :: thresh
    integer :: vrb
    integer :: i
    real(kind=wp) :: diff, sum, reldiff, maxreldiff
#ifndef NDEBUG
    !> observables cached?
    if (.not. kin(npar1)%qcachedObs .or. .not. kin(npar1)%qcachedObs) then
      print *, "compCacheObs_kin: tried to read from an uncached array"
      stop
    end if
#endif
    !> threshold
    thresh = 1e-3_wp  ! default value: ~ 3-4 digit agreement
    if (present(threshold)) thresh = threshold
    !> verbosity
    vrb = 1
    if (present(verbose)) vrb = verbose
    !> compute maxreldiff
    maxreldiff = 0._wp
    do i = 1, nObs
      diff = kin(npar1)%cacheObs(i) - kin(npar2)%cacheObs(i)
      sum = kin(npar1)%cacheObs(i) + kin(npar2)%cacheObs(i)
      reldiff = 0._wp
      if (sum .ne. 0._wp) reldiff = diff/sum
      if (abs(reldiff) > maxreldiff) maxreldiff = abs(reldiff)
      !print*, i, kin(npar1)%cacheObs(i), kin(npar2)%cacheObs(i), reldiff
    end do
    !> flag compatibility
    if (present(fail)) fail = (maxreldiff > thresh)
    !> optional output
    if (vrb > 0) then
      if (maxreldiff > thresh) then
        if (vrb > 1) call print_momenta(npar1)
        if (vrb > 1) call print_momenta(npar2)
        do i = 1, nObs
          diff = kin(npar1)%cacheObs(i) - kin(npar2)%cacheObs(i)
          sum = kin(npar1)%cacheObs(i) + kin(npar2)%cacheObs(i)
          reldiff = 0._wp
          if (sum .ne. 0._wp) reldiff = diff/sum
          print *, i, kin(npar1)%cacheObs(i), kin(npar2)%cacheObs(i), reldiff
        end do
      end if
    end if
  end subroutine compCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> copy over observables from one kinematics to the other
  !
  !> @param[in] nparfrom defines kinematics by the numer of particles
  !> @param[in] nparto defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine copyCacheObs_kin(nparfrom, nparto)
    integer, intent(in) :: nparfrom, nparto
    integer :: i, j
#ifndef NDEBUG
    if (.not. kin(nparfrom)%qcachedObs) then
      print *, "copyCacheObs_kin: tried to copy from an uncached momentum set"
      stop
    end if
#endif
    !> !!! do not copy `npar_from`: mapping must remain in tact !!!
    !> copy over the information of the jet clustering
    kin(nparto)%RconeSq = kin(nparfrom)%RconeSq
    kin(nparto)%njets = kin(nparfrom)%njets
    kin(nparto)%flav_njets = kin(nparfrom)%flav_njets
    !---- if I copy the observables, no need to copy jet momenta...
    !     one might think. but we should copy!!!
    !> force de- & re-allocation
    ! kin(nparto)%pjets = kin(nparfrom)%pjets
    !> quicker but maybe unsafe when njets[from] > nproto[to]?
    !> nope: now storage is adjusted to accommodate pjets full arrays
    do j = 1, kin(nparfrom)%njets
      do i = 1, 4
        kin(nparto)%pjets(i, j) = kin(nparfrom)%pjets(i, j)
      end do
      kin(nparto)%qflavjet(j) = kin(nparfrom)%qflavjet(j)
      kin(nparto)%sumqflavjet(j) = kin(nparfrom)%sumqflavjet(j)
    end do
    kin(nparto)%nphotons = kin(nparfrom)%nphotons
    if (allocated(kin(nparto)%pphotons) .and. allocated(kin(nparfrom)%pphotons)) then
      do j = 1, kin(nparfrom)%nphotons
        do i = 1, 4
          kin(nparto)%pphotons(i, j) = kin(nparfrom)%pphotons(i, j)
        end do
      end do
    end if
    kin(nparto)%ifrag = kin(nparfrom)%ifrag
    kin(nparto)%i2frag = kin(nparfrom)%i2frag
    kin(nparto)%zfrag = kin(nparfrom)%zfrag
    kin(nparto)%z2frag = kin(nparfrom)%z2frag
    !----
    !> copy over the cached observables
    do i = 1, nObs
      kin(nparto)%cacheObs(i) = kin(nparfrom)%cacheObs(i)
    end do
    kin(nparto)%qcachedObs = .true.
    !> copy over to the "real" cache if exists
    call storeCacheObs_kin(nparto)
  end subroutine copyCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> clear the Observables cache
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine clearCacheObs_kin(npar)
    integer, intent(in) :: npar
    !kin(npar)%npar_from  = npar
    Kin(npar)%RconeSq = -1._wp
    kin(npar)%njets = -1
    kin(npar)%flav_njets = -1
    kin(npar)%nphotons = -1
    !kin(npar)%ifrag      = -1  ! NEVER RESET HERE (done in unset_frag)
    kin(npar)%qcachedObs = .false.
  end subroutine clearCacheObs_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> create a new empty KinCache_t object on top of the cache-stack.
  !> The `Kin_t` sub-entry is not allocated at this point!
  !> This routine is local to the module and not accessible from outside
  !-----------------------------------------------------------------------------
  subroutine createNewCache()
    type(KinCache_t), pointer :: newCache
    allocate (newCache)
    !newCache%prev => null()
    newCache%next => headCache
    !headCache%prev => newCache
    headCache => newCache
  end subroutine createNewCache

  !-----------------------------------------------------------------------------
  !> @brief
  !> create copy of `npar`-kinematics in kin_p and allocate memory if necessary
  !
  !> @param[in] npar specify which `npar`-kinematics should be copied
  !> @param[inout] kin_p KinData pointer to copied object
  !>                     new allocation if null
  !>                     content overwritten if already allocated
  !-----------------------------------------------------------------------------
  subroutine copy_kin(npar, kin_p)
    integer, intent(in) :: npar
    type(Kin_t), intent(inout), pointer :: kin_p
    integer :: i, j
    !> make sure memory is allocated
    if (.not. associated(kin_p)) then
      call createNewCache() ! create new object on the cache stack
      kin_p => headCache%cache
      !> headCache%Kin_t members are not yet properly allocated since they
      !> depend on npar, do it here
      kin_p%npar = npar
      kin_p%npar_from = npar
      allocate (kin_p%p(4, npar))
      allocate (kin_p%s(npar, npar))
      allocate (kin_p%zA_data(npar, npar))
      allocate (kin_p%zB_data(npar, npar))
      kin_p%zA => kin_p%zA_data(:, :)
      kin_p%zB => kin_p%zB_data(:, :)
      !> we copy over the original array and by this allocate
      !> the pjets array with the correct size
      kin_p%pjets = kin(npar)%pjets
      kin_p%qflavjet = kin(npar)%qflavjet
      kin_p%sumqflavjet = kin(npar)%sumqflavjet
      !> same for the photon array
      if (allocated(kin(npar)%pphotons)) then
        kin_p%pphotons = kin(npar)%pphotons
      end if
      ! allocate(kin_p%pjets(4, size(kin(npar)%pjets, 2) ))
      allocate (kin_p%cacheObs(nObs))
      !> the new cache entry gets a new id
      if (associated(headCache%next)) then
        kin_p%cacheID = headCache%next%cache%cacheID + 1
      else
        kin_p%cacheID = 1
      end if
    end if
    !> this was a mapped momentum so overwrite the ID in the current set
    kin(npar)%cacheID = kin_p%cacheID
    !> set pointer to the refernece we copy to
    kin(npar)%cachePTR => kin_p
    nullify (kin_p%cachePTR)
#ifndef NDEBUG
    !> check for compatibility
    if (kin_p%npar /= kin(npar)%npar) then
      stop "copy_kin: npars not matching up!"
    end if
#endif
    !> copy over the content
    !>----- The do-loops look ugly but this is the efficient copy that avoids an implicit allocation
    kin_p%npar_from = kin(npar)%npar_from
    do j = 1, npar
    do i = 1, 4
      kin_p%p(i, j) = kin(npar)%p(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin_p%s(i, j) = kin(npar)%s(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin_p%zA_data(i, j) = kin(npar)%zA_data(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin_p%zB_data(i, j) = kin(npar)%zB_data(i, j)
    end do
    end do
    kin_p%RconeSq = kin(npar)%RconeSq
    kin_p%njets = kin(npar)%njets
    kin_p%flav_njets = kin(npar)%flav_njets
    kin_p%ijets_lower = kin(npar)%ijets_lower
    kin_p%ijets_upper = kin(npar)%ijets_upper
    kin_p%nproto = kin(npar)%nproto
    kin_p%nphotons = kin(npar)%nphotons
    kin_p%ifrag = kin(npar)%ifrag
    kin_p%i2frag = kin(npar)%i2frag
    kin_p%zfrag = kin(npar)%zfrag
    kin_p%z2frag = kin(npar)%z2frag

    kin_p%qcachedObs = kin(npar)%qcachedObs
    if (kin(npar)%qcachedObs) then
      do i = 1, nObs
        kin_p%cacheObs(i) = kin(npar)%cacheObs(i)
      end do
      do j = 1, kin(npar)%njets
      do i = 1, 4
        kin_p%pjets(i, j) = kin(npar)%pjets(i, j)
      end do
      kin_p%qflavjet(j) = kin(npar)%qflavjet(j)
      kin_p%sumqflavjet(j) = kin(npar)%sumqflavjet(j)
      end do
      if (allocated(kin_p%pphotons) .and. allocated(kin(npar)%pphotons)) then
        do j = 1, kin(npar)%nphotons
        do i = 1, 4
          kin_p%pphotons(i, j) = kin(npar)%pphotons(i, j)
        end do
        end do
      end if
    end if

  end subroutine copy_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> overwrite `npar`-particle kinematics from the cache
  !
  !> @param[in] kin_p KinData pointer object overwriting `npar`-kinematics
  !-----------------------------------------------------------------------------
  subroutine overwrite_kin(kin_p)
    type(Kin_t), intent(in), pointer :: kin_p
    ! Kin_t already knows npar, no need to pass it as argument
    integer :: i, j, npar
#ifndef NDEBUG
    if (.not. associated(kin_p)) then
      stop "overwrite_kin: kin_p not associated!"
    end if
#endif
    npar = kin_p%npar ! just for cleaner notation (compiler will optimize it out)
    !>----- The do-loops look ugly but this is the efficient copy that avoids an implicit allocation
    kin(npar)%npar_from = kin_p%npar_from
    do j = 1, npar
    do i = 1, 4
      kin(npar)%p(i, j) = kin_p%p(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin(npar)%s(i, j) = kin_p%s(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin(npar)%zA_data(i, j) = kin_p%zA_data(i, j)
    end do
    end do
    do j = 1, npar
    do i = 1, npar
      kin(npar)%zB_data(i, j) = kin_p%zB_data(i, j)
    end do
    end do
    kin(npar)%zA => kin(npar)%zA_data(:, :)
    kin(npar)%zB => kin(npar)%zB_data(:, :)

    kin(npar)%cacheID = kin_p%cacheID
    kin(npar)%cachePTR => kin_p
    kin(npar)%RconeSq = kin_p%RconeSq
    kin(npar)%njets = kin_p%njets
    kin(npar)%flav_njets = kin_p%flav_njets
    kin(npar)%ijets_lower = kin_p%ijets_lower
    kin(npar)%ijets_upper = kin_p%ijets_upper
    kin(npar)%nproto = kin_p%nproto
    kin(npar)%nphotons = kin_p%nphotons
    kin(npar)%ifrag = kin_p%ifrag
    kin(npar)%i2frag = kin_p%i2frag
    kin(npar)%zfrag = kin_p%zfrag
    kin(npar)%z2frag = kin_p%z2frag

    kin(npar)%qcachedObs = kin_p%qcachedObs
    if (kin_p%qcachedObs) then
      do i = 1, nObs
        kin(npar)%cacheObs(i) = kin_p%cacheObs(i)
      end do
      do j = 1, kin_p%njets
      do i = 1, 4
        kin(npar)%pjets(i, j) = kin_p%pjets(i, j)
      end do
      kin(npar)%qflavjet(j) = kin_p%qflavjet(j)
      kin(npar)%sumqflavjet(j) = kin_p%sumqflavjet(j)
      end do
      if (allocated(kin_p%pphotons) .and. allocated(kin(npar)%pphotons)) then
        do j = 1, kin_p%nphotons
        do i = 1, 4
          kin(npar)%pphotons(i, j) = kin_p%pphotons(i, j)
        end do
        end do
      end if
    end if

  end subroutine overwrite_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> clear the cache and free up all the memory
  !-----------------------------------------------------------------------------
  subroutine clearCache_kin()
    type(KinCache_t), pointer :: it, next
    it => headCache

    do while (associated(it))
      !> save next before we destroy it
      next => it%next
      !> deallocate sub-entries and then the entry itself
      deallocate (it%cache%p)
      deallocate (it%cache%s)
      deallocate (it%cache%zA_data)
      deallocate (it%cache%zB_data)
      deallocate (it%cache%pjets)
      deallocate (it%cache%qflavjet)
      deallocate (it%cache%sumqflavjet)
      if (allocated(it%cache%pphotons)) deallocate (it%cache%pphotons)
      deallocate (it%cache%cacheObs)
      deallocate (it)
      !> loop
      it => next
    end do
    nullify (headCache)
  end subroutine clearCache_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> wrapper to clear to local cache of a given npar-kinematics
  !
  !> @param[in] npar defines kinematics by the numer of particles
  !-----------------------------------------------------------------------------
  subroutine clearCacheLocal_kin(npar)
    integer, intent(in) :: npar
    call clearCacheObs_kin(npar)
  end subroutine clearCacheLocal_kin

  !-----------------------------------------------------------------------------
  !> @brief
  !> compute the dot product of four-vectors
  !
  !> @param[in] v1 1st four-vector
  !> @param[in] v2 2nd four-vector
  !> @return dot product v1.v2
  !-----------------------------------------------------------------------------
  pure real(kind=wp) function dot(v1, v2)
    use FPA_mod
    real(kind=wp), dimension(4), intent(in) :: v1, v2
    !----- Kahan summation for better numerical precision
    logical qkahan
    parameter(qkahan=.true.)
    real(kind=wp) :: err, acc_i, err_i
    integer :: i
    ! !-----
    ! real*16 :: dot_quad
    ! real*8 :: dot_dble, kahan_diff,dble_diff
    ! !-----
    if (qkahan) then
      dot = v1(4)*v2(4)
      err = 0._wp
      do i = 1, 3
        call twoSum(dot, -v1(i)*v2(i), acc_i, err_i)
        dot = acc_i
        err = err + err_i
      end do
      dot = dot + err
      !      !-----
      !      dot_dble = v1(4)*v2(4)-v1(1)*v2(1)-v1(2)*v2(2)-v1(3)*v2(3)
      !      dot_quad = real(v1(4)*v2(4), 16) &
      ! &             - real(v1(1)*v2(1), 16) &
      ! &             - real(v1(2)*v2(2), 16) &
      ! &             - real(v1(3)*v2(3), 16)
      !      kahan_diff = 2._wp * (dot      - dble(dot_quad)) / (dot      + dble(dot_quad))
      !      dble_diff  = 2._wp * (dot_dble - dble(dot_quad)) / (dot_dble + dble(dot_quad))
      !      if ( abs(kahan_diff)>1e-9_wp .or. abs(dble_diff)>1e-9_wp ) then
      !        print*,
      !        print*, "DOT: quad ", dot_quad
      !        print*, "DOT: kahan", dot, kahan_diff
      !        print*, "DOT: dble ", dot_dble, dble_diff
      !        print*,
      !      end if
      !      !-----
    else
      dot = v1(4)*v2(4) - v1(1)*v2(1) - v1(2)*v2(2) - v1(3)*v2(3)
    end if
  end function dot

  !-----------------------------------------------------------------------------
  !> @brief
  !> compute the Euclidean dot product of three-vectors
  !
  !> @param[in] v1 1st three-vector
  !> @param[in] v2 2nd three-vector
  !> @return dot product v1.v2
  !-----------------------------------------------------------------------------
  pure real(kind=wp) function dot_eucl(v1, v2)
    use FPA_mod
    real(kind=wp), dimension(3), intent(in) :: v1, v2
    !----- Kahan summation for better numerical precision
    logical qkahan
    parameter(qkahan=.true.)
    real(kind=wp) :: err, acc_i, err_i
    integer :: i
    if (qkahan) then
      dot_eucl = v1(1)*v2(1)
      err = 0._wp
      do i = 2, 3
        call twoSum(dot_eucl, v1(i)*v2(i), acc_i, err_i)
        dot_eucl = acc_i
        err = err + err_i
      end do
      dot_eucl = dot_eucl + err
    else
      dot_eucl = v1(1)*v2(1) + v1(2)*v2(2) + v1(3)*v2(3)
    end if
  end function dot_eucl

  !-----------------------------------------------------------------------------
  !> @brief
  !> boost into the rest frame of fv_ref (for direction > 0)
  !> boost_rest(fv, fv, +1._wp) ==> fv = (0,0,0,M)
  !
  !> @param[inout] fv four vector to be transformed
  !> @param[in] fv_ref reference four vector
  !> @param[in] direction optional direction of the boost
  !-----------------------------------------------------------------------------
  subroutine boost_rest(fv, fv_ref, direction)
    real(kind=wp), dimension(4), intent(inout) :: fv
    real(kind=wp), dimension(4), intent(in)    :: fv_ref
    real(kind=wp), optional, intent(in) :: direction
    real(kind=wp), dimension(4) :: fv_cpy
    real(kind=wp) :: dir, abs_ref
    real(kind=wp) :: bx, by, bz
    real(kind=wp) :: gamma, a, bp

    !> idea: make the transformation variables `SAVE`
    !> and also store fv_ref.
    !> first check if fv_rev unchanged and only compute

    dir = 1._wp
    if (present(direction)) dir = direction

    fv_cpy(1:4) = fv(1:4)

    !> determine variables for the boost
    abs_ref = sqrt(abs(dot(fv_ref, fv_ref)))
    bx = -fv_ref(1)/sign(abs_ref, dir)
    by = -fv_ref(2)/sign(abs_ref, dir)
    bz = -fv_ref(3)/sign(abs_ref, dir)
    gamma = fv_ref(4)/abs_ref
    a = 1._wp/(1._wp + gamma)
    bp = fv_cpy(1)*bx + fv_cpy(2)*by + fv_cpy(3)*bz

    !> perform the boost
    fv(4) = gamma*fv_cpy(4) + bp
    fv(1) = fv_cpy(1) + bx*fv_cpy(4) + a*bp*bx
    fv(2) = fv_cpy(2) + by*fv_cpy(4) + a*bp*by
    fv(3) = fv_cpy(3) + bz*fv_cpy(4) + a*bp*bz

  end subroutine boost_rest

  !-----------------------------------------------------------------------------
  !> @brief
  !> boost into the rest frame of 4 vector a
  !>
  !
  !> @param[in] four vector a to be boosted into rest frame
  !> @param[out] bmat transformation matrix from a -> a'
  !-----------------------------------------------------------------------------
  subroutine boostrest(a, bmat)
    implicit none
    real(kind=wp), intent(out) :: bmat(4, 4)
    real(kind=wp), intent(in) :: a(4)
    real(kind=wp) :: v(4), rma2, gam, rma, v2
    integer :: i, j

    rma2 = dot(a, a)

    if (rma2 .le. 0._wp) then
      write (*, *) 'undue boost:', rma2
      stop
    end if
    rma = sqrt(rma2)
    gam = a(4)/rma
    do i = 1, 4
      v(i) = a(i)/a(4)
    end do
    v2 = v(1)*v(1) + v(2)*v(2) + v(3)*v(3)
    bmat(4, 4) = gam
    do i = 1, 3
      bmat(i, 4) = -gam*v(i)
      bmat(4, i) = -gam*v(i)
      do j = 1, 3
        bmat(i, j) = (gam - 1._wp)*v(i)*v(j)/v2
      end do
      bmat(i, i) = bmat(i, i) + 1._wp
    end do
  end subroutine boostrest

  !-----------------------------------------------------------------------------
  !> @brief
  !> calculate inverse of bmat in boostrest
  !>
  !> @param[in] four vector a that was boosted into rest frame
  !> @param[out] bmat transformation matrix from a' -> a
  !-----------------------------------------------------------------------------
  subroutine unboostrest(a, bmat)
    implicit none
    real(kind=wp), intent(out) :: bmat(4, 4)
    real(kind=wp), intent(in) :: a(4)
    real(kind=wp) v(4), v2, rma2, rma, gam
    integer :: i, j

    rma2 = dot(a, a)
    if (rma2 .le. 0._wp) then
      write (*, *) 'undue boost'
      stop
    end if
    rma = sqrt(rma2)
    gam = a(4)/rma
    do i = 1, 4
      v(i) = -a(i)/a(4)
    end do
    v2 = v(1)*v(1) + v(2)*v(2) + v(3)*v(3)
    bmat(4, 4) = gam
    do i = 1, 3
      bmat(i, 4) = -gam*v(i)
      bmat(4, i) = -gam*v(i)
      do j = 1, 3
        bmat(i, j) = (gam - 1._wp)*v(i)*v(j)/v2
      end do
      bmat(i, i) = bmat(i, i) + 1._wp
    end do
  end subroutine unboostrest

  !-----------------------------------------------------------------------------
  !> @brief
  !> computes rotation into frame with a along z axis
  !>
  !> @param[in] four vector a that is rotated onto z axis
  !> @param[out] rmat: transformation matrix from a -> a'
  !-----------------------------------------------------------------------------
  subroutine rotatetoz(a, rmat)
    use Constants_mod, only: pi
    implicit none
    real(kind=wp), intent(in) :: a(4)
    real(kind=wp), intent(out) :: rmat(4, 4)
    real(kind=wp) :: b(4), rmatz(4, 4), rmaty(4, 4)
    real(kind=wp) :: phiz, cphiz, sphiz
    real(kind=wp) :: phiy, cphiy, sphiy
    real(kind=wp) :: b00
    integer :: i, j, k

    if (a(1) .eq. 0._wp .and. a(2) .eq. 0._wp .and. a(3) .ge. 0._wp) then
      do i = 1, 4
        do j = 1, 4
          rmat(i, j) = 0._wp
        end do
        rmat(i, i) = 1._wp
      end do
      return
    end if

    rmatz(4, 4) = 1._wp
    do i = 1, 3
      rmatz(4, i) = 0._wp
      rmatz(i, 4) = 0._wp
    end do

    if (a(1) .eq. 0._wp) then
      rmatz(1, 1) = 0._wp
      rmatz(1, 2) = 1._wp
      rmatz(2, 1) = -1._wp
      rmatz(2, 2) = 0._wp
    else
      phiz = atan(a(2)/a(1))
      cphiz = cos(phiz)
      sphiz = sin(phiz)
      rmatz(1, 1) = cphiz
      rmatz(1, 2) = sphiz
      rmatz(2, 1) = -sphiz
      rmatz(2, 2) = cphiz
    end if
    rmatz(3, 3) = 1._wp
    do i = 1, 2
      rmatz(i, 3) = 0._wp
      rmatz(3, i) = 0._wp
    end do

    do i = 1, 4
      b(i) = 0._wp
      do j = 1, 4
        b(i) = b(i) + rmatz(i, j)*a(j)
      end do
    end do

    rmaty(4, 4) = 1._wp
    do i = 1, 3
      rmaty(4, i) = 0._wp
      rmaty(i, 4) = 0._wp
    end do

    if (b(3) .eq. 0._wp) then
      b00 = 1._wp
      if (b(1) .le. 0._wp) b00 = -1._wp
      rmaty(1, 1) = 0._wp
      rmaty(1, 3) = -b00
      rmaty(3, 1) = b00
      rmaty(3, 3) = 0._wp
    else
      phiy = atan(b(1)/b(3))
      if (b(3) .lt. 0._wp) phiy = phiy + pi
      cphiy = cos(phiy)
      sphiy = sin(phiy)
      rmaty(1, 1) = cphiy
      rmaty(1, 3) = -sphiy
      rmaty(3, 1) = sphiy
      rmaty(3, 3) = cphiy
    end if
    rmaty(2, 2) = 1._wp
    do i = 1, 3, 2
      rmaty(i, 2) = 0._wp
      rmaty(2, i) = 0._wp
    end do

    do i = 1, 4
      do j = 1, 4
        rmat(i, j) = 0._wp
        do k = 1, 4
          rmat(i, j) = rmat(i, j) + rmaty(i, k)*rmatz(k, j)
        end do
      end do
    end do
  end subroutine rotatetoz

  !-----------------------------------------------------------------------------
  !> @brief
  !> computes rotation from a frame where a is along z
  !> to frame where a is at its actual value
  !>
  !> @param[in] four vector a that is rotated onto z axis
  !> @param[out] rmat: transformation matrix from a -> a'
  !-----------------------------------------------------------------------------
  subroutine unrotatetoz(a, rmat)
    implicit none
    real(kind=wp), intent(in) :: a(4)
    real(kind=wp), intent(out) :: rmat(4, 4)
    real(kind=wp) :: rmatinv(4, 4)
    integer :: i, j

    call rotatetoz(a, rmatinv)
    do i = 1, 4
      do j = 1, 4
        rmat(i, j) = rmatinv(j, i)
      end do
    end do

  end subroutine unrotatetoz

  !> fill the openloops momenta
  subroutine fill_pOL(npar, map)
    integer, intent(in) :: npar
    integer, dimension(npar), intent(in) :: map
    integer :: i
#ifdef USE_OL
    do i = 1, npar
      kin(npar)%pOL(1:3, i) = kin(npar)%p(1:3, map(i))
      kin(npar)%pOL(0, i) = kin(npar)%p(4, map(i))
    end do
#else
    stop "Not compiled with open loops should not call fill_pOL()"
#endif
  end subroutine fill_pOL

  !> fill the openloops momenta, mapping the last two momentum entries onto their sum
  !> This is needed for some Higgs processes, where we combine the two decay momenta,
  !> to use on-shell Higgs OL matrix elements
  subroutine fill_pOLHinc(npar, map)
    integer, intent(in) :: npar
    integer, dimension(npar), intent(in) :: map
    integer :: i
#ifdef USE_OL
    do i = 1, npar - 2
      kin(npar)%pOLHinc(1:3, i) = kin(npar)%p(1:3, map(i))
      kin(npar)%pOLHinc(0, i) = kin(npar)%p(4, map(i))
    end do
    kin(npar)%pOLHinc(1:3, npar - 1) = kin(npar)%p(1:3, map(npar - 1)) + kin(npar)%p(1:3, map(npar))
    kin(npar)%pOLHinc(0, npar - 1) = kin(npar)%p(4, map(npar - 1)) + kin(npar)%p(4, map(npar))
#else
    stop "Not compiled with open loops should not call fill_pOLHinc()"
#endif
  end subroutine fill_pOLHinc

  !> fill the openloops momenta for Higgs --> l-(npar-2), l+(npar-1), photon(npar)
  subroutine fill_pOLH3decay(npar)
    integer, intent(in) :: npar
    integer :: i
#ifdef USE_OL
    do i = 2, 4
      kin(npar)%pOLH3decay(1:3, i) = kin(npar)%p(1:3, npar + i - 4)
      kin(npar)%pOLH3decay(0, i) = kin(npar)%p(4, npar + i - 4)
    end do
    kin(npar)%pOLH3decay(1:3, 1) = kin(npar)%p(1:3, npar - 2) + kin(npar)%p(1:3, npar - 1) + kin(npar)%p(1:3, npar)
    kin(npar)%pOLH3decay(0, 1) = kin(npar)%p(4, npar - 2) + kin(npar)%p(4, npar - 1) + kin(npar)%p(4, npar)
#else
    stop "Not compiled with open loops should not call fill_pOLH3decay()"
#endif
  end subroutine fill_pOLH3decay

end module KinData_mod

!----- code cemetery

!> this only works as an access function but only after the 2008 stadard
!> the function pointer is allowed to stand on the left-hand-side
!> of an assignment operator as a variable
! function p(npar,i,j)
!   real(kind=wp), pointer :: p
!   integer, intent(in) :: npar, i, j
!   p => kin(npar)%p(i,j)
! end function p
