!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: LHAPDF interface for NNLOJET [external members: `_lhapdf`]
!
! DESCRIPTION:
!> Interface to the LHAPDF_NNLOJET.cxx routines
!> needed to make LHAPDF properly threadsafe using OpenMP
!> module-private members: `m_...`
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module LHAPDF_mod
!-------------------------------------------------------------------------------
  use, intrinsic :: iso_c_binding
  implicit none
  private

  public :: init_lhapdf, destroy_lhapdf
  public :: xMin_lhapdf, xMax_lhapdf
  public :: qMin_lhapdf, qMax_lhapdf
  public :: nf_lhapdf
  public :: alphasQ_lhapdf
  public :: evolve_xfxQ_lhapdf, evolve_xf2xQ_lhapdf
  public :: has_photon_lhapdf, evolve_xfphxQ_lhapdf, evolve_xfph2xQ_lhapdf


  interface

    function m_mkPDF(setname, member) result(pdf_p) bind(c, name='mkPDF_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr) :: pdf_p
      character(kind=c_char), dimension(*) :: setname
      integer(c_int) :: member
    end function m_mkPDF

    subroutine m_rmPDF(pdf_p) bind(c, name='rmPDF_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p  ! pass by value; otherwise double-ref in cxx: (LHAPDF::PDF**)
    end subroutine m_rmPDF

    real(c_double) function m_xMin(pdf_p) bind(c, name='xMin_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_xMin

    real(c_double) function m_xMax(pdf_p) bind(c, name='xMax_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_xMax

    real(c_double) function m_qMin(pdf_p) bind(c, name='qMin_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_qMin

    real(c_double) function m_qMax(pdf_p) bind(c, name='qMax_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_qMax

    integer(c_int) function m_nf(pdf_p) bind(c, name='nf_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
    end function m_nf

    logical(c_bool) function m_hasFlavor(pdf_p, id) bind(c, name='hasFlavor_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
      integer(c_int) :: id
    end function m_hasFlavor

    real(c_double) function m_alphasQ(pdf_p, Q) bind(c, name='alphasQ_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
      real(c_double), intent(in) :: Q
    end function m_alphasQ

    real(c_double) function m_xfxQ(pdf_p, id, x, Q) bind(c, name='xfxQ_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
      integer(c_int) :: id
      real(c_double), intent(in) :: x, Q
    end function m_xfxQ

    subroutine m_evolveQ(pdf_p, x, Q, xfxQ) bind(c, name='evolvePDFQ_lhapdf_nnlojet_')
      use, intrinsic :: iso_c_binding
      type(c_ptr), value :: pdf_p
      real(c_double), intent(in) :: x, Q
      real(c_double), dimension(-6:6), intent(out) :: xfxQ
    end subroutine m_evolveQ

  end interface


  !> pointer to LHAPDF objects
  type(c_ptr) :: m_pdf1 = c_null_ptr
  type(c_ptr) :: m_pdf2 = c_null_ptr
  !$omp threadprivate(m_pdf1, m_pdf2)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise LHAPDF
  !
  !> @param[in] set  optional string for the name of the PDF (1st)
  !> @param[in] mem  optional id for the set member (1st)
  !> @param[in] set2 optional string for the name of the PDF (2nd)
  !> @param[in] mem2 optional id for the set member (2nd)
  !-----------------------------------------------------------------------------
  subroutine init_lhapdf(set, mem, set2, mem2)
    character(len=*), optional :: set, set2
    integer, optional :: mem, mem2
    integer :: member

    if ( c_associated(m_pdf1) .or. c_associated(m_pdf2) ) then
      print*, "init_lhapdf: m_pdf[1|2] is not null?!"
      print*, "calling destroy_lhapdf... (you should be doing this yourself!)"
      call destroy_lhapdf()
    end if

    if ( present(set) ) then
      member = 0
      if ( present(mem) ) member = mem
      m_pdf1 = m_mkPDF(set // c_null_char, member)
    end if

    if ( present(set2) ) then  !
      member = 0
      if ( present(mem2) ) member = mem2
      m_pdf2 = m_mkPDF(set2 // c_null_char, member)
    end if

  end subroutine init_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy dynamically allocated stuff from LHAPDF
  !-----------------------------------------------------------------------------
  subroutine destroy_lhapdf()
    if ( c_associated(m_pdf1) ) call m_rmPDF(m_pdf1)
    if ( c_associated(m_pdf2) ) call m_rmPDF(m_pdf2)
    m_pdf1 = c_null_ptr  ! no dangling pointer...
    m_pdf2 = c_null_ptr
  end subroutine destroy_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> minimum value for the momentum fraction `x`
  !
  !> @return xmin
  !-----------------------------------------------------------------------------
  function xMin_lhapdf() result(xmin)
    real(c_double) :: xmin
    xmin = m_xMin(m_pdf1)
    if ( c_associated(m_pdf2) ) then
      xmin = max(xmin, m_xMin(m_pdf2))
    end if
  end function xMin_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> maximum value for the momentum fraction `x`
  !
  !> @return xmax
  !-----------------------------------------------------------------------------
  function xMax_lhapdf() result(xmax)
    real(c_double) :: xmax
    xmax = m_xMax(m_pdf1)
    if ( c_associated(m_pdf2) ) then
      xmax = min(xmax, m_xMax(m_pdf2))
    end if
  end function xMax_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> minimum value for the factorization scale [GeV]
  !
  !> @return qmin
  !-----------------------------------------------------------------------------
  function qMin_lhapdf() result(qmin)
    real(c_double) :: qmin
    qmin = m_qMin(m_pdf1)
    if ( c_associated(m_pdf2) ) then
      qmin = max(qmin, m_qMin(m_pdf2))
    end if
  end function qMin_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> maximum value for the factorization scale [GeV]
  !
  !> @return qmax
  !-----------------------------------------------------------------------------
  function qMax_lhapdf() result(qmax)
    real(c_double) :: qmax
    qmax = m_qMax(m_pdf1)
    if ( c_associated(m_pdf2) ) then
      qmax = min(qmax, m_qMax(m_pdf2))
    end if
  end function qMax_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> number of active flavours
  !
  !> @return nf
  !-----------------------------------------------------------------------------
  function nf_lhapdf() result(nf)
    integer :: nf
    nf = m_nf(m_pdf1)
  end function nf_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if PDF set has a photon
  !
  !> @return .true. if there is a photon
  !-----------------------------------------------------------------------------
  function has_photon_lhapdf() result(has_flv)
    logical :: has_flv
    has_flv = m_hasFlavor(m_pdf1, 22)
  end function has_photon_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> the strong coupling constant from LHAPDF
  !
  !> @param[in] Q renormalization scale [GeV]
  !> @return alphas
  !-----------------------------------------------------------------------------
  function alphasQ_lhapdf(Q) result(alphas)
    real(c_double) :: alphas
    real(c_double), intent(in) :: Q
    alphas = m_alphasQ(m_pdf1, Q)
  end function alphasQ_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> evolve the PDF in (x,Q)
  !
  !> @param[in] x momentum fraction
  !> @param[in] Q factorization scale [GeV]
  !> @param[out] xfxQ PDF set (multiplied by x)
  !-----------------------------------------------------------------------------
  subroutine evolve_xfxQ_lhapdf(x, Q, xfxQ)
    real(c_double), intent(in) :: x, Q
    real(c_double), dimension(-6:6), intent(out) :: xfxQ
    call m_evolveQ(m_pdf1, x, Q, xfxQ)
  end subroutine evolve_xfxQ_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> evolve the photon PDF in (x,Q)
  !
  !> @param[in] x momentum fraction
  !> @param[in] Q factorization scale [GeV]
  !> @param[out] xfphxQ PDF set + photon (multiplied by x)
  !-----------------------------------------------------------------------------
  subroutine evolve_xfphxQ_lhapdf(x, Q, xfphxQ)
    real(c_double), intent(in) :: x, Q
    real(c_double), dimension(-6:7), intent(out) :: xfphxQ
    call m_evolveQ(m_pdf1, x, Q, xfphxQ(-6:6))
    if ( m_hasFlavor(m_pdf1, 22) ) then
      xfphxQ(7) = m_xfxQ(m_pdf1, 22, x, Q)
    else
      xfphxQ(7) = 0d0
    end if
  end subroutine evolve_xfphxQ_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> evolve the 2nd PDF in (x,Q) <-- default: 1st PDF
  !
  !> @param[in] x momentum fraction
  !> @param[in] Q factorization scale [GeV]
  !> @param[out] xfxQ PDF set (multiplied by x)
  !-----------------------------------------------------------------------------
  subroutine evolve_xf2xQ_lhapdf(x, Q, xfxQ)
    real(c_double), intent(in) :: x, Q
    real(c_double), dimension(-6:6), intent(out) :: xfxQ
    if ( c_associated(m_pdf2) ) then
      call m_evolveQ(m_pdf2, x, Q, xfxQ)
    else
      call evolve_xfxQ_lhapdf(x, Q, xfxQ)
    end if
  end subroutine evolve_xf2xQ_lhapdf


  !-----------------------------------------------------------------------------
  !> @brief
  !> evolve the 2nd photon PDF in (x,Q) <-- default: 1st PDF
  !
  !> @param[in] x momentum fraction
  !> @param[in] Q factorization scale [GeV]
  !> @param[out] xfphxQ PDF set + photon (multiplied by x)
  !-----------------------------------------------------------------------------
  subroutine evolve_xfph2xQ_lhapdf(x, Q, xfphxQ)
    real(c_double), intent(in) :: x, Q
    real(c_double), dimension(-6:7), intent(out) :: xfphxQ
    if ( c_associated(m_pdf2) ) then
      call m_evolveQ(m_pdf2, x, Q, xfphxQ(-6:6))
      if ( m_hasFlavor(m_pdf2, 22) ) then
        xfphxQ(7) = m_xfxQ(m_pdf2, 22, x, Q)
      else
        xfphxQ(7) = 0d0
      end if
    else
      call evolve_xfphxQ_lhapdf(x, Q, xfphxQ)
    end if
  end subroutine evolve_xfph2xQ_lhapdf


end module LHAPDF_mod





