!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Lorentz Invariant Phase Space [external members: `_lips`]
!
! DESCRIPTION:
!> Collection of routines to manage the phase-space generators
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module LIPS_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_lips, destroy_lips
  public :: isActive_reduce_lips, setActive_reduce_lips
  public :: weight_lips, qpass_check_lips

  interface weight_lips
    module procedure weight_1set
    module procedure weight_2set
  end interface

  logical :: qreduce_lips = .false.  ! default off

  !> information on:  iphase = 1...nphase_lips
  integer :: nphase_lips = -1  ! max value `Nphase` can be small if ang rotn are switched off.
  integer, allocatable, dimension(:) :: igroup_lips
  integer, allocatable, dimension(:,:) :: permutation_lips


contains

  !> take data i1/.../
  !> for hadron-hadron: (1,2) -> (npar-1,npar)
  !> add (1,2) @ front
  !> re-order first perm to (1,2,3,...,npar)
  subroutine init_lips(npar, iregion)
      use Process_mod, only : name_proc
      integer, intent(in) :: npar, iregion
      integer :: i, i_rot, i_sub, i_grp, i_prm

      if ( .not. isActive_reduce_lips() ) return

      print*, "init_lips", npar, "[",name_proc,"]"
      nphase_lips = -1  ! invalid state

      if ( (name_proc(1:2) == 'HJ') .or.  &
         & (name_proc(1:6) == 'HTO2PJ') .or.  &
         & (name_proc(1:8) == 'HTO2TAUJ') .or.  &
         & (name_proc(1:3) == 'WPJ') .or.  &
         & (name_proc(1:3) == 'WMJ') .or.  &
         & (name_proc(1:3) == 'GGJ') .or.  &
         & (name_proc(1:2) == 'ZJ') ) then
        !> phasenewh phasenewGG
        select case(npar)
        case (4)
          !> phase2h
          nphase_lips = 1
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          igroup_lips(1) = 1
          permutation_lips(:,1) = (/ 1,2,3,4 /)
        case (5)
          !> phase3h, phase3GG
          nphase_lips = 1
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          igroup_lips(1) = 1
          permutation_lips(:,1) = (/ 1,2,3,4,5 /)
        case (6)
          !> phase4h, phase4GG
          nphase_lips = 2 * 2
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          do i=1,nphase_lips
            igroup_lips(i) = i
            permutation_lips(:,i) = (/ 1,2,3,4,5,6 /)
          end do
        case (7)
          !> phase5ah, phase5bh, phase5allh, phase5aGG, phase5bGG, phase5allGG
          nphase_lips = 6 * 4
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          do i=1,nphase_lips
            i_rot = (i-1) / 6 + 1
            i_sub = mod((i-1), 6) + 1
            i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
            i_prm = (i_sub-1) / 2 + 1
            print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
            igroup_lips(i) = i_grp
            select case(i_prm)
            case(1)
              permutation_lips(:,i) = (/ 1,2,3,4,5,6,7 /)
            case(2)
              permutation_lips(:,i) = (/ 1,2,5,3,4,6,7 /)
            case(3)
              permutation_lips(:,i) = (/ 1,2,4,5,3,6,7 /)
            case default
              stop "init_lips: invalid permutation index for phase5[a|b|all]h!"
            end select
          end do
        case default
          print*, "init_lips: invalid npar for phasenewh:", npar
          stop
        end select

      else if ( name_proc(2:)  == 'JET' .or.  &
              & name_proc(2:)  == 'JETFC' .or.  &
              & name_proc(1:2) == 'GJ' .or.  &
              & name_proc      == 'G_EW' .or.  &
              & name_proc      == 'GG' .or.  &
              & name_proc      == 'GG_FRAG' .or.  &
              & name_proc      == 'GGG' .or.  &
              & name_proc      == 'GG_EW' ) then
        !> phasenew
        select case(npar)

        case (4)
          !> phase2n
          nphase_lips = 1
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          igroup_lips(1) = 1
          permutation_lips(:,1) = (/ 1,2,3,4 /)

        case (5)
          !> phase3n
          nphase_lips = 6 * 2
          allocate(igroup_lips(1:nphase_lips))
          allocate(permutation_lips(1:npar,1:nphase_lips))
          do i=1,nphase_lips
            i_rot = (i-1) / 6 + 1
            i_sub = mod((i-1), 6) + 1
            i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
            i_prm = (i_sub-1) / 2 + 1
            print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
            igroup_lips(i) = i_grp
            select case(i_prm)
            case (1)
              permutation_lips(:,i) = (/ 1,2,3,4,5 /)
            case (2)
              permutation_lips(:,i) = (/ 1,2,4,5,3 /)
            case (3)
              permutation_lips(:,i) = (/ 1,2,5,3,4 /)
            case default
              stop "init_lips: invalid permutation index for phase5[a|b|all]h!"
            end select
          end do

        case (6)
          select case (iregion)
          case (1)
            !> phase4an
            nphase_lips = 24 * 4
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 24 + 1
              i_sub = mod((i-1), 24) + 1
              i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
              i_prm = (i_sub-1) / 2 + 1
              print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
              igroup_lips(i) = i_grp
              select case(i_prm)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,3,4,6,5 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,6,4,3,5 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,6,4,5,3 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,6,3,4,5 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,6,3,5,4 /)
              case (7)
                permutation_lips(:,i) = (/ 1,2,5,4,3,6 /)
              case (8)
                permutation_lips(:,i) = (/ 1,2,5,4,6,3 /)
              case (9)
                permutation_lips(:,i) = (/ 1,2,5,3,4,6 /)
              case (10)
                permutation_lips(:,i) = (/ 1,2,5,3,6,4 /)
              case (11)
                permutation_lips(:,i) = (/ 1,2,5,6,4,3 /)
              case (12)
                permutation_lips(:,i) = (/ 1,2,5,6,3,4 /)
              case default
                stop "init_lips: invalid permutation index for phase4an!"
              end select
            end do
          case (2)
            !> phase4bn
            nphase_lips = 6 * 4
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 6 + 1
              i_sub = mod((i-1), 6) + 1
              print*, i, "rot", i_rot, "sub", i_sub
              igroup_lips(i) = i_rot
              select case(i_sub)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,6,4,5,3 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,6,3,5,4 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,5,4,6,3 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,5,3,6,4 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,5,6,3,4 /)
              case default
                stop "init_lips: invalid permutation index for phase4bn!"
              end select
            end do
          case (0)
            !> phase4alln
            nphase_lips = 6 * 4
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 6 + 1
              i_sub = mod((i-1), 6) + 1
              print*, i, "rot", i_rot, "sub", i_sub
              igroup_lips(i) = i_rot
              select case(i_sub)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,6,4,5,3 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,6,3,5,4 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,5,4,6,3 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,5,3,6,4 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,5,6,3,4 /)
              case default
                stop "init_lips: invalid permutation index for phase4alln!"
              end select
            end do
          case default
            stop "init_lips: invalid region for phase4[a|b|alln]!"
          end select

        case (7)
          select case (iregion)
          case (1)
            !> phase5an
            nphase_lips = 60 * 4
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 60 + 1
              i_sub = mod((i-1), 60) + 1
              i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
              i_prm = (i_sub-1) / 2 + 1
              print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
              igroup_lips(i) = i_grp
              select case(i_prm)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6,7 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,4,3,5,6,7 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,5,3,4,6,7 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,7,4,5,6,3 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,4,7,5,6,3 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,5,7,4,6,3 /)
              case (7)
                permutation_lips(:,i) = (/ 1,2,7,3,5,6,4 /)
              case (8)
                permutation_lips(:,i) = (/ 1,2,3,7,5,6,4 /)
              case (9)
                permutation_lips(:,i) = (/ 1,2,5,3,7,6,4 /)
              case (10)
                permutation_lips(:,i) = (/ 1,2,7,3,4,6,5 /)
              case (11)
                permutation_lips(:,i) = (/ 1,2,3,7,4,6,5 /)
              case (12)
                permutation_lips(:,i) = (/ 1,2,4,3,7,6,5 /)
              case (13)
                permutation_lips(:,i) = (/ 1,2,6,4,5,7,3 /)
              case (14)
                permutation_lips(:,i) = (/ 1,2,4,6,5,7,3 /)
              case (15)
                permutation_lips(:,i) = (/ 1,2,5,6,4,7,3 /)
              case (16)
                permutation_lips(:,i) = (/ 1,2,6,3,5,7,4 /)
              case (17)
                permutation_lips(:,i) = (/ 1,2,3,6,5,7,4 /)
              case (18)
                permutation_lips(:,i) = (/ 1,2,5,6,3,7,4 /)
              case (19)
                permutation_lips(:,i) = (/ 1,2,6,3,4,7,5 /)
              case (20)
                permutation_lips(:,i) = (/ 1,2,3,6,4,7,5 /)
              case (21)
                permutation_lips(:,i) = (/ 1,2,4,6,3,7,5 /)
              case (22)
                permutation_lips(:,i) = (/ 1,2,6,7,5,3,4 /)
              case (23)
                permutation_lips(:,i) = (/ 1,2,7,6,5,3,4 /)
              case (24)
                permutation_lips(:,i) = (/ 1,2,5,6,7,3,4 /)
              case (25)
                permutation_lips(:,i) = (/ 1,2,6,7,4,3,5 /)
              case (26)
                permutation_lips(:,i) = (/ 1,2,7,6,4,3,5 /)
              case (27)
                permutation_lips(:,i) = (/ 1,2,4,6,7,3,5 /)
              case (28)
                permutation_lips(:,i) = (/ 1,2,6,7,3,4,5 /)
              case (29)
                permutation_lips(:,i) = (/ 1,2,7,6,3,4,5 /)
              case (30)
                permutation_lips(:,i) = (/ 1,2,3,6,7,4,5 /)
              case default
                stop "init_lips: invalid permutation index for phase5an!"
              end select
            end do
          case (2)
            !> phase5bn
            nphase_lips = 30 * 4
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 30 + 1
              i_sub = mod((i-1), 30) + 1
              i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
              i_prm = (i_sub-1) / 2 + 1
              print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
              igroup_lips(i) = i_grp
              select case(i_prm)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6,7 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,4,3,5,6,7 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,5,3,4,6,7 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,7,4,5,6,3 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,4,7,5,6,3 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,5,7,4,6,3 /)
              case (7)
                permutation_lips(:,i) = (/ 1,2,7,3,5,6,4 /)
              case (8)
                permutation_lips(:,i) = (/ 1,2,3,7,5,6,4 /)
              case (9)
                permutation_lips(:,i) = (/ 1,2,5,3,7,6,4 /)
              case (10)
                permutation_lips(:,i) = (/ 1,2,7,3,4,6,5 /)
              case (11)
                permutation_lips(:,i) = (/ 1,2,3,7,4,6,5 /)
              case (12)
                permutation_lips(:,i) = (/ 1,2,4,3,7,6,5 /)
              case (13)
                permutation_lips(:,i) = (/ 1,2,6,4,5,7,3 /)
              case (14)
                permutation_lips(:,i) = (/ 1,2,6,3,5,7,4 /)
              case (15)
                permutation_lips(:,i) = (/ 1,2,6,3,4,7,5 /)
              case default
                stop "init_lips: invalid permutation index for phase5bn!"
              end select
            end do
          case (0)
            !> phase5alln
            nphase_lips = 20 * 8
            allocate(igroup_lips(1:nphase_lips))
            allocate(permutation_lips(1:npar,1:nphase_lips))
            do i=1,nphase_lips
              i_rot = (i-1) / 20 + 1
              i_sub = mod((i-1), 20) + 1
              i_grp = 2*i_rot - 1 + mod((i_sub-1), 2)
              i_prm = (i_sub-1) / 2 + 1
              print*, i, "rot", i_rot, "sub", i_sub, "grp", i_grp, "prm", i_prm
              igroup_lips(i) = i_grp
              select case(i_prm)
              case (1)
                permutation_lips(:,i) = (/ 1,2,3,4,5,6,7 /)
              case (2)
                permutation_lips(:,i) = (/ 1,2,7,4,5,6,3 /)
              case (3)
                permutation_lips(:,i) = (/ 1,2,7,3,5,6,4 /)
              case (4)
                permutation_lips(:,i) = (/ 1,2,7,3,4,6,5 /)
              case (5)
                permutation_lips(:,i) = (/ 1,2,6,4,5,7,3 /)
              case (6)
                permutation_lips(:,i) = (/ 1,2,6,3,5,7,4 /)
              case (7)
                permutation_lips(:,i) = (/ 1,2,6,3,4,7,5 /)
              case (8)
                permutation_lips(:,i) = (/ 1,2,6,7,5,3,4 /)
              case (9)
                permutation_lips(:,i) = (/ 1,2,6,7,4,3,5 /)
              case (10)
                permutation_lips(:,i) = (/ 1,2,6,7,3,4,5 /)
              case default
                stop "init_lips: invalid permutation index for phase5alln!"
              end select
            end do
          case default
            stop "init_lips: invalid region for phase5alln!"
          end select

        case default
          print*, "init_lips: invalid npar for phasenew:", npar
          stop
        end select

      else
        print*, "init_lips: so far, only hard-coded stuff..."
      end if

      !> digest the output for checks
      do i=1,nphase_lips
        print*, i, "[",igroup_lips(i),"]: ", permutation_lips(:,i)
      end do

  end subroutine init_lips


  subroutine destroy_lips()
    nphase_lips = -1
    if ( allocated(igroup_lips) )      deallocate(igroup_lips)
    if ( allocated(permutation_lips) ) deallocate(permutation_lips)
  end subroutine destroy_lips



  logical function qpass_check_lips()
    use KinData_mod
    use FPA_mod
    integer :: i, j, k, npar, i_1st, i_grp
    logical :: qfail
    double precision, save, allocatable, dimension(:,:,:) :: p_sav
!$omp threadprivate(p_sav)
    !> the current status
    integer :: Nphase, iphase
    common /Nphase/ Nphase
!$omp threadprivate(/Nphase/)
    common /iphase/ iphase
!$omp threadprivate(/iphase/)

    qpass_check_lips = .true.

    if ( .not. isActive_reduce_lips() ) return

    npar = size(permutation_lips, 1)

    if (iphase == 1) then
      !> use first entry to allocate & reset
      !> not at all optimized for performance
      if (allocated(p_sav)) deallocate(p_sav)
    end if

    if (.not.allocated(p_sav)) then
      allocate(p_sav(1:4, 1:npar, 1:nphase_lips))
      p_sav = 0d0
    end if


    !> copy KinData entry
    do i=1,npar
      p_sav(1:4, i, iphase) = kin(npar)%p(1:4, i)
    end do

    !> match to first in the group
    i_1st = iphase
    i_grp = igroup_lips(iphase)
    do i = 1, Nphase
      if (igroup_lips(i) /= i_grp) cycle
      i_1st = i
      exit
    end do

    print*, "---"
    qfail = .false.
    do i=1,npar
      print*, " > map: ", permutation_lips(i,iphase), "->", permutation_lips(i,i_1st)
      do k=1,4
        !> check if values match
        if ( .not. is_eq( p_sav(k,permutation_lips(i,iphase),iphase),  &
                        & p_sav(k,permutation_lips(i,i_1st),i_1st) ) ) then
          qfail = .true.
        end if
      end do
    end do

    if (qfail) then
      print*, "qpass_check_lips: mismatch for iphase = ", iphase, " > ", i_1st
      print*, " > nth", iphase, ">", permutation_lips(:,iphase)
      do i=1,npar
        print*, i,":", p_sav(1:4, i, iphase)
      end do
      print*, " > 1st", i_1st, ">", permutation_lips(:,i_1st)
      do i=1,npar
        print*, i,":", p_sav(1:4, i, i_1st)
      end do
      ! stop
      qpass_check_lips = .false.
    end if

  end function qpass_check_lips


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of LIPS reduction
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_reduce_lips()
    isActive_reduce_lips = qreduce_lips
    !> would kill init_lips .and. (nphase_lips >= 0)
  end function isActive_reduce_lips


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate LIPS reduction
  !
  !> @param[in] active boolean flag to switch on/off
  !-----------------------------------------------------------------------------
  subroutine setActive_reduce_lips(active)
    logical, intent(in) :: active
    ! print*, "setActive_reduce_lips: setting qreduce_lips to: ", active
    qreduce_lips = active
  end subroutine setActive_reduce_lips


!  !-----------------------------------------------------------------------------
!  !> @brief
!  !> map to the 1st match (permutation)
!  !
!  !> @param[in] idset array of indices associated with the same particle
!  !> @param[out] mult multiplicity of the iphase that was
!  !> @return iphase of the reduced LIPS
!  !-----------------------------------------------------------------------------
!  integer function map_1st_1set(idset)
!    integer, intent(in) :: idset(:)
!    integer, intent(out) :: mult
!    integer :: i, i_grp
!    logical, allocatable, dimension(:) :: qperm
!    !> the current status
!    integer :: Nphase, iphase
!    common /Nphase/ Nphase
!!$omp threadprivate(/Nphase/)
!    common /iphase/ iphase
!!$omp threadprivate(/iphase/)
!
!    map_1st_1set = iphase ! always match yourself (not matched in loop below)
!    mult = 1
!    if ( .not. isActive_reduce_lips() .or. (nphase_lips <= 1) ) return
!
!    allocate(qperm(size(permutation_lips,1)))
!    qperm(:) = .false.
!    do i=1,size(idset)
!      qperm(:) = qperm(:) .or. ( permutation_lips(:,iphase) == idset(i) )
!    end do
!    ! print*, "> idset >", idset
!    ! print*, "> perm  >", permutation_lips(:,iphase)
!    ! print*, "> qperm >", qperm
!
!    i_grp = igroup_lips(iphase)
!    ! print*, "mult: ", iphase, "<>", i_grp
!    do i = 1, Nphase  ! only loop over active entries
!      if (igroup_lips(i) /= i_grp) cycle
!      if (i == iphase) cycle
!      ! print*, ">", i, ">", ( permutation_lips(:,iphase) /= permutation_lips(:,i) )
!      ! print*, ( permutation_lips(:,iphase) /= permutation_lips(:,i) ) .eqv. qperm(:)
!      if ( all( qperm(:) .or.  &
!         & ( (permutation_lips(:,iphase) /= permutation_lips(:,i)) .eqv. qperm(:) ) ) ) then
!        if (i < map_1st_1set) map_1st_1set = i
!        mult = mult + 1  ! counts total number in the perm set
!      end if
!    end do
!
!  end function map_1st_1set


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine the weight of the contribution based on the permutation symmetry
  !
  !> @param[in] idset array of indices associated with the same particle
  !> @return weight = 0 (reduced away), # of configurations reduced
  !-----------------------------------------------------------------------------
  integer function weight_1set(idset)
    integer, intent(in) :: idset(:)
    integer :: i, i_grp, i_1st
    logical, allocatable, dimension(:) :: qperm
    !> the current status
    integer :: Nphase, iphase
    common /Nphase/ Nphase
!$omp threadprivate(/Nphase/)
    common /iphase/ iphase
!$omp threadprivate(/iphase/)

    weight_1set = 1
    if ( .not. isActive_reduce_lips() .or. (nphase_lips <= 1) ) return

    allocate(qperm(size(permutation_lips,1)))
    qperm(:) = .false.
    do i=1,size(idset)
      qperm(:) = qperm(:) .or. ( permutation_lips(:,iphase) == idset(i) )
    end do
    ! print*, "> idset >", idset
    ! print*, "> perm  >", permutation_lips(:,iphase)
    ! print*, "> qperm >", qperm

    i_1st = iphase  ! always match yourself (not matched in loop below)
    i_grp = igroup_lips(iphase)
    ! print*, "weight_1set: ", iphase, "<>", i_grp
    do i = 1, Nphase
      if (igroup_lips(i) /= i_grp) cycle
      if (i == iphase) cycle
      ! print*, ">", i, ">", ( permutation_lips(:,iphase) /= permutation_lips(:,i) )
      ! print*, ( permutation_lips(:,iphase) /= permutation_lips(:,i) ) .eqv. qperm(:)
      if ( all( qperm(:) .or.  &
         & ( (permutation_lips(:,iphase) /= permutation_lips(:,i)) .eqv. qperm(:) ) ) ) then
        if (i < i_1st) i_1st = i
        weight_1set = weight_1set + 1  ! counts total number in the perm set
      end if
    end do

    !> iphase is mapped away -> reject
    if (i_1st /= iphase) weight_1set = 0

    ! if ( .not. qpass_check_lips() ) then
    !   print*, "weight_1set: ", iphase, "<>", i_grp
    !   print*, "<", permutation_lips(:,iphase)
    !   print*, ">", permutation_lips(:,i_1st)
    !   print*, " >> ", weight_1set
    !   read(*,*)
    ! end if

  end function weight_1set


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine the weight of the contribution based on the permutation symmetry
  !
  !> @param[in] idset1 array of indices associated with the same particle1
  !> @param[in] idset1 array of indices associated with the same particle2
  !> @return weight = 0 (reduced away), # of configurations reduced
  !-----------------------------------------------------------------------------
  integer function weight_2set(idset1, idset2)
    integer, intent(in) :: idset1(:), idset2(:)
    weight_2set = 1
    if ( .not. isActive_reduce_lips() .or. (nphase_lips <= 1) ) return
    !write(*,*)"WARNING! weight_2set: not implemented yet, no optimisation is active for this sub process"
  end function weight_2set


end module LIPS_mod


