!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: LeptonIsolation [external members: `_lepisol`]
!
! DESCRIPTION:
!> Module providing various prescriptions to isolate leptons from jets.
!> This is not yet compatible with MultiRun!
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module LeptonIsolation_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: setActive_lepisol, isActive_lepisol
  public :: setLeptonIsol_lepisol, digest_lepisol
  public :: registerLepton_lepisol, init_lepisol, destroy_lepisol
  public :: setDressLepton_lepisol
  public :: apply_lepisol

  logical :: qactive_lepisol = .false.

  !> for leptions we only care about fixed-cone isolations
  !> @todo: EW corrections <-> dressed/bare leptons
  double precision :: R_lepisol = 0d0
  double precision :: ETmax_const_lepisol = 0d0
  double precision :: ETmax_epsilon_lepisol = 0d0
  double precision :: R_min_lepisol = 0d0
  double precision :: ET_threshold_lepisol = 0d0

  !> possibility to dress leptons
  double precision :: R_dress_leptons = 0d0

  !> store number and location (in momentum set) of all leptons
  integer :: nleptons = 0
  integer, allocatable, dimension(:) :: ileptons


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate lepton isolation
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_lepisol(active)
    logical, intent(in), optional :: active
    if (present(active)) then
      qactive_lepisol = active
    else
      qactive_lepisol = .true.
    end if
  end subroutine setActive_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of lepton isolation
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_lepisol()
    isActive_lepisol = qactive_lepisol
  end function isActive_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the lepton isolation criterion
  !
  !> @param[in] R: the cone size for the isoilation
  !> @param[in] ETmax_const, ETmax_epsilon: defines maximum of transverse hadronic energy:
  !>            ETmax = ETmax_const + ETmax_epsilon * pT_gamma
  !> @param[in] R_min, ET_threshold: fixed-cone parameters
  !-----------------------------------------------------------------------------
  subroutine setLeptonIsol_lepisol( R, &
                                  & ETmax_const, ETmax_epsilon,  &
                                  & R_min, ET_threshold)
    double precision, optional, intent(in) :: R
    double precision, optional, intent(in) :: ETmax_const, ETmax_epsilon
    double precision, optional, intent(in) :: R_min, ET_threshold

    if ( present(R) )             R_lepisol = R
    if ( present(ETmax_const) )   ETmax_const_lepisol = ETmax_const
    if ( present(ETmax_epsilon) ) ETmax_epsilon_lepisol = ETmax_epsilon
    if ( present(R_min) )         R_min_lepisol = R_min
    if ( present(ET_threshold) )  ET_threshold_lepisol = ET_threshold

    call setActive_lepisol()

  end subroutine setLeptonIsol_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> print out a list of active isolation criteria
  !-----------------------------------------------------------------------------
  subroutine digest_lepisol(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_var =  &
      & '(1x,"*",16x,A15,"  = ",1f8.4  ,T103,"*")'
    character(len=*), parameter :: fmt_min_max =  &
      & '(1x,"*",16x,1f8.4," <= ",A15," <= ",1f8.4,T103,"*")'
    character(len=*), parameter :: fmt_min =  &
      & '(1x,"*",3x,A6,":",1f12.2," <= ",A15,59x,"*")'
    character(len=*), parameter :: fmt_max =  &
      & '(1x,"*",3x,A6,":",16x,A15," <= ",1f12.2,43x,"*")'
    character(len=*), parameter :: fmt_sel =  &
      & '(1x,"*",3x,A6,":",1f12.2," <= ",A15," <= ",1f12.2,43x,"*")'

    if (.not.isActive_lepisol()) return

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    write(iout,fmt_str) ""
    write(iout,fmt_str) "> lepton isolation <"
    write(iout,fmt_str) ""

    write(iout,fmt_min_max) R_min_lepisol, "R", R_lepisol
    write(iout,fmt_var) "ETmax_const", ETmax_const_lepisol
    write(iout,fmt_var) "ETmax_epsilon", ETmax_epsilon_lepisol
    write(iout,fmt_var) "ET_threshold", ET_threshold_lepisol

  end subroutine digest_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> register the particle indices that are associated with the leptons.
  !> negative indices correpsond to a "reverse" indexing like in Python:
  !>   -1 <--> npar
  !>   -2 <--> npar-1
  !>      ....
  !
  !> @param[in] ilepton index of a lepton in the KinData momentum array
  !-----------------------------------------------------------------------------
  subroutine registerLepton_lepisol(ilepton)
    integer, intent(in) :: ilepton
    integer :: i
    integer, allocatable, dimension(:) :: idummy
    !> we don't care about performance in this routine, so we rather make the
    !> external interface as simple as possible than asking the user to supply 
    !> an array with the entries or something similar

    !> we already have a non-empty list 
    if ( allocated(ileptons) ) then
  
      !> sanity check
      if (nleptons /= size(ileptons)) then
        print*, "registerLepton_lepisol: nleptons mismatch"
        stop
      end if

      !> check if we already know this lepton
      do i = 1,nleptons
        if (ileptons(i)==ilepton) return
      end do

      !> it's a new index: add it
      allocate(idummy(nleptons))
      idummy(1:nleptons) = ileptons(1:nleptons)
      deallocate(ileptons)
      allocate(ileptons(nleptons+1))
      ileptons(1:nleptons) = idummy(1:nleptons)
      ileptons(nleptons+1) = ilepton

    !> we have an empty list
    else  

      !> sanity check
      if (nleptons /= 0) then
        print*, "registerLepton_lepisol: nleptons mismatch"
        stop
      end if

      !> valid state: allocate & store
      allocate(ileptons(1))
      ileptons(1) = ilepton

    end if

    nleptons = nleptons + 1
  end subroutine registerLepton_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialisation routine for lepton isolation: registers
  !> leptons to test for isolation in a process-dependent array.
  !-----------------------------------------------------------------------------
  subroutine init_lepisol()
    use Process_mod, only : name_proc

    !> first destroy any allocation just in case
    call destroy_lepisol()

    !> have to register the leptons on a proces--by-process basis
    select case (name_proc)

      !> H -> ZZ* -> 4l (last four entries in the momentum array)
      case ( "HTO4E", "HTO4EJ", "HTO4EJJ", "HTO4EJJJ", &
           & "HTO4EM", "HTO4EJM",  &
           & "VFHTO4E", "VFHTO4EJ",  &
           & "HTO4EINC", &
           & "HTO2E2MU", "HTO2E2MUJ", "HTO2E2MUJJ", "HTO2E2MUJJJ",  &
           & "HTO2E2MUM", "HTO2E2MUJM",  &
           & "VFHTO2E2MU", "VFHTO2E2MUJ",  &
           & "HTO2E2MUINC" )
        call registerLepton_lepisol(-1)
        call registerLepton_lepisol(-2)
        call registerLepton_lepisol(-3)
        call registerLepton_lepisol(-4)

      !> H -> Z/gamma* gamma -> 2l gamma (last entry: photon, then: 2 leptons)
      case ( "HTO2L1P", "HTO2L1PJ", "HTO2L1PJJ", "HTO2L1PJJJ",  &
           & "HTO2L1PM", "HTO2L1PJM",  &
           & "VFHTO2L1P", "VFHTO2L1PJ",  &
           & "HTO2L1PINC" )
        call registerLepton_lepisol(-2)
        call registerLepton_lepisol(-3)

      !> H -> WW
      case ( "HTO2L2N", "HTO2L2NJ", "HTO2L2NJJ", "HTO2L2NJJJ",  &
           & "HTO2L2NM", "HTO2L2NJM",  &
           & "VFH2L2N", "VFH2L2NJ",  &
           & "HTO2L2NINC" )
        !> function HWW2l2n0(i1,i2,i3,i4,ipset)
        !> i1=l1+(l1-), i2=n1(n1bar), i3=l2-(l2+), i4=n2bar(n2).
        call registerlepton_lepisol(-2)
        call registerlepton_lepisol(-4)

      ! case default
      !   print*, "init_lepisol: process ", name_proc, " has no leptons?!"
      !   stop

    end select

  end subroutine init_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> clear leptons and isolation algorithms
  !-----------------------------------------------------------------------------
  subroutine destroy_lepisol()
    if (allocated(ileptons)) deallocate(ileptons)
    nleptons = 0
  end subroutine destroy_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the cone size for photon recombination
  !
  !> @param[in] Rcone Delta R cone
  !-----------------------------------------------------------------------------
  subroutine setDressLepton_lepisol(Rcone)
    double precision, intent(in) :: Rcone
    if (Rcone < 0d0) then
      print*, "setDressLepton_lepisol: invalid Rcone = ", Rcone
      stop
    end if
    R_dress_leptons = Rcone
  end subroutine setDressLepton_lepisol


  !-----------------------------------------------------------------------------
  !> @brief
  !> Applies fixed-cone isolation to leptons
  !
  !> @param[in] npar kinematics specified by the number of external partlicles
  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
  !-----------------------------------------------------------------------------
  subroutine apply_lepisol(npar, ipass)
    use KinData_mod
    use ObsHelper_mod
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer :: i,j, j_kd, ilepton
    double precision :: Rgj, sumET, ETmax, ETj
    type(Kin_t), pointer :: kd

    if ( (ipass == 0) .or. (.not.isActive_lepisol()) ) return ! nothing to do

    !> prettify code
    kd => kin(npar)

    !> apply isolation criterion to each lepton sequentially
    loop_leptons: do i = 1,nleptons
      ilepton = ileptons(i)
      if (ilepton < 0) then ! see "reverse" counting convention above
        ilepton = npar + ilepton + 1
      end if
      !> maximum allowed hadronic (partonic) ET inside the lepton cone
      ETmax = ETmax_const_lepisol  &
        &   + ETmax_epsilon_lepisol * v1_et(kd%p(:,ilepton))
      !> accumulate ET of partons inside the delta_etsum cone around the lepton
      sumET = 0d0
      do j = 1, kd%nproto
        !> index in KinData
        j_kd = kd%ijets_lower + j - 1
        Rgj = v2_delta_R( kd%p(:,ilepton) , kd%p(:,j_kd) )
        !> skip partons outside of the cone
        if ( (Rgj > R_lepisol) .or. (Rgj < R_min_lepisol) ) cycle
        ETj = v1_et( kd%p(:,j_kd) )
        !> if an ET threshold is set (>0), only keep partons above that theshold
        if ( (ET_threshold_lepisol > 0d0) .and. (ETj < ET_threshold_lepisol) ) cycle
        !> passed parton requirements => accumulate
        sumET = sumET + ETj
      end do  ! j <-> partons
      !> check condition: sumET < ETmax
      if ( sumET > ETmax ) then
        ipass = 0
        return
      end if
    end do loop_leptons

  end subroutine apply_lepisol


end module LeptonIsolation_mod

