!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Mapping [external members: `_map`]
!
! DESCRIPTION:
!> Store the full information on the mappings.
!> @todo:
!> eventually port over all libmap routines into this module
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Mapping_mod
!-------------------------------------------------------------------------------
  use KinData_mod
  use PhotonFragmentation_mod
  use HadronFragmentation_mod
  implicit none
  private

  public :: init_map, destroy_map
  public :: set_map, unset_map
  public :: getCurrent_map, idxUpCurrent_map
  public :: hasParent_map, getParent_map, idxUpParent_map

  public :: j_map, k_map
  public :: j1,j2,j3,j4,j5,j6,j7,j8,j9
  public :: k1,k2,k3,k4,k5,k6,k7,k8

  public :: isActiveCache_map, setActiveCache_map, flushCache_map

  type Mapping_t
    integer :: npar_from = -1
    integer :: npar_to   = -1
    !> the antenna defining the mapping
    integer, allocatable, dimension(:) :: antenna
    !> for mapping `j<-i` we have: `mapping(j) = i` (up-ward mapping)
    integer, allocatable, dimension(:) :: mapping
    !> type: 1=single, 11=single-single, 2=double
    integer :: type = -1
    integer, dimension(2) :: reduced
    integer :: hfrag = -1
  end type Mapping_t


  !> the four different mappings that can occur:
  !>   * i: highest momentum set
  !>   * j: single-unresolved momentum set
  !>   * k: double-unresolved momentum set
  !> (X3)      i --> j
  !> (X3 x X3) i --> j --> k
  !> (X4)      i --> k
  type(Mapping_t), target :: mapitoj, mapjtok, mapitok
!$OMP THREADPRIVATE(mapitoj, mapjtok, mapitok)

  !> pointers for the single- & double-unresolved mappings
  !> that are currently active.
  type(Mapping_t), pointer :: current_p => null()
  ! type(Mapping_t), pointer :: singleUnresolved_p => null()
  ! type(Mapping_t), pointer :: doubleUnresolved_p => null()
!$OMP THREADPRIVATE(current_p)

  integer, allocatable, dimension(:), target :: j_map, k_map
!$OMP THREADPRIVATE(j_map,k_map)
  !> prettification & legacy-compatibility
  integer, pointer :: j1,j2,j3,j4,j5,j6,j7,j8,j9
!$OMP THREADPRIVATE(j1,j2,j3,j4,j5,j6,j7,j8,j9)
  integer, pointer :: k1,k2,k3,k4,k5,k6,k7,k8
!$OMP THREADPRIVATE(k1,k2,k3,k4,k5,k6,k7,k8)


  !////////////////////////
  !//  Momentum Caching  //
  !////////////////////////

  logical :: qcache_map = .false.  ! off by default

  type MapCache_t
    logical :: exists = .false.
    integer :: ipass = 1
    type(Kin_t), pointer :: kd => null()
  end type MapCache_t

  type(MapCache_t), allocatable, dimension(:,:,:,:,:,:), target :: cache_map
!$OMP THREADPRIVATE(cache_map)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the mapping module
  !> (all dynamical allocation at run-time so not much happening here)
  !
  !> @param[in] max_npar optional set highest momentum set
  !-----------------------------------------------------------------------------
  subroutine init_map(max_npar)
    integer, intent(in), optional :: max_npar
    integer :: N_
    N_ = 9  ! largest set we support so far
    if ( present(max_npar) ) N_ = max_npar  ! optimize
    !> allocation is done dynamically:
    !> we don't allocate jtok, itok for R, neither itoj in LO, etc.
    nullify(current_p)
    !> actually, just call destroy...
    call destroy_map()
    !> initialisae the types of the three possible mappings
    mapitoj%type = 1
    mapjtok%type = 11
    mapitok%type = 2
    !> init cache size
    if (qcache_map)  &
      &  allocate(cache_map(0:N_,0:N_,0:N_,0:N_,0:N_,0:N_))
  end subroutine init_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy any allocated memory
  !> (call at the end of an parallel block)
  !-----------------------------------------------------------------------------
  subroutine destroy_map()
    !> reset to a "clean" state and deallocate everything
    mapitoj%npar_from = -1
    mapitoj%npar_to   = -1
    if(allocated(mapitoj%antenna)) deallocate(mapitoj%antenna)
    if(allocated(mapitoj%mapping)) deallocate(mapitoj%mapping)

    mapjtok%npar_from = -1
    mapjtok%npar_to   = -1
    if(allocated(mapjtok%antenna)) deallocate(mapjtok%antenna)
    if(allocated(mapjtok%mapping)) deallocate(mapjtok%mapping)

    mapitok%npar_from = -1
    mapitok%npar_to   = -1
    if(allocated(mapitok%antenna)) deallocate(mapitok%antenna)
    if(allocated(mapitok%mapping)) deallocate(mapitok%mapping)

    if(allocated(j_map)) deallocate(j_map)
    if(allocated(k_map)) deallocate(k_map)
    nullify(j1,j2,j3,j4,j5,j6,j7,j8,j9)
    nullify(k1,k2,k3,k4,k5,k6,k7,k8)

    nullify(current_p)

    if (allocated(cache_map)) deallocate(cache_map)
  end subroutine destroy_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status on kinematics caching
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActiveCache_map()
    isActiveCache_map = qcache_map
  end function isActiveCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate caching of mapped momentum sets
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActiveCache_map(active)
    logical, intent(in) :: active
    qcache_map = active
    if (qcache_map) then
      print*, "setActiveCache_map: turning on momentum cache..."
    end if
  end subroutine setActiveCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset momentum cache (without deallocation)
  !> needs to be called for each phase-space point
  !-----------------------------------------------------------------------------
  subroutine flushCache_map()
    if (.not.qcache_map) return
    cache_map%exists = .false.
  end subroutine flushCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> retrieve the cache location (pointer) of the active mapping (`current_p`)
  !-----------------------------------------------------------------------------
  subroutine getCurrentCache_map(cache_p)
    type(MapCache_t), pointer :: cache_p

    nullify(cache_p)
    if (.not.qcache_map) return

    select case (current_p%type)
    case (1)   ! single
#ifndef NDEBUG
      !> assert
      if ( .not.associated(current_p, mapitoj)  .or.  &
         & (size(current_p%antenna) /= 3)       ) then
        stop "getCurrentCache_map: assert failed for type 1"
      end if
#endif
      cache_p => cache_map( current_p%antenna(1),  &
                          & current_p%antenna(2),  &
                          & current_p%antenna(3),  &
                          & 0, 0, 0)
    case (11)  ! single-single
#ifndef NDEBUG
      !> assert
      if ( .not.associated(current_p, mapjtok)       .or.  &
        &  (size(current_p%antenna) /= 3)            .or.  &
        &  (current_p%npar_from /= mapitoj%npar_to)  .or.  &
        &  (size(mapitoj%antenna) /= 3)              ) then
        stop "getCurrentCache_map: assert failed for type 11"
      end if
#endif
      cache_p => cache_map( current_p%antenna(1),  &
                          & current_p%antenna(2),  &
                          & current_p%antenna(3),  &
                          & mapitoj%antenna(1),  &
                          & mapitoj%antenna(2),  &
                          & mapitoj%antenna(3) )
    case (2)   ! double
#ifndef NDEBUG
      !> assert
      if ( .not.associated(current_p, mapitok)  .or.  &
         & (size(current_p%antenna) /= 4)       ) then
        stop "getCurrentCache_map: assert failed for type 2"
      end if
#endif
      cache_p => cache_map( current_p%antenna(1),  &
                          & current_p%antenna(2),  &
                          & current_p%antenna(3),  &
                          & current_p%antenna(4),  &
                          & 0, 0)
    case default
      print*, "getCurrentCache_map: invalid type: ", current_p%type
      stop
    end select

  end subroutine getCurrentCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> read in kinematics from cache if stored
  !
  !> @param[inout] ipass flag to reject mapped momentum set
  !> @return success boolean flag indicating successful load from cache
  !-----------------------------------------------------------------------------
  function readCache_map(ipass) result(success)
    logical :: success
    integer, intent(inout) :: ipass
    type(MapCache_t), pointer :: cache_p

    success = .false.
    if ( (.not.qcache_map) .or. (ipass == 0) ) return

    call getCurrentCache_map(cache_p)

    success = cache_p%exists
    if (.not.success) return

    ipass = cache_p%ipass
    !> no copy needed if set was rejected anyway
    if (ipass == 0) return

    !> valid momentum set => copy to KinData
    !print*, "reading from cache", current_p%npar_to
    call overwrite_kin(cache_p%kd)

  end function readCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> read in kinematics from cache if stored
  !
  !> @param[in] ipass flag to reject mapped momentum set
  !-----------------------------------------------------------------------------
  subroutine writeCache_map(ipass)
    integer, intent(in) :: ipass
    type(MapCache_t), pointer :: cache_p

    if (.not.qcache_map) return

    call getCurrentCache_map(cache_p)

    cache_p%exists = .true.
    cache_p%ipass = ipass
    !> no copy needed if set was rejected anyway
    if (ipass == 0) return

    !> valid momentum set => copy to cache
    !print*, "writing into cache", current_p%npar_to
    call copy_kin(current_p%npar_to, cache_p%kd)

  end subroutine writeCache_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the current state of the mapping
  !-----------------------------------------------------------------------------
  subroutine unset_map()
    !> null pointer means unset...
    !> deallocation of stuff is done in destroy_map for performance
    nullify(current_p)
  end subroutine unset_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> perform a mapping
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] antenna array of indices defining the mapping (in npar_from)
  !> @param[in] pmap array to define the mapping
  !>            red[a] <-> parent[b]: pmap(a) = b  (no parent[{unresolved}])
  !> @param[inout] ipass flag to reject mapped momentum set
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)
  !-----------------------------------------------------------------------------
  subroutine set_map(npar_from, npar_to, antenna, pmap, ipass, frag, hfrag)
    use KinData_mod, only : kin
    integer, intent(in) :: npar_from, npar_to
    integer, intent(in), dimension(:) :: antenna, pmap
    integer, intent(inout) :: ipass
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    integer :: delta, ninitial, i,j,k, ishift
    integer, dimension(2) :: IF_swap
    integer, dimension(:), pointer :: red_set

    ! print*, "set_map:", npar_from, npar_to, ">", pmap
    ! if ( present(frag) ) print*, "  > frag", frag

    if (ipass==0) return

#ifndef NDEBUG
    if ( present(frag) ) then
      if ( size(frag) .lt. 2 .or. size(frag) .gt. 3 ) then
        stop "set_map: wrong size of frag array?!"
      end if
      !> @note: this is absolutely wrong!!
      !>  (look @ qgB1g0GGSNLO_EW: would not work for incl. photon setup)
      !>  left for future reference
      ! !> no need to evaluate frag blocks if fragmentation is inactive
      ! if ( .not.isActive_frag() ) then
      !   ipass = 0
      !   return
      ! end if
    end if

    if ( present(hfrag) ) then
      if ( kin(npar_from)%ifrag /= hfrag ) then
        print*, "set_map: inconsistent hfrag index vs. kin%ifrag", hfrag, kin(npar_from)%ifrag
        stop
      end if
    end if
#endif
    
    delta = npar_from - npar_to

#ifndef NDEBUG
    !> check compatibility of assumed-shape array sizes
    if ( size(antenna) /= 2+delta ) then
      print*, "set_map: incompatible antenna size:", npar_from, npar_to, antenna
      stop
    end if
    if ( size(pmap) /= npar_to ) then
      return
      print*, "set_map: incompatible pmap size:", npar_to, size(pmap), pmap
      stop
    end if
#endif

    nullify(red_set)
    select case(delta)
    case(1)
      if ( associated(current_p) ) then
        if ( npar_from == current_p%npar_to ) then
          current_p => mapjtok  ! 2nd single-unresolved
          call set_k_map(npar_to)
          ! print*, "set_map: jtok mapping:", antenna
        else
          print*, "set_map: forgot to unset_map?",  &
            & npar_from, npar_to, current_p%npar_from, current_p%npar_to
          stop
        end if
      else
        current_p => mapitoj  ! single-unresolved
        call set_j_map(npar_to)
        ! print*, "set_map: itoj mapping:", antenna
      end if
    case(2)
      current_p => mapitok  ! double-unresolved
      call set_k_map(npar_to)
      ! print*, "set_map: itok mapping:", antenna
    case default
      print*, "set_map: invalid input", npar_from, npar_to
      stop
    end select

#ifndef NDEBUG
    !> check if type matches delta
    if ( mod(current_p%type,10) /= delta ) then
      print*, "set_map: type mismatch", current_p%type, delta
      stop
    end if
    !> check set_X_map were called correclty
    if ( .not. associated(red_set) ) then
      print*, "set_map: red_set not associated"
      stop
    end if
#endif

    !> initialise Mapping & store info (type set in init_map)
    current_p%npar_from  = npar_from
    current_p%npar_to    = npar_to
    if (.not.allocated(current_p%antenna)) then
      allocate(current_p%antenna(2+delta))
    ! else check size compatibility?!
    end if
    if (.not.allocated(current_p%mapping)) then
      allocate(current_p%mapping(npar_to))
    ! else check size compatibility?!
    end if
    current_p%reduced(:) = -1
    if ( present(hfrag) ) then
      current_p%hfrag = hfrag
    else
      current_p%hfrag = -1
    end if
    ninitial = count( antenna <= 2 )

    !> start by copying antenna array
    do i=1,2+delta
      current_p%antenna(i) = antenna(i)
    end do
    !> then enforce a "normal ordering" of the mapping identifier
    !> ATTENTION: this is for the mapping identification!
    !>            only the IF swap can affect pmap because
    !>            it swaps a hard with an unresolved index!
    !>            for IF we use `IF_swap` to patch `pmap`
    IF_swap(:) = -1  ! no swap
    select case (ninitial)
    case (0)  ! FF
      if ( antenna(1) > antenna(2+delta) ) then
        !> invert the ordering (note: relative order matters!)
        do i=1,2+delta
          current_p%antenna(i) = antenna(delta+3-i)
        end do
      end if
    case (1)  ! IF
      !> sort the final-state particles
      do i=2,2+delta
        !> find the biggest entry j in [i, i+1, ...] (including hard)
        j = i
        do k=i+1,2+delta
          if ( current_p%antenna(k) > current_p%antenna(j) ) j = k
        end do
        if ( j /= i ) then
          !> swap
          k = current_p%antenna(i)
          current_p%antenna(i) = current_p%antenna(j)
          current_p%antenna(j) = k
        end if
      end do
      !> if the hard FS radiatior was swapped: save!
      if ( antenna(2+delta) /= current_p%antenna(2+delta) ) then
        IF_swap(1) = antenna(2+delta)            ! original
        IF_swap(2) = current_p%antenna(2+delta)  ! new
      end if
    case (2)
      if ( antenna(1) > antenna(2+delta) ) then
        !> swap the two incoming hard radiators
        k = current_p%antenna(1)
        current_p%antenna(1) = current_p%antenna(2+delta)
        current_p%antenna(2+delta) = k
      end if
      !> sort the unresolved particles
      do i=2,1+delta
        !> find the biggest entry j in [i, i+1, ...]
        j = i
        do k=i+1,1+delta
          if ( current_p%antenna(k) > current_p%antenna(j) ) j = k
        end do
        if ( j /= i ) then
          !> swap
          k = current_p%antenna(i)
          current_p%antenna(i) = current_p%antenna(j)
          current_p%antenna(j) = k
        end if
      end do
    case default
      print*, "set_map: invalid # of intial states", ninitial
      stop
    end select
    !print*, "          antenna", antenna
    !print*, "current_p%antenna", current_p%antenna

    !> the reduced set is shifted by the unresolved partons
    do j=1,npar_to
      ishift = 0
      i = pmap(j)
      if ( i == IF_swap(1) ) i = IF_swap(2)
      if ( (delta>=1) .and. (i>current_p%antenna(2)) ) ishift = ishift+1
      if ( (delta>=2) .and. (i>current_p%antenna(3)) ) ishift = ishift+1
      red_set(j) = i - ishift
    end do

#ifndef NDEBUG
    !> red_set is a permutation of {1,2,...,npar_to}
    do j=1,npar_to
      if ( .not.any(red_set == j) ) then
        print*, "set_map: red_set not a permutation:", red_set
        stop
      end if
    end do
#endif

    ! print*, "# red_set", red_set
    ! !> check DIS: next-to-last entry == FS lepton
    ! if ( red_set(npar_to-1) /= 3 ) then
    !   print*, "set_map: DIS not set up properly!"
    !   stop
    ! end if
    !> @todo more checks?!

    current_p%reduced(:) = -1
    do j=1,npar_to
      i = pmap(j)
      if ( i == IF_swap(1) ) i = IF_swap(2)
      current_p%mapping( red_set(j) ) = i
      if ( i == current_p%antenna(1) )       current_p%reduced(1) = red_set(j)
      if ( i == current_p%antenna(2+delta) ) current_p%reduced(2) = red_set(j)
    end do
#ifndef NDEBUG
    !> sanity check
    if ( any(current_p%reduced < 0) ) then
      print*, "set_map: reduced not set properly:", current_p%reduced
      stop
    end if
#endif
    ! print*, "current_p%mapping", current_p%mapping
    ! print*, "-----------"

    !> apply the mapping & check if it passes cuts
    if ( qcache_map .and.  &
       & ( present(frag) .or. present(hfrag) ) ) then
      !> didn't think carefully about cases with fragmentation yet...
      stop "set_map: momentum caching not supported with fragmentation"
    end if
    if ( readCache_map(ipass) ) return
    call apply_map(frag, hfrag)
    call sijcuts(npar_to, ipass)
    call writeCache_map(ipass)

  contains

    subroutine set_j_map(npar)
      integer, intent(in) :: npar
      if (.not.allocated(j_map)) then
        allocate(j_map(npar))
        if (npar >= 1) j1 => j_map(1)
        if (npar >= 2) j2 => j_map(2)
        if (npar >= 3) j3 => j_map(3)
        if (npar >= 4) j4 => j_map(4)
        if (npar >= 5) j5 => j_map(5)
        if (npar >= 6) j6 => j_map(6)
        if (npar >= 7) j7 => j_map(7)
        if (npar >= 8) j8 => j_map(8)
        if (npar >= 9) j9 => j_map(9)
      end if
      red_set => j_map(:)
    end subroutine set_j_map

    subroutine set_k_map(npar)
      integer, intent(in) :: npar
      if (.not.allocated(k_map)) then
        allocate(k_map(npar))
        if (npar >= 1) k1 => k_map(1)
        if (npar >= 2) k2 => k_map(2)
        if (npar >= 3) k3 => k_map(3)
        if (npar >= 4) k4 => k_map(4)
        if (npar >= 5) k5 => k_map(5)
        if (npar >= 6) k6 => k_map(6)
        if (npar >= 7) k7 => k_map(7)
        if (npar >= 8) k8 => k_map(8)
      end if
      red_set => k_map(:)
    end subroutine set_k_map

    ! function fixed_pmap(j)
    !   integer :: fixed_pmap
    !   integer, intent(in) :: j
    !   fixed_pmap = pmap(j)
    !   if ( (antenna(1)/=current_p%antenna(1)) .and.  &
    !      & (fixed_pmap==antenna(1)) ) then
    !     fixed_pmap = current_p%antenna(1)
    !   else if ( (antenna(2+delta)/=current_p%antenna(2+delta)) .and.  &
    !      & (fixed_pmap==antenna(2+delta)) ) then
    !     fixed_pmap = current_p%antenna(2+delta)
    !   end if
    ! end function fixed_pmap

  end subroutine set_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply the mapping of `current_p`
  !> dispatch to the 6 generic mapping routines
  !
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)  
  !-----------------------------------------------------------------------------
  subroutine apply_map(frag, hfrag)
    use KinData_mod
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    integer :: count_in, delta
    integer :: a,u1,u2,b  ! antenna: parent momentum set
    integer :: ap,bp      ! hard momenta in the reduced momentum set
    integer :: i

    if (.not.associated(current_p)) return

    delta = current_p%npar_from - current_p%npar_to
    select case (delta)
    case (1)
      a  = current_p%antenna(1)
      u1 = current_p%antenna(2)
      b  = current_p%antenna(3)
      u2 = -1
    case (2)
      a  = current_p%antenna(1)
      u1 = current_p%antenna(2)
      u2 = current_p%antenna(3)
      b  = current_p%antenna(4)
    case default
      print*, "apply_map: invalid antenna, delta = ", delta
      stop
    end select

    !> find the indices of the hard radiators in the reduced momentum set
    ap = -1
    bp = -1
    do i = 1,current_p%npar_to
      if ( current_p%mapping(i) == a ) then
        ap = i
        cycle
      end if
      if ( current_p%mapping(i) == b ) then
        bp = i
        cycle
      end if
      if ( (ap > 0) .and. (bp > 0) ) exit
    end do
#ifndef NDEBUG
    if ( (ap < 0) .or. (bp < 0) ) then
      print*, "apply_map: couldn't find mapped hard radiators:", ap,bp
      stop
    end if
#endif

    call resetFrag_kin(current_p%npar_to)

    !> dispatch to the different mapping routines
    count_in = count(current_p%antenna <= 2)
    select case (count_in)
    case (0)
      !> FF
      if (delta==1) then
        if ( present(hfrag) ) then
          call apply_single_FH_map( current_p%npar_from,current_p%npar_to,  &
                                        & a,u1,b, ap,bp, current_p%mapping, hfrag)
        else
          call apply_single_FF_map( current_p%npar_from,current_p%npar_to,  &
                                  & a,u1,b, ap,bp, current_p%mapping, frag)
        end if
      else
        if ( present(hfrag) ) then
          call apply_double_FH_map( current_p%npar_from,current_p%npar_to,  &
                                        & a,u1,u2,b, ap,bp, current_p%mapping, hfrag)
        else
          call apply_double_FF_map( current_p%npar_from,current_p%npar_to,  &
                                  & a,u1,u2,b, ap,bp, current_p%mapping, frag)
        end if
      end if
    case (1)
      !> IF
      if (delta==1) then
        call apply_single_IF_map( current_p%npar_from,current_p%npar_to,  &
                                & a,u1,b, ap,bp, current_p%mapping, frag, hfrag)
      else
        call apply_double_IF_map( current_p%npar_from,current_p%npar_to,  &
                                & a,u1,u2,b, ap,bp, current_p%mapping, frag, hfrag)
      end if
    case (2)
      !> II
      if (delta==1) then
        call apply_single_II_map( current_p%npar_from,current_p%npar_to,  &
                                & a,u1,b, ap,bp, current_p%mapping, frag, hfrag)
      else
        call apply_double_II_map( current_p%npar_from,current_p%npar_to,  &
                                & a,u1,u2,b, ap,bp, current_p%mapping, frag, hfrag)
      end if
    case default
      print*, "apply_map: invalid mapping, count_in = ", count_in
      stop
    end select

    call fillS_kin(current_p%npar_to)
    call clearCacheLocal_kin(current_p%npar_to)
    !> store mapping hierarchy
    kin(current_p%npar_to)%npar_from = current_p%npar_from

  end subroutine apply_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> single-unresolved final-final mapping
  !> (hfrag has a different mapping => separate routine)
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !-----------------------------------------------------------------------------
  subroutine apply_single_FF_map(npar_from,npar_to, a,u1,b, ap,bp, map, frag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: x,y,z,omx,omy,omz
    double precision :: zfrag
    integer :: i
    integer :: ifrag

#ifndef NDEBUG
    if ( (a<=2) .or. (b<=2) .or. (u1<=2) .or. (ap<=2) .or. (bp<=2) ) then
      print*, "apply_single_FF_map: invalid input:", a,u1,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_single_FF_map: inconsistent mapping:", a,u1,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)
        
    call DAK_single( kd%s(a,u1), kd%s(u1,b), kd%s(a,b),  &
                   & x,y,z,omx,omy,omz )

    kd_red%p(:,ap) =   x*kd%p(:,a) +   y*kd%p(:,u1) +   z*kd%p(:,b)
    kd_red%p(:,bp) = omx*kd%p(:,a) + omy*kd%p(:,u1) + omz*kd%p(:,b)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do

    if ( isActive_frag() ) then
      if ( present(frag) ) then
        !> frag(1) is the parton index in npar_from
        !> frag(2) is the photon index in npar_from
        !> determine the spectator: either a or b
        !> Calculate zfrag and set ifrag accordingly
        if(frag(1)==a .or. frag(2)==a) then
          zfrag = kd%s(frag(2),b) / ( kd%s(frag(2),b) + kd%s(frag(1),b) )
          ifrag = ap
        else if(frag(1)==b .or. frag(2)==b) then
          zfrag = kd%s(frag(2),a) / ( kd%s(frag(2),a) + kd%s(frag(1),a) )
          ifrag = bp
        else
          print*, "apply_single_FF_map: Could not determine spectator index"
          stop
        end if
        if ( (kd%ifrag>0) .and. (kd%ifrag==a .or. kd%ifrag==b .or. kd%ifrag==u1))then
          !> secondary mapping or convolution
          !> use kd%zfrag to calculate momentum fraction
          zfrag = kd%zfrag * zfrag
        end if
        if ( (kd%ifrag<0) .and. nphotons_brems<0) then
          !> delta piece of convolution
          ifrag=-1
        end if
        !> store in KinData
        call setFrag_kin(npar_to, ifrag, z=zfrag)
      else if ( (.not. present(frag)) .and. (kd%ifrag > 0)) then
        !> primary antenna is photonic
        !> secondary antenna purly QCD
        !> pass fragmentation information to the k momentum set
#ifndef NDEBUG
        if( (kd%ifrag==a) .or. (kd%ifrag==b) .or. (kd%ifrag==u1)) then
          print*, "apply_single_FF_map: photonic cluster in sec. antenna, but no frag set given!"
          stop
        end if
#endif
        zfrag = kd%zfrag
        !> find the partent index in the reduced momentum set
        ifrag = -1
        do i = 1,npar_to
          if ( map(i) == kd%ifrag  ) then
            ifrag = i
            exit
          end if
        end do
#ifndef NDEBUG
        if ( ifrag < 0 ) then
          print*, "apply_single_FF_map: couldn't find fragmented parent:", frag
          stop
        end if
#endif
        !> store in KinData
        call setFrag_kin(npar_to, ifrag, z=zfrag)
      end if
    end if

  contains

    subroutine DAK_single(yau,yub,yab, x,y,z,omx,omy,omz)
      double precision, intent(in) :: yau,yub,yab
      double precision, intent(out) :: x,y,z,omx,omy,omz
      double precision :: yaub, rho

      yaub = yau+yub+yab

      y   = yub/(yau+yub)
      omy = yau/(yau+yub)

      rho = yaub*yab + 4d0*y*omy*yau*yub
      rho = sqrt(rho/yaub/yab)

      x   = (1d0+rho+yub*(1d0+rho-2d0*y)/(yau+yab))/2d0
      omx = 1d0-x

      !z = (1d0-rho+yau*(1d0-rho-2d0*y)/(yub+yab))/2d0
      omz = (1d0+rho+yau*(1d0+rho-2d0*omy)/(yub+yab))/2d0
      z   = 1d0-omz

    end subroutine DAK_single

  end subroutine apply_single_FF_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> single-unresolved final-final mapping with hadron fragmentation
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] hfrag i3 to indicate fragmentation of i3(npar_from)
  !-----------------------------------------------------------------------------
  subroutine apply_single_FH_map(npar_from,npar_to, a,u1,b, ap,bp, map, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in) :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: zfrag
    integer :: i
    integer :: ifrag, iother
    integer :: pf,jf,kf
    double precision :: numz

#ifndef NDEBUG
    if ( (a<=2) .or. (b<=2) .or. (u1<=2) .or. (ap<=2) .or. (bp<=2) ) then
      print*, "apply_single_FH_map: invalid input:", a,u1,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_single_FH_map: inconsistent mapping:", a,u1,b, ">", ap,bp, "|", map
      stop
    end if
    if ( .not. any([a,u1,b] == hfrag) ) then
      print*, "apply_single_FH_map: hfrag not in antenna:", a,u1,b, ">", hfrag
      stop
    end if
    if ( .not. isActive_hfrag() ) then
      stop "apply_single_FH_map: hfrag not active?!"
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    if (hfrag==a) then
      pf = a
      jf = b
      kf = u1
    else if (hfrag==b) then
      pf = b
      jf = a
      kf = u1
    else if (hfrag==u1) then
      pf = u1
      jf = a
      kf = b
    end if

    !> mapping based on apply_single_IF_map without fragmentation
    if ( map(ap) == hfrag ) then
      ifrag  = ap
      iother = bp
    else if ( map(bp) == hfrag ) then
      ifrag  = bp
      iother = ap
    else
      stop "apply_single_FH_map: hfrag not in hard mapped index?!"
    end if
    numz = kd%s(pf,jf)+kd%s(pf,kf)
    zfrag = numz/(numz+kd%s(jf,kf))
    kd_red%p(:,ifrag)  = kd%p(:,pf)/zfrag
    kd_red%p(:,iother) = kd%p(:,jf) + kd%p(:,kf) - (1d0-zfrag)*kd_red%p(:,ifrag)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do

    if (kd%ifrag > 0) then
#ifndef NDEBUG
      if (kd%ifrag /= hfrag) then
        print*, "apply_single_FH_map: ifrag vs. hfrag mismatch:", kd%ifrag, hfrag
        stop
      end if
#endif
      zfrag = zfrag * kd%zfrag
    end if

    !> store in KinData
    call setFrag_kin(npar_to, ifrag, z=zfrag)

  end subroutine apply_single_FH_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> double-unresolved final-final mapping
  !> (hfrag has a different mapping => separate routine)
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,u2,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !-----------------------------------------------------------------------------
  subroutine apply_double_FF_map(npar_from,npar_to, a,u1,u2,b, ap,bp, map, frag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,u2,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: x,r1,r2,y,omx,omr1,omr2,omy
    integer :: i

#ifndef NDEBUG
    if ( (a<=2) .or. (b<=2) .or. (u1<=2) .or. (u2<=2) .or. (ap<=2) .or. (bp<=2) ) then
      print*, "apply_double_FF_map: invalid input:", a,u1,u2,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_double_FF_map: inconsistent mapping:", a,u1,u2,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    call DAK_double( kd%s(a,u1), kd%s(a,u2),  &
                   & kd%s(u1,b), kd%s(u2,b),  &
                   & kd%s(u1,u2), kd%s(a,b),  &
                   & x,r1,r2,y,omx,omr1,omr2,omy )

    kd_red%p(:,ap) =   x*kd%p(:,a) +   r1*kd%p(:,u1) +   r2*kd%p(:,u2) +   y*kd%p(:,b)
    kd_red%p(:,bp) = omx*kd%p(:,a) + omr1*kd%p(:,u1) + omr2*kd%p(:,u2) + omy*kd%p(:,b)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do

#ifndef NDEBUG
    if ( present(frag) .and. isActive_frag() ) then
      stop "apply_double_FF_map: no fragmentation!"
    end if
#endif

  contains

    subroutine DAK_double( ya1,ya2,y1b,y2b,y12,yab,  &
                         &   x,  r1,  r2,  y,  &
                         & omx,omr1,omr2,omy )
      double precision, intent(in) :: ya1,ya2,y1b,y2b,y12,yab
      double precision, intent(out) :: x,r1,r2,y,omx,omr1,omr2,omy
      double precision :: ya12b, rho

      ya12b = ya1+ya2+y1b+y2b+y12+yab

      r1   = (y1b+y12)/(ya1+y1b+y12)
      omr1 =       ya1/(ya1+y1b+y12)
      r2   =       y2b/(ya2+y2b+y12)
      omr2 = (ya2+y12)/(ya2+y2b+y12)

      rho = 1d0  &
        &  +(r1-r2)**2/yab**2/ya12b**2*  &
        &  (yab**2*y12**2+ya1**2*y2b**2+ya2**2*y1b**2  &
        & -2d0*(yab*ya1*y2b*y12+yab*ya2*y1b*y12+ya1*ya2*y1b*y2b))  &
        &  +((r1*omr2+r2*omr1)  &
        & *2d0*(yab*ya1*y2b+yab*ya2*y1b-yab**2*y12)  &
        &  +4d0*r1*omr1*yab*ya1*y1b  &
        &  +4d0*r2*omr2*yab*ya2*y2b)/yab**2/ya12b
      rho = sqrt(rho)

      x = 1d0/2d0/(yab+ya1+ya2)*(  &
        &   (1d0+rho)*ya12b  &
        &     -(2d0*y1b+y12)*r1  &
        &     -(2d0*y2b+y12)*r2  &
        &   +(ya1*y2b-ya2*y1b)*(r1-r2)/yab)
      omx = 1d0-x

      omy = 1d0/2d0/(yab+y1b+y2b)*(  &
        &    (1d0+rho)*ya12b  &
        &      -(2d0*ya1+y12)*omr1  &
        &      -(2d0*ya2+y12)*omr2  &
        &    +(y1b*ya2-y2b*ya1)*(r2-r1)/yab)
      y = 1d0-omy

    end subroutine DAK_double

  end subroutine apply_double_FF_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> double-unresolved final-final mapping with hadron fragmentation
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,u2,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] hfrag i3 to indicate fragmentation of i3(npar_from)
  !-----------------------------------------------------------------------------
  subroutine apply_double_FH_map(npar_from,npar_to, a,u1,u2,b, ap,bp, map, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,u2,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in) :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: zfrag
    integer :: i
    integer :: ifrag, iother
    integer :: pf,jf,kf,lf
    double precision :: numz

#ifndef NDEBUG
    if ( (a<=2) .or. (b<=2) .or. (u1<=2) .or. (u2<=2) .or. (ap<=2) .or. (bp<=2) ) then
      print*, "apply_double_FH_map: invalid input:", a,u1,u2,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_double_FH_map: inconsistent mapping:", a,u1,u2,b, ">", ap,bp, "|", map
      stop
    end if
    if ( .not. any([a,u1,u2,b] == hfrag) ) then
      print*, "apply_double_FH_map: hfrag not in antenna:", a,u1,u2,b, ">", hfrag
      stop
    end if
    if ( .not. isActive_hfrag() ) then
      stop "apply_double_FH_map: hfrag not active?!"
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    if (hfrag==a) then
      pf = a
      jf = b
      kf = u1
      lf = u2
    else if (hfrag==b) then
      pf = b
      jf = a
      kf = u1
      lf = u2
    else if (hfrag==u1) then
      pf = u1
      jf = a
      kf = b
      lf = u2
    else if (hfrag==u2) then
      pf = u2
      jf = a
      kf = b
      lf = u1
    end if

    !> mapping based on apply_double_IF_map without fragmentation
    if ( map(ap) == hfrag ) then
      ifrag  = ap
      iother = bp
    else if ( map(bp) == hfrag ) then
      ifrag  = bp
      iother = ap
    else
      stop "apply_double_FH_map: hfrag not in hard mapped index?!"
    end if
    numz = kd%s(pf,jf)+kd%s(pf,kf)+kd%s(pf,lf)
    zfrag = numz / (numz+kd%s(jf,kf)+kd%s(jf,lf)+kd%s(kf,lf))
    kd_red%p(:,ifrag)  = kd%p(:,pf)/zfrag
    kd_red%p(:,iother) = kd%p(:,jf) + kd%p(:,kf) + kd%p(:,lf) - (1d0-zfrag)*kd_red%p(:,ifrag)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do

    if (kd%ifrag > 0) then
#ifndef NDEBUG
      if (kd%ifrag /= hfrag) then
        print*, "apply_double_FH_map: ifrag vs. hfrag mismatch:", kd%ifrag, hfrag
        stop
      end if
#endif
      zfrag = zfrag * kd%zfrag
    end if

    !> store in KinData
    call setFrag_kin(npar_to, ifrag, z=zfrag)

  end subroutine apply_double_FH_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> single-unresolved initial-final mapping
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)    
  !-----------------------------------------------------------------------------
  subroutine apply_single_IF_map(npar_from,npar_to, a,u1,b, ap,bp, map, frag, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: x,omx
    integer :: i
    integer :: ifrag
    double precision :: zfrag

#ifndef NDEBUG
    if ( (a>2) .or. (b<=2) .or. (u1<=2) .or. (ap>2) .or. (bp<=2) ) then
      print*, "apply_single_IF_map: invalid input:", a,u1,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_single_IF_map: inconsistent mapping:", a,u1,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    omx = -kd%s(u1,b)/(kd%s(a,u1)+kd%s(a,b))
    x   = 1d0-omx

    kd_red%p(:,ap) = x*kd%p(:,a)
    kd_red%p(:,bp) = kd%p(:,b) + kd%p(:,u1) - omx*kd%p(:,a)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do


    if ( isActive_frag() .or. isActive_hfrag() ) then
      if ( present(frag) .or. present(hfrag) ) then
        if ( present(frag) ) then
          !> initial-state `a`: reference momentum
          !> frag(1) is the parton index in npar_from
          !> frag(2) is the photon index in npar_from
          !> in secondary antenna functions
          !> frag(2) can be either b or u1
          zfrag = kd%s(frag(2),a) / ( kd%s(frag(2),a) + kd%s(frag(1),a) )
        else if ( present(hfrag) ) then
          !> initial-state `a`: reference momentum
          !> hfrag is a single index, which can be b or u1
          zfrag = kd%s(hfrag,a) / ( kd%s(u1,a) + kd%s(b,a) )
        end if
        !///remove for convolution case: if (hasParent_map()) then
        !> we're in the second mapping of a X30 x X30 term
        !> let's see if the parent mapping already had a fragmentation
        !> and if the corresponding parton is involved in the current mapping
        if ( (kd%ifrag>0) .and. (kd%ifrag==u1 .or. kd%ifrag==b) ) then
          !> we can use the momentum fraction in the preceeding mapping:
          !> `kd%zfrag` to compute zfrag here:
          zfrag = kd%zfrag * zfrag
        end if
        !///end if
        if ( (kd%ifrag<0) .and. nphotons_brems<0) then
          !> delta piece of convolution
          ifrag = -1
        else
          ifrag = bp
        end if
        !> store in KinData
        call setFrag_kin(npar_to, ifrag, z=zfrag)
      else if ( (.not. present(frag)) .and. (.not. present(hfrag))  &
              &  .and. (kd%ifrag > 0)) then
        !> primary antenna is photonic
        !> secondary antenna purly QCD
        !> pass fragmentation information to the k momentum set
#ifndef NDEBUG
        if( (kd%ifrag==a) .or. (kd%ifrag==b) .or. (kd%ifrag==u1)) then
          print*, "apply_single_IF_map: photonic cluster in sec. antenna, but no frag set given!"
          stop
        end if
#endif
        zfrag = kd%zfrag
        !> find the parent index in the reduced momentum set
        ifrag = -1
        do i = 1,npar_to
          if ( map(i) == kd%ifrag  ) then
            ifrag = i
            exit
          end if
        end do
#ifndef NDEBUG
        if ( ifrag < 0 ) then
          print*, "apply_single_IF_map: couldn't find fragmented parent:", frag
          stop
        end if
#endif
        !> store in KinData
        call setFrag_kin(npar_to, ifrag, z=zfrag)
      end if
    end if

  end subroutine apply_single_IF_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> double-unresolved initial-final mapping
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,u2,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)    
  !-----------------------------------------------------------------------------
  subroutine apply_double_IF_map(npar_from,npar_to, a,u1,u2,b, ap,bp, map, frag, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,u2,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: x,omx
    integer :: i
    integer :: ifrag
    double precision :: zfrag

#ifndef NDEBUG
    if ( (a>2) .or. (b<=2) .or. (u1<=2) .or. (u2<=2) .or. (ap>2) .or. (bp<=2) ) then
      print*, "apply_double_IF_map: invalid input:", a,u1,u2,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_double_IF_map: inconsistent mapping:", a,u1,u2,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    omx = - (kd%s(u1,u2)+kd%s(u1,b)+kd%s(u2,b)) / (kd%s(a,u1)+kd%s(a,u2)+kd%s(a,b))
    x   = 1d0-omx

    kd_red%p(:,ap) = x*kd%p(:,a)
    kd_red%p(:,bp) = kd%p(:,b) + kd%p(:,u1) + kd%p(:,u2) - omx*kd%p(:,a)
    !> remaining momenta unchanged
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      kd_red%p(:,i) = kd%p(:, map(i) )
    end do

    if ( present(frag) .and. isActive_frag() ) then
      !> initial-state(a): reference momentum
      !> frag(1): parton index in frag
      !> frag(2): parton index in frag
      !> frag(3): photon index in frag
      zfrag = kd%s(frag(3),a) / ( kd%s(frag(1),a) + kd%s(frag(2),a) + kd%s(frag(3),a))
      !> find the parent index in the reduced momentum set
      !> for the time being...
      !> parent is always bp???
      ifrag = bp
#ifndef NDEBUG
      if ( ifrag < 0 ) then
        print*, "apply_double_IF_map: couldn't find fragmented parent:", frag
        stop
      end if
#endif
      !> store in KinData
      call setFrag_kin(npar_to, ifrag, z=zfrag)
    end if

    if ( present(hfrag) .and. isActive_hfrag() ) then
      !> initial-state(a): reference momentum
      zfrag = kd%s(hfrag,a) / ( kd%s(b,a) + kd%s(u1,a) + kd%s(u2,a))
      !> In kin%zfrag we want the info of zfrag_kin*zfrag (i.e. eta*z)
      !> Initially initialised as eta=p_hadron/p_parton
      zfrag = kd%zfrag * zfrag
      !> find the parent index in the reduced momentum set
      !> for the time being...
      !> parent is always bp???
      ifrag = bp
#ifndef NDEBUG
      if ( ifrag < 0 ) then
        print*, "apply_double_IF_map: couldn't find fragmented parent:", hfrag
        stop
      end if
#endif
      !> store in KinData
      call setFrag_kin(npar_to, ifrag, z=zfrag)
    end if

  end subroutine apply_double_IF_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> single-unresolved initial-initial mapping
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)
  !
  !> The boost implemented in this subroutine is designed for massless particles!
  !-----------------------------------------------------------------------------
  subroutine apply_single_II_map(npar_from,npar_to, a,u1,b, ap,bp, map, frag, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: Q(4), Qt(4), r(4)
    double precision :: Q2, saQ, sbQ, xaOxb, xaTxb, xa, xb
    double precision :: sjQ, sjQt, r2
    integer :: i,j
    integer :: ifrag
    double precision :: zfrag

#ifndef NDEBUG
    if ( (a>2) .or. (b>2) .or. (u1<=2) .or. (ap>2) .or. (bp>2) ) then
      print*, "apply_single_II_map: invalid input:", a,u1,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_single_II_map: inconsistent mapping:", a,u1,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    Q(:) = kd%p(:,a) + kd%p(:,b) - kd%p(:,u1)
    Q2 = kd%s(a,b) + kd%s(a,u1) + kd%s(b,u1)  ! note *minus* in: s(a,i): sign = -1!
    xaTxb = Q2 / kd%s(a,b)
    saQ = kd%s(a,b) + kd%s(a,u1)
    sbQ = kd%s(b,a) + kd%s(b,u1)
    xaOxb = sbQ / saQ

    xa = sqrt( xaTxb * xaOxb )
    xb = sqrt( xaTxb / xaOxb )

    kd_red%p(:,ap) = xa*kd%p(:,a)
    kd_red%p(:,bp) = xb*kd%p(:,b)

    Qt(:) = kd_red%p(:,ap) + kd_red%p(:,bp)
    r(:) = Q(:) + Qt(:)
    r2 = 2d0*Q2 + xa*saQ + xb*sbQ

    ! print*, "Q2", Q2, dot(Q,Q), Q2/dot(Q,Q)
    ! print*, "saQ", saQ, 2d0*dot(kd%p(:,a),Q), saQ/(2d0*dot(kd%p(:,a),Q))
    ! print*, "sbQ", sbQ, 2d0*dot(kd%p(:,b),Q), sbQ/(2d0*dot(kd%p(:,b),Q))
    ! print*, "xa", xa
    ! print*, "xb", xb
    ! print*, "Qt2", Q2, dot(Qt,Qt), Q2/dot(Qt,Qt)
    ! print*, "r2", r2, dot(r,r), r2/dot(r,r)

    !> remaining momenta are boosted 
    !> NOTE: The calculation routine for sij ( 'fillS_kin' in KinData.f90 ) is designed
    !> for a pair of massless particles. The massive final state particle will cause  
    !> the calculation of sjQ and sjQt to go wrong, and then the boost will be incorrect!
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      j = map(i)
      sjQ  = - ( kd%s(a,j) + kd%s(b,j) + kd%s(u1,j) )  ! positive
      sjQt = - ( xa*kd%s(a,j) + xb*kd%s(b,j) )  ! positive
      ! print*, ">", i, "<->", j
      ! print*, "> sjQ ", sjQ,  2d0*dot(kd%p(:,j),Q),   sjQ/(2d0*dot(kd%p(:,j),Q))
      ! print*, "> sjQt", sjQt, 2d0*dot(kd%p(:,j),Qt), sjQt/(2d0*dot(kd%p(:,j),Qt))
      kd_red%p(:,i) = kd%p(:,j)  &
                    & -(sjQ+sjQt)/r2 * r(:)  &
                    & +sjQ/Q2 * Qt(:)
    end do

    if ( isActive_frag() .or. isActive_hfrag() ) then
      if ( present(frag) ) then
        print*, "apply_single_II_map: no fragmentation set allowed!" , frag
        stop
      else if ( present(hfrag) ) then
        print*, "apply_single_II_map: no fragmentation set allowed!" , hfrag
        stop
      else if ( (.not. present(frag)) .and. (.not. present(hfrag))  &
              & .and. kd%ifrag > 0) then
        !> primary antenna is photonic
        !> secondary antenna purly QCD
        !> pass fragmentation information to the k momentum set
        !>if ( .not. hasParent_map() ) then
        !>  print*, "apply_single_II_map: ifrag>0 without parent mapping?"
        !>  stop
        !>end if
#ifndef NDEBUG
        if( (kd%ifrag==a) .or. (kd%ifrag==b) .or. (kd%ifrag==u1)) then
          print*, "apply_single_II_map: photonic cluster in sec. antenna, but no frag set given!"
          stop
        end if
#endif
        zfrag = kd%zfrag
        !> find the partent index in the reduced momentum set
        ifrag = -1
        do i = 1,npar_to
          if ( map(i) == kd%ifrag ) then
            ifrag = i
            exit
          end if
        end do
#ifndef NDEBUG
        if ( ifrag < 0 ) then
          print*, "apply_single_II_map: couldn't find fragmented parent:", frag
          stop
        end if
#endif
        !> store in KinData
        call setFrag_kin(npar_to, ifrag, z=zfrag)
      end if
    end if

  end subroutine apply_single_II_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> double-unresolved initial-initial mapping
  !
  !> @param[in] npar_from parent momentum set
  !> @param[in] npar_to reduced momentum set
  !> @param[in] a,u1,u2,b the antenna defined in npar_from
  !> @param[in] ap,bp the hard radiators in npar_to
  !> @param[in] map array to define the mapping (NOTE: different to pmap above!)
  !>            red[a] <-> parent[b]: map(red[a]) = b  (no parent[{unresolved}])
  !> @param[in] frag optional array (i1,i2) to indicate fragmentation of:
  !>            [i,j](npar_to) --> parton@i1(npar_from) + photon@i2(npar_from)
  !> @param[in] hfrag optional i3 to indicate fragmentation of i3(npar_from)    
  !
  !> The boost implemented in this subroutine is designed for massless particles!
  !-----------------------------------------------------------------------------
  subroutine apply_double_II_map(npar_from,npar_to, a,u1,u2,b, ap,bp, map, frag, hfrag)
    use KinData_mod
    integer, intent(in) :: npar_from,npar_to
    integer, intent(in) :: a,u1,u2,b  ! antenna: parent momentum set
    integer, intent(in) :: ap,bp   ! hard momenta in the reduced momentum set
    integer, intent(in), dimension(:) :: map
    integer, intent(in), dimension(:), optional :: frag
    integer, intent(in), optional :: hfrag
    type(Kin_t), pointer :: kd, kd_red  ! parent & reduced momentum set
    double precision :: Q(4), Qt(4), r(4)
    double precision :: Q2, saQ, sbQ, xaOxb, xaTxb, xa, xb
    double precision :: sjQ, sjQt, r2
    integer :: i,j
    integer :: ifrag
    double precision :: zfrag

#ifndef NDEBUG
    if ( (a>2) .or. (b>2) .or. (u1<=2) .or. (u2<=2) .or. (ap>2) .or. (bp>2) ) then
      print*, "apply_double_II_map: invalid input:", a,u1,u2,b, ">", ap,bp
      stop
    end if
    if ( (a/=map(ap)) .or. (b/=map(bp)) ) then
      print*, "apply_double_II_map: inconsistent mapping:", a,u1,u2,b, ">", ap,bp, "|", map
      stop
    end if
#endif

    !> prettify code
    kd     => kin(npar_from)
    kd_red => kin(npar_to)

    Q(:) = kd%p(:,a) + kd%p(:,b) - kd%p(:,u1) - kd%p(:,u2)
    !> note *minus* in: s(a,i): sign = -1!
    Q2 = kd%s(a,b) + kd%s(a,u1) + kd%s(a,u2) + kd%s(b,u1) + kd%s(b,u2) + kd%s(u1,u2)
    xaTxb = Q2 / kd%s(a,b)
    saQ = kd%s(a,b) + kd%s(a,u1) + kd%s(a,u2)
    sbQ = kd%s(b,a) + kd%s(b,u1) + kd%s(b,u2)
    xaOxb = sbQ / saQ

    xa = sqrt( xaTxb * xaOxb )
    xb = sqrt( xaTxb / xaOxb )

    kd_red%p(:,ap) = xa*kd%p(:,a)
    kd_red%p(:,bp) = xb*kd%p(:,b)

    Qt(:) = kd_red%p(:,ap) + kd_red%p(:,bp)
    r(:) = Q(:) + Qt(:)
    r2 = 2d0*Q2 + xa*saQ + xb*sbQ

    ! print*, "Q2", Q2, dot(Q,Q), Q2/dot(Q,Q)
    ! print*, "saQ", saQ, 2d0*dot(kd%p(:,a),Q), saQ/(2d0*dot(kd%p(:,a),Q))
    ! print*, "sbQ", sbQ, 2d0*dot(kd%p(:,b),Q), sbQ/(2d0*dot(kd%p(:,b),Q))
    ! print*, "xa", xa
    ! print*, "xb", xb
    ! print*, "Qt2", Q2, dot(Qt,Qt), Q2/dot(Qt,Qt)
    ! print*, "r2", r2, dot(r,r), r2/dot(r,r)

    !> remaining momenta are boosted
    !> NOTE: The calculation routine for sij ( 'fillS_kin' in KinData.f90 ) is designed
    !> for a pair of massless particles. The massive final state particle will cause  
    !> the calculation of sjQ and sjQt to go wrong, and then the boost will be incorrect!
    do i = 1, npar_to
      if ( (i==ap) .or. (i==bp) ) cycle
      j = map(i)
      sjQ  = - ( kd%s(a,j) + kd%s(b,j) + kd%s(u1,j) + kd%s(u2,j) )  ! positive
      sjQt = - ( xa*kd%s(a,j) + xb*kd%s(b,j) )  ! positive
      ! print*, ">", i, "<->", j
      ! print*, "> sjQ ", sjQ,  2d0*dot(kd%p(:,j),Q),   sjQ/(2d0*dot(kd%p(:,j),Q))
      ! print*, "> sjQt", sjQt, 2d0*dot(kd%p(:,j),Qt), sjQt/(2d0*dot(kd%p(:,j),Qt))
      kd_red%p(:,i) = kd%p(:,j)  &
                    & -(sjQ+sjQt)/r2 * r(:)  &
                    & +sjQ/Q2 * Qt(:)
    end do

#ifndef NDEBUG
    if ( present(frag) .and. isActive_frag() ) then
      stop "apply_double_II_map: no fragmentation!"
    end if
#endif
    !> we need to pass zfrag_kin (i.e. eta) info to kin(npar_to) set
    if ( isActive_hfrag() ) then
        zfrag=kd%zfrag
        ifrag = kd%ifrag
        call setFrag_kin(npar_to, ifrag, z=zfrag)
    end if

  end subroutine apply_double_II_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the state of the crurrent mapping: `current_p`
  !
  !> @param[out] a,u1,u2,b the antenna
  !> @param[out] npar_from parent momentum set
  !> @param[out] npar_to reduced momentum set
  !> @param[in] npar_test inqure if `npar_test` is identical with npar_to (debug)
  !> @return signal success (there was a mapping)
  !-----------------------------------------------------------------------------
  logical function getCurrent_map(a,u1,u2,b, npar_from, npar_to, npar_test)
    integer, intent(out) :: a,u1,u2,b
    integer, optional, intent(out) :: npar_from, npar_to
    integer, optional, intent(in) :: npar_test
    integer :: delta

    getCurrent_map = .false.
    a  = -1
    u1 = -1
    u2 = -1
    b  = -1
    if (.not.associated(current_p)) return

    if ( present(npar_test) ) then
      if ( npar_test /= current_p%npar_to ) then
        !> inconsistent momenutm set: rest mapping state and return
        print*, "getCurrent_map: inconsistent npar state?! resetting..."
        print*, " > npar_test", npar_test
        call unset_map()
        return
      end if
    end if

    getCurrent_map = .true.  ! have a valid mapping

    delta = current_p%npar_from - current_p%npar_to
    select case(delta)
    case(1)
      a  = current_p%antenna(1)
      u1 = current_p%antenna(2)
      b  = current_p%antenna(3)
      u2 = -1
    case(2)
      a  = current_p%antenna(1)
      u1 = current_p%antenna(2)
      u2 = current_p%antenna(3)
      b  = current_p%antenna(4)
    case default
      print*, "getCurrent_map: invalid state"
      stop
    end select

    if (present(npar_from)) npar_from = current_p%npar_from
    if (present(npar_to))   npar_to   = current_p%npar_to

  end function getCurrent_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> take index in reduced momentum set and obtain the associated
  !> index in the parent momentum set
  !
  !> @param[in] idx index in reduced momentum set
  !> @return index in parent momentum set
  !-----------------------------------------------------------------------------
  integer function idxUpCurrent_map(idx)
    integer, intent(in) :: idx
    idxUpCurrent_map = -1
    if ( .not.associated(current_p) ) return
    if ( (idx <= 0) .or. (idx > current_p%npar_to) ) return
    idxUpCurrent_map = current_p%mapping(idx)
  end function idxUpCurrent_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if the current mapping has another mapping preceding it <-> X30 x X30
  !> (a grand-partent)
  !
  !> @return bool signaling if there is a parent
  !-----------------------------------------------------------------------------
  logical function hasParent_map()
    hasParent_map = .false.
    if (.not.associated(current_p)) return
    hasParent_map = (current_p%npar_from == mapitoj%npar_to)
  end function hasParent_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the state of the parent mapping
  !
  !> @param[out] a,u1,b the antenna  (can only be a X30)
  !> @param[out] npar_from parent momentum set
  !> @param[out] npar_to reduced momentum set
  !> @return signal success (there was a mapping)
  !-----------------------------------------------------------------------------
  logical function getParent_map(a,u1,b, npar_from, npar_to)
    integer, intent(out) :: a,u1,b
    integer, optional, intent(inout) :: npar_from, npar_to
    integer :: delta

    getParent_map = .false.
    a  = -1
    u1 = -1
    b  = -1
    if (.not.hasParent_map()) return
    getParent_map = .true.  ! have a mapping

    delta = mapitoj%npar_from - mapitoj%npar_to
    select case(delta)
    case(1)
      a  = mapitoj%antenna(1)
      u1 = mapitoj%antenna(2)
      b  = mapitoj%antenna(3)
    case default
      print*, "getParent_map: invalid state"
      stop
    end select

    if (present(npar_from)) npar_from = mapitoj%npar_from
    if (present(npar_to))   npar_to   = mapitoj%npar_to

  end function getParent_map


  !-----------------------------------------------------------------------------
  !> @brief
  !> take index in parent momentum set and map one more layer up <-> X30 x X30
  !
  !> @param[in] idx index in parent momentum set
  !> @return index in grand-parent momentum set
  !-----------------------------------------------------------------------------
  integer function idxUpParent_map(idx)
    integer, intent(in) :: idx
    idxUpParent_map = -1
    if ( .not.hasParent_map() ) return
    if ( (idx <= 0) .or. (idx > mapitoj%npar_to) ) return
    idxUpParent_map = mapitoj%mapping(idx)
  end function idxUpParent_map


end module Mapping_mod




