!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Facilitate multiple evaluations [external members: `_mult`]
!
! DESCRIPTION:
!> Module implementing a framework to allow for multiple evaluations of the
!> cross section with different settings and relative weights
!> typical use-case is the direct computation of cross section differences
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module MultiRun_mod
!-------------------------------------------------------------------------------
  use Reweight_mod
  use Selectors_mod
  use PhotonIsolation_mod
  implicit none
  private

  public :: init_mult, destroy_mult
  public :: register_mult, addopt_mult
  public :: digest_mult
  public :: isActive_mult, hasGroup_mult
  public :: resetCurrentRun_mult, loadNextRun_mult, currentRun_mult
  public :: applySelectors_mult
  public :: ntot_mult
  !public :: load_mult
  public :: min_npar_p2b_mult

  type :: Run_t
    !> unique id
    integer :: irun
    !> run group
    integer :: igroup = 0
    !> reweighting function
    type(ReweightFunction_t) :: rewgtfcn
    !> selectors
    ! selector options when an invalid observable is encountered
    !   0: do not bin the event to the histogram at all
    !   1: ignore the given selector and continue with the next & bin
    integer :: selOpt = 0
    type(Selector_t), pointer :: sel_p => null()
    type(Isol_t), pointer :: isol_p => null()
    type(RunOpt_t), pointer :: opt_p => null()
    type(Run_t), pointer :: next => null()
  end type Run_t

  !> available options
  integer, parameter :: optid_rcone = 1
  integer, parameter :: optid_mur   = 2
  integer, parameter :: optid_muf   = 3
  integer, parameter :: optid_p2b   = 4

  type :: RunOpt_t
    integer :: optid = 0
    integer :: valueINT = 0
    double precision :: valueDBLE = 0d0
    logical :: valueBOOL = .false.
    type(RunOpt_t), pointer :: next => null()
  end type RunOpt_t

  !> keep track of the head of the run stack and set counter
  type(Run_t), pointer :: head => null()
  integer :: nreg = 0
  integer :: current_irun = 0
  type(Run_t), pointer :: currentRun_p => null()
!$OMP THREADPRIVATE(currentRun_p, current_irun)

  !> for the Projection-To-Born method, we want to keep track of
  !> the smallest npar kinematics that has been registered
  integer :: min_npar_p2b_mult = -1


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the module settings
  !-----------------------------------------------------------------------------
  subroutine init_mult()
    if ( associated(head) ) then
      print*, "init_mult: already initialised, call destroy_mult first"
      stop
    end if
    nreg = 0
    current_irun = 0
  end subroutine init_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new run
  !
  !> @param[in] rewgtfcn reweighting function
  !> @param[in] run_group optional integer id for the run group
  !> @param[in] selectorsOption optional flag to define selectors treatment
  !> @param[in] selectors_p optional pointer to the head of selectors
  !>            (linked list)
  !> @param[in] isolation_p optional pointer to the head of isolation conditions
  !>            (linked list)
  !> @return unique id of the newly registered run
  !-----------------------------------------------------------------------------
  integer function register_mult( rewgtfcn, run_group,  &
                                & selectorsOption, selectors_p,  &
                                & isolation_p )
    type(ReweightFunction_t), intent(in) :: rewgtfcn
    integer, optional, intent(in) :: run_group
    integer, optional, intent(in) :: selectorsOption
    type(Selector_t), optional, pointer, intent(in) :: selectors_p
    type(Isol_t), optional, pointer, intent(in) :: isolation_p
    type(Run_t), pointer :: newRun, it

    register_mult = -1

    !> find the tail; we want the ordering:
    !>   head => "1" => ... => "nreg" => null()
    it => head
    if ( associated(it) ) then
      do while ( associated(it%next) )
        it => it%next
      end do
    end if

    !> consistency check
    if ( associated(it) ) then
      if ( nreg /= it%irun ) stop "register_mult: inconsistent state"
    else
      if ( nreg /= 0 ) stop "register_mult: nreg uninitialised?!"
    end if

    !> create the new run entry
    allocate(newRun)
    newRun%irun     =  nreg + 1
    newRun%rewgtfcn =  rewgtfcn
    newRun%opt_p    => null()
    newRun%next     => null()

    if (present(run_group)) then
      ! if (run_group==0) then
      !   print*, "register_mult: run_group = 0 (default) cannot be set manually"
      !   stop
      ! end if
      newRun%igroup = run_group
    else
      newRun%igroup = 0  ! default group id
    end if

    if (present(selectorsOption)) then
      newRun%selOpt = selectorsOption
    else
      newRun%selOpt = 0  ! do not bin event in case of `.not.valid`
    end if

    if (present(selectors_p)) then
      newRun%sel_p => selectors_p
    end if

    if (present(isolation_p)) then
      newRun%isol_p => isolation_p
    end if

    !> push it to the back
    if ( .not. associated(it) ) then
      head => newRun  ! new head
    else
      it%next => newRun  ! add to the tail
    end if
    nreg = nreg + 1

    register_mult = newRun%irun

  end function register_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new option to a run
  !
  !> @param[in] irun unique id for a registered run
  !> @param[in] optname name of the option to add
  !> @param[in] valueINT optional integer value of the option
  !> @param[in] valueDBLE optional double value of the option
  !> @param[in] valueBOOL optional bool value of the option
  !-----------------------------------------------------------------------------
  subroutine addopt_mult(irun, optname,  &
  &                      valueINT,  &
  &                      valueDBLE,  &
  &                      valueBOOL )
    use Process_mod, only : name_proc
    use StrHelper_mod, only : to_upper_str
    use Proj2Born_mod, only: init_p2b
    integer, intent(in) :: irun
    character(len=*), intent(in) :: optname
    integer, intent(in), optional :: valueINT
    double precision, intent(in), optional :: valueDBLE
    logical, intent(in), optional :: valueBOOL
    type(RunOpt_t), pointer :: newOpt, it_opt
    type(Run_t), pointer :: it

    if (irun < 1 .or. irun > nreg) then
      print*, "addopt_mult: invalid irun: ", irun, nreg
      stop
    end if

    !> find the matching run
    it => head
    do while ( associated(it) )
      if (it%irun == irun) exit
      it => it%next
    end do
    if (.not.associated(it)) then
      print*, "addopt_mult: could not find a match for irun: ", irun
      stop
    end if

    !> allocate the new option
    allocate(newOpt)
    newOpt%optid     = 0
    newOpt%valueINT  =  0
    newOpt%valueDBLE =  0d0
    newOpt%valueBOOL =  .false.
    newOpt%next      => null()

    !> set up the option
    select case (to_upper_str(optname))
    !----------
    case ("RCONE")
      if (.not.present(valueDBLE)) then
        print*, "addopt_mult: no valueDBLE for ", optname
        stop
      end if
      newOpt%optid     = optid_rcone
      newOpt%valueDBLE = valueDBLE
    !----------
    case ("MUR0")
      if (.not.present(valueINT)) then
        print*, "addopt_mult: no valueINT for ", optname
        stop
      end if
      newOpt%optid    = optid_mur
      newOpt%valueINT = valueINT
      if (present(valueDBLE)) then
        newOpt%valueDBLE = valueDBLE
      else
        newOpt%valueDBLE = 1d0
      end if
    !----------
    case ("MUF0")
      if (.not.present(valueINT)) then
        print*, "addopt_mult: no valueINT for ", optname
        stop
      end if
      newOpt%optid    = optid_muf
      newOpt%valueINT = valueINT
      if (present(valueDBLE)) then
        newOpt%valueDBLE = valueDBLE
      else
        newOpt%valueDBLE = 1d0
      end if
    !----------
    case ("P2B")
      ! if (.not.present(valueINT)) then
      !   print*, "addopt_mult: no valueINT for ", optname
      !   stop
      ! end if
      ! newOpt%valueINT = valueINT
      newOpt%optid = optid_p2b

      !> determine the born-level parton multiplicities
      !> @todo: use min_njets eventually to cope with this better
      select case (name_proc)
      case ( "EPEM" )
        newOpt%valueINT = 4
        call init_p2b("dec")
      case ( "DIS","DISWM","DISWP",  &
           & "DIS1J","DISWM1J","DISWP1J",  &
           & "SFDIS","SFDISWM","SFDISWP" )
        newOpt%valueINT = 4
        call init_p2b("dis")
      case ( "Z","ZJ","WM","WMJ","WP","WPJ","H","HJ","HM","HJM",  &
           & "HTO2PM","HTO2PJM","HTO2P","HTO2PJ",  &
           & "HTO2TAUM","HTO2TAUJM","HTO2TAU","HTO2TAUJ",  &
           & "GG","GG_FRAG","GGJ","GGJFC" )
        newOpt%valueINT = 4
        !> p p -> colour neutral has an optional `qt_cut` option
        if ( present(valueDBLE) ) then
          print*, "P2B: valueDBLE = ", valueDBLE
          newOpt%valueDBLE = valueDBLE
        end if
        call init_p2b("hdy")
      case ( "HTO2L1PM","HTO2L1PJM","HTO2L1P","HTO2L1PJ" )
        newOpt%valueINT = 5
      case ( "HTO4EM","HTO4EJM","HTO4E","HTO4EJ","HH","HHJ",  &
           & "HTO2E2MUM","HTO2E2MUJM","HTO2E2MU","HTO2E2MUJ",  &
           & "HTO2L2NM","HTO2L2NJM","HTO2L2N","HTO2L2NJ" )
        newOpt%valueINT = 6
      case ( "1JET","2JET","1JETFC","2JETFC","3JET" )
        newOpt%valueINT = 4
        call init_p2b("jet")
      case default
        print*, "addopt_mult: P2B not available for process: ", name_proc
        stop
      end select

      !> keep track of the smallest npar
      if ( ( min_npar_p2b_mult < 0 ) .or.  &
      &    ( min_npar_p2b_mult > 0 .and.   &
      &      newOpt%valueINT < min_npar_p2b_mult ) ) then
        min_npar_p2b_mult = newOpt%valueINT
      end if
    !----------
    case default
      print*, "addopt_mult: unknown option: ", optname
      stop
    end select

    !> check that the option has not been added yet
    it_opt => it%opt_p
    do while ( associated(it_opt) )
      if (it_opt%optid == newOpt%optid) then
        print*, "addopt_mult: already registered option ", optname
        stop
      end if
      it_opt => it_opt%next
    end do

    !> add the option to the option stack of the run
    newOpt%next => it%opt_p
    it%opt_p    => newOpt

  end subroutine addopt_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all registered runs
  !-----------------------------------------------------------------------------
  subroutine destroy_mult()
    type(Run_t), pointer :: it, next
    type(RunOpt_t), pointer :: it_opt, next_opt
    it => head
    do while ( associated(it) )
      !> save next before we destroy it
      next => it%next
      !> deallocate sub-entries and then the entry itself
      it_opt => it%opt_p
      do while ( associated(it_opt) )
        next_opt => it_opt%next  ! save next
        deallocate(it_opt)
        it_opt => next_opt  ! loop
      end do
      it%opt_p => null()
      deallocate(it)
      it => next  ! loop
    end do
    head => null()
    nreg = 0
    current_irun = 0
  end subroutine destroy_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the registered runs
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_mult(unit)
    use Observables_mod, only : getNameFromId_obs
    integer, optional, intent(in) :: unit
    integer :: iout
    type(Run_t), pointer :: it
    type(RunOpt_t), pointer :: it_opt
    character(len=128) :: obsname
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",A,T103,"*")'
    character(len=*), parameter :: fmt_run =  &
      & '(1x,"*",1x,"[",I3,"]",T15,"irun = ",I3,T40,A,T103,"*")'
    character(len=*), parameter :: fmt_opt_dble   =  &
      & '(1x,"*",5x,A," = ",1f6.1,T103,"*")'
    character(len=*), parameter :: fmt_opt_int   =  &
      & '(1x,"*",5x,A," = ",I6,T103,"*")'
    character(len=*), parameter :: fmt_opt_scl   =  &
      & '(1x,"*",5x,A," = ",1f6.1,A,T103,"*")'
    !> for selectors
    type(Selector_t), pointer :: it_sel
    !> for isolation
    type(Isol_t), pointer :: it_isol

    if ( .not.associated(head) ) return

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_str) ""
    write(iout,fmt_str) "> Multi runs <"
    write(iout,fmt_str) ""

    it => head
    do while ( associated(it) )
      write(iout,fmt_run) it%igroup, it%irun, toString_rewgt(it%rewgtfcn)

      !>--- digest options
      it_opt => it%opt_p
      do while ( associated(it_opt) )
        select case (it_opt%optid)
        case (optid_rcone)
          write(iout,fmt_opt_dble) "Rcone", it_opt%valueDBLE
        case (optid_mur)
          if (it_opt%valueINT==0) then
            obsname = ' GeV'
          else
            obsname = ' * ' // getNameFromId_obs(it_opt%valueINT)
          end if
          write(iout,fmt_opt_scl) "muR", it_opt%valueDBLE, obsname
        case (optid_muf)
          if (it_opt%valueINT==0) then
            obsname = ' GeV'
          else
            obsname = ' * ' // getNameFromId_obs(it_opt%valueINT)
          end if
          write(iout,fmt_opt_scl) "muF", it_opt%valueDBLE, obsname
        case (optid_p2b)
          write(iout,fmt_opt_int) "P2B", it_opt%valueINT
        case default
          print*, "digest_mult: unknown option: ", it_opt%optid
          stop
        end select
        it_opt => it_opt%next
      end do
      !>---

      !>--- digest Run selectors
      if ( associated(it%sel_p) ) then
        it_sel => it%sel_p
        do while ( associated(it_sel) )
          call digest_mult_print_selector(it_sel, it%selOpt)
          it_sel => it_sel%next
        end do
      end if
      !>---

      !>--- digest Run isolation
      if ( associated(it%isol_p) ) then
        it_isol => it%isol_p
        do while ( associated(it_isol) )
          call digest_mult_print_isolation(it_isol)
          it_isol => it_isol%next
        end do
      end if
      !>---

      write(iout, fmt_str) ""
      it => it%next
    end do

    contains

      subroutine digest_mult_print_selector(sel, opt)
        type(Selector_t) :: sel
        integer :: opt
        character(len=6) :: seltype
        character(len=*), parameter :: fmt_min_max = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &T30,1f12.3," <= ",T48,A15,T65," <= ",1f12.3,T103,"*")'
        character(len=*), parameter :: fmt_min     = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &T30,1f12.3," <= ",T48,A15,                  T103,"*")'
        character(len=*), parameter :: fmt_max     = '(1x,"*",1x,L1,10x,"> ",A6,"[",I1,"]:",&
        &                  T48,A15,T65," <= ",1f12.3,T103,"*")'

        seltype = ''
        if ( sel%qinvert ) then
          seltype = 'reject'
        else
          seltype = 'accept'
        endif
        if (sel%qlower .and. sel%qupper) then
          write(iout,fmt_min_max) sel%qvalid, seltype, opt, sel%lower, getNameFromId_obs(sel%obsId), sel%upper
        else if (sel%qlower .and. .not.sel%qupper) then
          write(iout,fmt_min) sel%qvalid, seltype, opt, sel%lower, getNameFromId_obs(sel%obsId)
        else if (.not.sel%qlower .and. sel%qupper) then
          write(iout,fmt_max) sel%qvalid, seltype, opt, getNameFromId_obs(sel%obsId), sel%upper
        else
          write(iout,*) "SOMETHING WENT WRONG"
        endif

      end subroutine digest_mult_print_selector

      subroutine digest_mult_print_isolation(isol)
        type(Isol_t) :: isol
        character(len=*), parameter :: fmt_isol =  &
          & '(1x,"*",3x,A8,":",1f12.2," <= ",A6," <= ",1f6.2,T103,"*")'
        character(len=*), parameter :: fmt_var =  &
          & '(1x,"*",16x,A8,"  = ",1f6.2  ,T103,"*")'

        write(iout,fmt_isol) getName_isol(isol), isol%R_min, "delta", isol%R
        select case(isol%ialgo)
        case(1)  ! fixed
          write(iout,fmt_var) "ETmax_const", isol%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", isol%ETmax_epsilon
          write(iout,fmt_var) "ET_threshold", isol%ET_threshold
        case(2)  ! smooth
          write(iout,fmt_var) "ETmax_const", isol%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", isol%ETmax_epsilon
          write(iout,fmt_var) "n", isol%n
        case(3)  ! hybrid
          write(iout,fmt_var) "ETmax_const", isol%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", isol%ETmax_epsilon
          write(iout,fmt_var) "n", isol%n
          write(iout,fmt_var) "R_inner", isol%R_inner
        case(4)  ! softdrop
          write(iout,fmt_var) "zcut", isol%zcut
          write(iout,fmt_var) "beta", isol%beta
          write(iout,fmt_var) "recl.alg", isol%ja2
        case default
          write(iout,*) "SOMETHING WENT WRONG"
        end select

      end subroutine digest_mult_print_isolation

  end subroutine digest_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter function for the number of cross section evaluations
  !
  !> @return total number of cross section evasluations
  !-----------------------------------------------------------------------------
  logical function isActive_mult()
    isActive_mult = associated(head)
  end function isActive_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a group id exists (used in histogram init)
  !
  !> @param[in] run_group integer id for the run group
  !> @return bool flag .true. if group id was registered
  !-----------------------------------------------------------------------------
  logical function hasGroup_mult(run_group)
    integer, intent(in) :: run_group
    type(Run_t), pointer :: newRun, it

    !> group id 0 is special: always exists
    if (run_group==0) then
      hasGroup_mult = .true.
      return
    end if

    hasGroup_mult = .false.

    if (.not.isActive_mult()) return

    it => head
    do while ( associated(it) )
      if (it%igroup == run_group) then
        hasGroup_mult = .true.
        return
      end if
      it => it%next
    end do

  end function hasGroup_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the pointer to the currently active run
  !-----------------------------------------------------------------------------
  subroutine resetCurrentRun_mult()
    nullify(currentRun_p)
    current_irun = 0
  end subroutine resetCurrentRun_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> load the next run
  !
  !> @return boolean flag indicating success (.true. = loaded next run)
  !-----------------------------------------------------------------------------
  logical function loadNextRun_mult()
    use Jets_mod, only : setRcone_jet
    use Proj2Born_mod, only : setActive_p2b
    use Scales_mod, only : setMu_scl, setMuR_scl, setMuF_scl, digest_scl
    type(RunOpt_t), pointer :: it_opt

    !> in case MultiRun is not active, we need to
    !> "loop" over one run with no additional options
    if (.not.isActive_mult()) then
      if (current_irun==0) then
        loadNextRun_mult = .true.
        current_irun = current_irun + 1
      else
        loadNextRun_mult = .false.
      end if
      return
    end if

    !> get the "next" element
    if (.not.associated(currentRun_p)) then
      currentRun_p => head
    else
      currentRun_p => currentRun_p%next
    end if

    !> success?
    if (.not.associated(currentRun_p)) then
      loadNextRun_mult = .false.  ! reached end
      return
    else
      loadNextRun_mult = .true.  ! success
    end if

    !>----- reset settings to default values
    !> Rcone = inputRcone
    call setRcone_jet()
    !> switch off P2B
    call setActive_p2b(.false.)
    !> reset scales to the input scales
    call setMu_scl()

    !> set isolation
    if ( associated(currentRun_p%isol_p) ) then
      call setPhotonIsol_isol(currentRun_p%isol_p)
    else
      call setPhotonIsol_isol()  ! reset to default
    end if

    !> set the local reweighting function
    call setLocal_rewgt(currentRun_p%rewgtfcn)

    !> apply all options
    it_opt => currentRun_p%opt_p
    do while ( associated(it_opt) )
      select case(it_opt%optid)
      case (optid_rcone)
        call setRcone_jet(it_opt%valueDBLE)
      case (optid_mur)
        call setMuR_scl(fac=it_opt%valueDBLE, obsId=it_opt%valueINT)
      case (optid_muf)
        call setMuF_scl(fac=it_opt%valueDBLE, obsId=it_opt%valueINT)
      case (optid_p2b)
        call setActive_p2b(.true., npar_born=it_opt%valueINT, qtcut=it_opt%valueDBLE)
      case default
        print*, "load_mult: unknown option"
        stop
      end select
      it_opt => it_opt%next
    end do

    current_irun = currentRun_p%irun

    ! print*, "irun = ", irun
    ! call digest_scl(qlocal=.true.)
    ! read(*,*)
  end function loadNextRun_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply all registered selectors of the current run
  !
  !> @param[inout] ipass will be set to zero if selectors not passed
  !-----------------------------------------------------------------------------
  subroutine applySelectors_mult(ipass)
    use Observables_mod, only : getValue_obs
    integer, intent(inout), optional :: ipass
    type(Selector_t), pointer :: it_sel
    double precision :: valObs
    logical :: valid

    if (.not.isActive_mult()) return

    !> apply all registered selectors for the histogram
    it_sel => currentRun_p%sel_p
    do while ( associated(it_sel) )
      valObs = getValue_obs(it_sel%obsId, valid)
      !> treatment of invalid observables
      if (.not.valid) then
        select case (currentRun_p%selOpt)
        case (0)
          !> reject full event if selector required to be valid
          if ( it_sel%qvalid ) ipass = 0
        case (1)
          !> just skip (do nothing)
        case default
          print*, "applySelectors_mult: unknown selOpt:", currentRun_p%selOpt
          stop
        end select
        !> invalid: either return or goto next
        if (ipass==0) then
          return
        else
          it_sel => it_sel%next
          cycle
        end if
      end if
      !> check if selector passes
      if ( .not.qpass_sel(it_sel, valObs) ) then
        ipass = 0
        return
      end if
      it_sel => it_sel%next
    end do

  end subroutine applySelectors_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the currently active run
  !
  !> @param[out] irun optional the unique id of the currently active run
  !> @param[out] run_group optional the run_group the current run belongs to
  !-----------------------------------------------------------------------------
  subroutine currentRun_mult(irun, run_group)
    integer, intent(out), optional :: irun, run_group

    if (.not.isActive_mult()) then
      if (current_irun/=1) stop "currentRun_mult: invalid state"
      if (present(irun))       irun      = current_irun
      if (present(run_group))  run_group = 0
      return
    end if

    if (.not.associated(currentRun_p)) then
      print*, "currentRun_mult: currentRun_p not set!"
      stop
    end if

    if (present(irun))       irun      = currentRun_p%irun
    if (present(run_group))  run_group = currentRun_p%igroup

  end subroutine currentRun_mult


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter function for the number of cross section evaluations
  !
  !> @return total number of cross section evasluations
  !-----------------------------------------------------------------------------
  integer function ntot_mult()
    if (nreg == 0) then
      ntot_mult = 1
    else
      ntot_mult = nreg
    end if
  end function ntot_mult


!   !-----------------------------------------------------------------------------
!   !> @brief
!   !> load the setting for a run
!   !> ATTENTION: DEPRECATED, use `loadNextRun_mult` instead
!   !
!   !> @param[in] irun unique id for a registered run
!   !-----------------------------------------------------------------------------
!   subroutine load_mult(irun)
!     use Jets_mod, only : setRcone_jet
!     use Proj2Born_mod, only : setActive_p2b
!     use Scales_mod, only : setMu_scl, setMuR_scl, setMuF_scl, digest_scl
!     integer, intent(in) :: irun
!     type(Run_t), pointer, save :: it => null()
! !$OMP THREADPRIVATE(it)
!     type(RunOpt_t), pointer :: it_opt
!
!     if (nreg == 0) then
!       if (irun /= 1) then
!         print*, "load_mult: no multi-run active, irun: ", irun
!         stop
!       end if
!       return
!     end if
!
!     if (irun < 1 .or. irun > nreg) then
!       print*, "load_mult: invalid irun: ", irun, nreg
!       stop
!     end if
!
!     !>----- reset settings to default values
!     !> Rcone = inputRcone
!     call setRcone_jet()
!     !> switch off P2B
!     call setActive_p2b(.false.)
!     !> reset scales to the input scales
!     call setMu_scl()
!
!     !> find the matching run
!     if (.not.associated(it)) it => head
!     do while ( it%irun /= irun )
!       it => it%next
!       if (.not.associated(it)) it => head
!     end do
!
!     !> set the local reweighting function
!     call setLocal_rewgt(it%rewgtfcn)
!
!     !> apply all options
!     it_opt => it%opt_p
!     do while ( associated(it_opt) )
!       select case(it_opt%optid)
!       case (optid_rcone)
!         call setRcone_jet(it_opt%valueDBLE)
!       case (optid_mur)
!         call setMuR_scl(fac=it_opt%valueDBLE, obsId=it_opt%valueINT)
!       case (optid_muf)
!         call setMuF_scl(fac=it_opt%valueDBLE, obsId=it_opt%valueINT)
!       case (optid_p2b)
!         call setActive_p2b(.true., npar_born=it_opt%valueINT)
!       case default
!         print*, "load_mult: unknown option"
!         stop
!       end select
!       it_opt => it_opt%next
!     end do
!
!     ! print*, "irun = ", irun
!     ! call digest_scl(qlocal=.true.)
!     ! read(*,*)
!
!   end subroutine load_mult


end module MultiRun_mod
