!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Observables Helper [external members: `v<N>_<obs>`]
!
! DESCRIPTION:
!> Module collecting a few core routines for computing observables
!> and decorator functions that can be applyied to observables
!-------------------------------------------------------------------------------

!-------------------------------------------------------------------------------
module ObsHelper_mod
!-------------------------------------------------------------------------------
  use Kinds_mod, only: wp
  implicit none
  public

contains

  !----------------------!
  !  1-particle Heplers  !
  !----------------------!

  pure function v1_pt2(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = v(1)**2 + v(2)**2
  end function v1_pt2

  pure function v1_pt(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = sqrt(v1_pt2(v))
  end function v1_pt

  pure function v1_p2(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = v(1)**2 + v(2)**2 + v(3)**2
  end function v1_p2

  pure function v1_p(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = sqrt(v1_p2(v))
  end function v1_p

  pure function v1_minv2(v) result(res)
    use FPA_mod
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    real(kind=wp) :: err, acc_i, err_i
    integer :: i
    res = v(4)**2
    err = 0._wp
    do i = 1, 3
      call twoSum(res, -v(i)**2, acc_i, err_i)
      res = acc_i
      err = err + err_i
    end do
    res = res + err
    !res = v(4)**2 - v(1)**2 - v(2)**2 - v(3)**2
  end function v1_minv2

  pure function v1_minv(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = sqrt(v1_minv2(v))
  end function v1_minv

  pure function v1_et2(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    !> MCFM seems to define ET = E * sqrt(px2+py2) / sqrt(px2+py2+pz2)
    res = v1_pt2(v) + v1_minv2(v)
  end function v1_et2

  pure function v1_et(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = sqrt(v1_et2(v))
  end function v1_et

  pure function v1_y(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    if (v(4) > abs(v(3))) then
      res = log((v(4) + v(3))/(v(4) - v(3)))/2._wp
    else
      !> safety implementation assuming m=0 when E <= |pz| (from num. rounding)
      if (v(4) <= v(3)) then
        res = log((v(4) + v(3))**2/v1_pt2(v))/2._wp  ! (v(4)-v(3)) ~ 0
      else
        res = log(v1_pt2(v)/(v(4) - v(3))**2)/2._wp  ! (v(4)+v(3)) ~ 0
      end if
    end if
  end function v1_y

  pure function v1_eta(v) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    real(kind=wp) :: p
    p = v1_p(v)
    res = log((p + v(3))/(p - v(3)))/2._wp
  end function v1_eta

  pure function v1_phi(v) result(res) ! in [0,2pi]
    use Constants_mod, only: twopi
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v
    res = atan2(v(2), v(1))
    if (res < 0._wp) res = res + twopi
  end function v1_phi

  !----------------------!
  !  2-particle Heplers  !
  !----------------------!

  pure function v2_delta_y(v1, v2) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = v1_y(v1) - v1_y(v2)
  end function v2_delta_y

  pure function v2_delta_eta(v1, v2) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = v1_eta(v1) - v1_eta(v2)
  end function v2_delta_eta

  pure function v2_delta_phi(v1, v2) result(res) ! in [0,pi]
    use Constants_mod, only: pi, twopi
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = abs(v1_phi(v1) - v1_phi(v2))
    if (res > pi) res = twopi - res
  end function v2_delta_phi

  pure function v2_delta_R2(v1, v2) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = v2_delta_phi(v1, v2)**2 + v2_delta_y(v1, v2)**2
  end function v2_delta_R2

  pure function v2_delta_R(v1, v2) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = sqrt(v2_delta_R2(v1, v2))
  end function v2_delta_R

  pure function v2_delta_R_exp(v1, v2) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in), dimension(4) :: v1, v2
    res = sqrt(v2_delta_phi(v1, v2)**2 + v2_delta_eta(v1, v2)**2)
  end function v2_delta_R_exp

  !--------------!
  !  Decorators  !
  !--------------!

  pure function abs_deco(val) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in) :: val
    res = abs(val)
  end function abs_deco

  pure function sqrt_deco(val) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in) :: val
    res = sqrt(val)
  end function sqrt_deco

  pure function sqrt_abs_deco(val) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in) :: val
    res = sqrt(abs(val))
  end function sqrt_abs_deco

  pure function sqr_deco(val) result(res)
    real(kind=wp) :: res
    real(kind=wp), intent(in) :: val
    res = val**2
  end function sqr_deco

end module ObsHelper_mod
