!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: PDF Cache [external members: `_pdfc`]
!
! DESCRIPTION:
!> Module implementing a very simple caching of PDFs
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module PDFCache_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: setActive_pdfc, isActive_pdfc
  public :: init_pdfc, digest_pdfc
  public :: initCache_pdfc, destroyCache_pdfc
  public :: evolvePDF_pdfc

  logical :: qactive = .true.


  ! !> Information stored for each cache entry
  ! type :: CacheData_t
  !   double precision :: muf = 0d0
  !   double precision, dimension(-6:7) :: f = 0d0
  ! end type CacheData_t

  !> the PDF cache object for muf entries
  type :: PDFCache_t
    integer :: icfg = 1  ! = [1|2] for which PDF set to use
    integer :: size = 0  ! how many values are used
    double precision :: x  ! momentum fraction associated with the cache
    double precision, allocatable, dimension(:)   :: muf
    double precision, allocatable, dimension(:,:) :: f
  end type PDFCache_t

  type(PDFCache_t), target :: cache1, cache2, cache1x1, cache2x2
!$OMP THREADPRIVATE(cache1, cache2, cache1x1, cache2x2)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the cache
  !
  !> @param[inout] pdfc PDF cache object
  !-----------------------------------------------------------------------------
  subroutine init_cache(pdfc)
    type(PDFCache_t), intent(inout) :: pdfc

    if ( allocated(pdfc%muf) .or. allocated(pdfc%f) ) then
      print*, "init_cache: cache already allocated!"
      ! call destroy_cache(pdfc)
      ! return
      stop
    end if

    pdfc%x = -1d0  !  init to invalid value
    pdfc%size = 0
    allocate( pdfc%muf(0) )
    allocate( pdfc%f(-6:7,0) )

  end subroutine init_cache


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate dynamic storage
  !
  !> @param[inout] pdfc PDF cache object
  !-----------------------------------------------------------------------------
  subroutine destroy_cache(pdfc)
    type(PDFCache_t), intent(inout) :: pdfc

    ! print*, "destroy_cache: ", pdfc%size

    pdfc%x = -1d0
    pdfc%size = 0
    if ( allocated(pdfc%muf) ) deallocate( pdfc%muf )
    if ( allocated(pdfc%f) )   deallocate( pdfc%f )

  end subroutine destroy_cache


  !-----------------------------------------------------------------------------
  !> @brief
  !> clear the status of chache entries
  !
  !> @param[inout] pdfc PDF cache object
  !-----------------------------------------------------------------------------
  subroutine clear_cache(pdfc)
    type(PDFCache_t), intent(inout) :: pdfc
    integer :: i

    pdfc%size = 0
    !> not needed (save performance):
    ! pdfc%muf(:) = 0d0

  end subroutine clear_cache


  !-----------------------------------------------------------------------------
  !> @brief
  !> resize the cache
  !
  !> @param[inout] pdfc PDF interpolation object
  !> @param[in] size_new new size
  !-----------------------------------------------------------------------------
  subroutine reserve_cache(pdfc, size_new)
    type(PDFCache_t), intent(inout) :: pdfc
    integer, intent(in) :: size_new
    integer :: i, size_old
    double precision, allocatable, dimension(:) :: tmp_muf
    double precision, allocatable, dimension(:,:) :: tmp_f

    size_old = size(pdfc%muf)
    if ( size_old >= size_new ) return

    ! print*, "reserve_cache:", size_old , ">", size_new

    allocate( tmp_muf(pdfc%size) )
    do i = 1,pdfc%size
      tmp_muf(i) = pdfc%muf(i)
    end do
    deallocate( pdfc%muf )
    allocate( pdfc%muf(size_new) )
    do i = 1,pdfc%size
      pdfc%muf(i) = tmp_muf(i)
    end do
    deallocate(tmp_muf)

    allocate( tmp_f(-6:7,pdfc%size) )
    do i = 1,pdfc%size
      tmp_f(-6:7,i) = pdfc%f(-6:7,i)
    end do
    deallocate( pdfc%f )
    allocate( pdfc%f(-6:7,size_new) )
    do i = 1,pdfc%size
      pdfc%f(-6:7,i) = tmp_f(-6:7,i)
    end do
    deallocate(tmp_f)

  end subroutine reserve_cache


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF set for a given x & muf value
  !
  !> @param[inout] pdfc PDF interpolation object
  !> @param[in] x momentum fraction of the parton
  !> @param[in] muf factorisation scale
  !> @param[out] f PDF
  !-----------------------------------------------------------------------------
  subroutine read_cache(pdfc, x,muf,f)
    use LHAPDF_mod
    use FPA_mod, only : is_approx!, is_approx_zero
    type(PDFCache_t), intent(inout) :: pdfc
    double precision, intent(in) :: x
    double precision, intent(in) :: muf
    double precision, dimension(-6:7), intent(out) :: f
    double precision :: rel_diff
    integer :: idx, i

    ! print*, "read_cache", x, muf

    if ( x /= pdfc%x ) then
      ! print*, "new x: ", pdfc%x, x, " CLEAR CACHE"
      pdfc%x = x
      call clear_cache(pdfc)
    end if

    !> check to see if we have the value in the cache
    !idx = findloc( is_approx_zero(pdfc%muf(1:pdfc%size) - muf), value=.true., dim=1)
    idx = 0
    do i = 1,pdfc%size
      ! if ( is_approx(pdfc%muf(i),muf) ) then
      rel_diff = (pdfc%muf(i)-muf) / (pdfc%muf(i)+muf)
      if ( abs(rel_diff) < 1d-6 ) then
        idx = i
        exit
      end if
    end do
    ! print*, "cache idx = ", idx
    if ( idx > 0 ) then
      !> found a match
      f(-6:7) = pdfc%f(-6:7,idx)
      return
    else
      !> write new entry
      idx = pdfc%size + 1
      if (idx > size(pdfc%muf)) call reserve_cache(pdfc, int(2*idx))
      pdfc%size = idx
      pdfc%muf(idx) = muf
      !> f from LHAPDF
      if ( pdfc%icfg == 2 ) then
        call evolve_xfph2xQ_lhapdf(pdfc%x, muf, pdfc%f(-6:7,idx))
      else
        call evolve_xfphxQ_lhapdf(pdfc%x, muf, pdfc%f(-6:7,idx))
      end if
      f(-6:7) = pdfc%f(-6:7,idx)
    end if

  end subroutine read_cache


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate PDF interpolation
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_pdfc(active)
    logical, intent(in) :: active
    qactive = active
  end subroutine setActive_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of PDF interpolation
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_pdfc()
    isActive_pdfc = qactive
  end function isActive_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise global settings
  !-----------------------------------------------------------------------------
  subroutine init_pdfc()
    ! nothing to do?
  end subroutine init_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the PDF interpolation settings
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_pdfc(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_act =  &
      & '(1x,"*",1x,"active:",T103,"*")'
    character(len=*), parameter :: fmt_inact   =  &
      & '(1x,"*",5x,"inactive",T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    if ( isActive_pdfc() ) then
      write(iout,fmt_act)
    else
      write(iout,fmt_inact)
    endif

  end subroutine digest_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise all dynamic arrays
  !-----------------------------------------------------------------------------
  subroutine initCache_pdfc()

    call init_cache( cache1 )
    call init_cache( cache2 )
    call init_cache( cache1x1 )
    call init_cache( cache2x2 )

    !> configuration
    cache1%icfg   = 1
    cache2%icfg   = 2
    cache1x1%icfg = 1
    cache2x2%icfg = 2

  end subroutine initCache_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all dynamic arrays
  !-----------------------------------------------------------------------------
  subroutine destroyCache_pdfc()

    call destroy_cache( cache1 )
    call destroy_cache( cache2 )
    call destroy_cache( cache1x1 )
    call destroy_cache( cache2x2 )

  end subroutine destroyCache_pdfc


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF set for a given x & muf value
  !
  !> @param[in] x momentum fraction of the parton
  !> @param[in] muf factorisation scale
  !> @param[out] f PDF
  !> @param[in] id optional integer identifier for the PDF
  !-----------------------------------------------------------------------------
  subroutine evolvePDF_pdfc(x, muf, f, id)
    ! use LHAPDF_mod
    double precision, intent(in) :: x
    double precision, intent(in) :: muf
    double precision, dimension(-6:7), intent(out) :: f
    integer, optional, intent(in) :: id
    type(PDFCache_t), pointer :: cache_p
    ! !> for debugging
    ! double precision, dimension(-6:7) :: f_debug

    if (.not.qactive) then
      print*, "evolvePDF_pdfc: PDF interpolation is inactive!"
      stop
    end if

    ! print*, "evolvePDF_pdfc", x, muf, id

    !> select the grid
    cache_p => cache1  ! default: take the first entry
    if ( present(id) ) then
      select case (id)
      case (1)
        cache_p => cache1
      case (2)
        cache_p => cache2
      case (-1)
        cache_p => cache1x1
      case (-2)
        cache_p => cache2x2
      case default
        print*, "evolvePDF_pdfc: invalid id:", id
        stop
      end select
    end if

    call read_cache(cache_p, x, muf, f)

    ! !> check precision against direct LHAPDF call
    ! !> we write to STDERR
    ! !> `./NNLOJET -run XYZ.run 2> err.txt`
    ! !> will pipe the output to a file
    ! call evolve_xfphxQ_lhapdf(x, muf, f_debug(-6:7))
    ! ! print*, x, muf
    ! ! print*, f(-5:5)
    ! ! print*, f_debug(-5:5)
    ! print*, f(-5:5)/f_debug(-5:5)
    ! ! read(*,*)
  end subroutine evolvePDF_pdfc


end module PDFCache_mod
