!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: PDF Interpolator [external members: `_pdfi`]
!
! DESCRIPTION:
!> Module implementing a qubic spline interpoation for the muF dependence
!> of parton distributions functions (PDF's)
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module PDFInterp_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: setActive_pdfi, isActive_pdfi
  public :: init_pdfi, digest_pdfi
  public :: initGrids_pdfi, destroyGrids_pdfi
  public :: evolvePDF_pdfi

  logical :: qactive = .false.

  !> step size of the grid (log(muF/muF0))
  double precision :: step

  !> Information stored for each grid point
  type :: GridData_t
    !> set the status of the grid point:
    !>   0: not set
    !>   1: only y-value set
    !>   2: y & y'' values set
    integer :: status = 0
    double precision :: x = 0d0  ! x-value = log(muF/muF0)
    double precision, dimension(-6:7) :: y  = 0d0  ! y-value: y(x) = PDF
    double precision, dimension(-6:7) :: y2 = 0d0  ! 2nd derivative: y''(x)
  end type GridData_t

  !> the PDF interpolation object, storing all information needed for the interpolation
  type :: PDFInterp_t
    integer :: icfg = 1  ! = [1|2] for which PDF set to use
    double precision :: xpdf
    double precision :: muf0
    type(GridData_t), allocatable, dimension(:) :: grid
  end type PDFInterp_t

  type(PDFInterp_t), target :: interp1, interp2, interp1x1, interp2x2
!$OMP THREADPRIVATE(interp1, interp2, interp1x1, interp2x2)

  !> the threadprivate tmp array used in the computation of y2
  !> more efficient keeping one than re-allocating it each time.
  !> Absorbing it into GridData_t is also not ideal, since the
  !> interpolation doesn't need it. Also, it's sufficient to have
  !> one for each thread which is large enough
  double precision, allocatable, dimension(:,:) :: u_tmp
!$OMP THREADPRIVATE(u_tmp)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the grid
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine init_grid(pdfi)
    type(PDFInterp_t), intent(inout) :: pdfi

    if ( allocated(pdfi%grid) ) then
      print*, "init_grid: grid already allocated!"
      ! call clear_grid(pdfi)
      ! return
      stop
    end if

    pdfi%xpdf = -1d0  !  init to invalid value
    pdfi%muf0 = -1d0  !  init to invalid value

    allocate( pdfi%grid(0:0) )  ! start with one entry (=central scale)

    pdfi%grid(0)%status = 0    ! PDFs not copied in yet
    pdfi%grid(0)%x      = 0d0  ! = log(muF0/muF0)
    pdfi%grid(0)%y      = 0d0
    pdfi%grid(0)%y2     = 0d0

  end subroutine init_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate dynamic storage
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine destroy_grid(pdfi)
    type(PDFInterp_t), intent(inout) :: pdfi

    pdfi%xpdf = -1d0
    pdfi%muf0 = -1d0

    if ( allocated(pdfi%grid) ) deallocate( pdfi%grid )

  end subroutine destroy_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> clear the status of grid entries
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine clear_grid(pdfi)
    type(PDFInterp_t), intent(inout) :: pdfi
    integer :: i

    do i = lbound(pdfi%grid,1), ubound(pdfi%grid,1)
      pdfi%grid(i)%status = 0
    end do

  end subroutine clear_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the y2-values of the grid
  !> we use the boundary condition for a "natural" spline: y2(lb)=y2(ub)=0
  !
  !> @param[inout] pdfi PDF interpolation object
  !> @param[in] il new lower bound
  !> @param[in] iu new upper bound
  !-----------------------------------------------------------------------------
  subroutine resize_grid(pdfi, il,iu)
    type(PDFInterp_t), intent(inout) :: pdfi
    integer, intent(in) :: il, iu
    integer :: i, lb, ub, lb_new, ub_new
    type(GridData_t), allocatable, dimension(:) :: tmp

    lb = lbound(pdfi%grid,1)
    ub = ubound(pdfi%grid,1)

    if ( il<lb .or. iu>ub ) then

      lb_new = min(il,lb)
      ub_new = max(iu,ub)

      print*, "resize_grid:", pdfi%icfg, pdfi%xpdf
      print*, " > before:", lb, ub
      print*, " > after: ", lb_new, ub_new

      !> save old x & y values
      allocate( tmp(lb:ub) )
      do i = lb,ub
        tmp(i) = pdfi%grid(i)
      end do

      !> re-allocate grid
      deallocate( pdfi%grid )
      allocate( pdfi%grid(lb_new:ub_new) )

      !> lower
      do i = lb_new,lb
        pdfi%grid(i)%status = 0
        pdfi%grid(i)%x      = i*step
        pdfi%grid(i)%y      = 0d0
        pdfi%grid(i)%y2     = 0d0
      end do
      !> old
      do i = lb,ub
        pdfi%grid(i)%status = 1
        pdfi%grid(i)%x      = tmp(i)%x
        pdfi%grid(i)%y      = tmp(i)%y
        pdfi%grid(i)%y2     = 0d0  ! can recycle?
      end do
      !> upper
      do i = ub,ub_new
        pdfi%grid(i)%status = 0
        pdfi%grid(i)%x      = i*step
        pdfi%grid(i)%y      = 0d0
        pdfi%grid(i)%y2     = 0d0
      end do

      deallocate(tmp)


      !> now it's time to resize the u_tmp array (if necessary)
      !> !!! need to do this before update_grid !!!
      !> it's size should be large enough to accomodate any
      !> pdfi object that is ever encountered (for each thread)
      lb = lbound(u_tmp,2)
      ub = ubound(u_tmp,2)
      if ( il<lb .or. iu>ub ) then
        lb_new = min(il,lb)
        ub_new = max(iu,ub)
        if ( allocated(u_tmp) ) deallocate( u_tmp )
        allocate( u_tmp(-6:7, lb_new:ub_new) )
      end if

      call update_grid(pdfi)

    end if

  end subroutine resize_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the grid
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine update_grid(pdfi)
    type(PDFInterp_t), intent(inout) :: pdfi

    call update_y_grid(pdfi)
    call update_y2_grid(pdfi)

  end subroutine update_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the y-values of the grid
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine update_y_grid(pdfi)
    use LHAPDF_mod
    type(PDFInterp_t), intent(inout) :: pdfi
    double precision :: muf_i
    integer :: i

    do i = lbound(pdfi%grid,1), ubound(pdfi%grid,1)
      if ( pdfi%grid(i)%status >= 1 ) cycle  ! PDF already stored
      !> the i-th scale
      muf_i = pdfi%muf0 * exp( pdfi%grid(i)%x )  ! is exp too expensive?
      !> update the grid y-value using LHAPDF
      if ( pdfi%icfg == 2 ) then
        call evolve_xfph2xQ_lhapdf(pdfi%xpdf, muf_i, pdfi%grid(i)%y(-6:7))
      else
        call evolve_xfphxQ_lhapdf(pdfi%xpdf, muf_i, pdfi%grid(i)%y(-6:7))
      end if
      !> update the status
      pdfi%grid(i)%status = 1
    end do

  end subroutine update_y_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> update the y2-values of the grid
  !> we use the boundary condition for a "natural" spline: y2(lb)=y2(ub)=0
  !
  !> @param[inout] pdfi PDF interpolation object
  !-----------------------------------------------------------------------------
  subroutine update_y2_grid(pdfi)
    type(PDFInterp_t), intent(inout) :: pdfi
    integer :: i, lb,ub
    double precision :: p(-6:7), sig

    lb = lbound(pdfi%grid,1)
    ub = ubound(pdfi%grid,1)

    ! !> DEBUG: linear?
    ! do i = lb+1, ub-1
    !   pdfi%grid(i)%y2(-6:7) = 0d0
    ! end do
    ! return

    !> lower boundary condition: "natural" cubic spline
    pdfi%grid(lb)%y2(-6:7) = 0d0
    pdfi%grid(lb)%status = 2
    u_tmp(-6:7,lb) = 0d0
    !> "central values"
    do i = lb+1, ub-1
      sig = + ( pdfi%grid(i)%x   - pdfi%grid(i-1)%x )  &
      &     / ( pdfi%grid(i+1)%x - pdfi%grid(i-1)%x )  ! == 1/2
      p(-6:7) = sig * pdfi%grid(i-1)%y2(-6:7) + 2d0
      pdfi%grid(i)%y2(-6:7) = (sig-1d0) / p(-6:7)
      u_tmp(-6:7,i) = + ( pdfi%grid(i+1)%y(-6:7) - pdfi%grid(i)%y(-6:7) )  &
      &               / ( pdfi%grid(i+1)%x       - pdfi%grid(i)%x )  &
      &               - ( pdfi%grid(i)%y(-6:7) - pdfi%grid(i-1)%y(-6:7) )  &
      &               / ( pdfi%grid(i)%x       - pdfi%grid(i-1)%x )
      u_tmp(-6:7,i) = ( + 6d0 * u_tmp(-6:7,i) / ( pdfi%grid(i+1)%x - pdfi%grid(i-1)%x )  &
      &                 - sig * u_tmp(-6:7,i-1)  &
      &               ) / p(-6:7)
    end do
    !> upper boundary condition: "natural" cubic spline
    pdfi%grid(ub)%y2(-6:7) = 0d0
    pdfi%grid(ub)%status = 2
    !> backsubstitution
    do i = ub-1, lb, -1
      pdfi%grid(i)%y2(-6:7) = pdfi%grid(i)%y2(-6:7) * pdfi%grid(i+1)%y2(-6:7) + u_tmp(-6:7,i)
      pdfi%grid(i)%status = 2
    end do

  end subroutine update_y2_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF set for a given x & muf value
  !
  !> @param[inout] pdfi PDF interpolation object
  !> @param[in] x momentum fraction of the parton
  !> @param[in] muf factorisation scale
  !> @param[out] f PDF
  !-----------------------------------------------------------------------------
  subroutine interpolate_grid(pdfi, x,muf,f)
    type(PDFInterp_t), intent(inout) :: pdfi
    double precision, intent(in) :: x
    double precision, intent(in) :: muf
    double precision, dimension(-6:7), intent(out) :: f
    double precision :: x_val
    integer :: kl, ku
    double precision :: h, a, b

    if (x /= pdfi%xpdf) then
      pdfi%xpdf = x
      pdfi%muf0 = muf
      call clear_grid(pdfi)
      call update_grid(pdfi)
    end if

    x_val = log( muf / pdfi%muf0 )

    !> find the range in the grid where x_val lies
    kl = int( x_val / step )
    if (x_val < 0d0) kl = kl-1
    ku = kl +1

    !> need to extend the range?
    call resize_grid(pdfi, kl,ku)

    if ( pdfi%grid(kl)%status /= 2 .or.  &
    &    pdfi%grid(ku)%status /= 2 ) then
      print*, "interpolate_grid: status /= 2!"
      stop
    end if

    h = pdfi%grid(ku)%x - pdfi%grid(kl)%x  ! == step
    a = (pdfi%grid(ku)%x - x_val) / h
    b = (x_val - pdfi%grid(kl)%x) / h

    f(-6:7) = + a * pdfi%grid(kl)%y(-6:7)  &
    &         + b * pdfi%grid(ku)%y(-6:7)  &
    &         + ( + (a**3-a) * pdfi%grid(kl)%y2(-6:7)  &
    &             + (b**3-b) * pdfi%grid(ku)%y2(-6:7)  &
    &           ) * h**2 / 6d0

  end subroutine interpolate_grid


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate PDF interpolation
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_pdfi(active)
    logical, intent(in) :: active
    qactive = active
  end subroutine setActive_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of PDF interpolation
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_pdfi()
    isActive_pdfi = qactive
  end function isActive_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise global settings
  !
  !> @param[in] stepfac optional argument setting the step factor
  !-----------------------------------------------------------------------------
  subroutine init_pdfi(stepfac)
    double precision, optional, intent(in) :: stepfac

    if ( present(stepfac) ) then
      step = log(stepfac)
      if (step <= 0d0) then
        print*, "init_pdfi: invalid step size:", step
        stop
      endif
    else
      ! step = log(2d0) / 2d0  ! sqrt(2)-factor steps from central scale
      !> factor 2 ist already good!
      step = log(2d0)  ! factor-two steps from central scale
    end if

  end subroutine init_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the PDF interpolation settings
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_pdfi(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_act =  &
      & '(1x,"*",1x,"active:",T30,"stepfac = ",1f6.1,T103,"*")'

    if ( .not.isActive_pdfi() ) return

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_str) ""
    write(iout,fmt_str) "> PDF Interpolation <"
    write(iout,fmt_str) ""

    write(iout,fmt_act) exp(step)

  end subroutine digest_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise all dynamic arrays
  !-----------------------------------------------------------------------------
  subroutine initGrids_pdfi()

    call init_grid( interp1 )
    call init_grid( interp2 )
    call init_grid( interp1x1 )
    call init_grid( interp2x2 )

    !> configuration
    interp1%icfg   = 1
    interp2%icfg   = 2
    interp1x1%icfg = 1
    interp2x2%icfg = 2

    allocate( u_tmp(-6:7,0:0) )

  end subroutine initGrids_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all dynamic arrays
  !-----------------------------------------------------------------------------
  subroutine destroyGrids_pdfi()

    call destroy_grid( interp1 )
    call destroy_grid( interp2 )
    call destroy_grid( interp1x1 )
    call destroy_grid( interp2x2 )

    deallocate( u_tmp )

  end subroutine destroyGrids_pdfi


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF set for a given x & muf value
  !
  !> @param[in] x momentum fraction of the parton
  !> @param[in] muf factorisation scale
  !> @param[out] f PDF
  !> @param[in] id optional integer identifier for the PDF
  !-----------------------------------------------------------------------------
  subroutine evolvePDF_pdfi(x, muf, f, id)
    double precision, intent(in) :: x
    double precision, intent(in) :: muf
    double precision, dimension(-6:7), intent(out) :: f
    integer, optional, intent(in) :: id
    type(PDFInterp_t), pointer :: interp_p
    !> for debugging
    !double precision, dimension(-6:7) :: f_debug

    if (.not.qactive) then
      print*, "evolvePDF_pdfi: PDF interpolation is inactive!"
      stop
    end if

    ! print*, "evolvePDF_pdfi", x, muf, id

    !> select the grid
    interp_p => interp1  ! default: take the first entry
    if ( present(id) ) then
      select case (id)
      case (1)
        interp_p => interp1
      case (2)
        interp_p => interp2
      case (-1)
        interp_p => interp1x1
      case (-2)
        interp_p => interp2x2
      case default
        print*, "evolvePDF_pdfi: invalid id:", id
        stop
      end select
    end if

    call interpolate_grid(interp_p, x, muf, f)

    ! !> check precision against direct LHAPDF call
    ! !> we write to STDERR
    ! !> `./NNLOJET -run XYZ.run 2> err.txt`
    ! !> will pipe the output to a file
    ! call evolvePDF(x, muf, f_debug(-6:7))
    ! write(0,'("PDFcheck ", 24e35.25)') x, muf, f(-5:5), f_debug(-5:5)
    ! !read(*,*)
  end subroutine evolvePDF_pdfi


end module PDFInterp_mod
