!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: store & manage all parameters [external members: `_param`]
!
! DESCRIPTION:
!> Module providing the storage of all SM parameters and their settings
!> eventuall will have the `broadcast_param` routine to fill al common blocks
!>
!> default values taken from: "2018 Review of Particle Physics"
!>
!> @todo: as_pdf, vs. internal as routine
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Parameters_mod
!-------------------------------------------------------------------------------
  use Constants_mod
  implicit none
  private

  !> set parameters
  public :: set_mass_param, set_width_param, set_value_param
  public :: set_ckm_param
  !> set and get scheme settings
  public :: set_scheme_param
  public :: mass_V_scheme_param, width_V_scheme_param
  public :: mt_scheme_param, alpha_scheme_param

  !> subroutines to set up
  public :: init_param, broadcast_param
#ifdef USE_OL
  public :: initOL_param
#endif

  !> getters: masses and widths
  public :: MZ_param, MW_param, MH_param
  public :: GZ_param, GW_param, GH_param
  public :: mt_param, Gt_param
  public :: mproton_param

  !> getters: couplings
  public :: alpha_param, alpha0_param
  public :: sw_param, sw2_param
  public :: cw_param, cw2_param

  !> anomalous dimensions
  public :: beta_param, gamma_param

  !> id's for the different mass schemes
  integer, parameter, public :: id_mass_scheme_OS    = -1
  integer, parameter, public :: id_mass_scheme_pole  =  0
  integer, parameter, public :: id_mass_scheme_MSbar = +1
  !> id's for the different width schemes
  integer, parameter, public :: id_width_scheme_naive =  1
  integer, parameter, public :: id_width_scheme_nwa   =  2
  integer, parameter, public :: id_width_scheme_OS    =  3
  integer, parameter, public :: id_width_scheme_CMS   =  4
  integer, parameter, public :: id_width_scheme_run   = -4


  !/////////////////////////////////////////
  !// PRIVATE member variables ("m_...")  //
  !/////////////////////////////////////////

  !>----- BOSONS -----

  !> Higgs-boson:  on-shell mass & width
  double precision :: m_MH = 125.18d0
  double precision :: m_GH = 0.004029643852d0  ! not from PDG
  !> Z-boson:  on-shell mass & width (BW with mass-dep. width)
  double precision :: m_MZ = 91.1876d0
  double precision :: m_GZ = 2.4952d0
  !> W-boson:  on-shell mass & width (BW with mass-dep. width)
  double precision :: m_MW = 80.379d0
  double precision :: m_GW = 2.085d0
  !> W/Z-boson mass scheme
  !>  -1: on-shell [default]
  !>   0: pole
  integer :: m_Z_scheme      = id_mass_scheme_OS  ! scheme of input values
  integer :: m_W_scheme      = id_mass_scheme_OS  ! scheme of input values
  integer :: m_mass_V_scheme = id_mass_scheme_OS  ! scheme used in the calculation
  !> W/Z-boson width scheme ( <-> inwa)
  !>   1: naive (fixed) [default]
  !>   2: narrow-width-approx.
  !>   3: OS
  !>   4: complex-mass-scheme
  !>  -4: running width
  integer :: m_width_V_scheme = id_width_scheme_naive

  !>----- QUARKS -----

  !> light quarks: MSbar mass
  double precision :: m_mu = 2.2d-3
  double precision :: m_md = 4.7d-3
  double precision :: m_ms = 95d-3
  !> charm mass: MSbar mass
  double precision :: m_mc = 1.275d0
  !> bottom mass: MSbar mass
  double precision :: m_mb = 4.18d0
  !> top mass: pole mass
  double precision :: m_mt = 173.1d0
  double precision :: m_Gt = 1.41d0  ! on-shell(?)
  !> top-mass scheme:
  !>   0: pole [default]
  !>  +1: MSbar
  integer :: m_mt_scheme = 0

  !>----- LEPTONS -----

  double precision :: m_mel  = 0.5109989461d-3
  double precision :: m_mmu  = 105.6583745d-3
  double precision :: m_mtau = 1.77686d0

  !>----- HADRONS -----

  double precision :: m_mproton = 0.9382720813d0

  !>----- PARAMETERS -----

  !> EW coupling
  double precision :: m_alpha   = 0d0  ! allow a direct override
  double precision :: m_alpha0  = 1d0/137.035999139d0
  double precision :: m_alphaMZ = 1d0/128.93d0
  double precision :: m_GF      = 1.1663787d-5  ! Fermi constant
  !> alpha scheme
  !>  -1: alpha override by runcard
  !>   0: alpha(0)
  !>   1: alpha(MZ)
  !>   2: alpha(Gmu) [default]
  !>   3: alpha(muR) (EPRC routine)
  integer :: m_alpha_scheme = 2
  !> for the complex-mass-scheme: variants in taking alpha real:
  !>   1: sqrt(2) GF / pi * ABS[ xMW2 xsw2 ]     (OpenLoops default)
  !>   2: sqrt(2) GF / pi * Re[ xMW2 xsw2 ]
  !>   3: sqrt(2) GF / pi * rMW2 (1 - rMW2/rMZ2) (Recola default)
  integer :: m_alpha_GF_variant = 1
  !> hard/"resolved" photons should use alpha0
  logical, public :: qalpha0_hard_photons = .false.

  !> CKM matrix
  !> (see options in CKM.f90)
  integer :: m_ckm = 0

  !>----- DERIVED QUANTITIES -----

  !> squared boson masses
  double precision :: m_MH2, m_MZ2, m_MW2
  !> alphaX = gX^2 / (4*pi)
  double precision :: m_el
  ! double precision :: m_gs
  !> weak mixing angle (s=sin, c=cos, 2=^2)
  double precision :: m_sw, m_sw2, m_cw, m_cw2

  !> complex qunatities
  complex(8) :: x_cw2, x_sw2, x_MZ2, x_MW2


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the mass for a particle
  !
  !> @param[in] particle string identifying the particle
  !> @param[in] mass the value for the particle mass in [GeV]
  !> @param[in] scheme optional string setting identifying the scheme
  !-----------------------------------------------------------------------------
  subroutine set_mass_param(particle, mass, scheme)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: particle
    double precision, intent(in) :: mass
    character(len=*), intent(in), optional :: scheme

    print*, "set_mass_param: ", particle, mass

    select case(to_upper_str( trim(adjustl( particle )) ))
      case('H')
        m_MH = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: H"
      case('Z')
        m_MZ = mass
        if ( present(scheme) ) m_Z_scheme = parse_mass_scheme_param(scheme)
      case('W')
        m_MW = mass
        if ( present(scheme) ) m_W_scheme = parse_mass_scheme_param(scheme)
      case('U')
        m_mu = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: u"
      case('D')
        m_md = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: d"
      case('S')
        m_ms = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: s"
      case('C')
        m_mc = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: c"
      case('B')
        m_mb = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: b"
      case('T')
        m_mt = mass
        if ( present(scheme) ) m_mt_scheme = parse_mass_scheme_param(scheme)
      case('EL')
        m_mel = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: el"
      case('MU')
        m_mmu = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: mu"
      case('TAU')
        m_mtau = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: tau"
      case('PROTON')
        m_mproton = mass
        if ( present(scheme) ) stop "set_mass_param: no mass scheme for: proton"
      case default
        print*, "set_mass_param: cannot set the mass for particle: ", particle
        stop
    end select

  end subroutine set_mass_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the width for a particle
  !
  !> @param[in] particle string identifying the particle
  !> @param[in] width the value for the particle width in [GeV]
  !-----------------------------------------------------------------------------
  subroutine set_width_param(particle, width)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: particle
    double precision, intent(in) :: width

    print*, "set_width_param: ", particle, width

    select case(to_upper_str( trim(adjustl( particle )) ))
      case('H')
        m_GH = width
      case('Z')
        m_GZ = width
      case('W')
        m_GW = width
      case('T')
        m_Gt = width
      case default
        print*, "set_width_param: cannot set the width for particle: ", particle
        stop
    end select

  end subroutine set_width_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the value for a parameter
  !
  !> @param[in] param string identifying the parameter
  !> @param[in] value the value for the parameter
  !-----------------------------------------------------------------------------
  subroutine set_value_param(param, value)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: param
    double precision, intent(in) :: value

    ! print*, "set_value_param: ", param, value

    select case(to_upper_str( trim(adjustl( param )) ))
      case('ALPHA')
        m_alpha = value
      case('ALPHA0')
        m_alpha0 = value
      case('ALPHAMZ')
        m_alphaMZ = value
      case('GF')
        m_GF = value
      case default
        print*, "set_value_param: cannot set the value for parameter: ", param
        stop
    end select

  end subroutine set_value_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> parse the mass scheme
  !
  !> @param[in] scheme flag specifying the scheme
  !-----------------------------------------------------------------------------
  integer function parse_mass_scheme_param(scheme)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: scheme

    select case( to_upper_str(trim(scheme)) )
    case ("OS")
      parse_mass_scheme_param = id_mass_scheme_OS
    case ("POLE")
      parse_mass_scheme_param = id_mass_scheme_pole
    case ("MSBAR")
      parse_mass_scheme_param = id_mass_scheme_MSbar
    case default
      print*, "parse_mass_scheme_param: invalid scheme ", scheme
      stop
    end select

  end function parse_mass_scheme_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> set scheme options (wrapper routine)
  !
  !> @param[in] name string identifying which scheme setting we want to set
  !> @param[in] scheme string for the scheme
  !> @param[in] suboptions optional string for additional sub-settings
  !-----------------------------------------------------------------------------
  subroutine set_scheme_param(name, scheme, suboptions)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: name, scheme
    character(len=*), intent(in), optional :: suboptions

    print*, "set_scheme_param: ", name, ">", scheme
    if (present(suboptions)) print*, ">>", suboptions, "<<"

    select case(to_upper_str( trim(adjustl( name )) ))
      case('MV', 'MASS_V')
        m_mass_V_scheme = parse_mass_scheme_param(scheme)
      case('MT')
        m_mt_scheme = parse_mass_scheme_param(scheme)
      case('GV', 'WIDTH_V')
        call set_width_V_scheme_param(scheme)
      case('ALPHA')
        call set_alpha_scheme_param(scheme, suboptions)
      case default
        print*, "set_scheme_param: unknown scheme name: ", name
        stop
    end select

  contains

      !> for on/off flags
      ! select case( to_upper_str(trim(scheme)) )
      ! case ("0", ".FALSE.", "OFF")
      !   option = 0
      ! case ("1", ".TRUE.", "ON")
      !   option = 1

    subroutine set_width_V_scheme_param(scheme_)
      use StrHelper_mod, only : to_upper_str
      character(len=*), intent(in) :: scheme_

      select case( to_upper_str(trim(scheme_)) )
      case ("NWA", "NARROW_WIDTH", "NARROW_WIDTH_APPROXIMATION", "ZERO")
        m_width_V_scheme = id_width_scheme_nwa
      case ("CMS", "COMPLEX_MASS_SCHEME")
        m_width_V_scheme = id_width_scheme_CMS
        if (m_mass_V_scheme /= id_mass_scheme_pole) then
          print*, "set_width_V_scheme_param: CMS uses pole masses!"
          m_mass_V_scheme = id_mass_scheme_pole
        end if
      case ("NAIVE", "FIX", "FIXED")
        m_width_V_scheme = id_width_scheme_naive
      case ("RUN", "RUNNING")
        m_width_V_scheme = id_width_scheme_run
        if (m_mass_V_scheme /= id_mass_scheme_OS) then
          print*, "set_width_V_scheme_param: RUN uses OS masses!"
          m_mass_V_scheme = id_mass_scheme_OS
        end if
      case ("POLE")
        !> pole scheme for width means mass scheme!
        m_mass_V_scheme = id_mass_scheme_pole
      case default
        print*, "set_width_V_scheme_param: invalid scheme ", scheme_
        stop
      end select

    end subroutine set_width_V_scheme_param

    subroutine set_alpha_scheme_param(scheme_, variant)
      use StrHelper_mod, only : to_upper_str
      character(len=*), intent(in) :: scheme_
      character(len=*), intent(in), optional :: variant

      select case( to_upper_str(trim(scheme_)) )
      case ("FIX", "FIXED")
        m_alpha_scheme = -1
      case ("ALPHA0", "0", "ZERO")
        m_alpha_scheme = 0
      case ("ALPHAMZ", "ALPHAZ", "MZ")
        m_alpha_scheme = 1
      case ("ALPHAGF", "ALPHAGM", "GF", "GMU")
        m_alpha_scheme = 2
      case ("RUN", "MUR", "EPRC")
        m_alpha_scheme = 3
      case default
        print*, "set_alpha_scheme_param: invalid scheme ", scheme_
        stop
      end select

      if (present(variant)) then
        if (m_alpha_scheme .eq. 2) then
          select case( to_upper_str(trim(variant)) )
          case ("", "1", "ABS", "OL", "OPENLOOPS")
            m_alpha_GF_variant = 1
          case ("2")
            m_alpha_GF_variant = 2
          case ("3", "RECOLA")
            m_alpha_GF_variant = 3
          case default
            print*, "set_alpha_scheme_param: invalid GF variant ", variant
            stop
          end select
        end if
      end if

    end subroutine set_alpha_scheme_param

  end subroutine set_scheme_param

  subroutine set_ckm_param(option_string)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: option_string

    write(*,*) option_string

    select case( to_upper_str(trim(option_string)) )
    case ("IDENTITY","UNITY","NOCKM")
       m_ckm = 0
    case ("FULL")
       m_ckm = 1
    case ("CABIBBO")
       m_ckm = 2
    case ("ONLY_VCS")
       m_ckm = 3
    case ("ONLY_VCS_VCD")
       m_ckm = 4
    case ("ABM")
       m_ckm = 5
    case ("MCFM")
       m_ckm = 6
    case default
       print*, "set_ckm_param: invalid option ", option_string
       stop
    end select

  end subroutine set_ckm_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the MV scheme
  !-----------------------------------------------------------------------------
  integer function mass_V_scheme_param()
    mass_V_scheme_param = m_mass_V_scheme
  end function mass_V_scheme_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the setting for the complex-mass-scheme
  !-----------------------------------------------------------------------------
  integer function width_V_scheme_param()
    width_V_scheme_param = m_width_V_scheme
  end function width_V_scheme_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the top-mass scheme
  !-----------------------------------------------------------------------------
  integer function mt_scheme_param()
    mt_scheme_param = m_mt_scheme
  end function mt_scheme_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the alpha scheme
  !-----------------------------------------------------------------------------
  integer function alpha_scheme_param()
    alpha_scheme_param = m_alpha_scheme
  end function alpha_scheme_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> set up all schemes and derived variables
  !-----------------------------------------------------------------------------
  subroutine init_param()
    use DIS_mod, only : use_EPRC
    double precision :: vfac

    !> conversion between OS and pole masse for MZ & MW

    if ( m_mass_V_scheme /= m_Z_scheme ) then
      if ( m_mass_V_scheme==0 .and. m_Z_scheme==-1 ) then
        !> OS to pole
        vfac = sqrt( 1d0 + (m_GZ/m_MZ)**2 )
      else if ( m_mass_V_scheme==-1 .and. m_Z_scheme==0 ) then
        !> pole to OS
        vfac = 1d0 / sqrt( 1d0 + (m_GZ/m_MZ)**2 )
      else
        print*, "init_param: unknown Z-boson mass scheme input!"
        stop
      end if
      m_MZ = m_MZ / vfac
      m_GZ = m_GZ / vfac
      m_Z_scheme = m_mass_V_scheme
    end if

    if ( m_mass_V_scheme /= m_W_scheme ) then
      if ( m_mass_V_scheme==0 .and. m_W_scheme==-1 ) then
        !> OS to pole
        vfac = sqrt( 1d0 + (m_GW/m_MW)**2 )
      else if ( m_mass_V_scheme==-1 .and. m_W_scheme==0 ) then
        !> pole to OS
        vfac = 1d0 / sqrt( 1d0 + (m_GW/m_MW)**2 )
      else
        print*, "init_param: unknown W-boson mass scheme input!"
        stop
      end if
      m_MW = m_MW / vfac
      m_GW = m_GW / vfac
      m_W_scheme = m_mass_V_scheme
    end if

    !> squared boson masses
    m_MZ2 = m_MZ**2
    m_MW2 = m_MW**2
    m_MH2 = m_MH**2

    !> weak mixing angle
    m_cw = m_MW / m_MZ
    m_cw2 = m_cw**2
    m_sw2 = (1d0-m_cw)*(1d0+m_cw)
    m_sw = sqrt(m_sw2)

    !> CMS
    x_MZ2 = DCMPLX(m_MZ2, -m_GZ*m_MZ)
    x_MW2 = DCMPLX(m_MW2, -m_GW*m_MW)
    x_cw2 = x_MW2 / x_MZ2
    x_sw2 = 1d0 - x_cw2

    !> fine structure constant
    select case (m_alpha_scheme)
    case (-1)
      if (m_alpha <= 0d0) stop "init_param: forgot to set alpha in runcard?!"
    case (0)
      m_alpha = m_alpha0
    case (1)
      m_alpha = m_alphaMZ
    case (2)
      if ( width_V_scheme_param() .eq. id_width_scheme_CMS ) then
        select case(m_alpha_GF_variant)
        case(1)
          m_alpha = sqrt(2d0) * m_GF / pi * abs(x_MW2 * x_sw2)
        case(2)
          m_alpha = sqrt(2d0) * m_GF / pi * real(x_MW2 * x_sw2)
        case(3)
          m_alpha = sqrt(2d0) * m_GF / pi * m_MW2 * (1d0-m_MW2/m_MZ2)
        case default
          stop "init_param: invalid alpha GF variant..."
        end select
      else
        m_alpha = sqrt(2d0) * m_GF / pi * m_MW2 * m_sw2
      end if
    case (3)
      !> @ todo
      stop "init_param: alpha in MSbar not supported yet..."
      m_alpha = 0d0
    case default
      stop "init_param: unknown alpha scheme?!"
    end select

    !> overrides...
    !> use EPRC routines
    if (use_EPRC()) then
      print*, "Parameters: initialising EPRC..."
      call EPRC_INIT(.false.)
    end if

    m_el = sqrt(4d0*pi*m_alpha)

    !> done initialising parameters: broadcast to all common blocks
    call broadcast_param()

  end subroutine init_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> broadcast module values to all common blocks
  !-----------------------------------------------------------------------------
  subroutine broadcast_param()
    use DIS_mod, only : use_EPRC
    use CKM_mod, only : CKM, fillckm
    use Flavour_mod, only: isActive_flav
    use Process_mod, only: ilepton_proc

    double precision :: nTau(-5:5)
    double precision :: amz, aem
    integer :: i, j
    complex(8) :: zcw2, zsw2, zMZ2, zMW2

    double precision :: emz,ezwidth
    common /Zmass/ emz,ezwidth
    double precision :: emw,ewwidth
    common /Wmass/ emw,ewwidth
    double precision :: emh,ehwidth
    common /Hmass/ emh,ehwidth
    double precision :: emt,etwidth,emb,ebwidth,emc,ecwidth,emtau,etauwidth
    common /Qmass/ emt,etwidth,emb,ebwidth,emc,ecwidth,emtau,etauwidth

    double precision :: amzW,stw,ewfac
    common /eweakW/ amzW,stw,ewfac
!$omp threadprivate(/eweakW/)

    double precision :: amzZ
    complex(8) :: zewfac(4),zewnull(4)
    common /eweakZ/ amzZ,zewfac,zewnull
!$omp threadprivate(/eweakZ/)

    double precision :: ele,ere,sin2w,enQ(-5:5),enL(5),enR(5)
    common /LR/ ele,ere,sin2w,enQ,enL,enR

    complex(8) :: zele,zere,zsin2w,zenL(5),zenR(5)
    common /zLR/ zele,zere,zsin2w,zenL,zenR

    integer :: ckmmode
    common /CKMmode/ ckmmode

    !> fill in the mass common blocks first
    emh       = m_MH
    ehwidth   = m_GH
    emz       = m_MZ
    ezwidth   = m_GZ
    emw       = m_MW
    ewwidth   = m_GW
    emt       = m_mt  ! should call mt_param() everywhere -> signaling NaN to trace
    etwidth   = m_Gt
    emb       = m_mb
    ebwidth   = 0d0
    emc       = m_mc
    ecwidth   = 0d0
    emtau     = m_mtau
    etauwidth = 0d0

    !> initialise variables
    zewfac=0d0
    zewnull=0d0
    amzW=0d0
    stw=0d0
    ewfac=0d0

    ! !> hardcoding parameters temporarily for Z+bjet flavtag runs
    ! if (isActive_flav()) then
    !   print*, "NOTE: using hardcoded values for m_sw2 and m_alpha for flavour-tagging runs!"
    !   m_sw2 = 0.2292834081112d0
    !   m_sw  = sqrt(m_sw2)
    !   m_cw2 = 1d0 - m_sw2
    !   m_cw  = sqrt(m_cw2)
    !   m_alpha = 7.77906438417484d-3
    ! end if

    amz = m_alpha
    amzW = amz
    amzZ = amz ! fortran doesn't like two parameters being the same in different common blocks

    stw = m_sw2

    ckmmode = m_ckm
    call fillckm(m_ckm)

    !> ewcharges,iQ(0) is not a valid flavour
    enQ(-5) = +0.333333333333333d0
    enQ(-4) = -0.666666666666667d0
    enQ(-3) = +0.333333333333333d0
    enQ(-2) = -0.666666666666667d0
    enQ(-1) = +0.333333333333333d0
    enQ(0)  = +0d0
    enQ(+1) = -0.333333333333333d0
    enQ(+2) = +0.666666666666667d0
    enQ(+3) = -0.333333333333333d0
    enQ(+4) = +0.666666666666667d0
    enQ(+5) = -0.333333333333333d0

    !> A parameter used in the coupling
    data nTau/1d0,-1d0,1d0,-1d0,1d0,0d0,-1d0,1d0,-1d0,1d0,-1d0/
    sin2w = 2d0 * m_sw * m_cw
    do j=1,5
       enL(j) = (ntau(j)-2d0*enQ(j)*stw)/sin2w
       enR(j) = -2d0*enQ(j)*stw/sin2w
    enddo

    ele=(-1d0-2d0*(-1d0)*stw)/sin2w
    ere=(    -2d0*(-1d0)*stw)/sin2w
    if ( ilepton_proc == 0 ) then
      !> neutrino's
      ele=+1d0/sin2w
      ere=0d0
      print*, "NEUTRINO COUPLINGS", ele, ere
    end if

!////////// the complex-mass scheme
    if ( width_V_scheme_param() == id_width_scheme_CMS ) then
    print*, "broadcast_param: CMS"
    ! read(*,*)
    zMW2 = DCMPLX(emw**2, -ewwidth*emw)
    zMZ2 = DCMPLX(emz**2, -ezwidth*emz)
    zcw2 = zMW2 / zMZ2
    zsw2 = 1d0 - zcw2

    zsin2w = 2d0*sqrt(zsw2*zcw2)

    do j=1,5
       zenL(j) = (ntau(j)-2d0*enQ(j)*zsw2)/zsin2w
       zenR(j) = -2d0*enQ(j)*zsw2/zsin2w
    enddo

    zele=(-1d0-2d0*(-1d0)*zsw2)/zsin2w
    zere=(    -2d0*(-1d0)*zsw2)/zsin2w
    if ( ilepton_proc == 0 ) then
      !> neutrino's
      zele=+1d0/zsin2w
      zere=0d0
      print*, "NEUTRINO COUPLINGS", zele, zere
    end if

    !> for W-production
    stw = abs(zsw2)

    end if


  end subroutine broadcast_param


#ifdef USE_OL
  !-----------------------------------------------------------------------------
  !> @brief
  !> broadcast module values to OpenLoops library
  !-----------------------------------------------------------------------------
  subroutine initOL_param()
    use openloops
    use Run_mod, only : pdf_nf_run
    double precision :: vfac
    integer :: nf_pdf

    !> conversion between OS and pole masse for MZ & MW

    if ( m_mass_V_scheme /= m_W_scheme ) then
      if ( m_mass_V_scheme==0 .and. m_W_scheme==-1 ) then
        !> OS to pole
        vfac = sqrt( 1d0 + (m_GW/m_MW)**2 )
      else if ( m_mass_V_scheme==-1 .and. m_W_scheme==0 ) then
        !> pole to OS
        vfac = 1d0 / sqrt( 1d0 + (m_GW/m_MW)**2 )
      else
        print*, "init_param: unknown W-boson mass scheme input!"
        stop
      end if
      m_MW = m_MW / vfac
      m_GW = m_GW / vfac
      m_W_scheme = m_mass_V_scheme
    end if

    !> will be overwritten per PS point
    call set_parameter("mu",1d0)
    call set_parameter("alphas",1d0)

    !> nf used in LHAPDF
    call set_parameter("minnf_alphasrun", pdf_nf_run);

    !> set everything properly
    if (qalpha0_hard_photons) then
      call set_parameter("alpha_qed_0",m_alpha0)
    else
      call set_parameter("alpha_qed_0",m_alpha)
    end if
    call set_parameter("alpha_qed_mz",m_alphaMZ)
    call set_parameter("gmu",m_GF)

    !> in case we want to go away from the default alpha-setting
    ! call set_parameter("onshell_photon_lsz",1)

    if ( alpha_scheme_param() /= 2 ) then
      print*, ""
      print*, "*************** WARNING ***************"
      print*, "WARNING: initOL_param: I only accept Gmu scheme!"
      print*, "*************** WARNING ***************"
      print*, ""
    else
      call set_parameter("ew_scheme",1)
    end if

    if ( width_V_scheme_param() /= id_width_scheme_CMS ) then
      print*, ""
      print*, "*************** WARNING ***************"
      print*, "WARNING: initOL_param: I only accept CMS!"
      print*, "*************** WARNING ***************"
      print*, ""
    else
      call set_parameter("use_cms",1) ! complex mass scheme, default=1
    end if

    !> for debug of OL event generatd and their accuracy statistics
    !call set_parameter("stability_log",1)
    !call set_parameter("stability_kill2",10)

    !> only OS mass for Z-boson? -> nope
    vfac = 1d0
    ! if ( m_Z_scheme /= id_mass_scheme_OS ) then
    !   if ( m_Z_scheme == id_mass_scheme_pole ) then
    !     !> pole to OS
    !     vfac = 1d0 / sqrt( 1d0 + (m_GZ/m_MZ)**2 )
    !   else
    !     print*, "initOL_param: unknown Z-boson mass scheme input!"
    !     stop
    !   end if
    ! end if
    call set_parameter("mass(23)",m_MZ / vfac)
    call set_parameter("width(23)",m_GZ / vfac)

    !> only OS mass for W-boson? -> nope
    vfac = 1d0
    ! if ( m_W_scheme /= id_mass_scheme_OS ) then
    !   if ( m_W_scheme == id_mass_scheme_pole ) then
    !     !> pole to OS
    !     vfac = 1d0 / sqrt( 1d0 + (m_GW/m_MW)**2 )
    !   else
    !     print*, "initOL_param: unknown W-boson mass scheme input!"
    !     stop
    !   end if
    ! end if
    call set_parameter("mass(24)",m_MW / vfac)
    call set_parameter("width(24)",m_GW / vfac)

    call set_parameter("mass(25)",MH_param())
    call set_parameter("width(25)",GH_param())

    call set_parameter("mass(6)",mt_param())
    call set_parameter("width(6)",Gt_param())


    ! call set_parameter("n_quarks",5)

    call set_parameter("me_cache",0)
    call set_parameter("verbose",0)

    !> output all OL parameters (ONCE)
    call set_parameter("parameters_verbose",1)
    call parameters_flush()
    call set_parameter("parameters_verbose",0)

    !> suppress PS point warnings
    call set_parameter("psp_tolerance ",1d-5)

  end subroutine initOL_param
#endif




  !> whole lot of getters

  double precision function alpha_param()
    alpha_param = m_alpha
  end function alpha_param

  double precision function alpha0_param()
    alpha0_param = m_alpha0
  end function alpha0_param

  double precision function sw_param()
    sw_param = m_sw
  end function sw_param

  double precision function sw2_param()
    sw2_param = m_sw2
  end function sw2_param

  double precision function cw_param()
    cw_param = m_cw
  end function cw_param

  double precision function cw2_param()
    cw2_param = m_cw2
  end function cw2_param

  double precision function MZ_param()
    MZ_param = m_MZ
  end function MZ_param

  double precision function MW_param()
    MW_param = m_MW
  end function MW_param

  double precision function MH_param()
    MH_param = m_MH
  end function MH_param

  double precision function GZ_param()
    GZ_param = m_GZ
  end function GZ_param

  double precision function GW_param()
    GW_param = m_GW
  end function GW_param

  double precision function GH_param()
    GH_param = m_GH
  end function GH_param

  double precision function Gt_param()
    Gt_param = m_Gt
  end function Gt_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the top mass
  !
  !> @param[in] muR optional argument for the renormalisation scale
  !>            default: muR = mt in the MSbar scheme
  !-----------------------------------------------------------------------------
  double precision function mt_param(muR)
    double precision, intent(in), optional :: muR

    select case(m_mt_scheme)
    case (0)
      !> pole mass: just return
      mt_param = m_mt
      return
    case (1)
      !> MSbar mass: check if we have a scale, otherwise muR = mt
      if ( present(muR) ) then
        if ( m_mt_scheme /= id_mass_scheme_MSbar ) then
          print*, "mt_param: passed scale but not MSbar scheme?!"
          stop
        end if
        mt_param = run_mq_MSbar(m_mt, muR)
      else
        mt_param = m_mt
      end if
    case default
      print*, "mt_param: unsupported scheme id ", m_mt_scheme
      stop
    end select

  end function mt_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the quark mass in the MSbar scheme evolved to a ren. scale
  !
  !> @param[in] mQ the MSbar mass mQ(mQ) in [GeV]
  !> @param[in] muR the renormalisation scale in [GeV]
  !-----------------------------------------------------------------------------
  double precision function run_mq_MSbar(mQ, muR)
    use LHAPDF_mod, only : alphasQ_lhapdf
    double precision, intent(in) :: mQ, muR
    double precision :: asQ, asR

    asQ = alphasQ_lhapdf(mQ)
    asR = alphasQ_lhapdf(muR)

    run_mq_MSbar =                                                     &
    & +gamma_param(0)/beta_param(0) * log(asR/asQ)                     &
    & +(-beta_param(1)*gamma_param(0)                                  &
    &   +beta_param(0)*gamma_param(1)                                  &
    & ) / beta_param(0)**2 * (asR-asQ)/(2d0*pi)                        &
    & +(+beta_param(1)**2*gamma_param(0)                               &
    &   -beta_param(0)*beta_param(2)*gamma_param(0)                    &
    &   -beta_param(0)*beta_param(1)*gamma_param(1)                    &
    &   +beta_param(0)**2*gamma_param(2)                               &
    & ) / beta_param(0)**3                                             &
    &   * (asR**2-asQ**2) / (2d0*pi)**2 / 2d0                          &
    & +(-beta_param(1)**3*gamma_param(0)                               &
    &   +2d0*beta_param(0)*beta_param(1)*beta_param(2)*gamma_param(0)  &
    &   -beta_param(0)**2*beta_param(3)*gamma_param(0)                 &
    &   +beta_param(0)*beta_param(1)**2*gamma_param(1)                 &
    &   -beta_param(0)**2*beta_param(2)*gamma_param(1)                 &
    &   -beta_param(0)**2*beta_param(1)*gamma_param(2)                 &
    &   +beta_param(0)**3*gamma_param(3)                               &
    & ) / beta_param(0)**4                                             &
    &   * (asR**3-asQ**3) / (2d0*pi)**3 / 3d0

    run_mq_MSbar = mq * exp(run_mq_MSbar)

  end function run_mq_MSbar


  double precision function mproton_param()
    mproton_param = m_mproton
  end function mproton_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> The coefficients of the beta function: [hep-ph:9703284]
  !>   beta = - as * ( beta0 * (as/2pi)
  !>                 + beta1 * (as/2pi)^2
  !>                 + beta2 * (as/2pi)^3
  !>                 + ... )
  !
  !> @param[in] n specify the n-th order coefficient
  !> @param[in] nf optional parameter to override the number of fermions
  !-----------------------------------------------------------------------------
  double precision function beta_param(n, nf)
    integer, intent(in) :: n
    integer, intent(in), optional :: nf
    double precision :: nf_val

    !> number of flavours as double
    if ( present(nf) ) then
      nf_val = nf
    else
      nf_val = 5d0
    end if

    !> switch for the order
    select case(n)
      case(0)
        beta_param = ( 11d0             &
                   &   -2d0/3d0*nf_val  &
                   & ) / 2d0
      case(1)
        beta_param = ( 102d0             &
                   &   -38d0/3d0*nf_val  &
                   & ) / 4d0
      case(2)
        beta_param = ( 2857d0/2d0             &
                   &   -5033d0/18d0*nf_val    &
                   &   +325d0/54d0*nf_val**2  &
                   & ) / 8d0
      case(3)
        beta_param = ( (149753d0/6d0 +3564d0*zeta3)                   &
                   &   -(1078361d0/162d0 +6508d0/27d0*zeta3)*nf_val   &
                   &   +(50065d0/162d0 +6472d0/81d0*zeta3)*nf_val**2  &
                   &   +1093d0/729d0*nf_val**3                        &
                   & ) / 16d0
      case default
        print*, "beta_param: unsupported order:", n
        stop
    end select

  end function beta_param


  !-----------------------------------------------------------------------------
  !> @brief
  !> The coefficients of the quark mass anomalous dimension: [hep-ph:9703284]
  !>   gamma = - ( gamma0 * (as/2pi)
  !>             + gamma1 * (as/2pi)^2
  !>             + gamma2 * (as/2pi)^3
  !>             + ... )
  !
  !> @param[in] n specify the n-th order coefficient
  !> @param[in] nf optional parameter to override the number of fermions
  !-----------------------------------------------------------------------------
  double precision function gamma_param(n, nf)
    integer, intent(in) :: n
    integer, intent(in), optional :: nf
    double precision :: nf_val

    !> number of flavours as double
    if ( present(nf) ) then
      nf_val = nf
    else
      nf_val = 5d0
    end if

    !> switch for the order
    select case(n)
      case(0)
        gamma_param = 2d0
      case(1)
        gamma_param = ( 202d0/3d0         &
                    &   -20d0/9d0*nf_val  &
                    & ) / 4d0
      case(2)
        gamma_param = ( 1249d0                                   &
                    &   +(-2216d0/27d0 -160d0/3d0*zeta3)*nf_val  &
                    &   -140d0/81d0*nf_val**2                    &
                    & ) / 8d0
      case(3)
        gamma_param = ( 4603055d0/162d0        &
                    &   +135680d0/27d0*zeta3   &
                    &   -8800d0*zeta5          &
                    &   +( -91723d0/27d0       &
                    &      -34192d0/9d0*zeta3  &
                    &      +880d0*zeta4        &
                    &      +18400d0/9d0*zeta5  &
                    &   ) * nf_val             &
                    &   +( 5242d0/243d0        &
                    &      +800d0/9d0*zeta3    &
                    &      -160d0/3d0*zeta4    &
                    &   ) * nf_val**2          &
                    &   +( -332d0/243d0        &
                    &      +64d0/27d0*zeta3    &
                    &   ) * nf_val**3          &
                    & ) / 16d0
      case default
        print*, "gamma_param: unsupported order:", n
        stop
    end select

  end function gamma_param


end module Parameters_mod





