!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: PhotonFragmentation [external members: `_frag`]
!
! DESCRIPTION:
!> Module providing infrastructure and functions for photon fragmentation
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module PhotonFragmentation_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: setActive_frag, isActive_frag
  public :: unset_frag, set_frag, isSet_frag
  public :: nz_frag, nf1_frag
  public :: iz_frag,  ip1_frag
  public :: iz2_frag, ip2_frag

  public :: D1_frag

  logical :: qactive_frag = .false.  ! off by default
  integer :: nz_frag = 0  ! number of fragmentations (process dependent)
  !> sub-settings to switch on/off pieces of the fragmentation calculation
  !> number of fragmentation funcitons
  !>  -1 (default) : include all contributions
  !>   0           : only "direct" contribution
  !>   1-2         : number of Dfrag's
  integer :: nDfrag_ = -1

  !> store local state if we're dealing with a fragmentation term
  !> number of convolutions following same convention as `nDfrag_` (s.o.)
  integer :: nset_frag = 0  ! no fragmentation by default
  !$OMP THREADPRIVATE(nset_frag)
  integer :: fctn_frag  ! pick the parametrisation/extraction we want

!> decided to store in KinData
!  !> threadlocal variables defining the current state for the fragmentation
!  integer :: i_frag = 0  ! index of the "parent" (0: inactive)
!  integer :: z_frag      ! momentum fraction of the *photon* --> quark ~ (1-z)
!!$OMP THREADPRIVATE(i_frag,z_frag)

  !> better in KinData => cache & check?!

  !> analogue of ix:  (separate for photon 1 & 2)
  !>  1: ~ delta(1-zreg)  <->  Dfrag(z)
  !>  2: ~ regular        <->  Dfrag(z/zreg)
  integer :: iz_frag = -1, iz2_frag = -1
!$OMP THREADPRIVATE(iz_frag, iz2_frag)
  !> the parton flavour that fragments
  integer :: nf1_frag, ip1_frag, ip2_frag
!$OMP THREADPRIVATE(nf1_frag, ip1_frag, ip2_frag)

  !> fragmentation type:
  !> -1: ALPEH [default]
  !>  0: BFG only the perturbative component
  !>  1: BFG set I
  !>  2: BFG set II
  integer :: iparam_frag = -1

  !> [1] ALEPH parametrisation:
  !>     D. Buskulic et al. [ALEPH Collaboration], Z. Phys. C 69 (1996) 365.
  !>     * one-parameter fit:
  double precision :: mu0_aleph = 0.14d0  ! GeV
  double precision :: C_aleph = -13.26d0
  !>     * simulatneous fit:
  ! double precision :: mu0_aleph = 0.22d0  ! GeV
  ! double precision :: C_aleph = -12.1d0


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate fragmentation
  !
  !> @param[in] active boolean flag to switch on/off
  !> @param[in] fctn which parametrisation/extraction to use
  !> @param[in] frag integer to select # of fragmentations
  !-----------------------------------------------------------------------------
  subroutine setActive_frag(active, fctn, frag)
    use StrHelper_mod, only : to_upper_str
    logical, intent(in) :: active
    character(len=*), intent(in), optional :: fctn
    integer, intent(in), optional :: frag
    qactive_frag = active
    if ( present(frag) )   nDfrag_   = frag
    if ( nDfrag_ > 2 ) then
      print*, "setActive_frag: invalid value for nDfrag_:", nDfrag_
      stop
    end if
    print*, "setActive_frag:", active, "[",nDfrag_,"]"
    if ( present(fctn) ) then
      print*, "setActive_frag: parametrisation = ", fctn
      select case (to_upper_str(fctn))
      case ('ALEPH')
        iparam_frag = -1
      case ('BFG0', 'BFGPERT')
        iparam_frag = 0
      case ('BFG1', 'BFGI')
        iparam_frag = 1
      case ('BFG2', 'BFGII')
        iparam_frag = 2
      case default
        print*, "setActive_frag: unknown fctn = ", fctn
        stop
      end select
    end if
  end subroutine setActive_frag


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of the fragmentation
  !
  !> @return active boolean flag
  !> @param[in] ipass_direct set (ipass = 0) if direct component not active
  !> @param[in] ipass_frag set (ipass = 0) if frag component not active
  !-----------------------------------------------------------------------------
  logical function isActive_frag(ipass_direct, ipass_frag)
    integer, intent(inout), optional :: ipass_direct, ipass_frag
    isActive_frag = qactive_frag
    if ( present(ipass_direct) .and. (nDfrag_ > 0)  ) ipass_direct = 0
    if ( present(ipass_frag)   .and. (nDfrag_ == 0) ) ipass_frag   = 0
  end function isActive_frag


  !> @TODO: DIGEST & add to writer!


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the current state of the fragmentation
  !
  !> @param[in] npar momentum set
  !-----------------------------------------------------------------------------
  subroutine unset_frag(npar)
    use KinData_mod, only : resetFrag_kin, clearCacheLocal_kin, nphotons_brems
    integer, intent(in) :: npar

    if ( .not.isActive_frag() ) return

    call resetFrag_kin(npar)
    call clearCacheLocal_kin(npar)

    !> for PhotonIsolation: we count fragmentation photons
    !> as *negative* bremsstrahlung photons!
    nphotons_brems = nphotons_brems + nset_frag

    nset_frag = 0

  end subroutine unset_frag


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the current state of the fragmentation
  !
  !> @param[in] npar momentum set
  !> @param[in] frag parton index that fragments
  !> @param[inout] ipass check ipass state
  !> @param[in] nDfrag optional integer count number of Dfrag's in a term
  !>            default = -1 (all)
  !> @param[in] frag2 optional second parton index that fragments
  !-----------------------------------------------------------------------------
  subroutine set_frag(npar, frag, ipass, nDfrag, frag2)
    use KinData_mod, only : setFrag_kin, setFrag2_kin, resetFrag_kin, &
      &  clearCacheLocal_kin, nphotons_brems
    integer, intent(in) :: npar, frag
    integer, intent(inout) :: ipass
    integer, intent(in), optional :: nDfrag, frag2
    integer :: nDfrag_in

    if ( .not.isActive_frag() ) then
      ipass = 0
      return
    end if

    if ( (iz_frag/=1) .and. (iz_frag/=2) ) then
      print*, "set_frag: invalid iz_frag", iz_frag
      stop
    end if

    !> @todo: no case where nDfrag is not set -> make it mandatory argument
    !> check if sub-contribution selection triggers
    nDfrag_in = -1  ! = default
    if ( present(nDfrag) ) nDfrag_in = nDfrag
    if ( nDfrag_in < 0 ) then
      print*, "nDfrag not set?!", nDfrag_in
      stop
    end if
    if ( (nDfrag_ >= 0) .and. (nDfrag_in/=nDfrag_) ) ipass = 0

    !> for PhotonIsolation: we count fragmentation photons
    !> as "negative" bremsstrahlung photons:
    !> They do not appear as photons in the MEs
    nphotons_brems = nphotons_brems - 1
    nset_frag = 1

    if (ipass==0) return

    !> I *need* the z-value stored in KinData (for conv function evaluation)
    call setFrag_kin(npar, frag)  ! z = convolution
    !> but for the delta-piece, need to reset iflag so obs are computed correctly
    !> iz_frag == 1 is now the endpoint z=1
    !>if ( iz_frag == 1 ) call resetFrag_kin(npar)  ! delta piece!

    if ( present(frag2) ) then
      !> there's a second convolution!
      nphotons_brems = nphotons_brems - 1
      nset_frag = 2
      call setFrag2_kin(npar, frag2)
    end if

    call clearCacheLocal_kin(npar)

  end subroutine set_frag


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire if we're in a fragmentation convolution block
  !
  !> @param[in] ifrag optional check if `ifrag`-th convolution is set
  !>            default: check if any is set
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isSet_frag(ifrag)
    integer, intent(in), optional :: ifrag
    isSet_frag = .false.
    if ( .not. isActive_frag() ) return
    if ( present(ifrag) ) then
      isSet_frag = (nset_frag >= ifrag)
    else
      isSet_frag = (nset_frag > 0)
    end if
  end function isSet_frag


  !-----------------------------------------------------------------------------
  !> @brief
  !> finite (after MSbar factorization) part of the NLO q->gamma fragmentation
  !> ALEPH parametrisation
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA2 [GeV^2] fragmentation scale
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D1_aleph(pid,z,muA2)
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA2
    double precision :: oMz, PqtoG
    D1_aleph = 0d0
    if (pid == 0) return
    oMz   = 1d0-z
    PqtoG = (1d0+oMz**2)/z
    D1_aleph = PqtoG * log( muA2/(oMz*mu0_aleph)**2 ) + C_aleph
  end function D1_aleph


  !-----------------------------------------------------------------------------
  !> @brief
  !> finite (after MSbar factorization) part of the NLO q->gamma fragmentation
  !> BFG parametrisation
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA2 [GeV^2] fragmentation scale
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D1_bfg(pid, z, muA2)
    use BFGfrag_mod
    use Charges_EW_mod
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA2
    !> old_BFG: double precision :: xglue, xup, xdown, xstrange, xcharm, xbottom  !, aleph
    !> hard coded (alpha0/2pi) value in PHOX family
    double precision, parameter :: al = 1d0/(137.d0*2d0*3.1415926535897932385d0)

    ! print*, "D1_bfg:", pid, z, muA2
    D1_bfg = 0d0

    !> old_BFG:
    ! call FonFra(z, muA2, iparam_frag,  &
    !   &         xglue, xup, xdown, xstrange, xcharm, xbottom)

    select case (pid)
      case (0)
        !> old_BFG: D1gtoG_bfg = xglue
        call frag_bfg(z,muA2,iparam_frag,1,D1_bfg)
      case (1)
        !> old_BFG: D1_bfg = xdown
        call frag_bfg(z,muA2,iparam_frag,3,D1_bfg)
      case (2)
        !> old_BFG: D1_bfg = xup
        call frag_bfg(z,muA2,iparam_frag,2,D1_bfg)
      case (3)
        !> old_BFG: D1_bfg = xstrange
        call frag_bfg(z,muA2,iparam_frag,4,D1_bfg)
      case (4)
        !> old_BFG: D1_bfg = xcharm
        call frag_bfg(z,muA2,iparam_frag,5,D1_bfg)
      case (5)
        !> old_BFG: D1_bfg = xbottom
        call frag_bfg(z,muA2,iparam_frag,6,D1_bfg)
      case default
        print*, "D1_bfg: unknown flavour = ", pid
        stop
    end select

    D1_bfg = D1_bfg / al / z
    if (pid /= 0) D1_bfg = D1_bfg / Qq(pid)**2

  end function D1_bfg


  !-----------------------------------------------------------------------------
  !> @brief
  !> NLO q->gamma fragmentation (stripped from alpha/(2pi) & Qq^2)
  !
  !> @param[in] pid PDG id code for parton
  !> @param[in] z momentum fraction of the photon
  !> @param[in] muA2 [GeV^2] fragmentation scale
  !> @return fragmentation function
  !-----------------------------------------------------------------------------
  double precision function D1_frag(pid, z, muA2)
    integer, intent(in) :: pid
    double precision, intent(in) :: z, muA2

    select case (iparam_frag)
    case (-1)
      D1_frag = D1_aleph(pid,z,muA2)
    case (0,1,2)
      D1_frag = D1_bfg(pid,z,muA2)
    case default
      stop "D1_frag: shouldn't get here"
    end select

  end function D1_frag


end module PhotonFragmentation_mod

