!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: PhotonIsolation [external members: `_isol`]
!
! DESCRIPTION:
!> Module providing various prescriptions to isolate photons from jets.
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module PhotonIsolation_mod
!-------------------------------------------------------------------------------
  use Selectors_mod
  implicit none
  private

  public :: getiAlgo_isol, getName_isol
  public :: registerPhotonIsol_isol, setPhotonIsol_isol, registerSelector_isol
  public :: init_isol, destroy_isol
  public :: setRecombPhoton_isol
  public :: digest_isol, digestObs_isol
  public :: min_shat_isol
  public :: isActive_isol
  public :: applyAll_isol
  public :: initBuffer_isol, destroyBuffer_isol
  public :: setFrag_isol
  
  public :: Isol_t

  type Isol_t
    integer :: ialgo = 0
    integer :: iopt  = 0
    double precision :: R = 0d0
    double precision :: ETmax_const = 0d0
    double precision :: ETmax_epsilon = 0d0
    double precision :: R_min = 0d0
    double precision :: ET_threshold = 0d0
    double precision :: n = 0d0
    double precision :: R_inner = 0d0
    double precision :: zcut = 0d0
    double precision :: beta = 0d0
    integer :: ja2 = -1
    type(Isol_t), pointer :: next => null()
  end type Isol_t

  type(Isol_t), pointer :: head => null()
  type(Isol_t), pointer :: curr_head => null()
  type(Isol_t), pointer :: curr_isol => null()
!$OMP THREADPRIVATE(curr_head, curr_isol)

  integer :: min_nphotons = 0

  !> possibility to recombine photons
  double precision :: R_recomb_photons = 0d0

  !> observables for photon selectors
  integer, parameter :: idp_et      = 1
  integer, parameter :: idp_pt      = 2
  integer, parameter :: idp_e       = 3
  integer, parameter :: idp_y       = 4
  integer, parameter :: idp_abs_y   = 5
  integer, parameter :: idp_eta     = 6
  integer, parameter :: idp_abs_eta = 7
  !> number of photon-observables
  integer, parameter :: n_obsp = 7
  !> keep track of the head of the selectors list
  type(Selector_t), pointer :: sel_p => null()

  !> type to collect information on the Photons
  type Photon_t
    !> flag to identify status
    !>   < 0: Invalid
    !>   = 0: Copied to KinData
    !>   > 0: Photon (identified as a hard, isolated photon)
    integer :: isPhoton
    double precision, dimension(4) :: p
    double precision :: et
  end type Photon_t

  !> threadprivate buffer to minimize re-allocation.
  type(Photon_t), allocatable, dimension(:) :: photons
  double precision, allocatable, dimension(:) :: arr_ETj, arr_Rgj
  integer :: nproto_isol, nphotons_isol
!$OMP THREADPRIVATE(photons)
!$OMP THREADPRIVATE(arr_ETj, arr_Rgj)
!$OMP THREADPRIVATE(nproto_isol, nphotons_isol)

  !> threadprivate buffer for democratic clustering
  integer, allocatable, dimension(:) :: idxphotons_dem
  double precision, allocatable, dimension(:,:) :: pphotons_dem
!$OMP THREADPRIVATE(idxphotons_dem, pphotons_dem)

  !> if true, assume that no photon fragmentation is taking place
  !> (required to run isolated photon + identified hadron)
  logical :: qisfrag = .true.  


contains


  subroutine setFrag_isol(in_qisfrag)
    logical, intent(in) :: in_qisfrag    
    qisfrag = in_qisfrag
  end subroutine setFrag_isol
  

  !-----------------------------------------------------------------------------
  !> @brief
  !> Convert a name string to the isolation algorithm identifier, iAlgo.
  !
  !> @param[in] name string containing the name of the isolation algorithm
  !> @param[out] iopt return the variation type of the algorithm
  !> @return algorithm id (-1 if unsuccessful)
  !-----------------------------------------------------------------------------
  function getiAlgo_isol(algorithm, iopt)
    use StrHelper_mod, only : to_upper_str
    integer :: getiAlgo_isol
    character(len=*), intent(in), optional :: algorithm
    integer, intent(out), optional :: iopt

    if (present(iopt)) iopt = 0

    if (present(algorithm)) then
      select case (to_upper_str(algorithm))

        case ("NONE", "OFF", "FALSE", "0")
          getiAlgo_isol = 0

        case ("FIXED", "ETSUM", "STANDARD")
          getiAlgo_isol = 1

        case ("SMOOTH", "FRIX", "FRIXIONE")
          getiAlgo_isol = 2
          iopt = 0
        case ("SMOOTH_HAD", "FRIX_HAD", "FRIXIONE_HAD")
          getiAlgo_isol = 2
          iopt = 1
        case ("SMOOTHER")
          getiAlgo_isol = 2
          iopt = 2

        case ("HYBRID_SMOOTH", "HYBRID_FRIX", "HYBRID_FRIXIONE")
          getiAlgo_isol = 3
          iopt = 0
        case ("HYBRID_SMOOTH_HAD", "HYBRID_FRIX_HAD", "HYBRID_FRIXIONE_HAD")
          getiAlgo_isol = 3
          iopt = 1
        case ("HYBRID_SMOOTHER")
          getiAlgo_isol = 3
          iopt = 2

        case ("SOFTDROP", "SOFT_DROP", "SOFT-DROP")
          getiAlgo_isol = 4

        case ("DEMOCRATIC", "DEMO")
          getiAlgo_isol = 5

        case default
          getiAlgo_isol = -1  ! signal failure
      end select
    else
      if ( associated(curr_isol) ) then
        !> act as getter
        getiAlgo_isol = curr_isol%ialgo
        if (present(iopt)) iopt = curr_isol%iopt
      else
        getiAlgo_isol = 0
      end if
    end if

  end function getiAlgo_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> The isolation algorithm to a name string
  !
  !> @param[in] isol isolation object
  !> @return
  !-----------------------------------------------------------------------------
  function getName_isol(isol)
    use StrHelper_mod
    character(:), allocatable :: getName_isol
    type(Isol_t), intent(in) :: isol

    select case (isol%ialgo)

    case (0)
      !> none
      select case (isol%iopt)
      case (0)
        getName_isol =  "none"
      case default
        print*, "getName_isol: invalid none option", isol%iopt
        stop
      end select

    case (1)
      !> fixed cone
      select case (isol%iopt)
      case (0)
        getName_isol =  "fixed"
      case default
        print*, "getName_isol: invalid fixed-cone option", isol%iopt
        stop
      end select

    case (2)
      !> smooth cone
      select case (isol%iopt)
      case (0)
        getName_isol =  "smooth"
      case (1)
        getName_isol =  "smooth_had"
      case (2)
        getName_isol =  "smoother"
      case default
        print*, "getName_isol: invalid smooth-cone option", isol%iopt
        stop
      end select

    case (3)
      !> hybrid cone
      select case (isol%iopt)
      case (0)
        getName_isol =  "hybrid_smooth"
      case (1)
        getName_isol =  "hybrid_smooth_had"
      case (2)
        getName_isol =  "hybrid_smoother"
      case default
        print*, "getName_isol: invalid hybrid-cone option", isol%iopt
        stop
      end select

    case (4)
      !> soft drop
      select case (isol%iopt)
      case (0)
        getName_isol =  "softdrop"
      case default
        print*, "getName_isol: invalid soft-drop option", isol%iopt
        stop
      end select

    case (5)
      !> democratic
      select case (isol%iopt)
      case (0)
        getName_isol =  "democratic"
      case default
        print*, "getName_isol: invalid none option", isol%iopt
        stop
      end select

    case default
        print*, "getName_isol: invalid setting", isol%ialgo, isol%iopt
        stop
    end select


  end function getName_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new photon isolation criterion
  !
  !> @param[in] ialgo, iopt: to define the algorithm and its variant (see above)
  !> @param[in] R: the cone size for the isoilation
  !> @param[in] ETmax_const, ETmax_epsilon: defines maximum of transverse hadronic energy:
  !>            ETmax = ETmax_const + ETmax_epsilon * pT_gamma
  !> @param[in] R_min, ET_threshold: fixed-cone parameters
  !> @param[in] n: smooth-cone parameter controling the profile
  !> @param[in] R_inner: inner cone size for the hybrid prescription
  !> @param[in] zcut, beta, ja2 : softdrop parameters (pending merge from modules)
  !> @param[in] isol_p : optional list to add the isolation type (for multi-runs)
  !-----------------------------------------------------------------------------
  subroutine registerPhotonIsol_isol( ialgo, iopt,  &
                                    & R, &
                                    & ETmax_const, ETmax_epsilon,  &
                                    & R_min, ET_threshold,  &
                                    & n,  &
                                    & R_inner,  &
                                    & zcut, beta, ja2,  &
                                    & isol_p )
    ! use Softdrop_mod
    integer, intent(in) :: ialgo, iopt
    double precision, optional, intent(in) :: R
    double precision, optional, intent(in) :: ETmax_const, ETmax_epsilon
    double precision, optional, intent(in) :: R_min, ET_threshold
    double precision, optional, intent(in) :: n
    double precision, optional, intent(in) :: R_inner
    double precision, optional, intent(in) :: zcut, beta
    integer, optional, intent(in) :: ja2
    type(Isol_t), pointer, optional, intent(inout) :: isol_p
    type(Isol_t), pointer :: newIsol, it

    if ( present(isol_p) ) then
      it => isol_p
    else
      if ( ialgo == 0 ) return  ! no need to treat "none"
      it => head
    end if
    if ( associated(it) ) then
      do while ( associated(it%next) )
        it => it%next
      end do
    end if

    allocate(newIsol)

    !> copy over all the passed input
    !> consistency checks already done in IOHelper!
    newIsol%ialgo = ialgo
    newIsol%iopt = iopt
    if( present(R) )             newIsol%R = R
    if( present(ETmax_const) )   newIsol%ETmax_const = ETmax_const
    if( present(ETmax_epsilon) ) newIsol%ETmax_epsilon = ETmax_epsilon
    if( present(R_min) )         newIsol%R_min = R_min
    if( present(ET_threshold) )  newIsol%ET_threshold = ET_threshold
    if( present(n) )             newIsol%n = n
    if( present(R_inner) )       newIsol%R_inner = R_inner
    if( present(zcut) )          newIsol%zcut = zcut
    if( present(beta) )          newIsol%beta = beta
    if( present(ja2) )           newIsol%ja2 = ja2

    if ( .not.associated(it) ) then
      if ( present(isol_p) ) then
        isol_p => newIsol
      else
        head => newIsol
      end if
    else
      if ( present(isol_p) ) then
        if ( it%ialgo == 0 ) then
          !> kill previous `none`
          if ( .not.associated(it,isol_p) ) stop "registerPhotonIsol_isol: non-head `none`?!"
          deallocate(it)
          isol_p => newIsol
        else if ( ialgo ==0 ) then
          !> no need to add `none` to existing list
          deallocate(newIsol)
        else
          it%next => newIsol  ! add to the tail
        end if
      else
        it%next => newIsol  ! add to the tail
      end if
    end if

  end subroutine registerPhotonIsol_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all registered isolation algorithms
  !-----------------------------------------------------------------------------
  subroutine destroyPhotonIsol_isol()
    type(Isol_t), pointer :: it, next
    it => head
    do while ( associated(it) )
      !> save next before we destroy it
      next => it%next
      deallocate(it)
      it => next  ! loop
    end do
    head => null()
    curr_head => null()
  end subroutine destroyPhotonIsol_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the Isolation (default -> head)
  !
  !> @param[in] isol_p ooptional head to be set (for MultiRun)
  !-----------------------------------------------------------------------------
  subroutine setPhotonIsol_isol(isol_p)
    type(Isol_t), pointer, intent(in), optional :: isol_p

    if ( present(isol_p) ) then
      curr_head => isol_p
    else
      curr_head => head
    end if

  end subroutine setPhotonIsol_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new selector for photons
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] lower optional lower cut
  !> @param[in] upper optional upper cut
  !> @param[in] invert optional flag controlling whether we "accept" or "reject"
  !-----------------------------------------------------------------------------
  subroutine registerSelector_isol(obsName, lower, upper, invert)
    use StrHelper_mod, only : to_upper_str
    character(len=*), intent(in) :: obsName
    double precision, optional, intent(in) :: lower, upper
    logical, optional, intent(in) :: invert
    integer :: obsId
    type(Selector_t), pointer :: newSel

    select case (to_upper_str(obsName))
    case ("PHOTONS_ET")
      obsId = idp_et
    case ("PHOTONS_PT")
      obsId = idp_pt
    case ("PHOTONS_E")
      obsId = idp_e
    case ("PHOTONS_Y")
      obsId = idp_y
    case ("PHOTONS_ABS_Y")
      obsId = idp_abs_y
    case ("PHOTONS_ETA")
      obsId = idp_eta
    case ("PHOTONS_ABS_ETA")
      obsId = idp_abs_eta
    case default
      print*, "registerSelector_isol: unknown observable: ", obsName
      stop
    end select

    allocate(newSel)
    newSel%obsId  = obsId
    if (present(lower)) newSel%lower = lower
    if (present(upper)) newSel%upper = upper
    newSel%qlower = present(lower)
    newSel%qupper = present(upper)
    if (present(invert)) then
      newSel%qinvert = invert
    end if
    newSel%qisint = .false.  ! all double observables

    !> push it to the stack
    newSel%next => sel_p
    sel_p => newSel

  end subroutine registerSelector_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all registered selectors
  !-----------------------------------------------------------------------------
  subroutine destroySelectors_isol()
    type(Selector_t), pointer :: it, next
    it => sel_p
    do while ( associated(it) )
      !> save next before we destroy it
      next => it%next
      deallocate(it)
      it => next  ! loop
    end do
    sel_p => null()
  end subroutine destroySelectors_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialisation routine
  !-----------------------------------------------------------------------------
  subroutine init_isol()
    use Run_mod
    use Process_mod
    use Observables_mod, only : isUsed_obs
    double precision :: min_sel
    integer :: id

    !> init using process information
    min_nphotons = n_fs_photons_proc  ! default

    !> override from selectors?
    min_sel = -1d0  ! start with invalid value
    id = isUsed_obs("nphotons")
    if (id > 0)  call getExtremes_sel(id, minVal=min_sel)
    if ( min_sel >= 0d0 ) then
      print*, "found an `nphotons` selector"
      print*, "this will potentially override the global settings for"
      min_nphotons = INT( min_sel )
      print*, "> minnphotons:", min_nphotons, "[default:",n_fs_photons_proc,"]"
    end if

    !> now that min # photons is set, we can update shat_min from photon cuts
    call set_smin_run(min_shat_isol())

    !> this should be also done in the parallel block
    call setPhotonIsol_isol()

  end subroutine init_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> clear photons and isolation algorithms
  !-----------------------------------------------------------------------------
  subroutine destroy_isol()
    call destroyPhotonIsol_isol()
    call destroySelectors_isol()
  end subroutine destroy_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the cone size for photon recombination
  !
  !> @param[in] Rcone Delta R cone
  !-----------------------------------------------------------------------------
  subroutine setRecombPhoton_isol(Rcone)
    double precision, intent(in) :: Rcone
    if (Rcone < 0d0) then
      print*, "setRecombPhoton_isol: invalid Rcone = ", Rcone
      stop
    end if
    R_recomb_photons = Rcone
  end subroutine setRecombPhoton_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> print out a list of active isolation criteria & cuts
  !-----------------------------------------------------------------------------
  subroutine digest_isol(unit)
    integer, optional, intent(in) :: unit
    type(Isol_t), pointer :: it
    type(Selector_t), pointer :: it_sel
    character(len=6)  :: seltype
    character(len=32) :: name
    integer :: iout
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_isol =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_var =  &
      & '(1x,"*",16x,A15,"  = ",1f8.4  ,T103,"*")'
    character(len=*), parameter :: fmt_int =  &
      & '(1x,"*",16x,A15,"  = ",I8,T103,"*")'
    character(len=*), parameter :: fmt_min_max =  &
      & '(1x,"*",16x,1f8.4," <= ",A15," <= ",1f8.4,T103,"*")'
    character(len=*), parameter :: fmt_min =  &
      & '(1x,"*",3x,A6,":",1f12.2," <= ",A15,59x,"*")'
    character(len=*), parameter :: fmt_max =  &
      & '(1x,"*",3x,A6,":",16x,A15," <= ",1f12.2,43x,"*")'
    character(len=*), parameter :: fmt_sel =  &
      & '(1x,"*",3x,A6,":",1f12.2," <= ",A15," <= ",1f12.2,43x,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    if ( associated(head) ) then
      write(iout,fmt_str) ""
      write(iout,fmt_str) "> photon isolation <"
      write(iout,fmt_str) ""
      it => head
      do while ( associated(it) )
        write(iout,fmt_isol) getName_isol(it)
        write(iout,fmt_min_max) it%R_min, "R", it%R
        select case(it%ialgo)
        case(1)  ! fixed
          write(iout,fmt_var) "ETmax_const", it%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", it%ETmax_epsilon
          write(iout,fmt_var) "ET_threshold", it%ET_threshold
        case(2)  ! smooth
          write(iout,fmt_var) "ETmax_const", it%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", it%ETmax_epsilon
          write(iout,fmt_var) "n", it%n
        case(3)  ! hybrid
          write(iout,fmt_var) "ETmax_const", it%ETmax_const
          write(iout,fmt_var) "ETmax_epsilon", it%ETmax_epsilon
          write(iout,fmt_var) "n", it%n
          write(iout,fmt_var) "R_inner", it%R_inner
        case(4)  ! softdrop
          write(iout,fmt_var) "zcut", it%zcut
          write(iout,fmt_var) "beta", it%beta
          write(iout,fmt_int) "recl.alg", it%ja2
        case(5)  ! democratic
          write(iout,fmt_var) "zcut", it%zcut
        case default
          write(iout,*) "SOMETHING WENT WRONG"
        end select
        it => it%next
      end do
    endif

    if ( associated(sel_p) ) then
      write(iout,fmt_str) ""
      write(iout,fmt_str) "> photon cuts <"
      write(iout,fmt_str) ""
      it_sel => sel_p
      do while ( associated(it_sel) )
        seltype = ''
        if ( it_sel%qinvert ) then
          seltype = 'reject'
        else
          seltype = 'accept'
        endif
        name = ''
        select case (it_sel%obsId)
        case (idp_et)
          name = "photons_et"
        case (idp_pt)
          name = "photons_pt"
        case (idp_e)
          name = "photons_e"
        case (idp_y)
          name = "photons_y"
        case (idp_abs_y)
          name = "photons_abs_y"
        case (idp_eta)
          name = "photons_eta"
        case (idp_abs_eta)
          name = "photons_abs_eta"
        case default
          print*, "digest_isol: unknown id:", it_sel%obsId
        end select
        if (it_sel%qlower .and. it_sel%qupper) then
          write(iout,fmt_sel) seltype, it_sel%lower, name, it_sel%upper
        else if (it_sel%qlower .and. .not.it_sel%qupper) then
          write(iout,fmt_min) seltype, it_sel%lower, name
        else if (.not.it_sel%qlower .and. it_sel%qupper) then
          write(iout,fmt_max) seltype, name, it_sel%upper
        endif
        it_sel => it_sel%next
      end do
    endif

  end subroutine digest_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> print out a list of photon-isolation observables
  !-----------------------------------------------------------------------------
  subroutine digestObs_isol(unit)
    integer, optional, intent(in) :: unit
    integer :: i
    character(len=32) :: name
    integer :: iout
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    write(iout,fmt_str) ""
    write(iout,fmt_str) "> photon cuts <"
    write(iout,fmt_str) ""

    do i=1,n_obsp
      select case (i)
      case (idp_et)
        name = "photons_et"
      case (idp_pt)
        name = "photons_pt"
      case (idp_e)
        name = "photons_e"
      case (idp_y)
        name = "photons_y"
      case (idp_abs_y)
        name = "photons_abs_y"
      case (idp_eta)
        name = "photons_eta"
      case (idp_abs_eta)
        name = "photons_abs_eta"
      case default
        print*, "digestObs_isol: unknown id:", i
      end select
      write(iout,fmt_str) name
    end do

  end subroutine digestObs_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine from the photon selectors the minimal partonic
  !> centre-of-mass energy (squared)
  !
  !> @return minimal value for shat [GeV^2]
  !-----------------------------------------------------------------------------
  double precision function min_shat_isol()
    double precision :: min_val
    type(Selector_t), pointer :: it_sel
    min_shat_isol = 0d0
    !> no photons required => no smin contraint
    if (min_nphotons == 0) return
    !> loop over all selectors
    it_sel => sel_p
    do while ( associated(it_sel) )
      !> only consider non-inverted selectors
      if ( .not. it_sel%qinvert ) then
        min_val = it_sel%lower
        !> min depends on observable type
        select case (it_sel%obsId)
        case (idp_et, idp_pt, idp_e)
          min_shat_isol = max(min_shat_isol, (2d0*min_val)**2)
        ! case ("invariant-mass type")
        !   min_shat_isol = max(min_shat_isol, min_val**2)
        end select
      end if
      it_sel => it_sel%next
    end do  ! it_sel
  end function min_shat_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to enquire after status of photon isolation
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_isol()
    isActive_isol = associated(head)
  end function isActive_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to initialize the (threadprivate) buffer for the jet-algorithm
  !
  !> @param[inout] maxnphotons largest array size needed to store all photons
  !> @param[inout] maxnpartons largest array size needed to store all partons
  !-----------------------------------------------------------------------------
  subroutine initBuffer_isol(maxnphotons, maxnpartons)
    use PhotonFragmentation_mod
    integer, intent(inout) :: maxnphotons, maxnpartons
    type(Isol_t), pointer :: it
    logical :: qdemo
    curr_head => head
    !> if fragmentation is active, increase buffer size
    if ( isActive_frag() ) then
      maxnphotons = maxnphotons + nz_frag
      maxnpartons = maxnpartons + nz_frag
    end if
    !> if democratic clustering registered, accomodate photons
    qdemo = .false.
    if ( associated(head) ) then
      it => head
      do while ( associated(it) )
        if (it%ialgo == 5) qdemo = .true.
        it => it%next
      end do
    endif
    if (qdemo) maxnpartons = maxnpartons + maxnphotons
    !> time to allocate...
    allocate( photons(maxnphotons) )
    allocate( arr_ETj(maxnpartons) )
    allocate( arr_Rgj(maxnpartons) )
  end subroutine initBuffer_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to deallocate the (threadprivate) buffer
  !-----------------------------------------------------------------------------
  subroutine destroyBuffer_isol()
    curr_head => null()
    if (allocated(photons)) deallocate(photons)
    if (allocated(arr_ETj)) deallocate(arr_ETj)
    if (allocated(arr_Rgj)) deallocate(arr_Rgj)

    if (allocated(idxphotons_dem)) deallocate(idxphotons_dem)
    if (allocated(pphotons_dem))   deallocate(pphotons_dem)
  end subroutine destroyBuffer_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> Applies all registered members of the Isol_t type sequentially.
  !
  !> @param[in] npar kinematics specified by the number of external partlicles
  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
  !-----------------------------------------------------------------------------
  recursive subroutine applyAll_isol(npar, ipass)
    use Process_mod
    use KinData_mod
    use ObsHelper_mod
    use Mapping_mod
    use Proj2Born_mod
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer :: i,j, ialgo, iphoton
    integer :: iphoton_low, iphoton_upp
    type(Kin_t), pointer :: kd
    double precision :: eti, etj
    !> for mapping
    logical :: qmap
    integer :: npar_up, ipass_up
    integer :: ih1,iu1,iu2,ih2, jh1,js1,jh2
    type(Selector_t), pointer :: it_sel
    double precision :: valObs
    !> for photon recombination
    integer :: imin, jmin
    double precision :: dRmin, dRij
    !> g1 - g2 matching
    !> threshold = 1d-X -> require at least ~ X digit agreement before rand
    double precision, parameter :: threshold = 1d-7
    double precision :: et1, et2, del1, del2

    ! print*, "applyAll_isol:"
    ! print*, "  > npar", npar
    ! print*, "  > ipass", ipass
    ! print*, "  > n_fs_photons_proc", n_fs_photons_proc
    ! print*, "  > min_nphotons", min_nphotons
    ! print*, "  > nphotons_brems", nphotons_brems
    ! print*, "  > max_nphotons_brems", max_nphotons_brems

    if ( ipass == 0 ) return
    !> negative number <--> fragmentation photons
    ! if ( nphotons_brems < 0 ) then
    !   print*, "applyAll_isol: negative # of unresolved photons?!:", nphotons_brems
    !   stop
    ! end if
    if ( nphotons_brems > max_nphotons_brems ) then
      print*, "applyAll_isol: more photons than expected:", nphotons_brems, max_nphotons_brems
      stop
    end if

    !> prettify code
    kd => kin(npar)

    !> determine the number of photons
    nphotons_isol = n_fs_photons_proc + nphotons_brems
    !> subtract all photons that became unresolved in a mapping
    !> @todo: maybe keep track in set_map of the photons that are killed?
    qmap = .false.
    npar_up = -1
    !> the deepest level we can have comes from a P2B mapping
    if ( isActive_p2b() .and. (npar==npar_p2b) ) then
      npar_up = kd%npar_from
      !> nothing to check for `npar_up` as only QCD emissions mapped
    end if
    !> can't check mapping consistency (`npar_test`) directly
    !> when P2B is on or in a recursive call for the matching -> `npar_lv`
    qmap = getCurrent_map(ih1,iu1,iu2,ih2, npar_from=npar_up)
    if ( qmap ) then
      if ( npar_up == npar ) then
        !> we live at the higher momentum set => ignore mappings below!
        ih1 = -1
        iu1 = -1
        iu2 = -1
        ih2 = -1
      end if
      ! print*, "ih1,iu1,iu2,ih2", ih1,iu1,iu2,ih2
      if ( getParent_map(jh1,js1,jh2, npar_from=npar_up) ) then
        !> all indices in the highest real-emission momentum set!
        ih1 = idxUpParent_map(ih1)
        iu1 = idxUpParent_map(iu1)
        iu2 = idxUpParent_map(iu2)
        ih2 = idxUpParent_map(ih2)
        ! print*, "idxUpParent:", ih1,iu1,iu2,ih2
        if ( npar_up == npar ) then
          !> we live at the higher momentum set => ignore mappings below!
          jh1 = -1
          js1 = -1
          jh2 = -1
        end if
      end if
      ! print*, "jh1,js1,jh2", jh1,js1,jh2
      !> bounds in the highest momentum set
      if ( ibegin_photons_proc > 0 ) then
        iphoton_low = ibegin_photons_proc
        iphoton_upp = iphoton_low + (nphotons_isol-1)
      else
        iphoton_upp = npar_up + ibegin_photons_proc + 1
        iphoton_low = iphoton_upp - (nphotons_isol-1)
      end if
      ! print*, "  > iphoton_low", iphoton_low
      ! print*, "  > iphoton_upp", iphoton_upp
      ! print*, "  > KILLING", count( ((/ih1,iu1,iu2,ih2,jh1,js1,jh2/) >= iphoton_low) .and.  &
      !                             & ((/ih1,iu1,iu2,ih2,jh1,js1,jh2/) <= iphoton_upp) )
      nphotons_isol = nphotons_isol  &
        &           - count( ((/ih1,iu1,iu2,ih2,jh1,js1,jh2/) >= iphoton_low) .and.  &
        &                    ((/ih1,iu1,iu2,ih2,jh1,js1,jh2/) <= iphoton_upp) )
    end if  ! end: qmap
    nproto_isol = nphotons_isol  ! without fragmentation -> for loop below
    !> do we have a parton fragmenting?
    if ( qisfrag .and. kd%ifrag  > 0 ) nphotons_isol = nphotons_isol + 1
    if ( qisfrag .and. kd%i2frag > 0 ) nphotons_isol = nphotons_isol + 1
    !> do we have enough photons?
    if ( nphotons_isol < min_nphotons ) then
      ipass = 0
      return
    end if


    ! print*, "  > nphotons_isol", nphotons_isol
    ! print*, "  > cached?", ( (kd%nphotons >= 0) .and. (count(kd%iphotons>=0) == nphotons) )
    ! !> cached?
    ! if ( (kd%nphotons >= 0) .and. (count(kd%iphotons>=0) == nphotons) ) then
    !   if ( kd%nphotons < min_nphotons ) ipass = 0
    !   return
    ! end if
    !> either totally new or the # of photons has changed: clear the cache!
    call clearCacheLocal_kin(npar)


    !> number of photons changed => parton indices need to be adjusted as well!
    ! print*, "########## ", npar
    ! print*, "before", kd%ijets_lower, kd%ijets_upper, kd%nproto
    kd%ijets_lower = ibegin_partons_proc
    kd%nproto      = (npar-2) - n_fs_spectators_proc - nproto_isol  ! don't count fragmentation!
    kd%ijets_upper = ibegin_partons_proc + kd%nproto - 1
    ! print*, "after ", kd%ijets_lower, kd%ijets_upper, kd%nproto
    ! print*, "####################"


    !> copy over the momenta and initialize
    photons(:)%isPhoton = -1
    do i=1,nproto_isol
      iphoton = ibegin_photons_proc + sign(i-1,ibegin_photons_proc)
      if (iphoton < 0) then  ! see "reverse" counting convention above
        iphoton = npar + iphoton + 1
      end if
      ! print*, "photon", i, ">", iphoton, ":", kd%p(:,iphoton)
      photons(i)%isPhoton = 1
      photons(i)%p(:)     = kd%p(:,iphoton)
      photons(i)%et       = v1_et( kd%p(:,iphoton) )
    end do
    if ( qisfrag .and. kd%ifrag > 0 ) then
      nproto_isol = nproto_isol + 1
      photons(nproto_isol)%isPhoton = 1
      photons(nproto_isol)%p(:)     = kd%zfrag * kd%p(:,kd%ifrag)
      photons(nproto_isol)%et       = v1_et( photons(nproto_isol)%p )
    end if
    if ( qisfrag .and. kd%i2frag > 0 ) then
      nproto_isol = nproto_isol + 1
      photons(nproto_isol)%isPhoton = 1
      photons(nproto_isol)%p(:)     = kd%z2frag * kd%p(:,kd%i2frag)
      photons(nproto_isol)%et       = v1_et( photons(nproto_isol)%p )
    end if
    if ( nproto_isol /= nphotons_isol ) then
      print*, "applyAll_isol: nproto_isol /= nphotons_isol", nproto_isol, nphotons_isol
      stop
    end if


    !> apply the photon recombination
    if ( R_recomb_photons > 0d0 ) then
      do while (.true.)
        dRmin = -1d0
        imin  = -1
        jmin  = -1
        do i=1,nproto_isol-1
          if ( photons(i)%isPhoton <= 0 ) cycle
          do j=i+1,nproto_isol
            if ( photons(j)%isPhoton <= 0 ) cycle
            dRij = v2_delta_R( photons(i)%p , photons(j)%p )
            if ( (dRij<dRmin) .or. (dRmin<0d0) ) then
              dRmin = dRij
              imin  = i
              jmin  = j
            end if
          end do  ! j
        end do  ! i
        if ( (dRmin>0d0) .and. (dRmin<R_recomb_photons) ) then
          !> recombine imin & jmin to one photon (j into i)
          photons(imin)%isPhoton = 1
          photons(imin)%p(:)     = photons(imin)%p(:) + photons(jmin)%p(:)
          photons(imin)%et       = v1_et( photons(imin)%p )
          photons(jmin)%isPhoton = 0
          photons(jmin)%p(:)     = 0d0
          photons(jmin)%et       = 0d0
          nphotons_isol = nphotons_isol -1
        else
          exit  ! no pairs left to recombine
        end if
      end do  ! while
    end if  ! R_recomb_photons


    !> apply all isolation criteria
    curr_isol => curr_head
    do while ( associated(curr_isol) )
      ialgo = curr_isol%ialgo
      select case(ialgo)
      case (0)   ! none
        !> do nothing... (can't cycle)
      case (1)   ! fixed
        call apply_fixed(npar, ipass)
      case (2,3) ! smooth & hybrid
        call apply_smooth(npar, ipass)
      case (4)   ! softdrop
        stop "SOFTDROP: TODO!"
        ! call apply_softdrop(npar, ipass)
      case (5) ! democratic
        call apply_democratic(npar, ipass)
      case default
        stop "applyAll_isol: unknown algorithm"
      end select
      if ( ipass == 0 ) return
      curr_isol => curr_isol%next
    end do  ! <-> isol
    !> did enough photons survive?
    if ( nphotons_isol < min_nphotons ) then
      kd%nphotons = 0  ! for the cache
      ipass = 0
      return
    end if


    !> apply photon cuts
    it_sel => sel_p
    do while ( associated(it_sel) )
      !> loop over all photons
      do i = 1,nproto_isol
        if ( photons(i)%isPhoton <= 0 ) cycle
        !> compute the observable
        select case (it_sel%obsId)
        case (idp_et)
          valObs = photons(i)%et
        case (idp_pt)
          valObs = v1_pt( photons(i)%p )
        case (idp_e)
          valObs = photons(i)%p(4)
        case (idp_y)
          valObs = v1_y( photons(i)%p )
        case (idp_abs_y)
          valObs = abs( v1_y( photons(i)%p ) )
        case (idp_eta)
          valObs = v1_eta( photons(i)%p )
        case (idp_abs_eta)
          valObs = abs( v1_eta( photons(i)%p ) )
        case default
          print*, "applyAll_isol: unknown obsId in photon selectors?!"
          stop
        end select
        !> kill the photons that don't pass the cuts
        if ( .not. qpass_sel(it_sel, valObs) ) then
          photons(i)%isPhoton = -1
          nphotons_isol = nphotons_isol - 1
          !print*, " > applyAll_isol: kill", i
          !read(*,*)
        end if
      end do  ! iphoton
      it_sel => it_sel%next
    end do  ! it_sel
    !> did enough photons survive?
    if ( nphotons_isol < min_nphotons ) then
      kd%nphotons = 0  ! for the cache
      ipass = 0
      return
    end if


    !> write into KinData sorted in et
    kd%nphotons = 0
    et1 = -1d0  ! g1 & g2 for matching later...
    et2 = -1d0
    do i=1,nphotons_isol
      iphoton = -1
      eti = -1d0
      do j=1,nproto_isol
        if ( photons(j)%isPhoton <= 0 ) cycle
        etj = photons(j)%et
        if ( (iphoton < 0) .or. (etj > eti) ) then
          iphoton = j
          eti = etj
        end if
      end do
      if ( iphoton < 0 ) stop "applyAll_isol: ran out of photons?!"
      !> copy over to KinData
      kd%pphotons(:,i) = photons(iphoton)%p(:)
      kd%nphotons = kd%nphotons + 1
      photons(iphoton)%isPhoton = 0  ! used up
      if (i==1) et1 = eti
      if (i==2) et2 = eti
      ! print*, "----- photon ", iphoton, " -----"
      ! print*, " y  ", v1_y( photons(iphoton)%p ), v1_eta( photons(iphoton)%p )
      ! print*, " pt ", v1_pt( photons(iphoton)%p ), v1_et( photons(iphoton)%p )
    end do
    !> sanity check during debugging phase
    if ( nphotons_isol /= kd%nphotons ) then
      print*, "applyAll_isol: inconsistent state after KinData copy",  &
        & nphotons_isol, kd%nphotons
      do i=1,nproto_isol
        print*, i, ">", photons(i)%isPhoton, photons(i)%et
      end do
      stop
    end if

    !> match leading & sub-leading if both same
    npar_up = kd%npar_from  ! this also deals with P2B mappings
    if ( (kd%nphotons == 2) .and.  &
       & (abs((et1-et2)/(et1+et2)) < threshold) ) then
      ! print*, "g1 - g2 matching: ", et1, et2, "|", npar, npar_up
      if ( npar < npar_up ) then
        !> subtraction term => match to higher level
        !> make sure higher momentum set has photons identified
        ipass_up = 1
        if ( kin(npar_up)%nphotons < 0 ) then
          call applyAll_isol(npar_up, ipass_up)
        end if
        !> do the matching
        if ( (ipass_up/=0) .and. (kin(npar_up)%nphotons>=1) ) then
          del1 = deltaGG(kd%pphotons(:,1), kin(npar_up)%pphotons(:,1))
          del2 = deltaGG(kd%pphotons(:,2), kin(npar_up)%pphotons(:,1))
          ! print*, ">>", del1, del2
          ! read(*,*)
          if (del2 < del1) call flipGG()
        else
          ! print*, "no photons to match against"
          call randGG()
        end if
      else
        !> no subtraction  => randomise
        ! print*, "no subtraction  => randomise"
        call randGG()
      end if
    else
      ! print*, "g1 - g2 matching: none"
    end if


  contains


    !> funtion to measure "closeness" of two photon momenta (for matching)
    pure double precision function deltaGG(p1, p2)
      double precision, dimension(4), intent(in) :: p1, p2
      integer :: i

      ! deltaGG = 0d0
      ! do i=1,4
      !   deltaGG = deltaGG + (p1(i)-p2(i))**2
      ! end do

      deltaGG = v2_delta_R(p1,p2)

      ! deltaGG = 0d0
      ! do i=1,3
      !   deltaGG = deltaGG + (p1(i)-p2(i))**2
      ! end do

      !deltaGG = abs( dot(p1,p2) )

    end function deltaGG

    !> routine to flip g1 <-> g2
    subroutine flipGG()
      double precision :: p_dummy(4)
      ! print*, "FLIP"
      p_dummy(:)       = kd%pphotons(:,1)
      kd%pphotons(:,1) = kd%pphotons(:,2)
      kd%pphotons(:,2) = p_dummy(:)
    end subroutine flipGG

    !> routine to randomise g1, g2
    subroutine randGG
      double precision :: rnd
      !!> integer, save :: ntot = 0, nflp = 0
      !> this is dangerous and bad!!!
      !> For virtual pieces I need to correlate ix=1,..,4!!!
      ! call random_number(rnd)
      !> WRONG: this will no generate 50% flip probability! (why?)
      ! !> use et1 & et2 (offset by 2 digits) around threshold digit
      ! rnd = (et1/10d0 + et2*10d0) / threshold
      ! rnd = (rnd) - INT(rnd)
      double precision xvegas
      common /vdata/xvegas(20)
!$omp threadprivate(/vdata/)
      rnd = xvegas(1) / threshold  ! take the ~7th digit
      rnd = rnd - INT(rnd)
      ! print*, "RAND", rnd, xvegas(1)
      ! read(*,*)
      !> flip the two with a 50% probability
      if (rnd < 0.5d0) call flipGG()
      !!> !> get some stats to see 50% is correct
      !!> ntot = ntot+1
      !!> if (rnd < 0.5d0) nflp = nflp+1
      !!> if (mod(ntot,1000)==0) print*, "RAND", nflp, ntot, DBLE(nflp)/DBLE(ntot)
    end subroutine randGG

  end subroutine applyAll_isol


  !-----------------------------------------------------------------------------
  !> @brief
  !> impose a cut on the maximum sum(ET) in a cone around the photons
  !
  !> @note
  !> This is an isolation criterion that is not collinear safe.
  !> For H->gg it's fine but, in general, there's a collienar singularity
  !> because the fragmentation componend is not suppressed like in the
  !> smooth cone isolation.
  !
  !> @param[in] npar kinematics specified by the number of external partlicles
  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
  !-----------------------------------------------------------------------------
  subroutine apply_fixed(npar, ipass)
    use KinData_mod
    use ObsHelper_mod
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer :: i,j, j_kd
    double precision :: Rgj, sumET, ETmax, ETj
    type(Kin_t), pointer :: kd

    if ( ipass == 0 ) return ! nothing to do
    if ( curr_isol%ialgo /= 1 ) stop "apply_fixed: ialgo mismatch"

    !> prettify code
    kd => kin(npar)

    !> apply isolation criterion to each photon sequentially
    loop_photons: do i = 1,nproto_isol
      if ( photons(i)%isPhoton <= 0 ) cycle
      !> maximum allowed hadronic (partonic) ET inside the photon cone
      ETmax = curr_isol%ETmax_const  &
        &   + curr_isol%ETmax_epsilon * photons(i)%et
      !> (no need to use a buffer)
      !> accumulate ET of partons inside the delta_etsum cone around the photon
      sumET = 0d0
      do j = 1, kd%nproto
        !> index in KinData
        j_kd = kd%ijets_lower + j - 1
        Rgj = v2_delta_R( photons(i)%p , kd%p(:,j_kd) )  ! direction indep. of (1-z)
        !> skip partons outside of the cone
        if ( (Rgj > curr_isol%R) .or. (Rgj < curr_isol%R_min) ) cycle
        ETj = v1_et( kd%p(:,j_kd) )
        if ( qisfrag .and. kd%ifrag ==  j_kd ) ETj = (1d0-kd%zfrag)  * ETj  ! frag. linear
        if ( qisfrag .and. kd%i2frag == j_kd ) ETj = (1d0-kd%z2frag) * ETj  ! frag. linear
        !> if an ET threshold is set (>0), only keep partons above that theshold
        if ( (curr_isol%ET_threshold > 0d0) .and. (ETj < curr_isol%ET_threshold) ) cycle
        !> passed parton requirements => accumulate
        sumET = sumET + ETj
      end do  ! j <-> partons
      !> check condition: sumET < ETmax
      if ( sumET > ETmax ) then
        photons(i)%isPhoton = -1
        nphotons_isol = nphotons_isol - 1
        ! print*, " > apply_fixed: kill", i
        !> do not modify inside the loop: kd%nphotons = kd%nphotons - 1
        if ( nphotons_isol < min_nphotons ) then
          kd%nphotons = 0  ! for the cache
          ipass = 0
          return
        end if
      end if
    end do loop_photons

  end subroutine apply_fixed


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply the smooth cone isolation criterion
  !
  !> @note
  !> smooth cone photon isolation criterion by Frixione [arXiv:hep-ph/9801442].
  !> We implement the general algorithm that corresponds to Eq.(3.9) in that
  !> paper. Note that MCFM (./src/User/frix.f) has a much simpler
  !> implementation which however implicitly assumes that there is at most one
  !> additional unresolved emission, i.e. is only applicable at NLO.
  !
  !> @param[in] npar kinematics specified by the number of external partlicles
  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
  !-----------------------------------------------------------------------------
  subroutine apply_smooth(npar, ipass)
    use KinData_mod
    use ObsHelper_mod
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer :: i,j,k, curr_j, j_kd
    double precision :: preX, ETmax
    double precision :: curr_R, calX, sumET
    type(Kin_t), pointer :: kd

    if ( ipass == 0 ) return ! nothing to do
    if ( (curr_isol%ialgo /= 2) .and. (curr_isol%ialgo /= 3)  )  &
      & stop "apply_smooth: ialgo mismatch"

    !> prettify code
    kd => kin(npar)

    !>----- apply isolation criterion to each photon sequentially
    loop_photons: do i = 1,nproto_isol
      if ( photons(i)%isPhoton <= 0 ) cycle
      !> maximum allowed hadronic (partonic) ET inside the photon cone
      ETmax = curr_isol%ETmax_const  &
        &   + curr_isol%ETmax_epsilon * photons(i)%et
      !>--- pre-compute and store in buffer for efficiency
      do j = 1, kd%nproto
        !> index in KinData
        j_kd = kd%ijets_lower + j - 1
        arr_ETj(j) = v1_et( kd%p(:,j_kd) )
        if ( qisfrag .and. kd%ifrag  == j_kd ) arr_ETj(j) = (1d0-kd%zfrag)  * arr_ETj(j)  ! frag. linear
        if ( qisfrag .and. kd%i2frag == j_kd ) arr_ETj(j) = (1d0-kd%z2frag) * arr_ETj(j)  ! frag. linear
        arr_Rgj(j) = v2_delta_R( photons(i)%p , kd%p(:,j_kd) )  ! direction indep. of (1-z)
        if ( (arr_Rgj(j) > curr_isol%R) .or. (arr_Rgj(j) < curr_isol%R_min) ) then
          arr_ETj(j) =  0d0
          arr_Rgj(j) = -1d0  ! reject parton: cannot contribute
        end if
        ! print*, "!@ parton", j, "->" , j_kd
        ! print*, "!@ parton", "ETj", arr_ETj(j)
        ! print*, "!@ parton", "Rgj", arr_Rgj(j)
      end do  ! j <-> partons
      !>--- the actual rejection
      !> the accumulator for the hadronic transverse energy
      sumET = 0d0
      !> if we have `n` partons there can be up to `n` "steps" in delta to consider
      do j = 1, kd%nproto
        !> find the currently smallest `Rgj` <-> sets current cone size
        curr_R = curr_isol%R
        curr_j = -1
        do k = 1, kd%nproto
          if ( arr_Rgj(k) < 0d0 ) cycle  ! invalidated (out of delta-range or already eaten up)
          if ( (arr_Rgj(k) <= curr_R) .or. (curr_j < 0) ) then
            curr_R = arr_Rgj(k)
            curr_j = k
          end if
        end do
        !> no more valid delta-steps: done
        if ( curr_j < 0 ) exit  ! j-loop
        !> accumulate the energy for the current step & invalidate
        sumET = sumET + arr_ETj(curr_j)
        arr_ETj(curr_j) =  0d0
        arr_Rgj(curr_j) = -1d0  ! invalidate
        !> check smooth-cone criterion for this step
        calX = ETmax * profile(curr_R)
        !> check `I_i`
        if ( sumET > calX ) then
          photons(i)%isPhoton = -1
          nphotons_isol = nphotons_isol - 1
          !print*, " > apply_smooth: kill", i
          !read(*,*)
          !> do not modify inside the loop: kd%nphotons = kd%nphotons - 1
          if ( nphotons_isol < min_nphotons ) then
            kd%nphotons = 0  ! for the cache
            ipass = 0
            return
          end if
          cycle loop_photons
        end if
      end do  ! j <-> parton

    end do loop_photons

  contains ! functions local to `apply_smooth`

    double precision function profile(Rval)
      double precision, intent(in) :: Rval
      double precision :: Rcone, x
      !> if in hybrid mode, the profile is simply a ONE above R_inner
      if ( curr_isol%ialgo == 3 ) then
        if ( Rval > curr_isol%R_inner ) then
          profile = 1d0
          return
        end if
        !> Rval < R_inner => set the dynamical cone to R_inner
        Rcone = curr_isol%R_inner
      else
        Rcone = curr_isol%R  ! normal smooth isolation
      end if
      !> check validity
      if ( (Rval>Rcone) .or.(Rval<0d0) ) then
        print*, "apply_smooth: invalid input to profile function:", Rval, Rcone
        stop
      end if
      !> dispatch to different profile functions
      x = Rval / Rcone  ! convenience  variable in [0,1]
      select case (curr_isol%iopt)
      case (0)  ! smooth
        profile = ( (1d0-cos(Rval)) / (1d0-cos(Rcone)) )**curr_isol%n
      case (1)  ! smooth_had
        profile = ( x**2 )**(curr_isol%n)
      case (2)  ! smoother
        select case (int(curr_isol%n))
        case (1)
          profile = x * x * (3d0 - 2d0 * x)  ! smoother
        case (2)
          profile = x * x * x * (x * (x * 6d0 - 15d0) + 10d0)  ! even smoother
        case default
          print*, "apply_smooth: smoother isolation not inplemented for n =", int(curr_isol%n)
        end select
      case default
        print*, "getName_isol: invalid smooth-cone option", curr_isol%iopt
        stop
      end select

    end function profile

  end subroutine apply_smooth




!!!!! TODO:  !-----------------------------------------------------------------------------
!!!!! TODO:  !> @brief
!!!!! TODO:  !> apply soft-drop isolation criterion from arXiv:1805.11622
!!!!! TODO:  !> In principle, this routine can be applied to processes without photons to recluster
!!!!! TODO:  !> them using soft-drop, since the isolation/veto part of the subroutine is skipped for
!!!!! TODO:  !> nphotons = 0.
!!!!! TODO:  !> To do this, the check forbidding isolation-initialisation for photon-free processes
!!!!! TODO:  !> must be overridden.
!!!!! TODO:  !
!!!!! TODO:  !> @param[in] npar kinematics specified by the number of external particles
!!!!! TODO:  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
!!!!! TODO:  !-----------------------------------------------------------------------------
!!!!! TODO:  subroutine apply_softdrop(npar, ipass)
!!!!! TODO:    use KinData_mod
!!!!! TODO:    use EvalObs_mod
!!!!! TODO:    use Softdrop_mod
!!!!! TODO:    integer, intent(in) :: npar
!!!!! TODO:    integer, intent(inout) :: ipass
!!!!! TODO:    integer :: i,j, iphoton
!!!!! TODO:
!!!!! TODO:    if ( (ipass == 0) .or. (ialgo_isol == 0) ) return ! nothing to do
!!!!! TODO:    if (ialgo_isol /= 4) stop "apply_softdrop: ialgo mismatch"
!!!!! TODO:    if ( allocated(clusterHist_phot) ) deallocate(clusterHist_phot)
!!!!! TODO:    if ( allocated(clusterHist_orig) ) deallocate(clusterHist_orig)
!!!!! TODO:
!!!!! TODO:    call initRecomb_jet(npar)
!!!!! TODO:    clusterHist_jet = -1
!!!!! TODO:    call cluster_jet()
!!!!! TODO:    clusterHist_orig = clusterHist_jet
!!!!! TODO:    njets_orig = njets_jet
!!!!! TODO:
!!!!! TODO:    call recluster_jet()
!!!!! TODO:    clusterHist_phot = clusterHist_jet
!!!!! TODO:
!!!!! TODO:    ! perform veto
!!!!! TODO:    ! could separate out case in which 2 photons are clustered together
!!!!! TODO:    do i=1,nphotons
!!!!! TODO:       iphoton = iphotons(i)
!!!!! TODO:       if (iphoton < 0) then
!!!!! TODO:          ! photons are the last entries in the clusterHist_phot array.
!!!!! TODO:          ! (ordering is inherited from range kd%ijets_lower, kd%ijets_upper)
!!!!! TODO:          iphoton = nproto_jet + iphoton + 1
!!!!! TODO:       end if
!!!!! TODO:!       write(*,*) iphoton, npar
!!!!! TODO:       if (clusterHist_phot(iphoton) .eq. 0) then
!!!!! TODO:          ! photon groomed away by soft-drop (merge didn't meet SD condition)
!!!!! TODO:          ipass = 0
!!!!! TODO:!          write(*,*) "photon", iphoton, "merge failed softdrop"
!!!!! TODO:          return
!!!!! TODO:       else if (clusterHist_phot(iphoton) .ne. iphoton) then
!!!!! TODO:          ! merged into another jet ( > 0)
!!!!! TODO:          ipass = 0
!!!!! TODO:!          write(*,*) "photon", iphoton, "not isolated: was merged into jet"
!!!!! TODO:          return
!!!!! TODO:       else if (clusterHist_phot(iphoton) .eq. iphoton) then
!!!!! TODO:          do j = 1,nproto_jet
!!!!! TODO:             if (j .eq. iphoton) then
!!!!! TODO:                cycle
!!!!! TODO:             else if ( clusterHist_phot(j) .eq. iphoton ) then
!!!!! TODO:                ! photon not alone in its jet
!!!!! TODO:                ipass = 0
!!!!! TODO:!                write(*,*) "photon", iphoton, "not isolated: was merged into by protojet"
!!!!! TODO:                return
!!!!! TODO:             end if
!!!!! TODO:          end do
!!!!! TODO:       else
!!!!! TODO:          write(*,*) "Logic error in soft drop? shouldn't fail previous if clauses"
!!!!! TODO:       end if
!!!!! TODO:    enddo
!!!!! TODO:
!!!!! TODO:    ! if we reach this point, photons are isolated: clean-up
!!!!! TODO:    do i = 1,nphotons
!!!!! TODO:       iphoton = iphotons(i)
!!!!! TODO:       if (iphoton < 0) then
!!!!! TODO:          ! iphoton = npar + iphoton + 1
!!!!! TODO:          iphoton = nproto_jet + iphoton + 1
!!!!! TODO:       end if
!!!!! TODO:       clusterHist_jet(iphoton) = 0
!!!!! TODO:       jets(iphoton)%isJet = 0
!!!!! TODO:       njets_jet = njets_jet - 1
!!!!! TODO:    end do
!!!!! TODO:!    write(*,*) "clusterHist_orig", clusterHist_orig
!!!!! TODO:!    write(*,*) "clusterHist_phot", clusterHist_phot
!!!!! TODO:!    write(*,*) "clusterHist_jet", clusterHist_jet
!!!!! TODO:
!!!!! TODO:    ! could potentially define observables using these; leaving them here in case.
!!!!! TODO:    deallocate(clusterHist_orig)
!!!!! TODO:    deallocate(clusterHist_phot)
!!!!! TODO:
!!!!! TODO:  end subroutine apply_softdrop


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply democratic clusterion & isolation criterion from:
  !>    [1] E. W. N. Glover and A. G. Morgan, Z. Phys. C62 (1994) 311.
  !>    [2] E. W. N. Glover and A. G. Morgan, Phys. Lett. B334 (1994) 208.
  !
  !> @param[in] npar kinematics specified by the number of external particles
  !> @param[inout] ipass set to 0 if isolation criterion not fulfilled
  !-----------------------------------------------------------------------------
  subroutine apply_democratic(npar, ipass)
    use KinData_mod
    use ObsHelper_mod
    use Jets_mod, only : recombine_democratic_jet
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    integer :: i,j, iphoton
    integer :: nphotons

    if ( ipass == 0 ) return ! nothing to do
    if ( curr_isol%ialgo /= 5 )  &
      & stop "apply_democratic: ialgo mismatch"

    !> initialise local photon variables and arrays
    nphotons = 0
    if (.not.allocated(idxphotons_dem)) allocate(idxphotons_dem( size(photons) ))
    if (.not.allocated(pphotons_dem))   allocate(pphotons_dem(1:4, size(photons) ))
    do i = 1,nproto_isol
      if ( photons(i)%isPhoton <= 0 ) cycle
      nphotons = nphotons + 1
      idxphotons_dem(nphotons) = i
      pphotons_dem(1:4,nphotons) = photons(i)%p(1:4)
    end do

    !> apply the democratic clustering
    !>  =>  rejected photons have energy zero in pphotons_dem
    call recombine_democratic_jet(npar, nphotons, pphotons_dem, curr_isol%zcut)

    !> copy back photons that survived
    do i = 1, nphotons
      iphoton = idxphotons_dem(i)
      if ( pphotons_dem(4,i) <= 0d0 ) then
        photons(iphoton)%isPhoton = -1
        nphotons_isol = nphotons_isol - 1
      else
        photons(iphoton)%p(1:4) = pphotons_dem(1:4,i)
        photons(iphoton)%et     = v1_et( photons(iphoton)%p(1:4) )
      end if
    end do

  end subroutine apply_democratic


end module PhotonIsolation_mod

