!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Process [external members: `_proc`]
!
! DESCRIPTION:
!> Module collecting all process-specific settings
!> (run settings are stored in Run_mod)
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Process_mod
!-------------------------------------------------------------------------------
  use, intrinsic :: iso_c_binding
  ! use dlfcn_mod
  implicit none
  public

  !> process name
  character(32) :: in_name_proc  ! stripped backend name from runcard
  character(32) :: name_proc_usr  ! stripped user-friendly name from runcard
  character(32) :: name_proc  ! converted to uppercase

  !> collider defined through beams:
  !> particle id's follow the monte carlo numbering scheme:
  !> http://pdg.lbl.gov/2017/reviews/rpp2016-rev-monte-carlo-numbering.pdf
  !>   beam = +2212: proton
  !>   beam = -2212: anti-proton
  !>   beam = +11: electron
  !>   beam = -11: positron
  integer, parameter :: pid_p  =  2212
  integer, parameter :: pid_ap = -2212
  integer, parameter :: pid_em =  11
  integer, parameter :: pid_ep = -11
  integer, parameter :: pid_Pb = 1000822080
  integer, parameter :: pid_Au = 1000791970
  integer, parameter :: pid_O  = 1000080160
  integer, parameter :: pid_Cu = 1000290630
  integer, parameter :: pid_He = 1000020040
  integer, parameter :: pid_Li = 1000030060
  integer, parameter :: pid_C  = 1000060120
  integer :: A_beam1, Z_beam1, A_beam2, Z_beam2

  !> default: LHC
  integer :: beam1_proc = pid_p
  integer :: beam2_proc = pid_p
  integer :: nPDFs_proc = 2  ! is (2-nonpis)
  !> collider energy
  double precision :: sqrts_proc, s_proc, Ebeam1_proc, Ebeam2_proc

   !> inwa -- heavy particle decay type
   !>   0 : use ds12/s12
   !>   1 : use resonance parametrization (with mass rm12, width rg12)
   !>  -4 : use resonance parametrization (with mass rm12, width rg12), energy dependent width
   !>   2 : narrow width inserted: delta function
   !>   3 : on-shell heavy particle no decay
   !>   4 : differential distribution in the virtuality of the particle
  integer :: decay_type_proc = 1
  double precision  :: Q_OS_proc = 0d0  ! for type = 4, the invariant mass

  !> info on the (QCD) Born process
  !>   indices & # defining the index ranges of all final-state particle types
  !>   real-emission kinematics will be filled up with partons (& photons)
  integer :: ibegin_spectators_proc = 0,  n_fs_spectators_proc = 0
  integer :: ibegin_photons_proc    = 0,  n_fs_photons_proc    = 0
  integer :: ibegin_partons_proc    = 0,  n_fs_partons_proc    = 0
  integer :: njets_proc
  integer :: n_alphas_proc = -1, n_alpha_proc = -1
  integer :: nsub_proc = -1
  logical :: qexclusive_proc = .false.

  !> in the `npar` kinematics:
  !>
  !>    [(1 + partons_offset_lower_proc), (npar - partons_offset_upper_proc)]
  !>
  !> we therfore have `npar - partons_offset_lower_proc - partons_offset_upper_proc`
  !> proto-jets in any given `npar` kinematics
  !> and `nspec` (# colour-neutral particles) can be derived as
  !>    nspec =  partons_offset_lower_proc - 2 + partons_offset_upper_proc
  !>
  !> Z+J:
  !>    partons_offset_lower_proc = 2 (incoming particles)
  !>    partons_offset_upper_proc = 2 (last two momenta always leptons)
  !> DIS:
  !>    partons_offset_lower_proc = 3 (2 incoming particles + first lepton in `3`)
  !>    partons_offset_upper_proc = 0 (no trailing colour-neutral particles)
  integer :: partons_offset_lower_proc, partons_offset_upper_proc


  !/////////////////////////////////
  !//  Process-specific settings  //
  !/////////////////////////////////

  !> boson flag
  !>  -1 = W-
  !>   0 = Z/H
  !>  +1 = W+
  integer :: iboson_proc = 0

  !> neutral-current gauge-boson processes (Z, DIS, G, VH)
  !>   0 = Z & gamma
  !>   1 = only gamma  (for photon-induced processes & DIS)
  !>   2 = only Z  (for ZH amplitudes)
  integer :: igamma_proc = 0
  !>   0 = neutrino <--> observables still use lepton names
  !>   1 = charged lepton (e, mu)
  integer :: ilepton_proc = 1

  !> # extra dimensions or Vegas (needed, e.g., for qt sampling)
  integer :: ndim_extra_proc = 0


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the process module using default values
  !-----------------------------------------------------------------------------
  subroutine init_proc(process)
    use StrHelper_mod
    character(len=*), intent(in) :: process

    !> store user-friendly process name
    name_proc_usr =  trim(adjustl(process))

    !> convert user-friendly name to conventional proc name
    !> Naming convention for users:
    !> * JJ    -> 2jetfc 
    !> * eeJJ  -> EPEMJJ
    !> * eeJJJ -> EPEM
    !> * epL*  -> DIS*
    !> * epLJJ -> DIS
    !> * epN*  -> DISWm*
    !> * epNb* -> DISWp*
    !> * Default: in_name_proc = name_proc_usr

    !> WF: EPEM would coincide with EP processes. Add an if statement to check for EPEM

    !> If using development names already, drop the conversion
    if ( to_upper_str(name_proc_usr(:4)) .eq. "EPEM" .or.  &
       & to_upper_str(name_proc_usr(:3)) .eq. "DIS" ) then
      in_name_proc = name_proc_usr
    else
      select case (to_upper_str(name_proc_usr(:2)))
      case ("EE")
        !> Address the prefix 'ee' and convert them to conventional names
        select case (to_upper_str(name_proc_usr(3:)))
        case ("JJ")
          in_name_proc = trim("EPEM")//trim(to_upper_str(name_proc_usr(3:)))
        case ("JJJ")
          in_name_proc = "EPEM"
        case ("JJ_H")
          in_name_proc = "EPEMJJ_H"
        case DEFAULT
          print *, "init_proc: unknown process name for ee collision: ", name_proc_usr
          stop
        end select

      case ("EP")
        !> Address the prefix 'ep' and convert them to conventional names
        !> TODO: also include positron-proton for neutral current processes
        select case (to_upper_str(name_proc_usr(3:)))
        case ("LJ")
          in_name_proc = "DIS1j"
        case ("NJ")
          in_name_proc = "DISWm1j"
        case ("NBJ")
          in_name_proc = "DISWp1j"
        case ("LJJ")
          in_name_proc = "DIS"
        case ("NJJ")
          in_name_proc = "DISWm"
        case ("NBJJ")
          in_name_proc = "DISWp"
        case default
          print *, "init_proc: unknown process name for ep collision: ", name_proc_usr
          stop
        end select

      case ("JJ")
        !> Address processes starting with "JJ" at LHC, usually having exotic conventional names
        select case (to_upper_str(name_proc_usr))
        case ("JJ")
          in_name_proc = "2jetfc"
          ! case ("JJJ")
          !   in_name_proc = "3jet"
        case default
          print *, "init_proc: unknown process name for jet production: ", name_proc_usr
          stop
        end select
      case default
        in_name_proc = name_proc_usr
      end select
    end if

    !> convert to upper case for uniqueness
    name_proc = to_upper_str(in_name_proc)

    !> default init for the collider
    select case (name_proc)
      case ( "DIS1J", "SFDIS", "DIS",  &
           & "DISWM1J", "SFDISWM", "DISWM" )
        beam1_proc = pid_p
        beam2_proc = pid_em
        nPDFs_proc = 1
      case ( "DISWP1J", "SFDISWP", "DISWP" )
        beam1_proc = pid_p
        beam2_proc = pid_ep
        nPDFs_proc = 1
      case ( "EPEM", "EPEM_H", "EPEMJJ", "EPEMJJ_H", "EPEMINC" )
        beam1_proc = pid_ep
        beam2_proc = pid_em
        nPDFs_proc = 0
      case default
        beam1_proc = pid_p
        beam2_proc = pid_p
        nPDFs_proc = 2
    end select

    !> @todo: set default `decay_type_proc` values for the processes
    decay_type_proc = 1
    select case (name_proc)
      case ( "GG", "GG_FRAG", "GG_EW", "GJ", "GJ_FRAG", "G_EW", "GGG","GJJ", "GJJJ", "GGJ", "GGJFC", "GGJJ", &
           & "GGJJFC","GGJJJFC", "GGQT", "DIS1J", "DIS", "SFDIS",  &
           & "DISG", "GJ_H", "HH", "HHJ", "HHJJ", "HHJJJ" )
        decay_type_proc = 0
      case ( "H", "HJ", "HJJ", "HJJJ",  &
           & "HM", "HJM", "HJMREW", "HJJM", "HJJMREW", "HJJJM", &
           & "HTO2P", "HTO2PJ", "HTO2PJJ", "HTO2PJJJ",  &
           & "HTO2PM", "HTO2PJM",  &
           & "HTO2TAU", "HTO2TAUJ", "HTO2TAUJJ", "HTO2TAUJJJ",  &
           & "HTO2TAUM", "HTO2TAUJM",  &
           & "HINC", "HTO2PINC", "HTO2TAUINC",  &
           & "HQT", "HTO2PQT", "HTO2TAUQT")
        decay_type_proc = 3
      case ( "WMH", "WPH", "WMHJ", "WPHJ", "ZH", "ZHJ" )
        decay_type_proc = -20
      case default
        igamma_proc = 0  ! Z & gamma
    end select

    !> default init for igamma switch
    select case (name_proc)
      case ( "GG", "GG_FRAG", "GG_EW", "GJ", "GJ_FRAG", "G_EW", "GGG","GJJ", "GJJJ", "GGJ", "GGJFC", "GGJJ", &
           & "GGJJFC","GGJJJFC", "GGQT", "DIS1J", "DIS", "SFDIS",  &
           & "DISG", "GJ_H" )
        igamma_proc = 1  ! only gamma
      case default
        igamma_proc = 0  ! Z & gamma
    end select
    iboson_proc = 0

    !> default init
    sqrts_proc  = 0d0
    s_proc      = 0d0
    Ebeam1_proc = 0d0
    Ebeam2_proc = 0d0

    !> extra dimensions for Vegas
    select case (name_proc)
      case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
           & "HTO2L1PQT",  &
           & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT",  &
           & "ZQT", "WPQT", "WMQT", "GGQT" )
        ndim_extra_proc = 2  ! qT & phi sampling
      case default
        ndim_extra_proc = 0  ! no extra dimensions by default
    end select

  end subroutine init_proc
  !>>>>>> after init_proc, we'll read in the setting (also process dependent ones)
  !>>>>>> THEN we call init_param
  !>>>>>> THEN we call setup_proc (where igamma is REMOVED)


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the initial-state beam types
  !
  !> @param[in] beam1 string for the 1st beam (+z direction)
  !> @param[in] beam2 string for the 2nd beam (-z direction)
  !-----------------------------------------------------------------------------
  subroutine setBeams_proc(beam1, beam2)
    character(len=*), intent(in) :: beam1, beam2

    call set_beam_properties(1, beam1)
    call set_beam_properties(2, beam2)

  contains

     subroutine set_beam_properties(idx_, beam_)
      use StrHelper_mod
      integer, intent(in) :: idx_
      character(len=*), intent(in) :: beam_
      integer :: pid_, A_, Z_
      select case (to_upper_str(beam_))
        case ('P')
          pid_ = pid_p
          A_ = 1
          Z_ = 1
        case ('AP')
          pid_ = pid_ap
          A_ = 1
          Z_ = 1
        case ('EM')
          pid_ = pid_em
          A_ = 0
          Z_ = 1
        case ('EP')
          pid_ = pid_ep
          A_ = 0
          Z_ = 1
        case ('PB')
          pid_ = pid_Pb
          A_ = 208
          Z_ = 82
        case ('AU')
          pid_ = pid_Au
          A_ = 197
          Z_ = 79
        case ('O')
          pid_ = pid_O
          A_ = 16
          Z_ = 8
        case ('CU')
          pid_ = pid_Cu
          A_ = 64
          Z_ = 29
        case ('HE')
          pid_ = pid_He
          A_ = 4
          Z_ = 2
        case ('LI')
          pid_ = pid_Li
          A_ = 7
          Z_ = 3
        case ('C')
          pid_ = pid_C
          A_ = 12
          Z_ = 6
        case default
          print*, "setBeams_proc: unknown beam specifier:", beam_
          stop
      end select
      select case (idx_)
      case (1)
        beam1_proc = pid_
        A_beam1 = A_
        Z_beam1 = Z_
      case (2)
        beam2_proc = pid_
        A_beam2 = A_
        Z_beam2 = Z_
      case default
        print*, "setBeams_proc: invalid index:", idx_
        stop
      end select
    end subroutine set_beam_properties

  end subroutine setBeams_proc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the collider energy by the centre-of-mass energy (lab=cms fram)
  !
  !> @param[in] sqrts centre-of-mass energy of the collider
  !-----------------------------------------------------------------------------
  subroutine setSqrts_proc(sqrts)
    double precision, intent(in) :: sqrts
    sqrts_proc  = sqrts
    s_proc      = sqrts**2
    Ebeam1_proc = sqrts / 2d0
    Ebeam2_proc = sqrts / 2d0
  end subroutine setSqrts_proc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the collider energy by the beam energies (lab can be != cms frame)
  !
  !> @param[in] E1 energy of beam1 (+z direction)
  !> @param[in] E2 energy of beam2 (-z direction)
  !-----------------------------------------------------------------------------
  subroutine setEnergies_proc(E1, E2)
    double precision, intent(in) :: E1, E2
    if ( (abs(beam1_proc) /= abs(beam2_proc)) .and.  &
       & ( A_beam1 > 1 .or. A_beam2 > 1 ) ) then
      print*, "setEnergies_proc: asymmetric beams not supported for nucleons!"
      !> @todo: might need to be careful about DIS with nucleon target...
      stop
    end if
    Ebeam1_proc = E1
    Ebeam2_proc = E2
    s_proc      = 4d0*E1*E2
    sqrts_proc  = sqrt(s_proc)
  end subroutine setEnergies_proc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set up the process (needs to be called *after* parameters are set)
  !> + propagate information to the various common blocks
  !-----------------------------------------------------------------------------
  subroutine setup_proc()
    !> common blocks to set process-specific varibales
    double precision :: rm12, rg12
    integer :: inwa
    common /nwa/ rm12, rg12, inwa
    double precision :: emz, ezwidth
    common /Zmass/  emz, ezwidth
    double precision :: emw, ewwidth
    common /Wmass/  emw, ewwidth
    double precision :: emh, ehwidth
    common /Hmass/  emh, ehwidth
    double precision :: rmv1, rgv1, rmv2, rgv2
    integer :: ios1, ios2, isym
    common /vdecay/ rmv1, rgv1, rmv2, rgv2, ios1, ios2, isym
    !> for DIS swap
    integer :: itmp
    double precision :: rtmp

    !> for the moment, keep the common block
    integer :: npars,nspec,n
    common /npars/ npars,nspec,n

    !> initialise process-specific varibales
    n_fs_spectators_proc = 0  !  == not (partons || photons)
    n_fs_photons_proc    = 0
    n_fs_partons_proc    = 0
    njets_proc           = 0
    !> default: all final-state particles are proto-jets
    ibegin_spectators_proc = 0
    ibegin_photons_proc    = 0
    ibegin_partons_proc    = 3  ! 2 IS particles


    select case (name_proc)

    !> JET
    case ("1JET", "2JET", "1JETFC", "2JETFC", "3JET", "4JET", "5JET")
      !> this class better have decay_type_proc = 0
      if (decay_type_proc /= 0) then
        print*, "WARNING: resetting decay_type_proc = 0"
        decay_type_proc = 0
      end if
      select case (name_proc)
      case ("1JET")
        n_fs_partons_proc = 2
        njets_proc        = 1
      case ("2JET")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ("1JETFC")
        n_fs_partons_proc = 2
        njets_proc        = 1
      case ("2JETFC")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ("3JET")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case ("4JET")
        n_fs_partons_proc = 4
        njets_proc        = 4
      case ("5JET")
        n_fs_partons_proc = 5
        njets_proc        = 5
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> GJ
    case ("GJ", "GJ_FRAG", "G_EW", "GJJ", "GJJJ", "GJ_H")
      !> this class better have decay_type_proc = 0
      if (decay_type_proc /= 0) then
        print*, "WARNING: resetting decay_type_proc = 0"
        decay_type_proc = 0
      end if
      !> initialise helicity orderings for jet production MEs
      call inithel()
      iboson_proc = 0
      rm12 = 0d0
      rg12 = 0d0
      ibegin_photons_proc       = -1  ! last entry: photon
      n_fs_photons_proc         =  1
      select case (name_proc)
      case ("GJ", "GJ_FRAG", "G_EW")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ("GJ_H")
        n_fs_partons_proc = 1
        njets_proc        = 0
      case ("GJJ")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ("GJJJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> GG
    case ("GG", "GG_FRAG", "GG_EW", "GGJ", "GGJFC", "GGJJ", "GGJJFC","GGJJJFC", "GGQT")
      !> this class better have decay_type_proc = 0 (linear sample for general) or 10 (log sample for colourless final state)
      if ((decay_type_proc /= 0).and.(decay_type_proc /= 10)) then
        print*, "WARNING: resetting decay_type_proc = 0"
        decay_type_proc = 0
      end if
      !> initialise helicity orderings for jet production MEs
      call inithel()
      iboson_proc = 0
      ibegin_photons_proc       = -1  ! last two entries: photons
      n_fs_photons_proc         =  2
      select case (name_proc)
      case("GG", "GG_FRAG", "GG_EW", "GGQT")
        n_fs_partons_proc = 0
        njets_proc        = 0
      case("GGJ")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case("GGJFC")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case("GGJJ")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("GGJJFC")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("GGJJJFC")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> GGG
    case ("GGG")
      !> this class better have decay_type_proc = 0
      if (decay_type_proc /= 0) then
        print*, "WARNING: resetting decay_type_proc = 0"
        decay_type_proc = 0
      end if
      !> initialise helicity orderings for jet production MEs
      call inithel()
      iboson_proc = 0
      ibegin_photons_proc       = -1  ! last two entries: photons
      n_fs_photons_proc         =  3
      select case (name_proc)
      case("GGG")
        n_fs_partons_proc = 0
        njets_proc        = 0
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> Z
    case ("Z", "ZJ", "ZJ_H", "ZJJ", "ZJJJ", "ZQT", "Z_EW")
      iboson_proc = 0
      rm12 = emz
      rg12 = ezwidth
      ibegin_spectators_proc    = -1  ! last two entries: leptons
      n_fs_spectators_proc      =  2
      ibegin_photons_proc       = -3  ! photons before the leptons (EW corr)
      n_fs_photons_proc         =  0
      select case (name_proc)
      case("Z", "ZQT", "Z_EW")
        n_fs_partons_proc = 0
        njets_proc        = 0
      case("ZJ", "ZJ_H")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case("ZJJ")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("ZJJJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> Wm
    case ("WM", "WMJ", "WMJUNSYM", "WMJ_H", "WMJJ", "WMJJJ", "WMQT")
      iboson_proc = -1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last two entries: leptons
      n_fs_spectators_proc      =  2
      select case (name_proc)
      case("WM", "WMQT")
        n_fs_partons_proc = 0
        njets_proc        = 0
      case("WMJ", "WMJUNSYM", "WMJ_H")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case("WMJJ")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("WMJJJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> Wp
    case ("WP", "WPJ", "WPJUNSYM", "WPJ_H", "WPJJ", "WPJJJ", "WPQT")
      iboson_proc = +1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last two entries: leptons
      n_fs_spectators_proc      =  2
      select case (name_proc)
      case("WP", "WPQT")
        n_fs_partons_proc = 0
        njets_proc        = 0
      case("WPJ", "WPJUNSYM", "WPJ_H")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case("WPJJ")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("WPJJJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> VFH & VFH2
    case ( "VFH", "VFHFULL", "VFHJ",  &
         & "VFHTO2P", "VFHTO2PJ",   &
         & "VFHTO2TAU", "VFHTO2TAUJ" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      ibegin_spectators_proc    = -1  ! last two entries: Higgs
      n_fs_spectators_proc      =  2
      select case (name_proc)
      case("VFH", "VFHFULL", "VFHTO2P", "VFHTO2TAU")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("VFHJ", "VFHTO2PJ", "VFHTO2TAUJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> VFH3
    case ( "VFHTO2L1P", "VFHTO2L1PJ" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emz
      rgv1 = ezwidth
      rmv2 = 0d0
      rgv2 = 0d0
      ios1 = 1
      !--
      ibegin_photons_proc       = -1  ! last entry: photon
      n_fs_photons_proc         =  1
      ibegin_spectators_proc    = -2  ! followed by two leptons
      n_fs_spectators_proc      =  2
      select case (name_proc)
      case("VFHTO2L1P")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("VFHTO2L1PJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> VFH4
    case ( "VFHTO2E2MU", "VFHTO2E2MUJ", "VFHTO4E", "VFHTO4EJ" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emz
      rgv1 = ezwidth
      rmv2 = emz
      rgv2 = ezwidth
      ios1 = -1
      ios2 = 1
      isym = 0
      !--
      ibegin_spectators_proc    = -1  ! last four entries: leptons
      n_fs_spectators_proc      =  4
      select case (name_proc)
      case("VFHTO2E2MU", "VFHTO4E")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case("VFHTO2E2MUJ", "VFHTO4EJ")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> H & H2
    case ( "H", "HJ", "HJJ", "HJJJ",  &
         & "HM", "HJM", "HJMREW", "HJJM", "HJJMREW", "HJJJM", &
         & "HTO2P", "HTO2PJ", "HTO2PJJ", "HTO2PJJJ",  &
         & "HTO2PM", "HTO2PJM",  &
         & "HTO2TAU", "HTO2TAUJ", "HTO2TAUJJ", "HTO2TAUJJJ",  &
         & "HTO2TAUM", "HTO2TAUJM" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      if ( name_proc(1:5) == "HTO2P" ) then
        ibegin_photons_proc       = -1  ! last two entries: photons
        n_fs_photons_proc         =  2
      else
        ibegin_spectators_proc    = -1  ! last two entries: Higgs [+decay]
        n_fs_spectators_proc      =  2
      end if
      select case (name_proc)
      case ( "H",  &
           & "HM",  &
           & "HTO2P",  &
           & "HTO2PM",  &
           & "HTO2TAU",  &
           & "HTO2TAUM" )
        n_fs_partons_proc = 0
        njets_proc        = 0
      case ( "HJ",  &
           & "HTO2PJ",  &
           & "HTO2PJM",  &
           & "HTO2TAUJ",  &
           & "HTO2TAUJM" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HJM","HJMREW")
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HJJ",  &
           & "HTO2PJJ",  &
           & "HTO2TAUJJ" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HJJM","HJJMREW" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HJJJM" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case ( "HJJJ",  &
           & "HTO2PJJJ",  &
           & "HTO2TAUJJJ" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> H3
    case ( "HTO2L1P", "HTO2L1PJ", "HTO2L1PJJ", "HTO2L1PJJJ",  &
         & "HTO2L1PM", "HTO2L1PJM" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emz
      rgv1 = ezwidth
      rmv2 = 0d0
      rgv2 = 0d0
      ios1 = 1
      !--
      ibegin_photons_proc       = -1  ! last entry: photon
      n_fs_photons_proc         =  1
      ibegin_spectators_proc    = -2  ! followed by two leptons
      n_fs_spectators_proc      =  2
      select case (name_proc)
      case ( "HTO2L1P",  &
           & "HTO2L1PM" )
        n_fs_partons_proc = 0
        njets_proc        = 0
      case ( "HTO2L1PJ",  &
           & "HTO2L1PJM" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HTO2L1PJJ" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HTO2L1PJJJ" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> H -> ZZ
    case ( "HTO2E2MU", "HTO2E2MUJ", "HTO2E2MUJJ", "HTO2E2MUJJJ",  &
         & "HTO4E", "HTO4EJ", "HTO4EJJ", "HTO4EJJJ",  &
         & "HTO2E2MUM", "HTO2E2MUJM",  &
         & "HTO4EM", "HTO4EJM" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emz
      rgv1 = ezwidth
      rmv2 = emz
      rgv2 = ezwidth
      ios1 = -1
      ios2 = 1
      isym = 0
      !--
      ibegin_spectators_proc    = -1  ! last four entries: leptons
      n_fs_spectators_proc      =  4
      select case (name_proc)
      case ( "HTO2E2MU",  &
           & "HTO2E2MUM",  &
           & "HTO4E",  &
           & "HTO4EM" )
        n_fs_partons_proc = 0
        njets_proc        = 0
      case ( "HTO2E2MUJ",  &
           & "HTO2E2MUJM",  &
           & "HTO4EJ",  &
           & "HTO4EJM" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HTO2E2MUJJ",  &
           & "HTO4EJJ" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HTO2E2MUJJJ",  &
           & "HTO4EJJJ" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> H -> WW  (& ZZ?)
    case ( "HTO2L2N", "HTO2L2NJ", "HTO2L2NJJ", "HTO2L2NJJJ",  &
         & "HTO2L2NM", "HTO2L2NJM" )
      print*, "setup_proc: WARNING Hto2l2n only contains the H->WW contribution!"
      ! stop
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emw
      rgv1 = ewwidth
      rmv2 = emw
      rgv2 = ewwidth
      ios1 = -1
      ios2 = 1
      isym = 0
      !--
      ibegin_spectators_proc    = -1  ! last four entries: leptons
      n_fs_spectators_proc      =  4
      select case (name_proc)
      case ( "HTO2L2N",  &
           & "HTO2L2NM" )
        n_fs_partons_proc = 0
        njets_proc        = 0
      case ( "HTO2L2NJ",  &
           & "HTO2L2NJM" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HTO2L2NJJ" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HTO2L2NJJJ" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> Hinc
    case ( "HINC", "HTO2PINC", "HTO2TAUINC",  &
         & "HTO2L1PINC",  &
         & "HTO2E2MUINC", "HTO4EINC", "HTO2L2NINC" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      n_fs_partons_proc = 0
      njets_proc        = 0

      !> do some checks here
      ! only imethod=2
      ! only decay_type_proc = 3

      select case (name_proc)
      case ( "HINC",  &
           & "HTO2TAUINC" )
        ibegin_spectators_proc    = -1  ! last two entries: Higgs [+decay]
        n_fs_spectators_proc      =  2
      case ( "HTO2PINC" )
        ibegin_photons_proc       = -1  ! last two entries: photons
        n_fs_photons_proc         =  2
      case ( "HTO2L1PINC" )
        ibegin_photons_proc       = -1  ! last entry: photon
        n_fs_photons_proc         =  1
        ibegin_spectators_proc    = -2  ! followed by two leptons
        n_fs_spectators_proc      =  2
        !--
        rmv1 = emz
        rgv1 = ezwidth
        rmv2 = 0d0
        rgv2 = 0d0
        ios1 = 1
        !--
      case ( "HTO2E2MUINC", "HTO4EINC" )
        ibegin_spectators_proc    = -1  ! last four entries: leptons
        n_fs_spectators_proc      =  4
        !--
        rmv1 = emz
        rgv1 = ezwidth
        rmv2 = emz
        rgv2 = ezwidth
        ios1 = -1
        ios2 = 1
        isym = 0
        !--
      case ( "HTO2L2NINC" )
        print*, "setup_proc: Hto2l2ninc should also have H->ZZ contributions?!"
        stop
        ibegin_spectators_proc    = -1  ! last four entries: leptons
        n_fs_spectators_proc      =  4
        !--
        rmv1 = emw
        rgv1 = ewwidth
        rmv2 = emw
        rgv2 = ewwidth
        ios1 = -1
        ios2 = 1
        isym = 0
        !--
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select


    !> Hqt
    case ( "HQT", "HTO2PQT", "HTO2TAUQT",  &
         & "HTO2L1PQT",  &
         & "HTO2E2MUQT", "HTO4EQT", "HTO2L2NQT" )
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      n_fs_partons_proc = 0
      njets_proc        = 0

      !> do some checks here
      ! only imethod=2
      ! only decay_type_proc = 3

      select case (name_proc)
      case ( "HQT",  &
           & "HTO2TAUQT" )
        ibegin_spectators_proc    = -1  ! last two entries: Higgs [+decay]
        n_fs_spectators_proc      =  2
      case ( "HTO2PQT" )
        ibegin_photons_proc       = -1  ! last two entries: photons
        n_fs_photons_proc         =  2
      case ( "HTO2L1PQT" )
        ibegin_photons_proc       = -1  ! last entry: photon
        n_fs_photons_proc         =  1
        ibegin_spectators_proc    = -2  ! followed by two leptons
        n_fs_spectators_proc      =  2
        !--
        rmv1 = emz
        rgv1 = ezwidth
        rmv2 = 0d0
        rgv2 = 0d0
        ios1 = 1
        !--
      case ( "HTO2E2MUQT", "HTO4EQT" )
        ibegin_spectators_proc    = -1  ! last four entries: leptons
        n_fs_spectators_proc      =  4
        !--
        rmv1 = emz
        rgv1 = ezwidth
        rmv2 = emz
        rgv2 = ezwidth
        ios1 = -1
        ios2 = 1
        isym = 0
        !--
      case ( "HTO2L2NQT" )
        print*, "setup_proc: Hto2l2nqt should also have H->ZZ contributions?!"
        stop
        ibegin_spectators_proc    = -1  ! last four entries: leptons
        n_fs_spectators_proc      =  4
        !--
        rmv1 = emw
        rgv1 = ewwidth
        rmv2 = emw
        rgv2 = ewwidth
        ios1 = -1
        ios2 = 1
        isym = 0
        !--
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select


    !> HH
    case ( "HH","HHJ","HHJJ",'HHJJJ')
      iboson_proc = 0
      rm12 = emh
      rg12 = ehwidth
      !--
      rmv1 = emh
      rgv1 = ehwidth
      rmv2 = emh
      rgv2 = ehwidth
      ios1 = 3
      ios2 = 3
      isym = 0
      !--
      ibegin_spectators_proc    = -1  ! last four entries: two Higgs[decay]
      n_fs_spectators_proc      =  4
      !--
      !> fix up to be 4 eventually
      select case (name_proc)
      case ( "HH" )
        n_fs_partons_proc = 0
        njets_proc        = 0
      case ( "HHJ" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "HHJJ" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case ( "HHJJJ" )
        n_fs_partons_proc = 3
        njets_proc        = 3
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select


    !> DIS
    case ( "DIS1J", "DIS", "SFDIS" )
      !> this class better have decay_type_proc = 0 if only photon-exchange
      if ( (decay_type_proc/=0) .and. (igamma_proc==1) ) then
        print*, "WARNING: resetting decay_type_proc = 0"
        decay_type_proc = 0
      end if
      iboson_proc = 0
      !> mass & width are not used if igamma_proc=1
      rm12 = emz
      rg12 = ezwidth
      ibegin_spectators_proc    = 3  ! FS lepton @ 3
      n_fs_spectators_proc      = 1
      ibegin_partons_proc       = 4
      select case (name_proc)
      case ( "DIS1J",  &
           & "SFDIS" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "DIS" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> DISWm
    case ( "DISWM1J", "SFDISWM", "DISWM" )
      iboson_proc = +1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = 3  ! FS lepton @ 3
      n_fs_spectators_proc      = 1
      ibegin_partons_proc       = 4
      select case (name_proc)
      case ( "DISWM1J",  &
           & "SFDISWM" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "DISWM" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> DISWp
    case ( "DISWP1J", "SFDISWP", "DISWP" )
      iboson_proc = -1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = 3  ! FS lepton @ 3
      n_fs_spectators_proc      = 1
      ibegin_partons_proc       = 4
      select case (name_proc)
      case ( "DISWP1J",  &
           & "SFDISWP" )
        n_fs_partons_proc = 1
        njets_proc        = 1
      case ( "DISWP" )
        n_fs_partons_proc = 2
        njets_proc        = 2
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> DISG
    case ( "DISG" )
      print*, "setup_proc: please double-check DISG case ;)"
      stop
      !> initialise helicity orderings for jet production MEs
      call inithel()
      iboson_proc = 0
      rm12 = 0d0
      rg12 = 0d0
      ibegin_spectators_proc    = 3  ! FS lepton @ 3
      n_fs_spectators_proc      = 1
      ibegin_partons_proc       = 4
      select case (name_proc)
      case ("DISG")
        n_fs_partons_proc = 2
        njets_proc        = 2
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> epem
    case ( "EPEM", "EPEM_H", "EPEMJJ", "EPEMJJ_H", "EPEMINC" )
      iboson_proc = 0
      !> mass & width are not used if igamma_proc=1
      rm12 = emz
      rg12 = ezwidth
      select case (name_proc)
      case ("EPEM", "EPEM_H")
        n_fs_partons_proc = 3
        njets_proc        = 3
      case ("EPEMJJ", "EPEMJJ_H", "EPEMINC")
        n_fs_partons_proc = 2
        njets_proc        = 2        
      case default
        print*, "setup_proc: unknown process ", name_proc, " in group"
        stop
      end select

    !> WMH
    case ("WMH")
      iboson_proc = -1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last four entries: leptons and Higgs
      n_fs_spectators_proc      =  4
      n_fs_partons_proc = 0
      njets_proc        = 0

      rmv1 = emh
      rgv1 = ehwidth
      rmv2 = emw
      rgv2 = ewwidth
      ios1 = 3
      ios2 = 1
      isym = 0

    !> WPH
    case ("WPH")
      iboson_proc = 1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last four entries: leptons and Higgs
      n_fs_spectators_proc      =  4
      n_fs_partons_proc = 0
      njets_proc        = 0

      rmv1 = emh
      rgv1 = ehwidth
      rmv2 = emw
      rgv2 = ewwidth
      ios1 = 3
      ios2 = 1
      isym = 0

    !> WMHJ
    case ("WMHJ")
      iboson_proc = -1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last four entries: leptons and Higgs
      n_fs_spectators_proc      =  4
      n_fs_partons_proc = 1
      njets_proc        = 1

      rmv1 = emh
      rgv1 = ehwidth
      rmv2 = emw
      rgv2 = ewwidth
      ios1 = 3
      ios2 = 1
      isym = 0

    !> WPHJ
    case ("WPHJ")
      iboson_proc = 1
      rm12 = emw
      rg12 = ewwidth
      ibegin_spectators_proc    = -1  ! last four entries: leptons and Higgs
      n_fs_spectators_proc      =  4
      n_fs_partons_proc = 1
      njets_proc        = 1

      rmv1 = emh
      rgv1 = ehwidth
      rmv2 = emw
      rgv2 = ewwidth
      ios1 = 3
      ios2 = 1
      isym = 0


    !> error
    case default
      print*, "setup_proc: unknown process ", name_proc
      stop

    end select


    !> some consistency checks and process-specific setups
    if (sqrts_proc <= 0d0) stop "setup_proc: forgot to set collider?!"
    select case (name_proc)
      case ( "DIS1J", "DIS", "SFDIS",  &
           & "DISWM1J", "SFDISWM", "DISWM",  &
           & "DISWP1J", "SFDISWP", "DISWP",  &
           & "DISG" )
        if ( (beam1_proc == pid_em .or. beam1_proc == pid_ep) .and.  &
           & (beam2_proc == pid_p  .or. beam2_proc == pid_ap) ) then
          !> we need to swap it to our DIS ordering:
          !> beam1 = hadron
          !> beam2 = lepton
          itmp        = beam1_proc
          beam1_proc  = beam2_proc
          beam2_proc  = itmp
          rtmp        = Ebeam1_proc
          Ebeam1_proc = Ebeam2_proc
          Ebeam2_proc = rtmp
        end if
        if ( (beam1_proc /= pid_p  .and. beam1_proc /= pid_ap) .or.  &
           & (beam2_proc /= pid_em .and. beam2_proc /= pid_ep) .or.  &
           & (nPDFs_proc /= 1)  &
           & ) stop "setup_proc: invalid state for DIS"
        !> call legacy init routine for DIS (now member routine)
        !> order swapped because DIS routine wants (Ee, Ep)
        call setCollider_dis(Ebeam2_proc, Ebeam1_proc)
      case ( "EPEM", "EPEM_H", "EPEMJJ", "EPEMJJ_H", "EPEMINC" )
        if ( (beam1_proc /= pid_ep) .or.  &
           & (beam2_proc /= pid_em) .or.  &
           & (nPDFs_proc /= 0)  &
           & ) stop "setup_proc: invalid state for epem"
        !> call legacy init routine for epem (now member routine)
        call setCollider_epem(Ebeam1_proc, Ebeam2_proc)
      case default
        if ( beam1_proc == pid_em .or. beam1_proc == pid_ep .or.  &
           & beam2_proc == pid_em .or. beam2_proc == pid_ep .or.  &
           & nPDFs_proc /= 2  )  &
        &  stop "setup_proc: invalid state for hadrons"
    end select

    if ( (n_fs_photons_proc/=0) .and. (ibegin_photons_proc==0) ) then
      print*, "setup_proc: setting for photons ", ibegin_photons_proc, n_fs_photons_proc
    end if

    if ( (n_fs_spectators_proc/=0) .and. (ibegin_spectators_proc==0) ) then
      print*, "setup_proc: setting for spectators ", ibegin_spectators_proc, n_fs_spectators_proc
    end if

    !> broadcast for the moment
    npars = n_fs_partons_proc
    nspec = n_fs_spectators_proc + n_fs_photons_proc

    !> broadcast
    inwa = decay_type_proc
    !> Q-differential? overrive rm12
    if (inwa == 4) then
      rm12 = Q_OS_proc
      rg12 = 0d0
    end if

  contains

    subroutine setCollider_epem(Ee,Ep)
      double precision :: Ee,Ep
      double precision :: pa,pb
      common /pinee/pa(4),pb(4)
      pa(4)=Ee
      pa(1)=0d0
      pa(2)=0d0
      pa(3)=Ee
      pb(4)=Ep
      pb(1)=0d0
      pb(2)=0d0
      pb(3)=-Ep
    end subroutine setCollider_epem

    subroutine setCollider_dis(sEe,sEp)
      use DIS_mod, only : setetaep,setEs
      double precision sEe,sEp
      call setEs(sEe,sEp)
      call setetaep(sEp/sEe)
    end subroutine setCollider_dis

  end subroutine setup_proc


  !-----------------------------------------------------------------------------
  !> @brief
  !> output info on the chanels settings
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_proc(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",1x,A,T103,"*")'
    character(len=*), parameter :: fmt_beam =  &
      & '(1x,"*",7x,"> beam",I1," = ",I12," [",1f6.1," GeV]",T103,"*")'
    character(len=*), parameter :: fmt_opt_str =  &
      & '(1x,"*",5x,A," = ",A,T103,"*")'
    character(len=*), parameter :: fmt_opt_dble =  &
      & '(1x,"*",5x,A," = ",1f6.1,T103,"*")'
    character(len=*), parameter :: fmt_opt_int =  &
      & '(1x,"*",5x,A," = ",I6,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_str) "Process: " // name_proc
    write(iout,fmt_opt_dble) "sqrt(s)", sqrts_proc
    if ((A_beam1>1) .or. (A_beam2>1) .or. (Z_beam1>1) .or. (Z_beam2>1)) then
      write(iout,fmt_beam) 1, beam1_proc, Ebeam1_proc*(Z_beam1*A_beam2/Z_beam2/A_beam1)**0.5
      write(iout,fmt_beam) 2, beam2_proc, Ebeam2_proc*(Z_beam2*A_beam1/Z_beam1/A_beam2)**0.5
    else
      write(iout,fmt_beam) 1, beam1_proc, Ebeam1_proc
      write(iout,fmt_beam) 2, beam2_proc, Ebeam2_proc
    end if

  end subroutine digest_proc


  subroutine inithel()
      implicit integer(a-z)
      common/nheljets/nhel(6,14,24)
      dimension j1(3),j2(8),j3(11),j4(23)

      NHEL=0
      NHEL(:,1,1) =  (/ 1,-1,-1, 1, 1, 1 /)
      NHEL(:,2,1) =  (/ 1,-1, 1,-1, 1, 1 /)
      NHEL(:,3,1) =  (/ 1,-1, 1, 1,-1, 1 /)
      NHEL(:,4,1) =  (/ 1,-1, 1, 1, 1,-1 /)

      NHEL(:,5,1) =  (/ 1,-1,-1,-1,-1, 1 /)
      NHEL(:,6,1) =  (/ 1,-1,-1,-1, 1,-1 /)
      NHEL(:,7,1) =  (/ 1,-1,-1, 1,-1,-1 /)
      NHEL(:,8,1) =  (/ 1,-1, 1,-1,-1,-1 /)

      NHEL(:,9,1) =  (/ 1,-1,-1,-1, 1, 1 /)
      NHEL(:,10,1) =  (/ 1,-1, 1, 1,-1,-1 /)
      NHEL(:,11,1) =  (/ 1,-1,-1, 1, 1,-1 /)
      NHEL(:,12,1) =  (/ 1,-1, 1,-1,-1, 1 /)
      NHEL(:,13,1) =  (/ 1,-1,-1, 1,-1, 1 /)
      NHEL(:,14,1) =  (/ 1,-1, 1,-1, 1,-1 /)

      do ii=1,6
      do jj=1,14
      do kk=2,24
      nhel(ii,jj,kk)=nhel(ii,jj,1)
      enddo
      enddo
      enddo


      do jj=1,14
      nhel(3,jj,2)=nhel(3,jj,1)
      nhel(4,jj,2)=nhel(4,jj,1)
      nhel(6,jj,2)=nhel(5,jj,1)
      nhel(5,jj,2)=nhel(6,jj,1)

      nhel(3,jj,3)=nhel(3,jj,1)
      nhel(5,jj,3)=nhel(4,jj,1)
      nhel(4,jj,3)=nhel(5,jj,1)
      nhel(6,jj,3)=nhel(6,jj,1)

      nhel(3,jj,4)=nhel(3,jj,1)
      nhel(5,jj,4)=nhel(4,jj,1)
      nhel(6,jj,4)=nhel(5,jj,1)
      nhel(4,jj,4)=nhel(6,jj,1)

      nhel(3,jj,5)=nhel(3,jj,1)
      nhel(6,jj,5)=nhel(4,jj,1)
      nhel(4,jj,5)=nhel(5,jj,1)
      nhel(5,jj,5)=nhel(6,jj,1)

      nhel(3,jj,6)=nhel(3,jj,1)
      nhel(6,jj,6)=nhel(4,jj,1)
      nhel(5,jj,6)=nhel(5,jj,1)
      nhel(4,jj,6)=nhel(6,jj,1)

      nhel(4,jj,7)=nhel(3,jj,1)
      nhel(3,jj,7)=nhel(4,jj,1)
      nhel(5,jj,7)=nhel(5,jj,1)
      nhel(6,jj,7)=nhel(6,jj,1)

      nhel(4,jj,8)=nhel(3,jj,1)
      nhel(3,jj,8)=nhel(4,jj,1)
      nhel(6,jj,8)=nhel(5,jj,1)
      nhel(5,jj,8)=nhel(6,jj,1)

      nhel(4,jj,9)=nhel(3,jj,1)
      nhel(5,jj,9)=nhel(4,jj,1)
      nhel(3,jj,9)=nhel(5,jj,1)
      nhel(6,jj,9)=nhel(6,jj,1)

      nhel(4,jj,10)=nhel(3,jj,1)
      nhel(5,jj,10)=nhel(4,jj,1)
      nhel(6,jj,10)=nhel(5,jj,1)
      nhel(3,jj,10)=nhel(6,jj,1)

      nhel(4,jj,11)=nhel(3,jj,1)
      nhel(6,jj,11)=nhel(4,jj,1)
      nhel(3,jj,11)=nhel(5,jj,1)
      nhel(5,jj,11)=nhel(6,jj,1)

      nhel(4,jj,12)=nhel(3,jj,1)
      nhel(6,jj,12)=nhel(4,jj,1)
      nhel(5,jj,12)=nhel(5,jj,1)
      nhel(3,jj,12)=nhel(6,jj,1)

      nhel(5,jj,13)=nhel(3,jj,1)
      nhel(3,jj,13)=nhel(4,jj,1)
      nhel(4,jj,13)=nhel(5,jj,1)
      nhel(6,jj,13)=nhel(6,jj,1)

      nhel(5,jj,14)=nhel(3,jj,1)
      nhel(3,jj,14)=nhel(4,jj,1)
      nhel(6,jj,14)=nhel(5,jj,1)
      nhel(4,jj,14)=nhel(6,jj,1)

      nhel(5,jj,15)=nhel(3,jj,1)
      nhel(4,jj,15)=nhel(4,jj,1)
      nhel(3,jj,15)=nhel(5,jj,1)
      nhel(6,jj,15)=nhel(6,jj,1)

      nhel(5,jj,16)=nhel(3,jj,1)
      nhel(4,jj,16)=nhel(4,jj,1)
      nhel(6,jj,16)=nhel(5,jj,1)
      nhel(3,jj,16)=nhel(6,jj,1)

      nhel(5,jj,17)=nhel(3,jj,1)
      nhel(6,jj,17)=nhel(4,jj,1)
      nhel(3,jj,17)=nhel(5,jj,1)
      nhel(4,jj,17)=nhel(6,jj,1)

      nhel(5,jj,18)=nhel(3,jj,1)
      nhel(6,jj,18)=nhel(4,jj,1)
      nhel(4,jj,18)=nhel(5,jj,1)
      nhel(3,jj,18)=nhel(6,jj,1)

      nhel(6,jj,19)=nhel(3,jj,1)
      nhel(3,jj,19)=nhel(4,jj,1)
      nhel(4,jj,19)=nhel(5,jj,1)
      nhel(5,jj,19)=nhel(6,jj,1)

      nhel(6,jj,20)=nhel(3,jj,1)
      nhel(3,jj,20)=nhel(4,jj,1)
      nhel(5,jj,20)=nhel(5,jj,1)
      nhel(4,jj,20)=nhel(6,jj,1)

      nhel(6,jj,21)=nhel(3,jj,1)
      nhel(4,jj,21)=nhel(4,jj,1)
      nhel(3,jj,21)=nhel(5,jj,1)
      nhel(5,jj,21)=nhel(6,jj,1)

      nhel(6,jj,22)=nhel(3,jj,1)
      nhel(4,jj,22)=nhel(4,jj,1)
      nhel(5,jj,22)=nhel(5,jj,1)
      nhel(3,jj,22)=nhel(6,jj,1)

      nhel(6,jj,23)=nhel(3,jj,1)
      nhel(5,jj,23)=nhel(4,jj,1)
      nhel(3,jj,23)=nhel(5,jj,1)
      nhel(4,jj,23)=nhel(6,jj,1)

      nhel(6,jj,24)=nhel(3,jj,1)
      nhel(5,jj,24)=nhel(4,jj,1)
      nhel(4,jj,24)=nhel(5,jj,1)
      nhel(3,jj,24)=nhel(6,jj,1)
      enddo

  end subroutine inithel


end module Process_mod

