!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: routines for the Projection-To-Born method [external members: `_p2b`]
!
! DESCRIPTION:
!> Module providing routines for the Projection-To-Born method
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Proj2Born_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_p2b, npar_p2b
  public :: setActive_p2b, isActive_p2b
  public :: project_epem_p2b ! e+e- (colour singlet decay in general)
  public :: project_dis_p2b  ! DIS
  public :: project_hdy_p2b  ! hdy: Higgs & Drell-Yan & GG
  public :: project_jet_p2b  ! funky technical implementation

  !> P2B observable id's
  integer :: id_x_dis_p2b = 0  ! dis
  integer :: id_qt_hdy_p2b = 0  ! hdy
  ! integer :: id_ht_part_ = 0, id_yb_part_ = 0, id_ys_part_ = 0  ! jet

  !> P2B specific settings
  double precision :: hdy_qtcut_ = -1d0  ! [GeV]

  !> each thread needs its own P2B setting
  integer :: npar_p2b = -1
  logical :: qactive = .false.
!$OMP THREADPRIVATE(npar_p2b, qactive)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise P2B
  !
  !> @param[in] type name of the mapping type
  !-----------------------------------------------------------------------------
  subroutine init_p2b(type)
    use StrHelper_mod, only: to_upper_str
    use Observables_mod, only : getIdFromName_obs
    character(len=*), intent(in) :: type

    select case(to_upper_str(type))
    case ("DIS")
      id_x_dis_p2b = getIdFromName_obs("x")
    case ("HDY")
      id_qt_hdy_p2b = getIdFromName_obs("p2b_qt")
    case ("DEC")
    case ("JET")
      ! id_ht_part_ = getIdFromName_obs("ht_part")
      ! id_yb_part_ = getIdFromName_obs("yb_part")
      ! id_ys_part_ = getIdFromName_obs("ys_part")
    case default
      print*, "init_p2b: invalid type: ", type
      stop
    end select

  end subroutine init_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate the projection to Born
  !
  !> @param[in] active boolean flag
  !> @param[in] npar_born optional argument to set the Born momentum set value
  !> @param[in] qtcut optional argument to set `hdy_qtcut_`
  !-----------------------------------------------------------------------------
  subroutine setActive_p2b(active, npar_born, qtcut)
    use Observables_mod, only : isUsed_obs, unsetManual_obs
    logical, intent(in) :: active
    integer, optional, intent(in) :: npar_born
    double precision, optional, intent(in) :: qtcut
    !> check if npar_p2b & process are valid combinations: DIS[1,2j] & npar=4
    qactive = active
    if (present(npar_born)) npar_p2b = npar_born
    !print*, "setActive_p2b: ", qactive, npar_born
    if (present(qtcut)) then
      hdy_qtcut_ = qtcut
    else
      hdy_qtcut_ = -1d0
    end if
    ! print*, "setActive_p2b: ", qactive, npar_p2b, hdy_qtcut_
    !> if a manual observable is active, reset it in P2B
    if ( id_qt_hdy_p2b > 0 ) call unsetManual_obs(id_qt_hdy_p2b)
  end subroutine setActive_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of projection
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_p2b()
    isActive_p2b = qactive
  end function isActive_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to perform the projection to Born kinematics for e+e-
  !
  !> @param[in] npar_from kinematics from which we map down to born
  !> @param[in] ipass (0=fail, 1=success)
  !-----------------------------------------------------------------------------
  subroutine project_epem_p2b(npar_from, ipass)
    use KinData_mod
    use ObsHelper_mod, only : v1_p
    integer, intent(in) :: npar_from
    integer, intent(inout) :: ipass
    type(Kin_t), pointer :: kd
    integer :: i, j
    integer :: npar_sav
    double precision, dimension(3) :: psum
    double precision :: Engy

    if (.not. isActive_p2b()) stop "project_epem_p2b: inactive!"

    if (ipass == 0) return

    kin(npar_p2b)%cacheID = -1  ! P2B indicator
    nullify( kin(npar_p2b)%cachePTR )  ! avoid cache inconcistency when: npar_from = npar_p2b

    !> prettify code
    kd => kin(npar_from)

    if (npar_from == npar_p2b) then
      npar_sav = kd%npar_from
      call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from
      kin(npar_p2b)%npar_from = npar_sav  ! must come *after* clearCacheLocal_kin
      return
    end if

    psum = 0d0
    do i = kd%ijets_lower, kd%ijets_upper
      if ( kd%p(3,i) > 0d0 ) then
        psum(1:3) = psum(1:3) + kd%p(1:3,i)
      else
        psum(1:3) = psum(1:3) - kd%p(1:3,i)
      end if
    end do
    psum = psum / sqrt( dot_eucl(psum,psum) )
    Engy = sqrt( kd%s(1,2) ) / 2d0

    ! print*, "Engy", Engy
    ! print*, "psum", psum

    !> store projection in Born phase space
    kin(npar_p2b)%p(1:2,1) = 0d0
    kin(npar_p2b)%p(1:2,2) = 0d0
    kin(npar_p2b)%p(3,1)   = +Engy
    kin(npar_p2b)%p(3,2)   = -Engy
    kin(npar_p2b)%p(4,1)   =  Engy
    kin(npar_p2b)%p(4,2)   =  Engy

    kin(npar_p2b)%p(1:3,3) = +psum * Engy
    kin(npar_p2b)%p(1:3,4) = -psum * Engy
    kin(npar_p2b)%p(4,3)   = Engy
    kin(npar_p2b)%p(4,4)   = Engy

    call fillS_kin(npar_p2b)
    call clearCacheLocal_kin(npar_p2b)

    !> save the mapping hierarchy
    kin(npar_p2b)%npar_from = npar_from    ! must come *after* clearCacheLocal_kin

    call sijcuts(npar_p2b,ipass)

    ! call print_momenta(npar_p2b)

  end subroutine project_epem_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to perform the projection to Born kinematics for DIS
  !
  !> @param[in] npar_from kinematics from which we map down to born
  !> @param[in] ipass (0=fail, 1=success)
  !-----------------------------------------------------------------------------
  subroutine project_dis_p2b(npar_from, ipass)
    use KinData_mod
    use DIS_mod
    use Observables_mod, only : getValue_obs
    integer, intent(in) :: npar_from
    integer, intent(inout) :: ipass
    double precision :: q(4), pin(4), pout(4)
    double precision :: xbj
    logical :: valid

    if (.not. isActive_p2b()) stop "project_dis_p2b: inactive!"

    if (ipass == 0) return

    kin(npar_p2b)%cacheID = -1  ! P2B indicator
    nullify( kin(npar_p2b)%cachePTR )  ! avoid cache inconcistency when: npar_from = npar_p2b

    if (npar_from == npar_p2b) return

    !> vector boson momentum
    q(:) = kin(npar_from)%p(:,2) - kin(npar_from)%p(:,3)

    !> Calculate the born kinematics
    !> Bjorken x
    xbj = getValue_obs(id_x_dis_p2b, valid)

    !> Formula (1) from VBF@NNLO PRL
    pin(:)  = kin(npar_from)%p(:,1) / dis%xx1 * xbj
    pout(:) = pin(:) + q(:)

    !> store projection in born phase space
    kin(npar_p2b)%p(:,2) = kin(npar_from)%p(:,2)
    kin(npar_p2b)%p(:,3) = kin(npar_from)%p(:,3)
    kin(npar_p2b)%p(:,1) = pin
    kin(npar_p2b)%p(:,4) = pout

    call fillS_kin(npar_p2b)
    call clearCacheLocal_kin(npar_p2b)

    !> save the mapping hierarchy
    kin(npar_p2b)%npar_from = npar_from    ! must come *after* clearCacheLocal_kin

    call sijcuts(npar_p2b,ipass)

  end subroutine project_dis_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to perform the projection to Born kinematics for Higgs & Drell-Yan
  !
  !> @param[in] npar_from kinematics from which we map down to born
  !> @param[in] ipass (0=fail, 1=success)
  !-----------------------------------------------------------------------------
  subroutine project_hdy_p2b(npar_from, ipass)
    use KinData_mod
    use ObsHelper_mod
    use Observables_mod, only : setManual_obs
    integer, intent(in) :: npar_from
    integer, intent(inout) :: ipass
    type(Kin_t), pointer :: kd
    integer :: i, j, k
    integer :: npar_sav
    double precision :: pX(4), pXt(4), r(4)
    double precision :: pX2, s1X, s2X, xi1Oxi2, xi1Txi2, xi1, xi2
    double precision :: sjX, sjXt, sXXt
    ! !> for debugging
    ! !> Higgs/gauge-boson kinematics (X = H, Z, Wp, Wm, ...)
    ! double precision :: yX, mX, ptX, mtX
    ! logical :: valid

    if (.not. isActive_p2b()) stop "project_hdy_p2b: inactive!"

    if (ipass == 0) return

    ! print*, "project_hdy_p2b:", npar_from, "->", npar_p2b
    ! read(*,*)
    kin(npar_p2b)%cacheID = -1  ! P2B indicator
    nullify( kin(npar_p2b)%cachePTR )  ! avoid cache inconcistency when: npar_from = npar_p2b

    ! if (npar_from == npar_p2b) return  ! P2B(Born) = identity transformation
   
    !> prettify code
    kd => kin(npar_from)

    if (npar_from == npar_p2b) then
      npar_sav = kd%npar_from
      call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from
      kin(npar_p2b)%npar_from = npar_sav  ! must come *after* clearCacheLocal_kin
      return
    end if
    
    !> the old cumbersone way of handling cache inconsistencies....
    !> left for future reference to explain the core issue
    ! !> by definition, P2B(Born) is the identity transformation
    ! !> why do we catch this case?
    ! !>   - performance
    ! !>   - I need to retain the npar_from information to get
    ! !>     (sub-) leading photon assignments correct
    ! !>     only relevant when I perform P2B on Born kinematics 
    ! !>     (only done for validation of inclusive calculation)
    ! !>   - just returning will mess up the npar_p2b cache:
    ! !>     6 -> 5 -> 4 -[p2b]-> 4 is ok
    ! !>     6 -> 5 -[p2b]-> 4 will modify the cache of "genuine" mappings to 4 like in the line above
    ! if (npar_from == npar_p2b) then
    !    ! npar_sav = kd%npar_from
    !    ! call fillS_kin(npar_p2b)
    !    ! !> need to clear the cache each time because the p2b mapping 
    !    ! !> does not gobally collapse to the same kinematics!
    !    ! !> II mappings change yH
    !    ! call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from
    !    ! kin(npar_p2b)%npar_from = npar_sav  ! must come *after* clearCacheLocal_kin
    !    return
    ! end if
    
    !> consistency check
    if ( npar_from-npar_p2b /= kd%ijets_upper-kd%ijets_lower+1 ) then
      print*, "project_hdy_p2b: npar mismatch: ", npar_from, npar_p2b, kd%ijets_upper, kd%ijets_lower
      stop
    end if

    pX(:) = 0d0
    pX2 = 0d0
    s1X = 0d0
    s2X = 0d0
    !> loop over all X-decay products (not QCD partons)
    do i = 3, npar_from
      if ( i >= kd%ijets_lower .and. i <= kd%ijets_upper ) cycle
      pX(:) = pX(:) + kd%p(:,i)
      ! attention: s(a,i) are negative: s(a,i) = (pa-pi)^2 = -2pa.pi
      s1X = s1X + kd%s(1,i)
      s2X = s2X + kd%s(2,i)
      do j = i+1, npar_from
        if ( j >= kd%ijets_lower .and. j <= kd%ijets_upper ) cycle
        pX2 = pX2 + kd%s(i,j)
      end do
    end do
    xi1Oxi2 = s2X / s1X

    !> apply optional qt cut
    if ( hdy_qtcut_ > 0d0 ) then
      if ( (pX(1)**2 + pX(2)**2) < hdy_qtcut_**2 ) then
        ipass = 0
        return
      end if
    end if
    call setManual_obs(id_qt_hdy_p2b, sqrt(pX(1)**2 + pX(2)**2))
    !> check cuts (if set) here?

    xi1Txi2 = kd%s(1,2)
    !> loop over all QCD partons
    do i = kd%ijets_lower, kd%ijets_upper
      ! for initial-final s(a,i): sign = -1!
      xi1Txi2 = xi1Txi2 + kd%s(1,i) + kd%s(2,i)
      do j = i+1, kd%ijets_upper
        ! for final-final s(i,j): sign = +1!
        xi1Txi2 = xi1Txi2 + kd%s(i,j)
      end do
    end do
    xi1Txi2 = xi1Txi2 / kd%s(1,2)

    !> splitting
    xi1 = sqrt(xi1Txi2 * xi1Oxi2)
    xi2 = sqrt(xi1Txi2 / xi1Oxi2)

    !> mapped X momentum 
    pXt(:) = xi1*kd%p(:,1) + xi2*kd%p(:,2)
    r(:) = pX(:) + pXt(:)

    !> store projection in born phase space
    kin(npar_p2b)%p(:,1) = xi1 * kin(npar_from)%p(:,1)
    kin(npar_p2b)%p(:,2) = xi2 * kin(npar_from)%p(:,2)
    ! X decay products receive a boost
    sXXt = - ( xi1*s1X + xi2*s2X )  ! positive
    do i = 3, npar_p2b
      ! the mapping that needs to use the kinematics structure
      j = kd%ijets_upper + (i-2)  ! decay products tail the partons in the kinematics array
      ! get all scalar products
      sjX = 0d0
      do k = 3, npar_from
        if ( k == j ) cycle
        if ( k >= kd%ijets_lower .and. k <= kd%ijets_upper ) cycle
        sjX = sjX + kd%s(j,k)
      end do
      sjXt = - ( xi1*kd%s(1,j) + xi2*kd%s(2,j) )  ! positive
      ! print*, j, "sjX  ", sjX,  2d0*dot(kin(npar_from)%p(:,j), pX)
      ! print*, j, "sjXt ", sjXt, 2d0*dot(kin(npar_from)%p(:,j), pXt)
      kin(npar_p2b)%p(:,i) = kin(npar_from)%p(:,j)  &
                           & - (sjX + sjXt) / (2d0*pX2 + sXXt) * r(:)  &
                           & + sjX / pX2 * pXt(:)
    end do

    call fillS_kin(npar_p2b)
    call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from

    ! print*, "pX ", pX,  ">", v1_minv2(pX), "vs.", pX2
    ! print*, "pXt", pXt, ">", v1_minv2(pXt)
    ! call print_momenta(npar_p2b)
    ! read(*,*)
    !> save the mapping hierarchy
    kin(npar_p2b)%npar_from = npar_from  ! must come *after* clearCacheLocal_kin

    call sijcuts(npar_p2b,ipass)

  end subroutine project_hdy_p2b


  !-----------------------------------------------------------------------------
  !> @brief
  !> routine to perform the projection to Born kinematics for jet production
  !
  !> @param[in] npar_from kinematics from which we map down to born
  !> @param[in] ipass (0=fail, 1=success)
  !-----------------------------------------------------------------------------
  subroutine project_jet_p2b(npar_from, ipass)
    use KinData_mod
    use ObsHelper_mod
    ! use Observables_mod
    ! use Constants_mod
    integer, intent(in) :: npar_from
    integer, intent(inout) :: ipass
    type(Kin_t), pointer :: kd
    integer :: i, j, k
    integer :: npar_sav
    ! Born variables
    double precision :: ht, yb, ys  !, ph
    double precision :: pt
    ! projection
    double precision :: xi1, xi2, xi1Oxi2, xi1Txi2
    ! ! debug
    ! double precision :: t1, t2

    if (.not. isActive_p2b()) stop "project_jet_p2b: inactive!"

    if (ipass == 0) return

    ! print*, "project_jet_p2b:", npar_from, "->", npar_p2b
    ! read(*,*)
    kin(npar_p2b)%cacheID = -1  ! P2B indicator
    nullify( kin(npar_p2b)%cachePTR )  ! avoid cache inconcistency when: npar_from = npar_p2b
   
    !> prettify code
    kd => kin(npar_from)

    if (npar_from == npar_p2b) then
      npar_sav = kd%npar_from
      call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from
      kin(npar_p2b)%npar_from = npar_sav  ! must come *after* clearCacheLocal_kin
      return
    end if

    ht = 0d0
    yb = 0d0
    ! ph = 0d0
    do i = kd%ijets_lower, kd%ijets_upper
      pt = v1_pt(kd%p(:,i))
      ht = ht + pt
      yb = yb + pt * v1_y(kd%p(:,i))
      ! ph = ph + pt * v1_phi(kd%p(:,i))
    end do
    yb = yb / ht
    ! ph = ph / ht
    ys = 0d0
    do i = kd%ijets_lower, kd%ijets_upper
      ys = ys + v1_pt(kd%p(:,i)) * abs( v1_y(kd%p(:,i)) - yb )
    end do
    ys = ys / ht

    xi1Oxi2 = 1d0
    do i = kd%ijets_lower, kd%ijets_upper
      pt = v1_pt(kd%p(:,i))
      ! attention: s(a,i) are negative: s(a,i) = (pa-pi)^2 = -2pa.pi
      xi1Oxi2 = xi1Oxi2 * (kd%s(2,i)/kd%s(1,i))**(pt/ht)
    end do

    xi1Txi2 = (ht*cosh(ys))**2 / kd%s(1,2)

    !> splitting
    xi1 = sqrt(xi1Txi2 * xi1Oxi2)
    xi2 = sqrt(xi1Txi2 / xi1Oxi2)

    !> store projection in born phase space
    kin(npar_p2b)%p(:,1) = xi1 * kin(npar_from)%p(:,1)
    kin(npar_p2b)%p(:,2) = xi2 * kin(npar_from)%p(:,2)
    !> outgoing 3
    pt = ht / 2d0
    kin(npar_p2b)%p(1,3) = pt   ! pt * cos(ph+pi/2d0)
    kin(npar_p2b)%p(2,3) = 0d0  ! pt * sin(ph+pi/2d0)
    kin(npar_p2b)%p(3,3) = pt * sinh(+ys+yb)
    kin(npar_p2b)%p(4,3) = pt * cosh(+ys+yb)
    !> outgoing 4
    kin(npar_p2b)%p(1,4) = -pt  ! pt * cos(ph-pi/2d0)
    kin(npar_p2b)%p(2,4) = 0d0  ! pt * sin(ph-pi/2d0)
    kin(npar_p2b)%p(3,4) = pt * sinh(-ys+yb)
    kin(npar_p2b)%p(4,4) = pt * cosh(-ys+yb)

    call fillS_kin(npar_p2b)
    call clearCacheLocal_kin(npar_p2b)  ! attention: this guy resets npar_from


    ! !> debug
    ! t1 = sqrt(x2_kin/x1_kin) * ht * cosh(ys) * exp(+yb) / sqrt(kd%s(1,2))
    ! t2 = sqrt(x1_kin/x2_kin) * ht * cosh(ys) * exp(-yb) / sqrt(kd%s(1,2))
    ! print*, "xi1", xi1, t1, xi1/t1
    ! print*, "xi2", xi2, t2, xi2/t2
    ! call print_momenta(npar_p2b)
    ! print*, "------------------------"
    ! print*, "ht", ht
    ! print*, "yb", yb
    ! print*, "ys", ys
    ! print*, "------------------------"
    ! call printAll_obs(npar_from, skipCache=.true.)
    ! call printAll_obs(npar_p2b,  skipCache=.true.)
    ! read(*,*)
    !> save the mapping hierarchy
    kin(npar_p2b)%npar_from = npar_from  ! must come *after* clearCacheLocal_kin

    call sijcuts(npar_p2b,ipass)

  end subroutine project_jet_p2b


end module Proj2Born_mod
