!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Reweight [external members: `_rewgt`]
!
! DESCRIPTION:
!> Module to isolate the routines for a re-weighting of the weight returned to
!> Vegas. The re-weighting function is specified by an observable and the power
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Reweight_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: ReweightFunction_t
  public :: setActive_rewgt
  public :: isActive_rewgt, isActiveLocal_rewgt
  public :: set_rewgt, setLocal_rewgt
  public :: eval_rewgt
  public :: get_rewgt, getLocal_rewgt
  public :: toString_rewgt

  !> this object corresponds to a function:
  !>  fac * [obs]^pow  + off
  type ReweightFunction_t
    logical :: qactive = .false.
    integer :: obsId = -1  ! ngeative value: invalid
    double precision :: fac = 1d0
    double precision :: pow = 1d0
    double precision :: off = 0d0
  end type ReweightFunction_t

  !> the global reweighting is only applied to the weight returned to VEGAS
  !> and only affects the adaption but not the histograms
  type(ReweightFunction_t) :: global_rewgt

  !> this is an optional reweighting function set by the MultiRun module
  !> it is applied to *everything*, both to VEGAS weights and histograms
  type(ReweightFunction_t) :: local_rewgt
!$OMP THREADPRIVATE(local_rewgt)

  !> @interface
  !> @brief
  !> this is the external interface to set  the "local" reweighting function 
  !> for multi-runs which is overloaded to the two calling options
  interface setLocal_rewgt
    module procedure setLocalByFunc
    module procedure setLocalByObs
  end interface

contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the reweighting function
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_rewgt(active)
    logical, intent(in) :: active
    if (active .and. global_rewgt%obsId<0) then
      print*, "setActive_rewgt: invalid observale; cannot activate!"
      return
    end if
    global_rewgt%qactive = active
  end subroutine setActive_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of re-weighting
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_rewgt()
    isActive_rewgt = ( global_rewgt%qactive .or. isActiveLocal_rewgt() )
  end function isActive_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of the "local" re-weighting
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActiveLocal_rewgt()
    isActiveLocal_rewgt = local_rewgt%qactive
  end function isActiveLocal_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the reweighting function by specifying the observable and power
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] fac optional factor to take of the observable
  !> @param[in] pow optional power to take of the observable
  !> @param[in] off optional offset: must be a constant
  !> @param[out] rewgtfcn optional ReweightFunction_t for output, 
  !>             otherwise global reweighting function will be set
  !-----------------------------------------------------------------------------
  subroutine set_rewgt(obsName, fac, pow, off, rewgtfcn)
    use Observables_mod, only: getIdFromName_obs
    character(len=*), intent(in) :: obsName
    double precision, optional, intent(in) :: fac, pow, off
    type(ReweightFunction_t), optional, intent(out) :: rewgtfcn
    integer :: obs_val
    double precision :: fac_val, pow_val, off_val

    !> observable
    if (trim(obsName) .eq. '') then
      obs_val = 0  ! only const = off active
    else
      obs_val = getIdFromName_obs(obsName)
    end if

    !> default values
    fac_val = 1d0
    pow_val = 1d0
    off_val = 0d0
    !> optional input
    if (present(fac)) fac_val = fac
    if (present(pow)) pow_val = pow
    if (present(off)) off_val = off

    !> copy over the values
    if (present(rewgtfcn)) then
      rewgtfcn%qactive     = .true.
      rewgtfcn%obsId       = obs_val
      rewgtfcn%fac         = fac_val
      rewgtfcn%pow         = pow_val
      rewgtfcn%off         = off_val
    else
      global_rewgt%qactive = .true.
      global_rewgt%obsId   = obs_val
      global_rewgt%fac     = fac_val
      global_rewgt%pow     = pow_val
      global_rewgt%off     = off_val
      print*, "setting a global reweighting function: ", toString_rewgt(global_rewgt)
    end if

  end subroutine set_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the "local" reweighting function for multi-runs
  !
  !> @param[in] rewgtfcn the reweighting function to set
  !-----------------------------------------------------------------------------
  subroutine setLocalByFunc(rewgtfcn)
    type(ReweightFunction_t), intent(in) :: rewgtfcn
    local_rewgt = rewgtfcn
  end subroutine setLocalByFunc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the set the "local" reweighting function for multi-runs 
  !> by specifying the observable and power
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] fac optional factor to take of the observable
  !> @param[in] pow optional power to take of the observable
  !> @param[in] off optional offset: must be a constant
  !-----------------------------------------------------------------------------
  subroutine setLocalByObs(obsName, fac, pow, off)
    character(len=*), intent(in) :: obsName
    double precision, optional, intent(in) :: fac, pow, off
    type(ReweightFunction_t) :: rewgtfcn

    call set_rewgt(obsName, fac=fac, pow=pow, off=off, rewgtfcn=rewgtfcn) 
    call setLocalByFunc(rewgtfcn)

  end subroutine setLocalByObs


  !-----------------------------------------------------------------------------
  !> @brief
  !> evaluate a reweighting function
  !
  !> @param[in] rewgtfcn ReweightFunction_t to use
  !> @param[inout] iscl optional id for the scale dependence of rewgtfcn:
  !>     * scale independent rewgtfcn: iscl[out]=0
  !>     * scale dependent rewgtfcn:
  !>         iscl[in]=0: return central scale and iscl[out]=1
  !>         iscl[in]>0: return iscl-th scale and iscl[out]=iscl[in]
  !> @return weight function
  !-----------------------------------------------------------------------------
  double precision function eval_rewgt(rewgtfcn, iscl)
    use Observables_mod, only : getValue_obs
    type(ReweightFunction_t), intent(in) :: rewgtfcn
    integer, optional, intent(inout) :: iscl
    integer :: iscl_ret
    double precision :: valObs
    logical :: valid

    !> inactive
    if (.not.rewgtfcn%qactive) then
      if ( present(iscl) ) iscl = 0
      eval_rewgt = 1d0
      return
    end if
  
    !> only a constant
    if (rewgtfcn%obsId==0) then
      if ( present(iscl) ) iscl = 0
      eval_rewgt = rewgtfcn%off
      return
    end if

    !> normal observable
    valObs = getValue_obs(rewgtfcn%obsId, valid)
    if (.not.valid) then
      print*, "eval_rewgt: invalid observable"
      stop
    end if
    if ( present(iscl) ) iscl = 0
    eval_rewgt = rewgtfcn%fac * valObs**rewgtfcn%pow + rewgtfcn%off

  end function eval_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the reweighting function (global & local)
  !
  !> @param[inout] iscl optional id for the scale dependence of rewgtfcn:
  !>     * scale independent rewgtfcn: iscl[out]=0
  !>     * scale dependent rewgtfcn:
  !>         iscl[in]=0: return central scale and iscl[out]=1
  !>         iscl[in]>0: return iscl-th scale and iscl[out]=iscl[in]
  !> @return weight function
  !-----------------------------------------------------------------------------
  double precision function get_rewgt(iscl)
    integer, optional, intent(inout) :: iscl
    integer :: iscl_ret, iscl_global, iscl_local

    iscl_global = 0
    iscl_local  = 0

    if ( present(iscl) ) then
      iscl_global = merge(iscl, 0, global_rewgt%qactive)
      iscl_local  = merge(iscl, 0, isActiveLocal_rewgt())
    end if

    !> global & local reweighting function
    if (.not.isActive_rewgt()) then
      print*, "get_rewgt: global & local reweighting not active"
      stop
    end if

    get_rewgt = 1d0

    !> apply the global reweighting
    if (global_rewgt%qactive) then
      get_rewgt = get_rewgt * eval_rewgt(global_rewgt, iscl_global)
    end if

    !> apply the local reweighting
    if (isActiveLocal_rewgt()) then
      get_rewgt = get_rewgt * getLocal_rewgt(iscl_local)
    end if

    !> determine return iscl
    if (iscl_global*iscl_local == 0) then
      !> at least one is scale-independent
      iscl_ret = iscl_global + iscl_local
    else if (iscl_global == iscl_local) then
      !> if both are scale-dependent, better be the same
      iscl_ret = iscl_global
    else
      !> error
      print*, "get_rewgt: iscl mismatch: ", iscl_global, iscl_local
      stop
    end if

    if ( present(iscl) ) iscl = iscl_ret
    
  end function get_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the "local" reweighting function
  !
  !> @param[inout] iscl optional id for the scale dependence of rewgtfcn:
  !>     * scale independent rewgtfcn: iscl[out]=0
  !>     * scale dependent rewgtfcn:
  !>         iscl[in]=0: return central scale and iscl[out]=1
  !>         iscl[in]>0: return iscl-th scale and iscl[out]=iscl[in]
  !> @return weight function
  !-----------------------------------------------------------------------------
  double precision function getLocal_rewgt(iscl)
    integer, optional, intent(inout) :: iscl
    integer :: iscl_ret, iscl_local

    iscl_local  = 0

    if ( present(iscl) ) then
      iscl_local = merge(iscl, 0, local_rewgt%qactive)
    end if

    if (.not.local_rewgt%qactive) then
      print*, "getLocal_rewgt: local reweighting not active"
      stop
    end if

    getLocal_rewgt = 1d0

    if (local_rewgt%qactive) then
      getLocal_rewgt = getLocal_rewgt * eval_rewgt(local_rewgt, iscl_local)
    end if

    iscl_ret = iscl_local

    if ( present(iscl) ) iscl = iscl_ret

  end function getLocal_rewgt


  !-----------------------------------------------------------------------------
  !> @brief
  !> Convert ReweightFunction_t to a string
  !
  !> @param[in] rewgtfcn optional ReweightFunction_t, 
  !>            otherwise global reweighting function will be used
  !> @return name string
  !-----------------------------------------------------------------------------
  function toString_rewgt(rewgtfcn)
    use Observables_mod, only: getNameFromId_obs
    character(len=100) :: toString_rewgt
    type(ReweightFunction_t), optional, intent(in) :: rewgtfcn
    !> local var for simplicity and no performance-critical function
    type(ReweightFunction_t) :: locRF
    character(len=100) :: dummy

    locRF = global_rewgt
    if (present(rewgtfcn)) then
      locRF = rewgtfcn
    end if

    toString_rewgt = ''

    if (.not.locRF%qactive .or. locRF%obsId < 0) return

    !> a constant
    if (locRF%obsId == 0) then
      write(toString_rewgt,'(F10.2)') locRF%off
      return
    end if

    !> build up the general case
    if (locRF%fac /= 1d0) then
      write(dummy,'(F5.2)') locRF%fac
      toString_rewgt = trim(toString_rewgt) // trim(dummy) // " * "
    end if

    toString_rewgt = trim(toString_rewgt) // getNameFromId_obs(locRF%obsId)

    if (locRF%pow /= 1d0) then
      write(dummy,'(F5.2)') locRF%pow
      toString_rewgt = trim(toString_rewgt) // " ** " // trim(dummy)
    end if

    if (locRF%off /= 0d0) then
      write(dummy,'(F5.2)') locRF%off
      toString_rewgt = trim(toString_rewgt) // " + " // trim(dummy)
    end if

  end function toString_rewgt


end module Reweight_mod

