!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Run [external members: `_run`]
!
! DESCRIPTION:
!> Module collecting all run settings
!> (process-specific settings are stored in Process_mod)
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Run_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  !> global stuff
  public :: init_run, setup_run, destroy_run, digest_run
  public :: set_pdf_run, init_pdf_run, destroy_pdf_run
  public :: name_run
  public :: pdf_set_run, pdf_mem_run, pdf_nf_run
  public :: pdf_xmin_run, pdf_xmax_run, pdf_qmin_run, pdf_qmax_run
  public :: iseed_run, imethod_run, iplot_run
  public :: set_lips_run, ilips_run
  public :: qtsample_th_run
  public :: tcut_run, qtcut_veto_run
  public :: angular_average_run
  public :: reset_vegas_grid_run
  public :: print_max_weight_run
  public :: pole_check_run

  public :: smin_run, set_smin_run

  !> MatchObs interface
  public :: setActiveMatchObs_run
  public :: isActiveMatchObs_run
  public :: qTriggerMatchObs_run

  !> management of sweeps to be performed
  public :: registerSweep_run, currentSweep_run
  public :: resetCurrentSweep_run, loadNextSweep_run
  public :: qproductionSweep_run

  !> run name
  character(128) :: name_run

  !> PDF info
  character(128) :: pdf_set_run
  integer :: pdf_mem_run = -1  ! init to an invalid state
  integer :: pdf_nf_run = 0
  double precision :: pdf_xmin_run = 0d0
  double precision :: pdf_xmax_run = 0d0
  double precision :: pdf_qmin_run = -1d0  ! negative: unset
  double precision :: pdf_qmax_run = -1d0  ! negative: unset
  !> if we want to use two different PDF sets
  character(128) :: pdf2_set_run
  integer :: pdf2_mem_run = -1

  !> more run settings
  double precision :: tcut_run = 1d-8
  logical :: qtcut_veto_run = .false.

  integer :: iseed_run   = 1
  integer :: imethod_run = 2
  !> zero for MC, 2 for point test, 3 for spikes, 4 for inclusive calc
  integer :: iplot_run   = 0

  !> switch for the phase-space generator
  integer :: ilips_run   = 0  ! = default
  !>    QT:  (ilips_run = 1)  transition from lin to tanh sampling
  double precision :: qtsample_th_run = 0d0
  !> (1-T):  (ilips_run = 1)  optimize for unresolved region

  logical :: angular_average_run  = .true.
  ! logical :: cache_kinematics_run = .true. ! --> Mapping_mod > qcache_map
  logical :: reset_vegas_grid_run = .false.
  logical :: print_max_weight_run = .false.
  logical :: pole_check_run       = .false.

  double precision :: smin_run = 0d0

  !>----- observables matching
  logical :: qactive_MatchObs_run = .false.  ! off by default
  double precision :: threshold_MatchObs_run = 1d-8  ! default value

  !>----- sweep data structure
  !> type defining a sweep (can be warmup or production)
  type Sweep_t
    !> type
    !>   1: warmup
    !>   2: production
    integer :: type = 0
    !> number of calls ()
    integer :: ncall
    integer :: nit
    type(Sweep_t), pointer :: next => null()
  end type Sweep_t

  !> the linked list to store *distinct* contributions
  type(Sweep_t), pointer :: headSweep_p => null()
  !> the pointer for the currently active contribution
  type(Sweep_t), pointer :: currentSweep_p => null()
  logical :: qproductionSweep_run


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the run module
  !
  !> @param[in] runname string for the name of the run (=/= process name)
  !-----------------------------------------------------------------------------
  subroutine init_run(runname)
    use StrHelper_mod
    character(len=*), intent(in) :: runname

    !> save run name
    name_run = trim(adjustl( runname ))

    if (trim(name_run) == '') then
      print*, "init_run: empty run name?!"
      stop
    end if

    !> reset smin
    call set_smin_run(0d0, reset=.true.)

  end subroutine init_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the parton distribution function(s)
  !
  !> @param[in] set  optional string for the name of the PDF (1st)
  !> @param[in] mem  optional id for the set member (1st)
  !> @param[in] set2 optional string for the name of the PDF (2nd)
  !> @param[in] mem2 optional id for the set member (2nd)
  !-----------------------------------------------------------------------------
  subroutine set_pdf_run(set, mem, set2, mem2)
    use StrHelper_mod
    use LHAPDF_mod
    character(len=*), intent(in), optional :: set, set2
    integer, intent(in), optional :: mem, mem2

    if ( present(set) ) then
      pdf_set_run = trim(adjustl( set ))
      pdf_mem_run = 0  ! default
      if ( present(mem) ) pdf_mem_run = mem
    end if

    if ( present(set2) ) then
      pdf2_set_run = trim(adjustl( set2 ))
      pdf2_mem_run = 0
      if ( present(mem2) ) pdf2_mem_run = mem2
    end if

    !> store some PDF info
    if ( pdf_mem_run >= 0 ) then
      call init_pdf_run()
      !> xmin is the minimum allowed momentum fraction, straight from the PDF set.
      pdf_xmin_run = xMin_lhapdf()
      !> xmax is the maximum allowed momentum fraction, straight from the PDF set.
      pdf_xmax_run = xMax_lhapdf()
      if (pdf_xmax_run > 1d0) then
        print*, "set_pdf_run: PDF fraction beyond NNLOJET convention of 1: ", pdf_xmax_run
        stop 
      end if
      !> muF range
      pdf_qmin_run = qMin_lhapdf()
      pdf_qmax_run = qMax_lhapdf()
      !> nf in running
      pdf_nf_run = nf_lhapdf()
      call destroy_pdf_run()
    end if

  end subroutine set_pdf_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise LHAPDF
  !-----------------------------------------------------------------------------
  subroutine init_pdf_run()
    use LHAPDF_mod

    if ( pdf_mem_run < 0 ) then
      stop "init_pdf_run: PDF set 1 not initialised?!"
    end if

    !> @todo Parameters.f90: as_pdf, vs. internal as routine
    ! if (nPDFs_proc > 0) then

    if ( pdf2_mem_run >= 0 ) then
      call init_lhapdf( set =trim(pdf_set_run),  mem =pdf_mem_run,  &
                      & set2=trim(pdf2_set_run), mem2=pdf2_mem_run)
    else
      call init_lhapdf(set=trim(pdf_set_run), mem=pdf_mem_run)
    end if

    ! end if

  end subroutine init_pdf_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate dynamic memory from LHAPDF
  !-----------------------------------------------------------------------------
  subroutine destroy_pdf_run()
    use LHAPDF_mod
    call destroy_lhapdf()
  end subroutine destroy_pdf_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the run module
  !-----------------------------------------------------------------------------
  subroutine setup_run()
    use Process_mod, only : s_proc, nPDFs_proc, decay_type_proc
    !> broadcast to common block for now
    integer :: imethod
    common /method/ imethod
    integer :: iang
    common /techpar/ iang
    double precision :: rm12, rg12
    integer :: inwa
    common /nwa/ rm12,rg12,inwa
    double precision :: ymin, y0
    common /tcuts/ ymin,y0
    integer :: iplot
    common /plotmode/ iplot

    if ( imethod_run/=1 .and. imethod_run/=2 ) then
      print*, "setup_run: invalid imethod: ", imethod_run
      stop
    end if

    if ( iseed_run<0 ) then
      print*, "setup_run: invalid iseed: ", iseed_run
      stop
    end if

    !> determine smin
    call set_smin_run(pdf_xmin_run**2*s_proc)
    ! call set_smin_run(pdf_xmin_run**nPDFs_proc * s_proc) ! better
    if (decay_type_proc >= 3 .or. decay_type_proc == -20) call set_smin_run(rm12**2)
    !> selectors can further change the min value

    !> for now, broadcast to common blocks
    imethod = imethod_run
    if (angular_average_run) then
      iang = 1
    else
      iang = 0
    end if
    y0 = tcut_run
    iplot = iplot_run

  end subroutine setup_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch to a different phase-space parametrisation/generator
  !
  !> @param[in] name string identifying the LIPS
  !> @param[in] optDBLE optional double setting of the option
  !-----------------------------------------------------------------------------
  subroutine set_lips_run(name, optDBLE)
    use StrHelper_mod, only : to_upper_str
    use Process_mod, only : name_proc
    character(len=*), intent(in) :: name
    double precision, intent(in), optional :: optDBLE

    select case (name_proc)
    !>---------------------------------
    case ("HJ", "ZJ", "WPJ", "WMJ", "GGJFC", "GG", "GG_FRAG")
      select case (to_upper_str(name))
      case ("QT", "Q_T", "PT", "P_T")
        print*, "set_lips_run: switching to the qt-optimised phase-space parametrisation"
        ilips_run = 1
        if (.not.present(optDBLE)) then
          print*, "set_lips_run: ", name, " requires optDBLE!"
          stop
        end if
        qtsample_th_run = optDBLE
        if (qtsample_th_run.ne.0d0) then
          print*, "set_lips_run: qt-optimised phase-space with qtsample_th_run = ",qtsample_th_run
        endif
      case default
        print*, "set_lips_run: unknown phase-space option for ", name_proc ," ", name
        ilips_run = 0  ! default
        qtsample_th_run = 0d0  ! default
      end select
    !>---------------------------------
    case ( "EPEM", "EPEM_H", "EPEMJJ", "EPEMJJ_H", "EPEMINC" )
      select case (to_upper_str(name))
      case ("THRUST", "IR", "IRC", "1MT", "OMT")
        print*, "set_lips_run: switching to the (1-T)-optimised phase-space parametrisation"
        ilips_run = 1
        if (present(optDBLE)) then
          print*, "set_lips_run: ", name, " require no optDBLE?!"
          stop
        end if
      case default
        print*, "set_lips_run: unknown phase-space option for ", name_proc ," ", name
        ilips_run = 0  ! default
      end select
    !>---------------------------------
    case default
      print*, "set_lips_run: ignoring phase-space option ", name
      ilips_run = 0  ! default
      qtsample_th_run = 0d0 ! default
    end select

  end subroutine set_lips_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the observables-matching feature
  !
  !> @param[in] smin new minimal value
  !> @param[in] reset optional flag to reset smin
  !-----------------------------------------------------------------------------
  subroutine set_smin_run(smin, reset)
    double precision, intent(in) :: smin
    logical, intent(in), optional :: reset
    logical :: qreset
    !> broadcast
    double precision :: s12_min
    common /s12_min/ s12_min

    qreset = .false.  ! default
    if (present(reset)) qreset = reset

    if (qreset) then
      print*, "set_smin_run: resetting smin"
      smin_run = 0d0
    end if

    if (smin > smin_run) then
      smin_run = smin
    end if

    print*, "set_smin_run: smin = ", smin_run
    s12_min = smin_run

  end subroutine set_smin_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> destroy the run module by de-allocating everything we allocated
  !> dynamically on the course of the run
  !-----------------------------------------------------------------------------
  subroutine destroy_run()
    type(Sweep_t), pointer :: it, next

    !> de-allocate the sweep-list
    it => headSweep_p
    do while ( associated(it) )
      next => it%next  ! save for next step
      deallocate(it)
      it => next
    end do
    nullify(headSweep_p)
    nullify(currentSweep_p)

  end subroutine destroy_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the observables-matching feature
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActiveMatchObs_run(active, threshold)
    logical, intent(in) :: active
    double precision, intent(in), optional :: threshold
    qactive_MatchObs_run = active
    if ( present(threshold) ) then
      threshold_MatchObs_run = threshold
    end if
  end subroutine setActiveMatchObs_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status of observables-matching
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActiveMatchObs_run()
    isActiveMatchObs_run = qactive_MatchObs_run
  end function isActiveMatchObs_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> check if a value for the dimensionless variable yval triggers obs-matching
  !
  !> @param[in] yval is the value to be tested
  !> @return trigger boolean flag
  !-----------------------------------------------------------------------------
  logical function qTriggerMatchObs_run(yval)
    double precision, intent(in) :: yval
    if (.not. qactive_MatchObs_run) then
      qTriggerMatchObs_run = .false.  ! not active: never trigger
    else
      qTriggerMatchObs_run = ( yval < threshold_MatchObs_run )
    end if
  end function qTriggerMatchObs_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new sweep
  !
  !> @param[in] type
  !> @param[in] ncall
  !> @param[in] nit
  !-----------------------------------------------------------------------------
  !> @todo to register different types (time, auto etc.) make this an interface!
  subroutine registerSweep_run(type, ncall, nit)
    integer, intent(in) :: type, ncall, nit
    type(Sweep_t), pointer :: it, last
    logical :: found_prod
    integer :: max_ncall

    print*, "registerSweep_run:", type, ncall, nit

    !> sanity checks
    if ( ncall<=0 .or. nit<=0 ) then
      print*, "registerSweep_run: negative numbers?!", ncall, nit
      stop
    end if

    found_prod = .false.
    max_ncall  = 0

    !> find the tail; we want the ordering:
    !>   head => "1" => ... => "N" => null()
    it => headSweep_p
    nullify(last)
    do while ( associated(it) )
      found_prod = found_prod .or. (it%type==2)
      if ( it%type==1 .and. it%ncall>max_ncall ) max_ncall = it%ncall
      last => it  ! save it so I can find the tail
      it   => it%next
    end do


    !> sanity checks: if there's a production it has to be the last sweep
    if (qproductionSweep_run .neqv. found_prod) then
      print*, "registerSweep_run: you should never get here"
      stop
    end if
    if (found_prod .and. type==2) then
      print*, "registerSweep_run: at most one production sweep"
      stop
    end if
    if (found_prod .and. type==1) then
      print*, "registerSweep_run: no warmups after a production; fix your order"
      stop
    end if
    if (type==1 .and. ncall<max_ncall) then
      print*, "registerSweep_run: WARNING: later warmups should have higher ncall..."
      ! stop
    end if

    qproductionSweep_run = (type==2)

    !> allocate new tail element & connect
    allocate(it)
    nullify(it%next)
    if (.not.associated(headSweep_p)) headSweep_p => it
    if (associated(last)) last%next => it

    !> set up new entry
    it%type  = type
    it%ncall = ncall
    it%nit   = nit

  end subroutine registerSweep_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset the pointer to the current sweep
  !-----------------------------------------------------------------------------
  subroutine resetCurrentSweep_run()
    print*, "resetCurrentSweep_run"
    nullify(currentSweep_p)
  end subroutine resetCurrentSweep_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> load the next sweep
  !
  !> @return boolean flag indicating success (.true. = loaded next contrib)
  !-----------------------------------------------------------------------------
  logical function loadNextSweep_run()

    !> get the "next" element
    if (.not.associated(currentSweep_p)) then
      currentSweep_p => headSweep_p
    else
      currentSweep_p => currentSweep_p%next
    end if

    !> success?
    if (.not.associated(currentSweep_p)) then
      loadNextSweep_run = .false.  ! reached end
      return
    else
      loadNextSweep_run = .true.  ! success
    end if

    !> @todo: load all the sweep-specific stuff here...
    print*, "loadNextSweep_run: ", currentSweep_p%type

  end function loadNextSweep_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> get information on the current sweep
  !
  !> @param[out] type optional
  !> @param[out] ncall optional
  !> @param[out] nit optional
  !-----------------------------------------------------------------------------
  subroutine currentSweep_run(type, ncall, nit)
    integer, intent(out), optional :: type, ncall, nit

    if (present(type))  type  = currentSweep_p%type
    if (present(ncall)) ncall = currentSweep_p%ncall
    if (present(nit))   nit   = currentSweep_p%nit

  end subroutine currentSweep_run


  !-----------------------------------------------------------------------------
  !> @brief
  !> output info on the run settings
  !-----------------------------------------------------------------------------
  subroutine digest_run(unit)
    integer, optional, intent(in) :: unit
    integer :: iout
    character(len=*), parameter :: fmt_blank =  &
      & '(1x,"*",T103,"*")'
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_sweep =  &
      & '(1x,"*",T10,A," = ",I0,"[",I0,"]",T103,"*")'
    character(len=*), parameter :: fmt_pdf =  &
      & '(1x,"*",5x,"PDF",I0," = ",A," [mem = ",I0,"]",T103,"*")'
    character(len=*), parameter :: fmt_opt_str =  &
      & '(1x,"*",5x,A," = ",A,T103,"*")'
    character(len=*), parameter :: fmt_opt_dble =  &
      & '(1x,"*",5x,A," = ",1f6.1,T103,"*")'
    character(len=*), parameter :: fmt_opt_int =  &
      & '(1x,"*",5x,A," = ",I6,T103,"*")'
    integer :: type, ncall, nit

    if (present(unit)) then
      iout = unit
    else
      iout = 6  ! output to screen
    endif

    write(iout,fmt_opt_str) "Run:", trim(name_run)
    write(iout,fmt_pdf) 1, trim(pdf_set_run), pdf_mem_run
    if (pdf2_mem_run >= 0) write(iout,fmt_pdf) 2, trim(pdf2_set_run), pdf2_mem_run
    if (qactive_MatchObs_run) then
      write(iout,fmt_opt_str) "MatchObs", "ON"
      write(iout,fmt_opt_dble) "threshold", threshold_MatchObs_run
    else
      write(iout,fmt_opt_str) "MatchObs", "OFF"
    end if

    write(iout,fmt_blank)
    write(iout,fmt_str) "> Sweeps <"
    write(iout,fmt_blank)
    call resetCurrentSweep_run()
    do while (loadNextSweep_run())
      !> get information on the current sweep
      call currentSweep_run(type=type, ncall=ncall, nit=nit)
      select case (type)
        case(1)
          write(iout,fmt_sweep) "warmup", ncall, nit
        case(2)
          write(iout,fmt_sweep) "production", ncall, nit
        case default
          stop 'digest_run: something went wrong'
      end select
    end do

    if (smin_run < 1d0) then
      write(iout,fmt_blank)
      write(iout,fmt_str) "> Warnings <"
      write(iout,fmt_blank)
      write(iout,fmt_opt_dble) "smin", smin_run
    endif

    !> @todo: warning if tcut_run too small/large


!    if (smin_run < 1d0) then
!       print*, "WARNING"
!       print*, "the smallest partonic centre-of-mass energy (squared)"
!       print*, "determined from the selectors is smaller than 1 GeV:"
!       print*, "smin_run = ", smin_run
!    ! print*, "will use y0*s_proc instead:"
!    ! smin_run = y0*s_proc
!    ! print*, "smin_run = ", smin_run
!    else
!       print*, "smin_run = ", smin_run
!    endif

    contains

    function flag_on_off(flag) result(str)
      use StrHelper_mod
      logical, intent(in) :: flag
      character(:), allocatable :: str
      if (flag) then
        str = "ON"
      else
        str = "OFF"
      end if
    end function flag_on_off

  end subroutine digest_run


end module Run_mod

