!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: manage coefficients of log(muR,muF) [external members: `_sclc`]
!
! DESCRIPTION:
!> Module to facilitate the management of coefficients of scale logs
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module ScaleCoeff_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_sclc, extract_sclc
  public :: setActive_sclc, isActive_sclc

  !> public gtters
  public :: nscl_sclc, ndim_sclc, idmap_sclc
  public :: facR_sclc, facF_sclc

  !> flag to switch on/off the scale-coefficient decomposition
  logical :: qactive = .false.  ! default off
  !> flag to switch on/off reduction into a minimal set
  logical :: qreduce = .true.  ! default on

  !> const data structures for the coefficient extraction
  integer, parameter :: ndim_sclc = 6  ! at NNLO (& nSubR_scl == 1) need 6 points
  integer, parameter :: nscl_sclc = ndim_sclc + 1  ! +1 for closure test

  !> the vector facR & facF factors
  double precision, dimension(nscl_sclc), parameter :: vec_facR =  &
    & (/ 1d0 , exp(+1d0) , 1d0       , exp(+1d0) , exp(-1d0) , 1d0       , 0.5d0 /)
  double precision, dimension(nscl_sclc), parameter :: vec_facF =  &
    & (/ 1d0 , 1d0       , exp(+1d0) , exp(+1d0) , 1d0       , exp(-1d0) , 2.0d0 /)

  !> mappings
  integer, dimension(nscl_sclc), parameter :: idmap_muR_sclc = (/ 1, 2, 1, 2, 5, 1, 7 /)
  integer, dimension(nscl_sclc), parameter :: idmap_muF_sclc = (/ 1, 1, 3, 3, 1, 6, 7 /)

  !> need to transpose because of the fortran ordering
  double precision, dimension(ndim_sclc,ndim_sclc), parameter :: mat_scale2coeff =  &
    &  transpose(reshape(  &
    &    (/ 1d0      , 0d0      , 0d0      , 0d0     , 0d0      , 0d0      ,  &
    &       0d0      , 1d0/4d0  , 0d0      , 0d0     , -1d0/4d0 , 0d0      ,  &
    &       0d0      , 0d0      , 1d0/4d0  , 0d0     , 0d0      , -1d0/4d0 ,  &
    &       -1d0/4d0 , 1d0/8d0  , 0d0      , 0d0     , 1d0/8d0  , 0d0      ,  &
    &       1d0/4d0  , -1d0/4d0 , -1d0/4d0 , 1d0/4d0 , 0d0      , 0d0      ,  &
    &       -1d0/4d0 , 0d0      , 1d0/8d0  , 0d0     , 0d0      , 1d0/8d0  /),  &
    &    (/ndim_sclc,ndim_sclc/) ))
  double precision, dimension(ndim_sclc,ndim_sclc), parameter :: mat_coeff2scale =  &
    &  transpose(reshape(  &
    &    (/ 1d0 , 0d0  , 0d0  , 0d0 , 0d0 , 0d0 ,  &
    &       1d0 , 2d0  , 0d0  , 4d0 , 0d0 , 0d0 ,  &
    &       1d0 , 0d0  , 2d0  , 0d0 , 0d0 , 4d0 ,  &
    &       1d0 , 2d0  , 2d0  , 4d0 , 4d0 , 4d0 ,  &
    &       1d0 , -2d0 , 0d0  , 4d0 , 0d0 , 0d0 ,  &
    &       1d0 , 0d0  , -2d0 , 0d0 , 0d0 , 4d0 /),  &
    &    (/ndim_sclc,ndim_sclc/) ))

  !> actually: NO! we don't have catani-subtracted (IR finite) pieces at NLO (both `V` & `TNLO`)
  !> so we can have higher powers in LR:  LR^|deepest-eps-pole|
  ! !> at NLO, we can look at a reduced set of equations
  ! integer, parameter :: ndim_sclc_NLO = 3  ! at NLO (& nSubR_scl == 1) need 3 points
  ! double precision, dimension(ndim_sclc_NLO,ndim_sclc_NLO), parameter :: mat_scale2coeff_NLO =  &
  !   &  transpose(reshape(  &
  !   &    (/ 1d0      , 0d0      , 0d0      ,  &
  !   &       -1d0/2d0 , 1d0/2d0  , 0d0      ,  &
  !   &       -1d0/2d0 , 0d0      , 1d0/2d0  /),  &
  !   &    (/ndim_sclc_NLO,ndim_sclc_NLO/) ))


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> setter routine to (de-)activate use of scale coefficients
  !
  !> @param[in] active boolean flag
  !-----------------------------------------------------------------------------
  subroutine setActive_sclc(active)
    logical, intent(in) :: active
    qactive = active
    if (qactive) then
      print*, "setActive_sclc: turning on scale-coefficient recosntruction..."
    end if
  end subroutine setActive_sclc


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter routine to inquire status on use of scale coefficients
  !
  !> @return active boolean flag
  !-----------------------------------------------------------------------------
  logical function isActive_sclc()
    isActive_sclc = qactive
  end function isActive_sclc


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the renormalisation scale multiplier
  !
  !> @param[in] iscl index of the scale entry
  !> @return scale multiplier for the coefficient recosntruction
  !-----------------------------------------------------------------------------
  double precision function facR_sclc(iscl)
    integer, intent(in) :: iscl
    if ( iscl<1 .or. iscl>nscl_sclc ) then
      print*, "facR_sclc: index out of bounds: ", iscl
      stop
    end if
    facR_sclc = vec_facR(iscl)
  end function facR_sclc


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the factorisation scale multiplier
  !
  !> @param[in] iscl index of the scale entry
  !> @return scale multiplier for the coefficient recosntruction
  !-----------------------------------------------------------------------------
  double precision function facF_sclc(iscl)
    integer, intent(in) :: iscl
    if ( iscl<1 .or. iscl>nscl_sclc ) then
      print*, "facF_sclc: index out of bounds: ", iscl
      stop
    end if
    facF_sclc = vec_facF(iscl)
  end function facF_sclc


  !-----------------------------------------------------------------------------
  !> @brief
  !> map to the index of the first occurrence
  !
  !> @param[in] iscl index for a scale entry [1,nAct_scl]
  !> @return index mapping to the first occurence of the scale
  !-----------------------------------------------------------------------------
  integer function idmap_sclc(iscl, withR, withF)
    integer, intent(in) :: iscl
    logical, optional, intent(in) :: withR, withF
    logical :: qR, qF
    integer :: iR, iF

    if ( iscl<1 .or. iscl>nscl_sclc ) then
      print*, "idmap_sclc: index out of bounds: ", iscl
      stop
    end if

    qR = .false.
    if (present(withR)) qR = withR

    qF = .false.
    if (present(withF)) qF = withF

    iR = iscl
    if (qR) iR = idmap_muR_sclc(iscl)

    iF = iscl
    if (qF) iF = idmap_muF_sclc(iscl)

    if (qR.and.qF) then
      idmap_sclc = iscl
      if (iR==iF) idmap_sclc = iR
    else if (qR) then
      idmap_sclc = iR
    else if (qF) then
      idmap_sclc = iF
    else
      idmap_sclc = iscl
    end if

  end function idmap_sclc


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise scale-coefficient module
  !-----------------------------------------------------------------------------
  subroutine init_sclc()

    print*, "init_sclc"

    print*, matmul( mat_scale2coeff, mat_coeff2scale )
    print*, matmul( mat_coeff2scale, mat_scale2coeff )


    ! call extract_sclc( (/ 2d0, 36d0, 64d0, 142d0, 24d0, 44d0 /) )

    stop "init_sclc: die"
  end subroutine init_sclc



  !>
  !> scale decomposition:
  !>
  !>    wt ~ A00
  !>       + B10 * log(muR^2/muR0^2)
  !>       + B01 * log(muF^2/muF0^2)
  !>       + C20 * log(muR^2/muR0^2)^2
  !>       + C11 * log(muR^2/muR0^2) * log(muF^2/muF0^2)
  !>       + C02 * log(muF^2/muF0^2)^2
  !>
  !>   * A00 = coeffs(1)
  !>   * B10 = coeffs(2)
  !>   * B01 = coeffs(3)
  !>   * C20 = coeffs(4)
  !>   * C11 = coeffs(5)
  !>   * C02 = coeffs(6)

  subroutine extract_sclc(vec, coeffs, nV)
    double precision, dimension(:), intent(in) :: vec
    double precision, dimension(ndim_sclc), intent(out) :: coeffs
    integer, intent(in), optional :: nV
#ifndef NDEBUG
    double precision, dimension(ndim_sclc) :: logs
    double precision :: check
#endif

    coeffs = 0d0
    if (qreduce .and. present(nV)) then
      ! select case (nV)
      ! case (0)
      !   coeffs(1) = vec(1)
      ! case (1)
      !   coeffs(1:ndim_sclc_NLO) = matmul( mat_scale2coeff_NLO, vec(1:ndim_sclc_NLO) )
      ! case (2)
      !   coeffs = matmul( mat_scale2coeff, vec(1:ndim_sclc) )
      ! case default
      !   print*, "extract_sclc: invalid nV:", nV
      !   stop
      ! end select
      if (nV == 0) then
        coeffs(1) = vec(1)
      else
        coeffs = matmul( mat_scale2coeff, vec(1:ndim_sclc) )
      end if
    else
      coeffs = matmul( mat_scale2coeff, vec(1:ndim_sclc) )
    end if

#ifndef NDEBUG
    logs(:) = 0d0
    logs(1) = 1d0                            ! x A00
    logs(2) = 2D0*log(facR_sclc(nscl_sclc))  ! x B10
    logs(3) = 2D0*log(facF_sclc(nscl_sclc))  ! x B01
    logs(4) = logs(2)**2                     ! x C20
    logs(5) = logs(2)*logs(3)                ! x C11
    logs(6) = logs(3)**2                     ! x C02
    check = dot_product(logs, coeffs)

    if ( abs( (check-vec(nscl_sclc)) / (check+vec(nscl_sclc)) ) > 1d-6 ) then
      print*, "extract_sclc: closure check failed:"
      print*, "vec", vec(1:ndim_sclc)
      print*, "---"
      print*, "A00:", coeffs(1), "*", logs(1)
      print*, "B10:", coeffs(2), "*", logs(2)
      print*, "B01:", coeffs(3), "*", logs(3)
      print*, "C20:", coeffs(4), "*", logs(4)
      print*, "C11:", coeffs(5), "*", logs(5)
      print*, "C02:", coeffs(6), "*", logs(6)
      print*, "---"
      print*, "CHK", vec(nscl_sclc), check
      ! read(*,*)
    end if
#endif

  end subroutine extract_sclc


end module ScaleCoeff_mod


