!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Scales [external members: `_scl`]
!
! DESCRIPTION:
!> collcet everything that has to do with scales
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Scales_mod
!-------------------------------------------------------------------------------
  use ScaleCoeff_mod
  implicit none
  private

  !>----- routines & functions

  public :: init_scl, destroy_scl
  !public :: registerMuR_scl, registerMuF_scl
  public :: registerPair_scl, nAct_scl
  public :: digest_scl!, purge_scl
  public :: resetInputMuR_scl, resetInputMuF_scl
  public :: setMu_scl, unsetMu_scl
  public :: setMuR_scl, setMuF_scl
  ! needed? public :: getMuR_scl, getMuF_scl
  public :: update_scl
  !> access to the PDFs
  public :: updatePDF_scl
  public :: get_lumi_scl
  public :: f1_scl, f2_scl
  public :: f1x1_scl, f2x2_scl
  !> access to the alphas
  public :: updateAlphaS_scl
  public :: alphassub_scl, alphas_scl
  !> directly get alphas array for new scales
  public :: get_alphas_scl
  public :: get_muR2_scl, get_muF2_scl

  !>----- variables

  public :: Scale_t
  public :: nMax_scl
  public :: nInp_scl
  public :: nSubR_scl, nSubF_scl  ! # sub-scales
  public :: nMuR_scl, nMuF_scl    ! = # *distinct* scales

  !>----- external interface

  public :: muR_scl, muR2_scl, facR_scl  ! = renscale2
  public :: muF_scl, muF2_scl, facF_scl  ! = facscale2
  public :: muA_scl, muA2_scl  ! = fragscale2 (== facscale2[isub=2] )
  public :: idmap_scl  ! a wrapper for muR & muF

  !>----- MiNLO access, eventually needs to be solved better
  public :: input_muRred_scl, input_muFred_scl
  public :: active_muRred_scl, active_muFred_scl
  public :: map_muRsub_to_red_scl, map_muFsub_to_red_scl


  type :: Scale_t
    !> constant  scale: obsId=0  &  fac ~ absolute value [GeV]
    !> dynamical scale: obdId>0  &  fac ~ scale factor
    double precision :: fac = 0d0
    integer :: obsId = 0
  end type Scale_t

  !> we need a static upper bound for the non-dynamic part of the driver code
  !> (mostly the auto-generated bits)
  integer, parameter :: nMax_scl = 7

  !> # of active scale-paris (muR,muF)
  integer :: nInp_scl = 0  ! # of input scales  ( = nAct_scl when ScaleCoeff off)
  integer :: nAct_scl = 0  ! # of active scales
  !> scale arrays
  integer :: nSubR_scl = -1
  integer :: nSubF_scl = -1
  !> # of *distinct* scales
  integer :: nMuR_scl = 0
  integer :: nMuF_scl = 0


  !> input scales
  type(Scale_t), allocatable, dimension(:), target :: input_muRred_scl, input_muFred_scl
  !> the active scale can be overridden (MultiRun)
  type(Scale_t), allocatable, dimension(:), target :: active_muRred_scl, active_muFred_scl
  !> a global prefactor to muR (used for warmup sampling)
  double precision :: global_facR_scl
!$OMP THREADPRIVATE(active_muRred_scl, active_muFred_scl, global_facR_scl)

  !> the reduced values in [GeV] & [GeV^2] for *distinct* scales
  double precision, allocatable, dimension(:) :: muRred_scl, muR2red_scl
  double precision, allocatable, dimension(:) :: muFred_scl, muF2red_scl
  logical, allocatable, dimension(:) :: valid_muRred_scl, valid_muFred_scl
!$OMP THREADPRIVATE(muRred_scl,muR2red_scl,valid_muRred_scl, muFred_scl,muF2red_scl,valid_muFred_scl)


  !> mappings:
  !>     muR: [1,nAct_scl][1:nSubR_scl] --> [1,nMuR_scl]
  !>     muF: [1,nAct_scl][1:nSubF_scl] --> [1,nMuF_scl]
  integer, allocatable, dimension(:,:) :: map_muRsub_to_red_scl, map_muFsub_to_red_scl

  !> map to *first* occurence of the distinct scale (array!)
  !>     [1,nAct_scl] --> [1,nAct_scl]
  !>     idmapMuX(i) = i   (first occurence)
  !>     idmapMuX(i) < i   (maps to first occurence)
  integer, dimension(nMax_scl) :: idmap_muR_scl, idmap_muF_scl


  !> PDF storage
  double precision, dimension(-6:7,nMax_scl), target :: f1_scl, f2_scl, f1x1_scl, f2x2_scl
!$OMP THREADPRIVATE(f1_scl, f2_scl, f1x1_scl, f2x2_scl)

  !> alphas storage (alphas = g2s/(4 pi))
  double precision, allocatable, dimension(:,:), target :: alphassub_scl
  double precision, dimension(:), pointer :: alphas_scl
!$OMP THREADPRIVATE(alphassub_scl, alphas_scl)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialize scales module
  !-----------------------------------------------------------------------------
  subroutine init_scl()
    use Run_mod, only : qproductionSweep_run

    if (nInp_scl < 1) then
      print*, "init_scl: forgot to register scales?!"
      stop
    end if

    !> set nAct_scl
    if ( isActive_sclc() ) then
      !> scale arrays not supported for the moment
      if ( nSubR_scl/=1 .or. nSubF_scl/=1 ) then
        print*, "init_scl: scale arrays together with scale coefficients not supported!"
        stop
      end if
      !> override
      nAct_scl = nscl_sclc
    else
      nAct_scl = nInp_scl
    end if

    !> if only warmup run, kill all additional scales & scale coefficients
    if ( .not.qproductionSweep_run .and. nAct_scl>1 ) then
      print*, "init_scl: only warmup run -> switch off all but the central scale"
      nInp_scl=1
      nAct_scl=1
      !> this is actually WRONG when using scale-arrays:
      ! nMuR_scl=1
      ! nMuF_scl=1
      call setActive_sclc(.false.)
    end if

    !> initialise threadloacl variables once outside the parallel block
    !> getMu[R|F]_scl needs this to work
    call setMu_scl()

  end subroutine init_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> Initialize storage for the scales
  !-----------------------------------------------------------------------------
  subroutine setup_scl(nsubR, nsubF)
    use PhotonFragmentation_mod
    use HadronFragmentation_mod
    integer, intent(in), optional :: nsubR, nsubF

    call destroy_scl()  ! make sure everything is deallocated

    !> begin with the largest possible size
    !> will be trimmed later to appropriate sizes --> purge_scl
    ! allocate(inputMuR(nMax_scl))
    ! ...

    if ( present(nsubR) ) then
      nSubR_scl = nsubR
    else
      nSubR_scl = 1
    end if
    allocate( map_muRsub_to_red_scl(nMax_scl,nSubR_scl) )
    map_muRsub_to_red_scl(:,:) = -1
    idmap_muR_scl(:)        = -1

    if ( present(nsubF) ) then
      nSubF_scl = nsubF
    else
      nSubF_scl = 1
    end if
    allocate( map_muFsub_to_red_scl(nMax_scl,nSubF_scl) )
    map_muFsub_to_red_scl(:,:) = -1
    idmap_muF_scl(:)        = -1

    if (.not.isActive_hfrag() .and. .not.isActive_frag() .and. nSubF_scl /= 1) then
      print*, "setup_scl: nSubF_scl must be 1 (reserved for muA = fragmentaiton scale)", nSubF_scl
      stop
    end if

    nInp_scl = 0
    nAct_scl = 0
    nMuR_scl = 0
    nMuF_scl = 0

  end subroutine setup_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> Deallocate all dynamic arrays
  !-----------------------------------------------------------------------------
  subroutine destroy_scl()
    !> hm... maybe switch to dynamic allocation with a purge_scl routine?

    if (allocated(input_muRred_scl)) deallocate(input_muRred_scl)
    if (allocated(input_muFred_scl)) deallocate(input_muFred_scl)

    if (allocated(map_muRsub_to_red_scl)) deallocate(map_muRsub_to_red_scl)
    if (allocated(map_muFsub_to_red_scl)) deallocate(map_muFsub_to_red_scl)

    nSubR_scl = -1
    nSubF_scl = -1

  end subroutine destroy_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a renormalization scale (we don't care about performance)
  !
  !> @param[in] muR the renormalization scale to be registered
  !> @return index of the array input_muRred_scl pointing to this scale
  !-----------------------------------------------------------------------------
  integer function registerMuR_scl(muR)
    type(Scale_t), intent(in) :: muR
    integer :: i
    type(Scale_t), allocatable, dimension(:) :: reduced_tmp

    !> look if we already registered the scale
    do i=1,nMuR_scl
      if ( input_muRred_scl(i)%fac   == muR%fac .and.  &
         & input_muRred_scl(i)%obsId == muR%obsId ) then
        registerMuR_scl = i
        return
      end if
    end do

    !> no match: create a new entry
    if ( nMuR_scl == 0 ) then  ! first entry
      allocate( input_muRred_scl(1) )
    else  ! already allocated
      allocate( reduced_tmp(nMuR_scl) )
      reduced_tmp = input_muRred_scl
      deallocate(input_muRred_scl)
      allocate( input_muRred_scl(nMuR_scl+1) )
      input_muRred_scl(1:nMuR_scl) = reduced_tmp
      deallocate(reduced_tmp)
    end if

    !> and copy
    nMuR_scl = nMuR_scl + 1
    input_muRred_scl(nMuR_scl)%fac   = muR%fac
    input_muRred_scl(nMuR_scl)%obsId = muR%obsId
    registerMuR_scl = nMuR_scl

  end function registerMuR_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a factorization scale (we don't care about performance)
  !
  !> @param[in] muF the factorization scale to be registered
  !> @return index of the array input_muFred_scl pointing to this scale
  !-----------------------------------------------------------------------------
  integer function registerMuF_scl(muF)
    type(Scale_t), intent(in) :: muF
    integer :: i
    type(Scale_t), allocatable, dimension(:) :: reduced_tmp

    !> look if we already registered the scale
    do i=1,nMuF_scl
      if ( input_muFred_scl(i)%fac   == muF%fac .and.  &
         & input_muFred_scl(i)%obsId == muF%obsId ) then
        registerMuF_scl = i
        return
      end if
    end do

    !> no match: create a new entry
    if ( nMuF_scl == 0 ) then  ! first entry
      allocate( input_muFred_scl(1) )
    else  ! already allocated
      allocate( reduced_tmp(nMuF_scl) )
      reduced_tmp = input_muFred_scl
      deallocate(input_muFred_scl)
      allocate( input_muFred_scl(nMuF_scl+1) )
      input_muFred_scl(1:nMuF_scl) = reduced_tmp
      deallocate(reduced_tmp)
    end if

    !> and copy
    nMuF_scl = nMuF_scl + 1
    input_muFred_scl(nMuF_scl)%fac   = muF%fac
    input_muFred_scl(nMuF_scl)%obsId = muF%obsId
    registerMuF_scl = nMuF_scl

  end function registerMuF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a pair of scale arrays ([muR1,muR2,...],[muF1,muF2,...])
  !
  !> @param[in] arrayR
  !> @param[in] arrayF
  !-----------------------------------------------------------------------------
  subroutine registerPair_scl(arrayR, arrayF)
    type(Scale_t), dimension(:), intent(in) :: arrayR, arrayF
    integer :: i

    !> init and compatibility check
    if ( (nSubR_scl < 0) .and. (nSubF_scl < 0) ) then
      call setup_scl(nsubR=size(arrayR), nsubF=size(arrayF))
    else
      if ( (size(arrayR) /= nSubR_scl) .or.  &
         & (size(arrayF) /= nSubF_scl) ) then
        print*, "registerPair_scl: array size mismatch!"
        stop
      end if
    end if

    !> sanity check
    if ( any(arrayR(:)%obsId<0) .or. any(arrayF(:)%obsId<0) ) then
      print*, "registerPair_scl: found invalid obsId's: "
      print*, "arrayR: ", arrayR(:)%obsId
      print*, "arrayF: ", arrayF(:)%obsId
      stop
    end if
    if ( .not.isActive_sclc() .and. nInp_scl==nMax_scl) then
      print*, "registerPair_scl: no slot left to register a new scale pair [nMax_scl]"
      stop
    end if

    !> so far: restrict to only [1|2] for muF scales
    !> implicit assumptions used below (sub = 1  <--> muA scale)
    print*, "nSubF_scl = ", nSubF_scl
    if ( (nSubF_scl/=1) .and. (nSubF_scl/=2) ) then
      print*, "registerPair_scl: nSubF_scl must be [1|2]!"
      stop
    end if

    nInp_scl = nInp_scl + 1

    !> set up renormalization scale
    do i=1,nSubR_scl
      map_muRsub_to_red_scl(nInp_scl,i) = registerMuR_scl(arrayR(i))
    end do
    do i=1,nInp_scl
      if ( all(map_muRsub_to_red_scl(nInp_scl,:)==map_muRsub_to_red_scl(i,:)) ) then
        idmap_muR_scl(nInp_scl) = i
        exit
      end if
    end do

    !> set up factorization scale
    do i=1,nSubF_scl
      map_muFsub_to_red_scl(nInp_scl,i) = registerMuF_scl(arrayF(i))
    end do
    do i=1,nInp_scl
      if ( all(map_muFsub_to_red_scl(nInp_scl,:)==map_muFsub_to_red_scl(i,:)) ) then
        idmap_muF_scl(nInp_scl) = i
        exit
      end if
    end do

  end subroutine registerPair_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset renormalisation scales
  !
  !> @param[in] fac
  !> @param[in] obsId
  !-----------------------------------------------------------------------------
  subroutine resetInputMuR_scl(fac, obsId)
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i
    !> info on the input central scale
    double precision :: fac0
    integer :: obsId0

    !> get muR for the central scale
    fac0   = input_muRred_scl(1)%fac
    obsId0 = input_muRred_scl(1)%obsId
    do i=1,nMuR_scl
      !> swap out scales only when obsId same as central scale
      if ( input_muRred_scl(i)%obsId == obsId0 ) then
        if ( present(obsId) )  &
        &  input_muRred_scl(i)%obsId = obsId
        if ( present(fac) )  &
        &  input_muRred_scl(i)%fac   = (fac/fac0) * input_muRred_scl(i)%fac
      else
        print*, "resetInputMuR_scl: obsId mismatch for imuR = ", i
        stop
      end if
    end do

  end subroutine resetInputMuR_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> reset factorization scales
  !
  !> @param[in] fac
  !> @param[in] obsId
  !-----------------------------------------------------------------------------
  subroutine resetInputMuF_scl(fac, obsId)
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i
    !> info on the input central scale
    double precision :: fac0
    integer :: obsId0

    !> get muR for the central scale
    fac0   = input_muFred_scl(1)%fac
    obsId0 = input_muFred_scl(1)%obsId
    do i=1,nMuF_scl
      !> swap out scales only when obsId same as central scale
      if ( input_muFred_scl(i)%obsId == obsId0 ) then
        if ( present(obsId) )  &
        &  input_muFred_scl(i)%obsId = obsId
        if ( present(fac) )  &
        &  input_muFred_scl(i)%fac   = (fac/fac0) * input_muFred_scl(i)%fac
      else
        print*, "resetInputMuF_scl: obsId mismatch for imuF = ", i
        stop
      end if
    end do

  end subroutine resetInputMuF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update threadlocal scale-settings to the input-scales
  !-----------------------------------------------------------------------------
  subroutine setMu_scl()
    call setMuR_scl()
    call setMuF_scl()
  end subroutine setMu_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> de-allocate threadlocal storage
  !-----------------------------------------------------------------------------
  subroutine unsetMu_scl()
    !> muR
    if (allocated(alphassub_scl))     deallocate(alphassub_scl)
    if (allocated(muRred_scl))        deallocate(muRred_scl)
    if (allocated(muR2red_scl))       deallocate(muR2red_scl)
    if (allocated(valid_muRred_scl))  deallocate(valid_muRred_scl)
    if (allocated(active_muRred_scl)) deallocate(active_muRred_scl)
    !> muF
    if (allocated(muFred_scl))        deallocate(muFred_scl)
    if (allocated(muF2red_scl))       deallocate(muF2red_scl)
    if (allocated(valid_muFred_scl))  deallocate(valid_muFred_scl)
    if (allocated(active_muFred_scl)) deallocate(active_muFred_scl)
  end subroutine unsetMu_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update / override (MultiRun) local renormalization scales
  !
  !> @param[in] fac
  !> @param[in] obsId
  !> @param[in] global_fac
  !-----------------------------------------------------------------------------
  subroutine setMuR_scl(fac, obsId, global_fac)
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    double precision, optional, intent(in) :: global_fac
    integer :: i
    !> info on the input central scale
    double precision :: fac0
    integer :: obsId0

    !> for muR we also need to allocate the strong coupling cache
    if (.not.allocated(alphassub_scl)) then
      allocate(alphassub_scl(nMax_scl,nSubR_scl))
      alphassub_scl(:,:) = 0d0
      alphas_scl => alphassub_scl(:,1)  ! legacy compatibility
    end if

    !> allocate nMuR_scl dependent arrays
    if (.not.allocated(muRred_scl)) then
      allocate(muRred_scl(nMuR_scl))
      muRred_scl(:)  = 0d0
    end if
    if (.not.allocated(muR2red_scl)) then
      allocate(muR2red_scl(nMuR_scl))
      muR2red_scl(:) = 0d0
    end if
    if (.not.allocated(valid_muRred_scl)) then
      allocate(valid_muRred_scl(nMuR_scl))
      valid_muRred_scl(:) = .false.
    end if
    if (.not.allocated(active_muRred_scl)) then
      allocate(active_muRred_scl(nMuR_scl))
    end if

    !> set the global muR sampling factor
    global_facR_scl = 1d0
    if ( present(global_fac) ) global_facR_scl = global_fac

    !> get muR for the central scale
    fac0   = input_muRred_scl(1)%fac
    obsId0 = input_muRred_scl(1)%obsId
    do i=1,nMuR_scl
      !> default: use the input scales (+ possible global muR factor)
      active_muRred_scl(i)%fac   = global_facR_scl * input_muRred_scl(i)%fac
      active_muRred_scl(i)%obsId = input_muRred_scl(i)%obsId
      !> swap out scales only when obsId same as central scale
      if ( input_muRred_scl(i)%obsId == obsId0 ) then
        if ( present(fac) )   active_muRred_scl(i)%fac   = (fac/fac0) * active_muRred_scl(i)%fac
        if ( present(obsId) ) active_muRred_scl(i)%obsId = obsId
      end if
    end do

  end subroutine setMuR_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update / override (MultiRun) local factorization scales
  !
  !> @param[in] fac
  !> @param[in] obsId
  !-----------------------------------------------------------------------------
  subroutine setMuF_scl(fac, obsId)
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i
    !> info on the input central scale
    double precision :: fac0
    integer :: obsId0

    !> allocate nMuF_scl dependent arrays
    if (.not.allocated(muFred_scl)) then
      allocate(muFred_scl(nMuF_scl))
      muFred_scl(:)  = 0d0
    end if
    if (.not.allocated(muF2red_scl)) then
      allocate(muF2red_scl(nMuF_scl))
      muF2red_scl(:) = 0d0
    end if
    if (.not.allocated(valid_muFred_scl)) then
      allocate(valid_muFred_scl(nMuF_scl))
      valid_muFred_scl(:) = .false.
    end if
    if (.not.allocated(active_muFred_scl)) allocate(active_muFred_scl(nMuF_scl))

    !> get muF for the central scale
    fac0   = input_muFred_scl(1)%fac
    obsId0 = input_muFred_scl(1)%obsId
    do i=1,nMuF_scl
      !> default: use the input scales
      active_muFred_scl(i) = input_muFred_scl(i)
      !> swap out scales only when obsId same as central scale
      if ( input_muFred_scl(i)%obsId == obsId0 ) then
        if ( present(fac) )   active_muFred_scl(i)%fac   = (fac/fac0) * active_muFred_scl(i)%fac
        if ( present(obsId) ) active_muFred_scl(i)%obsId = obsId
      end if
    end do

  end subroutine setMuF_scl


  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> get the renormalisation scale
  ! !
  ! !> @param[in] iscl index of the scales pair
  ! !> @param[in] isub optional index of the scale array
  ! !-----------------------------------------------------------------------------
  ! type(Scale_t) function getMuR_scl(iscl, isub)
  !   integer, intent(in)           :: iscl
  !   integer, optional, intent(in) :: isub
  !   if (present(isub)) then
  !     getMuR_scl = active_muRred_scl(map_muRsub_to_red_scl(iscl,isub))
  !   else
  !     getMuR_scl = active_muRred_scl(map_muRsub_to_red_scl(iscl,1))
  !   end if
  ! end function getMuR_scl


  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> get the factorisation scale
  ! !
  ! !> @param[in] iscl index of the scales pair
  ! !> @param[in] isub optional index of the scale array
  ! !-----------------------------------------------------------------------------
  ! type(Scale_t) function getMuF_scl(iscl, isub)
  !   integer, intent(in)           :: iscl
  !   integer, optional, intent(in) :: isub
  !   if (present(isub)) then
  !     getMuF_scl = active_muFred_scl(map_muFsub_to_red_scl(iscl,isub))
  !   else
  !     getMuF_scl = active_muFred_scl(map_muFsub_to_red_scl(iscl,1))
  !   end if
  ! end function getMuF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the renormalisation scale [GeV]
  !
  !> @param[in] iscl index of the scales pair
  !> @param[in] isub optional index of the scale array
  !-----------------------------------------------------------------------------
  double precision function muR_scl(iscl, isub)
    integer, intent(in)           :: iscl
    integer, optional, intent(in) :: isub
    if (isActive_sclc()) then
      muR_scl = facR_sclc(iscl) * muRred_scl(1)  ! central scale
      return
    end if
    if (present(isub)) then
      muR_scl = muRred_scl(map_muRsub_to_red_scl(iscl,isub))
    else
      muR_scl = muRred_scl(map_muRsub_to_red_scl(iscl,1))
    end if
  end function muR_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the renormalisation scale squared [GeV^2]
  !
  !> @param[in] iscl index of the scales pair
  !> @param[in] isub optional index of the scale array
  !-----------------------------------------------------------------------------
  double precision function muR2_scl(iscl, isub)
    integer, intent(in)           :: iscl
    integer, optional, intent(in) :: isub
    if (isActive_sclc()) then
      muR2_scl = facR_sclc(iscl)**2 * muR2red_scl(1)  ! central scale
      return
    end if
    if (present(isub)) then
      muR2_scl = muR2red_scl(map_muRsub_to_red_scl(iscl,isub))
    else
      muR2_scl = muR2red_scl(map_muRsub_to_red_scl(iscl,1))
    end if
  end function muR2_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the relative factor w.r.t. the central scale (0 if obs different)
  !
  !> @param[in] iscl index of the scales pair
  !-----------------------------------------------------------------------------
  double precision function facR_scl(iscl)
    integer, intent(in) :: iscl
    integer :: iR

#ifndef NDEBUG
    if (nSubR_scl > 1) stop "facR_scl: scale arrays unsupported"
    if (iscl < 1 .or. iscl > nInp_scl) stop "facR_scl: index out of bound"
#endif

    iR = map_muRsub_to_red_scl(iscl,1)

    if (active_muRred_scl(iR)%obsId == active_muRred_scl(1)%obsId) then
      facR_scl = active_muRred_scl(iR)%fac / active_muRred_scl(1)%fac
    else
      facR_scl = 0d0
    end if

  end function facR_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the factorisation scale [GeV]
  !
  !> @param[in] iscl index of the scales pair
  !> @param[in] isub optional index of the scale array
  !-----------------------------------------------------------------------------
  double precision function muF_scl(iscl, isub)
    integer, intent(in)           :: iscl
    integer, optional, intent(in) :: isub
    if (isActive_sclc()) then
      muF_scl = facF_sclc(iscl) * muFred_scl(1)  ! central scale
      return
    end if
    if (present(isub)) then
      muF_scl = muFred_scl(map_muFsub_to_red_scl(iscl,isub))
    else
      muF_scl = muFred_scl(map_muFsub_to_red_scl(iscl,1))
    end if
  end function muF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the factorisation scale squared [GeV^2]
  !
  !> @param[in] iscl index of the scales pair
  !> @param[in] isub optional index of the scale array
  !-----------------------------------------------------------------------------
  double precision function muF2_scl(iscl, isub)
    integer, intent(in)           :: iscl
    integer, optional, intent(in) :: isub
    if (isActive_sclc()) then
      muF2_scl = facF_sclc(iscl)**2 * muF2red_scl(1)  ! central scale
      return
    end if
    if (present(isub)) then
      muF2_scl = muF2red_scl(map_muFsub_to_red_scl(iscl,isub))
    else
      muF2_scl = muF2red_scl(map_muFsub_to_red_scl(iscl,1))
    end if
  end function muF2_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the relative factor w.r.t. the central scale (0 if obs different)
  !
  !> @param[in] iscl index of the scales pair
  !-----------------------------------------------------------------------------
  double precision function facF_scl(iscl)
    integer, intent(in) :: iscl
    integer :: iF

#ifndef NDEBUG
    if (nSubF_scl > 1) stop "facF_scl: scale arrays unsupported"
    if (iscl < 1 .or. iscl > nInp_scl) stop "facF_scl: index out of bound"
#endif

    iF = map_muFsub_to_red_scl(iscl,1)

    if (active_muFred_scl(iF)%obsId == active_muFred_scl(1)%obsId) then
      facF_scl = active_muFred_scl(iF)%fac / active_muFred_scl(1)%fac
    else
      facF_scl = 0d0
    end if

  end function facF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the fragmentation scale [GeV]
  !
  !> @param[in] iscl index of the scales pair
  !-----------------------------------------------------------------------------
  double precision function muA_scl(iscl)
    integer, intent(in) :: iscl
    !> we're making a dirty assumption here!
    !> nSubF_scl is either 1 or 2; in case it's 2, muA is in slot 2
    !> if not, we by default map the fragmentation scale to muF
    muA_scl = muF_scl(iscl, nSubF_scl)
  end function muA_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the fragmentation scale squared [GeV^2]
  !
  !> @param[in] iscl index of the scales pair
  !-----------------------------------------------------------------------------
  double precision function muA2_scl(iscl)
    integer, intent(in) :: iscl
    !> we're making a dirty assumption here!
    !> nSubF_scl is either 1 or 2; in case it's 2, muA is in slot 2
    !> if not, we by default map the fragmentation scale to muF
    muA2_scl = muF2_scl(iscl, nSubF_scl)
  end function muA2_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> map the iscl-th entry to the first occurence
  !
  !> @param[in] iscl index for a scale entry [1,nAct_scl]
  !> @return index mapping to the first occurence of the scale
  !-----------------------------------------------------------------------------
  integer function idmap_scl(iscl, withR, withF)
    integer, intent(in) :: iscl
    logical, optional, intent(in) :: withR, withF
    logical :: qR, qF
    integer :: iR, iF

    if (isActive_sclc()) then
      idmap_scl = idmap_sclc(iscl, withR, withF)
      return
    end if

    qR = .false.
    if (present(withR)) qR = withR

    qF = .false.
    if (present(withF)) qF = withF

    iR = iscl
    if (qR) iR = idmap_muR_scl(iscl)

    iF = iscl
    if (qF) iF = idmap_muF_scl(iscl)

    if (qR.and.qF) then
      idmap_scl = iscl
      if (iR==iF) idmap_scl = iR
    else if (qR) then
      idmap_scl = iR
    else if (qF) then
      idmap_scl = iF
    else
      idmap_scl = iscl
    end if

  end function idmap_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the registered scales
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !> @param[in] qlocal optional flag to use the threadlocal settings
  !-----------------------------------------------------------------------------
  subroutine digest_scl(unit, qlocal)
    use Observables_mod
    integer, optional, intent(in) :: unit
    logical, optional, intent(in) :: qlocal
    integer :: iout
    type(Scale_t), dimension(:), pointer :: useMuR, useMuF
    integer :: i,j
    !> new scale arrays
    type(Scale_t), allocatable, dimension(:) :: muR_arr, muF_arr
    character(:), allocatable :: muR_str, muF_str
    character(len=*), parameter :: fmt_label = '(1x,"*",3x,"ID",T35,"muR",T75,"muF",T103,"*")'
    character(len=*), parameter :: fmt_scale = '(1x,"*",3x,I0,T25,A,T65,A,T103,"*")'

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    allocate(muR_arr(nSubR_scl))
    allocate(muF_arr(nSubF_scl))

    !> default: use the input settings
    useMuR => input_muRred_scl(:)
    useMuF => input_muFred_scl(:)
    if ( present(qlocal) ) then
      if (qlocal) then
        useMuR => active_muRred_scl(:)
        useMuF => active_muFred_scl(:)
      end if
    end if

    write(iout,fmt_label)
    do i=1,nInp_scl
      !> get muR string
      do j=1,nSubR_scl
        muR_arr(j)%fac   = useMuR( map_muRsub_to_red_scl(i,j) )%fac
        muR_arr(j)%obsId = useMuR( map_muRsub_to_red_scl(i,j) )%obsId
      end do
      muR_str = digest_scl_to_str(muR_arr)
      !> get muF string
      do j=1,nSubF_scl
        muF_arr(j)%fac   = useMuF( map_muFsub_to_red_scl(i,j) )%fac
        muF_arr(j)%obsId = useMuF( map_muFsub_to_red_scl(i,j) )%obsId
      end do
      muF_str = digest_scl_to_str(muF_arr)
      !>
      write(iout,fmt_scale) i, muR_str, muF_str
    end do

    deallocate(muR_arr)
    deallocate(muF_arr)

  contains

    function digest_scl_to_str(mu_arr) result (string)
      type(Scale_t), dimension(:), intent(in) :: mu_arr
      character(:), allocatable :: string, sep_str, tmp_str
      character(len=128) :: input_str
      integer :: i

      string  = ""
      sep_str = ""

      if (size(mu_arr)<1) return
      if (size(mu_arr)>1) string  = "["

      do i=1,size(mu_arr)
        if ( mu_arr(i)%obsId == 0 ) then
          write(input_str,'(ES10.3E2)') mu_arr(i)%fac
          tmp_str = trim(adjustl(input_str)) // " GeV"
        else
          write(input_str,'(ES10.3E2)') mu_arr(i)%fac
          tmp_str = trim(adjustl(input_str)) // " x " // trim(getNameFromId_obs(mu_arr(i)%obsId))
        end if
        string = string // sep_str // tmp_str
        sep_str = ", "  ! for n>1 want separators
      end do

      if (size(mu_arr)>1) string  = trim(string) // "]"

    end function digest_scl_to_str

  end subroutine digest_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the alphas array where fac & obsId are swapped out
  !> e.g. used in ScaleCoeff
  !> e.g. used in MiNLO to get underlying-Born alphas
  !
  !> @param[out] alphas nMax_scl dimensional array of alphas values
  !> @param[in] fac optional
  !> @param[in] obsId optional
  !-----------------------------------------------------------------------------
  subroutine get_alphas_scl(alphas, fac, obsId)
    use Constants_mod
    use Observables_mod
    use LHAPDF_mod
    double precision, dimension(nMax_scl), intent(out) :: alphas
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i, iR, id
    !> info on the input central scale
    logical :: valid
    double precision :: fac0, faciR, muiR, val
    integer :: obsId0, obsIdiR

#ifndef NDEBUG
    if (nSubR_scl > 1) stop "get_alphas_scl: scale arrays unsupported"
#endif

    !> no scale override: default fallback
    if ( .not.present(fac) .and. .not.present(obsId) ) then
      if (isActive_sclc()) then
        !> for coefficients in as/(2pi)
        alphas = twopi
      else
        !> no coefficients: cached values (only nSubR_scl = 1)
        alphas(:) = alphassub_scl(:,1)
      end if
      return
    end if

    alphas = 0d0

    !> get muR for the central scale
    fac0   = active_muRred_scl(1)%fac
    obsId0 = active_muRred_scl(1)%obsId

    ! print*, "get_alphas_scl:"
    ! print*, "fac ",fac,  "obsId ",obsId
    ! print*, "fac0",fac0, "obsId0",obsId0

    do i=1,nInp_scl
      !> this is where nSubR_scl == 1 enters!
      iR = map_muRsub_to_red_scl(i,1)
      ! print*, "i",i, "iR",iR
      !>----- already evolved the scale before!
      id = idmap_muR_scl(i)
      ! print*, "> id", id
      if ( i /= id ) then
        alphas(i) = alphas(id)
        cycle
      end if
      !>----- actually compute the scale
      muiR = 0d0
      obsIdiR = active_muRred_scl(iR)%obsId
      faciR   = active_muRred_scl(iR)%fac
      ! print*, "> [before] faciR",faciR, "obsIdiR",obsIdiR
      !> swap out scales only when obsId same as central scale
      if ( obsIdiR == obsId0 ) then
        if ( present(obsId) ) obsIdiR = obsId
        if ( present(fac) ) then
          faciR = fac * faciR
          if ( (obsIdiR > 0) .or. present(obsId) )  &
            &  faciR = faciR / fac0
        end if
      end if
      ! print*, "> [after]  faciR",faciR, "obsIdiR",obsIdiR
      !> set muiR
      if (obsIdiR == 0) then  ! fixed scale
        muiR = faciR
      else  ! dynamic scale
        val = getValue_obs(obsIdiR, valid)
        if (.not.valid) then
          alphas(i) = 0d0
          cycle
        end if
        muiR = faciR * val
      end if
      if ( isnan(muiR) ) then
        print*, "get_alphas_scl: muR = NaN?!"
        print*, " > obs", obsIdiR, ">", val, valid
        print*, " > fac", faciR
        call print_xvegas()
        cycle  ! alphas = zero
      end if
      !> call LHAPDF alphas routine
      alphas(i) = alphasQ_lhapdf(muiR)
      ! print*, "get_alphas_scl", i, "muR", muiR, ">", alphas(i)
    end do

  end subroutine get_alphas_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the renormalisation scale squared [GeV^2]
  !
  !> @param[out] muR2 array of the scales
  !> @param[in] fac optional
  !> @param[in] obsId optional
  !-----------------------------------------------------------------------------
  subroutine get_muR2_scl(muR2, fac, obsId)
    use Observables_mod
    double precision, dimension(nMax_scl), intent(out) :: muR2
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i, iR, id
    !> info on the input central scale
    logical :: valid
    double precision :: fac0, faciR, muiR, val
    integer :: obsId0, obsIdiR

#ifndef NDEBUG
    if (nSubR_scl > 1) stop "get_muR2_scl: scale arrays unsupported"
#endif

    muR2 = 0d0

    if ( .not.present(fac) .and. .not.present(obsId) ) then
      do i=1,nInp_scl
        muR2(i) = muR2_scl(i)
      end do
      return
    end if

    !> get muR for the central scale
    fac0   = active_muRred_scl(1)%fac
    obsId0 = active_muRred_scl(1)%obsId

    do i=1,nInp_scl
      !> this is where nSubR_scl == 1 enters!
      iR = map_muRsub_to_red_scl(i,1)
      !>----- already evolved the scale before!
      id = idmap_muR_scl(i)
      if ( i /= id ) then
        muR2(i) = muR2(id)
        cycle
      end if
      !>----- actually compute the scale
      muiR = 0d0
      obsIdiR = active_muRred_scl(iR)%obsId
      faciR   = active_muRred_scl(iR)%fac
      !> swap out scales only when obsId same as central scale
      if ( obsIdiR == obsId0 ) then
        if ( present(obsId) ) obsIdiR = obsId
        if ( present(fac) ) then
          faciR = fac * faciR
          if ( (obsIdiR > 0) .or. present(obsId) )  &
            &  faciR = faciR / fac0
        end if
      end if
      !> set muiR
      if (obsIdiR == 0) then  ! fixed scale
        muiR = faciR
      else  ! dynamic scale
        val = getValue_obs(obsIdiR, valid)
        if (.not.valid) then
          muR2(i) = 0d0
          cycle
        end if
        muiR = faciR * val
      end if
      muR2(i) = muiR**2
    end do

  end subroutine get_muR2_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value of the factorisation scale squared [GeV^2]
  !
  !> @param[out] muF2 array of the scales
  !> @param[in] fac optional
  !> @param[in] obsId optional
  !-----------------------------------------------------------------------------
  subroutine get_muF2_scl(muF2, fac, obsId)
    use Observables_mod
    double precision, dimension(nMax_scl), intent(out) :: muF2
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    integer :: i, iF, id
    !> info on the input central scale
    logical :: valid
    double precision :: fac0, faciF, muiF, val
    integer :: obsId0, obsIdiF

#ifndef NDEBUG
    ! if (nSubF_scl > 1) stop "get_muF2_scl: scale arrays unsupported"
#endif

    muF2 = 0d0

    if ( .not.present(fac) .and. .not.present(obsId) ) then
      do i=1,nInp_scl
        muF2(i) = muF2_scl(i)
      end do
      return
    end if

    !> get muF for the central scale
    fac0   = active_muFred_scl(1)%fac
    obsId0 = active_muFred_scl(1)%obsId

    do i=1,nInp_scl
      !> this is where nSubF_scl == 1 enters!
      iF = map_muFsub_to_red_scl(i,1)
      !>----- already evolved the scale before!
      id = idmap_muF_scl(i)
      if ( i /= id ) then
        muF2(i) = muF2(id)
        cycle
      end if
      !>----- actually compute the scale
      muiF = 0d0
      obsIdiF = active_muFred_scl(iF)%obsId
      faciF   = active_muFred_scl(iF)%fac
      !> swap out scales only when obsId same as central scale
      if ( obsIdiF == obsId0 ) then
        if ( present(obsId) ) obsIdiF = obsId
        if ( present(fac) ) then
          faciF = fac * faciF
          if ( (obsIdiF > 0) .or. present(obsId) )  &
            &  faciF = faciF / fac0
        end if
      end if
      !> set muiF
      if (obsIdiF == 0) then  ! fixed scale
        muiF = faciF
      else  ! dynamic scale
        val = getValue_obs(obsIdiF, valid)
        if (.not.valid) then
          muF2(i) = 0d0
          cycle
        end if
        muiF = faciF * val
      end if
      muF2(i) = muiF**2
    end do

  end subroutine get_muF2_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update all scales (muR2_scl,muF2_scl) using current Observables values
  !
  !> @param[inout] ipass optional set to 0 if at least one cut applies
  !-----------------------------------------------------------------------------
  subroutine update_scl(ipass)
    use Observables_mod
    integer, intent(inout), optional :: ipass
    integer :: i, j
    double precision :: val
    logical :: valid

    if (present(ipass)) then
      if (ipass==0) return
    end if

    if (isActive_sclc()) then
      !> only need central scale when ScaleCoeff is active
      muRred_scl(:) = 0d0
      valid_muRred_scl(:) = .false.
      call update_scl_eval(active_muRred_scl(1), muRred_scl(1), valid_muRred_scl(1))
      muFred_scl(:) = 0d0
      valid_muFred_scl(:) = .false.
      call update_scl_eval(active_muFred_scl(1), muFred_scl(1), valid_muFred_scl(1))
    else
      do i =1,nMuR_scl
        call update_scl_eval(active_muRred_scl(i), muRred_scl(i), valid_muRred_scl(i))
      end do
      do i =1,nMuF_scl
        call update_scl_eval(active_muFred_scl(i), muFred_scl(i), valid_muFred_scl(i))
      end do
    end if

    ! print*, "[1] muR:", muRred_scl(1:nMuR_scl)
    if (present(ipass) .and. .not.update_scl_qpass()) then
      ipass = 0
      return
    end if

    call update_scl_mu2()

  contains

    !> get value & valid state
    subroutine update_scl_eval(scale, value, valid)
      type(Scale_t), intent(in) :: scale
      double precision, intent(out) :: value
      logical, intent(out) :: valid
      double precision :: loc_value
      logical :: loc_valid
      !> fixed scale
      if ( scale%obsId == 0 ) then
        value = scale%fac
        valid = .true.
      !> dynamic scale
      else
        loc_value = getValue_obs(scale%obsId, loc_valid)
        if (.not.loc_valid) then
          !>----- option 1: exit out of the program
          ! print*, "update_scl: all dynamical scales must use valid observables"
          ! stop
          !>----- do nothing, alphas & PDFs will be set to zero later
          !>      however, enter some default value, so ME,... can be evaluated
          loc_value = 1d0
        end if
        value = scale%fac * loc_value
        valid = loc_valid
      end if
    end subroutine update_scl_eval

    !> update muR2 & muF2 (everything that is [GeV^2])
    subroutine update_scl_mu2()
      muR2red_scl(1:nMuR_scl) = muRred_scl(1:nMuR_scl)**2
      muF2red_scl(1:nMuF_scl) = muFred_scl(1:nMuF_scl)**2
    end subroutine update_scl_mu2

    !> check valid states to determine ipass return
    logical function update_scl_qpass()
      logical :: qvalidR, qvalidF
      logical :: qnoNaNR, qnoNaNF
      qvalidR = any(valid_muRred_scl(1:nMuR_scl))
      qvalidF = any(valid_muFred_scl(1:nMuF_scl))
      !> NaN: by definition fails self-comparison
      qnoNaNR = .not. any( muRred_scl(1:nMuR_scl) /= muRred_scl(1:nMuR_scl) )
      qnoNaNF = .not. any( muFred_scl(1:nMuF_scl) /= muFred_scl(1:nMuF_scl) )
      update_scl_qpass = qvalidR .and. qvalidF .and. qnoNaNR .and. qnoNaNF
    end function update_scl_qpass

  end subroutine update_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update internal PDF storage
  !> optionally swap out fac & obsId from the active scale setting
  !
  !> @param[in] withXRegions optional flag to include fixi's in the link
  !> @param[in] fac optional
  !> @param[in] obsId optional
  !-----------------------------------------------------------------------------
  subroutine updatePDF_scl(withXRegions, fac, obsId)
    use Process_mod, only : nPDFs_proc, beam1_proc, beam2_proc, pid_ap
    use LHAPDF_mod
    use PDFCache_mod
    use PDFInterp_mod
    use KinData_mod, only : x1_kin, x2_kin, xreg1_kin, xreg2_kin
    use Run_mod, only: pdf_xmin_run, pdf_qmin_run, pdf_qmax_run
    ! use PDFInterp_mod
    logical, intent(in), optional :: withXRegions
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    logical :: withXRegions_, qswap

#ifndef NDEBUG
    ! if (nSubF_scl > 1) stop "updatePDF_scl: scale arrays unsupported"
#endif

    ! !> if we use the PDF interpolation, skip the caching!
    ! if ( isActive_pdfi() ) return

    if ( isActive_sclc() .and.  &
       & .not.present(fac) .and.  &
       & .not.present(obsId) ) then
      return
    end if

    withXRegions_ = .false.  ! default
    if (present(withXRegions)) withXRegions_ = withXRegions

    if (nPDFs_proc >= 1) then
      qswap = (beam1_proc==pid_ap)
      call updatePDF_scl_evolve(x1_kin, f1_scl, 1, qswap, fac, obsId)
      if (withXRegions_)  &
      &  call updatePDF_scl_evolve(x1_kin/xreg1_kin, f1x1_scl, -1, qswap, fac, obsId)
    end if
    if (nPDFs_proc >= 2) then
      qswap = (beam2_proc==pid_ap)
      call updatePDF_scl_evolve(x2_kin, f2_scl, 2, qswap, fac, obsId)
      if (withXRegions_)  &
      &  call updatePDF_scl_evolve(x2_kin/xreg2_kin, f2x2_scl, -2, qswap, fac, obsId)
    end if

  contains

    !> tiny wrapper for the cases
    subroutine updatePDF_scl_evolve(xval, f, itype, qswap_, fac_, obsId_)
      double precision, intent(in) :: xval
      double precision, dimension(-6:7,nMax_scl), intent(out) :: f
      logical, intent(in) :: qswap_
      integer, intent(in) :: itype  ! for PDF interpolation
      double precision, optional, intent(in) :: fac_
      integer, optional, intent(in) :: obsId_
      if ( (xval>1d0) .or. (xval<0d0) ) then
        f = 0d0
        return
      end if
      if ( present(fac_) .or. present(obsId_) ) then
        call updatePDF_scl_evolve_active(xval, f, itype, fac_, obsId_)
      else
        call updatePDF_scl_evolve_cache(xval, f, itype)
      end if
      !> for anti-protons we need to swap q <-> qb
      if (qswap_) call updatePDF_scl_swap(f)
    end subroutine updatePDF_scl_evolve

    !> convert proton to anti-protons
    subroutine updatePDF_scl_swap(f)
      double precision, dimension(-6:7,nMax_scl), intent(inout) :: f
      integer :: iscl, iq
      double precision :: f_tmp
      do iscl=1,nInp_scl
        do iq=1,6
          f_tmp       = f(iq,iscl)
          f(iq,iscl)  = f(-iq,iscl)
          f(-iq,iscl) = f_tmp
        end do
      end do
    end subroutine updatePDF_scl_swap

    !> evolve a PDF using cached muF values
    subroutine updatePDF_scl_evolve_cache(xval, f, itype)
      double precision, intent(in) :: xval
      double precision, dimension(-6:7,nMax_scl), intent(out) :: f
      integer, intent(in) :: itype  ! for PDF interpolation
      integer :: i, iF, id

      f = 0d0  ! reset

      do i=1,nInp_scl
        !> this is where nSubF_scl == 1 enters!
        iF = map_muFsub_to_red_scl(i,1)
        !>----- invalid scale: PDF's zero
        if ( .not. valid_muFred_scl(iF) ) cycle
        !>----- already evolved the scale before!
        id = idmap_muF_scl(i)
        if ( i /= id ) then
          f(-6:7,i) = f(-6:7,id)
          cycle
        end if
        !> valid & new distinct scale
        if ( isActive_pdfc() ) then
          call evolvePDF_pdfc(xval, muFred_scl(iF), f(-6:7,i), id=itype)
        elseif ( isActive_pdfi() ) then
          if ( (xval < pdf_xmin_run) .or.  &
             & (muFred_scl(iF) < pdf_qmin_run) .or.  &
             & (muFred_scl(iF) > pdf_qmax_run) ) then
            print*, "updatePDF_scl: pdfi out of bounds", iF
            print*, " > xval:", xval, pdf_xmin_run
            print*, " > q:   ", muFred_scl(iF), pdf_qmin_run, pdf_qmax_run
            call print_xvegas()
            cycle  ! f = 0d0
          end if
          call evolvePDF_pdfi(xval, muFred_scl(iF), f(-6:7,i), id=itype)
        else
          select case (itype)
          case(1,-1)
            call evolve_xfphxQ_lhapdf(xval, muFred_scl(iF), f(-6:7,i))
          case(2,-2)
            call evolve_xfph2xQ_lhapdf(xval, muFred_scl(iF), f(-6:7,i))
          end select
        end if
        ! if (i==1) then
        !   print*, "C: f   ", f(-5:5,1)
        !   print*, "C: xval", xval
        !   print*, "C: muF ", muFred_scl(iF)
        ! end if
      end do
      f = f / xval
    end subroutine updatePDF_scl_evolve_cache

    !> evolve a PDF ignoring the cache and using the "active" scale (for sclc)
    subroutine updatePDF_scl_evolve_active(xval, f, itype, fac_, obsId_)
      use Observables_mod
      double precision, intent(in) :: xval
      double precision, dimension(-6:7,nMax_scl), intent(out) :: f
      integer, intent(in) :: itype  ! for PDF interpolation
      double precision, optional, intent(in) :: fac_
      integer, optional, intent(in) :: obsId_
      integer :: i, iF, id
      !> info on the input central scale
      logical :: valid
      double precision :: fac0, faciF, muiF, val
      integer :: obsId0, obsIdiF

      f = 0d0  ! reset

      !> get muF for the central scale
      fac0   = active_muFred_scl(1)%fac
      obsId0 = active_muFred_scl(1)%obsId

      do i=1,nInp_scl
        !> this is where nSubF_scl == 1 enters!
        iF = map_muFsub_to_red_scl(i,1)
        !>----- already evolved the scale before!
        id = idmap_muF_scl(i)
        if ( i /= id ) then
          f(-6:7,i) = f(-6:7,id)
          cycle
        end if
        !>----- actually compute the scale
        muiF = 0d0
        obsIdiF = active_muFred_scl(iF)%obsId
        faciF   = active_muFred_scl(iF)%fac
        !> swap out scales only when obsId same as central scale
        if ( obsIdiF == obsId0 ) then
          if ( present(obsId_) ) obsIdiF = obsId_
          if ( present(fac_) ) then
            faciF = fac_ * faciF
            if ( (obsIdiF > 0) .or. present(obsId_) )  &
              &  faciF = faciF / fac0
          end if
        end if
        !> set muiF
        if (obsIdiF == 0) then  ! fixed scale
          muiF = faciF
        else  ! dynamic scale
          val = getValue_obs(obsIdiF, valid)
          if (.not.valid) cycle  ! PDF = zero
          muiF = faciF * val
        end if
        if ( isnan(muiF) ) then
          print*, "updatePDF_scl_evolve_active: muF = NaN?!"
          print*, " > obs", obsIdiF, ">", val, valid
          print*, " > fac", faciF
          call print_xvegas()
          cycle  ! PDF = zero
        end if
        !> call PDF evolution routine
        if ( isActive_pdfc() ) then
          call evolvePDF_pdfc(xval, muiF, f(-6:7,i), id=itype)
        elseif ( isActive_pdfi() ) then
          if ( (xval < pdf_xmin_run) .or.  &
             & (muiF < pdf_qmin_run) .or.  &
             & (muiF > pdf_qmax_run) ) then
            print*, "updatePDF_scl: pdfi out of bounds", iF
            print*, " > xval:", xval, pdf_xmin_run
            print*, " > q:   ", muiF, pdf_qmin_run, pdf_qmax_run
            print*, " > obs", obsIdiF, ">", val, valid
            call print_xvegas()
            cycle  ! f = 0d0
          end if
          call evolvePDF_pdfi(xval, muiF, f(-6:7,i), id=itype)
        else
          select case (itype)
          case(1,-1)
            call evolve_xfphxQ_lhapdf(xval, muiF, f(-6:7,i))
          case(2,-2)
            call evolve_xfph2xQ_lhapdf(xval, muiF, f(-6:7,i))
          end select
        end if
        ! if (i==1) then
        !   print*, "I: f   ", f(-5:5,1)
        !   print*, "I: xval", xval
        !   print*, "I: muF ", muiF
        ! end if
      end do
      f = f / xval
    end subroutine updatePDF_scl_evolve_active

  end subroutine updatePDF_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> compute the parton luminosity
  !
  !> @param[in] ix
  !> @param[in] ip1
  !> @param[in] ip2
  !> @param[in] partons
  !> @param[in] fac optional
  !> @param[in] obsId optional
  !-----------------------------------------------------------------------------
  subroutine get_lumi_scl(ix, ip1, ip2, partons, fac, obsId)
    use KinData_mod, only : x1_kin, x2_kin, xreg1_kin, xreg2_kin
    integer, intent(in) :: ix, ip1, ip2
    double precision, intent(out) :: partons(nMax_scl)
    double precision, optional, intent(in) :: fac
    integer, optional, intent(in) :: obsId
    !>
    logical jflag
    common /jacobian/jflag
!$omp threadprivate(/jacobian/)
    !>
    integer imethod
    common /method/imethod

#ifndef NDEBUG
    if ((ip1<-6).or.(ip1>7).or.(ip2<-6).or.(ip2>7)) then
      print*, "get_lumi_scl: invalid ip's:", ip1, ip2
      stop
    end if
#endif

    !> for coefficient functions: not needed
    if ( isActive_sclc() .and.  &
       & .not.present(fac) .and.  &
       & .not.present(obsId) ) then
      partons(:) = 1d0  ! luminosity build up in bino
      return
    end if

    if ( present(fac) .or. present(obsId) ) then
      call updatePDF_scl(withXRegions=(ix/=1), fac=fac, obsId=obsId)
    end if

    !> convolute PDFs
    if (imethod.eq.2) then
       select case (ix)
       case(1)
          partons(1:nInp_scl) = f1_scl(ip1,1:nInp_scl)   * f2_scl(ip2,1:nInp_scl)
       case(2)
          partons(1:nInp_scl) = f1_scl(ip1,1:nInp_scl)   * f2x2_scl(ip2,1:nInp_scl)
       case(3)
          partons(1:nInp_scl) = f1x1_scl(ip1,1:nInp_scl) * f2_scl(ip2,1:nInp_scl)
       case(4)
          partons(1:nInp_scl) = f1x1_scl(ip1,1:nInp_scl) * f2x2_scl(ip2,1:nInp_scl)
       end select
    else
       partons(1:nInp_scl) = f1_scl(ip1,1:nInp_scl) * f2_scl(ip2,1:nInp_scl)
    end if

  end subroutine get_lumi_scl


  !-----------------------------------------------------------------------------
  !> @brief
  !> update internal alphas storage
  !-----------------------------------------------------------------------------
  subroutine updateAlphaS_scl()
    use Constants_mod
    use LHAPDF_mod
    integer :: i,j, iR, id

    if ( isActive_sclc() ) then
      alphassub_scl = twopi  ! we expand in: alphas/(2pi)
      return
    end if

    do j=1,nSubR_scl
    do i=1,nInp_scl
      ! !> sclc override
      ! if ( isActive_sclc() .and. i>=1 .and. i<=nscl_sclc ) then
      !   alphassub_scl(i,j) = 1d0
      !   cycle
      ! end if
      iR = map_muRsub_to_red_scl(i,j)
      !> invalid scale: set alphas to zero
      if ( .not. valid_muRred_scl(iR) ) then
        alphassub_scl(i,j) = 0d0
        cycle
      end if
      !> already evolved the scale before!
      id = idmap_muR_scl(i)
      if ( i /= id ) then
        alphassub_scl(i,j) = alphassub_scl(id,j)
        cycle
      end if
      !> valid scale
      ! print*, "[2] muR:", muRred_scl(1:nMuR_scl)
      ! print*, "updateAlphaS_scl:", iR, ":", muRred_scl(iR)
      alphassub_scl(i,j) = alphasQ_lhapdf(muRred_scl(iR))
    end do
    end do

  end subroutine updateAlphaS_scl


end module Scales_mod


