!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Selectors [external members: `_sel`]
!
! DESCRIPTION:
!> Module maintining all registered event selection cuts for a process
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module Selectors_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_sel
  public :: register_sel, destroy_sel
  public :: initBuffer_sel, destroyBuffer_sel
  public :: digest_sel
  public :: qpass_sel, applyAll_sel
  public :: min_shat_sel, getExtremes_sel
  public :: Selector_t

  logical, public :: fully_inclusive = .false. !> False by default

  integer :: minnjets = -1
  logical :: qexclusive = .false.

  type Selector_t
    integer :: obsId = 0
    double precision :: lower = 0d0
    double precision :: upper = 0d0
    logical :: qlower = .false.
    logical :: qupper = .false.
    logical :: qinvert = .false.
    logical :: qisint = .false.
    logical :: qvalid = .true.  ! must be valid?
    integer :: iORgroup = 0  ! 0: no OR group
    type(Selector_t), pointer :: next => null()
  end type Selector_t

  logical, allocatable, dimension(:) :: qpassORgroup
!$OMP THREADPRIVATE(qpassORgroup)
  integer :: nORgroup = 0

  !> keep track of the head of the selectors list
  type(Selector_t), pointer :: head => null()


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the minimum number of jets that is required and whether the
  !> computation should be exclusive/inclusive w.r.t. additional jet emissions.
  !-----------------------------------------------------------------------------
  subroutine init_sel()
    use Process_mod
    use Jets_mod, only : algo_jet

    minnjets = njets_proc
    qexclusive = qexclusive_proc

    !> if the jet algorithm "none" (id = 0) is chosen: no jet requirements
    !> it's the user's responsibility to have an IR finite setup
    if ( algo_jet() == 0 ) then
      print*, "no jet algorithm active"
      print*, "we will therefore switch to the inclusive case"
      print*, "with minnjets = 0"
      minnjets = 0
      qexclusive = .false.
    end if

  end subroutine init_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> register a new selector: lower < obs < upper
  !
  !> @param[in] obsName unique identifyer of the observable
  !> @param[in] lower optional lower cut
  !> @param[in] upper optional upper cut
  !> @param[in] invert optional flag controlling whether we "accept" or "reject"
  !> @param[in] valid optional flag to enfoce a valid state on the observable
  !> @param[in] ORgroup optional identifier for the OR group
  !-----------------------------------------------------------------------------
  subroutine register_sel(obsName, lower, upper, invert, valid, ORgroup)
    use Observables_mod, only : getIdFromName_obs, isUsed_obs, isInteger_obs
    use Flavour_mod, only: minnjets_flav
    use Jets_mod
    character(len=*), intent(in) :: obsName
    double precision, optional, intent(in) :: lower, upper
    logical, optional, intent(in) :: invert, valid
    integer, optional, intent(in) :: ORgroup
    integer :: obsId
    type(Selector_t), pointer :: newSel

    obsId = getIdFromName_obs(obsName)

    if (.not.present(lower) .and. .not.present(upper)) then
      print*, "ignoring selector that does nothing!"
      return
    end if

    allocate(newSel)
    newSel%obsId  = obsId
    if (present(lower)) newSel%lower = lower
    if (present(upper)) newSel%upper = upper
    newSel%qlower = present(lower)
    newSel%qupper = present(upper)
    if (present(invert)) then
      newSel%qinvert = invert
    end if ! default init is .false. (see above)
    if (present(valid)) then
      newSel%qvalid = valid
    end if ! default init is .true. (see above)
    newSel%qisint = isInteger_obs(obsId)

    !> OR group setup
    if (present(ORgroup) .and. ORgroup > 0) then
      newSel%iORgroup = ORgroup ! default init is 0 (see above)
      if ( nORgroup < ORgroup ) nORgroup = ORgroup
    end if

    !> push it to the stack
    newSel%next => head
    head => newSel

    !> in case the selector is `njets` we should adjust
    !>   minnjets
    !>   qexclusive
    !> to even be able to accommodate the selector
    !> (we assume the user knows what he/she is doing when invoking this)
    if ( newSel%obsId == isUsed_obs("njets") ) then
      print*, "registered an `njets` selector"
      print*, "we will therefore switch to the inclusive case and"
      print*, "this will potentially override the global settings for"
      print*, " > qinvert ", newSel%qinvert
      print*, " > lower   ", newSel%lower, newSel%qlower
      print*, " > upper   ", newSel%upper, newSel%qupper
      qexclusive = .false.
      if ( .not. newSel%qinvert ) then
        if ( newSel%qlower ) then
          !> we assume that the user has made the setup IR safe
          minnjets = min( minnjets, INT(newSel%lower) )
        end if
        if ( newSel%qupper ) then
          minnjets = min( minnjets, INT(newSel%upper) )
        end if
      else
        if ( newSel%qupper ) then
          minnjets = min( minnjets, INT(newSel%upper)+1 )
        end if
        if ( newSel%qlower ) then
          !> we assume that the user has made the setup IR safe
          minnjets = 0
        end if
      end if
      minnjets = max(minnjets, 0)
      print*, "> minnjets: ", minnjets
    end if

    !> done in PhotonIsolation_mod using getExtremes_sel
    ! !> in case the selector is `nphotons` we should adjust
    ! !>   minnphotons
    ! if (newSel%obsId == id_nphotons) then
    !   print*, "registered an `nphotons` selector"
    !   print*, "this will potentially override the global settings for"
    !   print*, "> minnphotons"
    !   if ( .not. newSel%qinvert ) then
    !     if ( newSel%qlower ) then
    !       minnphotons = min( minnphotons, INT(newSel%lower) )
    !     end if
    !     if ( newSel%qupper ) then
    !       minnphotons = min( minnphotons, INT(newSel%upper) )
    !     end if
    !   else
    !     if ( newSel%qupper ) then
    !       minnphotons = min( minnphotons, INT(newSel%upper)+1 )
    !     end if
    !     if ( newSel%qlower ) then
    !       !> we assume that the user has made the setup IR safe
    !       minnphotons = 0
    !     end if
    !   end if
    !   minnphotons = max(minnphotons, 0)
    !   print*, "> minnphotons: ", minnphotons
    ! end if

    !> in case the selector is `flav_njets` we should adjust
    !>   minnjets_flav
    !> if the lower selector exists
    if ( newSel%obsId == isUsed_obs("flav_njets") ) then
      print*, "registered an `flav_njets` selector"
      print*, "we will therefore switch to the inclusive case and"
      print*, "this will potentially override the global settings for"
      qexclusive = .false.
      if ( .not. newSel%qinvert .and. newSel%qlower ) then
        minnjets_flav = INT(newSel%lower)
      end if
      minnjets_flav = max(minnjets_flav, 0)
      print*, "> minnjets_flav: ", minnjets_flav
    end if

  end subroutine register_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate all registered selectors
  !-----------------------------------------------------------------------------
  subroutine destroy_sel()
    type(Selector_t), pointer :: it, next
    it => head
    do while ( associated(it) )
      ! save next before we destroy it
      next => it%next
      deallocate(it)
      ! loop
      it => next
    end do
    head => null()
    if ( allocated(qpassORgroup) ) deallocate(qpassORgroup)
    nORgroup = 0
  end subroutine destroy_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate threadprivate dynamical arrays
  !-----------------------------------------------------------------------------
  subroutine initBuffer_sel()
    allocate(qpassORgroup(nORgroup))
  end subroutine initBuffer_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> deallocate threadprivate dynamical arrays
  !-----------------------------------------------------------------------------
  subroutine destroyBuffer_sel()
    if ( allocated(qpassORgroup) ) deallocate(qpassORgroup)
  end subroutine destroyBuffer_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> digest information on the registered selectors
  !
  !> @param[in] unit optional logical unit connected for formatted output
  !-----------------------------------------------------------------------------
  subroutine digest_sel(unit)
    use Process_mod
    use Jets_mod
    use Observables_mod
    integer, optional, intent(in) :: unit
    type(Selector_t), pointer :: it
    integer :: iout
    character(len=*), parameter :: fmt_str =  &
      & '(1x,"*",3x,A,T103,"*")'
    character(len=*), parameter :: fmt_njets =  &
      & '(1x,"*",3x,"njets: ",10x,I2," (",A9,")",T103,"*")'
    character(len=*), parameter :: fmt_min_max =  &
      & '(1x,"*",1x,L1,I1,1x,A6,":",1f12.2," <= ",A15," <= ",1f12.2,T103,"*")'
    character(len=*), parameter :: fmt_min =  &
      & '(1x,"*",1x,L1,I1,1x,A6,":",1f12.2," <= ",A15,T103,"*")'
    character(len=*), parameter :: fmt_max =  &
      & '(1x,"*",1x,L1,I1,1x,A6,":",16x,A15," <= ",1f12.2,T103,"*")'
    character(len=6) :: seltype

    if (present(unit)) then
      iout = unit
    else
      iout = 6 ! output to screen
    endif

    if ( associated(head) ) then
      it => head
      do while ( associated(it) )
        seltype = ''
        if ( it%qinvert ) then
          seltype = 'reject'
        else
          seltype = 'accept'
        endif
        if (it%qlower .and. it%qupper) then
          write(iout,fmt_min_max) it%qvalid, it%iORgroup, seltype, it%lower, getNameFromId_obs(it%obsId), it%upper
        else if (it%qlower .and. .not.it%qupper) then
          write(iout,fmt_min) it%qvalid, it%iORgroup, seltype, it%lower, getNameFromId_obs(it%obsId)
        else if (.not.it%qlower .and. it%qupper) then
          write(iout,fmt_max) it%qvalid, it%iORgroup, seltype, getNameFromId_obs(it%obsId), it%upper
        else
          write(iout,*) "SOMETHING WENT WRONG"
        endif
        it => it%next
      end do
    else
      print*, "* no selectors registered"
    endif


    write(iout,fmt_str) ""
    write(iout,fmt_str) "> Jet settings <"
    write(iout,fmt_str) ""

    if (qexclusive) then
      write(iout,fmt_njets) minnjets, "exclusive"
    else
      write(iout,fmt_njets) minnjets, "inclusive"
    end if

    if (minnjets < njets_proc) then
      write(iout,fmt_str) "WARNING: minnjets < njets_proc (potential IR divergence)"
    end if
    if (njets_proc > 0 .and. algo_jet() == 0) then
      write(iout,fmt_str) "WARNING: jet algorithm = 'none' (potential IR divergence)"
    end if

    call digest_jet(iout)

  end subroutine digest_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply all event selection cuts
  !
  !> @param[in] sel Selector to be tested
  !> @param[in] val value of the observable associated with `sel`
  !-----------------------------------------------------------------------------
  function qpass_sel(sel, val)
    logical :: qpass_sel
    type(Selector_t), intent(in) :: sel
    double precision, intent(in) :: val

    qpass_sel = .true.
    if ( sel%qisint ) then
      !> INT
      if ( .not. sel%qinvert .and.  &
         & (  &
         &   (sel%qlower .and. INT(val)<INT(sel%lower)) .or.  &
         &   (sel%qupper .and. INT(val)>INT(sel%upper))  &
         & )  &
         & .or.  &
         & sel%qinvert .and.  &
         & (  &
         &   (sel%qlower .and. INT(val)>=INT(sel%lower) .and.  &
         &    sel%qupper .and. INT(val)<=INT(sel%upper)) .or.  &
         &   (sel%qlower .and. INT(val)>=INT(sel%lower) .and.  &
         &    .not. sel%qupper ) .or.  &
         &   (sel%qupper .and. INT(val)<=INT(sel%upper) .and.  &
         &    .not. sel%qlower )  &
         & ) ) then
        qpass_sel = .false.
      end if
    else
      !> DBLE
      if ( .not. sel%qinvert .and.  &
         & (  &
         &   (sel%qlower .and. val<sel%lower) .or.  &
         &   (sel%qupper .and. val>sel%upper)  &
         & )  &
         & .or.  &
         & sel%qinvert .and.  &
         & (  &
         &   (sel%qlower .and. val>=sel%lower .and.  &
         &    sel%qupper .and. val<=sel%upper) .or.  &
         &   (sel%qlower .and. val>=sel%lower .and.  &
         &    .not. sel%qupper ) .or.  &
         &   (sel%qupper .and. val<=sel%upper .and.  &
         &    .not. sel%qlower )  &
         & ) ) then
        qpass_sel = .false.
      end if
    end if
  end function qpass_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> apply all event selection cuts
  !> @note selectors active during jet clusitering are applied in JetAlgo_mod
  !
  !> @param[in] npar kinematics specified by the number of external partlicles
  !> @param[inout] ipass set to 0 if at least one cut applies
  !> @param[in] onlyFlav optional flag to only consider flavour observables
  !> @param[in] skipFlav optional flag to skip all flavour-sensitive observables
  !-----------------------------------------------------------------------------
  subroutine applyAll_sel(npar, ipass, onlyFlav, skipFlav)
    use KinData_mod
    use Observables_mod
    use Flavour_mod, only: isActive_flav, minnjets_flav, is_leading_flav
    use UnitPhase_mod, only : phase_unit, nocuts
    integer, intent(in) :: npar
    integer, intent(inout) :: ipass
    logical, intent(in), optional :: onlyFlav, skipFlav
    type(Selector_t), pointer :: it
    double precision :: valObs
    logical :: valid, qonlyF, qskipF
    !> composite observables
    integer :: isub, nsub

    if (ipass == 0) return  ! nothing to do

    if (phase_unit .and. nocuts) return  ! UnitPhase test with no cuts


    qonlyF = .false.
    if (present(onlyFlav)) qonlyF = onlyFlav

    qskipF = .false.
    if (present(skipFlav)) qskipF = skipFlav

    if ( kin(npar)%njets.lt.minnjets .or. &
       & ( qexclusive .and. kin(npar)%njets.gt.minnjets ) ) then
      ipass = 0
      return
    end if

    if ( isActive_flav() .and. .not.qskipF ) then
      !> satisfy minimum # of flavoured jets?
      if ( kin(npar)%flav_njets.lt.minnjets_flav ) then
        ipass = 0
        return
      end if
      !> satisfy condition that leading jet be flavoured
      if ( is_leading_flav .and. .not.kin(npar)%qflavjet(1) ) then
        ipass = 0
        return
      end if
    end if

    !> OR group init qpass=.false. and look for at least one
    !> selector in the group that passes to set it to .true.
    if ( nORgroup > 0 ) qpassORgroup(:) = .false.
    ! call print_momenta(npar)

    it => head
    do while ( associated(it) )

      !> Skip extra, flav-dep(optional)
      !> if only global we skip all OR-blocks (only safe solution)
      if ( (qonlyF .and. .not.needsFlav_obs(it%obsId)) .or. &
         & (qskipF .and. needsFlav_obs(it%obsId)) ) then
        it => it%next
        cycle
      end if

      valObs = getValue_obs(it%obsId, valid)
      if ( .not.valid ) then
        !> observable must be valid
        if ( it%qvalid ) then
          print*, getNameFromId_obs(it%obsId)
          print*, "applyAll_sel: all selectors must use valid observables", it%obsId
          print*, kin(npar)%njets, " vs. ", minnjets
          stop
        end if
        !> optional: `select_if_valid`, etc. --> go to next
        it => it%next
        cycle
      end if

      if ( .not. isComposite_obs(it%obsId) ) then
        !>----- [1] non-composite observables
        if ( .not. qpass_sel(it,valObs) ) then
          if (it%iORgroup==0) then
            ipass = 0
            return
          end if
        else
          if (it%iORgroup/=0) then
            qpassORgroup(it%iORgroup) = .true.
          end if
        end if

      else
        !>----- [2] composite observables
        nsub = int(valObs)
        loopsub: do isub = 1, nsub
          valObs = getValueComposite_obs(it%obsId, valid, isub)
          ! print*, "SELECT:", isub, "/", nsub, ":", valObs
          if ( .not. qpass_sel(it,valObs) ) then
            if (it%iORgroup==0) then
              ipass = 0
              return
            end if
          else
            if (it%iORgroup/=0) then
              qpassORgroup(it%iORgroup) = .true.
            end if
          end if
        end do loopsub

      end if  ! composite

      it => it%next
    end do  ! it

    !> check that all OR groups pass
    ! print*, qpassORgroup
    ! print*, size(qpassORgroup)
    !> no OR-blocks when we evaluate only a sub-set of cuts (only safe solution)
    if ( nORgroup > 0 ) then
      if (.not. all(qpassORgroup)) then
        ! print*, "not all pass"
        ipass = 0
        return
      end if
    end if

  end subroutine applyAll_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> determine from the registered selectors the minimal partonic
  !> centre-of-mass energy (squared)
  !
  !> @return minimal value for shat [GeV^2]
  !-----------------------------------------------------------------------------
  function min_shat_sel()
    use StrHelper_mod, only : to_upper_str
    use Observables_mod, only : getNameFromId_obs, max_name_length_obs
    use Jets_mod
    double precision :: min_shat_sel
    double precision :: min_valObs
    type(Selector_t), pointer :: it
    double precision, allocatable, dimension(:) :: min_shat_grp
    logical, allocatable, dimension(:) :: qhit_grp
    !>---- info on the observable
    character(len=max_name_length_obs) :: name
    integer :: ifac

    min_shat_sel = 0d0
    !> OR groups
    allocate(min_shat_grp(0:nORgroup))
    min_shat_grp = 0d0
    allocate(qhit_grp(nORgroup))
    qhit_grp = .false.


    it => head
    do while ( associated(it) )
      !> check what we're dealing with
      name = getNameFromId_obs(it%obsId, ifac=ifac)
      !> not an observable that helps in determining smin
      if ( ifac <= 0 ) then
        it => it%next
        cycle
      end if
      print*, "<sel>"
      print*, "name: ", name, ifac

      min_valObs = 0d0
      if ( .not.it%qinvert .and. it%qlower ) min_valObs = it%lower
      !>---------------------------------------------------
      !> !!! this is WRONG !!!
      !> never uncomment! Left in so no one get's the idea of implementing such a line
      !> think about:
      !> select mll min=10 max=100
      !> reject mll min=50 max=60
      !> this line would set s12_min=60**2, which is not correct!
      !if (      it%qinvert .and. it%qupper ) min_valObs = it%upper
      !>---------------------------------------------------
      ! found a cut on a lower boundary which also is "meaningful", i.e. positive
      ! because cuts on "energies" better be positive
      if ( min_valObs > 0d0 ) then
        min_valObs = (ifac*min_valObs)**2

        !> is it a global selector or in an OR-group?
        if ( it%iORgroup == 0 ) then
          min_shat_grp(0) = max( min_shat_grp(0), min_valObs )
        else
          if (.not. qhit_grp(it%iORgroup) ) then
            min_shat_grp(it%iORgroup) = min_valObs
            qhit_grp(it%iORgroup) = .true.
          else
            min_shat_grp(it%iORgroup) = min( min_shat_grp(it%iORgroup), min_valObs )
          end if
        end if

      end if

      ! and iterate
      it => it%next
    end do
    !> finish off with the normal selectors.
    min_shat_sel = maxval( min_shat_grp )
    !> deallocate stuff
    deallocate(min_shat_grp)
    deallocate(qhit_grp)

    !> if minnjets > 0 also look in the jet selectors
    if (minnjets > 0) then
      min_shat_sel = max(min_shat_sel, min_shat_jet())
    end if

    return
  end function min_shat_sel


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the minimum and maximum values for an observable as determined by
  !> all the registered Selectors that are *not* in an OR group
  !
  !> @param[in] obsId identifyer for the observable
  !> @param[inout] minVal minimum value allowed by the selectors
  !> @param[inout] maxVal maximum value allowed by the selectors
  !-----------------------------------------------------------------------------
  subroutine getExtremes_sel(obsId, minVal, maxVal)
    use Observables_mod
    integer, intent(in) :: obsId
    double precision, intent(inout), optional :: minVal, maxVal
    type(Selector_t), pointer :: it
    double precision :: minLoc, maxLoc

    it => head
    do while ( associated(it) )
      if ( it%obsId /= obsId .or. it%iORgroup /= 0) then
        it => it%next
        cycle
      end if

      if (present(minVal)) then
        minLoc = minVal
        if ( .not.it%qinvert .and. it%qlower ) then
          minLoc = it%lower
        end if
        if ( it%qinvert .and. it%qupper .and. .not.it%qlower ) then
          minLoc = it%upper
          if ( isInteger_obs(it%obsId) ) minLoc = DBLE( INT(minLoc) + 1 )
        end if
        minVal = max(minVal, minLoc)
      end if

      if (present(maxVal)) then
        maxLoc = maxVal
        if ( .not.it%qinvert .and. it%qupper ) then
          maxLoc = it%upper
        end if
        if ( it%qinvert .and. it%qlower .and. .not.it%qupper ) then
          maxLoc = it%lower
          if ( isInteger_obs(it%obsId) ) maxLoc = DBLE( INT(maxLoc) - 1 )
        end if
        maxVal = min(maxVal, maxLoc)
      end if

      it => it%next
    end do

    return
  end subroutine getExtremes_sel


end module Selectors_mod
