!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: String Helper [external members: `_str`]
!
! DESCRIPTION:
!> Module collecting routines to manipulate strings
!>
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module StrHelper_mod
!-------------------------------------------------------------------------------
  implicit none
  public


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> Conversion of strings into upper-case letters
  !
  !> @param[in] str input string to be converted to upper case
  !> @return string converted to upper case
  !-----------------------------------------------------------------------------
  pure function to_upper_str(str) result (string)
    character(*), intent(in) :: str
    character(len(str)) :: string
    integer :: i, j
    character(26), parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    character(26), parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

    string = str
    do i = 1, len_trim(str)
       j = index(low, str(i:i))
       if (j > 0) string(i:i) = cap(j:j)
    end do

  end function to_upper_str


  !-----------------------------------------------------------------------------
  !> @brief
  !> Conversion of strings into lower-case letters
  !
  !> @param[in] str input string to be converted to lower case
  !> @return string converted to lower case
  !-----------------------------------------------------------------------------
  pure function to_lower_str(str) result (string)
    character(*), intent(in) :: str
    character(len(str)) :: string
    integer :: i, j
    character(26), parameter :: cap = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    character(26), parameter :: low = 'abcdefghijklmnopqrstuvwxyz'

    string = str
    do i = 1, len_trim(str)
       j = index(cap, str(i:i))
       if (j > 0) string(i:i) = low(j:j)
    end do

  end function to_lower_str


  !-----------------------------------------------------------------------------
  !> @brief
  !> Conversion of strings to double precision numbers
  !
  !> @param[in] str input string to be converted to a double
  !> @param[out] qfail optional flag to signal success of conversion
  !> @return valueDBLE
  !-----------------------------------------------------------------------------
  function to_dble_str(str, qfail) result (valueDBLE)
    double precision :: valueDBLE
    character(*), intent(in) :: str
    logical, optional, intent(out) :: qfail
    integer :: ierr

    if (present(qfail)) qfail = .false.

    read(str,*,iostat=ierr) valueDBLE
    if (ierr/=0) then
      if (present(qfail)) qfail = .true.
      valueDBLE = 0d0  ! default value
    end if

  end function to_dble_str


  !-----------------------------------------------------------------------------
  !> @brief
  !> Conversion of strings to integer numbers
  !
  !> @param[in] str input string to be converted to an int
  !> @param[out] qfail optional flag to signal success of conversion
  !> @return valueINT
  !-----------------------------------------------------------------------------
  function to_int_str(str, qfail) result (valueINT)
    integer :: valueINT
    character(*), intent(in) :: str
    logical, optional, intent(out) :: qfail
    integer :: ierr

    if (present(qfail)) qfail = .false.

    read(str,*,iostat=ierr) valueINT
    if (ierr/=0) then
      if (present(qfail)) qfail = .true.
      valueINT = 0  ! default value
    end if

  end function to_int_str


  !-----------------------------------------------------------------------------
  !> @brief
  !> Conversion of strings to booleans
  !
  !> @param[in] str input string to be converted to a bool
  !> @param[out] qfail optional flag to signal success of conversion
  !> @return valueBOOL
  !-----------------------------------------------------------------------------
  function to_bool_str(str, qfail) result (valueBOOL)
    logical :: valueBOOL
    character(*), intent(in) :: str
    logical, optional, intent(out) :: qfail
    integer :: ierr

    if (present(qfail)) qfail = .false.

    read(str,*,iostat=ierr) valueBOOL
    if (ierr/=0) then
      if (present(qfail)) qfail = .true.
      valueBOOL = .false.  ! default value
    end if

  end function to_bool_str


end module StrHelper_mod

