!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: Dynamic Libraries [external members: `_dl`]
!
! DESCRIPTION:
!> Module for handling runtime loading of dynamical libraries
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module dlfcn_mod
!-------------------------------------------------------------------------------
  use, intrinsic :: iso_c_binding
  implicit none
  private

  !> wrap C routines for convenience
  public :: open_dl, close_dl, sym_dl, error_dl

  !> modes for dlopen:
  integer(c_int), public, bind(c,name="rtld_lazy_nnlojet_")   :: rtld_lazy_dl
  integer(c_int), public, bind(c,name="rtld_now_nnlojet_")    :: rtld_now_dl
  integer(c_int), public, bind(c,name="rtld_global_nnlojet_") :: rtld_global_dl
  integer(c_int), public, bind(c,name="rtld_local_nnlojet_")  :: rtld_local_dl
  type(c_ptr), public, bind(c,name="shared_module_suffix_nnlojet_") :: shared_module_suffix_dl

  interface  ! from dlfcn_nnlojet.cxx:
    subroutine mangle_fortran(in_name, out_name) bind(c,name="mangle_nnlojet_")
      use iso_c_binding
      character(c_char), dimension(*), intent(in)  :: in_name
      character(c_char), dimension(*), intent(out) :: out_name
    end subroutine
  end interface

  interface  ! from <dlfcn.h>:
    function c_dlopen(file,mode) result(handle) bind(c,name="dlopen")
      !> void *dlopen(const char *file, int mode);
      use iso_c_binding
      character(c_char), dimension(*), intent(in) :: file
      integer(c_int), value :: mode
      type(c_ptr) :: handle
    end function
    function c_dlsym(handle,name) result(funptr) bind(c,name="dlsym")
      !> void *dlsym(void *handle, const char *name);
      use iso_c_binding
      type(c_ptr), value :: handle
      character(c_char), dimension(*), intent(in) :: name
      type(c_funptr) :: funptr ! a function pointer
    end function
    function c_dlclose(handle) result(status) bind(c,name="dlclose")
      !> int dlclose(void *handle);
      use iso_c_binding
      type(c_ptr), value :: handle
      integer(c_int) :: status
    end function
    function c_dlerror() result(error) bind(c,name="dlerror")
      !> char *dlerror(void);
      use iso_c_binding
      type(c_ptr) :: error
    end function
  end interface

  interface  ! from <string.h>:
    function strlen(string) result(len) bind(c,name="strlen")
      !> int strlen(char *string)
      use iso_c_binding
      type(c_ptr), value :: string
      integer(c_int) :: len
    end function
  end interface

contains

  function c_f_string(cptr) result(f_string)
    ! convert a null-terminated C string into a fortran character array pointer
    type(c_ptr), intent(in) :: cptr
    character(len=:), allocatable :: f_string
    integer :: i, slength
    character(kind=c_char), dimension(:), pointer :: fptr
    if ( .not. c_associated(cptr) ) then
      f_string = ""
      return
    end if
    slength = strlen(cptr)
    call c_f_pointer(fptr=fptr, cptr=cptr, shape=[slength])
    allocate(character(len=slength)::f_string)
    do i=1,slength
      f_string(i:i) = fptr(i)
    end do
  end function c_f_string

  function error_dl() result(error)
    character(len=:), allocatable :: error
    error = c_f_string(c_dlerror())
  end function error_dl

  function open_dl(file, suffix, mode) result(handle)
    character(len=*), intent(in) :: file
    character(len=*), intent(in), optional :: suffix
    integer(c_int), intent(in), optional :: mode
    type(c_ptr) :: handle
    character(:), allocatable :: suffix_

    integer(c_int) :: mode_
    suffix_ = ""
    if (present(suffix)) then
      suffix_ = suffix
    else
      suffix_ = c_f_string(shared_module_suffix_dl)
    end if
    mode_ = rtld_lazy_dl  ! default mode
    if (present(mode)) mode_ = mode
    handle = c_dlopen(trim(file)//trim(suffix_)//c_null_char, mode_)
    if ( .not. c_associated(handle) ) then
      write(*,*) "dlfcn_mod::open_dl: ", error_dl()
      stop
    end if
  end function open_dl

  subroutine close_dl(handle)
    type(c_ptr), intent(in) :: handle
    integer(c_int) :: status
    status = c_dlclose(handle)
    if (status /= 0) then
      write(*,*) "dlfcn_mod::close_dl: ", error_dl()
      stop
    end if
  end subroutine close_dl

  function sym_dl(handle, name, status, mangle) result(c_fptr)
    type(c_ptr), intent(in) :: handle
    character(len=*), intent(in) :: name
    integer, optional, intent(out) :: status
    logical, optional, intent(in) :: mangle
    logical :: mangle_
    type(c_funptr) :: c_fptr
    character(:), allocatable :: name_
    integer :: idx_
    mangle_ = .false.  ! default
    if (present(mangle)) mangle_ = mangle
    allocate(character(len=2*len(name)+10) :: name_)
    if (mangle_) then
      call mangle_fortran(trim(name)//c_null_char, name_)
    else
      name_ = trim(name)//c_null_char
    end if
    idx_ = index(name_,c_null_char)
    if (idx_ > 0) then
      name_(idx_+1:) = " "
    else
      write(*,*) "dlfcn_mod::magle: couldn't find NULL char?!"
      stop
    end if
    if (present(status))  status = 0  ! success
    c_fptr = c_dlsym(handle, name_)
    if ( .not. c_associated(c_fptr) ) then
      if (present(status)) then
        status = 1  ! fail
      else
        write(*,*) "dlfcn_mod::sym_dl: ", error_dl()
        stop
      end if
    end if
  end function sym_dl

end module dlfcn_mod

