!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: sigHinc [external members: `_Hinc`]
!
! DESCRIPTION:
!> Module providing the Higgs-production cross section differential in [yH]
!> first experimental implementation at NLO based on: hep-ph/0211141
!> you better be running these calculations using the NWA using inwa=3
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module sigHinc_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_Hinc, set_parameters_Hinc
  public :: set_kinematics_Hinc, get_mass_Hinc, get_rapidity_Hinc
  !> the differential cross section to be integrated
  public :: sigHinc

  !----- local setting for the cross ection:

  !> the higgs-boson decay
  !>  0) no decay
  !>  1) H -> 2 photons
  !>  2) H -> 2 tau
  !>  3) H -> 2 electrons 2 muons
  integer, parameter :: decay_H_Hinc        = 0
  integer, parameter :: decay_Hto2p_Hinc    = 1
  integer, parameter :: decay_Hto2tau_Hinc  = 2
  integer, parameter :: decay_Hto2l1p_Hinc  = 3
  integer, parameter :: decay_Hto2e2mu_Hinc = 4
  integer, parameter :: decay_Hto4e_Hinc    = 5
  integer, parameter :: decay_Hto2l2n_Hinc  = 6
  integer :: decay_Hinc = -1
  integer :: npar_Hinc  = -1
  !> vacuum-expectation-value of the higgs potential [GeV]
  double precision :: vev_Hinc = 246.21846d0

  !> variables setting the kinematics
  double precision :: M_Hinc, Y_Hinc, r1_Hinc, r2_Hinc
!$OMP THREADPRIVATE(M_Hinc, Y_Hinc, r1_Hinc, r2_Hinc)

  !> interface to the wrapper routines in `rapidix_wrapper.cxx`
  interface

    subroutine init_rapidix(roots) bind(c, name="init_rapidix_")
      use, intrinsic :: iso_c_binding
      real(c_double), intent(in) :: roots
    end subroutine init_rapidix

    subroutine set_MH_rapidix(MH) bind(c, name="set_MH_rapidix_")
      use, intrinsic :: iso_c_binding
      real(c_double), intent(in) :: MH
    end subroutine set_MH_rapidix

    subroutine set_mt_rapidix(mt) bind(c, name="set_mt_rapidix_")
      use, intrinsic :: iso_c_binding
      real(c_double), intent(in) :: mt
    end subroutine set_mt_rapidix

    subroutine set_scales_rapidix(mur, muf, as_muf, as_mt, as_mur) bind(c, name="set_scales_rapidix_")
      use, intrinsic :: iso_c_binding
      real(c_double), intent(in) :: mur, muf, as_muf, as_mt, as_mur
    end subroutine set_scales_rapidix

    subroutine set_order_rapidix(order) bind(c, name="set_order_rapidix_")
      use, intrinsic :: iso_c_binding
      integer(c_int), intent(in) :: order
    end subroutine set_order_rapidix

    subroutine set_WilsonCoeff_rapidix(WilsonCoefficientMode) bind(c, name="set_WilsonCoeff_rapidix_")
      use, intrinsic :: iso_c_binding
      integer(c_int), intent(in) :: WilsonCoefficientMode
    end subroutine set_WilsonCoeff_rapidix

    subroutine update_rapidix() bind(c, name="update_rapidix_")
    end subroutine update_rapidix

    subroutine evaluate_rapidix(x1, x2) bind(c, name="evaluate_rapidix_")
      use, intrinsic :: iso_c_binding
      !> corresponds to x[i]reg in NNLOJET
      real(c_double), intent(in) :: x1, x2
    end subroutine evaluate_rapidix

    subroutine get_results_rapidix(retMatrix) bind(c, name="get_results_rapidix_")
      use, intrinsic :: iso_c_binding
      !> return array
      !>  dimension 1: order = (0:3)
      !>               0 = LO
      !>               1 = NLO
      !>               2 = NNLO
      !>               3 = N3LO
      !>  dimension 2: process = (1:6)
      !>               1 = gg
      !>               2 = qg = qbarg
      !>               3 = gq = gqbar
      !>               4 = qqbar = qbarq
      !>               5 = qq = qbarqbar
      !>               6 = qQ = qbarQbar
      !>  dimension 3: ix = (1:4)  <-> already in NNLOJET convention (*but* xireg different!!!)
      !>               1 = delta(1-x1) delta(1-x2)
      !>               2 = delta(1-x1)
      !>               3 = delta(1-x2)
      !>               4 = 1
      real(c_double), intent(out) :: retMatrix(0:3,1:6,1:4)
    end subroutine get_results_rapidix

  end interface


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set parmeters of the cross section
  !-----------------------------------------------------------------------------
  subroutine init_Hinc()
    use Process_mod, only : name_proc, sqrts_proc
    use Parameters_mod
    use Scales_mod

    double precision :: emh,ehwidth
    common /Hmass/ emh,ehwidth

    double precision :: emt,etwidth,emb,ebwidth,emc,ecwidth,emtau,etauwidth
    common /Qmass/emt,etwidth,emb,ebwidth,emc,ecwidth,emtau,etauwidth

    integer :: i
    type(Scale_t) :: mu
    double precision, dimension(nMax_scl) :: muR, muF, alphas

    print*, "init_Hinc: setting up RapidiX..."

    !> initialise the decay settings
    select case (name_proc)
    case ('HINC')
      decay_Hinc = decay_H_Hinc
      npar_Hinc  = 4
    case ('HTO2PINC')
      decay_Hinc = decay_Hto2p_Hinc
      npar_Hinc  = 4
    case ('HTO2TAUINC')
      decay_Hinc = decay_Hto2tau_Hinc
      npar_Hinc  = 4
    case ('HTO2L1PINC')
      decay_Hinc = decay_Hto2l1p_Hinc
      npar_Hinc  = 5
    case ('HTO2E2MUINC')
      decay_Hinc = decay_Hto2e2mu_Hinc
      npar_Hinc  = 6
    case ('HTO4EINC')
      decay_Hinc = decay_Hto4e_Hinc
      npar_Hinc  = 6
    case ('HTO2L2NINC')
      decay_Hinc = decay_Hto2l2n_Hinc
      npar_Hinc  = 6
    case default
      print*, "init_Hinc: unknown process name: ", name_proc
      stop
    end select

#ifdef RAPIDIX

    call init_rapidix(sqrts_proc)
    call set_MH_rapidix(emh)

    if (mt_scheme_param() /= 1) then
      print*, "init_Hinc: I only accept mt in the MSbar scheme"
      stop
    end if
    call set_mt_rapidix(mt_param())

    !> default always MSbar
    select case (mt_scheme_param())
    case (0)  ! pole / on-shell scheme
      call set_WilsonCoeff_rapidix(2)
    case (1)  ! MSbar scheme
      call set_WilsonCoeff_rapidix(1)
    case default
      stop ""
    end select

    !> initialise some random scale?
    !call set_scales_alphas_rapidix(???)

    call update_rapidix()

#else
    stop "you can't use the Hinc process without the RapidiX library!"
#endif

  end subroutine init_Hinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set parmeters of the cross section
  !
  !> @param[in] vev optional input to set the Higgs vev
  !-----------------------------------------------------------------------------
  subroutine set_parameters_Hinc(vev)
    double precision, optional, intent(in) :: vev
    if (present(vev)) vev_Hinc = vev
  end subroutine set_parameters_Hinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the variables that define the Born kinematics
  !
  !> @param[in] M the invariant mass of the colour-less final state (= H)
  !> @param[in] Y the rapidity of the colour-less final state (= H)
  !-----------------------------------------------------------------------------
  subroutine set_kinematics_Hinc(M, Y)
    double precision, intent(in) :: M, Y
    M_Hinc = M
    Y_Hinc = Y
    ! print*, "set_kinematics_Hinc: ", M, Y
  end subroutine set_kinematics_Hinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter function of the invariant mass of the colour-less final state (= H)
  !-----------------------------------------------------------------------------
  integer function get_mass_Hinc()
    get_mass_Hinc = M_Hinc
  end function get_mass_Hinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter function of the rapidity of the colour-less final state (= H)
  !-----------------------------------------------------------------------------
  integer function get_rapidity_Hinc()
    get_rapidity_Hinc = Y_Hinc
  end function get_rapidity_Hinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> the differential cross section
  !
  !> @param[in] rnorm
  !> @param[in] x1
  !> @param[in] x2
  !> @param[in] wgt
  !-----------------------------------------------------------------------------
  double precision function sigHinc(rnorm, x1, x2, wgt)
    use Process_mod, only : sqrts_proc
    use Constants_mod
    use KinData_mod
    use Scales_mod
    use Channels_mod
    use Parameters_mod
    use LHAPDF_mod
    double precision, intent(in) :: rnorm, x1, x2, wgt

    double precision :: x1curr,x2curr
    common /parfrac/x1curr,x2curr
!$omp threadprivate(/parfrac/)

    double precision :: xreg1,xreg2
    logical :: virtsflag
    common /xregions/xreg1,xreg2,virtsflag
!$omp threadprivate(/xregions/)

    logical :: jflag
    common /jacobian/jflag
!$omp threadprivate(/jacobian/)

    double precision :: rnormcurr,wgtcurr
    common /histonorm/rnormcurr,wgtcurr
!$omp threadprivate(/histonorm/)

    double precision, external :: bino

    ! double precision, parameter :: fbGeV2 = 389379365600d0
    double precision, dimension(nMax_scl) :: g2qcd, partons
    double precision, dimension(nMax_scl) :: kinwt
    double precision :: wt, wtsum
    double precision :: ave, fac, tau
    integer :: ipass, ix, iorder, norder, iscl
    double precision :: as_R, as_F, as_mt
    double precision :: result(1:nMax_scl, 0:3,1:6,1:4)

    ! print*, "sigHinc: ", rnorm, x1, x2, wgt

    current_chan = 1  ! only one channel

    x1curr = x1
    x2curr = x2
    rnormcurr = 1d0  ! override rnorm
    wgtcurr = wgt

    sigHinc = 0d0
    wtsum = 0d0
    kinwt = 0d0
    wt    = 0d0
    fac   = 1d3  ! pb -> fb
    virtsflag=.true.
    jflag=.false.

    !> THIS IS WRONG:
    !> RapidiX needs the full [0,1] range!
    !>  -->  set PDF=0 for x[i]/xreg[i] > 1 instead
    ! if (xreg1_kin < x1) return
    ! if (xreg2_kin < x2) return

    ipass = 1
    call ecuts_vj(0,npar_Hinc,ipass)
    if (ipass==0) return

    tau = (M_Hinc / sqrts_proc)**2
    fac = fac * (-log(tau)) * tau * decayME()


#ifdef RAPIDIX
    call currentContrib_chan(nO=norder)
    call set_order_rapidix(norder)
!$omp critical(rapidix)
    ! print*, "calling evaluate_rapidix ..."
    call evaluate_rapidix(xreg1_kin, xreg2_kin)
    result = 0d0
    do iscl=1,nAct_scl
      as_R  = alphasQ_lhapdf(muR_scl(iscl))
      as_F  = alphasQ_lhapdf(muF_scl(iscl))
      as_mt = as_F  ! no running induced when scales are the same
      if (mt_scheme_param() == 1) then
        !> alphas at mt(mt): running is done inside of rapidix
        as_mt = alphasQ_lhapdf(mt_param())
      end if
      call set_scales_rapidix(muR_scl(iscl), muF_scl(iscl), as_F, as_mt, as_R)
      call get_results_rapidix(result(iscl,:,:,:))
    end do
  ! print*, "done: evaluate_rapidix"
!$omp end critical(rapidix)
#endif


    do iorder=0,3
    ! print*, "iorder = ", iorder, get_order_inc_chan(iorder)
    if (.not.get_order_inc_chan(iorder)) cycle
    do ix=1,4
      ! print*, "ix = ", ix

      !--- g g -> H + X
      ip1_chan = 0
      ip2_chan = 0
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,1,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "g g -> H + X: ", wt
      wtsum = wtsum + wt

      !--- q g -> H + X
      ip1_chan = 30
      ip2_chan = 0
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,2,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "q g -> H + X: ", wt
      wtsum = wtsum + wt

      !--- qb g -> H + X
      ip1_chan = -30
      ip2_chan = 0
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,2,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "qb g -> H + X: ", wt
      wtsum = wtsum + wt

      !--- g q -> H + X
      ip1_chan = 0
      ip2_chan = 30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,3,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "g q -> H + X: ", wt
      wtsum = wtsum + wt

      !--- g qb -> H + X
      ip1_chan = 0
      ip2_chan = -30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,3,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "g qb -> H + X: ", wt
      wtsum = wtsum + wt

      !--- q qb -> H + X
      ip1_chan = 30
      ip2_chan = -30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,4,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "q qb -> H + X: ", wt
      wtsum = wtsum + wt

      !--- qb q -> H + X
      ip1_chan = -30
      ip2_chan = 30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,4,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "qb q -> H + X: ", wt
      wtsum = wtsum + wt

      !--- q q -> H + X
      ip1_chan = 30
      ip2_chan = 30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,5,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "q q -> H + X: ", wt
      wtsum = wtsum + wt

      !--- qb qb -> H + X
      ip1_chan = -30
      ip2_chan = -30
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,5,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "qb qb -> H + X: ", wt
      wtsum = wtsum + wt

      !--- q Qb -> H + X
      ip1_chan = 32
      ip2_chan = -32
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,6,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "q Qb -> H + X: ", wt
      wtsum = wtsum + wt

      !--- qb Q -> H + X
      ip1_chan = -32
      ip2_chan = 32
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,6,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "qb Q -> H + X: ", wt
      wtsum = wtsum + wt

      !--- q Q -> H + X
      ip1_chan = 32
      ip2_chan = 32
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,6,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "q Q -> H + X: ", wt
      wtsum = wtsum + wt

      !--- qb Qb -> H + X
      ip1_chan = -32
      ip2_chan = -32
      call get_lumi_scl(ix, ip1_chan, ip2_chan, partons)
      kinwt = fac * result(:,iorder,6,ix)
      wt = bino(ix, partons, kinwt)
      ! print*, "qb Qb -> H + X: ", wt
      wtsum = wtsum + wt

    end do  ! end do ix
    end do  ! end do iorder

    sigHinc = wtsum

    ! print*, "sigHinc = ", sigHinc
    ! read(*,*)
  contains

    double precision function decayME()
      double precision :: Htautau, HZg2l1g0, HZZ2e2mu0, HZZ4e0, HWW2l2n0
      select case (decay_Hinc)
      case (decay_H_Hinc)
        decayME = 1d0
      case (decay_Hto2p_Hinc)
        decayME = 2.35d-3
      case (decay_Hto2tau_Hinc)
        decayME = Htautau(3,4,npar_Hinc)
      case (decay_Hto2l1p_Hinc)
        decayME = HZg2l1g0(3,4,5,npar_Hinc)
      case (decay_Hto2e2mu_Hinc)
        decayME = HZZ2e2mu0(3,4,5,6,npar_Hinc)
      case (decay_Hto4e_Hinc)
        decayME = HZZ4e0(3,4,5,6,npar_Hinc)
      case (decay_Hto2l2n_Hinc)
        decayME = HWW2l2n0(3,4,5,6,npar_Hinc)
      case default
        print*, "sigHinc:decayME: invalid decay type", decay_Hinc
        stop
      end select
    end function decayME

  end function sigHinc


end module sigHinc_mod


