!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: sigHqt [external members: `_Hqt`]
!
! DESCRIPTION:
!> Module providing qT-subtraction components for Higgs-production uo to N3LO
!-------------------------------------------------------------------------------

!#define HQTN3LO

!-------------------------------------------------------------------------------
module sigHqt_mod
!-------------------------------------------------------------------------------
  use Constants_mod
  use Scales_mod, only : nMax_scl
  implicit none
  private

  public :: init_Hqt
  public :: set_qtcut_Hqt, qtcut_Hqt
  public :: set_boost_Hqt
  public :: set_kinematics_Hqt, sample_recoil_Hqt
  !> the differential cross section to be integrated
  public :: sigHqt

  !> run settings
  double precision :: qtcut_ = 1d0  ! [GeV]
  integer :: qtcutmultiplier_ = 1  ! multiplier of qtcut_ if boost > 1, default value is 1
  logical :: qboost_ = .true.

  !----- local setting for the cross section:

  !> the higgs-boson decay
  !>  0) no decay
  !>  1) H -> 2 photons
  !>  2) H -> 2 tau
  !>  3) H -> 2 electrons 2 muons
  integer, parameter :: decay_H_        = 0
  integer, parameter :: decay_Hto2p_    = 1
  integer, parameter :: decay_Hto2tau_  = 2
  integer, parameter :: decay_Hto2l1p_  = 3
  integer, parameter :: decay_Hto2e2mu_ = 4
  integer, parameter :: decay_Hto4e_    = 5
  integer, parameter :: decay_Hto2l2n_  = 6
  integer :: decay_ = -1
  integer :: npar_  = -1
  !> vacuum-expectation-value of the higgs potential [GeV]
  double precision :: vev_ = 246.21846d0  ! GeV
  integer, parameter :: nc = 3
  integer, parameter :: nf = 5

  !> variables setting the kinematics
  double precision :: Q_, Y_
!$OMP THREADPRIVATE(Q_, Y_)
  !> variables for the recoil kinematics (& jacobian weight)
  double precision :: QT_, PHI_, WGT_
!$OMP THREADPRIVATE(QT_, PHI_, WGT_)

  !> variables setting the channels
  !>  we use bit flags to switch on/off partonic channels
  integer, parameter :: ibit_gg_   = 0
  integer, parameter :: ibit_qg_   = 1
  integer, parameter :: ibit_qbg_  = 2
  integer, parameter :: ibit_gq_   = 3
  integer, parameter :: ibit_gqb_  = 4
  integer, parameter :: ibit_qq_   = 5
  integer, parameter :: ibit_qqb_  = 6
  integer, parameter :: ibit_qbq_  = 7
  integer, parameter :: ibit_qbqb_ = 8
  integer, parameter :: nbit_chan_ = 8

  !> local cache of variables that are expensive to compute
  double precision :: x1_, x2_, z1_, z2_
  double precision :: omx1_, omx2_, dlomx1_, dlomx2_
  !# double precision :: denz1_, denz2_, denomz1_, denomz2_, denopz1_, denopz2_
  double precision, dimension(nMax_scl) :: dlQTcut_, dlQ_, dlQT_, dlMT_, dlR_
  double precision, dimension(nMax_scl) :: as_, h1_, h2_, h3_
  double precision :: hra1_(-1:1),hra2_(-1:1,-1:1),hra3_(-1:1,-1:1,-1:1),  &
                    & hra4_(-1:1,-1:1,-1:1,-1:1),  &
                    & hra5_(-1:1,-1:1,-1:1,-1:1,-1:1)
  double precision :: hrb1_(-1:1),hrb2_(-1:1,-1:1),hrb3_(-1:1,-1:1,-1:1),  &
                    & hrb4_(-1:1,-1:1,-1:1,-1:1),  &
                    & hrb5_(-1:1,-1:1,-1:1,-1:1,-1:1)
!$OMP THREADPRIVATE(x1_, x2_, z1_, z2_)
!$OMP THREADPRIVATE(omx1_, omx2_, dlomx1_, dlomx2_)
!# $OMP THREADPRIVATE(denz1_, denz2_, denomz1_, denomz2_, denopz1_, denopz2_)
!$OMP THREADPRIVATE(dlQTcut_, dlQ_, dlQT_, dlR_, dlMT_)
!$OMP THREADPRIVATE(as_, h1_, h2_, h3_)
!$OMP THREADPRIVATE(hra1_, hra2_, hra3_, hra4_, hra5_)
!$OMP THREADPRIVATE(hrb1_, hrb2_, hrb3_, hrb4_, hrb5_)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set parmeters of the cross section
  !-----------------------------------------------------------------------------
  subroutine init_Hqt()
    use Process_mod, only : name_proc
    integer :: i

    print*, "init_Hqt: setting up SCET..."

    !> initialise the decay settings
    select case (name_proc)
    case ('HQT')
      decay_ = decay_H_
      npar_  = 4
    case ('HTO2PQT')
      decay_ = decay_Hto2p_
      npar_  = 4
    case ('HTO2TAUQT')
      decay_ = decay_Hto2tau_
      npar_  = 4
    case ('HTO2L1PQT')
      decay_ = decay_Hto2l1p_
      npar_  = 5
    case ('HTO2E2MUQT')
      decay_ = decay_Hto2e2mu_
      npar_  = 6
    case ('HTO4EQT')
      decay_ = decay_Hto4e_
      npar_  = 6
    case ('HTO2L2NQT')
      decay_ = decay_Hto2l2n_
      npar_  = 6
    case default
      print*, "init_Hqt: unknown process name: ", name_proc
      stop
    end select

  end subroutine init_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the upper value for the qtcut value (cumulant)
  !
  !> @param[in] Q the invariant mass
  !-----------------------------------------------------------------------------
  subroutine set_qtcut_Hqt(qtcut)
    double precision, intent(in) :: qtcut
    if (qtcut <= 0d0) then
      print*, "set_qtcut_Hqt: invalid qt cut:", qtcut
      stop
    end if
    print*, "set_qtcut_Hqt: ", qtcut
    ! read(*,*)
    qtcut_ = qtcut
    ! this is the value used in accumulated contribution
    ! never book histogram selector with pth lower than this value
  end subroutine set_qtcut_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter for the qt cut value
  !-----------------------------------------------------------------------------
  double precision function qtcut_Hqt()
    qtcut_Hqt = qtcut_*dble(qtcutmultiplier_)
  end function qtcut_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the boost and the associated local counterterm
  !> If multiplier is presented, we expect small qtcut_ value:
  !>    The allowed pT sampling is [qtcut_, qtcut_*qtcutmultiplier_]
  !>    This is to avoid large power correction cancellation in the zero bin in region [tcut,qtcut_]
  !>    By only taking accumHqt_ result in the zero bin and only bin diffHqt_ in the [qtcut_, qtcut_*qtcutmultiplier_] region.
  !>    To enable histograms from multiple qtcut_ values in one run with stable zero bin
  !
  !> @param[in] active
  !> @param[in] multiplier
  !-----------------------------------------------------------------------------
  subroutine set_boost_Hqt(active,multiplier)
    logical, intent(in) :: active
    integer, intent(in), optional :: multiplier
    qboost_ = active
    if ( (present(multiplier)) .and. qboost_) then
      if (multiplier.le.1) then
         write(*,*) "Invalid (<1) boost multiplier in set_boost_Hqt: ", multiplier
         stop
      else
        qtcutmultiplier_ = multiplier
      endif
    end if
    print*, "set_boost_Hqt qboost_: ", qboost_
    print*, "set_boost_Hqt qtcutmultiplier_: ", qtcutmultiplier_
  end subroutine set_boost_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the variables that define the Born kinematics
  !> of the colour-less final state (= H)
  !
  !> @param[in] Q the invariant mass
  !> @param[in] Y the rapidity
  !-----------------------------------------------------------------------------
  subroutine set_kinematics_Hqt(Q, Y)
    double precision, intent(in) :: Q, Y
    ! print*, "set_kinematics_Hqt: ", Q, Y
    Q_   = Q
    Y_   = Y
  end subroutine set_kinematics_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> sample the recoil kinematics (qt, phi)
  !> of the colour-less final state (= H)
  !
  !> @param[in] r1 random number [0,1] for qT  sampling in  [0, qtcut_] or [qtcut_,qtcut_*boostmultiply] region
  !> @param[in] r2 random number [0,1] for phi sampling in  [0, 2pi]
  !-----------------------------------------------------------------------------
  subroutine sample_recoil_Hqt(r1, r2)
    use Run_mod, only : tcut_run
    double precision, intent(in) :: r1, r2
    double precision :: qtmin,qtmax
    ! print*, "sample_recoil_Hqt: ", r1, r2
    qtmax = qtcut_Hqt()
    if (qtmax.ne.qtcut_) then
      !> qtcutmultiplier_ is active and larger than 1, only sample QT_ from qtcut_
      qtmin = max(tcut_run * Q_, qtcut_)
    else
      qtmin = tcut_run * Q_
    endif

    WGT_ = 1d0
    !>----- QT
    !> linear mapping
    ! QT_  = r1 * qtcut_
    ! WGT_ = WGT_ * qtcut_
    !> log mapping (requires non-vanishing lower bound)
    QT_ = qtmin * (qtmax/qtmin)**r1
    WGT_ = WGT_ * QT_ * log(qtmax/qtmin)
    !>----- PHI
    PHI_ = r2 * twopi
    !> this is WRONG, as PHI was inclusive!
    !> WGT_ = WGT_ * twopi
    ! print*, " > QT  = ", QT_
    ! print*, " > PHI = ", PHI_
  end subroutine sample_recoil_Hqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> the differential cross section
  !
  !> @param[in] rnorm
  !> @param[in] x1
  !> @param[in] x2
  !> @param[in] wgt
  !-----------------------------------------------------------------------------
  double precision function sigHqt(rnorm, x1_in, x2_in, wgt)
    use KinData_mod
    use Channels_mod
    use Scales_mod
    double precision, intent(in) :: rnorm, x1_in, x2_in, wgt

    double precision :: x1curr,x2curr
    common /parfrac/x1curr,x2curr
!$omp threadprivate(/parfrac/)

    double precision :: xreg1,xreg2
    logical :: virtsflag
    common /xregions/xreg1,xreg2,virtsflag
!$omp threadprivate(/xregions/)

    logical :: jflag
    common /jacobian/jflag
!$omp threadprivate(/jacobian/)

    double precision :: rnormcurr,wgtcurr
    common /histonorm/rnormcurr,wgtcurr
!$omp threadprivate(/histonorm/)

    double precision, external :: bino

    double precision, dimension(nMax_scl) :: born, kinwt, partons
    double precision :: fac, wt, wtsum
    integer :: ipass, ix, iorder, norder, itype

    ! print*, "sigHqt: ", rnorm, x1_in, x2_in, wgt

    current_chan = 1  ! only one channel

    x1curr = x1_in
    x2curr = x2_in
    rnormcurr = rnorm  ! 1d0  ! <- override?
    wgtcurr = wgt

    sigHqt = 0d0
    wtsum = 0d0
    kinwt = 0d0
    wt    = 0d0
    fac   = 1d0
    virtsflag = .true.
    ! jflag = .false.  ! <- already have the 1/(1-xi) in our expressions
    jflag = .true.  ! <- for stability, we pulled it out

    ! if ( (abs(1d0-xreg1_kin) < 1d-7) .or. (abs(1d0-xreg2_kin) < 1d-7) ) return

    type_loop : do itype=1,3
    ! print*,  "itype = ", itype
    if ( (itype>1) .and. (.not.qboost_) ) cycle type_loop
    select case (itype)
    case (1)
      ipass = 1
      call ecuts_vj(0,npar_,ipass)
      if (ipass==0) cycle type_loop
      fac = 1d0
      !> Born ME & cache varibles depend on scales => AFTER ecuts!
      born = bornME() * alphas_scl**2
      call init_cache_()
      if (qtcut_<QT_) cycle type_loop
      !skip zero pt bin (accumXX) if one uses qtcutmultiplier_ > 1
      !But still need Born ME initialised above for diffXX in case(3)
    case (2)
      !> eveything same as case 1, just include Jac
      !> if sampled QT_ value is larger than qtcut_, we accumulate the weight from diffHqt.inc
      !> if sampled QT_ value is smaller than qtcut_, we subtract the weight from diffHqt.inc,
      !> since it has been included in accumHqt.inc.
      if ( (ipass==0) .or. (qtcut_<QT_) ) cycle type_loop
      fac = - WGT_
    case (3)
      call apply_boost_()  ! must come last & only once!
      ipass = 1
      call ecuts_vj(0,npar_,ipass)
      if ( ipass==0 ) cycle type_loop
      fac = + WGT_
      !> need to recompute scale-dependent ME & cached vars
      born = bornME() * alphas_scl**2
      call init_cache_()
    end select

    order_loop : do iorder=0,3
    ! print*, "iorder = ", iorder, get_order_inc_chan(iorder)
    if (.not.get_order_inc_chan(iorder)) cycle order_loop
    do ix=1,4
      ! print*, "ix = ", ix

      !--- g g -> H + X
      select case(itype)
      case (1)
        continue  !#killed: call accumHqt_gg(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffHqt_gg(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gg_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q g -> H + X
      select case(itype)
      case (1)
        continue  !#killed: call accumHqt_qg(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffHqt_qg(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qg_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt
      call getlumi_(ibit_qbg_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q -> H + X
      select case(itype)
      case (1)
        continue  !#killed: call accumHqt_gq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffHqt_gq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt
      call getlumi_(ibit_gqb_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q q -> H + X
      select case(itype)
      case (1)
        continue  !#killed: call accumHqt_qq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffHqt_qq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qq_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt
      call getlumi_(ibit_qqb_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt
      call getlumi_(ibit_qbq_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt
      call getlumi_(ibit_qbqb_, ix, partons)
      wt = bino(ix, partons, fac * born * kinwt * as_**iorder)
      wtsum = wtsum + wt

    end do  ! end do ix
    end do order_loop
    end do type_loop

    sigHqt = rnorm * wtsum

    !> in case of a NaN, reject the full event
    if ( sigHqt /= sigHqt ) then
      ireject_event_chan = -2
      sigHqt = 0d0
    end if

    ! print*, "sigHqt = ", sigHqt
    ! read(*,*)
    contains

    double precision function bornME()
      !> assemble Born Matrix Element
      !> (stripped by `alphas_scl**2` to make it scalar)
      bornME = (1d0/6d0/pi/vev_)**2 * (nc**2-1)  ! C0
      bornME = bornME * 2D0*(kin(npar_)%s(1,2))**2  !  A2g0H
      bornME = bornME / (8*8*2*2)  ! avg factor (colour & spin)
      bornME = bornME * decayME()
    end function bornME

    double precision function decayME()
      double precision :: Htautau, HZg2l1g0, HZZ2e2mu0, HZZ4e0, HWW2l2n0
      select case (decay_)
      case (decay_H_)
        decayME = 1d0
      case (decay_Hto2p_)
        decayME = 2.35d-3
      case (decay_Hto2tau_)
        decayME = Htautau(3,4,npar_)
      case (decay_Hto2l1p_)
        decayME = HZg2l1g0(3,4,5,npar_)
      case (decay_Hto2e2mu_)
        decayME = HZZ2e2mu0(3,4,5,6,npar_)
      case (decay_Hto4e_)
        decayME = HZZ4e0(3,4,5,6,npar_)
      case (decay_Hto2l2n_)
        decayME = HWW2l2n0(3,4,5,6,npar_)
      case default
        print*, "sigHqt:decayME: invalid decay type", decay_
        stop
      end select
    end function decayME

  end function sigHqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> boost the final state using QT_ & PHI_
  !-----------------------------------------------------------------------------
  subroutine apply_boost_()
    use KinData_mod
    type(Kin_t), pointer :: kd
    integer :: i
    double precision :: Q2, chY, shY, MT
    double precision :: pX(4), pXb(4), r(4)
    double precision :: sXXb, siX, sir

    Q2 = Q_**2
    chY = cosh(Y_)
    shY = sinh(Y_)
    MT  = sqrt( Q2 + QT_**2 )

    pX(1) = 0d0
    pX(2) = 0d0
    pX(3) = Q_ * shY
    pX(4) = Q_ * chY

    pXb(1) = QT_ * cos(PHI_)
    pXb(2) = QT_ * sin(PHI_)
    pXb(3) = MT * shY
    pXb(4) = MT * chY

    r(:) = pX(:) + pXb(:)

    sXXb = 2d0*Q_*(MT+Q_)

    !> prettify code
    kd => kin(npar_)

    do i = 3, npar_
      !> scalar products
      sir = 2d0*dot(kd%p(:,i), r)
      siX = 2d0*dot(kd%p(:,i), pX)
      kd%p(:,i) = kd%p(:,i)  &
                & - sir / sXXb * r(:)  &
                & + siX / Q2 * pXb(:)
    end do

    call fillS_kin(npar_)
    call clearCacheLocal_kin(npar_)

  end subroutine apply_boost_


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the local variables
  !-----------------------------------------------------------------------------
  subroutine init_cache_()
    use KinData_mod, only : x1_kin, x2_kin, xreg1_kin, xreg2_kin
    use Scales_mod
    use Parameters_mod, only : mt_param
    use LHAPDF_mod, only : alphasQ_lhapdf
    integer :: iscl, idmap
    !> dummy arguments for calling hplog5
    integer :: nw  ! max weight for hplog: [1,5]
    complex(8) :: hc1(-1:1),  &
                & hc2(-1:1,-1:1),  &
                & hc3(-1:1,-1:1,-1:1),  &
                & hc4(-1:1,-1:1,-1:1,-1:1),  &
                & hc5(-1:1,-1:1,-1:1,-1:1,-1:1)
    double precision :: hi1(-1:1),  &
                      & hi2(-1:1,-1:1),  &
                      & hi3(-1:1,-1:1,-1:1),  &
                      & hi4(-1:1,-1:1,-1:1,-1:1),  &
                      & hi5(-1:1,-1:1,-1:1,-1:1,-1:1)

    ! print*, "init_cache_"

    x1_ = x1_kin
    x2_ = x2_kin
    z1_ = xreg1_kin
    z2_ = xreg2_kin

    omx1_ = (1D0 - x1_)
    omx2_ = (1D0 - x2_)
    dlomx1_ = log(omx1_)
    dlomx2_ = log(omx2_)

    ! denz1_ = 1d0 / z1_
    ! denz2_ = 1d0 / z2_
    ! denomz1_ = 1D0 / (1D0 - z1_)
    ! denomz2_ = 1D0 / (1D0 - z2_)
    ! denopz1_ = 1D0 / (1D0 + z1_)
    ! denopz2_ = 1D0 / (1D0 + z2_)

    dlQ_(:)     = 0d0
    dlQT_(:)    = 0d0
    dlQTcut_(:) = 0d0
    dlMT_(:)    = 0d0
    do iscl=1,nAct_scl
      idmap = idmap_scl(iscl, withF=.true.)
      if (idmap == iscl) then
        dlQ_(iscl)     =  2D0 * log(Q_         / muF_scl(iscl))
        dlQT_(iscl)    =  2D0 * log(QT_        / muF_scl(iscl))
        dlQTcut_(iscl) =  2D0 * log(qtcut_     / muF_scl(iscl))
        dlMT_(iscl)    = -2D0 * log(mt_param() / muF_scl(iscl))
      else
        dlQ_(iscl)     = dlQ_(idmap)
        dlQT_(iscl)    = dlQT_(idmap)
        dlQTcut_(iscl) = dlQTcut_(idmap)
        dlMT_(iscl)    = dlMT_(idmap)
      end if
    enddo

    dlR_(:)    = 0d0
    do iscl=1,nAct_scl
      dlR_(iscl) = 2D0 * log(muR_scl(iscl) / muF_scl(iscl))
    end do

    as_ = alphas_scl / (4d0*pi)

    !> The Wilson coefficients:
    !>   wilson2 = 1 + 2*3  h1*as + 4*9 h2*as^2 + 8*27  h3*as^3
    h1_ = 11d0/3d0
    h2_ = 1933d0/162d0  &
      & + (19d0 * dlMT_)/18d0  &
      & - (67d0*nf)/108d0  &
      & + (8d0*dlMT_*nf)/27d0
    h3_ = -(1792967d0/69984d0)  &
      & + (5347d0*dlMT_)/972d0  &
      & + (209d0*dlMT_**2)/108d0  &
      & + (493d0*nf)/34992d0  &
      & + (836d0*dlMT_*nf)/729d0  &
      & + (23d0*dlMT_**2*nf)/54d0  &
      & - (6865d0*nf**2)/52488d0 &
      & + (77d0*dlMT_*nf**2)/2916d0  &
      & - (8d0*dlMT_**2*nf**2)/243d0  &
      & + (897943d0*zeta3)/15552d0  &
      & - (110779d0*nf*zeta3)/23328d0

    !> harmonic polylogs
    !> @todo optimise depending on the order that is requested!
    nw = 5

    hra1_ = 0d0
    hra2_ = 0d0
    hra3_ = 0d0
    hra4_ = 0d0
    hra5_ = 0d0
    call hplog5( z1_, nw,  &
               & hc1,   hc2,   hc3,   hc4,   hc5,     &
               & hra1_, hra2_, hra3_, hra4_, hra5_,   &
               & hi1,   hi2,   hi3,   hi4,   hi5,     &
               & -1, +1)

    hrb1_ = 0d0
    hrb2_ = 0d0
    hrb3_ = 0d0
    hrb4_ = 0d0
    hrb5_ = 0d0
    call hplog5( z2_, nw,  &
               & hc1,   hc2,   hc3,   hc4,   hc5,     &
               & hrb1_, hrb2_, hrb3_, hrb4_, hrb5_,   &
               & hi1,   hi2,   hi3,   hi4,   hi5,     &
               & -1, +1)

  end subroutine init_cache_


!> load the qT-accumulated expressions
!#killed: #include "accumHqt.inc"


!> load the qT-differential expressions
!#killed: #include "diffHqt.inc"


  subroutine getlumi_(ichan, ix, partons)
    use Scales_mod
    use Channels_mod
    use CKM_mod
    use Process_mod
    integer, intent(in) :: ichan, ix
    double precision, intent(out), dimension(nMax_scl) :: partons
    double precision, dimension(:,:), pointer :: f1, f2
    double precision :: f1lumi(-6:7,1:nMax_scl), f2lumi(-6:7,1:nMax_scl)

    select case (ix)
    case (1)
       f1(-6:7,1:nMax_scl) => f1_scl
       f2(-6:7,1:nMax_scl) => f2_scl
    case (2)
       f1(-6:7,1:nMax_scl) => f1_scl
       f2(-6:7,1:nMax_scl) => f2x2_scl
    case (3)
       f1(-6:7,1:nMax_scl) => f1x1_scl
       f2(-6:7,1:nMax_scl) => f2_scl
    case (4)
       f1(-6:7,1:nMax_scl) => f1x1_scl
       f2(-6:7,1:nMax_scl) => f2x2_scl
    end select

    f1lumi(-6:7,1:nMax_scl) = f1(-6:7,1:nMax_scl)
    f2lumi(-6:7,1:nMax_scl) = f2(-6:7,1:nMax_scl)

    if ( beam1_proc==pid_ap ) then
      f1lumi(-6,:) = f1( 6,:)
      f1lumi(-5,:) = f1( 5,:)
      f1lumi(-4,:) = f1( 4,:)
      f1lumi(-3,:) = f1( 3,:)
      f1lumi(-2,:) = f1( 2,:)
      f1lumi(-1,:) = f1( 1,:)
      f1lumi( 6,:) = f1(-6,:)
      f1lumi( 5,:) = f1(-5,:)
      f1lumi( 4,:) = f1(-4,:)
      f1lumi( 3,:) = f1(-3,:)
      f1lumi( 2,:) = f1(-2,:)
      f1lumi( 1,:) = f1(-1,:)    
    end if

    if ( beam2_proc==pid_ap ) then
      f2lumi(-6,:) = f2( 6,:)
      f2lumi(-5,:) = f2( 5,:)
      f2lumi(-4,:) = f2( 4,:)
      f2lumi(-3,:) = f2( 3,:)
      f2lumi(-2,:) = f2( 2,:)
      f2lumi(-1,:) = f2( 1,:)
      f2lumi( 6,:) = f2(-6,:)
      f2lumi( 5,:) = f2(-5,:)
      f2lumi( 4,:) = f2(-4,:)
      f2lumi( 3,:) = f2(-3,:)
      f2lumi( 2,:) = f2(-2,:)
      f2lumi( 1,:) = f2(-1,:)    
    end if

    partons = 0d0

    select case (ichan)
    case ( ibit_gg_ )
      ip1_chan = 0
      ip2_chan = 0
      partons(1:nInp_scl) =  &
        &   f1lumi(0,1:nInp_scl)*f2lumi(0,1:nInp_scl)
    case ( ibit_qg_ )
      ip1_chan = 1
      ip2_chan = 0
      partons(1:nInp_scl) =  &
        &   f1lumi(+1,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(+2,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(+3,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(+4,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(+5,1:nInp_scl)*f2lumi(0,1:nInp_scl)
    case ( ibit_qbg_ )
      ip1_chan = -1
      ip2_chan = 0
      partons(1:nInp_scl) =  &
        &   f1lumi(-1,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(0,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(0,1:nInp_scl)
    case ( ibit_gq_ )
      ip1_chan = 0
      ip2_chan = 1
      partons(1:nInp_scl) =  &
        &   f1lumi(0,1:nInp_scl)*f2lumi(+1,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(+2,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(+3,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(+4,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(+5,1:nInp_scl)
    case ( ibit_gqb_ )
      ip1_chan = 0
      ip2_chan = -1
      partons(1:nInp_scl) =  &
        &   f1lumi(0,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(0,1:nInp_scl)*f2lumi(-5,1:nInp_scl)
    case ( ibit_qq_ )
      ip1_chan = 1
      ip2_chan = 1
      partons(1:nInp_scl) =  &
        &   f1lumi(1,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(5,1:nInp_scl)
    case ( ibit_qqb_ )
      ip1_chan = 1
      ip2_chan = -1
      partons(1:nInp_scl) =  &
        &   f1lumi(1,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(1,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(2,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(3,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(4,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(5,1:nInp_scl)*f2lumi(-5,1:nInp_scl)
    case ( ibit_qbq_ )
      ip1_chan = -1
      ip2_chan = 1
      partons(1:nInp_scl) =  &
        &   f1lumi(-1,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(5,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(1,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(3,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(5,1:nInp_scl)
    case ( ibit_qbqb_ )
      ip1_chan = -1
      ip2_chan = -1
      partons(1:nInp_scl) =  &
        &   f1lumi(-1,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(-1,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(-2,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(-3,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(-4,1:nInp_scl)*f2lumi(-5,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(-1,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(-2,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(-3,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(-4,1:nInp_scl)  &
        & + f1lumi(-5,1:nInp_scl)*f2lumi(-5,1:nInp_scl)
    case default
      print*, "sigHqt getlumi_: invalid ichan:", ichan
      stop
    end select
  end subroutine getlumi_


end module sigHqt_mod


