!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: sigWqt [external members: `_Wqt`]
!
! DESCRIPTION:
!> Module providing qT-subtraction components for W-production up to N3LO
!-------------------------------------------------------------------------------

!#define WQTN3LO

!-------------------------------------------------------------------------------
module sigWqt_mod
!-------------------------------------------------------------------------------
  use Constants_mod
  use Scales_mod, only : nMax_scl
  implicit none
  private

  public :: init_Wqt
  public :: set_qtcut_Wqt, qtcut_Wqt
  public :: set_boost_Wqt
  public :: set_kinematics_Wqt, sample_recoil_Wqt
  !> the differential cross section to be integrated
  public :: sigWqt

  !> run settings
  double precision :: qtcut_ = 1d0  ! [GeV]
  integer :: qtcutmultiplier_ = 1  ! multiplier of qtcut_ if boost > 1, default value is 1
  logical :: qboost_ = .true.

  !----- local setting for the cross section:

  integer, parameter :: npar_  = 4
  integer, parameter :: nc = 3
  integer, parameter :: nf = 5
  integer, parameter :: nup = 2
  integer, parameter :: ndn = 3

  !> variables setting the kinematics
  double precision :: Q_, Y_
!$OMP THREADPRIVATE(Q_, Y_)
  !> variables for the recoil kinematics (& jacobian weight)
  double precision :: QT_, PHI_, WGT_
!$OMP THREADPRIVATE(QT_, PHI_, WGT_)

  !> variables setting the channels
  !>  we use bit flags to switch on/off partonic channels
  integer, parameter :: ibit_gg_     = 0   !fAgBg
  integer, parameter :: ibit_gq_     = 1   !fAgBq
  integer, parameter :: ibit_gq1_    = 2   !fAgBq1
  integer, parameter :: ibit_gqp_    = 3   !fAgBqp
  integer, parameter :: ibit_qg_     = 4   !fAqBg
  integer, parameter :: ibit_q1g_    = 5   !fAq1Bg
  integer, parameter :: ibit_qpg_    = 6   !fAqpBg
  integer, parameter :: ibit_gqb_    = 7   !fAgBqb
  integer, parameter :: ibit_gq1b_   = 8   !fAgBq1b
  integer, parameter :: ibit_gqbp_   = 9   !fAgBqbp
  integer, parameter :: ibit_qbg_    = 10  !fAqbBg
  integer, parameter :: ibit_q1bg_   = 11  !fAq1bBg
  integer, parameter :: ibit_qbpg_   = 12  !fAqbpBg
  integer, parameter :: ibit_qq1b_   = 13  !fAqBq1b
  integer, parameter :: ibit_qqbp_   = 14  !fAqBq1bp
  integer, parameter :: ibit_qpq1b_  = 15  !fAqpBq1b
  integer, parameter :: ibit_q1bq_   = 16  !fAq1bBq
  integer, parameter :: ibit_qbpq_   = 17  !fAq1bpBq
  integer, parameter :: ibit_q1bqp_  = 18  !fAq1bBqp
  integer, parameter :: ibit_q1q_    = 19  !fAq1Bq
  integer, parameter :: ibit_qpq_    = 20  !fAq1pBq
  integer, parameter :: ibit_qq1_    = 21  !fAqBq1
  integer, parameter :: ibit_qqp_    = 22  !fAqBq1p
  integer, parameter :: ibit_q1bqb_  = 23  !fAq1bBqb
  integer, parameter :: ibit_q1bqbp_ = 24  !fAq1bBqbp
  integer, parameter :: ibit_qbq1b_  = 25  !fAqbBq1b
  integer, parameter :: ibit_qbpq1b_ = 26  !fAqbpBq1b
  integer, parameter :: nbit_chan_   = 26

  !> local cache of variables that are expensive to compute
  double precision :: x1_, x2_, z1_, z2_
  double precision :: omx1_, omx2_, dlomx1_, dlomx2_
  !# double precision :: denz1_, denz2_, denomz1_, denomz2_, denopz1_, denopz2_
  double precision, dimension(nMax_scl) :: dlQTcut_, dlQ_, dlQT_, dlR_
  double precision, dimension(nMax_scl) :: as_
  double precision :: hra1_(-1:1),hra2_(-1:1,-1:1),hra3_(-1:1,-1:1,-1:1),  &
                    & hra4_(-1:1,-1:1,-1:1,-1:1),  &
                    & hra5_(-1:1,-1:1,-1:1,-1:1,-1:1)
  double precision :: hrb1_(-1:1),hrb2_(-1:1,-1:1),hrb3_(-1:1,-1:1,-1:1),  &
                    & hrb4_(-1:1,-1:1,-1:1,-1:1),  &
                    & hrb5_(-1:1,-1:1,-1:1,-1:1,-1:1)
!$OMP THREADPRIVATE(x1_, x2_, z1_, z2_)
!$OMP THREADPRIVATE(omx1_, omx2_, dlomx1_, dlomx2_)
!# $OMP THREADPRIVATE(denz1_, denz2_, denomz1_, denomz2_, denopz1_, denopz2_)
!$OMP THREADPRIVATE(dlQTcut_, dlQ_, dlQT_, dlR_)
!$OMP THREADPRIVATE(as_)
!$OMP THREADPRIVATE(hra1_, hra2_, hra3_, hra4_, hra5_)
!$OMP THREADPRIVATE(hrb1_, hrb2_, hrb3_, hrb4_, hrb5_)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set parmeters of the cross section
  !-----------------------------------------------------------------------------
  subroutine init_Wqt()
    use Process_mod, only : name_proc
    integer :: i

    print*, "init_Wqt: setting up SCET... ", name_proc

  end subroutine init_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the upper value for the qtcut value (cumulant)
  !
  !> @param[in] Q the invariant mass
  !-----------------------------------------------------------------------------
  subroutine set_qtcut_Wqt(qtcut)
    double precision, intent(in) :: qtcut
    if (qtcut <= 0d0) then
      print*, "set_qtcut_Wqt: invalid qt cut:", qtcut
      stop
    end if
    print*, "set_qtcut_Wqt: ", qtcut
    ! read(*,*)
    qtcut_ = qtcut
    ! this is the value used in accumulated contribution
    ! never book histogram selector with ptw lower than this value
  end subroutine set_qtcut_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> getter for the qt cut value
  !-----------------------------------------------------------------------------
  double precision function qtcut_Wqt()
    qtcut_Wqt = qtcut_*dble(qtcutmultiplier_)
  end function qtcut_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> switch on/off the boost and the associated local counterterm
  !> If multiplier is presented, we expect small qtcut_ value:
  !>    The allowed pT sampling is [qtcut_, qtcut_*qtcutmultiplier_]
  !>    This is to avoid large power correction cancellation in the zero bin in region [tcut,qtcut_]
  !>    By only taking accumWqt_ result in the zero bin and only bin diffWqt_ in the [qtcut_, qtcut_*qtcutmultiplier_] region.
  !>    To enable histograms from multiple qtcut_ values in one run with stable zero bin
  !
  !> @param[in] active
  !> @param[in] multiplier
  !-----------------------------------------------------------------------------
  subroutine set_boost_Wqt(active,multiplier)
    logical, intent(in) :: active
    integer, intent(in), optional :: multiplier
    qboost_ = active
    if ( (present(multiplier)) .and. qboost_) then
      if (multiplier.le.1) then
         write(*,*) "Invalid (<1) boost multiplier in set_boost_Wqt: ", multiplier
         stop
      else
        qtcutmultiplier_ = multiplier
      endif
    end if
    print*, "set_boost_Wqt qboost_: ", qboost_
    print*, "set_boost_Wqt qtcutmultiplier_: ", qtcutmultiplier_
  end subroutine set_boost_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the variables that define the Born kinematics
  !> of the colour-less final state (= H)
  !
  !> @param[in] Q the invariant mass
  !> @param[in] Y the rapidity
  !-----------------------------------------------------------------------------
  subroutine set_kinematics_Wqt(Q, Y)
    double precision, intent(in) :: Q, Y
    ! print*, "set_kinematics_Wqt: ", Q, Y
    Q_   = Q
    Y_   = Y
  end subroutine set_kinematics_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> sample the recoil kinematics (qt, phi)
  !> of the colour-less final state (= H)
  !
  !> @param[in] r1 random number [0,1] for qT  sampling in  [0, qtcut_]
  !> @param[in] r2 random number [0,1] for phi sampling in  [0, 2pi]
  !-----------------------------------------------------------------------------
  subroutine sample_recoil_Wqt(r1, r2)
    use Run_mod, only : tcut_run
    double precision, intent(in) :: r1, r2
    double precision :: qtmin, qtmax
    ! print*, "sample_recoil_Wqt: ", r1, r2
    qtmax = qtcut_Wqt()
    if (qtmax.ne.qtcut_) then
      !qtcutmultiplier_ is active and larger than 1, only sample QT_ from qtcut_
        qtmin = max(tcut_run * Q_, qtcut_)
    else
        qtmin = tcut_run * Q_
    endif

    WGT_ = 1d0
    !>----- QT
    !> linear mapping
    ! QT_  = r1 * qtcut_
    ! WGT_ = WGT_ * qtcut_
    !> log mapping (requires non-vanishing lower bound)
    QT_ = qtmin * (qtmax/qtmin)**r1
    WGT_ = WGT_ * QT_ * log(qtmax/qtmin)
    !>----- PHI
    PHI_ = r2 * twopi
    !> this is WRONG, as PHI was inclusive!
    !> WGT_ = WGT_ * twopi
    ! print*, " > qtmin = ", qtmin
    ! print*, " > QT    = ", QT_
    ! print*, " > PHI   = ", PHI_
  end subroutine sample_recoil_Wqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> the differential cross section
  !
  !> @param[in] rnorm
  !> @param[in] x1
  !> @param[in] x2
  !> @param[in] wgt
  !-----------------------------------------------------------------------------
  double precision function sigWqt(rnorm, x1_in, x2_in, wgt)
    use KinData_mod
    use Channels_mod
    use Scales_mod
    use Parameters_mod, only : alpha_param
    double precision, intent(in) :: rnorm, x1_in, x2_in, wgt

    double precision :: x1curr,x2curr
    common /parfrac/x1curr,x2curr
!$omp threadprivate(/parfrac/)

    double precision :: xreg1,xreg2
    logical :: virtsflag
    common /xregions/xreg1,xreg2,virtsflag
!$omp threadprivate(/xregions/)

    logical :: jflag
    common /jacobian/jflag
!$omp threadprivate(/jacobian/)

    double precision :: rnormcurr,wgtcurr
    common /histonorm/rnormcurr,wgtcurr
!$omp threadprivate(/histonorm/)

    !> the flavour-dependence of the Born MEs & couplings
    double precision, external :: B0g0W
    integer :: nfB
    common /BZFlav/ nfB
!$omp threadprivate(/BZFlav/)
    double precision :: facB, born_qqb, born_qbq
    double precision :: amzW,stw,ewfac
    common /eweakW/ amzW,stw,ewfac
!$omp threadprivate(/eweakW/)

    double precision, external :: bino

    double precision, dimension(nMax_scl) :: kinwt, partons
    double precision :: fac, wt, wtsum
    integer :: ipass, ix, iorder, norder, itype

    ! print*, "sigWqt: ", rnorm, x1_in, x2_in, wgt
    current_chan = 1  ! only one channel

    x1curr = x1_in
    x2curr = x2_in
    rnormcurr = rnorm  ! 1d0  ! <- override?
    wgtcurr = wgt

    sigWqt = 0d0
    wtsum = 0d0
    kinwt = 0d0
    wt    = 0d0
    fac   = 1d0
    virtsflag = .true.
    ! jflag = .false.  ! <- already have the 1/(1-xi) in our expressions
    jflag = .true.  ! <- for stability, we pulled it out

    ! if ( (abs(1d0-xreg1_kin) < 1d-7) .or. (abs(1d0-xreg2_kin) < 1d-7) ) return

    type_loop : do itype=1,3
    ! print*,  "itype = ", itype
    if ( (itype>1) .and. ( .not.qboost_ ) ) cycle type_loop
    select case (itype)
    case (1)
      ipass = 1
      call ecuts_vj(0,npar_,ipass)
      if (ipass==0) cycle type_loop
      fac = 1d0
      !> cache varibles depend on scales => AFTER ecuts!
      call init_cache_()
      !> cache the Born MEs (diffevernt flavours & permutations):
      facB = (4d0*pi*alpha_param()/stw)**2*nc / (nc*nc*2*2)
      nfB = 1  !  u db -> W+
      born_qqb = facB * B0g0W(1,2,3,4)
      born_qbq = facB * B0g0W(2,1,3,4)
      if (qtcut_<QT_) cycle type_loop
      !skip zero pt bin (accumXX) if one uses qtcutmultiplier_ > 1
      !But still need Born ME initialised above for diffXX in case(3)
    case (2)
      !> eveything same as case 1, just include Jac
      !> if sampled QT_ value is larger than qtcut_, we accumulate the weight from diffWqt.inc
      !> if sampled QT_ value is smaller than qtcut_, we subtract the weight from diffWqt.inc,
      !> since it has been included in accumWqt.inc.
      if ( (ipass==0) .or. (qtcut_<QT_) ) cycle type_loop
      fac = - WGT_
    case (3)
      call apply_boost_()  ! must come last & only once!
      ipass = 1
      call ecuts_vj(0,npar_,ipass)
      if ( ipass==0 ) cycle type_loop
      fac = + WGT_
      !> need to recompute scale-dependent cached vars
      call init_cache_()
    end select



    order_loop : do iorder=0,3
    ! print*, "iorder = ", iorder, get_order_inc_chan(iorder)
    if (.not.get_order_inc_chan(iorder)) cycle order_loop
    do ix=1,4
      ! print*, "ix = ", ix

      !--- q q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qq1b_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qq1b_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- q q1b -> W + X  (this is zero up to N3LO)
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qq1b_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qq1b_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qq1b_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q qbp -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qqbp_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qqbp_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qqbp_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- q qbp -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qqbp_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qqbp_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qqbp_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- qp q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qpq1b_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qpq1b_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qpq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- qp q1b -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qpq1b_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qpq1b_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qpq1b_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      ! !--- q1b q -> W + X  (this is zero up to N3LO)
      ! select case(itype)
      ! case (1)
      !   call accumWqt_q1bq_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_q1bq_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_q1bq_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q1b q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bq_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bq_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bq_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- q1b q' -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_q1bqp_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_q1bqp_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_q1bqp_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q1b q' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bqp_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bqp_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bqp_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- qb' q -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qbpq_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qbpq_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qbpq_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- qb' q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbpq_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbpq_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbpq_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q1 g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1g_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1g_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1g_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q1 g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1g_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1g_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1g_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q' g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qpg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qpg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qpg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q' g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qpg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qpg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qpg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q1 -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq1_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq1_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq1_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q1 -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq1_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq1_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq1_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqp_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqp_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqp_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqp_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqp_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqp_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q1b g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q1b g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb' g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbpg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbpg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbpg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb' g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbpg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbpg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbpg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g qb -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqb_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqb_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqb_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g qb -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqb_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqb_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqb_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq1b_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq1b_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gq1b_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gq1b_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g qb' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqbp_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqbp_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqbp_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g qb' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gqbp_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gqbp_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gqbp_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gg_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gg_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gg_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- g g -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_gg_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_gg_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_gg_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q q1 -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qq1_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qq1_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qq1_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- q q1 -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qq1_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qq1_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qq1_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      ! !--- q1 q -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_q1q_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_q1q_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_q1q_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q1 q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1q_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1q_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1q_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- q qp -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qqp_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qqp_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qqp_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- q qp -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qqp_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qqp_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qqp_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      ! !--- qp q -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qpq_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qpq_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qpq_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- qp q -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qpq_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qpq_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qpq_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbq1b_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbq1b_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- qb q1b -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qbq1b_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qbq1b_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qbq1b_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      ! !--- q1b qb -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_q1bqb_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_q1bqb_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_q1bqb_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q1b qb -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bqb_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bqb_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bqb_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

      !--- qb' q1b -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_qbpq1b_qqb(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_qbpq1b_qqb(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_qbpq1b_, ix, partons)
      wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      wtsum = wtsum + wt

      ! !--- qb' q1b -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_qbpq1b_qbq(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_qbpq1b_qbq(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_qbpq1b_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      ! !--- q1b qb' -> W + X
      ! select case(itype)
      ! case (1)
      !   call accumWqt_q1bqbp_qqb(ix, iorder, kinwt)
      ! case (2,3)
      !   call diffWqt_q1bqbp_qqb(ix, iorder, kinwt)
      ! end select
      ! call getlumi_(ibit_q1bqbp_, ix, partons)
      ! wt = bino(ix, partons, fac * born_qqb * kinwt * as_**iorder)
      ! wtsum = wtsum + wt

      !--- q1b qb' -> W + X
      select case(itype)
      case (1)
        continue  !#killed: call accumWqt_q1bqbp_qbq(ix, iorder, kinwt)
      case (2,3)
        continue  !#killed: call diffWqt_q1bqbp_qbq(ix, iorder, kinwt)
      end select
      call getlumi_(ibit_q1bqbp_, ix, partons)
      wt = bino(ix, partons, fac * born_qbq * kinwt * as_**iorder)
      wtsum = wtsum + wt

    end do  ! end do ix
    end do order_loop
    end do type_loop

    sigWqt = rnorm * wtsum

    !> in case of a NaN, reject the full event
    if ( sigWqt /= sigWqt ) then
      ireject_event_chan = -2
      sigWqt = 0d0
    end if

    ! print*, "sigWqt = ", sigWqt
    ! read(*,*)
  end function sigWqt


  !-----------------------------------------------------------------------------
  !> @brief
  !> boost the final state using QT_ & PHI_
  !-----------------------------------------------------------------------------
  subroutine apply_boost_()
    use KinData_mod
    type(Kin_t), pointer :: kd
    integer :: i
    double precision :: Q2, chY, shY, MT
    double precision :: pX(4), pXb(4), r(4)
    double precision :: sXXb, siX, sir

    Q2 = Q_**2
    chY = cosh(Y_)
    shY = sinh(Y_)
    MT  = sqrt( Q2 + QT_**2 )

    pX(1) = 0d0
    pX(2) = 0d0
    pX(3) = Q_ * shY
    pX(4) = Q_ * chY

    pXb(1) = QT_ * cos(PHI_)
    pXb(2) = QT_ * sin(PHI_)
    pXb(3) = MT * shY
    pXb(4) = MT * chY

    r(:) = pX(:) + pXb(:)

    sXXb = 2d0*Q_*(MT+Q_)

    !> prettify code
    kd => kin(npar_)

    do i = 3, npar_
      !> scalar products
      sir = 2d0*dot(kd%p(:,i), r)
      siX = 2d0*dot(kd%p(:,i), pX)
      kd%p(:,i) = kd%p(:,i)  &
                & - sir / sXXb * r(:)  &
                & + siX / Q2 * pXb(:)
    end do

    call fillS_kin(npar_)
    call clearCacheLocal_kin(npar_)

  end subroutine apply_boost_


  !-----------------------------------------------------------------------------
  !> @brief
  !> initialise the local variables
  !-----------------------------------------------------------------------------
  subroutine init_cache_()
    use KinData_mod, only : x1_kin, x2_kin, xreg1_kin, xreg2_kin
    use Scales_mod
    use Parameters_mod, only : mt_param
    use LHAPDF_mod, only : alphasQ_lhapdf
    integer :: iscl, idmap
    !> dummy arguments for calling hplog5
    integer :: nw  ! max weight for hplog: [1,5]
    complex(8) :: hc1(-1:1),  &
                & hc2(-1:1,-1:1),  &
                & hc3(-1:1,-1:1,-1:1),  &
                & hc4(-1:1,-1:1,-1:1,-1:1),  &
                & hc5(-1:1,-1:1,-1:1,-1:1,-1:1)
    double precision :: hi1(-1:1),  &
                      & hi2(-1:1,-1:1),  &
                      & hi3(-1:1,-1:1,-1:1),  &
                      & hi4(-1:1,-1:1,-1:1,-1:1),  &
                      & hi5(-1:1,-1:1,-1:1,-1:1,-1:1)

    ! print*, "init_cache_"

    x1_ = x1_kin
    x2_ = x2_kin
    z1_ = xreg1_kin
    z2_ = xreg2_kin

    omx1_ = (1D0 - x1_)
    omx2_ = (1D0 - x2_)
    dlomx1_ = log(omx1_)
    dlomx2_ = log(omx2_)

    ! denz1_ = 1d0 / z1_
    ! denz2_ = 1d0 / z2_
    ! denomz1_ = 1D0 / (1D0 - z1_)
    ! denomz2_ = 1D0 / (1D0 - z2_)
    ! denopz1_ = 1D0 / (1D0 + z1_)
    ! denopz2_ = 1D0 / (1D0 + z2_)

    dlQ_(:)     = 0d0
    dlQT_(:)    = 0d0
    dlQTcut_(:) = 0d0
    do iscl=1,nAct_scl
      idmap = idmap_scl(iscl, withF=.true.)
      if (idmap == iscl) then
        dlQ_(iscl)     =  2D0 * log(Q_     / muF_scl(iscl))
        dlQT_(iscl)    =  2D0 * log(QT_    / muF_scl(iscl))
        dlQTcut_(iscl) =  2D0 * log(qtcut_ / muF_scl(iscl))
      else
        dlQ_(iscl)     = dlQ_(idmap)
        dlQT_(iscl)    = dlQT_(idmap)
        dlQTcut_(iscl) = dlQTcut_(idmap)
      end if
    enddo

    dlR_(:)    = 0d0
    do iscl=1,nAct_scl
      dlR_(iscl) = 2D0 * log(muR_scl(iscl) / muF_scl(iscl))
    end do

    as_ = alphas_scl / (4d0*pi)

    !> harmonic polylogs
    !> @todo optimise depending on the order that is requested!
    nw = 5

    hra1_ = 0d0
    hra2_ = 0d0
    hra3_ = 0d0
    hra4_ = 0d0
    hra5_ = 0d0
    call hplog5( z1_, nw,  &
               & hc1,   hc2,   hc3,   hc4,   hc5,     &
               & hra1_, hra2_, hra3_, hra4_, hra5_,   &
               & hi1,   hi2,   hi3,   hi4,   hi5,     &
               & -1, +1)

    hrb1_ = 0d0
    hrb2_ = 0d0
    hrb3_ = 0d0
    hrb4_ = 0d0
    hrb5_ = 0d0
    call hplog5( z2_, nw,  &
               & hc1,   hc2,   hc3,   hc4,   hc5,     &
               & hrb1_, hrb2_, hrb3_, hrb4_, hrb5_,   &
               & hi1,   hi2,   hi3,   hi4,   hi5,     &
               & -1, +1)

  end subroutine init_cache_


!> load the qT-accumulated expressions
!#killed: #include "accumWqt.inc"


!> load the qT-differential expressions
!#killed: #include "diffWqt.inc"


  subroutine getlumi_(ichan, ix, partons)
    use Scales_mod
    use Channels_mod
    use CKM_mod
    use Process_mod
    integer, intent(in) :: ichan, ix
    double precision, intent(out), dimension(nMax_scl) :: partons
    double precision, dimension(:,:), pointer :: f1, f2
    double precision :: f1lumi(-6:7,1:nMax_scl), f2lumi(-6:7,1:nMax_scl)

    select case (ix)
    case (1)
       f1(-6:7,1:nMax_scl) => f1_scl
       f2(-6:7,1:nMax_scl) => f2_scl
    case (2)
       f1(-6:7,1:nMax_scl) => f1_scl
       f2(-6:7,1:nMax_scl) => f2x2_scl
    case (3)
       f1(-6:7,1:nMax_scl) => f1x1_scl
       f2(-6:7,1:nMax_scl) => f2_scl
    case (4)
       f1(-6:7,1:nMax_scl) => f1x1_scl
       f2(-6:7,1:nMax_scl) => f2x2_scl
    end select

    f1lumi(-6:7,1:nMax_scl) = f1(-6:7,1:nMax_scl)
    f2lumi(-6:7,1:nMax_scl) = f2(-6:7,1:nMax_scl)

    if ( beam1_proc==pid_ap ) then
      f1lumi(-6,:) = f1( 6,:)
      f1lumi(-5,:) = f1( 5,:)
      f1lumi(-4,:) = f1( 4,:)
      f1lumi(-3,:) = f1( 3,:)
      f1lumi(-2,:) = f1( 2,:)
      f1lumi(-1,:) = f1( 1,:)
      f1lumi( 6,:) = f1(-6,:)
      f1lumi( 5,:) = f1(-5,:)
      f1lumi( 4,:) = f1(-4,:)
      f1lumi( 3,:) = f1(-3,:)
      f1lumi( 2,:) = f1(-2,:)
      f1lumi( 1,:) = f1(-1,:)    
    end if

    if ( beam2_proc==pid_ap ) then
      f2lumi(-6,:) = f2( 6,:)
      f2lumi(-5,:) = f2( 5,:)
      f2lumi(-4,:) = f2( 4,:)
      f2lumi(-3,:) = f2( 3,:)
      f2lumi(-2,:) = f2( 2,:)
      f2lumi(-1,:) = f2( 1,:)
      f2lumi( 6,:) = f2(-6,:)
      f2lumi( 5,:) = f2(-5,:)
      f2lumi( 4,:) = f2(-4,:)
      f2lumi( 3,:) = f2(-3,:)
      f2lumi( 2,:) = f2(-2,:)
      f2lumi( 1,:) = f2(-1,:)    
    end if

    partons = 0d0
!We assume CABIBBO mixing for the first two quark generations
!Full CKM is not yet supported and is in the comments
    select case (ichan)
    case( ibit_gg_ ) !fAgBg
      select case (name_proc)
      case('WPQT','WMQT')
        ip1_chan = 0
        ip2_chan = 0
        partons(1:nInp_scl) = &
        &  2d0*f1lumi(0,1:nInp_scl)*f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gq_ ) !fAgBq
      select case (name_proc)
      case('WPQT')
        ip1_chan = 0
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(2,5)*f2lumi(2,1:nInp_scl) + CKM(2,1)*f2lumi(2,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(2,1:nInp_scl) + CKM(4,5)*f2lumi(4,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(4,1:nInp_scl) + CKM(4,3)*f2lumi(4,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(2,1:nInp_scl) + f2lumi(4,1:nInp_scl))
      case('WMQT')
        ip1_chan = 0
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,1)*f2lumi(1,1:nInp_scl) + CKM(2,1)*f2lumi(1,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(3,1:nInp_scl) + CKM(2,3)*f2lumi(3,1:nInp_scl) + &
!         &     CKM(4,5)*f2lumi(5,1:nInp_scl) + CKM(2,5)*f2lumi(5,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(1,1:nInp_scl) + f2lumi(3,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gq1_ ) !fAgBq1
      select case (name_proc)
      case('WPQT')
        ip1_chan = 0
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,1)*f2lumi(1,1:nInp_scl) + CKM(2,1)*f2lumi(1,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(3,1:nInp_scl) + CKM(2,3)*f2lumi(3,1:nInp_scl) + &
!         &     CKM(4,5)*f2lumi(5,1:nInp_scl) + CKM(2,5)*f2lumi(5,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(1,1:nInp_scl) + f2lumi(3,1:nInp_scl))
      case('WMQT')
        ip1_chan = 0
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(2,5)*f2lumi(2,1:nInp_scl) + CKM(2,1)*f2lumi(2,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(2,1:nInp_scl) + CKM(4,5)*f2lumi(4,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(4,1:nInp_scl) + CKM(4,3)*f2lumi(4,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(2,1:nInp_scl) + f2lumi(4,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gqp_ ) !fAgBqp
      select case (name_proc)
      case('WPQT','WMQT')
        ip1_chan = 0
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5) + CKM(4,1) + CKM(4,3) + &
!         &     CKM(2,5) + CKM(2,1) + CKM(2,3))*f1lumi(0,1:nInp_scl)* &
!         &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
!         &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
        &  2d0*f1lumi(0,1:nInp_scl)* &
        &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
        &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qg_ ) !fAqBg
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(2,1:nInp_scl) + CKM(2,1)*f1lumi(2,1:nInp_scl) + &
!         &   CKM(2,3)*f1lumi(2,1:nInp_scl) + CKM(4,5)*f1lumi(4,1:nInp_scl) + &
!         &   CKM(4,1)*f1lumi(4,1:nInp_scl) + CKM(4,3)*f1lumi(4,1:nInp_scl))*f2lumi(0,1:nInp_scl)
        &  (f1lumi(2,1:nInp_scl) + f1lumi(4,1:nInp_scl))*f2lumi(0,1:nInp_scl)
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
!         &   CKM(4,5)*f1lumi(5,1:nInp_scl) + CKM(2,5)*f1lumi(5,1:nInp_scl))*f2lumi(0,1:nInp_scl)
        &  (f1lumi(1,1:nInp_scl) + f1lumi(3,1:nInp_scl))*f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1g_ ) !fAq1Bg
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
!         &   CKM(4,5)*f1lumi(5,1:nInp_scl) + CKM(2,5)*f1lumi(5,1:nInp_scl))*f2lumi(0,1:nInp_scl)
        &  (f1lumi(1,1:nInp_scl) +  f1lumi(3,1:nInp_scl))*f2lumi(0,1:nInp_scl)
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(2,1:nInp_scl) + CKM(2,1)*f1lumi(2,1:nInp_scl) + &
!         &   CKM(2,3)*f1lumi(2,1:nInp_scl) + CKM(4,5)*f1lumi(4,1:nInp_scl) + &
!         &   CKM(4,1)*f1lumi(4,1:nInp_scl) + CKM(4,3)*f1lumi(4,1:nInp_scl))*f2lumi(0,1:nInp_scl)
        &  (f1lumi(2,1:nInp_scl) + f1lumi(4,1:nInp_scl))*f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qpg_ ) !fAqpBg
      select case (name_proc)
      case('WPQT','WMQT')
        ip1_chan = 1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5) + CKM(4,1) + CKM(4,3) +  &
!         &     CKM(2,5) + CKM(2,1) + CKM(2,3))* &
!         &   (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
!         &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
!         &   f2lumi(0,1:nInp_scl)
        &  2d0* &
        &   (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
        &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
        &   f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gqb_ ) !fAgBqb
      select case (name_proc)
      case('WPQT')
        ip1_chan = 0
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,5)*f2lumi(-4,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(-4,1:nInp_scl) + CKM(4,3)*f2lumi(-4,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(-2,1:nInp_scl) + CKM(2,1)*f2lumi(-2,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(-2,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(-4,1:nInp_scl) + f2lumi(-2,1:nInp_scl))
      case('WMQT')
        ip1_chan = 0
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,5)*f2lumi(-5,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(-5,1:nInp_scl) + CKM(4,3)*f2lumi(-3,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(-3,1:nInp_scl) + CKM(4,1)*f2lumi(-1,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(-1,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*( &
        &  f2lumi(-3,1:nInp_scl) + f2lumi(-1,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gq1b_ ) !fAgBq1b
      select case (name_proc)
      case('WPQT')
        ip1_chan = 0
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,5)*f2lumi(-5,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(-5,1:nInp_scl) + CKM(4,3)*f2lumi(-3,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(-3,1:nInp_scl) + CKM(4,1)*f2lumi(-1,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(-1,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*( & 
        &  f2lumi(-3,1:nInp_scl) + f2lumi(-1,1:nInp_scl))
      case('WMQT')
        ip1_chan = 0
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  f1lumi(0,1:nInp_scl)*(CKM(4,5)*f2lumi(-4,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(-4,1:nInp_scl) + CKM(4,3)*f2lumi(-4,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(-2,1:nInp_scl) + CKM(2,1)*f2lumi(-2,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(-2,1:nInp_scl))
        &  f1lumi(0,1:nInp_scl)*(f2lumi(-4,1:nInp_scl) + f2lumi(-2,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_gqbp_ ) !fAgBqbp
      select case (name_proc)
      case('WPQT','WMQT')
        ip1_chan = 0
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5) + CKM(4,1) + CKM(4,3) + &
!         &     CKM(2,5) + CKM(2,1) + CKM(2,3))*f1lumi(0,1:nInp_scl)* &
!         &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) + &
!         &     f2lumi(-1,1:nInp_scl))
        &  2d0*f1lumi(0,1:nInp_scl)* &
        &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) + &
        &     f2lumi(-1,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qbg_ ) !fAqbBg
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-4,1:nInp_scl) + CKM(4,1)*f1lumi(-4,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(-4,1:nInp_scl) + CKM(2,5)*f1lumi(-2,1:nInp_scl) + &
!         &   CKM(2,1)*f1lumi(-2,1:nInp_scl) + CKM(2,3)*f1lumi(-2,1:nInp_scl))* &
!         &   f2lumi(0,1:nInp_scl)
        &  (f1lumi(-4,1:nInp_scl) + f1lumi(-2,1:nInp_scl))* &
        &   f2lumi(0,1:nInp_scl)
      case('WMQT')
        ip1_chan = -1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-5,1:nInp_scl) + CKM(2,5)*f1lumi(-5,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
!         &   CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
!         &   f2lumi(0,1:nInp_scl)
        &  (f1lumi(-3,1:nInp_scl) + &
        &   f1lumi(-1,1:nInp_scl))* &
        &   f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1bg_ ) !fAq1bBg
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-5,1:nInp_scl) + CKM(2,5)*f1lumi(-5,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
!         &   CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
!         &   f2lumi(0,1:nInp_scl)
        &  (f1lumi(-3,1:nInp_scl) + &
        &   f1lumi(-1,1:nInp_scl))* &
        &   f2lumi(0,1:nInp_scl)
      case('WMQT')
        ip1_chan = -1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-4,1:nInp_scl) + CKM(4,1)*f1lumi(-4,1:nInp_scl) + &
!         &   CKM(4,3)*f1lumi(-4,1:nInp_scl) + CKM(2,5)*f1lumi(-2,1:nInp_scl) + &
!         &   CKM(2,1)*f1lumi(-2,1:nInp_scl) + CKM(2,3)*f1lumi(-2,1:nInp_scl))* &
!         &   f2lumi(0,1:nInp_scl)
        &  (f1lumi(-4,1:nInp_scl) + f1lumi(-2,1:nInp_scl))* &
        &   f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qbpg_ ) !fAqbpBg
      select case (name_proc)
      case('WPQT','WMQT')
        ip1_chan = -1
        ip2_chan = 0
        partons(1:nInp_scl) = &
!         &  (CKM(4,5) + CKM(4,1) + CKM(4,3) + &
!         &     CKM(2,5) + CKM(2,1) + CKM(2,3))* &
!         &   (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
!         &     f1lumi(-1,1:nInp_scl))*f2lumi(0,1:nInp_scl)
        &   2d0*(f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
        &     f1lumi(-1,1:nInp_scl))*f2lumi(0,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qq1b_ ) !fAqBq1b
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(2,5)*f1lumi(2,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(4,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
        &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
      case('WMQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(2,5)*f1lumi(5,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(5,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(-4,1:nInp_scl)
        &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(-4,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qqbp_ ) !fAqBq1bp
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(2,1:nInp_scl) + CKM(2,1)*f1lumi(2,1:nInp_scl) +  &
!         &     CKM(2,3)*f1lumi(2,1:nInp_scl) + CKM(4,5)*f1lumi(4,1:nInp_scl) +  &
!         &     CKM(4,1)*f1lumi(4,1:nInp_scl) + CKM(4,3)*f1lumi(4,1:nInp_scl))* &
!         &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
!         &     f2lumi(-1,1:nInp_scl))
        &  (f1lumi(2,1:nInp_scl) + f1lumi(4,1:nInp_scl))* &
        &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
        &     f2lumi(-1,1:nInp_scl))
      case('WMQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(5,1:nInp_scl) + CKM(2,5)*f1lumi(5,1:nInp_scl) + &
!         &     CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
!         &     CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl))* &
!         &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
!         &     f2lumi(-1,1:nInp_scl))
        &  (CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
        &     CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl))* &
        &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
        &     f2lumi(-1,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qpq1b_ ) !fAqpBq1b
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
!         &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
!         &   (CKM(4,5)*f2lumi(-5,1:nInp_scl) + CKM(2,5)*f2lumi(-5,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(-3,1:nInp_scl) + CKM(2,3)*f2lumi(-3,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(-1,1:nInp_scl) + CKM(2,1)*f2lumi(-1,1:nInp_scl))
        &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
        &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
        &   ( CKM(4,3)*f2lumi(-3,1:nInp_scl) + CKM(2,3)*f2lumi(-3,1:nInp_scl) + &
        &     CKM(4,1)*f2lumi(-1,1:nInp_scl) + CKM(2,1)*f2lumi(-1,1:nInp_scl))
      case('WMQT')
        ip1_chan = 1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
!         &   f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
!         &   (CKM(2,5)*f2lumi(-2,1:nInp_scl) + CKM(2,1)*f2lumi(-2,1:nInp_scl) +  &
!         &     CKM(2,3)*f2lumi(-2,1:nInp_scl) + CKM(4,5)*f2lumi(-4,1:nInp_scl) +  &
!         &     CKM(4,1)*f2lumi(-4,1:nInp_scl) + CKM(4,3)*f2lumi(-4,1:nInp_scl))
        &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
        &   f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
        &   ( f2lumi(-2,1:nInp_scl) + f2lumi(-4,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1bq_ ) !fAq1bBq
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,5)*f1lumi(-5,1:nInp_scl)*f2lumi(2,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(2,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(2,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(-5,1:nInp_scl)*f2lumi(4,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(4,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(4,1:nInp_scl)
        &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(2,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(2,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(4,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(4,1:nInp_scl)
      case('WMQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,5)*f1lumi(-2,1:nInp_scl)*f2lumi(5,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(-4,1:nInp_scl)*f2lumi(5,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(1,1:nInp_scl)
        &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(1,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qbpq_ ) !fAq1bpBq
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
!         &     f1lumi(-1,1:nInp_scl))*(CKM(2,5)*f2lumi(2,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(2,1:nInp_scl) + CKM(2,3)*f2lumi(2,1:nInp_scl) + &
!         &     CKM(4,5)*f2lumi(4,1:nInp_scl) + CKM(4,1)*f2lumi(4,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(4,1:nInp_scl))
        &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
        &     f1lumi(-1,1:nInp_scl))*( f2lumi(2,1:nInp_scl) + f2lumi(4,1:nInp_scl))
      case('WMQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
!         &     f1lumi(-1,1:nInp_scl))*(CKM(4,5)*f2lumi(5,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(5,1:nInp_scl) + CKM(4,3)*f2lumi(3,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(3,1:nInp_scl) + CKM(4,1)*f2lumi(1,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(1,1:nInp_scl))     
        &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
        &     f1lumi(-1,1:nInp_scl))*(CKM(4,3)*f2lumi(3,1:nInp_scl) + &
        &     CKM(2,3)*f2lumi(3,1:nInp_scl) + CKM(4,1)*f2lumi(1,1:nInp_scl) + &
        &     CKM(2,1)*f2lumi(1,1:nInp_scl))    
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1bqp_ ) !fAq1bBqp
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-5,1:nInp_scl) + CKM(2,5)*f1lumi(-5,1:nInp_scl) + &
!         &     CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
!         &     CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
!         &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
!         &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
        &  (  CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
        &     CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
        &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
        &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
      case('WMQT')
        ip1_chan = -1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(-2,1:nInp_scl) + CKM(2,1)*f1lumi(-2,1:nInp_scl) +  &
!         &     CKM(2,3)*f1lumi(-2,1:nInp_scl) + CKM(4,5)*f1lumi(-4,1:nInp_scl) +  &
!         &     CKM(4,1)*f1lumi(-4,1:nInp_scl) + CKM(4,3)*f1lumi(-4,1:nInp_scl))* &
!         &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
!         &    f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))   
        &  (f1lumi(-2,1:nInp_scl) + f1lumi(-4,1:nInp_scl))* &
        &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
        &    f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))    
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1q_ ) !fAq1Bq
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(2,5)*f1lumi(5,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
!         &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
!         &  + CKM(4,5)*f1lumi(5,1:nInp_scl)*f2lumi(4,1:nInp_scl)
        &    CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(4,1:nInp_scl) 
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(2,1:nInp_scl)*f2lumi(5,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(4,1:nInp_scl)*f2lumi(5,1:nInp_scl)
        &    CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(3,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qpq_ ) !fAq1pBq
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
!         &   f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
!         &   (CKM(2,5)*f2lumi(2,1:nInp_scl) + CKM(2,1)*f2lumi(2,1:nInp_scl) +  &
!         &     CKM(2,3)*f2lumi(2,1:nInp_scl) + CKM(4,5)*f2lumi(4,1:nInp_scl) +  &
!         &     CKM(4,1)*f2lumi(4,1:nInp_scl) + CKM(4,3)*f2lumi(4,1:nInp_scl))
        &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
        &   f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
        &   ( f2lumi(2,1:nInp_scl) + f2lumi(4,1:nInp_scl))
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
!         &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
!         &   (CKM(4,5)*f2lumi(5,1:nInp_scl) + CKM(2,5)*f2lumi(5,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(3,1:nInp_scl) + CKM(2,3)*f2lumi(3,1:nInp_scl) + &
!         &     CKM(4,1)*f2lumi(1,1:nInp_scl) + CKM(2,1)*f2lumi(1,1:nInp_scl))
        &  (f1lumi(1,1:nInp_scl) + f1lumi(2,1:nInp_scl) + f1lumi(3,1:nInp_scl) + &
        &      f1lumi(4,1:nInp_scl) + f1lumi(5,1:nInp_scl))* &
        &   ( CKM(4,3)*f2lumi(3,1:nInp_scl) + CKM(2,3)*f2lumi(3,1:nInp_scl) + &
        &     CKM(4,1)*f2lumi(1,1:nInp_scl) + CKM(2,1)*f2lumi(1,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qq1_ ) !fAqBq1
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(2,1:nInp_scl)*f2lumi(5,1:nInp_scl) &
!         &  + CKM(4,5)*f1lumi(4,1:nInp_scl)*f2lumi(5,1:nInp_scl)
        &    CKM(2,1)*f1lumi(2,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(4,1:nInp_scl)*f2lumi(1,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(2,1:nInp_scl)*f2lumi(3,1:nInp_scl) &
        &  + CKM(4,3)*f1lumi(4,1:nInp_scl)*f2lumi(3,1:nInp_scl) 
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &    CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(2,5)*f1lumi(5,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
!         &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
!         &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
!         &  + CKM(4,5)*f1lumi(5,1:nInp_scl)*f2lumi(4,1:nInp_scl)
        &    CKM(2,1)*f1lumi(1,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        &  + CKM(2,3)*f1lumi(3,1:nInp_scl)*f2lumi(2,1:nInp_scl)  &
        &  + CKM(4,1)*f1lumi(1,1:nInp_scl)*f2lumi(4,1:nInp_scl)  &
        &  + CKM(4,3)*f1lumi(3,1:nInp_scl)*f2lumi(4,1:nInp_scl)  
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qqp_ ) !fAqBq1p
      select case (name_proc)
      case('WPQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(2,1:nInp_scl) + CKM(2,1)*f1lumi(2,1:nInp_scl) +  &
!         &     CKM(2,3)*f1lumi(2,1:nInp_scl) + CKM(4,5)*f1lumi(4,1:nInp_scl) +  &
!         &     CKM(4,1)*f1lumi(4,1:nInp_scl) + CKM(4,3)*f1lumi(4,1:nInp_scl))* &
!         &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
!         &    f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
        &  (f1lumi(2,1:nInp_scl) + f1lumi(4,1:nInp_scl))* &
        &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
        &    f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
      case('WMQT')
        ip1_chan = 1
        ip2_chan = 1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(5,1:nInp_scl) + CKM(2,5)*f1lumi(5,1:nInp_scl) + &
!         &     CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
!         &     CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl))* &
!         &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
!         &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
        &   ( CKM(4,3)*f1lumi(3,1:nInp_scl) + CKM(2,3)*f1lumi(3,1:nInp_scl) + &
        &     CKM(4,1)*f1lumi(1,1:nInp_scl) + CKM(2,1)*f1lumi(1,1:nInp_scl))* &
        &   (f2lumi(1,1:nInp_scl) + f2lumi(2,1:nInp_scl) + f2lumi(3,1:nInp_scl) + &
        &      f2lumi(4,1:nInp_scl) + f2lumi(5,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1bqb_ ) !fAq1bBqb
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(4,5)*f1lumi(-5,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(-5,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)
        &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)
      case('WMQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(4,5)*f1lumi(-4,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(-2,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
        &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_q1bqbp_ ) !fAq1bBqbp
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (CKM(4,5)*f1lumi(-5,1:nInp_scl) + CKM(2,5)*f1lumi(-5,1:nInp_scl) + &
!         &     CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
!         &     CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
!         &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) + &
!         &     f2lumi(-1,1:nInp_scl))
        &   ( CKM(4,3)*f1lumi(-3,1:nInp_scl) + CKM(2,3)*f1lumi(-3,1:nInp_scl) + &
        &     CKM(4,1)*f1lumi(-1,1:nInp_scl) + CKM(2,1)*f1lumi(-1,1:nInp_scl))* &
        &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) + &
        &     f2lumi(-1,1:nInp_scl))
      case('WMQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (CKM(2,5)*f1lumi(-2,1:nInp_scl) + CKM(2,1)*f1lumi(-2,1:nInp_scl) +  &
!         &     CKM(2,3)*f1lumi(-2,1:nInp_scl) + CKM(4,5)*f1lumi(-4,1:nInp_scl) +  &
!         &     CKM(4,1)*f1lumi(-4,1:nInp_scl) + CKM(4,3)*f1lumi(-4,1:nInp_scl))* &
!         &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
!         &     f2lumi(-1,1:nInp_scl))
        &  (f1lumi(-2,1:nInp_scl) + f1lumi(-4,1:nInp_scl))* &
        &   (f2lumi(-5,1:nInp_scl) + f2lumi(-4,1:nInp_scl) + f2lumi(-3,1:nInp_scl) + f2lumi(-2,1:nInp_scl) +  &
        &     f2lumi(-1,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qbq1b_ ) !fAqbBq1b
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(4,5)*f1lumi(-4,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(-2,1:nInp_scl)*f2lumi(-5,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
        &  + CKM(4,3)*f1lumi(-4,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(-2,1:nInp_scl)*f2lumi(-3,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-4,1:nInp_scl)*f2lumi(-1,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-2,1:nInp_scl)*f2lumi(-1,1:nInp_scl)
      case('WMQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &    CKM(4,5)*f1lumi(-5,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
!         &  + CKM(2,5)*f1lumi(-5,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
!         &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)
        &  + CKM(4,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
        &  + CKM(4,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-4,1:nInp_scl) &
        &  + CKM(2,3)*f1lumi(-3,1:nInp_scl)*f2lumi(-2,1:nInp_scl) &
        &  + CKM(2,1)*f1lumi(-1,1:nInp_scl)*f2lumi(-2,1:nInp_scl)
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case( ibit_qbpq1b_ ) !fAqbpBq1b
      select case (name_proc)
      case('WPQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
!         &     f1lumi(-1,1:nInp_scl))*(CKM(4,5)*f2lumi(-5,1:nInp_scl) + &
!         &     CKM(2,5)*f2lumi(-5,1:nInp_scl) + CKM(4,3)*f2lumi(-3,1:nInp_scl) + &
!         &     CKM(2,3)*f2lumi(-3,1:nInp_scl) + CKM(4,1)*f2lumi(-1,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(-1,1:nInp_scl))
        &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
        &     f1lumi(-1,1:nInp_scl))*( CKM(4,3)*f2lumi(-3,1:nInp_scl) + &
        &     CKM(2,3)*f2lumi(-3,1:nInp_scl) + CKM(4,1)*f2lumi(-1,1:nInp_scl) + &
        &     CKM(2,1)*f2lumi(-1,1:nInp_scl))
      case('WMQT')
        ip1_chan = -1
        ip2_chan = -1
        partons(1:nInp_scl) = &
!         &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
!         &     f1lumi(-1,1:nInp_scl))*(CKM(2,5)*f2lumi(-2,1:nInp_scl) + &
!         &     CKM(2,1)*f2lumi(-2,1:nInp_scl) + CKM(2,3)*f2lumi(-2,1:nInp_scl) + &
!         &     CKM(4,5)*f2lumi(-4,1:nInp_scl) + CKM(4,1)*f2lumi(-4,1:nInp_scl) + &
!         &     CKM(4,3)*f2lumi(-4,1:nInp_scl))
        &  (f1lumi(-5,1:nInp_scl) + f1lumi(-4,1:nInp_scl) + f1lumi(-3,1:nInp_scl) + f1lumi(-2,1:nInp_scl) + &
        &     f1lumi(-1,1:nInp_scl))*(f2lumi(-2,1:nInp_scl) + f2lumi(-4,1:nInp_scl))
      case default
         write(*,*) "Invalid process in sigWqt.f90 getlumi_: ", name_proc
         stop
      end select
    case default
      print*, "sigHqt getlumi_: invalid ichan:", ichan
      stop
    end select
  end subroutine getlumi_


end module sigWqt_mod


