!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: sigEEinc [external members: `_EEinc`]
!
! DESCRIPTION:
!> Module providing inclusive hadron production in epem
!> Useful references:
!>    * https://arxiv.org/pdf/1210.3594.pdf
!>    * https://arxiv.org/pdf/1707.01044.pdf
!>    * https://arxiv.org/pdf/1201.5804.pdf
!>    * https://arxiv.org/pdf/hep-ph/9609377.pdf
!>    * ...
!> first start with of-shell photons only!
!> you better run with `V_NC = gamma`
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module sigEEinc_mod
!-------------------------------------------------------------------------------
  implicit none
  private

  public :: init_EEinc
  !> the differential cross section to be integrated
  public :: sigEEinc

  double precision, parameter :: NC = 3d0
  double precision, parameter :: CF = 4d0/3d0
  double precision, parameter :: CA = 3d0
  double precision, parameter :: NF = 5d0
  double precision, parameter :: dFdF = 5d0/6d0


  !----- local setting for the cross ection:

  double precision :: sum_q, sum_q2
  double precision :: sum_a, sum_a2
  double precision :: sum_v, sum_v2


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> set parmeters of the cross section
  !-----------------------------------------------------------------------------
  subroutine init_EEinc()
    use Process_mod, only : name_proc, sqrts_proc
    use Parameters_mod
    use Charges_EW_mod
    integer :: iq

    print*, "init_EEinc: setting up EE..."

    sum_q  = 0d0
    sum_q2 = 0d0
    sum_a  = 0d0
    sum_a2 = 0d0
    sum_v  = 0d0
    sum_v2 = 0d0

    do iq = 1,5
      sum_q  = sum_q  + Qq(iq)
      sum_q2 = sum_q2 + Qq(iq)**2
    end do

  end subroutine init_EEinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> the differential cross section
  !
  !> @param[in] rnorm
  !> @param[in] x1
  !> @param[in] x2
  !> @param[in] wgt
  !-----------------------------------------------------------------------------
  double precision function sigEEinc(rnorm, x1, x2, wgt)
    use Process_mod, only : sqrts_proc
    use Constants_mod
    use KinData_mod
    use Scales_mod
    use Channels_mod
    use Parameters_mod
    use LHAPDF_mod
    double precision, intent(in) :: rnorm, x1, x2, wgt

    double precision :: x1curr,x2curr
    common /parfrac/x1curr,x2curr
!$omp threadprivate(/parfrac/)

    double precision :: xreg1,xreg2
    logical :: virtsflag
    common /xregions/xreg1,xreg2,virtsflag
!$omp threadprivate(/xregions/)

    logical :: jflag
    common /jacobian/jflag
!$omp threadprivate(/jacobian/)

    double precision :: rnormcurr,wgtcurr
    common /histonorm/rnormcurr,wgtcurr
!$omp threadprivate(/histonorm/)

    double precision, external :: bino

    double precision, dimension(nMax_scl) :: as_, partons_
    double precision, dimension(nMax_scl) :: kinwt
    double precision :: wt, wtsum
    double precision :: fac, accum
    integer :: ipass, ix, iorder, norder, iscl
    double precision :: result(1:nMax_scl, 0:3,1:6,1:4)

    ! print*, "sigEEinc: ", rnorm, x1, x2, wgt

    current_chan = 1  ! only one channel

    x1curr = x1
    x2curr = x2
    rnormcurr = rnorm  ! should we override rnorm???
    wgtcurr = wgt

    sigEEinc = 0d0
    wtsum = 0d0
    kinwt = 0d0
    wt    = 0d0
    virtsflag=.false.
    jflag=.false.


    ipass = 1
    call ecuts_epem(0,4,ipass)
    if (ipass==0) return


    fac = BornME() ! the Born amplitude!
    as_ = alphas_scl / (4d0*pi)
    partons_ = 1d0  ! lumi always 1 for epem
    ix = 1  ! no IS convolution

    do iorder=0,3
      ! print*, "iorder = ", iorder, get_order_inc_chan(iorder)
      if (.not.get_order_inc_chan(iorder)) cycle

      accum = sum_q2 * r_NS_V(iorder) + sum_q * r_S_V(iorder)
      kinwt = fac * accum * as_**iorder
      wt = bino(ix, partons_, kinwt)
      wtsum = wtsum + wt

    end do  ! end do iorder

    sigEEinc = rnorm * wtsum

    ! print*, "sigEEinc = ", sigEEinc
    ! read(*,*)
  contains

    !> photon-exchange case with quark charges stripped
    double precision function BornME()
      use Constants_mod

      !> avg factor
      BornME = 0.25d0
      !> Born pre-factor
      BornME = BornME * (4d0*pi*alpha_param())**2*NC*4d0
      !> Born amplitude (B0g0Z without Z exchange & quark charges)
      BornME = BornME * 2d0 * ( s4(1,3)**2 + s4(1,4)**2 ) / ( s4(1,2)**2 )

    end function BornME

  end function sigEEinc


  !-----------------------------------------------------------------------------
  !> @brief
  !> non-singlet vector contribution to R-ratio
  !> expansion in terms of as/(4pi)
  !
  !> @param[in] iorder order in alphas
  !> @return coefficient of the iorder-th term
  !-----------------------------------------------------------------------------
  double precision function r_NS_V(iorder)
    use Constants_mod
    integer, intent(in) :: iorder
    r_NS_V = 0d0
    select case(iorder)
    case(0)
      r_NS_V = 1d0
    case(1)
      r_NS_V = 3d0 * CF
    case(2)
      r_NS_V = -3d0/2d0 * CF**2  &
             & +CA*CF*( 123d0/2d0 - 44d0*zeta3 )  &
             & -CF*NF*( 11d0 - 8d0*zeta3 )
    case(3)
      r_NS_V = -69d0/2d0*CF**3  &
             & -CA*CF**2*( 127d0 + 572*zeta3 - 880d0*zeta5 )  &
             & +CA**2*CF*( 90445d0/54d0 - 242d0/3d0*zeta2 - 10948d0/9d0*zeta3 - 440d0/3d0*zeta5 )  &
             & -CF**2*NF*( 29d0/2d0 - 152d0*zeta3 + 160d0*zeta5 )  &
             & -CA*CF*NF*( 15520d0/27d0 - 88d0/3d0*zeta2 - 3584d0/9d0*zeta3 - 80d0/3d0*zeta5 )  &
             & +CF*NF**2*( 1208d0/27d0 - 8d0/3d0*zeta2 - 304d0/9d0*zeta3 )
    case default
      stop 'sigEEinc_mod: r_NS_V: order not implemented'
    end select
  end function r_NS_V


  !-----------------------------------------------------------------------------
  !> @brief
  !> singlet vector contribution to R-ratio
  !> expansion in terms of as/(4pi)
  !
  !> @param[in] iorder order in alphas
  !> @return coefficient of the iorder-th term
  !-----------------------------------------------------------------------------
  double precision function r_S_V(iorder)
    use Constants_mod
    integer, intent(in) :: iorder
    r_S_V = 0d0
    select case(iorder)
    case(0,1,2)
      r_S_V = 0d0
    case(3)
      r_S_V = +dFdF/NC*( 176d0/3d0 - 128d0*zeta3 )
    case default
      stop 'sigEEinc_mod: r_S_V: order not implemented'
    end select
  end function r_S_V

end module sigEEinc_mod


