!-------------------------------------------------------------------------------
! NNLOJET
!-------------------------------------------------------------------------------
!
! MODULE: nnlo_bridge [external members: `_hook`]
!
! DESCRIPTION:
!> Hooks used in the nnlo-bridge code to interface to APPLgrid & fastNLO
!-------------------------------------------------------------------------------


!-------------------------------------------------------------------------------
module nnlo_bridge_mod
!-------------------------------------------------------------------------------
  use, intrinsic :: iso_c_binding
  use Scales_mod, only : nMax_scl
  implicit none
  public

  integer(c_int), bind(C) :: maxscl_hook = nMax_scl
  integer(c_int), bind(C) :: iseed_hook = 0

  !> thredlocal storage of "current" scale
  private :: muR2_current, muF2_current
  double precision, dimension(nMax_scl) :: muR2_current, muF2_current
!$OMP THREADPRIVATE(muR2_current, muF2_current)

  !> thredlocal storage of "current" observable id
  private :: obsId_current
  integer :: obsId_current
!$OMP THREADPRIVATE(obsId_current)


contains


  !-----------------------------------------------------------------------------
  !> @brief
  !> collider energy in GeV
  !
  !> @return c-o-m collider energy
  !-----------------------------------------------------------------------------
  function sqrts_hook() bind(C)
    use Process_mod, only : sqrts_proc
    real(c_double) :: sqrts_hook
    sqrts_hook = sqrts_proc
  end function sqrts_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> number of PDFs = incoming hadrons
  !
  !> @return number of PDFs
  !-----------------------------------------------------------------------------
  function n_pdf_hook() bind(C)
    use Process_mod, only : nPDFs_proc
    integer(c_int) :: n_pdf_hook
    n_pdf_hook = nPDFs_proc
  end function n_pdf_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> is the unit-phase option on?
  !
  !> @return flag for unit-phase
  !-----------------------------------------------------------------------------
  function unit_phase_hook() bind(C)
    use UnitPhase_mod, only : phase_unit
    logical(c_bool) :: unit_phase_hook
    unit_phase_hook = phase_unit
  end function unit_phase_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> jet algorithm id:
  !>   0: none
  !>   1: anti-kt
  !>   2: C/A
  !>   3: kt
  !>   4: jade
  !
  !> @return jet algo id
  !-----------------------------------------------------------------------------
  function jet_algo_hook() bind(C)
    use Jets_mod, only : jetalg
    integer(c_int) :: jet_algo_hook
    jet_algo_hook = jetalg
  end function jet_algo_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> jet Rcut value
  !
  !> @return jet cone size
  !-----------------------------------------------------------------------------
  function jet_rcut_hook() bind(C)
    use Jets_mod, only : Rcone
    real(c_double) :: jet_rcut_hook
    jet_rcut_hook = Rcone
  end function jet_rcut_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> current channel id
  !
  !> @return integer id for the currnt channel
  !-----------------------------------------------------------------------------
  function n_scales_hook() bind(C)
    use Scales_mod, only : nInp_scl
    integer(c_int) :: n_scales_hook
    n_scales_hook = nInp_scl
  end function n_scales_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> current channel id
  !
  !> @return integer id for the currnt channel
  !-----------------------------------------------------------------------------
  function iproc_hook() bind(C)
    use Channels_mod, only : current_chan
    integer(c_int) :: iproc_hook
    iproc_hook = current_chan
  end function iproc_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> parton a (in NNLOjet convention)
  !
  !> @return integer id for parton a (cann be correlated to b)
  !-----------------------------------------------------------------------------
  function ip1_hook() bind(C)
    use Channels_mod, only : ip1_chan
    integer(c_int) :: ip1_hook
    ip1_hook = ip1_chan
  end function ip1_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> parton b (in NNLOjet convention)
  !
  !> @return integer id for parton b (cann be correlated to a)
  !-----------------------------------------------------------------------------
  function ip2_hook() bind(C)
    use Channels_mod, only : ip2_chan
    integer(c_int) :: ip2_hook
    ip2_hook = ip2_chan
  end function ip2_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> momentum fraction for parton a
  !
  !> @return momentum fraction x1
  !-----------------------------------------------------------------------------
  function x1_hook() bind(C)
    use Channels_mod, only : ix_chan
    use KinData_mod, only : x1_kin, xreg1_kin
    real(c_double) :: x1_hook
    select case (ix_chan)
    case(1,2)
       x1_hook = x1_kin
    case(3,4)
       x1_hook = x1_kin / xreg1_kin
    case default
      print*, "x1_hook: invalid ix value:", ix_chan
      stop 
    end select
  end function x1_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> momentum fraction for parton b
  !
  !> @return momentum fraction x2
  !-----------------------------------------------------------------------------
  function x2_hook() bind(C)
    use Channels_mod, only : ix_chan
    use KinData_mod, only : x2_kin, xreg2_kin
    real(c_double) :: x2_hook
    select case (ix_chan)
    case(1,3)
       x2_hook = x2_kin
    case(2,4)
       x2_hook = x2_kin / xreg2_kin
    case default
      print*, "x2_hook: invalid ix value:", ix_chan
      stop 
    end select
  end function x2_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> ix - the NNLOjet id for vortual contributions
  !
  !> @return integer id for ix
  !-----------------------------------------------------------------------------
  function ix_hook() bind(C)
    use Channels_mod, only : ix_chan
    integer(c_int) :: ix_hook
    ix_hook = ix_chan
  end function ix_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> alphas power associated with the current weight
  !
  !> @return integer for the as/(2pi) power
  !-----------------------------------------------------------------------------
  function n_as_hook() bind(C)
    use Channels_mod, only : currentContrib_chan
    integer(c_int) :: n_as_hook
    call currentContrib_chan(nas=n_as_hook)
  end function n_as_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> the order associated with the current weight
  !
  !> @return 0: LO, 1: NLO(V,R), 2: NNLO(VV,RV,RR)
  !-----------------------------------------------------------------------------
  function n_order_hook() bind(C)
    use Channels_mod, only : currentContrib_chan
    integer(c_int) :: n_order_hook
    call currentContrib_chan(nO=n_order_hook)
  end function n_order_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> alphas power associated with the Born process
  !
  !> @return integer for the as/(2pi) power
  !-----------------------------------------------------------------------------
  function n_as_born_hook() bind(C)
    use Process_mod, only : n_alphas_proc
    integer(c_int) :: n_as_born_hook
    n_as_born_hook = n_alphas_proc
  end function n_as_born_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the current observable id
  !
  !> @param[in] obsId
  !-----------------------------------------------------------------------------
  subroutine set_obsId(obsId)
    integer, intent(in) :: obsId
    obsId_current = obsId
  end subroutine set_obsId


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the observable id of the current weight
  !
  !> @return observable id
  !-----------------------------------------------------------------------------
  function current_obs_id_hook() bind(C)
    integer(c_int) :: current_obs_id_hook
    current_obs_id_hook = obsId_current
  end function current_obs_id_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> set the scales
  !
  !> @param[in] muR2 array of ren scales
  !> @param[in] muF2 array of fac scales
  !-----------------------------------------------------------------------------
  subroutine set_mu2(muR2, muF2)
    double precision, dimension(nMax_scl), intent(in) :: muR2, muF2
    muR2_current = muR2
    muF2_current = muF2
  end subroutine set_mu2


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the scales
  !
  !> @param[out] muR2 array of ren scales
  !> @param[out] muF2 array of fac scales
  !-----------------------------------------------------------------------------
  subroutine get_mu2(muR2, muF2)
    double precision, dimension(nMax_scl), intent(out) :: muR2, muF2
    muR2 = muR2_current
    muF2 = muF2_current
  end subroutine get_mu2


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the renormalisation scale
  !
  !> @param[in] idx # indexing the different scale-pairs
  !> @return renormalisation scale [GeV^2]
  !-----------------------------------------------------------------------------
  function mur2_hook(idx) bind(C)
    use Scales_mod, only : muR2_scl
    real(c_double) :: mur2_hook
    integer(c_int) :: idx
    if (idx==0) then
      mur2_hook = muR2_scl(1)
    else
      mur2_hook = muR2_current(idx)      
    end if
  end function mur2_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the factorisation scale
  !
  !> @param[in] idx # indexing the different scale-pairs
  !> @return factorisation scale [GeV^2]
  !-----------------------------------------------------------------------------
  function muf2_hook(idx) bind(C)
    use Scales_mod, only : muF2_scl
    real(c_double) :: muf2_hook
    integer(c_int) :: idx
    if (idx==0) then
      muf2_hook = muF2_scl(1)
    else
      muf2_hook = muF2_current(idx)      
    end if
  end function muf2_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the renormalisation scale prefactor
  !
  !> @param[in] idx # indexing the different scale-pairs
  !> @return renormalisation scale multiplier
  !-----------------------------------------------------------------------------
  function facr_hook(idx) bind(C)
    use Scales_mod, only : facR_scl
    real(c_double) :: facr_hook
    integer(c_int) :: idx
    facr_hook = facR_scl(idx)
  end function facr_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the factorisation scale prefactor
  !
  !> @param[in] idx # indexing the different scale-pairs
  !> @return factorisation scale multiplier
  !-----------------------------------------------------------------------------
  function facf_hook(idx) bind(C)
    use Scales_mod, only : facF_scl
    real(c_double) :: facf_hook
    integer(c_int) :: idx
    facf_hook = facF_scl(idx)
  end function facf_hook


  ! !-----------------------------------------------------------------------------
  ! !> @brief
  ! !> get the current cache id
  ! !
  ! !> @return cache identifier
  ! !-----------------------------------------------------------------------------
  ! function currentCacheID_hook() bind(C)
  !   use Observables_mod, only : npar_current
  !   use KinData_mod
  !   integer(c_int) :: currentCacheID_hook
  !   currentCacheID_hook = kin(npar_current)%cacheID
  ! end function currentCacheID_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the observable id
  !
  !> @param[in] obsname name of the observable
  !> @return integer id for obsname
  !-----------------------------------------------------------------------------
  function obs_id_hook(obsname) bind(C)
    use Observables_mod, only : getIdFromName_obs
    integer(c_int) :: obs_id_hook
    character(kind=c_char, len=1), intent(in) :: obsname(*)
    character(:), allocatable :: f_obsname
    integer :: i, nchars
    i = 1
    do
      if (obsname(i) == c_null_char) exit
      i = i + 1
    end do
    nchars = i - 1  ! Exclude null character from Fortran string
    allocate(character(len=nchars) :: f_obsname)
    f_obsname = transfer(obsname(1:nchars), f_obsname)
    ! f_obsname = ''
    ! i = 1
    ! do while ( obsname(i) /= c_null_char ) 
    !   f_obsname = f_obsname // obsname(i)
    !   i = i+1
    ! end do
    obs_id_hook = getIdFromName_obs(f_obsname)
    deallocate(f_obsname)
  end function obs_id_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the value for an observable
  !
  !> @param[in] obsid integer id for the observable
  !> @return value of the observable
  !-----------------------------------------------------------------------------
  function obs_val_hook(obsid) bind(C)
    use Observables_mod, only : getValue_obs
    real(c_double) :: obs_val_hook
    integer(c_int) :: obsid
    logical :: valid
    obs_val_hook = getValue_obs(obsid, valid)
    !if (.not.valid) stop "obs_val_hook: invalid observable"
    if (.not.valid) obs_val_hook = 0d0
  end function obs_val_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the name for an observable
  !
  !> @param[in] obsid integer id for the observable
  !> @param[out] obsname 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine obs_name_hook(obsid, obsname, cstrlen) bind(C)
    use Observables_mod, only : getNameFromId_obs
    integer(c_int) :: obsid
    character(kind=c_char, len=1), intent(out) :: obsname(*)
    integer(c_int), intent(in) :: cstrlen
    integer :: strlen, i
    character(len=64) :: name

    name = getNameFromId_obs(obsid)

    if (cstrlen <= 0) return
    strlen = len(trim(name))
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      obsname(i) = name(i:i)
    end forall
    obsname(strlen+1) = c_null_char
  end subroutine obs_name_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the process name
  !
  !> @param[out] process 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine process_name_hook(process, cstrlen) bind(C)
    use Process_mod, only : in_name_proc
    character(kind=c_char, len=1), intent(out) :: process(*)
    integer(c_int), intent(in) :: cstrlen
    integer :: strlen, i

    if (cstrlen <= 0) return
    strlen = len(trim(in_name_proc))
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      process(i) = in_name_proc(i:i)
    end forall
    process(strlen+1) = c_null_char
  end subroutine process_name_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the job name
  !
  !> @param[out] jobname 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine job_name_hook(jobname, cstrlen) bind(C)
    use Run_mod, only : name_run
    character(kind=c_char, len=1), intent(out) :: jobname(*)
    integer(c_int), intent(in) :: cstrlen
    integer :: strlen, i

    if (cstrlen <= 0) return
    strlen = len(trim(name_run))
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      jobname(i) = name_run(i:i)
    end forall
    jobname(strlen+1) = c_null_char
  end subroutine job_name_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the channel string
  !
  !> @param[out] channel 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine channel_hook(channel, cstrlen) bind(C)
    use Channels_mod, only : currentContrib_chan
    character(kind=c_char, len=1), intent(out) :: channel(*)
    integer(c_int), intent(in) :: cstrlen
    character(:), allocatable :: str
    integer :: strlen, i

    call currentContrib_chan(str=str)

    strlen = len(trim(str))
    if (cstrlen <= 0) return
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      channel(i) = str(i:i)
    end forall
    channel(strlen+1) = c_null_char
  end subroutine channel_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF name
  !
  !> @param[out] pdfname 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine pdf_set_hook(pdfname, cstrlen) bind(C)
    use Run_mod, only : pdf_set_run
    character(kind=c_char, len=1), intent(out) :: pdfname(*)
    integer(c_int), intent(in) :: cstrlen
    integer :: strlen, i

    strlen = len(trim(pdf_set_run))
    if (cstrlen <= 0) return
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      pdfname(i) = pdf_set_run(i:i)
    end forall
    pdfname(strlen+1) = c_null_char
  end subroutine pdf_set_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the PDF member
  !
  !> @return integer id for pdf mem
  !-----------------------------------------------------------------------------
  function pdf_mem_hook() bind(C)
    use Run_mod, only : pdf_mem_run
    integer(c_int) :: pdf_mem_hook
    pdf_mem_hook = pdf_mem_run
  end function pdf_mem_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the citation list
  !
  !> @param[out] citation 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine citation_hook(citation, cstrlen) bind(C)
    use Process_mod, only : name_proc
    character(kind=c_char, len=1), intent(out) :: citation(*)
    integer(c_int), intent(in) :: cstrlen
    character(:), allocatable :: fstr
    integer :: strlen, i

    select case(name_proc)
    case("1JET")
      fstr = "Phys. Rev. Lett. 118 (2017) 072002 [arXiv:1611.01460] + more?"
    case("ZJ")
      fstr = "Phys. Rev. Lett. 117 (2016) 022001 [arXiv:1507.02850] + more?"
    case("WJ")
      fstr = "Phys. Rev. Lett. 120 (2018) 122001 [arXiv:1712.07543] + more?"
    case default
      fstr = "not set yet..."
    end select

    strlen = len(trim(fstr))
    if (cstrlen <= 0) return
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      citation(i) = fstr(i:i)
    end forall
    citation(strlen+1) = c_null_char
  end subroutine citation_hook


  !-----------------------------------------------------------------------------
  !> @brief
  !> get the NNLOJET version string
  !
  !> @param[out] version 
  !> @param[in] cstrlen length of the char array
  !-----------------------------------------------------------------------------
  subroutine version_hook(version, cstrlen) bind(C)
    use NNLOJET_mod, only : version_nnlojet
    character(kind=c_char, len=1), intent(out) :: version(*)
    integer(c_int), intent(in) :: cstrlen
    integer :: strlen, i

    strlen = len(trim(version_nnlojet))
    if (cstrlen <= 0) return
    strlen = min(strlen, cstrlen-1)  ! NULL

    forall (i=1:strlen)
      version(i) = version_nnlojet(i:i)
    end forall
    version(strlen+1) = c_null_char
  end subroutine version_hook


end module nnlo_bridge_mod
