**************************************************************
* A small module for storing all of our momenta configurations
*
      module pmap
      use KinData_mod
      implicit none



!> type `store` should be replaced by Kin_t pointers:
!>     type(Kin_t), pointer :: psX30(.....)
!>     type(Kin_t), pointer :: psX40(.....)
!> which have to be all initialized to null(), nullify is
!> also required whenever the cache is cleared, i.e. when
!> `destroy_kin()` or `clearCache_kin()` is called.
!> this happens after every vegas iteration
!>     see vegasnr.f:245
!> So one should provide a routine `clear_pstore()` to set
!> all pointers in psX30 and psX40 to null() and call it
!> directly after `call destroy_kin()` in vegasnr.f


!> Inside `storepsetX30` and `storepsetX40` one should call
!>     call copy_kin(npar, pointer)
!> where `pointer` wil be the the appropriate entry in the
!> arrays `psX30` and `psX30`. If the entry is null(), new
!> memory will be allocated and the `npar`-kinematics will
!> be copied into it and the pointer will contain the
!> reference to the copy. If pointer is not null, this means
!> memory was previously allocated and the `npar`-kinematics
!> stored there will be overwritten by whatever is currently
!> stored in `kin`.

!> Inside `checkpsetX30` and `checkpsetX40` one can call
!>     call overwrite_kin(pointer)
!> where `pointer` wil be the the appropriate entry in the
!> arrays `psX30` and `psX30`. This will automatically
!> overwrite the correct `npar` entry in `kin` with whatever
!> is stored in the cache behind the pointer.



      type store
         logical :: exists = .false.
         integer :: ipass = 0
         type(Kin_t), pointer :: Kin => null()
      end type store

      type (store), target :: psX30(6,3:7,7,0:5,0:6,0:6)
!$omp threadprivate(psX30)
      type (store), target :: psX40(6,6,7,7)
!$omp threadprivate(psX40)
      save psX30, psX40


      contains


      subroutine clear_pstore
      implicit none
      integer i,j,k,l,m,n

      call pflush()

!     nullify all pointers
      forall(n=0:6, m=0:6, l=0:5, k=1:7, j=3:7, i=1:6)
         psX30(i,j,k,l,m,n)%kin => null()
      end forall

      forall(l=1:7, k=1:7, j=1:6, i=1:6)
         psX40(i,j,k,l)%kin => null()
      end forall

      return
      end subroutine


      subroutine pflush
      use Process_mod, only : name_proc
      use astore
      implicit none
      integer ifh1,ifs1,ifh2
      integer jfh1,jfs1,jfh2
      integer ipsid
      logical pstoreflag
      common /currentmap/ ifh1,ifs1,ifh2,jfh1,jfs1,jfh2
!$omp threadprivate(/currentmap/)
      common/pstore/pstoreflag,ipsid


      if(pstoreflag)then
         psX30%exists = .false.
         psX40%exists = .false.

         ifh1 = 0
         ifs1 = 0
         ifh2 = 0

         jfh1 = 0
         jfs1 = 0
         jfh2 = 0
         if ( (name_proc(1:3).eq.'DIS') .or.
     .        (name_proc(1:1).eq.'Z') .or.
     .        (name_proc(1:1).eq.'W') .or.
     .        (name_proc(1:4).eq."EPEM") ) then ! ZJ,WJ 4quark amplitude calculations
            call aflush()
         endif

      endif

      return
      end subroutine pflush

      end module pmap

********************************************************************
* A subroutine for checking where a momentum configuration has been
* calculated previously. If it has it'll pull it from pstore
*
      subroutine checkpsetX40(npar,iha,isa,isb,ihb,exists,ipass)
      use KinData_mod
      use pmap
      implicit real*8(a-h,o-y)
      implicit complex*16(z)
      logical exists,pstoreflag
      common /currentmap/ ifh1,ifs1,ifh2,jfh1,jfs1,jfh2
!$omp threadprivate(/currentmap/)
      common/pstore/pstoreflag,ipsid
      type (store), pointer :: psTmp

      if(.not.pstoreflag)then ! do not store any momenta
         exists=.false.
         return
      endif

      stop "pstore deactivated for now!"

!     with a FF mapping the ordering of the unresolved partons matters
!     => must be treated appropriately
      if((iha.gt.2).and.(ihb.gt.2))then
         if(iha.lt.ihb)then
            ih1=iha
            ih2=ihb
            is1=isa
            is2=isb
         else
            ih1=ihb
            ih2=iha
            is1=isb
            is2=isa
         endif
      elseif((iha.le.2).and.(ihb.le.2))then !II
         if(iha.lt.ihb)then
            ih1=iha
            ih2=ihb
         else
            ih1=ihb
            ih2=iha
         endif
         if(isa.lt.isb)then
            is1=isa
            is2=isb
         else
            is1=isb
            is2=isa
         endif
      else !IF
         call sortval(isa,isb,ihb,is1,is2,ih2)
         ih1 = iha
      endif

      psTmp => psX40(ih1,is1,is2,ih2)

      if(.not.psTmp%exists)then
         exists=.false.
         !print*, npar, "generating X40 mapping: ", iha,isa,isb,ihb
      else
         ipass=psTmp%ipass
         if(ipass.eq.1)then
            call overwrite_kin(psTmp%kin)
         endif
         exists=.true.
         !print*, npar, "reading cached X40 mapping: ", iha,isa,isb,ihb
         ifh1=ih1
         ifs1=is1
         ifh2=is2
         jfh1=ih2
         jfs1=0
         jfh2=0
      endif

      return
      end subroutine


*******************************************************************
* A small sorting routine for the IFFF mappings
*
      subroutine sortval(i1,i2,i3,j1,j2,j3)
      implicit none
      integer i1,i2,i3,j1,j2,j3,k2,k3

      if((i1.lt.i2).and.(i1.le.i3))then
         j1=i1
         k2=i2
         k3=i3
      elseif((i2.lt.i1).and.(i2.lt.i3))then
         j1=i2
         k2=i1
         k3=i3
      else
         j1=i3
         k2=i1
         k3=i2
      endif

      if(k3.gt.k2)then
         j2=k2
         j3=k3
      else
         j2=k3
         j3=k2
      endif

      return
      end


********************************************************************
* A subroutine for checking where a momentum configuration has been
* calculated previously. If it has it'll pull it from pstore
*
      subroutine checkpsetX30(npar,iha,isa,ihb,exists,ipass)
      use KinData_mod
      use pmap
      implicit real*8(a-h,o-y)
      implicit complex*16(z)
      logical exists,pstoreflag
      common /currentmap/ ifh1,ifs1,ifh2,jfh1,jfs1,jfh2
!$omp threadprivate(/currentmap/)
      common/pstore/pstoreflag,ipsid
      type (store), pointer :: psTmp

      if(.not.pstoreflag)then ! do not store any momenta
         exists=.false.
         return
      endif

      stop "pstore deactivated for now!"

      if(iha.lt.ihb)then ! II mapping => exchange II
         ih1=iha
         ih2=ihb
      else
         ih1=ihb
         ih2=iha
      endif

      if((iha.le.2).and.(ihb.gt.2).and.(isa.gt.ihb))then ! IF mapping where isa > ihb => swap is and ihb
         is=ihb
         ih2=isa
      else
         is = isa
      endif

      !> checking for secondary mapping
      if(ifh1.ne.0)then
         psTmp => psX30(ifh1,ifs1,ifh2,ih1,is,ih2)
      else
         psTmp => psX30(ih1,is,ih2,0,0,0)
      endif

      if(.not.psTmp%exists)then
         exists=.false.         ! need to generate momenta
         !print*, npar, "generating X30 mapping: ", iha,isa,ihb
      else
         exists=.true.
         !print*, npar, "reading cached X30 mapping: ", iha,isa,ihb
         ipass=psTmp%ipass
         if(ipass.eq.1)then
            call overwrite_kin(psTmp%kin)
         endif
         if(ifh1.eq.0)then
            ifh1=ih1
            ifs1=is
            ifh2=ih2
            jfh1=0
            jfs1=0
            jfh2=0
         else
            jfh1=ih1
            jfs1=is
            jfh2=ih2
         endif
      endif

      return
      end subroutine


**********************************************************************
* This subroutine stores the momentum configuration following mapping
* so that it doesn't need to be recalculated. This is to store NLO
* type mappings
*
      subroutine storepsetX30(npar,iha,isa,ihb,ipass)
      use KinData_mod
      use Observables_mod
      use pmap
      implicit real*8(a-h,o-y)
      implicit complex*16(z)
      logical pstoreflag
      common /currentmap/ ifh1,ifs1,ifh2,jfh1,jfs1,jfh2
!$omp threadprivate(/currentmap/)
      common/pstore/pstoreflag,ipsid
      type (store), pointer :: psTmp

      if(.not.pstoreflag)then ! do not store any momenta
         return
      endif

      if(iha.lt.ihb)then ! II mapping => exchange II
         ih1=iha
         ih2=ihb
      else
         ih1=ihb
         ih2=iha
      endif

      if((iha.le.2).and.(ihb.gt.2).and.(isa.gt.ihb))then ! IF mapping where isa > ihb => swap is and ihb
         is=ihb
         ih2=isa
      else
         is = isa
      endif


      if(ifh1.ne.0)then
         psTmp => psX30(ifh1,ifs1,ifh2,ih1,is,ih2)
         jfh1 = ih1
         jfs1 = is
         jfh2 = ih2
      else
         psTmp => psX30(ih1,is,ih2,0,0,0)
         ifh1 = ih1
         ifs1 = is
         ifh2 = ih2
         jfh1 = 0
         jfs1 = 0
         jfh2 = 0
      endif

      psTmp%exists = .true.
      psTmp%ipass = ipass
      if (ipass.eq.1) then  ! write arrays if the momenta pass the technical cut
         call copy_kin(npar,psTmp%kin)
      end if

      return
      end subroutine


**********************************************************************
* This subroutine stores the momentum configuration following mapping
* so that it doesn't need to be recalculated. This is to store NNLO
* type mappings
*
      subroutine storepsetX40(npar,iha,isa,isb,ihb,ipass)
      use KinData_mod
      use Observables_mod
      use pmap
      implicit real*8(a-h,o-y)
      implicit complex*16(z)
      logical pstoreflag
      common /currentmap/ ifh1,ifs1,ifh2,jfh1,jfs1,jfh2
!$omp threadprivate(/currentmap/)
      common/pstore/pstoreflag,ipsid
      type (store), pointer :: psTmp

      if(.not.pstoreflag)then ! do not store any momenta
         return
      endif


!     with a FF mapping the ordering of the unresolved partons matters
!     => must be treated appropriately
      if((iha.gt.2).and.(ihb.gt.2))then
         if(iha.lt.ihb)then
            ih1=iha
            ih2=ihb
            is1=isa
            is2=isb
         else
            ih1=ihb
            ih2=iha
            is1=isb
            is2=isa
         endif
      elseif((iha.le.2).and.(ihb.le.2))then !II
         if(iha.lt.ihb)then
            ih1=iha
            ih2=ihb
         else
            ih1=ihb
            ih2=iha
         endif
         if(isa.lt.isb)then
            is1=isa
            is2=isb
         else
            is1=isb
            is2=isa
         endif
      else !IF
         call sortval(isa,isb,ihb,is1,is2,ih2)
         ih1=iha
      endif

      psTmp => psX40(ih1,is1,is2,ih2)

      psTmp%exists=.true.
      psTmp%ipass=ipass
      if (ipass.eq.1) then
         call copy_kin(npar,psTmp%kin)
      end if

      ifh1=iha
      ifs1=isa
      ifh2=isb
      jfh1=ihb
      jfs1=0
      jfh2=0

      return
      end subroutine


