      !-------------------------------------------------------------------------------
      ! NNLOJET
      !-------------------------------------------------------------------------------
      !
      ! MODULE: Vector Boson Higgs Production speficit functions
      !
      ! DESCRIPTION:
      !> Module containing all subroutines for computing Vector Boson Fusion
      !>    Matrix Elements
      !>    i1 --->--- i4
      !>    i2 --->--- i3
      !>    i5, i6 : radiated gluons / q/qb final pair
      !-------------------------------------------------------------------------------

      !-------------------------------------------------------------------------------
      module commonVBF_mod
      !-------------------------------------------------------------------------------
         implicit none
         public  :: propagator_vbf, c_propagator_vbf
         public  :: coupling_vbf

         private

         integer, parameter :: dp = kind(1.0d0)

         real(dp) :: ele, ere, sin2w, enQ, enL, enR
         common /LR/ ele,ere,sin2w,enQ(-5:5),enL(5),enR(5)
         real(dp) :: emz, ezwidth
         common /Zmass/emz, ezwidth
         real(dp) :: emw, ewwidth
         common /Wmass/emw, ewwidth
         real(dp) :: amzW, stw, ewfac
         common /eweakW/ amzW,stw,ewfac
!$omp threadprivate(/eweakW/)

      contains
         !> Wrapper for couplings
         subroutine coupling_vbf(n1, n2, iboson, glr)
            integer, intent(in) :: n1, n2, iboson
            real(dp), dimension(4), intent(out) :: glr(4)

            glr(:) = 0d0

            select case(iboson)
            case(0)
               glr(1) = enR(n1)*enR(n2) ! RR
               glr(2) = enL(n1)*enR(n2) ! LR
               glr(3) = enR(n1)*enL(n2) ! RL
               glr(4) = enL(n1)*enL(n2) ! LL
            case(1)
               glr(4) = 1d0
            case(2)
               !> Mix of Z/W
               glr(4) = enL(n1)*enL(n2) ! LL
            end select

            return
         end subroutine

         !> Wrapper to obtain the square of the
         !> VBF-H Matrix element x Propagators (legacy)
         subroutine propagator_vbf(s1, s2, iboson, prop)
            real(dp), intent(in) :: s1, s2
            integer, intent(in) :: iboson
            real(dp), intent(out) :: prop
            real(dp) :: prop1, prop2, vbf_amp

            select case(iboson)
            case(0)
               prop1 = rZ_prop(s1)
               prop2 = rZ_prop(s2)
            case(1)
               prop1 = rW_prop(s1)
               prop2 = rW_prop(s2)
            end select

            vbf_amp = VVH(iboson)
            prop = (vbf_amp)**2/prop1/prop2
         end subroutine

         !> Wrapper to obtain the
         !> VBF-H Matrix element x Propagators
         subroutine c_propagator_vbf(s1, s2, iboson, prop)
            real(dp), intent(in) :: s1, s2
            integer, intent(in) :: iboson
            complex(dp), intent(out) :: prop
            real(dp) :: vbf_amp
            complex(dp) :: prop1, prop2

            select case(iboson)
            case(0)
               prop1 = cZ_prop(s1)
               prop2 = cZ_prop(s2)
            case(1)
               prop1 = cW_prop(s1)
               prop2 = cW_prop(s2)
            end select

            vbf_amp = VVH(iboson)
            prop = vbf_amp/(prop1*prop2)
         end subroutine

         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
         !! Private functions / suborutines
         !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

         !> Generate VVH amplitude
         pure real(dp) function VVH(iboson)
            integer, intent(in) :: iboson
            select case(iboson)
            case(0) ! ZZH
               VVH = 4d0*emz/sin2w
            case(1) ! WWH
               VVH = sqrt(emw**2/stw**3)
            end select
         end function

         !> Wrappers to generate propators for the Z and the W
         !> r: legacy, propagator already squared
         pure real(dp) function rW_prop(s)
            real(dp), intent(in) :: s
            rW_prop = propagator(s, emw, ewwidth)
         end function
         pure real(dp) function rZ_prop(s)
            real(dp), intent(in) :: s
            rZ_prop = propagator(s, emz, ezwidth)
         end function
         !> c: complex, amplitude level
         pure complex(dp) function cW_prop(s)
            real(dp), intent(in) :: s
            cW_prop = c_propagator(s, emw, ewwidth)
         end function
         pure complex(dp) function cZ_prop(s)
            real(dp), intent(in) :: s
            cZ_prop = c_propagator(s, emz, ezwidth)
         end function


         !> Generic propagators
         pure complex(dp) function c_propagator(s, rm, w)
            real(dp), intent(in) :: s, rm, w
            !> returns (s - mass**2), width*mass
            c_propagator = dcmplx((s-rm**2), w*rm)
            return
         end function

         pure real(dp) function propagator(s, rm, w)
            real(dp), intent(in) :: s, rm, w
            real(dp) :: t1, t2

            !> TODO (should be exactly the same)
!              propagator = abs(c_propagator(s, rm, w))**2

            t1 = (s - rm**2)**2
            t2 = (rm*w)**2
            propagator = t1 + t2

            return
         end function

      end module commonVBF_mod


