c$$$************************************************************
c$$$* Returns the leading colour matrix element for the process
c$$$*
c$$$* q(i1) + Qbar(i4) ->  + q`(i3) + Qb`(i2) + nu(i5) + e+(i6)
c$$$*
c$$$* where quarks i1 and i4 couple to the W boson
c$$$*
c$$$* The common block /C0gWFlav/ denotes the flavours of the partons.
c$$$* In FullC0g0W, this common block denotes the flavours of the initial state.
c$$$* Once this is passed onto the C0g0W/D0g0W function the flavours now denote
c$$$* the flavours of the partons coupling to the W boson.
c$$$*
c$$$* Note: This function assumes that the Vckm matrix has been diagonalised,
c$$$* as with MCFM. If the CKM matrix is non-diagonal then it is a far more
c$$$* complicated function. Given that this function will only be used for
c$$$* validation we will ignore these extra terms.
c$$$*
c$$$
c$$$      function FullC0g0W(i1,i2,i3,i4,i5,i6)
c$$$      use KinData_mod
c$$$      use C0gZ, only : zampC0g0Z
c$$$      implicit double precision (a-h,o-y)
c$$$      implicit complex*16 (z)
c$$$      parameter(pi=3.141592653589793238d0)
c$$$      common /koppel/ nf,nc,b0
c$$$      common /strong/ as
c$$$!$omp threadprivate(/strong/)
c$$$      common /eweakW/ amz,stw,ewfac
c$$$!$omp threadprivate(/eweakW/)
c$$$      common /C0gWFlav/ nf1a,nf2a
c$$$
c$$$      fac = 0.25d0/nc**2 ! colour averaging is always qqb initial state
c$$$
c$$$      fac = fac*(nc**2-1)*(4*pi*as)**2*(4d0*pi*amz/stw)**2
c$$$
c$$$      FullC0g0W = 0d0
c$$$
c$$$      if((i1.le.2).and.(i2.le.2).and.(nf1a.ne.nf2a))then ! qqb initial state
c$$$! non-identical flavour initial state => could couple to the W boson
c$$$      FullC0g0W = C0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .           +C0g0W(i2,i1,i4,i3,i6,i5)
c$$$
c$$$
c$$$       FullC0g0W = FullC0g0W+nf*C0g0W(i1,i4,i3,i2,i5,i6)
c$$$     .            -(1d0/nc)*(D0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .                     +D0g0W(i2,i1,i4,i3,i6,i5))
c$$$
c$$$      elseif((i1.le.2).and.(i2.le.2).and.(nf1a.eq.nf2a))then ! qqb initial state
c$$$! identical flavour initial state => cannot both couple to the W boson.
c$$$
c$$$         initialFlav = nf1a
c$$$         nf1a = 2
c$$$         nf2a = 1
c$$$         FullC0g0W = C0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .              +2d0*C0g0W(i3,i2,i1,i4,i5,i6)
c$$$     .              -(1d0/nc)*D0g0W(i4,i3,i2,i1,i6,i5)
c$$$
c$$$!     Reset flavours
c$$$         nf1a = initialFlav
c$$$         nf2a = initialFlav
c$$$
c$$$      elseif((i3.le.2).and.(i4.le.2).and.(nf1a.eq.nf2a))then ! qqb initial state
c$$$
c$$$         initialFlav = nf1a
c$$$         nf1a = 2
c$$$         nf2a = 1
c$$$
c$$$         FullC0g0W=C0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .            +2d0*C0g0W(i1,i4,i3,i2,i5,i6)
c$$$     .            -(1d0/nc)*D0g0W(i1,i2,i3,i4,i5,i6)
c$$$
c$$$         nf1a = initialFlav
c$$$         nf2a = initialFlav
c$$$
c$$$      elseif((i1.le.2).and.(i3.le.2).and.(nf1a.eq.nf2a))then ! qq initial state
c$$$
c$$$         initialFlav = nf1a
c$$$         nf1a = 2
c$$$         nf2a = 1
c$$$         FullC0g0W=C0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .            +C0g0W(i3,i2,i1,i4,i5,i6)
c$$$     .            -(1d0/nc)*D0g0W(i4,i1,i2,i3,i6,i5)
c$$$         nf1a = initialFlav
c$$$         nf2a = initialFlav
c$$$
c$$$      elseif((i1.le.2).and.(i3.le.2).and.(nf1a.ne.nf2a))then ! qQ initial state
c$$$
c$$$         fac = 0.5d0*fac ! symmetry factor
c$$$
c$$$         FullC0g0W=C0g0W(i1,i2,i3,i4,i5,i6)
c$$$     .            +C0g0W(i1,i4,i3,i2,i5,i6)
c$$$     .            -(1d0/nc)*D0g0W(i1,i2,i3,i4,i5,i6)
c$$$
c$$$      endif
c$$$
c$$$
c$$$      FullC0g0W = fac*FullC0g0W
c$$$
c$$$      return
c$$$      end
c$$$
****************************************************
* Returns the leading colour matrix element for the process
*
* q(i1) + Qbar(i4) ->  + q`(i3) + Qb`(i2) + nu(i5) + e+(i6)
*
* Ignoring overall factors
*
* The W boson couples to the i1 and i4 pair of quarks
*

      function C0g0W(i1,i2,i3,i4,i5,i6)
      use KinData_mod
      use C0gZ, only : zampC0g0Z
      implicit double precision (a-h,o-y)
      implicit complex*16 (z)
      common /eweakZ/ amz,zewfac1(4),zewfac2(4)
!$omp threadprivate(/eweakZ/)
      common /CZFlav/nf1,nf2
!$omp threadprivate(/CZFlav/)

      C0g0W = 0d0

      call eweakV2qq(nf1,nf2,s6(i5,i6))

      C0g0W =+abs(zampC0g0Z(i4,i3,i2,i1,i6,i5))**2
     .       +abs(zampC0g0Z(i4,i2,i3,i1,i6,i5))**2


      C0g0W = zewfac1(1)**2*C0g0W

      return
      end



**********************************************************************
