#include <cstddef>
#include <fstream>
#include <iostream>
#include <sstream>
#include <string>
#include <cstring>
#include <pineappl_capi.h>


inline bool file_exists(const std::string& name) {
    std::ifstream ifs(name.c_str());
    return ifs.good();
}


int main(int argc, char const *argv[]) {
  if (argc < 3) {
    std::cerr << "wrong number of arguments!\n\n";
    std::cerr << "\tUsage:  " << argv[0] << " <weight_file> <out_file> [-v] [--skip] [--noopt]\n";
    return EXIT_FAILURE;
  }
  //> defaults
  bool verbose = false;
  bool skip_invalid = false;
  bool optimize = true;
  if (argc > 3) {
    for (int i = 3; i < argc; ++i) {
      if (strcmp(argv[i], "-v")      == 0)  verbose = true;
      if (strcmp(argv[i], "--skip")  == 0)  skip_invalid = true;
      if (strcmp(argv[i], "--noopt") == 0)  optimize = false;
    }
  }

  struct pineappl_grid *grid_accum = nullptr;
  size_t nbins = 0;

  if (!file_exists(argv[1])) {
    std::cerr << "weight file: " << argv[1] << " missing?" << std::endl;
    return EXIT_FAILURE;
  }
  if (verbose) std::cout << "loading weights from " << argv[1] << std::endl;
  std::ifstream weights_file(argv[1]);
  std::string line;
  while (std::getline(weights_file, line)) {
    if (line.rfind("#", 0) == 0) continue;  // comment lines
    std::istringstream iline(line);

    //> read in the grid
    std::string ifile;
    iline >> std::ws >> ifile;
    ifile.erase(ifile.rfind(".dat")).append(".pineappl.lz4");
    if (!file_exists(ifile)) {
      std::cerr << "grid file: " << ifile << " missing?" << std::endl;
        if (skip_invalid) {
          continue;
        } else {
          if (grid_accum) pineappl_grid_delete(grid_accum);
          return EXIT_FAILURE;
        }
    }
    if (verbose) std::cout << "reading grid " << ifile;
    struct pineappl_grid *igrid = pineappl_grid_read(ifile.c_str());

    //> check bin size compatibility
    if (nbins == 0) {
      nbins = pineappl_grid_bin_count(igrid);
    } else {
      if (nbins != pineappl_grid_bin_count(igrid)) {
        std::cerr << "grid file: " << ifile << ":" << std::endl;
        std::cerr << "bin size mismatch: " << nbins
                  << " != " << pineappl_grid_bin_count(igrid) << std::endl;
        pineappl_grid_delete(igrid);
        if (skip_invalid) {
          continue;
        } else {
          if (grid_accum) pineappl_grid_delete(grid_accum);
          return EXIT_FAILURE;
        }
      }
    }

    //> read in the weigths
    double weights[nbins];
    for (int i = 0; i < nbins; ++i) {
      iline >> std::ws >> weights[i];
    }
    if (verbose) std::cout << " weights[" << nbins << "] = [ "
      << weights[0] << " ... " << weights[nbins-1] << " ]" << std::endl;
    pineappl_grid_scale_by_bin(igrid, nbins, weights);

    //> accumulate
    if (grid_accum == nullptr) {
      grid_accum = igrid;
      igrid = nullptr;
    } else {
      pineappl_grid_merge_and_delete(grid_accum, igrid);
    }

  }  // end while [line]

  if (grid_accum == nullptr) {
    std::cerr << "no grid read in?!" << std::endl;
    return EXIT_FAILURE;
  }

  if (optimize) pineappl_grid_optimize(grid_accum);
  if (verbose) std::cout << "writing out " << argv[2] << "..." << std::endl;
  pineappl_grid_write(grid_accum, argv[2]);
  pineappl_grid_delete(grid_accum);
  return EXIT_SUCCESS;
}
